(* ::Package:: *)

(* This package is loaded in context "Theorema`Core`Trusted`Private`". *)

$ContextPath = Join[ $ContextPath, {"Theorema`Language`", "Theorema`Common`"}]	(* for symbols such as 'TVAR$', 'FORM$TM', 'STRING$', etc. *)

With[ {lang = "English"},

(* ::Section:: *)
(*Reasoning*)

coreMessageToList[ 1, {method_String}, _, lang] :=
	{"Proof method", STRING$[ method], "is invalid."};
coreMessageToList[ 2, _, _, lang] = {"The branch cannot be closed."};
coreMessageToList[ 3, {thm_}, _, lang] :=
	{"Theorem/formula", ToString[ List @@ thm], "does not exist."};
coreMessageToList[ 4, {thm_}, _, lang] :=
	{"No subgoal available for adding theorem/formula", ToString[ List @@ thm], "."};
coreMessageToList[ 5, {method_String}, _, lang] :=
	{"Inference", STRING$[ method], "is invalid."};
coreMessageToList[ 6, {param_}, _, lang] :=
	{"\"forget\" applied with illegal parameter:", param, "."};
coreMessageToList[ 7, {method_String, params___}, _, lang] :=
	Join[ {"Inference", STRING$[ method], ", with parameters"}, Riffle[ {params}, ","], {", is not applicable."}];
coreMessageToList[ 8, {method_String, params___, out_String}, _, lang] :=
	Join[ {"Inference", STRING$[ method], ", with parameters"}, Riffle[ {params}, ","], {", returned", STRING$[ out], "."}];
coreMessageToList[ 9, _, _, lang] = {"No subgoal to skip."};
coreMessageToList[ 10, _, _, lang] = {"Proof method is not applicable."};
coreMessageToList[ 11, _, _, lang] = {"Proof goal is ill-typed."};
coreMessageToList[ 12, _, _, lang] = {"Proof goal is invalid."};
coreMessageToList[ 13, _, _, lang] = {"Proof goal is no FML$-structure."};
coreMessageToList[ 14, {fml_, out_}, _, lang] :=
	{"Formula", fml, "cannot be added to the hypotheses:", out};
coreMessageToList[ 14, {fml_}, _, lang] :=
	{"Formula", fml, "cannot be added to the hypotheses."};
coreMessageToList[ 15, {n_Integer}, _, lang] :=
	{"+", n, "further messages."};
coreMessageToList[ 16, {src_String}, _, lang] :=
	{"Theory-source \"", src, "\" of proof goal differs from current theory."};
	

(* ::Section:: *)
(*Types*)

(*illFormedInvalidSym*)
coreMessageToList[ 101, {sym_}, _, lang] :=
	{"Symbol", TERMDISP$[ sym], "appears at invalid position."};
(*illFormedInvalidType*)
coreMessageToList[ 102, {typ_}, _, lang] :=
	{"Type", TYPEDISP$[ typ], "is invalid."};
(*illFormedInvalidAtom*)
coreMessageToList[ 103, {atm_}, _, lang] :=
	{"Atom", TERMDISP$[ atm], "is invalid."};
(*illFormedUnbound*)
coreMessageToList[ 104, {var_}, _, lang] :=
	{"Variable", TERMDISP$[ var], "is unbound."};
(*illTypedApplication*)
coreMessageToList[ 105, {typ1_, typ2_}, _, lang] :=
	{"Operator-type", TYPEDISP$[ typ1], "and operand-type", TYPEDISP$[ typ2], "do not match."};
(*illTypedTypeClash*)
coreMessageToList[ 106, {typ1_, typ2_}, _, lang] :=
	{"Inferred type", TYPEDISP$[ typ1], "cannot be unified with annotated type", TYPEDISP$[ typ2], "."};
(*illTypedInvalidInst*)
coreMessageToList[ 107, {typ_}, _, lang] :=
	{"Type-instance", TYPEDISP$[ typ], "is invalid."};
coreMessageToList[ 108, {var_}, _, lang] :=
	{"Illegal name of bound variable:", var};
coreMessageToList[ 109, {const_}, _, lang] :=
	{"Constant", TERMDISP$[ const], "is not declared."};
coreMessageToList[ 110, {expr_, type_List}, _, lang] :=
	{"Type-list", TYPEDISP$ /@ type, "does not yield a valid type instance of term", TERMDISP$[ expr], "."};
coreMessageToList[ 111, {expr_, type_}, _, lang] :=
	{"Type", TYPEDISP$[ type], "cannot be unified with declared type of constant", TERMDISP$[ expr], "."};
coreMessageToList[ 112, {const_}, _, lang] :=
	{"Constant", TERMDISP$[ const], "appears without type annotation."};


(* ::Section:: *)
(*Tools/IO, Tools/Toplevel and Tools/Reasoning*)

coreMessageToList[ 201, _, _, lang] = {"Pattern is ill-typed."};
coreMessageToList[ 202, {logic_String}, _, lang] :=
	{"Object-logic", STRING$[ logic], "not found."};
coreMessageToList[ 203, {k_List, lbl_String}, _, lang] :=
	{"Cannot add formula", FKEYDISP$[ k, lbl]};
coreMessageToList[ 203, _, _, lang] = {"Cannot add formula(s)."};
coreMessageToList[ 204, {tj_}, _, lang] :=
	{"Cannot declare truth judgment", TERMDISP$[ tj], "; use \"judgment\" instead."};
coreMessageToList[ 205, consts_List, _, lang] :=
	Join[ {"Cannot declare constants without type annotations:"}, Riffle[ TERMDISP$ /@ consts, ","]];
coreMessageToList[ 206, {lbl_String}, _, lang] :=
	{"Label", STRING$[ lbl], "is not unique."};
coreMessageToList[ 207, {tj_}, _, lang] :=
	{"Cannot unset truth judgment", TERMDISP$[ tj]};
coreMessageToList[ 208, {term_}, _, lang] :=
	{"Term", TERMDISP$[ term], "is ill-typed."};
coreMessageToList[ 209, _, _, lang] = {"Cannot parse proof notebook."};
coreMessageToList[ 210, {tj_, term_}, _, lang] :=
	{"Truth judgment", TERMDISP$[ tj], "appears in term", TERMDISP$[ term]};
coreMessageToList[ 211, {tj_}, _, lang] :=
	{"Cannot undo declaration of truth judgment", TERMDISP$[ tj], "; use \"undo judgment\" instead."};
coreMessageToList[ 212, {typ_}, _, lang] :=
	{"Type", TYPEDISP$[ typ], "is unknown."};
coreMessageToList[ 213, old_List, _, lang] :=
	Join[ {"Do you want to overwrite the declarations/definitions of type(s)"}, Riffle[ TYPEDISP$ /@ old, ","], {"?"}];
coreMessageToList[ 214, {typ_}, _, lang] :=
	{"Predicate in definition of type ", TYPEDISP$[ typ], "is ill-typed."};
coreMessageToList[ 215, {typ_}, _, lang] :=
	{"Cannot add some theorems related to the definition of type", TYPEDISP$[ typ]};
coreMessageToList[ 216, {const_, typ_}, _, lang] :=
	{"Do you want to overwrite the declaration", TTDISP$[ const, typ], "?"};
coreMessageToList[ 217, _, _, lang] = {"Do you want to terminate the current Core session?"};
coreMessageToList[ 218, _, _, lang] = {"\"facts\" cannot be undone."};
coreMessageToList[ 219, {lbl_}, _, lang] :=
	{"Ignoring label", STRING$[ lbl], "in the presence of an explicit non-emptiness proof."};
coreMessageToList[ 220, consts_List, _, lang] :=
	Join[ {"Cannot lock/unlock typed constants:"}, Riffle[ TTDISP$ @@@ consts, ","]];
coreMessageToList[ 221, {typ_, lbl_}, _, lang] :=
	{"Non-emptiness formula", STRING$[ lbl], "for definition of type", TYPEDISP$[ typ], "does not exist."};
coreMessageToList[ 222, _, _, lang] = {"\"import\" cannot be undone."};
coreMessageToList[ 223, {t_}, _, lang] :=
	{"Theory specification", t, "is no string."};
coreMessageToList[ 224, {expr_}, _, lang] :=
	{"Term", TERMDISP$[ expr], "is no constant."};
coreMessageToList[ 225, {c_}, _, lang] :=
	{TERMDISP$[ c], "is locked or protected."};
coreMessageToList[ 226, {c_}, _, lang] :=
	{TERMDISP$[ c], "is declared in the theory context."};
coreMessageToList[ 227, {l_, r_}, _, lang] :=
	{"The equality between", TERMDISP$[ l], "and", TERMDISP$[ r], "is ill-typed or contains a.b.f. constants or meta variables."};
coreMessageToList[ 228, {c_}, _, lang] :=
	{TERMDISP$[ c], "appears on the right-hand-side of the abbreviation."};
coreMessageToList[ 229, {c_}, _, lang] :=
	{"Left-hand-side", TERMDISP$[ c], "of the abbreviation lacks some type variables that appear on the right-hand-side."};
coreMessageToList[ 230, {c_}, _, lang] :=
	{"Unable to generate pretty-printing rule for", TERMDISP$[ c], ", since there are additional type variables on the left-hand-side of the abbreviation."};
coreMessageToList[ 231, {c_}, _, lang] :=
	{TERMDISP$[ c], "is declared as an abbreviation."};
coreMessageToList[ 232, {c_, keys_List}, _, lang] :=
	Join[ {"Overriding existing abbreviation(s)"}, Riffle[ KEYDISP$ /@ keys, ","], {"for", TERMDISP$[ c]}];
coreMessageToList[ 233, {c_, typ_}, _, lang] :=
	{"In abbreviation for", TERMDISP$[ c], ": Type", TYPEDISP$[ typ], "on the right-hand-side is not declared/defined."};
coreMessageToList[ 234, {c_, const_}, _, lang] :=
	{"In abbreviation for", TERMDISP$[ c], ": Constant", TERMDISP$[ const], "on the right-hand-side is not declared."};
coreMessageToList[ 235, {expr_}, _, lang] :=
	{"The head of left-hand-side", TERMDISP$[ expr], "in the abbreviation is no symbol."};
coreMessageToList[ 236, {typ_}, _, lang] :=
	{TYPEDISP$[ typ], "is declared as an abbreviation."};
coreMessageToList[ 237, {kind_String}, _, lang] :=
	{"Notation", STRING$[ kind], "does not affect input."};
coreMessageToList[ 238, {kind_String}, _, lang] :=
	{"Notation", STRING$[ kind], "does not affect output."};
coreMessageToList[ 239, _, _, lang] = {"Notations may not involve named type variables, named a.b.f. type constants and named syntactic variables."};
coreMessageToList[ 240, {kind_String}, _, lang] :=
	{"In notation", STRING$[ kind], ": ignoring terms with type-annotated compound subterms for pretty-printing."};
coreMessageToList[ 241, {kind_String, fun_}, _, lang] :=
	{"In notation", STRING$[ kind], ":", TERMDISP$[ fun], "must not involve syntactic variables."};
coreMessageToList[ 242, {kind_String}, _, lang] :=
	{"In notation", STRING$[ kind], ": ignoring terms involving syntactic variables for printing."};
coreMessageToList[ 243, {kind_String, constr_}, _, lang] :=
	{"In notation", STRING$[ kind], ": cannot create pretty-printing rules, because term", TERMDISP$[ constr], "contains type-annotated compound subterms."};
coreMessageToList[ 244, {kind_String}, _, lang] :=
	{"Notation", STRING$[ kind], "does not work for types."};
coreMessageToList[ 245, {noSym_}, _, lang] :=
	{"Object", noSym, "cannot be exported, because it is no symbol."};
coreMessageToList[ 245, noSyms_List, _, lang] :=
	Join[ {"Objects"}, Riffle[ noSyms, ","], {"cannot be exported, because they are no symbols."}];
coreMessageToList[ 246, protected_List, _, lang] :=
	Join[ {"Read-protected symbol(s)"}, Riffle[ protected, ","], {"cannot be exported."}];
coreMessageToList[ 247, {file_}, _, lang] :=
	{"Files imported into a theory must have file extension \"nb\"; this is violated by", STRING$[ file]};
coreMessageToList[ 248, _, _, lang] = {"Changing the internal default sort is not possible."};
coreMessageToList[ 249, {cat_}, _, lang] :=
	{"Rule category", cat, "is unknown. Possible categories are \"autored\", \"elim\", \"intro\", \"sym\", \"trans\" and \"simp\"."};
coreMessageToList[ 250, _, _, lang] = {"Fact collections must not contain instances of theorem schemas."};
coreMessageToList[ 251, _, _, lang] = {"Fact collections must not contain formula variants."};
coreMessageToList[ 252, {rule_}, _, lang] :=
	{"Formula not found:", rule};
coreMessageToList[ 253, {rule_}, _, lang] :=
	{"Formula comes after current input cell:", rule};
coreMessageToList[ 254, {fact_}, _, lang] :=
	{"Literal fact", fact, "is ill-typed."};
coreMessageToList[ 255, {coll_String}, _, lang] :=
	{"Fact collection", STRING$[ coll], "does not exist or is not in scope."};
coreMessageToList[ 256, _, _, lang] = {"Names of fact collections must be strings."};
coreMessageToList[ 257, _, _, lang] = {"Names of fact collections must not contain \"::\"."};
coreMessageToList[ 258, dups_List, _, lang] :=
	Join[
		{"Names of fact collections are not unique:"},
		Riffle[ FKEYDISP$ @@@ dups, ","],
		{"You can refer to the respective collections by prefixing their names with \"\[LeftAngleBracket]source\[RightAngleBracket]::\"."}
	];
coreMessageToList[ 259, dups_List, _, lang] :=
	Join[ {"Duplicate fact collections in the same theory:"}, Riffle[ FKEYDISP$ @@@ dups, ","]];
coreMessageToList[ 260, _, _, lang] = {"Facts can be added to and removed from automatic fact collections only in the theory they originate from."};
coreMessageToList[ 261, _, _, lang] = {"The meta-data of formulas and theorem schemas can only be changed in the theory they originate from."};
coreMessageToList[ 262, {file_String}, _, lang] :=
	{"Do you want to create an image of theory", STRING$[ file], "?\n(This might take a few minutes.)"};
coreMessageToList[ 263, _, _, lang] = {"Exporting the image failed."};
coreMessageToList[ 264, {file_String, log_String}, _, lang] :=
	{"Building theory", STRING$[ file], "failed. Consult", STRING$[ log], "for details."};
coreMessageToList[ 265, _, _, lang] = {"Cannot instantiate by ill-typed term."};
coreMessageToList[ 266, _, _, lang] = {"More instantiations than variables."};
coreMessageToList[ 267, _, _, lang] = {"Patterns do not match term."};
coreMessageToList[ 268, _, _, lang] = {"Patterns are ill-formed."};
coreMessageToList[ 269, uinst_, _, lang] :=
	Join[ {"Obtained matcher does not instantiate syntactic variables"}, Riffle[ Composition[ SYMDISP$, First] /@ uinst, ","]];
coreMessageToList[ 270, {META$[ x_, _]}, _, lang] :=
	{"Instance of syntactic variable", SYMDISP$[ x], "depends on other syntactic variables and/or meta variables."};


(* ::Section:: *)
(*Tools/Syntax and Tools/Interactive*)

coreMessageToList[ 301, {c_String}, _, lang] :=
	{STRING$[ c], "cannot be used as a top-level proof method."};
coreMessageToList[ 302, {}, _, lang] = {"Input cannot be parsed."};
coreMessageToList[ 303, {}, _, lang] = {"Notebook has not been saved."};
coreMessageToList[ 304, {}, _, lang] = {"Formula is not in the current theory context."};
coreMessageToList[ 305, {}, _, lang] = {"Evaluation notebook or -cell cannot be obtained."};
coreMessageToList[ 351, {h_}, _, lang] :=
	{"Expression", h, "appears at invalid position."};
coreMessageToList[ 352, _, _, lang] = {"Type contains illegal sort constraint."};
coreMessageToList[ 353, {v_}, _, lang] :=
	{"Variable", v, "is unbound."};
coreMessageToList[ 354, {e_}, _, lang] :=
	{"Term", e, "is ill-formed."};
coreMessageToList[ 355, {r_}, _, lang] :=
	{"Multi-range", r, "contains non-simple variable ranges."};
coreMessageToList[ 356, {r_}, _, lang] :=
	{"Cannot extract condition from variable range", r};
coreMessageToList[ 357, _, _, lang] = {"Binder with empty range has non-trivial conditions."};
coreMessageToList[ 358, dup_List, _, lang] :=
	Join[ {"Inconsistent sort constraints for"}, Riffle[ TYPEDISP$ /@ dup, ","]];
coreMessageToList[ 359, {typ_}, _, lang] :=
	{"Cannot add sort constraint to non-variable type", TYPEDISP$[ typ]};
coreMessageToList[ 360, _, _, lang] = {"Pattern-expressions (\"is\") are not allowed in the formula list."};
coreMessageToList[ 361, {s_String}, _, lang] :=
	{"Pattern-indicators (\"is\") must not be followed by", STRING$[ s]};
coreMessageToList[ 362, {s_String}, _, lang] :=
	{"Pattern-indicators (\"is\") must not be preceded by", STRING$[ s]};
coreMessageToList[ 363, {tags_List}, _, lang] :=
	{"List", tags, "is no well-formed list of formula tags."};
coreMessageToList[ 364, {k_String}, _, lang] :=
	{STRING$[ k], "is no valid option for type definitions."};
coreMessageToList[ 365, {s_String}, _, lang] :=
	{"Theory commands that manipulate the logical context, in particular", STRING$[ s], ", should be the first commands in cells."};


(* ::Section:: *)
(*Theorems*)

coreMessageToList[ 401, {thm_}, _, lang] :=
	Join[ {"Theorem", STRING$[ List @@ thm], "is ill-typed."}];


(* ::Section:: *)
(*Consistency*)

coreMessageToList[ 1001, {fun_String, args___}, _, lang] :=
	Join[ {"Function", STRING$[ fun], "called with invalid arguments:"}, Riffle[ {args}, ","]];
coreMessageToList[ 1002, {k_, l_}, _, lang] :=
	{"Formula", FKEYDISP$[ k, l], "is no definition."};
coreMessageToList[ 1003, {k_}, _, lang] :=
	{"Removing formulas from the theory is not permitted:", FKEYDISP$[ k]};
coreMessageToList[ 1004, {k_}, _, lang] :=
	{"Formula", FKEYDISP$[ k], "does not exist."};
coreMessageToList[ 1005, {k_}, _, lang] :=
	{"Replacing formulas in the theory is not permitted:", FKEYDISP$[ k]};
coreMessageToList[ 1006, {k_}, _, lang] :=
	{"Changing the status of formulas is not permitted:", FKEYDISP$[ k]};
coreMessageToList[ 1007, {const_}, _, lang] :=
	{"Re-declaring constants is not permitted:", TERMDISP$[ const]};
coreMessageToList[ 1008, {k_, stat_}, _, lang] :=
	{"Changing the status of formula", FKEYDISP$[ k], "to", stat, "is not permitted."};
coreMessageToList[ 1009, {const_}, _, lang] :=
	{"Declaring locked constants is not permitted:", TERMDISP$[ const]};
coreMessageToList[ 1010, {k_, l_}, _, lang] :=
	{"Formula", FKEYDISP$[ k, l], "is ill-typed."};
coreMessageToList[ 1011, {spec_}, _, lang] :=
	{"Replacing theorem schemas in the theory is not permitted:", spec};
coreMessageToList[ 1012, {}, _, lang] :=
	{"The declared type of constant", TERMDISP$[ TrueQ$TM], "must be of the form", TYPEDISP$[ FUN$TM[ Blank[], FORM$TM]],
		", with the argument type being free of", TYPEDISP$[ FORM$TM], "and type variables."};
coreMessageToList[ 1013, {spec_}, _, lang] :=
	{"Removing theorem schemas from the theory is not permitted:", spec};
coreMessageToList[ 1014, {spec_}, _, lang] :=
	{"Theorem schema", spec, "does not exist."};
coreMessageToList[ 1015, {const_, typ_}, _, lang] :=
	constDefMessage[ const, typ, {"defining locked constants is not permitted."}, lang];
coreMessageToList[ 1016, {typ_}, _, lang] :=
	{"Type", TYPEDISP$[ typ], "is not a type constructor applied to distinct type variables. (type overloading is not enabled)"};
coreMessageToList[ 1017, {fun_String, args___}, _, lang] :=
	Join[ {"Theory action", STRING$[ fun], "called with invalid arguments:"}, Riffle[ {args}, ","]];
coreMessageToList[ 1018, {fun_String}, _, lang] :=
	{"Theory action", STRING$[ fun], "called during a proof."};
coreMessageToList[ 1019, {head_}, _, lang] :=
	{"Declaring/defining protected type constructors is not permitted:", TERMDISP$[ head]};
coreMessageToList[ 1020, {typ_}, _, lang] :=
	{"Replacing type declarations/definitions is not permitted:", TYPEDISP$[ typ]};
coreMessageToList[ 1021, {typ_}, _, lang] :=
	{"Type", TYPEDISP$[ typ], "is not declared/defined."};
coreMessageToList[ 1022, {typ_, rt_}, _, lang] :=
	typeDefMessage[ typ, {"representation type", TYPEDSIP$[ rt], "must not contain", TYPEDSIP$[ FORM$TM]}, lang];
coreMessageToList[ 1023, {const_}, _, lang] :=
	{"Unlocking constants is not permitted:", TERMDISP$[ const]};
coreMessageToList[ 1024, {const_}, _, lang] :=
	{"Unlocking protected constants is not permitted:", TERMDISP$[ const]};
coreMessageToList[ 1025, {const_}, _, lang] :=
	{"Constant", TERMDISP$[ const], "is not locked."};
coreMessageToList[ 1026, {const_}, _, lang] :=
	{"Removing declarations of constants is not permitted:", TERMDISP$[ const]};
coreMessageToList[ 1027, {const_}, _, lang] :=
	{"Removing declarations of locked constants is not permitted:", TERMDISP$[ const]};
coreMessageToList[ 1028, {const_}, _, lang] :=
	{"Constant", TERMDISP$[ const], "is not declared."};
coreMessageToList[ 1028, consts:{_, __}, _, lang] :=
	Join[ {"Constants"}, Riffle[ TERMDISP$ /@ consts, ","], {"are not declared."}];
coreMessageToList[ 1029, {typ_, morph_}, _, lang] :=
	typeDefMessage[ typ, {"type morphism", TERMDISP$[ morph], "has been declared already."}, lang];
coreMessageToList[ 1030, {typ_, v_TFIX$}, _, lang] :=
	{"In declaration/definition of type", TYPEDISP$[ typ], ": a.b.f. type constant", TYPEDISP$[ v], "appears."};
coreMessageToList[ 1031, {typ_, t0_}, _, lang] :=
	{"In declaration/definition of type", TYPEDISP$[ typ], ": type", TYPEDISP$[ t0], "is unknown."};
coreMessageToList[ 1032, {typ_, v_TFIX$}, _, lang] :=
	typeDefMessage[ typ, {"a.b.f. type constant", TYPEDISP$[ v], "appears in defining predicate."}, lang];
coreMessageToList[ 1032, {typ_, _}, _, lang] :=
	typeDefMessage[ typ, {"type variable(s) in defining predicate do not occur in new type."}, lang];	(* it does not make sense to name variable *)
coreMessageToList[ 1033, {typ_, t0_}, _, lang] :=
	typeDefMessage[ typ, {"type", TYPEDISP$[ t0], "is unknown."}, lang];
coreMessageToList[ 1034, {const_, typ_, v_TVAR$}, _, lang] :=
	constDefMessage[ const, typ, {"Type variable", TYPEDISP$[ v], "appears on right-hand-side of definition, but not on left-hand-side."}, lang];
coreMessageToList[ 1034, {const_, typ_, v_TFIX$}, _, lang] :=
	constDefMessage[ const, typ, {"a.b.f. type constant", TYPEDISP$[ v], "appears on right-hand-side of definition."}, lang];
coreMessageToList[ 1035, {const_, typ1_, typ2_}, _, lang] :=
	constDefMessage[ const, typ1, {"type", TYPEDISP$[ typ2], "is unknown."}, lang];
coreMessageToList[ 1036, {const_, v_}, _, lang] :=
	{"Declaration of constant", TERMDISP$[ const], "involves illegal type variable/constant", TYPEDISP$[ v]};
coreMessageToList[ 1037, {const_, typ_}, _, lang] :=
	{"In declaration of", TERMDISP$[ const], ", type", TYPEDISP$[ typ], "is unknown."};
coreMessageToList[ 1038, {typ_}, _, lang] :=
	typeDefMessage[ typ, {"non-emptiness proof failed."}, lang];
coreMessageToList[ 1039, {typ_}, _, lang] :=
	typeDefMessage[ typ, {"axiom stating characteristic properties of type morphisms exists already."}, lang];
coreMessageToList[ 1040, {typ_}, _, lang] :=
	typeDefMessage[ typ, {"predicate does not type-check."}, lang];
coreMessageToList[ 1041, {typ_, morph_}, _, lang] :=
	typeDefMessage[ typ, {"type morphism", TERMDISP$[ morph], "is locked."}, lang];
coreMessageToList[ 1042, {typ_}, _, lang] :=
	typeDefMessage[ typ, {"type morphisms must be distinct."}, lang];
coreMessageToList[ 1043, {const_}, _, lang] :=
	{"Locking defined constants is not permitted:", TERMDISP$[ const]};
coreMessageToList[ 1044, {const_}, _, lang] :=
	{"Removing constant definitions is not permitted:", TERMDISP$[ const]};
coreMessageToList[ 1045, {typ_}, _, lang] :=
	{"Removing declarations/definitions of types not permitted:", TYPEDISP$[ typ]};
coreMessageToList[ 1046, {head_}, _, lang] :=
	{"Removing declarations/definitions of protected type constructors is not permitted:", TERMDISP$[ head]};
coreMessageToList[ 1047, {k_}, _, lang] :=
	{"Proof of formula", FKEYDISP$[ k], "is incomplete."};
coreMessageToList[ 1048, {k_}, _, lang] :=
	{"Proof of formula", FKEYDISP$[ k], "is circular."};
coreMessageToList[ 1049, {k_}, _, lang] :=
	{"Proof of formula", FKEYDISP$[ k], "is invalid."};
coreMessageToList[ 1050, {k_}, _, lang] :=
	{"Proving definitions or internal axioms is not permitted:", FKEYDISP$[ k]};
coreMessageToList[ 1051, {k_}, _, lang] :=
	{"Changing proofs of formulas is not permitted:", FKEYDISP$[ k]};
coreMessageToList[ 1052, _, _, lang] = {"Proof is invalid."};
coreMessageToList[ 1053, {typ_}, _, lang] :=
	{"Head of type", TYPEDISP$[ typ], "is invalid."};
coreMessageToList[ 1054, _, _, lang] = {"Disabling type definitions after types have been defined is not permitted."};
coreMessageToList[ 1055, _, _, lang] = {"Declaring truth judgment", TERMDISP$[ TrueQ$TM], "after types have been defined is not permitted."};
coreMessageToList[ 1056, {typ_}, _, lang] :=
	typeDefMessage[ typ, {"type definitions are not enabled."}, lang];
coreMessageToList[ 1057, _, _, lang] = {"Enabling type definitions again after disabling them is not permitted."};
coreMessageToList[ 1058, _, _, lang] = {"Theory dependency graph is cyclic."};
coreMessageToList[ 1059, {thy_String}, _, lang] :=
	{"Changing ancestors of theory \"", thy, "\" in imported theory image is not permitted."};
coreMessageToList[ 1060, _, _, lang] = {"Disabling type overloading after types have been overloaded is not permitted."};
coreMessageToList[ 1061, _, _, lang] = {"Removing ancestors of theories is not permitted."};
coreMessageToList[ 1062, _, _, lang] = {"Enabling type overloading again after disabling it is not permitted."};
(*UNUSED*)coreMessageToList[ 1063, _, _, lang] = {"Enabling type overloading again."};
coreMessageToList[ 1064, _, _, lang] = {"Disabling constant overloading after constants have been overloaded is not permitted."};
(*UNUSED*)coreMessageToList[ 1065, _, _, lang] = {"Disabling constant overloading after constants have been overloaded."};
coreMessageToList[ 1066, {const_, typ_}, _, lang] :=
	constDefMessage[ const, typ, {"overloaded constant definitions are not enabled."}, lang];
coreMessageToList[ 1067, _, _, lang] = {"Enabling constant overloading again after disabling it is not permitted."};
(*UNUSED*)coreMessageToList[ 1068, _, _, lang] = {"Enabling constant overloading again."};
coreMessageToList[ 1069, {const_, typ_, keys_List}, _, lang] :=
	constDefMessage[ const, typ, Join[ {"overlapping left-hand-sides are not permitted:"}, Riffle[ FKEYDISP$ /@ keys, ","]], lang];
coreMessageToList[ 1070, {const_, typ_, keys_List}, _, lang] :=
	constDefMessage[ const, typ, Join[ {"left-hand-side overlaps with left-hand-side(s) of"}, Riffle[ FKEYDISP$ /@ keys, ","]], lang];
coreMessageToList[ 1071, {const_, typ_, unif_}, _, lang] :=
	With[ {ft = getFullTypeNoSorts[ typ, HoldComplete[ const]]},
		constDefMessage[ const, ft, {"dependency relation for type instance", TYPEDISP$[ ft /. unif], "is not composable (and hence potentially cyclic)."}, lang]
	];
coreMessageToList[ 1072, {const_, typ_, unif_}, _, lang] :=
	With[ {ft = getFullTypeNoSorts[ typ, HoldComplete[ const]]},
		constDefMessage[ const, ft, {"dependency relation for type instance", TYPEDISP$[ ft /. unif], "is cyclic."}, lang]
	];
coreMessageToList[ 1073, {k_}, _, lang] :=
	{"The meta-data (label, tags, etc.) of formula", FKEYDISP$[ k], "can only be changed in the theory the formula originates from."};
coreMessageToList[ 1074, {k_}, _, lang] :=
	{"The label of theorem schema", THM$[ k], "can only be changed in the theory the theorem schema originates from."};
coreMessageToList[ 1075, l_List, _, lang] :=
	Join[ {"The following theorems are missing:"}, Riffle[ l, ","]];
coreMessageToList[ 1076, l_List, _, lang] :=
	Join[ {"The following theorems are ill-typed or have wrong type-variable positions:"}, Riffle[ l, ","]];
coreMessageToList[ 1077, l_List, _, lang] :=
	Join[ {"The following theorems have invalid proofs:"}, Riffle[ l, ","]];
coreMessageToList[ 1078, l_List, _, lang] :=
	Join[ {"The following theorems depend on theorems with invalid indices:"}, Riffle[ l, ","]];
coreMessageToList[ 1079, l_List, _, lang] :=
	Join[ {"The following theorems are unproved axioms:"}, Riffle[ l, ","]];
coreMessageToList[ 1080, _, _, lang] = {"Theory images must be exported to and imported from files with file extension \".mx\"."};
coreMessageToList[ 1081, _, _, lang] = {"Theory images can only be be exported in strict theory mode."};
coreMessageToList[ 1082, _, _, lang] = {"No theory scope set for exporting theory image."};
coreMessageToList[ 1083, _, _, lang] = {"Theory images cannot be exported with unproved theorem instances in the context."};
coreMessageToList[ 1084, _, _, lang] = {"Theory images must be imported into an empty context."};
coreMessageToList[ 1085, {file_String}, _, lang] :=
	{"File", STRING$[ file], "does not exist."};
coreMessageToList[ 1086, {file_String}, _, lang] :=
	{"Image-file", STRING$[ file], "has invalid format."};

typeDefMessage[ typ_, l_List, lang] :=
	Join[ {"In definition of type", TYPEDISP$[ typ], ":"}, l];

constDefMessage[ const_, l_List, lang] :=
	Join[ {"In definition of constant", TERMDISP$[ const], ":"}, l];
constDefMessage[ const_, typ_, l_List, lang] :=
	Join[ {"In definition of constant", TTDISP$[ const, getFullTypeNoSorts[ typ, HoldComplete[ const]]], ":"}, l];


(* ::Section:: *)
(*translate*)

Theorema`Common`translate[ "availKnowl", lang] = "Available Knowledge";
Theorema`Common`translate[ "axm", lang] = "axiom";
Theorema`Common`translate[ "axmTrusted", lang] = "trusted axiom";
Theorema`Common`translate[ "Calculation", lang] = "Calculation";
Theorema`Common`translate[ "chainedFacts", lang] = "Chained facts";
Theorema`Common`translate[ "checkingInsts", lang] = "Checking theorem instances";
Theorema`Common`translate[ "cmdPending", lang] = "command pending";
Theorema`Common`translate[ "defCyclic", lang] = "cyclic definition";
Theorema`Common`translate[ "defIllegalType", lang] = "definition with illegal type on RHS";
Theorema`Common`translate[ "defOverlapping", lang] = "definition with overlapping LHS";
Theorema`Common`translate[ "defRecursive", lang] = "recursive definition";
Theorema`Common`translate[ "defValid", lang] = "admissible definition";
Theorema`Common`translate[ "Goal", lang] = "Goal";
Theorema`Common`translate[ "No", lang] = "No";
Theorema`Common`translate[ "noSubgoals", lang] = "No subgoals";
Theorema`Common`translate[ "openBranches", lang] = "more branch(es)";
Theorema`Common`translate[ "prf", lang] = "proof";
Theorema`Common`translate[ "prfCyclic", lang] = "cyclic proof";
Theorema`Common`translate[ "prfFinished", lang] = "Proof finished";
Theorema`Common`translate[ "prfInvalid", lang] = "invalid proof";
Theorema`Common`translate[ "prfIncomplete", lang] = "incomplete proof";
Theorema`Common`translate[ "prfSkipped", lang] = "skipped proof";
Theorema`Common`translate[ "prfValid", lang] = "complete proof";
Theorema`Common`translate[ "selectItem", lang] = "select item";
Theorema`Common`translate[ "Subgoals", lang] = "Subgoals";
Theorema`Common`translate[ "thm", lang] = "theorem";
Theorema`Common`translate[ "unevalFmlCell", lang] = "unevaluated formula cell";
Theorema`Common`translate[ "Yes", lang] = "Yes";

] (* With *)

$ContextPath = Drop[ $ContextPath, -2]
