(* ::Package:: *)

(* Theorema
    Copyright (C) 2017 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

BeginPackage[ "Theorema`Core`Tools`Interactive`"]

Needs[ "Theorema`Common`"]
Needs[ "Theorema`Core`Trusted`"]
Needs[ "Theorema`Core`Tools`"]
Needs[ "Theorema`Language`"]

Begin["`Private`"]

(* ::Section:: *)
(*Language*)

(* Basic idea of the current setup:
	- "Interactive.m" defines the default syntax of the proof language and also defines its default semantics.
	- Object logics may add syntax and arbitrarily change the default semantics (they may not change existing syntax, though).
	The semantics of a proof language is characterized by how language elements are translated into plain proof commands, i.e.
	by the value of '$proofLanguageTranslator'. *)

registerProofMethodRules[ args___] :=
	(
		$proofMethodMapping =
			Association @@ (
				First[ #] -> {{}, Last[ #]}& /@ {
					"default" -> (Theorema`CoreProvers`coreRule[ {##}]&),
					"rule" -> (Theorema`CoreProvers`coreRule[ {##}]&),
					"dest" -> (Theorema`CoreProvers`coreDest[ {##}, True]&),
					"dest1" -> (Theorema`CoreProvers`coreDest[ {##}, False]&),
					"elim" -> (Theorema`CoreProvers`coreElim[ {##}, True]&),
					"elim1" -> (Theorema`CoreProvers`coreElim[ {##}, False]&),
					"intro" -> (Theorema`CoreProvers`coreIntro[ {##}, True]&),
					"intro1" -> (Theorema`CoreProvers`coreIntro[ {##}, False]&),
					"assumption" -> (Theorema`CoreProvers`internalFactHandler[ Theorema`CoreProvers`closeByAssm[ ##], First, 0]&),
					"this" -> Theorema`CoreProvers`coreThis,
					"fact" -> (Theorema`CoreProvers`internalFactHandler[ Theorema`CoreProvers`closeByFact[ {##}], First, 0]&),
					"simp" -> (Theorema`CoreProvers`directRewriteAll[ {##}]&),
					"autored" -> (Theorema`CoreProvers`autoReduce[ {##}]&),
					"triv" -> Theorema`CoreProvers`closeByTerminal,
					"use" ->
						(Replace[ {##},
							{
								{} -> Theorema`CoreProvers`voidTac[],
								{"-"} -> Theorema`CoreProvers`internalFactHandler[ Theorema`CoreProvers`voidTac[], None, 2],
								facts_List :> Theorema`CoreProvers`addKnowledgeImplicit[ facts, "chainedFacts", False, 3, True, None]
							}
						]&)
				}
			);
		addProofMethodRules[ args]
	)

addProofMethodRules[ args___] := addToMapping[ $proofMethodMapping, args]

registerFormulaVariantRules[ args___] :=
	(
		$formulaVariantMapping =
			Association @@ (
				First[ #] -> {{}, Last[ #]}& /@ {
					"beta" -> Theorema`CoreProvers`CoreProofMethods`Private`normVariant[ "beta"],
					"\[Beta]" -> Theorema`CoreProvers`CoreProofMethods`Private`normVariant[ "beta"],
					"canon" -> Theorema`CoreProvers`CoreProofMethods`Private`canonVariant,
					"eq" -> Theorema`CoreProvers`CoreProofMethods`Private`eqVariant,
					"elimination" -> Theorema`CoreProvers`CoreProofMethods`Private`eliminationVariant,
					"eta" -> Theorema`CoreProvers`CoreProofMethods`Private`normVariant[ "eta"],
					"\[Eta]" -> Theorema`CoreProvers`CoreProofMethods`Private`normVariant[ "eta"],
					"for" -> Theorema`CoreProvers`CoreProofMethods`Private`forVariant,
					"in" -> Theorema`CoreProvers`CoreProofMethods`Private`inVariant,
					"norm" -> Theorema`CoreProvers`CoreProofMethods`Private`normVariant[ "normalize"],
					"of" -> Theorema`CoreProvers`CoreProofMethods`Private`ofVariant,
					"OF" -> Theorema`CoreProvers`CoreProofMethods`Private`OFVariant,
					"simplified" -> Theorema`CoreProvers`CoreProofMethods`Private`simplifiedVariant,
					"symmetric" -> Theorema`CoreProvers`CoreProofMethods`Private`symmetricVariant,
					"transitive" -> Theorema`CoreProvers`CoreProofMethods`Private`transitiveVariant
				}
			);
		addFormulaVariantRules[ args]
	)

addFormulaVariantRules[ args___] := addToMapping[ $formulaVariantMapping, args]

commandPreSplit[ stream_] :=
	stream //.
		{
			Boxes[ "(", l_List, ")"] :> l,
			Boxes[ "\[LeftAngleBracket]", {n0:(_String|_STRING$), params___}, "\[RightAngleBracket]"] :>
				With[ {res = THM$ @@ Prepend[ DeleteCases[ {params}, ","|":"], fastThmFromLabel[ Replace[ n0, STRING$[ n_String] :> n]]]},
					res /; True
				],
			Boxes[ "[", {}, "]"] -> Sequence[],
			Boxes[ "\[LeftDoubleBracket]", {}, "\[RightDoubleBracket]"] -> Sequence[]
		} //.
		{
			h_[ pre___, a:Except[ _Boxes], Boxes[ "[", l_List, "]"], post___] :>
				With[ {res = parseVariants[ a, l]},
					h[ pre, res, post] /; res =!= $Failed
				],
			h_[ pre___, a:Except[ _Boxes], Boxes[ "\[LeftDoubleBracket]", l_List, "\[RightDoubleBracket]"], post___] :>
				With[ {res = ToExpression[ RowBox[ l]]},
					h[ pre, VARIANT$[ a, Theorema`CoreProvers`CoreProofMethods`Private`partVariant, res], post] /; MatchQ[ res, _Span|_Integer]
				]
		}


(* ::Subsection:: *)
(*Theory Language*)

$theorySeparatorList =
{
	";", "undo", "endundo", "stop", "import", "judgment", "typedecl", "typedef", "lock", "decl", "facts", "rules",
	"overloading", "syntax", "notation", "abbreviation", "default", "collection",
	"type", "term", "fml", "simplify"
};

$theoryInterpretationRules1 =
{
	{Hold[ "undo"], _, post_} :> {Hold[ "undo"], post},
	{Hold[ "endundo"], Hold[], post_} :> {Hold[ "!undo"], post},
	{Hold[ "stop", ___], _, post_} :> {Hold[], If[ MatchQ[ post, _Hold], Hold[ "stop"], post]},
	{Hold[ a:("import"|"judgment"|"typedecl"|"typedef"|"lock"|"decl"|"facts"|"rules"|"overloading"|"syntax"|"notation"|"abbreviation"|"default"|"collection")], t_Hold, post_} :>
		{getCoreVar[ $theoryLanguageTranslator][ a, Hold[], t], post},
	{Hold[ a:("type"|"term"|"fml"|"simplify")], t_Hold, post_} :>
		{getCoreVar[ $theoryLanguageTranslator][ a, Hold[], t], post}
};

$theoryInterpretationRules2 =
{
	{Hold[ ";"]|Null, Hold[], post_} :> {Hold[], post}
};

theoryCommandPreSplit[ stream_] :=
	commandPreSplit[ stream] //.
		h_[ pre___, Boxes[ "\[LeftDoubleBracket]", l_List, "\[RightDoubleBracket]"], post___] :>
			With[ {res =
					Block[ {Theorema`Interface`ProofScript`$commandSeparators = $proofCommandSeparators,
							Theorema`Interface`ProofScript`$commandInterpretationRules = Join[ $proofCommandInterpretationRules1, $proofCommandInterpretationRules2],
							Theorema`Interface`ProofScript`$commandPreSplit = commandPreSplit},
						Theorema`Interface`ProofScript`Private`interpretCommandStream[ Hold @@ l, False]
					]},
				If[ res === $Failed,
					$Failed,
				(*else*)
					h[ pre, res, post]
				]
			]

registerCustomTheoryInterpretationRules[ args___] :=
	(
		$theoryCommandSeparators = Alternatives @@ $theorySeparatorList;
		$theoryCommandInterpretationRules1 = $theoryInterpretationRules1;
		$theoryCommandInterpretationRules2 = $theoryInterpretationRules2;
		addCustomTheoryInterpretationRules[ args];
	)

addCustomTheoryInterpretationRules[] :=
	addCustomTheoryInterpretationRules[ {}]
addCustomTheoryInterpretationRules[ rules1_List] :=
	addCustomTheoryInterpretationRules[ rules1, {}]
addCustomTheoryInterpretationRules[ rules1_List, rules2_List] :=
	addCustomTheoryInterpretationRules[ rules1, rules2, {}]
addCustomTheoryInterpretationRules[ rules1_List, rules2_List, sep_List] :=
	addCustomTheoryInterpretationRules[ rules1, rules2, sep, EvalTheoryName[]]
addCustomTheoryInterpretationRules[ rules1_List, rules2_List, sep_List, thy_String] :=
	(
		addToConstPattern[ $theoryCommandSeparators, sep, thy];
		$theoryCommandInterpretationRules1 = Join[ $theoryCommandInterpretationRules1, addScopeCondition[ thy] /@ rules1];
		$theoryCommandInterpretationRules2 = Join[ $theoryCommandInterpretationRules2, addScopeCondition[ thy] /@ rules2];
		(* If new rules are prepended rather than appended, image imports must be changed accordingly. *)
	)

addCriticalTheoryCommands[ rules_List] :=
	addCriticalTheoryCommands[ rules, EvalTheoryName[]]
addCriticalTheoryCommands[ rules_List, thy_String] :=
	($criticalTheoryCommands =
		Join[
			DeleteCases[ $criticalTheoryCommands, (h_[ _?(MemberQ[ rules, # -> _]&), _] -> _) /; h === Condition],
			addScopeCondition[ thy] /@ rules
		];)

parseTheoremaCodeCell[ nb_NotebookObject, cell_CellObject, boxes_, ___] :=
	If[ Theorema`Interface`InteractivePalette`cellStyleActivatedQ[ "TheoremaCode"] === False,
		False,
	(*else*)
			Replace[ Quiet[ NotebookFileName[ nb], NotebookFileName::nosv],
				{
					nbFile_String :>
						If[ setTheory[ sourceLabel[ nbFile]] =!= $Failed,
							Block[ {Theorema`Interface`ProofScript`$commandSeparators = $theoryCommandSeparators,
									Theorema`Interface`ProofScript`$commandInterpretationRules = Join[ $theoryCommandInterpretationRules1, $theoryCommandInterpretationRules2],
									Theorema`Interface`ProofScript`$commandPreSplit = theoryCommandPreSplit},
								Theorema`Interface`ProofScript`Private`$pendingCommand = Hold[];
								Replace[ Reap[ Theorema`Interface`ProofScript`interpretBoxes[ boxes], "Message"],
									{ret_, msg_List} :> (coreNotification[ msg, False, {}, {nb, cell}]; {nb, cell, nbFile, ret})
								]
							],
						(*else*)
							False
						],
					_ :> (coreNotification[ {}, True, 303]; exitCoreCellEvaluation[ nb, cell, False]; False)
				}
			]
	]
parseTheoremaCodeCell[ args__] :=
	If[ Theorema`Interface`InteractivePalette`cellStyleActivatedQ[ "TheoremaCode"] === False,
		False,
	(*else*)
		With[ {nb = EvaluationNotebook[]},
			(* 'SelectionMove[]' is more cumbersome than 'EvaluationCell[]', but more reliable:
				sometimes 'EvaluationCell[]' gives wrong results -- especially when using the
				interactive palette, and especially when evaluating proof info cells.
				'EvaluationNotebook[]', on the other hand, seems to be fine. *)
			SelectionMove[ nb, All, EvaluationCell, AutoScroll -> False];
			Replace[ SelectedCells[ nb],
				{
					{cell_CellObject} :>
						With[ {res = parseTheoremaCodeCell[ nb, cell, args]},
							SelectionMove[ cell, After, Cell, AutoScroll -> False];
							res
						],
					_ -> Null
				}
			]
		]
	]

processTheoremaCell[ {nb_NotebookObject, cell_CellObject, nbFile_String, cmd0_Hold}] :=
	Block[ {$evaluationPosition = Null, $currentScope = Null},
		executeTheoryActions[
			Module[ {cmd = cmd0, undo = False, j = 0},
				List @@ Join @@ Flatten[
					Last[
						Reap[
							While[ cmd =!= Hold[],
								If[ undo,
									With[ {i = First[ FirstPosition[ cmd, "!undo", {0}, {1}, Heads -> False]]},
									With[ {out = If[ i === 0, cmd, Take[ cmd, i - 1]]},
										If[ i === 0, cmd = Hold[], undo = False; cmd = Drop[ cmd, i]];
										Sow[ Reverse[ Replace[ DeleteCases[ out, "undo"], a_ :> With[ {jj = ++j}, Hold[ a, True, jj] /; True], {1}]]]
									]],
								(*else*)
									With[ {i = First[ FirstPosition[ cmd, "undo", {0}, {1}, Heads -> False]]},
									With[ {out = If[ i === 0, cmd, Take[ cmd, i - 1]]},
										If[ i === 0, cmd = Hold[], undo = True; cmd = Drop[ cmd, i]];
										Sow[ Replace[ DeleteCases[ out, "!undo"], a_ :> With[ {jj = ++j}, Hold[ a, False, jj] /; True], {1}]]
									]]
								]
							]
						]
					],
					1
				]
			],
			nb,
			nbFile,
			cell
		]
	]
processTheoremaCell[ {nb_NotebookObject, cell_CellObject, __}] :=
	(coreNotification[ {}, True, 302, {nb, cell}]; exitCoreCellEvaluation[ nb, cell, False];)
processTheoremaCell[ False] := Null		(* Theorema code cells are not activated *)
processTheoremaCell[ _] :=
	(coreNotification[ {}, True, 305]; exitCoreCellEvaluation[ Null, Null, False];)

executeTheoryActions[ acts_List, nb_NotebookObject, src_String, cell_CellObject] :=
	With[ {id = CurrentValue[ cell, CellID]},
	Module[ {out = Null},
		enterCoreCellEvaluation[ nb, cell, 0, id];
		Catch[
			Scan[
				Replace[ #,
					Hold[ a_, undo_, i_Integer] :>
						With[ {crit = Replace[ Head[ Unevaluated[ a]], $criticalTheoryCommands]},
							If[ undo === False && i =!= 1 && StringQ[ crit], coreNotification[ {MSG$[ coreMessageToList, 365, {crit}, "Warning"]}]];
							If[ (out = a[ undo, i, id, src, cell, nb]) === $Failed,
								exitCoreCellEvaluation[ nb, cell, False];
								Throw[ i]
							]
						]
				]&,
				acts
			];
			exitCoreCellEvaluation[ nb, cell, True];
			out
		]
	]]

defaultTLT[ "import", Hold[], Hold[ thys___]] :=
	Hold[ importTheories[ {thys}]]
defaultTLT[ "judgment", Hold[], Hold[ term_]] :=
	Hold[ setTruthJudgment[ term]] /. EXPR$[ e_] :> parseCoreTermPre[ e]
defaultTLT[ "typedecl", Hold[], Hold[ types___]] :=
	Hold @@ {declareTypes[ Replace[ {types}, EXPR$[ e_] :> parseType[ e, False], {1}]]}
defaultTLT[ "typedef", Hold[], Hold[ oo:("on"|"off")]] :=
	Hold @@ {typedefOnOff[ oo === "on"]}
defaultTLT[ "typedef", Hold[], Hold[ typ_, ":=", rhs_, opts___List]] :=
	Replace[ getMorphsLabel[ opts],
		{
			{{abs_, rep_}, _} :>
				Hold @@ {copyType[ Replace[ typ, EXPR$[ e_] :> parseType[ e, False]], Replace[ rhs, EXPR$[ e_] :> parseType[ e]], abs, rep]},
			_ -> $Failed
		}
	]
defaultTLT[ "typedef", Hold[], Hold[ typ_, "\[TildeEqual]", pred_, opts___List, prf:(Null|_Hold)]] :=
	Replace[ getMorphsLabel[ opts],
		{
			{{abs_, rep_}, lbl:Except[ $Failed]} :>
				Hold @@ {defineType[ Replace[ typ, EXPR$[ e_] :> parseType[ e, False]], Replace[ pred, EXPR$[ e_] :> parseCoreTermPre[ e]], prf, abs, rep, lbl]},
			_ -> $Failed
		}
	]
defaultTLT[ "typedef", Hold[], Hold[ typ_, "\[TildeEqual]", pred_, opts___List]] :=
	defaultTLT[ "typedef", Hold[], Hold[ typ, "\[TildeEqual]", pred, opts, Null]]
defaultTLT[ "lock", Hold[], Hold[ consts___]] :=
	Hold[ lockConstants[ {consts}]] /. EXPR$[ e_] :> parseCoreTermPre[ e]
defaultTLT[ "decl", Hold[], Hold[ consts___]] :=
	Hold[ declareConstants[ {consts}]] /. EXPR$[ e_] :> parseCoreTermPre[ e]
defaultTLT[ "facts", Hold[], k_Hold] :=
	Hold @@ translateNote[ k, addFactsToTheory]
defaultTLT[ "rules", Hold[], Hold[ cat_List, k___]] :=
	Replace[ makeFmlList[ False][ k],
		l_List :> Hold @@ {addRulesToTheory[ cat, l]}
	]
defaultTLT[ "overloading", Hold[], Hold[ oo:("off"|"on")]] :=
	Hold @@ {overloadingOnOff[ {"const", "type"}, oo === "on"]}
defaultTLT[ "overloading", Hold[], Hold[ l_List, oo:("off"|"on")]] :=
	Hold @@ {overloadingOnOff[ l, oo === "on"]}
defaultTLT[ "syntax", Hold[], Hold[ mods_List, lhs_, dir:("\[DownLeftVector]"|"\[RightVector]"|"\[Equilibrium]"), rhs_]] :=
	With[ {typ = MemberQ[ mods, "type"]},
	With[ {meth = If[ typ, typeSyntax, termSyntax], parser = If[ typ, parseType, parseCoreTermPre]},
		Hold @@ {
			meth[
				Switch[ dir, "\[DownLeftVector]", 2, "\[RightVector]", 1, _, 3],
				Replace[ lhs, EXPR$[ e_] :> parser[ e]],
				Replace[ rhs, EXPR$[ e_] :> parser[ e]]
			]
		}
	]]
defaultTLT[ "syntax", Hold[], Hold[ lhs_, dir_, rhs_]] :=
	defaultTLT[ "syntax", Hold[], Hold[ {}, lhs, dir, rhs]]
defaultTLT[ "notation", Hold[], Hold[ mods_List, terms___, ":", spec0__]] :=
	With[ {spec = getNotationSpec[ {spec0}]},
		If[ ListQ[ spec],
			With[ {typ = MemberQ[ mods, "type"], input = MemberQ[ mods, "input"], output = MemberQ[ mods, "output"]},
			With[ {meth = If[ typ, typeNotation, termNotation], parser = If[ typ, parseType[ #, False]&, parseCoreTermPre]},
				Hold @@ {
					meth[
						If[ input,
							If[ output, 3, 1],
						(*else*)
							If[ output, 2, 3]
						],
						Replace[ {terms}, EXPR$[ e_] :> parser[ e], {1}],
						Replace[ spec, EXPR$[ e_] :> parser[ e], {1}]
					]
				}
			]],
		(*else*)
			$Failed
		]
	]
defaultTLT[ "notation", Hold[], Hold[ a:Except[ _List], b___]] :=
	defaultTLT[ "notation", Hold[], Hold[ {}, a, b]]
defaultTLT[ "abbreviation", Hold[], Hold[ mods_List, lhs_, dir:("\[DownLeftVector]"|"\[RightVector]"|"\[Equilibrium]"), rhs_]] :=
	If[ MemberQ[ mods, "type"],
		Hold @@ {
			abbrevType[
				Switch[ dir, "\[DownLeftVector]", 2, "\[RightVector]", 1, _, 3],
				Replace[ lhs, EXPR$[ e_] :> parseType[ e, False]],
				Replace[ rhs, EXPR$[ e_] :> parseType[ e]]
			]
		},
	(*else*)
		Hold @@ {
			abbrevTerm[
				Switch[ dir, "\[DownLeftVector]", 2, "\[RightVector]", 1, _, 3],
				Replace[ lhs, EXPR$[ e_] :> parseCoreTermPre[ e]],
				Replace[ rhs, EXPR$[ e_] :> parseCoreTermPre[ e]]
			]
		}
	]
defaultTLT[ "abbreviation", Hold[], Hold[ lhs_, dir_, rhs_]] :=
	defaultTLT[ "abbreviation", Hold[], Hold[ {}, lhs, dir, rhs]]
defaultTLT[ "default", Hold[], Hold[ "sort", int_, s_]] :=
	Hold @@ {setDefaultSort[ MemberQ[ Flatten[ {int}], "internal"], s]}
defaultTLT[ "default", Hold[], Hold[ "sort", s_]] :=
	defaultTLT[ "default", Hold[], Hold[ "sort", {}, s]]
defaultTLT[ "collection", Hold[], s_Hold] :=
	Hold[ Evaluate[ addGlobalFactList[ List @@ s]]]

defaultTLT[ "type", Hold[], Hold[ spec_List, types___]] :=
	Hold @@ {printTypes[ Replace[ {types}, EXPR$[ e_] :> parseType[ e], {1}], getPrintingSpec[ spec]]}
defaultTLT[ "type", Hold[], Hold[ types___]] :=
	defaultTLT[ "type", Hold[], Hold[ {}, types]]
defaultTLT[ "term", Hold[], Hold[ spec_List, terms___]] :=
	Hold @@ {printTerms[ Replace[ {terms}, EXPR$[ e_] :> parseCoreTermPre[ e], {1}], getPrintingSpec[ spec]]}
defaultTLT[ "term", Hold[], Hold[ terms___]] :=
	defaultTLT[ "term", Hold[], Hold[ {}, terms]]
defaultTLT[ "fml", Hold[], Hold[ spec_List, forms___]] :=
	Hold @@ translateNote[ Hold[ forms], printForms[ getPrintingSpec[ spec], MemberQ[ spec, "all"]]]
defaultTLT[ "fml", Hold[], Hold[ forms___]] :=
	defaultTLT[ "fml", Hold[], Hold[ {}, forms]]
defaultTLT[ "simplify", Hold[], Hold[ spec_List, args_List, terms___]] :=
	With[ {method = lookupCurrentValue @@ Lookup[ $proofMethodMapping, "simp", {{}, $Failed}]},
	With[ {prf = If[ method === $Failed, List, method] @@ args},
		Hold @@ {simplifyTerms[ Replace[ {terms}, EXPR$[ e_] :> parseCoreTermPre[ e], {1}], getPrintingSpec[ spec], MemberQ[ spec, "context"], Hold[ prf]]}
	]]
defaultTLT[ "simplify", Hold[], Hold[ args_List, terms___]] :=
	defaultTLT[ "simplify", Hold[], Hold[ {}, args, terms]]
defaultTLT[ "simplify", Hold[], Hold[ terms___]] :=
	defaultTLT[ "simplify", Hold[], Hold[ {}, {}, terms]]

getMorphsLabel[ opts___List] :=
	Replace[ parseOptsAsAssociation[ "morphisms"|"label", Rest][ opts],
		{
			assoc_Association :>
				{
					getTypeMorphisms[ Lookup[ assoc, "morphisms", {}]],
					Replace[ Lookup[ assoc, "label", {"$"}],
						{
							{"$"|"_"} -> Automatic,
							{STRING$[ l_String]} :> l,
							_ -> $Failed
						}
					]
				},
			Missing[ k_] :> (sowMessage[ 364, {k}, "Fatal"]; $Failed)
		}
	]

getTypeMorphisms[ spec0_List] :=
	With[ {spec1 = Replace[ spec0, {"_"|"$" -> Automatic, EXPR$[ e_] :> e}, {1}]},
	With[ {len = Length[ spec1]},
		If[ len > 2,
			$Failed,
		(*else*)
			Join[ spec1, Table[ Automatic, {2 - len}]]
		]
	]]

getNotationSpec[ {"binder"}] :=
	{0, True}
getNotationSpec[ {"binder", cond_}] :=
	Switch[ cond,
		"conditional",
		{0, True},
		"unconditional",
		{0, False},
		_,
		$Failed
	]
getNotationSpec[ {"binary"}] :=
	{1, Right}
getNotationSpec[ {"binary", fun_}] :=
	Switch[ fun,
		"right",
		{1, Right},
		"left",
		{1, Left},
		_EXPR$,
		{1, fun},
		_,
		$Failed
	]
getNotationSpec[ {"recursive", base_EXPR$, constr_EXPR$}] :=
	{2, base, constr, Right, False}
getNotationSpec[ {"recursive", base_EXPR$, constr_EXPR$, lr_}] :=
	Switch[ lr,
		"left",
		{2, base, constr, Left, False},
		"right",
		{2, base, constr, Right, False},
		"reverse",
		{2, base, constr, Right, True},
		_,
		$Failed
	]
getNotationSpec[ {"recursive", base_EXPR$, constr_EXPR$, lr_, "reverse"}] :=
	Switch[ lr,
		"left",
		{2, base, constr, Left, True},
		"right",
		{2, base, constr, Right, True},
		_,
		$Failed
	]
getNotationSpec[ _] :=
	$Failed

(*
	- "display": 0 ("internal"), 1 ("postproc"), 2 ("pretty")
	- "return": 'True' ("return"), 'False' ("print")
*)
getPrintingSpec[ spec_List] :=
	With[ {specRev = Reverse[ spec]},
	With[ {list =
			Join[
				Thread[ "display" -> Replace[ FirstCase[ specRev, "internal"|"postproc"|"pretty", {}, {1}], {"internal" -> {0}, "postproc" -> {1}, "pretty" -> {2}}]],
				Thread[ "return" -> Replace[ FirstCase[ specRev, "return"|"print", {}, {1}], {"return" -> {True}, "print" -> {False}}]]
			]},
		Association[ list]
	]]

parseOptsAsAssociation[ legalKeys_, fun_][ opts___List] :=
	Module[ {assoc = Association[]},
		Catch[
			Scan[
				Function[ o,
					AssociateTo[ assoc,
						Cases[
							Split[ o, #1 =!= "," && #2 =!= "," &],
							{k:Except[ ","], val___} :> If[ MatchQ[ k, legalKeys], k -> fun[ {k, val}], Throw[ Missing[ k]]]
						]
					]
				],
				{opts}
			]
		];
		assoc
	]

getEvaluationPosition[ id_Integer, nb_NotebookObject] :=
	If[ $evaluationPosition === $Failed || ListQ[ $evaluationPosition],
		$evaluationPosition,
	(*else*)
		$evaluationPosition = FirstPosition[ NotebookGet[ nb], Cell[ _, ___, CellID -> id, ___], $Failed, Infinity]
	]
getEvaluationPosition[ _, _] :=
	If[ ListQ[ $evaluationPosition],
		$evaluationPosition,
	(*else*)
		$Failed
	]

increaseEvaluationPosition[ val_Integer] :=
	If[ MatchQ[ $evaluationPosition, {__}],
		(* there is no need to increase the position if it has not been computed yet *)
		$evaluationPosition[[-1]] += val;
	]

getCurrentScope[ cell_CellObject, nb_NotebookObject] :=
	If[ $currentScope === $Failed || ListQ[ $currentScope],
		$currentScope,
	(*else*)
		With[ {id = CurrentValue[ cell, CellID], file = Quiet[ NotebookFileName[ nb], NotebookFileName::nosv]},
			$currentScope =
				If[ IntegerQ[ id] && StringQ[ file],
					{{cellIDLabel[ id], sourceLabel[ file]}, getEvaluationPosition[ id, nb]},
				(*else*)
					$Failed
				]
		]
	]
getCurrentScope[ id_Integer, nb_NotebookObject] :=
	If[ $currentScope === $Failed || ListQ[ $currentScope],
		$currentScope,
	(*else*)
		With[ {file = Quiet[ NotebookFileName[ nb], NotebookFileName::nosv]},
			$currentScope =
				If[ StringQ[ file],
					{{cellIDLabel[ id], sourceLabel[ file]}, getEvaluationPosition[ id, nb]},
				(*else*)
					$Failed
				]
		]
	]
getCurrentScope[ id_Integer, file_String] :=
	If[ $currentScope === $Failed || ListQ[ $currentScope],
		$currentScope,
	(*else*)
		$currentScope =
			If[ StringQ[ file],
				{{cellIDLabel[ id], sourceLabel[ file]}, getEvaluationPosition[ id, file]},
			(*else*)
				$Failed
			]
	]
getCurrentScope[ _, _] :=
	If[ ListQ[ $currentScope],
		$currentScope,
	(*else*)
		$Failed
	]
getCurrentScope[] :=
	If[ $currentScope === $Failed || ListQ[ $currentScope],
		$currentScope,
	(*else*)
		getCurrentScope[ EvalCell[], EvalNotebook[]]
	]


(* ::Subsection:: *)
(*Proof Language*)

$proofSeparatorList =
{
	";", ".", "..", "skip", "skip1", "apply", "subgoal", "proof", "by", "qed", "define", "let", "be", "using", "use",
	"unfolding", "unfold", "fix", "assume", "show", "have", "where", "obtain", "from", "with", "then",
	"hence", "thus", "context", "case", "note", "also", "finally", "moreover", "ultimately", "close"
};
	
$proofInterpretationRules1 =
{
	{Hold[ c:("skip"|"skip1"|"close")], Hold[], post_} :> {{INF$[ c]}, post},
	{Hold[ "apply"], m_Hold, post_} :> {getCoreVar[ $proofLanguageTranslator][ "apply", Hold[], m], post},
	{Hold[ "subgoal"], n_Hold, post_} :> {getCoreVar[ $proofLanguageTranslator][ "subgoal", Hold[], n], post},
	{Hold[ "proof"], Hold[], Hold[ "by"]} :> {Hold[], Hold[ "proof", "by"]},
	{Hold[ "proof"], m_Hold, post:Except[ Hold[ "by"]]} :> {getCoreVar[ $proofLanguageTranslator][ "proof", Hold[], m], post},
	{Hold[ "proof", "by"], m_Hold, post_Hold} :> getCoreVar[ $proofLanguageTranslator][ "proof by", Hold[], m, post],
	{Hold[ c:("by"|"qed")], m_Hold, post_} :> {getCoreVar[ $proofLanguageTranslator][ c, Hold[], m], post},
	{Hold[ "define"], m_Hold, post_} :> {getCoreVar[ $proofLanguageTranslator][ "define", Hold[], m], post},
	{Hold[ "let"], Hold[ x_], Hold[ "be"]} :> {Hold[], Hold[ x, "be"]},
	{Hold[ pre__, "be"], Hold[ t__], post_} :>
		Switch[ post,
			Except[ Hold[ "be"], _Hold],
			{getCoreVar[ $proofLanguageTranslator][ "let", Hold[], DeleteCases[ Hold[ pre, t], "and"]], post},
			_,
			{Hold[], Hold[ pre, t, "be"]}
		],
	{Hold[ "using"|"use"], fmls_Hold, post_Hold} :> {getCoreVar[ $proofLanguageTranslator][ "use", Hold[], fmls], post},
	{Hold[ "from"], fmls_Hold, post_Hold} :> {getCoreVar[ $proofLanguageTranslator][ "from", Hold[], fmls], post},
	{Hold[ "unfolding"|"unfold"], eqs_Hold, post_Hold} :> {getCoreVar[ $proofLanguageTranslator][ "unfold", Hold[], eqs], post},
	{Hold[ "fix"], consts_Hold, post_Hold} :> {getCoreVar[ $proofLanguageTranslator][ "fix", Hold[], consts], post},
	{Hold[ "assume"], assms_Hold, post_Hold} :> {getCoreVar[ $proofLanguageTranslator][ "assume", Hold[], assms], post},
	{Hold[ "show"], fmls_Hold, post_} :> {getCoreVar[ $proofLanguageTranslator][ "show", Hold[], fmls], post},
	{Hold[ "have"], fmls_Hold, post_Hold} :> {getCoreVar[ $proofLanguageTranslator][ "have", Hold[], fmls], post},
	{Hold[ pre___, "where"], fmls_Hold, post_Hold} :> {getCoreVar[ $proofLanguageTranslator][ "where", Hold[ pre], fmls], post},
	{Hold[ pre___, "obtain"], Hold[ args___], Hold[ c:("obtain"|"where")]} :> {Hold[], Hold[ pre, "obtain", args, c]},
	{Hold[ "context"], t_Hold, post_} :> {getCoreVar[ $proofLanguageTranslator][ "context", Hold[], t], post},
	{Hold[ "case"], t_Hold, post_Hold} :> {getCoreVar[ $proofLanguageTranslator][ "case", Hold[], t], post},
	{Hold[ "note"], t_Hold, post_Hold} :> {getCoreVar[ $proofLanguageTranslator][ "note", Hold[], t], post},
	{Hold[ c:("also"|"moreover"|"finally"|"ultimately")], t_Hold, post_Hold} :> {getCoreVar[ $proofLanguageTranslator][ c, Hold[], t], post},
	
	(*abbreviations*)
	{Hold[ "with"], fmls_Hold, post_Hold} :> {getCoreVar[ $proofLanguageTranslator][ "from", Hold[], Append[ fmls, "this"]], post},
	{Hold[ "then"], Hold[], post_} :> {getCoreVar[ $proofLanguageTranslator][ "from", Hold[], Hold[ "this"]], post},
	{Hold[ "thus"], fmls_Hold, post_} :>
		{Join[ getCoreVar[ $proofLanguageTranslator][ "from", Hold[], Hold[ "this"]], getCoreVar[ $proofLanguageTranslator][ "show", Hold[], fmls]], post},
	{Hold[ "hence"], fmls_Hold, post_Hold} :>
		{Join[ getCoreVar[ $proofLanguageTranslator][ "from", Hold[], Hold[ "this"]], getCoreVar[ $proofLanguageTranslator][ "have", Hold[], fmls]], post},
	{Hold[ "."], Hold[], post_} :> {getCoreVar[ $proofLanguageTranslator][ "by", Hold[], Hold[ "this"]], post},
	{Hold[ ".."], Hold[], post_} :> {getCoreVar[ $proofLanguageTranslator][ "by", Hold[], Hold[ "default"]], post}
};

$proofInterpretationRules2 =
{
	{Hold[ ";"]|Null, Hold[], Hold[ "."]} :> {getCoreVar[ $proofLanguageTranslator][ "by", Hold[], Hold[]], Null},
	{Hold[ ";"]|Null, Hold[], Hold[ ".."]} :> {getCoreVar[ $proofLanguageTranslator][ "by", Hold[], Hold[ "default"]], Null},
	{Hold[ ";"]|Null, Hold[], post_} :> {Hold[], post}
};

registerCustomProofInterpretationRules[ args___] :=
	(
		$proofCommandSeparators = Alternatives @@ $proofSeparatorList;
		$proofCommandInterpretationRules1 = $proofInterpretationRules1;
		$proofCommandInterpretationRules2 = $proofInterpretationRules2;
		addCustomProofInterpretationRules[ args];
	)

addCustomProofInterpretationRules[] :=
	addCustomProofInterpretationRules[ {}]
addCustomProofInterpretationRules[ rules1_List] :=
	addCustomProofInterpretationRules[ rules1, {}]
addCustomProofInterpretationRules[ rules1_List, rules2_List] :=
	addCustomProofInterpretationRules[ rules1, rules2, {}]
addCustomProofInterpretationRules[ rules1_List, rules2_List, sep_List] :=
	addCustomProofInterpretationRules[ rules1, rules2, sep, EvalTheoryName[]]
addCustomProofInterpretationRules[ rules1_List, rules2_List, sep_List, thy_String] :=
	(
		addToConstPattern[ $proofCommandSeparators, sep, thy];
		$proofCommandInterpretationRules1 = Join[ $proofCommandInterpretationRules1, addScopeCondition[ thy] /@ rules1];
		$proofCommandInterpretationRules2 = Join[ $proofCommandInterpretationRules2, addScopeCondition[ thy] /@ rules2];
		(* If new rules are prepended rather than appended, image imports must be changed accordingly. *)
		Theorema`Interface`ProofScript`$commandSeparators = $proofCommandSeparators;
		Theorema`Interface`ProofScript`$commandInterpretationRules = Join[ $proofCommandInterpretationRules1, $proofCommandInterpretationRules2];
		Theorema`Interface`ProofScript`$commandPreSplit = commandPreSplit;
	)

parseVariants[ a_, {}] := a
parseVariants[ a_, {m_, args___}] :=
	With[ {var = lookupCurrentValue @@ Lookup[ $formulaVariantMapping, m, {{}, $Failed}]},
		If[ var === $Failed,
			var,
		(*else*)
			parseVariants[ a, var, {args}]
		]
	]
parseVariants[ a_, m_, args_] :=
	With[ {i = LengthWhile[ args, # =!= "," &]},
		Switch[ Length[ args],
   			i,
   			VARIANT$[ a, m, args],
   		(*else*)
   			i + 1,
   			$Failed,
   		(*else*)
   			_,
   			parseVariants[ VARIANT$[ a, m, List @@ Take[ args, i]], Drop[ args, i + 1]]
   		]
  	]

(*ternary versions*)
defaultPLT[ "apply", Hold[], Hold[ m_]] :=
	{Replace[ makeProofCommand[ m], cmd:Except[ $Failed|Theorema`CoreProvers`voidTac[]] :> Theorema`CoreProvers`evalTac[ cmd, $generalizeFix = True]]}
defaultPLT[ "subgoal", Hold[], Hold[ i_Integer]] :=
	{Theorema`CoreProvers`selectSubgoal[ i]}
defaultPLT[ "subgoal", Hold[], _] :=
	{Theorema`CoreProvers`selectSubgoal[ 1]}
defaultPLT[ "proof", Hold[], Hold[ "-"]] :=
	{Theorema`CoreProvers`internalFactHandler[ Theorema`CoreProvers`voidTac[], All, 2]}
	(* "-" shall not be available as an ordinary proof method that can be combined by tacticals, that's why we don't
		call 'makeProofCommand' here. Similar in "use -". *)
defaultPLT[ "proof", Hold[], Hold[ m_]] :=
	{
		With[ {post = Theorema`CoreProvers`internalFactHandler[ Theorema`CoreProvers`voidTac[], None, 2]},
			Replace[ makeProofCommand[ m],
				{
					Theorema`CoreProvers`voidTac[] -> post,
					cmd:Theorema`CoreProvers`internalFactHandler[ __, 2] :> cmd,
					cmd:Theorema`CoreProvers`andTac[ ___, Theorema`CoreProvers`internalFactHandler[ __, 2]] :>
						Theorema`CoreProvers`evalTac[ cmd, $generalizeFix = True],
					cmd_Theorema`CoreProvers`andTac :> Theorema`CoreProvers`evalTac[ Append[ cmd, post], $generalizeFix = True],
					cmd:Except[ $Failed] :> Theorema`CoreProvers`evalTac[ Theorema`CoreProvers`andTac[ cmd, post], $generalizeFix = True]
				}
			]
		]
	}
defaultPLT[ "proof", Hold[], Hold[]] :=
	getCoreVar[ $proofLanguageTranslator][ "proof", Hold[], Hold[ "default"]]
defaultPLT[ "by", Hold[], Hold[ m_]] :=
	{
		Replace[ makeProofCommand[ m],
			{
				Theorema`CoreProvers`voidTac[] -> Theorema`CoreProvers`closeProofSit[],
				cmd_Theorema`CoreProvers`andTac :> Theorema`CoreProvers`evalTac[ Append[ cmd, Theorema`CoreProvers`closeProofSit[]], $generalizeFix = False],
				cmd:Except[ $Failed] :> Theorema`CoreProvers`evalTac[ Theorema`CoreProvers`andTac[ cmd, Theorema`CoreProvers`closeProofSit[]], $generalizeFix = False]
			}
		]
	}
defaultPLT[ "qed", Hold[], Hold[ m_]] :=
	{
		Replace[ makeProofCommand[ m],
			{
				Theorema`CoreProvers`voidTac[] -> Theorema`CoreProvers`closeProofSit[],
				cmd:Except[ $Failed] :> Theorema`CoreProvers`evalTac[ Theorema`CoreProvers`closeProofSit[ cmd], $generalizeFix = False]
			}
		]
	}
defaultPLT[ "by"|"qed", Hold[], Hold[]] :=
	{Theorema`CoreProvers`internalFactHandler[ Theorema`CoreProvers`closeProofSit[], All, 0]}	(* chained facts are made explicit only if no proof methods are provided *)
defaultPLT[ "define", Hold[], def_Hold] :=
	Replace[ makeFmlList[ True] @@ def,
		{
			{d_} :> {Theorema`CoreProvers`defineConst[ d]},
			_ -> $Failed
		}
	]
defaultPLT[ "let", Hold[], abbrev_Hold] :=
	If[ EvenQ[ Length[ abbrev]],
		Replace[
			makeFmlList[ True] @@ Replace[ Partition[ abbrev, 2],
				{
					Hold[ EXPR$[ x_], EXPR$[ t_]] :> EXPR$[ CoreEqual$TM[x, t]],
					Hold[ EXPR$[ x_], t_] :> EXPR$[ CoreEqual$TM[x, t]],
					Hold[ x_, EXPR$[ t_]] :> EXPR$[ CoreEqual$TM[x, t]],
					Hold[ x_, t_] :> CoreEqual$TM[x, t]
				},
				{1}
			],
			{
				a:{__EXPR$} :> Theorema`CoreProvers`addExplicitTermAbbrev /@ a,
				_ -> $Failed
			}
		],
	(*else*)
		$Failed
	]
defaultPLT[ "show", Hold[], fml_Hold] :=
	Replace[ makeFmlList[ True] @@ fml,
		{
			{f_} :> {Theorema`CoreProvers`selectSubgoal[ f, {}, 0]},
			_ -> $Failed
		}
	]
defaultPLT[ "have", hyps_Hold, fmls_Hold] :=
	Replace[ makeFmlList[ True] @@ fmls,
		l0:Except[ {}, _List] :> {Theorema`CoreProvers`setupSubgoals[ l0]}
	]
defaultPLT[ "where", consts_Hold, fmls_Hold] :=
	Replace[ makeFmlList[ True] @@ fmls,
		l0:Except[ {}, _List] :>
			{Theorema`CoreProvers`obtainConstants[ parseCoreTermPre /@ DeleteCases[ List @@ consts, "obtain"|"and"], l0]}
	]
defaultPLT[ "use", Hold[], Hold[ "-"]] :=
	{Theorema`CoreProvers`internalFactHandler[ Theorema`CoreProvers`voidTac[], None, 2]}
defaultPLT[ "use", Hold[], fmls_Hold] :=
	Replace[ makeFmlList[ True] @@ fmls,
		l_List :> {Theorema`CoreProvers`addKnowledgeImplicit[ l, "chainedFacts", True, 3, True, None]}
		(* ATTENTION! Command "use <facts>" does not the same as proof method "use <facts>":
			The former *adds* the given facts to the chained facts, whereas the latter *replaces*
			the chained facts. *)
	]
defaultPLT[ "from", Hold[], fmls_Hold] :=
	Replace[ makeFmlList[ True] @@ fmls,
		l_List :> {Theorema`CoreProvers`addKnowledgeImplicit[ l, "chainedFacts", True, 3, True, None]}	(* here one could set 'acc' to 'False' *)
	]
defaultPLT[ "unfold", Hold[], eqs_Hold] :=
	{Theorema`CoreProvers`rewriteAll[ List @@ Prepend[ DeleteCases[ eqs, "and"] /. EXPR$[ e_] :> parseCoreTermPre[ e], ":"]]}
defaultPLT[ "fix", Hold[], consts_Hold] :=
	{Theorema`CoreProvers`introduceFix[ DeleteCases[ List @@ consts, "and"]]}
defaultPLT[ "assume", Hold[], Hold[ assms___, "..."]] :=
	Replace[ makeFmlList[ True][ assms],
		l_List :> {Theorema`CoreProvers`selectSubgoal[ "_", l, All]}
	]
defaultPLT[ "assume", Hold[], assms_Hold] :=
	Replace[ makeFmlList[ True] @@ assms,
		l_List :> {Theorema`CoreProvers`selectSubgoal[ "_", l, 0]}
	]
defaultPLT[ "context", Hold[], Hold[ lbl_String]] :=
	{Theorema`CoreProvers`introduceAssmsFix[ Null, lbl, {}]}
defaultPLT[ "context", Hold[], Hold[ {lbl_String, consts___}]] :=
	{Theorema`CoreProvers`introduceAssmsFix[ Null, lbl, {consts}]}
defaultPLT[ "case", Hold[], Hold[ lbl_String, ":", {case_String, consts___}]] :=
	{Theorema`CoreProvers`introduceAssmsFix[ case, lbl, {consts}]}
defaultPLT[ "case", Hold[], Hold[ lbl_String, ":", case_String]] :=
	{Theorema`CoreProvers`introduceAssmsFix[ case, lbl, {}]}
defaultPLT[ "case", Hold[], Hold[ {case_String, consts___}]] :=
	{Theorema`CoreProvers`introduceAssmsFix[ case, Automatic, {consts}]}
defaultPLT[ "case", Hold[], Hold[ case_String]] :=
	{Theorema`CoreProvers`introduceAssmsFix[ case, Automatic, {}]}
defaultPLT[ "case", Hold[], assms_Hold] :=
	defaultPLT[ "assume", Hold[], assms]	(* This version of "case" is a mere synonym for "assume". *)
defaultPLT[ "note", Hold[], k_Hold] :=
	translateNote[ k, Theorema`CoreProvers`addKnowledgeImplicit]
defaultPLT[ "also", Hold[], rules_Hold] :=
	Replace[ makeFmlList[ True] @@ rules,
		l_List :> {Theorema`CoreProvers`combineCalculation[ l, False]}
	]
defaultPLT[ "finally", Hold[], rules_Hold] :=
	Replace[ makeFmlList[ True] @@ rules,
		l_List :> {Theorema`CoreProvers`combineCalculation[ l, True]}
	]
defaultPLT[ "moreover", Hold[], Hold[]] :=
	{Theorema`CoreProvers`accumulateCalculation[ False]}
defaultPLT[ "ultimately", Hold[], Hold[]] :=
	{Theorema`CoreProvers`accumulateCalculation[ True]}
(*4-ary versions*)
defaultPLT[ "proof by", Hold[], m_Hold, post_Hold] :=
	{{}, post}	(* TODO: Get proof technique (backchaining, induction, rewriting, ...). *)
defaultPLT[ __] := $Failed

translateNote[ Hold[ (a_String)|STRING$[ a_String], s:("="|"+="), k___], fun_] :=
	Replace[ makeFmlList[ True][ k],
		l_List :> {fun[ l, a, s === "+=", 3, False, None]}
	]
translateNote[ Hold[ (a_String)|STRING$[ a_String], Boxes[ "[", tags0_List, "]"], s:("="|"+="), k___], fun_] :=
	Replace[ {makeTagList[ tags0], makeFmlList[ True][ k]},
		{
			{tags_List, l_List} :> {fun[ addTags[ tags, l], a, s === "+=", 3, False, None]},
			_ -> $Failed
		}
	]
translateNote[ Hold[ Boxes[ "[", tags0_List, "]"], "=", k___], fun_] :=
	Replace[ {makeTagList[ tags0], makeFmlList[ True][ k]},
		{
			{tags_List, l_List} :> {fun[ addTags[ tags, l], Null, False, 3, False, None]},
			_ -> $Failed
		}
	]
translateNote[ k_Hold, fun_] :=
	Replace[ makeFmlList[ True] @@ k,
		l_List :> {fun[ l, Null, False, 3, False, None]}
	]

makeFmlList[ allowPatt_][ fmls___] :=
	Module[ {new = DeleteCases[ Flatten[ {fmls}], "and"], changed = True},
		Catch[
			While[ changed,
				changed = False;
				new = Replace[ new,
						{pre___, a_, "is", b_, post___} :>
							Which[
								allowPatt =!= True,
								sowMessage[ 360, {}, "Fatal"];
								Throw[ $Failed],
							(*else*)
								b === "is" || b === ":",
								sowMessage[ 361, {b}, "Fatal"];
								Throw[ $Failed],
							(*else*)
								a === "is" || a === ":",
								sowMessage[ 362, {a}, "Fatal"];
								Throw[ $Failed],
							(*else*)
								True,
								changed = True;
								Replace[ a,
									{
										PEXPR$[ x_, y__] :> {pre, PEXPR$[ x, y, b], post},
										_ -> {pre, PEXPR$[ a, b], post}
									}
								]
							]
					]
			];
			changed = True;
			While[ changed,
				changed = False;
				new = Replace[ new,
						{
							{pre___, lbl:(STRING$[ _String]|"$"), Boxes[ "[", tags0_List, "]"], ":", b_, post___} :>
								With[ {tags = makeTagList[ tags0]},
									If[ ListQ[ tags],
										changed = True;
										{
											pre,
											If[ MatchQ[ b, _LABELED$],
												b,
											(*else*)
												LABELED$[ b, If[ lbl === "$", Automatic, First[ lbl]], "tags" -> First[ tags], "remtags" -> Last[ tags]]
											],
											post
										},
									(*else*)
										sowMessage[ 363, {tags0}, "Fatal"];
										Throw[ $Failed]
									]
								],
							{pre___, STRING$[ lbl_String], ":", b_, post___} :>
								(changed = True; {pre, If[ MatchQ[ b, _LABELED$], b, LABELED$[ b, lbl]], post})
						}
					]
			];
			If[ MemberQ[ new, ":"|"is"],
				$Failed,
			(*else*)
				new /. EXPR$[ e_] :> parseCoreTermPre[ e]
			]
		]
	]

makeTagList[ tags0_List] :=
	With[ {tags1 = Replace[ Flatten[ tags0], STRING$[ t_String] :> t, {1}]},
		If[ MatchQ[ tags1, {___String}],
			Module[ {add = {}, rem = {}},
				Scan[
					Replace[ #,
						{
							{"-", l___, ","} :> With[ {t = StringJoin[ Riffle[ {l}, " "]]}, add = DeleteCases[ add, t]; AppendTo[ rem, t]],
							{"-", l___} :> With[ {t = StringJoin[ Riffle[ {l}, " "]]}, add = DeleteCases[ add, t]; AppendTo[ rem, t]],
							{l___, ","} :> With[ {t = StringJoin[ Riffle[ {l}, " "]]}, rem = DeleteCases[ rem, t]; AppendTo[ add, t]],
							l_List :> With[ {t = StringJoin[ Riffle[ l, " "]]}, rem = DeleteCases[ rem, t]; AppendTo[ add, t]]
						}
					]&,
					Split[ tags1, #1 =!= "," &]
				];
				{add, rem}
			],
		(*else*)
			$Failed
		]
	]
makeTagList[ _] := $Failed

addTags[ {{}, {}}, l_] :=
	l
addTags[ {add_List, rem_List}, l_List] :=
	Replace[ l,
		{
			LABELED$[ expr_, lbl_, data___] :>
				With[ {add0 = Replace[ "tags", {data, _ -> {}}], rem0 = Replace[ "remtags", {data, _ -> {}}]},
					LABELED$[
						expr,
						lbl,
						(* (add n rem == {} & add0 n rem0 == {}) ==> (((T \ rem) u add) \ rem0) u add0 == (T \ ((rem \ add0) u rem0)) u ((add \ rem0) u add0) *)
						"tags" -> DeleteDuplicates[ Join[ DeleteCases[ add, Alternatives @@ rem0], add0]],
						"remtags" -> DeleteDuplicates[ Join[ DeleteCases[ rem, Alternatives @@ add0], rem0]],
						Sequence @@ DeleteCases[ {data}, "tags"|"remtags" -> _]
					]
				],
			expr_ :> LABELED$[ expr, Automatic, "tags" -> add, "remtags" -> rem]
		},
		{1}
	]

makeProofCommand[ "!"] :=
	"!"	(* only makes sense inside 'andTac' *)
makeProofCommand[ m_String] :=
	makeProofCommand[ {m}, Null]
makeProofCommand[ Subscript[ m_, sub:(_Integer?NonNegative|All|"all")]] :=
	Theorema`CoreProvers`mapAllTac[ makeProofCommand[ m], sub]
makeProofCommand[ Superscript[ m_, "?"]] :=
	Theorema`CoreProvers`tryTac[ makeProofCommand[ m]]
makeProofCommand[ Superscript[ m_, "*"]] :=
	Theorema`CoreProvers`repeatTac[ makeProofCommand[ m]]
makeProofCommand[ Superscript[ m_, "+"]] :=
	With[ {cmd = makeProofCommand[ m]},
		Theorema`CoreProvers`andTac[ cmd, Theorema`CoreProvers`repeatTac[ cmd]]
	]
makeProofCommand[ Superscript[ m_, n_Integer?NonNegative]] :=
	Theorema`CoreProvers`repeatNTac[ makeProofCommand[ m], n]
makeProofCommand[ m_List] :=
	makeProofCommand[ m, ","]
makeProofCommand[ m_] :=
	m
makeProofCommand[ {m_String, args___}, Null] :=
	With[ {method = lookupCurrentValue @@ Lookup[ $proofMethodMapping, m, {{}, $Failed}]},
		Which[
			method =!= $Failed,
			method[ args],
		(*else*)
			MemberQ[ {"close", "cut", "split", "rearrangePS", "if", "skip", "skip1"}, m],
			(* "close", "cut" etc. cannot be used as top-level proof methods, because when combined by certain
				tacticals they might lead to problems with subsequent 'internalFactHandler': it might refer to
				a different proof situation.
				Actually, this applies to all proof methods that change the current proof situation (like
				'setupSubgoals' etc.) ... they should not be called from the top level directly. *)
			Theorema`CoreProvers`failTac[ sowMessage[ 301, {m}, "Fatal"]],
		(*else*)
			MemberQ[ {"psData", "insert", "hypData", "makeExplicit", "makeImplicit", "drop", "rearrangeG",
					"meta", "inst", "sorts", "instT", "allE", "allI", "forgetFix", "impE", "refl", "alphabetaeta", "replace", "ext", "eqI",
					"theorem", "formula", "forget"}, m],
			INF$[ m, args],
		(*else*)
			True,
			Symbol[ "Theorema`CoreProvers`" <> m][ args]
		] /. EXPR$[ e_] :> parseCoreTermPre[ e]
	]
makeProofCommand[ {m_}, Null] :=
	makeProofCommand[ m]
makeProofCommand[ m0_List, delim_] :=
	With[ {nextDelim = Replace[ delim, {"," -> ";", ";" -> "\[Rule]"|"->", "|" -> Null, _ -> "|"}]},
		Replace[ DeleteCases[ Split[ m0, (!MatchQ[ #1, delim] && !MatchQ[ #2, delim])&], {delim}],
			{
				{} -> Theorema`CoreProvers`voidTac[],
				{m_List} :> makeProofCommand[ m, nextDelim],
				m:{__List} :>
					With[ {cmds = makeProofCommand[ #, nextDelim]& /@ m},
						Switch[ delim,
							",",
							Theorema`CoreProvers`andTac @@ cmds,
							";",
							Theorema`CoreProvers`andNewTac[ cmds],
							"|",
							Theorema`CoreProvers`orElseTac @@ cmds,
							_,
							Theorema`CoreProvers`thenTac @@ cmds
						]
					]
			}
		]
	]


(* ::Section:: *)
(*ProofScript Interface*)

(* In interactive Core proofs, '$TMAproofObject' is always a pair '{state, trust}'. *)

initInteractiveCoreProof[ init_Hold][ goal:FML$[ {_, src_String}, __], assms_List, interactive_] :=
	With[ {ps0 = Reap[ If[ setTheory[ src] === $Failed, $Failed, initiateProof[ goal, assms, $proofMode]], "Message"]},
	With[ {ps = First[ ps0], msg = printMessage[ Join @@ Last[ ps0], True]},
		If[ MatchQ[ ps, _PRFSTAT$],
			With[ {res = executeCoreProofCommands[ getCoreVar[ $initialProofSteps], {ps, True}, interactive]},
				{First[ res], init, Join[ msg, Last[ res]]}
			],
		(*else*)
			{$Failed, Hold[], msg}
		]
	]]

executeCoreProofCommands[ cmd_Hold, state_List, interactive_] :=
	MapAt[ printMessage[ interactive], ReplacePart[ Reap[ executeProofCommands[ cmd, state, $proofMode, True], "Message"], {2, 0} -> Join], {2}]

displayCoreKnowledge[] :=
	With[ {pos = $TMAproofObject[[1, 2]]},
		Replace[ pos,
			{
				{p_List, ___} :> displayCoreKnowledge[ Extract[ $TMAproofObject[[1, -1]], p]],
				{ps_PRFSIT$, ___} :> displayCoreKnowledge[ ps],
				_ -> Null
			}
		]
	]
displayCoreKnowledge[ PRFSIT$[ _, {}, __]] :=
	CreateDocument[
		TextCell[ translate[ "noKnowl"], Italic, Gray],
		StyleDefinitions -> makeColoredStylesheet[ "Proof State"],
		WindowTitle -> translate[ "availKnowl"],
		Saveable -> False,
		Editable -> False,
		Selectable -> False,
		ShowCellBracket -> False,
		ShowCellLabel -> False,
		WindowElements -> {}
	]
displayCoreKnowledge[ ps:PRFSIT$[ _, hyps_List, _, fix_, meta_, tfix_, tvar_, __]] :=
	With[ {keys = Replace[ getProofSitData[ ps, "chainedFacts", {}], {{k:{_, _}, perm_} :> (k -> perm), k_ :> (k -> True)}, {1}]},
		CreateDocument[
			Grid[
				Map[
					Function[ f,
						{
							TextCell[
								Switch[ Replace[ key@f, keys],
									True,
									"\[FilledRightTriangle]",
									False,
									"\[RightTriangle]",
									_,
									Row[ {}]
								],
								"Text"
							],
							DisplayForm @ selectableExpressionBox[ printCoreTerm[ formula@f, {fix, meta, tfix, tvar}]],
							TextCell[ displayLabel[ f], "Text"]
						},
						{HoldFirst}
					],
					hyps
				],
				Spacings -> {2, 0},
				ColumnAlignments -> Left
			],
			StyleDefinitions -> makeColoredStylesheet[ "Proof State"],
			WindowTitle -> translate[ "availKnowl"],
			Saveable -> False,
			Editable -> False,
			Selectable -> False,
			ShowCellBracket -> False,
			ShowCellLabel -> False,
			WindowElements -> {}
		]
	]

displayCoreProofState[ Hold[ pending_]][ $id_] :=
	With[ {pos = $TMAproofObject[[1, 2]]},
		Replace[ pos,
			{
				{} :>
					displayCoreProofSituation[ $id, Null, 0, pending],
				{p_List, ___List} :>
					displayCoreProofSituation[ $id, Extract[ $TMAproofObject[[1, -1]], p], Length[ pos], pending],
				{ps_PRFSIT$, ___PRFSIT$} :>
					displayCoreProofSituation[ $id, ps, Length[ pos], pending],
				_ -> $Failed
			}
		]
	]

displayCoreProofSituation[ $id_, Null, 0, pending_] :=
	With[ {statSize = CurrentValue[ SequentialEvaluation`Private`$stateWindow[$id], "WindowSize"]},
		Pane[
			TextCell[ "\[Checkmark] " <> translate[ "prfFinished"], "Text"],
			Scrollbars -> {False, Automatic},
			AppearanceElements -> None,
			ImageSize -> {Full, Last[ statSize] - 90}
		]
	]
displayCoreProofSituation[ $id_, ps_PRFSIT$, n_Integer, pending_] :=
	With[ {statSize = CurrentValue[ SequentialEvaluation`Private`$stateWindow[$id], "WindowSize"]},
		Pane[
			Column[
				Join[
					displayCurrentGoals[
						$id,
						ps,
						Replace[ {n, pending},
							{
								{1, True} :> "  [" <> translate[ "cmdPending"] <> "]",
								{1, _} -> "",
								_ :> "  [" <> ToString[ n - 1] <> " " <> translate[ "openBranches"] <> If[ TrueQ[ pending], "; " <> translate[ "cmdPending"] <> "]", "]"]
							}
						]
					],
					displayCurrentFacts[ $id, ps]
				],
				Left
			],
			Scrollbars -> {False, Automatic},
			AppearanceElements -> None,
			ImageSize -> {Full, Last[ statSize] - 20}
		]
	]

displayCurrentGoals[ $id_, PRFSIT$[ {}, __], txt_String] :=
	{TextCell[ translate[ "noSubgoals"] <> txt, "Text"]}
displayCurrentGoals[ $id_, PRFSIT$[ {G_}, _, _, fix_, meta_, tfix_, tvar_, __], txt_String] :=
	{
		TextCell[ translate[ "Goal"] <> txt, "Text"],
		DisplayForm @ selectableExpressionBox[ printCoreTerm[ G, {fix, meta, tfix, tvar}]]
	}
displayCurrentGoals[ $id_, PRFSIT$[ G:{_, __}, _, _, fix_, meta_, tfix_, tvar_, __], txt_String] :=
	{
		TextCell[ translate[ "Subgoals"] <> txt, "Text"],
		Grid[
			MapIndexed[
				Function[ {s, index},
					{ToString[ First[ index]], DisplayForm @ selectableExpressionBox[ printCoreTerm[ s, {fix, meta, tfix, tvar}]]},
					{HoldFirst}
				],
				G
			],
			ColumnAlignments -> {Left, Left}
		]
	}

displayCurrentFacts[ _, ps:PRFSIT$[ _, hyps_List, _, fix_, meta_, tfix_, tvar_, __]] :=
	With[ {facts = Join @@ Map[ Cases[ hyps, FML$[ Replace[ #, {k:{_, _}, _} :> k], __], {1}, 1]&, getProofSitData[ ps, "chainedFacts", {}]],
			data = {fix, meta, tfix, tvar}},
		Join[
			If[ facts === {},
				displayCurrentFacts[
					"this:",
					Join @@ Map[
						Cases[ hyps, FML$[ Replace[ #, {k:{_, _}, _} :> k], __], {1}, 1]&,
						Lookup[ getProofSitData[ ps, "factLists", Association[]], "this", {}]
					],
					data
				],
			(*else*)
				displayCurrentFacts[ translate[ "chainedFacts"] <> ":", facts, data]
			],
			displayCurrentFacts[
				translate[ "Calculation"] <> ":",
				Join @@ Map[
					Cases[ hyps, FML$[ Replace[ #, {k:{_, _}, _} :> k], __], {1}, 1]&,
					getProofSitData[ ps, "calc", {}]
				],
				data
			]
		]
	]
displayCurrentFacts[ _String, {}, _List] :=
	{}
displayCurrentFacts[ caption_String, facts_List, data_List] :=
	{
		Item[ TextCell[ caption, "Text"], ItemSize -> {Automatic, 3}],
		Grid[
			Map[
				{
					"\[FilledSmallSquare]",
					DisplayForm @ selectableExpressionBox[ printCoreTerm[ formula@#, data]],
					displayLabel[ #]
				}&,
				facts
			],
			ColumnAlignments -> {Left, Left, Left}
		]
	}

displayLabel[ FML$[ _, _, l_, ___, "origForm" -> orig_, ___]] :=
	Tooltip[ displayLabel[ l], theoremaDisplay[ orig]]
displayLabel[ FML$[ _, _, l_, ___]] :=
	displayLabel[ l]
displayLabel[ l:Except[ "?", _String]] :=
	makeLabel[ l]
displayLabel[ _] :=
	""

printCoreProofSituation[] :=
	With[ {open = $TMAproofObject[[1, 2]]},
	With[ {ps =
			Switch[ open,
				{_List, ___},
				Extract[ $TMAproofObject[[1, -1]], First[ open]],
				
				{_PRFSIT$, ___},
				First[ open],
				
				_,
				$Failed
			]},
		If[ MatchQ[ ps, _PRFSIT$],
			{CellGroup[ {TextCell[ "Proof Situation", "Subsubsection"], ps}, 1]},
		(*else*)
			{}
		]
	]]

exportCoreProof[ goal_, assms_List, {PRFSTAT$[ id_, pos_List, keys_List, thms_List, po_PRFOBJ$], trust_}, time_, num_, reg_, flag_] :=
	Replace[ exportCoreProof[ goal, assms, {PRFSTAT$[ id, pos, keys, thms], trust}, time, num, reg, flag],
			(* Of course, 'pos' is not the same as 'open', but anyway the only thing that matters is whether the list is empty or not. *)
		Null :>
			(
				Theorema`Interface`GUI`Private`$addProofKB = {};
				Theorema`Interface`ProofScript`Private`exportProof[
					goal,
					Join[ assms, getFormulas[ True] /@ keys],
					{po, pos, trust},
					time, num, reg, flag
				]
			)
	]
exportCoreProof[ goal:FML$[ k_, __], assms_List, {PRFSTAT$[ _, open_List, keys_List, _List], trust_}, time_, num_, reg_, flag_] :=
	If[ flag =!= 0 || open === {},
		If[ reg === "internal",		(* "internal" is used for testing only, when no proof-info cells shall be written to the notebook. *)
			Null,
		(*else*)
			printCoreProofInfo[
				goal, time, num, reg,
				Which[
					MemberQ[ keys, k] || MemberQ[ assms, FML$[ k, __]],
					2,
				(*else*)
					open =!= {},
					3,
				(*else*)
					trust,
					5,
				(*else*)
					True,
					4
				]
			]
		],
	(*else*)
		If[ reg =!= "internal",
			printCoreProofInfo[
				goal, time, num, reg,
				If[ MemberQ[ keys, k] || MemberQ[ assms, FML$[ k, __]],
					2,
				(*else*)
					3
				]
			]
		];
		$Failed
	]
exportCoreProof[ goal_FML$, _List, i_Integer, time_, num_, reg_, flag_] :=
	If[ flag === 2,
		If[ reg === "internal",
			i,
		(*else*)
			printCoreProofInfo[ goal, time, num, reg, 1]
		],
	(*else*)
		If[ reg =!= "internal", printCoreProofInfo[ goal, time, num, reg, 1]];
		$Failed
	]
exportCoreProof[ ___] := Null

printCoreProofInfo[ goal:FML$[ k:{id_String, src_String}, __], time_, num_, reg_, stat_] :=
	(
		If[ TrueQ[ reg] && getTheoryMode[] =!= True,
			Reap[ manipulateContext[ "fml stat", k, stat], "Message"]
		];
		Replace[ notebookFromFile[ sourceFile[ src]],
			{nb_NotebookObject, close_} :>
				With[ {prfTag = "Proof|" <> id <> "-"},
				With[ {cells = Cells[ nb, CellTags -> prfTag <> ToString[ num]]},
					If[ cells === {},
						(* no replacement of existing proof -> put after last proof after env *)
						NotebookFind[ nb, extractCellID[ id], All, CellID, AutoScroll -> False];
						Replace[ Cells[ NotebookFind[ nb, "CloseEnvironment", Next, CellStyle, AutoScroll -> False]],
							{
								{start_CellObject} :>
									SelectionMove[
										lastCell[ nb, start, (MatchQ[ NotebookRead[ #], Cell[ _, "ProofInfo", ___]])&, Next],
										After,
										Cell,
										AutoScroll -> False
									],
								_ :> SelectionMove[ nb, After, Cell, AutoScroll -> False]
							}
						];
						NotebookWrite[ nb,
							Cell[
								TextData[
									Cell[ BoxData[ ToBoxes[ Button[ Style[ translate["ShowProof"], FontVariations -> {"Underline" -> True}],
										displayCoreProof[], ImageSize -> Automatic, Appearance -> None, Method -> "Queued"]]]
									]
								],
								"ProofInfo",
								ShowCellBracket -> Automatic,
								Evaluatable -> True,
								CellEvaluationFunction -> evaluateProofCell,
								CellTags -> {prfTag <> ToString[ num], ToString[ stat]},	(* cell tags must be strings *)
								CellFrameLabels -> {makeProofInfoContents[ goal, num, stat, reg, time], {None, None}}
									(* Contents of 'CellFrameLabels' can easily be changed afterward. *)
							],
							None,
							AutoScroll -> False
						],
					(*else*)
						Scan[
							SetOptions[ #, CellFrameLabels -> {makeProofInfoContents[ goal, num, stat, reg, time], {None, None}}]&,
							cells
						]
					];
					If[ TrueQ[ close], NotebookSave[ nb]; NotebookClose[ nb]];
				]]
		];
	)

printTimeInfo[ time_] :=
	"[" <> DateString[] <> If[ NumberQ[ time], ", " <> ToString[ time] <> " sec", ""] <> "]"

makeRegisterProofButton[ False] :=
	Cell[ BoxData[ ToBoxes[ Button[ Style[ "\[FivePointedStar]", FontColor -> LightGray],
		registerCoreProof[], ImageSize -> Automatic, Appearance -> None, Evaluator -> Automatic]]]
	]
makeRegisterProofButton[ True] :=
	"\[FivePointedStar]"

makeProofInfoContents[ goal_FML$, num_Integer, stat_Integer, reg_, time_] :=
	(* ATTENTION! Don't forget to adjust 'replaceProofInfoCell' when changing the contents of the cell! *)
	{
		Cell[
			TextData[
				{
					Switch[ stat, 1, "\[Times]", 2, "\:21bb", 3, "\[WarningSign]", 4, "\[Ellipsis]", 5, "\[CheckmarkedBox]", _, "?"] <> " ",
					makeRegisterProofButton[ reg],
					" " <> translate[ "Proof of"] <> " ",
					formulaReference[ goal, Tooltip -> False],
					" \[NumberSign]" <> ToString[ num] <> "  ",
					StyleBox[ printTimeInfo[ time], FontSize -> Small]
				}
			]
		],
		Cell[ BoxData[ ButtonBox[ "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> removeCoreProof[]]]]
	}

evaluateProofCell[ nb_NotebookObject, cell_CellObject] :=
	If[ Theorema`Interface`InteractivePalette`cellStyleActivatedQ[ "ProofInfo"] === False,
		Null,
	(*else*)
		With[ {nbFile = Quiet[ Check[ NotebookFileName[ nb], Return[ Null]]]},
		With[ {src = sourceLabel[ nbFile]},
			If[ setTheory[ src] =!= $Failed,
				Replace[ extractGoalInfoFromProofCell[ cell],
					{
						{id_String, num_String} :>
							Replace[ formulaFromKey[ {id, src}],
								{
									{fml_FML$} :>
										(Theorema`Interface`ProofScript`prove[
											fml,
											{},
											If[ Theorema`Interface`InteractivePalette`$enterInteractiveOnFailure === False, 2, 0],
											Number -> ToExpression[ num],
											Trigger -> {nb, cell}
										];),
									_ :>
										(
											coreNotification[ {}, True, 304, {nbFile, cell}];
											exitCoreCellEvaluation[ nb, cell, False];
												(* Actually, proof cells are not evaluated during a theory build,
													so it should be fine to call 'exitCellEvaluation' from "InteractivePalette.m" directly. *)
										)
								}
							],
						_ -> Null
					}
				]
			]
		]]
	]
evaluateProofCell[ ___] :=
	If[ Theorema`Interface`InteractivePalette`cellStyleActivatedQ[ "ProofInfo"] === False,
		Null,
	(*else*)
		With[ {nb = EvaluationNotebook[]},
			SelectionMove[ nb, All, EvaluationCell, AutoScroll -> False];
			Replace[ SelectedCells[ nb],
				{cell_CellObject} :> evaluateProofCell[ nb, cell]
			];
			SelectionMove[ nb, After, Cell, AutoScroll -> False];
		]
	]

registerCoreProof[] :=
	Replace[ extractGoalInfoFromProofCell[],
		{
			{nb_NotebookObject, {id_String, num0_String}} :>
				With[ {nbFile = Quiet[ Check[ NotebookFileName[ nb], Return[ Null]]], num = ToExpression[ num0]},
				With[ {src = sourceLabel[ nbFile], dir = FileNameJoin[ {DirectoryName[ nbFile], FileBaseName[ nbFile]}]},
				With[ {assoc0 = getRegisteredProofNumbersOfFile[ nbFile, True, dir], k = {id, src}},
				With[ {regNum =
						If[ MatchQ[ assoc0, _Association],
							Lookup[ assoc0, Key[ k], getRegisteredProofNumber[ dir, extractCellID[ id], "-nb.nb"]],
						(*else*)
							getRegisteredProofNumber[ dir, extractCellID[ id], "-nb.nb"]
						],
						assoc = If[ MatchQ[ assoc0, _Association], assoc0, Association[]]},
					If[ regNum =!= num,		(* security check *)
						With[ {new = Append[ assoc, k -> num], co = SelectedCells[ nb]},
							AssociateTo[ $registeredProofNumbers, nbFile -> new];
							Put[ new, FileNameJoin[ {dir, "reg.m"}]];
							If[ IntegerQ[ regNum], replaceProofInfoCell[ nb, regNum, False, k]];
							If[ MatchQ[ co, {_CellObject}], replaceProofInfoCell[ First[ co], True, k]]
						]
					];
				]]]],
			_ -> Null
		}
	]

replaceProofInfoCell[ nb_NotebookObject, num_Integer, reg_, k:{id_String, _String}] :=
	With[ {cells = Cells[ nb, CellTags -> "Proof|" <> id <> "-" <> ToString[ num]]},
		If[ cells === {},
			If[ getTheoryMode[] =!= True,
				Reap[ manipulateContext[ "fml stat", k, 0], "Message"]
			],
		(*else*)
			Scan[
				replaceProofInfoCell[ #, reg, k]&,
				cells
			]
		];
	]
replaceProofInfoCell[ co_CellObject, reg_, k_] :=
	Replace[ CurrentValue[ co, CellFrameLabels],
		{
			{{Cell[ TextData[ {stat_String, _, rest___}]], r_}, tb_List} :>
				(
					FirstCase[ Flatten[ {CurrentValue[ co, CellTags]}],
						s_String /; StringMatchQ[s, DigitCharacter..] :>
							Reap[ manipulateContext[ "fml stat", k, ToExpression[ s]], "Message"],
						Null
					];
					SetOptions[ co, CellFrameLabels -> {{Cell[ TextData[ {stat, makeRegisterProofButton[ reg], rest}]], r}, tb}];
				),
			_ -> Null
		}
	]

displayCoreProof[] :=
	Replace[ extractGoalInfoFromProofCell[],
		{
			{nb_NotebookObject, {id_String, num_String}} :>
				With[ {nbFile = Quiet[ Check[ NotebookFileName[ nb], Return[ Null]]]},
				With[ {file = FileNameJoin[ {DirectoryName[ nbFile], FileBaseName[ nbFile], "p" <> extractCellID[ id] <> "-" <> num <> "-nb.nb"}]},
					If[ FileExistsQ[ file], NotebookOpen[ file]]
				]],
			_ -> Null
		}
	]

removeCoreProof[] :=
	Replace[ extractGoalInfoFromProofCell[],
		{
			{nb_NotebookObject, {id_String, num0_String}} :>
				With[ {nbFile = Quiet[ Check[ NotebookFileName[ nb], Return[ Null]]], num = ToExpression[ num0]},
				With[ {src = sourceLabel[ nbFile], dir = FileNameJoin[ {DirectoryName[ nbFile], FileBaseName[ nbFile]}]},
				With[ {assoc0 = getRegisteredProofNumbersOfFile[ nbFile, True, dir], k = {id, src}, cid = extractCellID[ id]},
				With[ {regNum =
						If[ MatchQ[ assoc0, _Association],
							Lookup[ assoc0, Key[ k], getRegisteredProofNumber[ dir, cid, "-nb.nb"]],
						(*else*)
							getRegisteredProofNumber[ dir, cid, "-nb.nb"]
						],
						assoc = If[ MatchQ[ assoc0, _Association], assoc0, Association[]]},
					NotebookDelete[ nb];
					DeleteFile[ FileNames[ FileNameJoin[ {dir, "p" <> cid <> "-" <> num0 <> "-nb.nb"}]]];
					If[ regNum === num,
						With[ {proofCells =
								Replace[ getRegisteredProofNumber[ dir, cid, "-nb.nb"],
									{
										n_Integer :> Cells[ nb, CellTags -> "Proof|" <> id <> "-" <> ToString[ n]],
										_ -> {}
									}
								],
								new = KeyDrop[ assoc, {k}]},
							AssociateTo[ $registeredProofNumbers, nbFile -> new];	(* we intentionally do this even if 'new' is empty *)
							If[ new === Association[],
								Quiet[ DeleteFile[ FileNameJoin[ {dir, "reg.m"}]], DeleteFile::fdnfnd],
							(*else*)
								Put[ new, FileNameJoin[ {dir, "reg.m"}]]
							];
							If[ proofCells === {},
								If[ getTheoryMode[] =!= True,
									Reap[ manipulateContext[ "fml stat", k, 0], "Message"]
								],
							(*else*)
								Scan[
									replaceProofInfoCell[ #, True, k]&,
									proofCells
								]
							]
						]
					];
				]]]],
			_ -> Null
		}
	]

(* 'lastCell[ nb, start, pred, dir]' starts from cell 'start' and then considers all subsequent (or previous, depending
	on 'dir') cells until the first one fails to satisfy 'pred'; in that case, the function returns the last visited
	cell that satisfies 'pred', or 'start' if no cell satisfies 'pred'. The selection or insertion point is not moved! *)
If[ $VersionNumber >= 10.,
	With[ {nextCell = Symbol[ "NextCell"], prevCell = Symbol[ "PreviousCell"]},
		lastCell[ _NotebookObject, start_, pred_, Next] :=
			Replace[ nextCell[ start],
				{
					co_CellObject :> Quiet[ Check[ NestWhile[ nextCell, co, pred, 1, Infinity, -1], start, NestWhile::nres], NestWhile::nres],
					_ -> start
				}
			];
		lastCell[ _NotebookObject, start_, pred_, Previous] :=
			Replace[ prevCell[ start],
				{
					co_CellObject :> Quiet[ Check[ NestWhile[ prevCell, co, pred, 1, Infinity, -1], start, NestWhile::nres], NestWhile::nres],
					_ -> start
				}
			]
	],
(*else*)
	lastCell[ nb_NotebookObject, start_, pred_, dir_] :=
		With[ {cells = Cells[ nb]},
			Replace[ First[ FirstPosition[ cells, start, {0}, {1}, Heads -> False]],
				{
					0 -> start,
					i_Integer :>
						With[ {sub = If[ dir === Previous, Reverse[ Take[ cells, i - 1]], Drop[ cells, i]]},
							Replace[ First[ FirstPosition[ sub, co_CellObject /; !pred[ co], {0}, {1}, Heads -> False]],
								{
									0 :> If[ sub === {}, start, Last[ sub]],
									1 -> start,
									j_Integer :> sub[[j - 1]]
								}
							]
						]
				}
			]
		]
]


(* ::Section:: *)
(*Theory Builds and Notebook Evaluation*)

(* 'buildTheory[ file]' builds theory 'file'. The function returns 'True' or 'False' depending on whether the building
	process finished with success or not. Note that invalid proofs (of formulas, type definitions, etc.) only trigger
	warning messages, but do not lead to failure! *)
buildTheory[ file0_, stream_, checkThms_] :=
	With[ {file =
			Switch[ file0,
				_String,
				If[ StringMatchQ[ file0, "Source" ~~ __],
					sourceFile[ file0],
				(*else*)
					makeTheoryFileName[ file0]
				],
				_Path,
				makeTheoryFileName[ file0],
				_,
				$Failed
			]},
		If[ StringQ[ file] && FileExtension[ file] === "nb" && FileExistsQ[ file],
			If[ TrueQ[ enterCore[ True, False]],
				$abortBuild = False;
				$buildStatus = True;
				$TMALogFile =
					Switch[ stream,
						_String,
						OpenWrite[ stream],
					(*else*)
						Automatic,
						OpenWrite[ FileBaseName[ file] <> "-" <> DateString[ {"Year", "-", "Month", "-", "Day", "-", "Hour24", "-", "Minute", "-", "Second"}] <> ".txt"],
					(*else*)
						_,
						$Failed
					];
				If[ Head[ $TMALogFile] === OutputStream,
					WriteString[ $TMALogFile,
						">>> Building theory " <> file <> " <<<\n",
						">>> Time: " <> DateString[ "DateTime"] <> " <<<\n"
					]
				];
				With[ {time = First[ AbsoluteTiming[ evaluateNotebook[ file, Automatic]]], write = (Head[ $TMALogFile] === OutputStream)},
					If[ write,
						WriteString[ $TMALogFile, "\n--- Problematic formulas ---\n"];
						WriteString @@ Prepend[
							Join @@ Map[
								Cases[ #,
									{k_, l_String} :>
										With[ {stat = getFmlEnv[ False][ k][[3]]},
											ToString[ stat] <> "  " <> makeLabel[ l] <> "  " <> ToString[ k] <> "\n" /; !MemberQ[ {5, 15, 25}, stat]
										]
								]&,
								getFormulas[ False][]
							],
							$TMALogFile
						];
						WriteString[ $TMALogFile,
							"\n>>> Build finished with " <> If[ TrueQ[ $buildStatus], "SUCCESS", "FAILURE"] <> " <<<\n",
							">>> Elapsed time: ", time, " sec <<<\n"
						]
					];
					If[ TrueQ[ $buildStatus && checkThms],
						Theorema`Interface`InteractivePalette`setCurrentEvaluation[ translate[ "checkingInsts"]];
						If[ write, WriteString[ $TMALogFile, "\n>>> Checking instances of theorem schemas <<<\n"]];
						Replace[ AbsoluteTiming[ checkTheoremInsts[ False, True]],
							{time2_, num_} :>
								(
									If[ !IntegerQ[ num], $buildStatus = False];
									If[ write,
										WriteString[ $TMALogFile,
											If[ IntegerQ[ num],
												"\n>>> Checking a total of " <> ToString[ num] <> " theorem instances finished with SUCCESS <<<\n",
											(*else*)
												"\n>>> Checking theorem instances finished with FAILURE <<<\n"
											]
										];
										WriteString[ $TMALogFile, ">>> Elapsed time: ", time2, " sec <<<"]
									]
								)
						]
					];
					If[ write,
						Close[ $TMALogFile];
						$TMALogFile = Null
					];
					Theorema`Interface`InteractivePalette`abortEvaluation[ False];
					With[ {res = $buildStatus},
						$buildStatus = None;
						res
					]
				]
			],
		(*else*)
			False
		]
	]

(* "ProofScript.m" does not call 'enterCoreCellEvaluation' and 'exitCoreCellEvaluation', but directly
	calls the corresponding functions in "InteractivePalette.m". This should be fine, since in a
	theory build the functions in "ProofScript.m" are not called anyway. *)
enterCoreCellEvaluation[ nb_, cell_, cat_Integer, spec_, params___] :=
	(
		If[ $buildStatus =!= None && Head[ nb] === NotebookObject && Head[ $TMALogFile] === OutputStream,
			WriteString[
				$TMALogFile,
				"\n*** ",
				DateString[ "Time"],
				"  ",
				Quiet[ Check[ FileBaseName[ NotebookFileName[ nb]] <> ".nb", "???"]],
				"  ",
				Switch[ cat,
					0,
					ToString[ spec] <> " [Theorema code]",
					1,
					ToString[ spec] <> " [Theorema input]",
					2,
					makeLabel[ spec],
					3,
					makeLabel[ spec] <> " [proof]",
					_,
					Sequence @@ {}
				],
				" ***\n"
			]
		];
		Theorema`Interface`InteractivePalette`enterCellEvaluation[ nb, cell, cat, spec, params]
	)

exitCoreCellEvaluation[ nb_, cell_, success_] :=
	If[ $buildStatus === None || TrueQ[ success],
		Theorema`Interface`InteractivePalette`exitCellEvaluation[ nb, cell, success],
	(*else*)
		$abortBuild = True;
		$buildStatus = False;
		Theorema`Interface`InteractivePalette`exitCellEvaluation[ nb, cell, True]
			(* 'success' is set to 'True', because otherwise the interactive palette aborts the evaluation. *)
	]

(* 'evaluateNotebook' tries to resemble built-in 'NotebookEvaluate', which for some reasons does not work properly:
	- 'EvaluationNotebook[]' and 'EvaluationCell[]' sometimes give wrong results.
	- '$PreRead', '$Pre' etc. are not reset, i.e. after evaluating a formula cell, all subsequent cells are interpreted
		as formula cells, too. *)
evaluateNotebook[ file_String, args___] :=
	Replace[ notebookFromFile[ file],
		{
			{nb_NotebookObject, close_} :>
				With[ {res = evaluateNotebook[ nb, args]},
					If[ TrueQ[ close], NotebookSave[ nb]; NotebookClose[ nb]];
					res
				],
			_ -> $Failed
		}
	]
evaluateNotebook[ nb_NotebookObject, Automatic] :=
	evaluateNotebook[ nb, Join[ {"Code", "TheoremaCode", "FormalTextInputFormula"}, If[ TrueQ[ getTheoryMode[]], {}, {"ProofInfo"}]]]
		(* In strict theory mode, "ProofInfo"-cells must not be evaluated. *)
evaluateNotebook[ nb_NotebookObject, cellStyles:{__String}] :=
	With[ {cells0 = Cells[ nb, CellStyle -> cellStyles]},
		CurrentValue[ cells0, CellID];
			(* If we do not evaluate the above command and the notebook is not visible, Mathematica erroneously thinks
				that all the cells (except "ProofInfo") are not evaluatable. *)
	With[ {cells = Pick[ cells0, CurrentValue[ cells0, Evaluatable]]},
		Theorema`Interface`InteractivePalette`pushToEvaluationStack[
			nb,
			Theorema`Interface`InteractivePalette`thinSelectedCells[ nb, cells, True, True, True]
		];
		evaluateNotebook[ nb, cells]
	]]
evaluateNotebook[ nb_NotebookObject, cells_List] :=
	Block[ {$evaluationNotebook = nb},
	Module[ {out = Null},
		Catch[
			Scan[
				(
					evaluateCell[ out, nb, #, NotebookRead[ #]];
					If[ TrueQ[ $abortBuild], Throw[ out]]
				)&,
				cells
			];
			out
		]
	]]

isNotebookEvaluation[] :=
	Head[ $evaluationNotebook] === NotebookObject

(* ATTENTION! 'evaluateCell' always evaluates "TheoremaCode" cells and "ProofInfo" cells according to the rules defined
	in Core, even if the current session is not a Core session. *)
SetAttributes[ evaluateCell, HoldFirst];
evaluateCell[ out_, nb_NotebookObject, cell_CellObject, Cell[ boxes_, "Computation", ___]] :=
	Replace[ Check[ ToExpression[ Theorema`Language`Session`Private`openComputation[ nb, boxes], StandardForm, Hold], out = $Failed],
		Hold[ expr_] :>
			Block[ {$evaluationCell = cell},
				out = processComputation[ expr];
				Theorema`Language`Session`Private`closeComputation[];
				out
			]
	]
evaluateCell[ out_, nb_NotebookObject, cell_CellObject, Cell[ boxes_, "GlobalDeclaration", ___]] :=
	Replace[ Check[ ToExpression[ Theorema`Language`Session`Private`openGlobalDeclaration[ boxes], StandardForm, Hold], out = $Failed],
		Hold[ expr_] :>
			Block[ {$evaluationCell = cell},
				out = Theorema`Language`Session`Private`processGlobalDeclaration[ nb, cell, expr];
				Theorema`Language`Session`Private`closeGlobalDeclaration[];
				out
			]
	]
evaluateCell[ out_, nb_NotebookObject, cell_CellObject, Cell[ boxes_, "FormalTextInputFormula", ___]] :=
	Replace[ Check[ ToExpression[ Theorema`Language`Session`Private`openEnvironment[ nb, boxes], StandardForm, Hold], out = $Failed],
		Hold[ expr_] :>
			Block[ {$evaluationCell = cell},
				out = processEnvironment[ nb, cell, expr];
				Theorema`Language`Session`Private`closeEnvironment[ nb];
				out
			]
	]
evaluateCell[ out_, nb_NotebookObject, cell_CellObject, Cell[ boxes_, "TheoremaCode", ___]] :=
	Block[ {$evaluationCell = cell},
		out = processTheoremaCell[ parseTheoremaCodeCell[ nb, cell, boxes]]
	]
evaluateCell[ out_, nb_NotebookObject, cell_CellObject, Cell[ _, "ProofInfo", ___]] :=
	Block[ {$evaluationCell = cell},
		out = evaluateProofCell[ nb, cell]
	]
evaluateCell[ out_, nb_NotebookObject, cell_CellObject, Cell[ boxes_, __]] :=
	Block[ {$evaluationCell = cell},
	With[ {res = (AbsoluteCurrentValue[ cell, CellProlog]; evaluateCell[ out, AbsoluteCurrentValue[ cell, CellEvaluationFunction][ boxes]])},
		AbsoluteCurrentValue[ cell, CellEpilog];
		res
	]]
evaluateCell[ out_, TextData[ l_List]] :=
	evaluateCell[ out, BoxData[ l]]
evaluateCell[ out_, (TextData|BoxData)[ bx:Except[ _List]]] :=
	evaluateCell[ out, BoxData[ {bx}]]
evaluateCell[ out_, bx_] :=
	evaluateCell[ out, BoxData[ {bx}]]
evaluateCell[ out_, BoxData[ l_List]] :=
	Scan[
		(* lines must be parsed individually, since some might change context *)
		Function[ bx,
			With[ {bx1 = If[ ValueQ[ $PreRead], $PreRead, Identity][ bx],
					head = If[ ValueQ[ $Pre], $Pre, Replace[ {##}, {___, x_} :> x]&]},
				out =
					If[ ValueQ[ $Post], $Post, Identity][
						Switch[ bx1,
							_String,
							ToExpression[ bx1, InputForm, head],
						(*else*)
							_RowBox|_GridBox|_SubscriptBox|_SuperscriptBox|_SubsuperscriptBox|_FractionBox|_SqrtBox|_RadicalBox|
								_UnderscriptBox|_OverscriptBox|_UnderoverscriptBox|_FrameBox|
								_TagBox|_InterpretationBox|_FormBox|_TemplateBox|_ButtonBox|_StyleBox|_ErrorBox,
							ToExpression[ bx1, StandardForm, head],
						(*else*)
							_,
							head[ bx1]
						]
					]
			]
		],
		l
	]

End[]

EndPackage[]
