(* ::Package:: *)

(* Theorema
    Copyright (C) 2017 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

BeginPackage[ "Theorema`Core`Tools`Messages`"]

Needs[ "Theorema`Common`"]
Needs[ "Theorema`Language`"]
Needs[ "Theorema`Core`Trusted`"]
Needs[ "Theorema`Core`Tools`"]

Begin["`Private`"]

coreNotification[ msg_List] :=
	coreNotification[ msg, False]
coreNotification[ msg_List, isErr_] :=
	coreNotification[ msg, isErr, {}]
coreNotification[ msg_List, isErr_, errMsg_] :=
	coreNotification[ msg, isErr, errMsg, None]
coreNotification[ msg_List, isErr_, errMsg_, src_] :=
	coreNotification[ msg, isErr, errMsg, src, ""]
coreNotification[ msg_List, isErr_, errMsg_, src_, defTitle_] :=
	coreNotification[ msg, isErr, errMsg, src, defTitle, True]
coreNotification[ msg_List, isErr_, errMsg_Integer, src_, defTitle_, printLine_] :=
	coreNotification[ msg, isErr, {MSG$[ coreMessageToList, errMsg, {}, "Fatal"]}, src, defTitle, printLine]
coreNotification[ msg_List, isErr_, errMsg_MSG$, src_, defTitle_, printLine_] :=
	coreNotification[ msg, isErr, {errMsg}, src, defTitle, printLine]
coreNotification[ msg_List, isErr_, errMsg_List, src_, defTitle_, printLine_] :=
	(
		(* 'defTitle' is currently unsed, but be keep it *)
		Which[
			Head[ $TMALogFile] === OutputStream,
			WriteString @@ Prepend[
				Cases[
					Join[
						Cases[ msg, m_MSG$ :> printMessage[ m, False], {0, Infinity}],
						If[ TrueQ[ isErr], printMessage[ False] /@ errMsg, {}]
					],
					s_String :> (s <> "\n")
				],
				$TMALogFile
			],
		(*else*)
			TrueQ[ $Notebooks],
			With[ {opts =
					Replace[ printLine,
						{
							lbl_String :> {CellLabel -> ToString[ StringForm[ translate[ "duringEval"], lbl]], ShowCellLabel -> True},
							True :> {CellLabel -> ToString[ StringForm[ translate[ "duringEval"], $Line]], ShowCellLabel -> True},
							_ -> {}
						}
					],
					linkF = makeLinkFunction[ src]},
				printToConsole[
					Map[
						Cell[ BoxData[ ToBoxes[ linkF[ #]]], "Print", Sequence @@ opts]&,
						Join[
							Cases[ msg, m_MSG$ :> printMessage[ m, True], {0, Infinity}],
							If[ TrueQ[ isErr], printMessage[ True] /@ errMsg, {}]
						]
					]
				]
			]
		];
		isErr
	)

dialogYesNo[ msgID_Integer, args_List] :=
	(
		$Notebooks === False ||
		DialogInput[
			Column[
				{
					printMessage[ MSG$[ coreMessageToList, msgID, args], "", True],
					Row[
						{
							DefaultButton[ translate[ "Yes"], DialogReturn[ True], ImageSize -> 70],
							CancelButton[ translate[ "No"], DialogReturn[ False], ImageSize -> 70]
						},
						Spacer[ 10]
					]
				},
				Center,
				2
			]
		]
	)

printMessage[ boxes_][ msg_] :=
	printMessage[ msg, boxes]
printMessage[ i_Integer -> msg_List, boxes_] :=
	With[ {i0 = ToString[ i] <> ":"},
		Flatten[ printMessage[ #, i0, boxes]& /@ msg]
	]
printMessage[ msg_List, boxes_] :=
	Flatten[ printMessage[ #, boxes]& /@ msg]
printMessage[ msg_, boxes_] :=
	printMessage[ msg, "", boxes]
printMessage[ msg_List, pre_String, boxes_] :=
	Flatten[ printMessage[ #, pre, boxes]& /@ msg]
printMessage[ MSG$[ text_String], pre_String, boxes_] :=
	printMessage[ MSG$[ #1&, text, {}, "", ""], pre, boxes]
printMessage[ MSG$[ text_String, typ_String], pre_String, boxes_] :=
	printMessage[ MSG$[ #1&, text, {}, typ, ""], pre, boxes]
printMessage[ MSG$[ fun_, id_], pre_String, boxes_] :=
	printMessage[ MSG$[ fun, id, {}, "", ""], pre, boxes]
printMessage[ MSG$[ fun_, id_, args_], pre_String, boxes_] :=
	printMessage[ MSG$[ fun, id, args, "", ""], pre, boxes]
printMessage[ MSG$[ fun_, id_, args_List, typ_], pre_String, boxes_] :=
	printMessage[ MSG$[ fun, id, args, typ, ""], pre, boxes]
printMessage[ MSG$[ fun_, id_, args_List, typ_String, src_String], pre_String, boxes_] :=
	With[ {prefix = Join[ msgHeaderToList[ typ, boxes], If[ pre === "", {}, {pre}], If[ src === "", {}, {src <> ":"}]]},
		If[ TrueQ[ boxes],
			With[ {list =
					Replace[ fun[ id, args, boxes, $TmaLanguage],
						{
							s_String :> {s},
							Except[ _List] :> {HoldForm[ fun[ id, args]]}
						}
					]},
				Row[
					Join[
						prefix,
						list /. expr_ :> With[ {res = printDisplayStruct[ expr, boxes]}, res /; res =!= $Failed]
					],
					Spacer[ 2]
				]
			],
		(*else*)
			With[ {list =
					Replace[ fun[ id, args, boxes, $TmaLanguage],
						{
							s_String :> {s},
							Except[ _List] :> {ToString[ fun] <> "[" <> ToString[ id] <> ", " <> ToString[ args] <> "]"}
						}
					]},
				StringJoin @@ Join[
					Riffle[ prefix, "  "],
					If[ prefix === {}, {}, {"  "}],
					Riffle[
						Replace[
							list /. expr_ :> With[ {res = printDisplayStruct[ expr, boxes]}, res /; StringQ[ res]],
							x:Except[ _STRING$] :> ToString[ x],
							{1}
						],
						" "
					]
				]
			]
		]
	]
printMessage[ msg_, pre_, boxes_] :=
	Which[
		pre === "",
		ToString[ msg],
	(*else*)
		boxes,
		Row[ {pre, ToString[ msg]}, Spacer[ 2]],
	(*else*)
		True,
		pre <> "  " <> ToString[ msg]
	]

printDisplayStruct[ TYPEDISP$[ type_], True] :=
	Style[ theoremaDisplay[ printType[ type]], FontColor -> Gray]
printDisplayStruct[ (TERMDISP$|EXPR$)[ term_, ___], True] :=
	theoremaDisplay[ printCoreTerm[ makeSTERM[ term]]]
printDisplayStruct[ TTDISP$[ term_, type_], True] :=
	Row[ {printDisplayStruct[ TERMDISP$[ term], True], "::", printDisplayStruct[ TYPEDISP$[ type], True]}]
printDisplayStruct[ arg:KEYDISP$[ k_List], True] :=
	fmlHyperlink[ printDisplayStruct[ arg, False], k]
printDisplayStruct[ FKEYDISP$[ {id_String, src0_String}, l_String], boxes_] :=
	With[ {src = sourceFile[ src0]},
	With[ {disp = "(" <> l <> ") [" <> shortenFileName[ src] <> "]"},
		If[ TrueQ[ boxes],
			fmlHyperlink[ disp, src, extractCellID[ id]],
		(*else*)
			disp
		]
	]]
printDisplayStruct[ FKEYDISP$[ k_], boxes_] :=
	Replace[ getFormulas[ False][ k],
		{
			FML$[ _, _, l_String, ___] :> printDisplayStruct[ FKEYDISP$[ k, l], boxes],
			_ :> printDisplayStruct[ KEYDISP$[ k], boxes]
		}
	]
printDisplayStruct[ SYMDISP$[ s_], _] :=
	tmaSymbolToString[ s]
printDisplayStruct[ STRING$[ s_], _] :=
	ToString[ s, InputForm]
printDisplayStruct[ THM$[ thm_, ___], _] :=
	ToString[ thm, InputForm]
printDisplayStruct[ (TYPEDISP$|TERMDISP$|EXPR$)[ expr_, ___], _] :=
	ToString[ expr]
printDisplayStruct[ TTDISP$[ term_, type_], False] :=
	printDisplayStruct[ TERMDISP$[ term], False] <> "::" <> printDisplayStruct[ TYPEDISP$[ type], False]
printDisplayStruct[ KEYDISP$[ k_], False] :=
	ToString[ k]
printDisplayStruct[ _, _] :=
	$Failed

End[]

EndPackage[]
