(* ::Package:: *)

(* Theorema
    Copyright (C) 2017 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

(* This package is about parsing and pretty-printing Core expressions. *)

BeginPackage[ "Theorema`Core`Tools`Syntax`"]

Needs[ "Theorema`Common`"]
Needs[ "Theorema`Core`Trusted`"]
Needs[ "Theorema`Core`Tools`"]
Needs[ "Theorema`Language`"]

Begin["`Private`"]

(* ::Section:: *)
(*Parsing*)

(* ::Subsection:: *)
(*Types*)

adjustTypes[ expr_] :=
	Replace[ Reap[ adjustTypes[ expr, MessageName$TM|TYPED$|$typeHeads], {TFIX$, TVAR$, Hold[ TFIX$], Hold[ TVAR$], False}],
		{res_, {tfixSorts_List, tvarSorts_List, tfix_List, tvar_List, termSorts_List}} :>
			makeSortAssociations[
				Fold[ TYPED$[ CoreImplies$TM, {}][ #2][ #1]&, res, Reverse[ Flatten[ termSorts]]],
				Flatten[ tfixSorts],
				Flatten[ tvarSorts],
				Flatten[ tfix],
				Flatten[ tvar]
			]
	]
adjustTypes[ expr_, patt_] :=
	expr /.
		{
			(h:patt)[ term_, type_, rest___] :>
				If[ h === MessageName$TM, TYPED$, h] @@ Join[ {adjustTypes[ term, patt], parseType[ type, True, True, Null]}, adjustTypes[ {rest}, patt]],
			(h:patt) :> (sowMessage[ 351, {h}, "Error", "parsing"]; Throw[ $Failed])
		}

adjustSorts[ type_, allowTermSorts_] :=
	(
		type /.
			Join[
				{
					(MessageName$TM|TYPED$)[ type0_, sort0_] :>
						With[ {type1 = parseType[ type0, True, allowTermSorts, Null]},
							Switch[ type1,
								_TFIX$,
								Sow[ type1 -> sort0, TFIX$];
								type1,
							(*else*)
								_TVAR$,
								Sow[ type1 -> sort0, TVAR$];
								type1,
							(*else*)
								_,
								sowMessage[ 359, {type1}, "Error", "parsing"];
								$Failed
							]
						],
					h:MessageName$TM|TYPED$ :> (sowMessage[ 351, {h}, "Error", "parsing"]; Throw[ $Failed])
				},
				If[ getDefaultSort[] === getCoreVar[ $defaultInputSort],
					{},
				(*else*)
					{x_TFIX$ :> (Sow[ x, Hold[ TFIX$]]; x), x_TVAR$ :> (Sow[ x, Hold[ TVAR$]]; x)}
				]
			]
	)

(* In case of succes, 'parseType' returns a structure of the form 'EXPR$[ _, ___]'. *)
parseType[ type_] :=
	parseType[ type, True]
parseType[ type_, unfoldAbbrevs_] :=
	parseType[ type, unfoldAbbrevs, False]
parseType[ type_, unfoldAbbrevs_, allowTermSorts_] :=
	Replace[ Reap[ parseType[ type, unfoldAbbrevs, allowTermSorts, Null], {TFIX$, TVAR$, Hold[ TFIX$], Hold[ TVAR$]}],
		{newType_, {tfixSorts_List, tvarSorts_List, tfix_List, tvar_List}} :>
			makeSortAssociations[ newType, Flatten[ tfixSorts], Flatten[ tvarSorts], Flatten[ tfix], Flatten[ tvar]]
	]
parseType[ t_String, unfoldAbbrevs_, allowTermSorts_, Null] :=
	If[ StringMatchQ[ t, __ ~~ "_"],
		Replace[ parseType[ "Theorema`Knowledge`VAR$" <> StringDrop[ t, -1], unfoldAbbrevs, allowTermSorts, Null],
			{
				name_Symbol :> (If[ getDefaultSort[] =!= getCoreVar[ $defaultInputSort], Sow[ TFIX$[ name, 0], Hold[ TFIX$]]]; TFIX$[ name, 0]),
				_ -> $Failed
			}
		],
	(*else*)
		parseType[ tmaStringToSymbol[ t], unfoldAbbrevs, allowTermSorts, Null]
	]
parseType[ type_, unfoldAbbrevs_, allowTermSorts_, Null] :=
	With[ {type0 = adjustSorts[ type, allowTermSorts] /.
					{Subscript$TM[ Pattern$TM[ VAR$[ v_Symbol], Blank$TM[]], i_Integer?NonNegative] :> TFIX$[ v, i],
					Pattern$TM[ VAR$[ v_Symbol], Blank$TM[]] :> TFIX$[ v, 0]}},
	With[ {type1 = ReplacePart[ type0,
					DeleteCases[
						Map[
							With[ {name = With[ {x = Extract[ type0, Append[ #, 1], Unevaluated]}, SymbolName[ x]]},
								If[ StringMatchQ[ name, "VAR$" ~~ __],
									Block[ {$Context = "Theorema`Knowledge`",
											$ContextPath = Join[ {"Theorema`Language`"}, If[ ListQ[ $TheoremaArchives], $TheoremaArchives, {}], $ContextPath]},
										{#} -> ToExpression[ StringDrop[ name, 4]]
									],
								(*else*)
									$Failed
								]
							]&,
							Position[ type0, VAR$[ _Symbol]]
						],
						$Failed
					]
				]
			},
	With[ {type2 =
			applyRuleBlocks[
				ReplacePart[ type1,
					DeleteCases[
						Map[
							With[ {name = With[ {x = Extract[ type1, Append[ #, 1], Unevaluated]}, SymbolName[ x]]},
								If[ StringMatchQ[ name, "VAR$" ~~ __],
									$Failed,
								(*else*)
									Replace[
										Replace[ If[ MatchQ[ #, {___, 1}], Extract[ type1, {Most[ #]}], 0],
											{
												{Subscript$TM[ _, j_Integer?NonNegative]} :> {ToString[ j], Most[ #]},
												_ -> {"0", #}
											}
										],
										{i_String, p_List} :>
											With[ {c = ToExpression[ "Theorema`Language`TFIX$[Theorema`Knowledge`VAR$" <> name <> "," <> i <> "]"]},
												If[ getDefaultSort[] =!= getCoreVar[ $defaultInputSort], Sow[ c, Hold[ TFIX$]]];
												{p} -> c
											]
									]
								]
							]&,
							Position[ type1, Pattern$TM[ _Symbol, Blank$TM[]]]
						],
						$Failed
					]
				] /. {Blank$TM[] -> TVAR$[], Rule$TM -> FUN$TM} //.
				FUN$TM[ a_, b__, c_] :> Fold[ FUN$TM[ #2, #1] &, c, Reverse[ {a, b}]] //.	(* should not occur, but just to be sure ... *)
				FUN$TM[ CompoundExpression$TM[ a___], b_] :> Fold[ FUN$TM[ #2, #1] &, b, Reverse[ {a}]],
				$typeParsingRules
			] //.
			$binaryTypeParsingRules},
		If[ TrueQ[ unfoldAbbrevs], type2 /. $abbrevTypeUnfoldingRules, type2]
	]]]

(* UNUSED *)
makeTermSort[ type_, sort_Set$TM] :=
	getCoreVar[ $termSortConstructor][
		(#[ TYPED$[ Theorema`Knowledge`TYPE$TM, {type}]])& /@
			adjustTypes[ Replace[ List @@ sort, s_String :> tmaStringToSymbol[ s], {1}], MessageName$TM|TYPED$|$typeHeads]
	]
makeTermSort[ type_, class_] :=
	makeTermSort[ type, Set$TM[ class]]

makeSortAssociations[ expr_, fixSorts0_List, varSorts0_List, tfix_List, tvar_List] :=
	With[ {fun =
			DeleteDuplicates[
				Replace[ #,
					{
						(lhs_ -> s:(_Set$TM|_List)) :> (lhs -> Union[ List @@ s]),
						(lhs_ -> s_String) :> (lhs -> {tmaStringToSymbol[ s]}),
						(lhs_ -> s_) :> (lhs -> {s})
					},
					{1}
				]
			]&},
	With[ {fixSorts = fun[ fixSorts0], varSorts = fun[ DeleteCases[ varSorts0, TVAR$[] -> _]]},
	With[ {dup = Cases[ Join[ Tally[ fixSorts[[All, 1]]], Tally[ varSorts[[All, 1]]]], {a_, Except[ 0|1]} :> a],
			defInputSort = getCoreVar[ $defaultInputSort]},
		If[ dup === {},
			EXPR$[
				expr,
				Join[ Association @@ Map[ (# -> defInputSort)&, tfix], Association[ fixSorts]],	(* keys might not be disjoint, explicit sort annotations overwrite default ones *)
				Join[ Association @@ Map[ (# -> defInputSort)&, tvar], Association[ varSorts]]
			],
		(*else*)
			sowMessage[ 358, dup, "Error", "parsing"];
			Throw[ $Failed]
		]
	]]]


(* ::Subsection:: *)
(*Terms*)

(* In case of success, 'parseCoreTermPre' returns a structure of the form 'EXPR$[ _, ___]'. *)
parseCoreTermPre[ EXPR$[ term_, ___]] :=
	parseCoreTermPre[ term]
parseCoreTermPre[ term_] :=
	Catch[
		Replace[
			adjustTypes[
				term /. barSlantBracketted$TM[ typ_] :> TYPED$[ Theorema`Knowledge`TYPE$TM, {typ}]
							(* This rule must be applied *before* adjusting types. *)
			],
			EXPR$[ term1_, sorts___] :>
				EXPR$[
					applyRuleBlocks[
						toDB[ curryTerm[
							applyRuleBlocks[
								splitLambda[ term1] /.
								{
									(* These rules must be applied *after* adjusting types. *)
									Blank$TM[] -> SVAR$[],
									LongRightArrow$TM -> CoreImplies$TM,
									EqualTilde$TM -> CoreEqual$TM
								} //.
								(h:(CoreImplies$TM|TYPED$[ CoreImplies$TM, _]))[ a_, b__] :> binaryApply[ h, {a, b}, Right] //.
								(h:(CoreImplies$TM|TYPED$[ CoreImplies$TM, _]))[ a_CompoundExpression$TM][ b_] :> Fold[ h[ #2][ #1] &, b, Reverse[ List @@ a]],
								$uncurriedNominalParsingRules	(* rules operating on uncurried terms with named variables *)
							] //.
							$binaryTermParsingRules //.
							q:(Except[ Lambda$TM][ _RNG$, ___]) :> getCoreVar[ $binderParser][ q]
								(* This rule must be applied before currying, for otherwise 'freeVariables' returns wrong results. *)
						]],
						$curriedDBParsingRules	(* rules operating on curried terms with de Bruijn indices *)
					],
					sorts
				]
		]
	]

parseCoreTermPost[ term_] :=
	parseCoreTermPost[ term, normalize]
parseCoreTermPost[ term_, norm_] :=
	parseCoreTermPost[ term, norm, True]
parseCoreTermPost[ term0_, norm_, unfoldAbbrevs_] :=
	Replace[ applyRuleBlocks[ term0, $postParsingRules],	(* rules operating on typed terms *)
		{{typ_, STERM$[ term1_, termTFix_Association, termTVar_Association], rest__}, tsubst_} :>
			{
				With[ {term = If[ TrueQ[ unfoldAbbrevs], norm[ Evaluate[ Block[ {$parsingCtxt = {termTFix, termTVar}}, term1 /. $abbrevTermUnfoldingRules]]], norm[ term1]]},
					{typ, STERM$[ term, termTFix, termTVar], rest}
				],
				tsubst
			}
	]

toDB[ expr_] :=
	Catch[
		ReplacePart[
			expr,
			reap[ toDB[ expr, {}, {}]]
		]
	]
toDB[ v_VAR$, abs_, pos_] :=
	Sow[ pos -> VAR$[ First[ FirstPosition[ abs, varToSymbol[ v], sowMessage[ 353, {v}, "Error", "parsing"]; Throw[ $Failed], {1}]] - 1]]
toDB[ Lambda$TM[ RNG$[ TYPRNG$[ v_, _]], e_], abs_, pos_] := toDB[ e, Prepend[ abs, varToSymbol[ v]], Append[ pos, 2]]
toDB[ TYPED$[ e_, _], abs_, pos_] := toDB[ e, abs, Append[ pos, 1]]
toDB[ f_[ a___], abs_, pos_] :=
	Module[ {i = -1},
		Scan[
			toDB[ #, abs, Append[ pos, ++i]]&,
			{f, a}
		]
	]
toDB[ ___] := Null

varToSymbol[ VAR$[ s_Symbol]] := s
varToSymbol[ s_Symbol] := s
varToSymbol[ _] := Null

splitLambda[ term_] :=
	term /. l_Lambda$TM :> splitLambda[ l, Null]

splitLambda[ Lambda$TM[ rng_RNG$, True, e_], Null] :=
	splitLambda[ Lambda$TM[ rng, e], Null]
splitLambda[ Lambda$TM[ RNG$[ SIMPRNG$[ var_], b___], e_], Null] :=
	Lambda$TM[ RNG$[ TYPRNG$[ varToSymbol[ var], TVAR$[]]], splitLambda[ Lambda$TM[ RNG$[ b], e], Null]]
splitLambda[ Lambda$TM[ RNG$[ TYPRNG$[ var_, typ_], b___], e_], Null] :=
	Lambda$TM[ RNG$[ TYPRNG$[ varToSymbol[ var], typ]], splitLambda[ Lambda$TM[ RNG$[ b], e], Null]]
splitLambda[ Lambda$TM[ RNG$[], e_], Null] :=
	splitLambda[ e]
splitLambda[ e_, Null] :=
	(sowMessage[ 354, {e}, "Error", "parsing"]; Throw[ $Failed])

curryTerm[ TYPED$[ x_, t_]] :=
	TYPED$[ curryTerm[ x], t]
curryTerm[ Lambda$TM[ rng_RNG$, expr_]] :=
	Lambda$TM[ rng, curryTerm[ expr]]
curryTerm[ expr:(_VAR$|_FIX$|_META$|_SVAR$|_RNG$|_Complex|_Rational|_DirectedInfinity)] :=
	expr
curryTerm[ h_[]] :=
	curryTerm[ h]
curryTerm[ h_[ a__]] :=
	Fold[ #1[ #2]&, curryTerm[ h], curryTerm /@ {a}]
curryTerm[ expr_] :=
	expr

defaultBinderParser[ q_[ r0_RNG$, f0_]] :=
	Replace[ normalizeRange[ r0, {}, f0],
		{
			{r_List, {{}...}, f_} :>
				With[ {args = Reverse[ r]},
					Fold[
						q[ Lambda$TM[ #2, #1]]&,
						q[ Lambda$TM[ First[ args], f]],
						Rest[ args]
					]
				],
			_ :> (sowMessage[ 355, {r0}, "Error", "parsing"]; Throw[ $Failed])
		}
	]
defaultBinderParser[ e_] :=
	(sowMessage[ 354, {e}, "Error", "parsing"]; Throw[ $Failed])

normalizeRange[ r_RNG$, form_] :=
	Module[ {allNames = Cases[ {r, form}, v:VAR$[ __] :> SymbolName[ varName[ v]], Infinity, Heads -> True],
			fv = {}, rename = False},
	With[ {new =
			Replace[
				List @@ r,
				_[ x_, rest___] :>
					If[ MemberQ[ (fv = Join[ fv, freeVariables[ Hold[ rest]]]), x],
						rename = True;
						With[ {n = varName[ x]},
							With[ {tmp = freshName[ SymbolName[ n], allNames]},
								AppendTo[ allNames, tmp];
								renameVar[ x, Symbol[ Context[ n] <> tmp]]
							]
						],
					(*else*)
						AppendTo[ fv, x];
						x
					],
				{1}
			]},
		If[ TrueQ[ rename],
			renameBound[ {r, form}, new, "checkTypes" -> False, "postprocessing" -> Identity],
		(*else*)
			{r, form}
		]
	]]
(* This variant immediately constructs type ranges, returning a list of conditions instead of the given single condition. *)
normalizeRange[ r0_RNG$, cond0_List, form0_] :=
	Replace[ normalizeRange[ r0, {cond0, form0}],
		{r_, {cond_, form_}} :>
			With[ {newCond =
					Map[
						With[ {rc = Theorema`Language`FormulaManipulation`Private`singleRngToCondition[ #]},
							Switch[ rc,
								_List,
								rc,
								_,
								sowMessage[ 356, {#}, "Error", "parsing"];
								Throw[ $Failed]
							]
						]&,
						List @@ r
					]},
				{
					Map[ RNG$[ TYPRNG$ @@ #]&, rngVarsTypes[ r]],
					Which[
						cond === {},
						newCond,
					(*else*)
						newCond === {},
						sowMessage[ 357, {}, "Error", "parsing"];
						Throw[ $Failed],
					(*else*)
						True,
						MapAt[ Join[ #, cond]&, newCond, -1]
					],
					form
				}
			]
	]

rngVarsTypes[ r_RNG$] :=
	rngVarsTypes /@ (List @@ r)
rngVarsTypes[ h_[ v_, t_, ___]] :=
	{varToSymbol[ v], If[ MatchQ[ h, $typeHeads], t, TVAR$[]]}
rngVarsTypes[ _[ v_]] :=
	{varToSymbol[ v], TVAR$[]}

binaryApply[ h_, a_] :=
	binaryApply[ h, a, Right]
binaryApply[ h1_, a_, h2_] :=
	binaryApply[ h1, a, h2, True]
binaryApply[ h_, _[], _, _] :=
	h
binaryApply[ h_, _[ a_], _, _] :=
	h[ a]
binaryApply[ h_, _[ a_, b__], Left, curried_] :=
	With[ {fun = If[ TrueQ[ curried], h[ #1][ #2]&, h]},
		Fold[ fun, a, {b}]
	]
binaryApply[ h_, _[ a__, b_], Right, curried_] :=
	With[ {fun = If[ TrueQ[ curried], h[ #2][ #1]&, h[ #2, #1]&]},
		Fold[ fun, b, Reverse[ {a}]]
	]
binaryApply[ h1_, _[ a_, b___, c_], h2_, curried_] :=
	With[ {fun = If[ TrueQ[ curried], h1[ #1][ #2]&, h1]},
		h2 @@ MapThread[ fun, {{a, b}, {b, c}}]
	]

(* The reason why we have 'recursiveApply' is that it appears on the RHS of parsing rules, and that makes it easy to
	determine whether a parsing rules originates from "recursive" notation or not. *)
recursiveApply[ args_List, base_, constr_] :=
	Fold[ constr, base, args]


(* ::Section:: *)
(*Printing*)

theoremaInputDisplay[ expr_] :=
	theoremaDisplay[ expr] //. (FormBox|StyleBox)[ x_, _] :> x


(* ::Subsection:: *)
(*Types*)

printType[ typ_] := applyRuleBlocks[ typ //. $abbrevTypeFoldingRules //. $binaryTypePrintingRules, $typePrintingRules] /. FUN$TM -> Rule$TM

registerFullTypeDisplayRules[ args___] :=
	(
		$fullTypeDisplayMapping =
			Association[
				LongRightArrow$TM -> {{}, If[ #1 === {}, FUN$TM[ FORM$TM, FUN$TM[ FORM$TM, FORM$TM]], #1]&},
				EqualTilde$TM -> {{}, Replace[ #1, {a_} :> FUN$TM[ a, FUN$TM[ a, FORM$TM]]]&},
				CoreAll$TM -> {{}, FUN$TM[ FUN$TM @@ Append[ #1, FORM$TM], FORM$TM]&}
			];
		addFullTypeDisplayRules[ args]
	)

addFullTypeDisplayRules[ args___] := addToMapping[ $fullTypeDisplayMapping, args]

(* 'getFullTypeForDisplay[ t]' tries to obtain the full type of the potentially short type 't'. *)
getFullTypeForDisplay[ TYPED$[ term_List, typ_List]] :=
	getFullTypeForDisplay[ term, typ]
getFullTypeForDisplay[ TYPED$[ term_, typ_List]] :=
	getFullTypeForDisplay[ {term}, typ]
getFullTypeForDisplay[ {term_, args___}, typ_List] :=
	With[ {fun = lookupCurrentValue @@ Lookup[ $fullTypeDisplayMapping, term, {{}, $Failed}]},
		If[ fun === $Failed,
			Replace[ getFullTypeNoSorts[ typ, HoldComplete[ term], getAllConstDecls[ True]], $Failed -> typ],
		(*else*)
			fun[ typ, args]
		]
	]
getFullTypeForDisplay[ TYPED$[ ___, t_]] := t

prepareTypesForDisplay[ expr_, _, _, True] :=
	prepareTypesForDisplay[ expr, Position[ expr, TYPED$[ _, __]]]
prepareTypesForDisplay[ expr_, tfix_, tvar_, _] :=
	With[ {defSort = getDefaultSort[], defInputSort = getCoreVar[ $defaultInputSort]},
		ReplacePart[
			expr,
			If[ defSort === defInputSort,
				Cases[
					Join[ Normal[ tfix], Normal[ tvar]],
					(v_ -> s:Except[ defInputSort]) :> With[ {p = Position[ expr, v]}, Last[ p] -> MessageName$TM[ v, Replace[ s, {x_} :> x]] /; p =!= {}]
				],
			(*else*)
				Cases[
					Join[
						Map[ #[[1, 1]] -> {Last[ #], Lookup[ tfix, Last[ #], defSort]} &, getPositions[ expr, _TFIX$]],
						Map[ #[[1, 1]] -> {Last[ #], Lookup[ tvar, Last[ #], defSort]} &, getPositions[ expr, _TVAR$]]
					],
					(p_ -> {v_, s:Except[ defInputSort]}) :> (p -> MessageName$TM[ v, Replace[ s, {x_} :> x]])
				]
			]
		]
	]
prepareTypesForDisplay[ expr_, tpos_List] :=
	MapAt[
		Replace[ #,
			{
				TYPED$[ e_, t_, e0_] :>
					TAG$[
						e,
						TYPED$[ e0, t]
					],
				TYPED$[ e_, t_List] :>
					TAG$[
						e,
						TYPED$[ e, t]
					],
				TYPED$[ e_, t_] :>
					TAG$[
						e,
						TYPED$[ t]	(* Storing symbols and integers should be more efficient than storing strings (in terms of memory consumption). *)
					]
			}
		]&,
		expr,
		tpos
	]

addTypeQuotation[ typ_] :=
	RowBox[ {TagBox[ "\:27ec", Identity, SyntaxForm -> "("], typ, TagBox[ "\:27ed", Identity, SyntaxForm -> ")"]}]

SetAttributes[ makeTypeBoxes, HoldAllComplete];
makeTypeBoxes[ typ_] :=
	makeTypeBoxes[ typ, System`TheoremaForm]
makeTypeBoxes[ typ_, frm_] :=
	With[ {tbox = MakeBoxes @@ Append[ printType[ HoldComplete[ typ]], frm]},
	With[ {tbox0 = NestWhile[ First, tbox, MatchQ[ #, _StyleBox|_FormBox|_TagBox]&]},
		If[ StringQ[ tbox0] || MatchQ[ tbox0, RowBox[ {_?Theorema`Language`Syntax`Private`isLeftDelimiter, ___, _?Theorema`Language`Syntax`Private`isRightDelimiter}]],
			tbox,
		(*else*)
			(* We add parentheses even if 'tbox' is something like 'RowBox[ {_, "[", ___, "]"}]',
				because "::" has a higher precedence than "[". *)
			RowBox[ {TagBox[ "(", "AutoParentheses"], tbox, TagBox[ ")", "AutoParentheses"]}]
		]
	]]

MakeBoxes[ (TAG$|Theorema`Computation`Language`TAG$)[ Theorema`Knowledge`TYPE$TM|Theorema`Computation`Knowledge`TYPE$,
				tag:((TYPED$|Theorema`Computation`Language`TYPED$)[ Theorema`Knowledge`TYPE$TM|Theorema`Computation`Knowledge`TYPE$TM, {_}])], System`TheoremaForm] :=
	MakeBoxes[ tag, System`TheoremaForm]
MakeBoxes[ (TYPED$|Theorema`Computation`Language`TYPED$)[ Theorema`Knowledge`TYPE$TM|Theorema`Computation`Knowledge`TYPE$, {typ_}], System`TheoremaForm] :=
	addTypeQuotation[ MakeBoxes @@ Append[ printType[ HoldComplete[ typ]], System`TheoremaForm]]	(* We don't use 'makeTypeBoxes', because no parentheses shall be added. *)

MakeBoxes[ (TYPED$|Theorema`Computation`Language`TYPED$)[ term_, typ0_], System`TheoremaForm] :=
	With[ {typ = getFullTypeForDisplay[ TYPED$[ term, typ0]]},
		RowBox[ {MakeBoxes[ term, System`TheoremaForm], "::", makeTypeBoxes[ typ, System`TheoremaForm]}]
	]

Scan[
	Function[ h,
		MakeBoxes[ h[], System`TheoremaForm] :=
			StyleBox[ "_", "ExpressionVariable"];
		MakeBoxes[ h[ s_String], System`TheoremaForm] :=
			StyleBox[ SuperscriptBox[ s, "*"], "ExpressionVariable"];
		MakeBoxes[ h[ i_Integer], System`TheoremaForm] :=
			With[ {str = ToString[ i]},
				MakeBoxes[ h[ str], System`TheoremaForm]
			];
		MakeBoxes[ h[ 0, s_Symbol], System`TheoremaForm] :=
			With[ {str = tmaSymbolToString[ s]},
				MakeBoxes[ h[ str], System`TheoremaForm]
			];
		MakeBoxes[ h[ i_Integer, s_Symbol], System`TheoremaForm] :=
			With[ {str = tmaSymbolToString[ s] <> "\[Breve]" <> ToString[ i]},
				MakeBoxes[ h[ str], System`TheoremaForm]
			]
	],
	{TVAR$, Theorema`Computation`Language`TVAR$}
]

Scan[
	Function[ h,
		MakeBoxes[ h[ s_Symbol, 0], System`TheoremaForm] :=
			StyleBox[ tmaSymbolToString[ s] <> "_", "ExpressionABF"];
		MakeBoxes[ h[ s_Symbol, i_Integer], System`TheoremaForm] :=
			StyleBox[ SubscriptBox[ tmaSymbolToString[ s] <> "_", ToString[ i]], "ExpressionABF"]
	],
	{TFIX$, Theorema`Computation`Language`TFIX$}
]


(* ::Subsection:: *)
(*Terms*)

printCoreTerm[ STERM$[ term_, tfix_, tvar_]] :=
	printCoreTerm[ term, {{}, {}, tfix, tvar}]
printCoreTerm[ STERM$[ term_, termTFix_, termTVar_], {fix_List, meta_List, tfix_, tvar_}, pretty___] :=
	printCoreTerm[ term, {fix, meta, Join[ tfix, termTFix], Join[ tvar, termTVar]}, pretty]	(* keys might not be disjoint, but then 'termT...' simply replaces 't...' *)
printCoreTerm[ term_] :=
	printCoreTerm[ term, {{}, {}, Association[], Association[]}]
printCoreTerm[ term_, data_] :=
	printCoreTerm[ term, data, True]
printCoreTerm[ term_, {fix_List, meta_List, tfix_, tvar_}, pretty_] :=
	Catch[
		Block[ {$printingCtxt = {tfix, tvar}},
		With[ {out0 =
				applyRuleBlocks[
					uncurryTerm[
						applyRuleBlocks[
							fromDB[
								If[ TrueQ[ pretty],
									applyRuleBlocks[
										(* It does not harm to remove 'TrueQ$TM' on output. *)
										term //. $abbrevTermFoldingRules //. TYPED$[ TrueQ$TM, {}][ x_] :> x //.
										{
											q:($condDefaultBinderPrinting[ _][ _]) :> printCoreBinder[ q, pretty],
											q:($uncondDefaultBinderPrinting[ _]) :> printCoreBinder[ q, pretty]
										} //.
										TYPED$[ CoreAll$TM, typ1_List][ r1_RNG$, TYPED$[ CoreAll$TM, typ2_List][ r2_RNG$, f_]] :>
											TYPED$[ CoreAll$TM, Join[ typ1, typ2]][ Join[ r1, r2], f],
										$curriedDBPrintingRules
									],
								(*else*)
									applyRuleBlocks[
										term //. $abbrevTermFoldingRules /. TYPED$[ x_, {}] :> x //.
										{
											q:($condDefaultBinderInputPrinting[ _][ _]) :> printCoreBinder[ q, pretty],
											q:($uncondDefaultBinderInputPrinting[ _]) :> printCoreBinder[ q, pretty]
										} //.
										CoreAll$TM[ r1_RNG$, CoreAll$TM[ r2_RNG$, f_]] :> CoreAll$TM[ Join[ r1, r2], f],
										$curriedDBInputPrintingRules
									]
								] //.
								q:((TYPED$[ _?isQuantifier, _]|_?isQuantifier)[ _][ _]) :> printCoreBinder[ q, pretty] //.
								If[ TrueQ[ pretty],
									Lambda$TM[ RNG$[ TYPRNG$[ v_Symbol, typ_]], t_] :>
										Lambda$TM[ RNG$[ SIMPRNG$[ TYPED$[ VAR$[ v], typ]]], True, mapVarsSlow[ t, (#2 === #3)&, TYPED$[ VAR$[ #3], typ]&]],
								(*else*)
									Lambda$TM[ RNG$[ TYPRNG$[ v_Symbol, typ_]], t_] :>
										Lambda$TM[ RNG$[ If[ typ === TVAR$[], SIMPRNG$[ VAR$[ v]], TYPRNG$[ VAR$[ v], typ]]], True, t]
								] //.
								Lambda$TM[ r1_RNG$, True, Lambda$TM[ r2_RNG$, True, t_]] :> Lambda$TM[ Join[ r1, r2], True, t] //.
								If[ TrueQ[ pretty],
									$binaryTermPrintingRules,
								(*else*)
									$binaryTermInputPrintingRules
								]
							],
							If[ TrueQ[ pretty], $curriedNominalPrintingRules, $curriedNominalInputPrintingRules]
						]
					] /.
					{
						f_FIX$ :>
							With[ {typ = Replace[ f, fix]},
								TYPED$[ f, typ] /; !MatchQ[ typ, _FIX$]
							],
						m:META$[ _, _] :>
							With[ {new = Append[ m, {}]},
								Replace[ Replace[ m, meta],
									{
										{typ_, _} :> TYPED$[ new, typ],
										_ -> new
									}
								]
							],
						CoreImplies$TM -> LongRightArrow$TM,
						CoreEqual$TM -> EqualTilde$TM
					},
					If[ TrueQ[ pretty], $uncurriedNominalPrintingRules, $uncurriedNominalInputPrintingRules]
				]},
			prepareTypesForDisplay[ out0, tfix, tvar, pretty]
		]]
	]

uncurryTerm[ term_] :=
	(
		term //. h_[ a1:Except[ _RNG$], a2___][ b1:Except[ _RNG$], b2___] :>
					With[ {ar = getDeclaredArity[ h, {a1, a2}]},
						If[ ar === Infinity || ar < 0,
							h[ a1, a2, b1, b2],
						(*else*)
							With[ {lb = Length[ {b2}]},
								If[ 0 < ar <= lb,
									Drop[ Join[ h[ a1, a2], Take[ h[ b1, b2], ar]][ b2], ar - 1],
								(*else*)
									h[ a1, a2, b1, b2]
								]
							]
						] /; ar =!= 0
					]
	)

(* unconditional, pretty *)
printCoreBinder[ q_[ Lambda$TM[ RNG$[ TYPRNG$[ x_Symbol, typ_]], f_]], True] :=
	q[ RNG$[ SIMPRNG$[ TYPED$[ VAR$[ x], typ]]], mapVarsSlow[ f, (#2 === #3)&, TYPED$[ VAR$[ #3], typ]&]]
printCoreBinder[ (q:TYPED$[ _, {typ_, ___}])[ f_], True] :=
	q[ RNG$[ SIMPRNG$[ TYPED$[ Null, typ]]], mapFreeVarsSlow[ f, VAR$[ #3 + 1]&][ TYPED$[ VAR$[ 0], typ]]]
printCoreBinder[ (q:TYPED$[ _, _])[ f_], True] :=
	q[ RNG$[ SIMPRNG$[ Null]], mapFreeVarsSlow[ f, VAR$[ #3 + 1]&][ VAR$[ 0]]]

(* unconditional, input *)
printCoreBinder[ TYPED$[ q_, _][ t_Lambda$TM], False] :=
	printCoreBinder[ q[ t], False]
printCoreBinder[ q_[ Lambda$TM[ RNG$[ TYPRNG$[ x_Symbol, typ_]], f_]], False] :=
	q[ RNG$[ If[ typ === TVAR$[], SIMPRNG$[ VAR$[ x]], TYPRNG$[ VAR$[ x], typ]]], f]
printCoreBinder[ (TYPED$[ q_, {typ_, ___}])[ f_], False] :=
	q[ RNG$[ TYPRNG$[ Null, typ]], mapFreeVarsSlow[ f, VAR$[ #3 + 1]&][ VAR$[ 0]]]
printCoreBinder[ q_[ f_], False] :=
	q[ RNG$[ SIMPRNG$[ Null]], mapFreeVarsSlow[ f, VAR$[ #3 + 1]&][ VAR$[ 0]]]

(* conditional, pretty *)
printCoreBinder[ q_[ Lambda$TM[ RNG$[ TYPRNG$[ y_Symbol, typ1_]], c_]][ Lambda$TM[ RNG$[ TYPRNG$[ x_Symbol, typ2_]], f_]], True] :=
	With[ {rng = getCoreVar[ $rangeReconstructionF][ c, False]},
		q @@ Prepend[ mapVarsSlow[ {Last[ rng], f}, (#2 === #3)&, TYPED$[ VAR$[ #3], typ2]&], RNG$[ printCoreRange[ First[ rng], {x, y}, {typ2, typ1}, True]]]
	]
printCoreBinder[ q_[ c:Except[ _Lambda$TM]][ Lambda$TM[ RNG$[ TYPRNG$[ x_Symbol, typ_]], f_]], True] :=
	q[ RNG$[ PREDRNG$[ TYPED$[ VAR$[ x], typ], c]], True, f]		(* this might look weird if 'c' is a compound expression, e.g. 'Less$TM[ 1]' *)
printCoreBinder[ q_[ Lambda$TM[ RNG$[ TYPRNG$[ x_Symbol, typ_]], c_]][ f:Except[ _Lambda$TM]], True] :=
	With[ {rng = getCoreVar[ $rangeReconstructionF][ c, False]},
		q[
			RNG$[ printCoreRange[ First[ rng], {x}, {typ}, True]],
			mapVarsSlow[ Last[ rng], (#2 === #3)&, TYPED$[ VAR$[ #3], typ]&],
			mapFreeVarsSlow[ f, VAR$[ #3 + 1]&][ TYPED$[ VAR$[ 0], typ]]
		]
	]
printCoreBinder[ (q:TYPED$[ _, {typ_, ___}])[ c:Except[ _Lambda$TM]][ f:Except[ _Lambda$TM]], True] :=
	q[ RNG$[ PREDRNG$[ TYPED$[ Null, typ], c]], True, mapFreeVarsSlow[ f, VAR$[ #3 + 1]&][ TYPED$[ VAR$[ 0], typ]]]
printCoreBinder[ q_[ c_][ f_], True] :=
	q[ RNG$[ PREDRNG$[ Null, c]], True, mapFreeVarsSlow[ f, VAR$[ #3 + 1]&][ VAR$[ 0]]]

(* conditional, input *)
printCoreBinder[ q_[ Lambda$TM[ RNG$[ TYPRNG$[ y_Symbol, typ1_]], c_]][ Lambda$TM[ RNG$[ TYPRNG$[ x_Symbol, typ2_]], f_]], False] :=
	With[ {rng = getCoreVar[ $rangeReconstructionF][ c, MemberQ[ {typ1, typ2}, Except[ TVAR$[]]]]},
		q[ RNG$[ printCoreRange[ First[ rng], {x, y}, {typ2, typ1}, False]], Last[ rng], f]
	]
printCoreBinder[ q_[ c:Except[ _Lambda$TM]][ Lambda$TM[ RNG$[ TYPRNG$[ x_Symbol, typ_]], f_]], False] :=
	q[ RNG$[ If[ typ === TVAR$[], PREDRNG$[ VAR$[ x], c], Theorema`Knowledge`TPREDRNG$[ VAR$[ x], typ, c]]], True, f]
printCoreBinder[ q_[ Lambda$TM[ RNG$[ TYPRNG$[ x_Symbol, typ_]], c_]][ f:Except[ _Lambda$TM]], False] :=
	With[ {rng = getCoreVar[ $rangeReconstructionF][ c, typ =!= TVAR$[]]},
		q[
			RNG$[ printCoreRange[ First[ rng], {x}, {typ}, False]],
			Last[ rng],
			mapFreeVarsSlow[ f, VAR$[ #3 + 1]&][ VAR$[ 0]]
		]
	]
printCoreBinder[ (TYPED$[ q_, {typ_, ___}])[ c:Except[ _Lambda$TM]][ f:Except[ _Lambda$TM]], False] :=
	q[ RNG$[ Theorema`Knowledge`TPREDRNG$[ Null, typ, c]], True, mapFreeVarsSlow[ f, VAR$[ #3 + 1]&][ VAR$[ 0]]]
printCoreBinder[ q_[ c_][ f_], False] :=
	q[ RNG$[ PREDRNG$[ Null, c]], True, mapFreeVarsSlow[ f, VAR$[ #3 + 1]&][ VAR$[ 0]]]

chooseRangeVar[ vars_List] :=
	FirstCase[ vars, Except[ Null, _Symbol], Null]

printCoreRange[ h_[ _, rest___], vars_List,  {typ_, ___}, True] :=
	h[ TYPED$[ VAR$[ chooseRangeVar[ vars]], typ], rest]
printCoreRange[ h_[ _, Null, rest___], vars_List, {___, typ:Except[ TVAR$[]], ___}, False] :=
	h[ VAR$[ chooseRangeVar[ vars]], typ, rest]
printCoreRange[ h_[ _, rest___], vars_List, _, False] :=
	h[ VAR$[ chooseRangeVar[ vars]], rest]

defaultReconstructRange[ P0_[ VAR$[ 0]], tr_] /; !containsVar[ P0, (#2 === #3)&] :=
	With[ {P = mapFreeVarsSlow[ P0, VAR$[ #3 - 1]&]},
		{
			If[ TrueQ[ tr],
				Theorema`Knowledge`TPREDRNG$[ Null, Null, P],
			(*else*)
				PREDRNG$[ Null, P]
			],
			True
		}
	]
defaultReconstructRange[ c_, tr_] :=
	If[ TrueQ[ tr],
		{TYPRNG$[ Null, Null], c},
	(*else*)
		{SIMPRNG$[ Null], c}
	]

fromDB[ expr_] :=
	Catch[
		ReplacePart[
			expr,
			reap[ fromDB[ expr, {}, {}]]
		]
	]
fromDB[ VAR$[ i_], abs_, pos_] :=
	If[ IntegerQ[ i] && i >= 0 && i < Length[ abs],
		Sow[ pos -> VAR$[ abs[[i + 1]]]],
	(*else*)
		Throw[ $Failed]
	]
fromDB[ q_[ rng_RNG$, e__], abs_, pos_] :=
	Module[ {newAbs = abs, i = 0},
		Scan[
			(
				++i;
				Replace[
					#,
					{
						TYPRNG$[ v_Symbol, _] :>
							With[ {new = If[ v === Null, Symbol[ "Theorema`Knowledge`" <> freshName[ "VAR$x$TM", newAbs]], freshName[ v, newAbs]]},
								If[ new =!= v, Sow[ Join[ pos, {1, i, 1}] -> VAR$[ new]]];
								PrependTo[ newAbs, new]
							],
						_[ TYPED$[ Null|VAR$[], _], rest___] :>
							With[ {new = Symbol[ "Theorema`Knowledge`" <> freshName[ "VAR$x$TM", newAbs]]},
								Sow[ Join[ pos, {1, i, 1, 1}] -> VAR$[ new]];
								fromDB[ {Null, rest}, newAbs, Join[ pos, {1, i}]];
								PrependTo[ newAbs, new]
							],
						_[ TYPED$[ VAR$[ v_Symbol, r___], _], rest___] :>
							With[ {new = If[ v === Null, Symbol[ "Theorema`Knowledge`" <> freshName[ "VAR$x$TM", newAbs]], freshName[ v, newAbs]]},
								If[ new =!= v || {r} =!= {}, Sow[ Join[ pos, {1, i, 1, 1}] -> VAR$[ new]]];
								fromDB[ {Null, rest}, newAbs, Join[ pos, {1, i}]];
								PrependTo[ newAbs, new]
							],
						_[ Null|VAR$[], rest___] :>
							With[ {new = Symbol[ "Theorema`Knowledge`" <> freshName[ "VAR$x$TM", newAbs]]},
								Sow[ Join[ pos, {1, i, 1}] -> VAR$[ new]];
								fromDB[ {Null, rest}, newAbs, Join[ pos, {1, i}]];
								PrependTo[ newAbs, new]
							],
						_[ VAR$[ v_Symbol, r___], rest___] :>
							With[ {new = If[ v === Null, Symbol[ "Theorema`Knowledge`" <> freshName[ "VAR$x$TM", newAbs]], freshName[ v, newAbs]]},
								If[ new =!= v || {r} =!= {}, Sow[ Join[ pos, {1, i, 1}] -> VAR$[ new]]];
								fromDB[ {Null, rest}, newAbs, Join[ pos, {1, i}]];
								PrependTo[ newAbs, new]
							]
					}
				]
			)&,
			List @@ rng
		];
		fromDB[ q, abs, Append[ pos, 0]];
		fromDB[ {Null, e}, newAbs, pos]
	]
fromDB[ TYPED$[ e_, _], abs_, pos_] := fromDB[ e, abs, Append[ pos, 1]]
fromDB[ f_[ a___], abs_, pos_] :=
	Module[ {i = -1},
		Scan[
			fromDB[ #, abs, Append[ pos, ++i]]&,
			{f, a}
		]
	]
fromDB[ ___] := Null

getDeclaredArity[ TYPED$[ s_, __], a_] :=
	getDeclaredArity[ s, a]
getDeclaredArity[ _VAR$|_FIX$|_META$, _] :=
	Infinity
getDeclaredArity[ s_, a_] :=
	If[ MemberQ[ {TYPED$, VAR$, FIX$, META$, TVAR$, TFIX$}, s] || isQuantifier[ s] || MatchQ[ s, $curriedHeads],
		0,
	(*else*)
		Replace[ getDeclaredType[ HoldComplete[ s], 0, getAllConstDecls[ True]],
			{
				{typ_, _List, _, _, _} :> Last[ resultType[ typ, 0]] - Length[ a],
				_ -> Infinity
			}
		]
	]


(* ::Section:: *)
(*Misc*)

(* 'mapVarsSlow' and 'mapFreeVarsSlow' cater for multi-ranges and ternary binders (i.e. binders with conditions) that
	may pop up when printing expressions, in contrast to 'mapVars' and 'mapFreeVars'. But of course, 'mapVarsSlow' also
	works on standard Core expressions. *)
mapVarsSlow[ pred_, fun_][ expr_] :=
	mapVarsSlow[ expr, pred, fun]
mapVarsSlow[ l_List, pred_, fun_] :=
	Map[ mapVarsSlow[ pred, fun], l]
mapVarsSlow[ expr_, pred_, fun_] :=
	With[ {pos = Position[ expr, VAR$[ _]]},
		ReplacePart[ expr,
			MapThread[
				Function[ {p, v},
					With[ {abs = countAbstractionsFromPosSlow[ p, expr]},
						If[ TrueQ[ pred[ p, abs, First[ v]]], {p} -> fun[ p, abs, First[ v]], Sequence @@ {}]
					]
				],
				{pos, Extract[ expr, pos]}
			]
		]
	]

mapFreeVarsSlow[ expr_, fun_] := mapVarsSlow[ expr, (#3 >= #2)&, fun]

countAbstractionsFromPosSlow[ p_List, expr_] :=
	Total[ Map[ Length[ First[ extractSingle[ expr, Take[ p, First[ #] - 1]]]]&, Position[ p, Except[ 0|1], {1}, Heads -> False]]]

(* 'applyRuleBlocks[ expr, rules]' applies lists of lists of transformation rules to 'expr', each list ("block") one after the other.
	By default, rules are applied repeatedly, but this can be changed. *)
applyRuleBlocks[ expr_, {}] := expr
applyRuleBlocks[ expr0_, blocks:{_List, ___}] :=
	Module[ {expr = expr0},
		Scan[
			Function[ block,
				Replace[ block,
					{
						{Replace, rules__} :> (expr = Replace[ expr, {rules}]),
						{All, rules__} :> (expr = expr /. {rules}),
						{Repeated, rules__} :> (expr = expr //. {rules}),
						rules_List :> (expr = expr //. rules)
					}
				]
			],
			blocks
		];
		expr
	]
applyRuleBlocks[ expr_, rules_List] :=
	expr //. rules
applyRuleBlocks[ rules_List][ expr_] :=
	applyRuleBlocks[ expr, rules]

SetAttributes[ addToRuleBlocks, HoldFirst];
addToRuleBlocks[ rb_Symbol, new_List, app_, i_Integer, mode_] :=
	With[ {j =
			Which[
				i === 0,
				Return[ $Failed],
			(*else*)
				Positive[ i],
				With[ {pos = Position[ rb, {app, ___}, {1}, i, Heads -> False]},
					If[ Length[ pos] === i,
						pos[[-1, 1]],
					(*else*)
						Return[ $Failed]
					]
				],
			(*else*)
				True,
				With[ {pos = Position[ rb, {app, ___}, {1}, Heads -> False]},
					If[ NonNegative[ Length[ pos] + i],
						pos[[i, 1]],
					(*else*)
						Return[ $Failed]
					]
				]
			]},
		rb =
			Switch[ mode,
				Before,
				Insert[ rb, Prepend[ new, app], j],
				After,
				Insert[ rb, Prepend[ new, app], j + 1],
				Prepend,
				MapAt[ Join[ {app}, new, Rest[ #]]&, rb, j],
				_,
				MapAt[ Join[ #, new]&, rb, j]
			]
	]
addToRuleBlocks[ rb_Symbol, new_List, app_, Front] :=
	PrependTo[ rb, Prepend[ new, app]]
addToRuleBlocks[ rb_Symbol, new_List, app_, Back] :=
	AppendTo[ rb, Prepend[ new, app]]
addToRuleBlocks[ rb_Symbol, rl:(_Rule|_RuleDelayed), rest___] :=
	addToRuleBlocks[ rb, {rl}, rest]

(* 'dropTypesFromPrintingRule[ rl]' returns a rule with the same LHS as 'rl', and optionally sows another rule with
	distinct LHS. *)
dropTypesFromPrintingRule[ rl:((h:(Rule|RuleDelayed))[ l0_, r0_])] :=
	With[ {l = Hold[ l0], r = Hold[ r0]},
		Which[
			FreeQ[ l, TYPED$[ _, ___]],
			If[ FreeQ[ r, TYPED$[ _, ___]],
				rl,
			(*else*)
				h @@ Join[ l, r //. TYPED$[ x_, ___] :> x]
			],
		(*else*)
			FreeQ[ l, TYPED$[ _, t_, ___] /;
				MemberQ[
					t //. (_?(# === Pattern &))[ _, p_] :> p,
					Except[ List|Blank|BlankSequence|BlankNullSequence|PatternSequence|Repeated|RepeatedNull|Except, _?AtomQ],
					{0, Infinity},
					Heads -> True
				]
			],
			Sow[ rl //. TYPED$[ x_, ___] :> x];
			rl,
		(*else*)
			FreeQ[ l, TYPED$[ _, {}, ___]],
			h @@ Join[ l, r //. TYPED$[ x_, {}, ___] :> x],
		(*else*)
			True,
			Sow[ rl //. TYPED$[ x_, {}, ___] :> x];
			rl
		]
	]
dropTypesFromPrintingRule[ arg:(Replace|All|Repeated)] :=
	(Sow[ arg]; arg)

translatePrettyPrintingRulesToInputPrintingRules[ rules_List] :=
	Replace[ rules,
		l_List :>
			Replace[ Reap[ dropTypesFromPrintingRule /@ l],
				{
					{old_List, {}|{{Replace|All|Repeated}}} :> old,
					{old_List, {new_List}} :> Sequence[ old, new]
				}
			],
		{1}
	]

(* ATTENTION! If 'addScopeCondition' is changed such that after applying it to a rule the LHS of that rule might not
	necessarily be of the form 'Condition[ _, _]', then function 'binaryFoldingPatt' in "Toplevel.m" must be
	changed accordingly. *)
addScopeCondition[ thy_String] :=
	addScopeCondition[ thy, {}]
addScopeCondition[ thy_String, excpt_List][ rule_] :=
	addScopeCondition[ rule, thy, excpt, True]
addScopeCondition[ rule:(_Rule|_RuleDelayed), thy_] :=
	addScopeCondition[ rule, thy, {}, True]
addScopeCondition[ rule_, thy_, excpt_] :=
	addScopeCondition[ rule, thy, excpt, True]
addScopeCondition[ (h:(Rule|RuleDelayed))[ l_, r_], thy_String, excpt_List, True] :=
	h[ addScopeCondition[ l, thy, excpt, False], r]	(* if 'l' is a rule itself, we must not dig further into the LHS of 'l' *)
addScopeCondition[ expr_, thy_String, excpt_List, _] :=
	Replace[ expr,
		{
			h2_[ patt_, cond_] /; h2 === Condition :>
				With[ {pos = FirstPosition[ Hold[ Null, cond], HoldPattern[ inTheoryScopeQ[ _, ___List]], $Failed]},
					If[ ListQ[ pos],
						Condition @@ ReplacePart[
							Hold[ patt, cond],
							If[ excpt === {}, pos :> inTheoryScopeQ[ thy], pos :> inTheoryScopeQ[ thy, excpt]]
						],
					(*else*)
						Condition @@ If[ excpt === {}, Hold[ patt, inTheoryScopeQ[ thy] && cond], Hold[ patt, inTheoryScopeQ[ thy, excpt] && cond]]
					]
				],
			_ :> Condition @@ If[ excpt === {}, Hold[ expr, inTheoryScopeQ[ thy]], Hold[ expr, inTheoryScopeQ[ thy, excpt]]]
		}
	]

addExcptToScopeCondition[ excpt_List][ rule_] :=
	addExcptToScopeCondition[ rule, excpt, True]
addExcptToScopeCondition[ rule_, {}, ___] := rule
addExcptToScopeCondition[ rule:(_Rule|_RuleDelayed), excpt_] :=
	addExcptToScopeCondition[ rule, excpt, True]
addExcptToScopeCondition[ (h:(Rule|RuleDelayed))[ l_, r_], excpt_List, True] :=
	With[ {aux = {addExcptToScopeCondition[ l, excpt, False]}},	(* if 'l' is a rule itself, we must not dig further into the LHS of 'l' *)
		If[ aux === {},
			Sequence @@ {},
		(*else*)
			h[ First[ aux], r]
		]
	]
addExcptToScopeCondition[ expr_, excpt_List, _] :=
	Replace[ expr,
		{
			h2_[ patt_, cond_] /; h2 === Condition :>
				With[ {pos = FirstPosition[ Hold[ Null, cond], HoldPattern[ inTheoryScopeQ[ _, ___List]], $Failed]},
					If[ ListQ[ pos],
						Replace[ Extract[ Hold[ Null, cond], pos, Hold],
							{
								Hold[ inTheoryScopeQ[ thy_]] :>
									If[ MemberQ[ excpt, thy],
										Sequence @@ {},
									(*else*)
										If[ thy =!= "$Core", Sow[ thy, "Theory"]];
										Condition @@ ReplacePart[
											Hold[ patt, cond],
											pos :> inTheoryScopeQ[ thy, excpt]
										]
									],
								Hold[ inTheoryScopeQ[ thy_, excpt0_]] :>
									If[ MemberQ[ excpt, thy],
										Sequence @@ {},
									(*else*)
										If[ thy =!= "$Core", Sow[ thy, "Theory"]];
										Condition @@ ReplacePart[
											Hold[ patt, cond],
											With[ {newExcpt = DeleteDuplicates[ Join[ excpt0, excpt]]},
												pos :> inTheoryScopeQ[ thy, newExcpt]
											]
										]
									]
							}
						],
					(*else*)
						Condition @@ Hold[ patt, inTheoryScopeQ[ "$Core", excpt] && cond]
					]
				],
			_ :> Condition @@ Hold[ expr, inTheoryScopeQ[ "$Core", excpt]]
		}
	]

remExcptFromScopeCondition[ excpt_List][ rule_] :=
	remExcptFromScopeCondition[ rule, excpt, True]
remExcptFromScopeCondition[ rule_, {}, ___] := rule
remExcptFromScopeCondition[ rule:(_Rule|_RuleDelayed), excpt_] :=
	remExcptFromScopeCondition[ rule, excpt, True]
remExcptFromScopeCondition[ (h:(Rule|RuleDelayed))[ l_, r_], excpt_List, True] :=
	h[ remExcptFromScopeCondition[ l, excpt, False], r]	(* if 'l' is a rule itself, we must not dig further into the LHS of 'l' *)
remExcptFromScopeCondition[ expr_, excpt_List, _] :=
	Replace[ expr,
		h2_[ patt_, cond_] /; h2 === Condition :>
			With[ {pos = FirstPosition[ Hold[ Null, cond], HoldPattern[ inTheoryScopeQ[ _, _List]], $Failed]},
				If[ ListQ[ pos],
					Replace[ Extract[ Hold[ Null, cond], pos, Hold],
						Hold[ inTheoryScopeQ[ thy_, excpt0_]] :>
							Condition @@ ReplacePart[
								Hold[ patt, cond],
								With[ {newExcpt = DeleteCases[ excpt0, Alternatives @@ excpt]},
									If[ newExcpt === {}, pos :> inTheoryScopeQ[ thy], pos :> inTheoryScopeQ[ thy, newExcpt]]
								]
							]
					],
				(*else*)
					expr
				]
			]
	]

SetAttributes[ addToConditionalRules, HoldFirst];
addToConditionalRules[ rules_, {}, _, _] := rules
addToConditionalRules[ rules_Symbol, add_List, thy_String, pos_List] :=
	Replace[ Reap[ MapAt[ addExcptToScopeCondition[ {thy}], rules, pos], "Theory"],
		{rules0_List, thys_List} :>
			(rules = Join[ addScopeCondition[ thy, Join @@ thys] /@ add, rules0])
	]
addToConditionalRules[ rules_Symbol, add_List, thy_String, patt_] :=
	addToConditionalRules[ rules, add, thy, Position[ rules, patt, {1}, Heads -> False]]

SetAttributes[ remFromConditionalRules, HoldFirst];
remFromConditionalRules[ rules_, _, {}] := rules
remFromConditionalRules[ rules_Symbol, thy_String, pos_List] :=
	(
		rules =
			MapAt[
				Replace[ #,
					{
						_[ h_[ _, cond_] /; h === Condition && MemberQ[ Hold[ cond], HoldPattern[ inTheoryScopeQ[ thy, ___]], Infinity], _] -> Sequence[],
						rl_ :> remExcptFromScopeCondition[ rl, thy]
					}
				]&,
				rules,
				pos
			]
	)
remFromConditionalRules[ rules_Symbol, thy_String, patt_] :=
	remFromConditionalRules[ rules, thy, Position[ rules, patt, {1}, Heads -> False]]

(* 'addToConstPattern' and 'remFromConstPattern' add and remove, respectively, constants to/from patterns of
	constants as, for instance, '$condDefaultBinderPrinting'.
	Attention! 'addToConstPattern' does not remove the given constants from theories other than the current one! *)
SetAttributes[ addToConstPattern, HoldFirst];
addToConstPattern[ patt_Symbol, {}, ___] := patt
addToConstPattern[ patt_Symbol, consts_] :=
	addToConstPattern[ patt, consts, EvalTheoryName[]]
addToConstPattern[ patt_Symbol, consts_List, thy_String] :=
	With[ {i = First[ FirstPosition[ patt, h_[ _, HoldPattern[ inTheoryScopeQ[ thy]]] /; h === Condition, {$Failed}, {1}]]},
		Which[
			IntegerQ[ i],
			patt =
				ReplacePart[
					patt,
					i ->
						Condition[
							Evaluate[
								With[ {old = patt[[i, 1]]},
									If[ Head[ old] === Alternatives,
										Join[ old, Alternatives @@ DeleteCases[ consts, old]],
									(*else*)
										With[ {new = DeleteCases[ consts, old]},
											If[ new === {},
												old,
											(*else*)
												Prepend[ Alternatives @@ new, old]
											]
										]
									]
								]
							],
							inTheoryScopeQ[ thy]
						]
				],
		(*else*)
			Length[ consts] === 1,
			PrependTo[ patt, Condition[ First[ consts], inTheoryScopeQ[ thy]]],
		(*else*)
			True,
			PrependTo[ patt, Condition[ Evaluate[ Alternatives @@ consts], inTheoryScopeQ[ thy]]]
		]
	]
addToConstPattern[ patt_Symbol, const_, thy_String] :=
	addToConstPattern[ patt, {const}, thy]

SetAttributes[ remFromConstPattern, HoldFirst];
remFromConstPattern[ patt_, {}, ___] := patt
remFromConstPattern[ patt_Symbol, consts_] :=
	remFromConstPattern[ patt, consts, EvalTheoryName[]]
remFromConstPattern[ patt_Symbol, consts_List, thy_String] :=
	With[ {cpatt = Alternatives @@ consts},
		patt =
			DeleteCases[
				Replace[
					patt,
					h2_[ l:(h_[ __]), HoldPattern[ inTheoryScopeQ[ thy]]] /; h2 === Condition && h === Alternatives :>
						With[ {new0 = DeleteCases[ l, cpatt]},
						With[ {new = If[ Length[ new0] === 1, First[ new0], new0]},
							Condition @@ Hold[ new, inTheoryScopeQ[ thy]]
						]],
					{1}
				],
				h2_[ cpatt|(h_[] /; h === Alternatives), HoldPattern[ inTheoryScopeQ[ thy]]] /; h2 === Condition
			]
	]
remFromConstPattern[ patt_Symbol, const_, thy_String] :=
	remFromConstPattern[ patt, {const}, thy]

SetAttributes[ addToMapping, HoldFirst];
addToMapping[ map_Symbol, thy_String][ new_] :=
	addToMapping[ map, new, thy]
addToMapping[ map_Symbol, new:(_List|_Association|_Rule|_RuleDelayed)] :=
	addToMapping[ map, new, EvalTheoryName[]]
addToMapping[ map_Symbol, new_Association, thy_String] :=
	addToMapping[ map, Normal[ new], thy]
addToMapping[ map_Symbol, new_List, thy_String] :=
	Scan[ addToMapping[ map, thy], new]
addToMapping[ map_Symbol, (h:(Rule|RuleDelayed))[ k_ , v_], thy_String] :=
	AssociateTo[ map,
		k ->
			Replace[ Lookup[ map, k, $Failed],
				{
					{l_List, d_} :>
						Replace[ First[ FirstPosition[ l, _[ thy, _], {0}, {1}, Heads -> False]],
							{
								0 :> {Prepend[ l, h[ thy, v]], d},
								i_Integer :> {ReplacePart[ l, i -> h[ thy, v]], d}
							}
						],
					{a_Association, d_} :> {Prepend[ a, h[ thy, v]], d},
					_ :> {{h[ thy, v]}, $Failed}
				}
			]
	]


(* ::Section:: *)
(*General Setup*)

registerQuantifier[ "CoreAll", "\[And]" -> {None, False, True, "SIMPRNG$"|"TYPRNG$"}];

registerRange[
	"TYPRNG$",
	{
		RowBox[ {RowBox[ {"(", "$var", ")"}], "::", t_}] :> {t},
		RowBox[ {"$var", "::", t_}] :> {t}
	},
	{x_, t_} :> RowBox[ {MakeBoxes[ x], "::", makeTypeBoxes[ t]}],
	{x_, t_} :> {}
];
registerRange[
	"TPREDRNG$",
	{
		RowBox[ {P_, "[", RowBox[ {RowBox[ {"(", "$var", ")"}], "::", t_}], "]"}] :> {t, P},
		RowBox[ {P_, "[", RowBox[ {"$var", "::", t_}], "]"}] :> {t, P}
	},
	{x_, t_, P_} :> RowBox[ {MakeBoxes[ P], "[", RowBox[ {MakeBoxes[ x], "::", makeTypeBoxes[ t]}], "]"}],
	{x_, _, P_} :> {P[ x]}
];
(* In contrast to set ranges and step ranges, predicate ranges make sense also in Core (no special theory is required).
	Hence, we introduce syntax of typed predicate ranges here.
	But still, Core expressions *cannot* contain any ranges other than plain type ranges! *)

Theorema`Interface`ProofScript`$formulaQuotingOperators =
	{
		{TagBox[ "\:27ec", Identity, SyntaxForm -> "("], TagBox[ "\:27ed", Identity, SyntaxForm -> ")"]} -> addTypeQuotation
	};

End[]

EndPackage[]
