(* ::Package:: *)

(* Theorema
    Copyright (C) 2017 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

BeginPackage[ "Theorema`Core`Tools`Toplevel`"]

Needs[ "Theorema`Common`"]
Needs[ "Theorema`Core`Trusted`"]
Needs[ "Theorema`Core`Tools`"]
Needs[ "Theorema`Language`"]

Begin["`Private`"]

(* ::Section:: *)
(*Default Theory Image*)

(* The default theory image is hard-coded. Change here if other theory shall become the default image. *)
$defaultTheoryImage = {Theorema`$TheoremaDirectory, "Theorema", "CoreTheories", "HOL", "HOL"};


(* ::Section:: *)
(*Global Functions*)

If[ !ListQ[ $coreContextPathStack], $coreContextPathStack = {}];

beginCore[ ctxt_String] :=
	beginCore[ ctxt, {}]
beginCore[ ctxt_String, needs_List] :=
	beginCore[ ctxt, needs, True]
beginCore[ ctxt_String, needs_List, priv_] :=
	With[ {cp = $ContextPath},
		Check[
			BeginPackage[ ctxt],
			Throw[ $Failed]
		];
		PrependTo[ $coreContextPathStack, cp];
		SetDirectory[ Theorema`$TheoremaDirectory];
		Needs[ "Theorema`Common`"];
		Needs[ "Theorema`Core`Trusted`"];
		Needs[ "Theorema`Core`Tools`"];
		Needs[ "Theorema`Language`"];
		Scan[
			Replace[ #,
				{
					{c_String, f_String} :> Needs[ c, makeFileName[ f, "m"]],
					c_ :> Needs[ c]
				}
			]&,
			needs
		];
		ResetDirectory[];
		If[ TrueQ[ priv], Begin[ "`Private`"]];
		setTheory[ getEvaluationTheoryName[]];
	]

endCore[] :=
	(
		If[ StringMatchQ[ $Context, __ ~~ "`Private`"], End[]];
		EndPackage[];
		If[ $coreContextPathStack =!= {},
			$ContextPath = First[ $coreContextPathStack];
			$coreContextPathStack = Rest[ $coreContextPathStack]
		];
	)

getEvaluationNotebook[] :=
	If[ Head[ $evaluationNotebook] === NotebookObject,
		$evaluationNotebook,
	(*else*)
		EvaluationNotebook[]
	]

getEvaluationCell[] :=
	If[ Head[ $evaluationCell] === CellObject,
		$evaluationCell,
	(*else*)
		EvaluationCell[]
	]

getEvaluationTheoryName[] :=
	getEvaluationTheoryName[ True]
getEvaluationTheoryName[ True] :=
	Quiet[ Check[ sourceLabel[ NotebookFileName[ getEvaluationNotebook[]]], $Failed], NotebookFileName::nosv]
getEvaluationTheoryName[ False] :=
	Quiet[ Check[ NotebookFileName[ getEvaluationNotebook[]], $Failed], NotebookFileName::nosv]

(* The following symbols are merely included for the user's convenience. *)
System`BeginCore = beginCore;
System`EndCore = endCore;
(* In Theorema notebooks, *ALWAYS* use 'EvalNotebook[]' and 'EvalCell[]' instead of 'EvaluationNotebook[]' and 'EvaluationCell[]'!
	The built-in functions return wrong results when notebooks are imported. *)
System`EvalNotebook = getEvaluationNotebook;
System`EvalCell = getEvaluationCell;
System`EvalTheoryName = getEvaluationTheoryName;
System`$TmaDir = FileNameJoin[ {Theorema`$TheoremaDirectory, "Theorema"}];


(* ::Section:: *)
(*Theories*)

enterCore[] :=
	enterCore[ False, True]
enterCore[ strict_] :=
	enterCore[ strict, True]
enterCore[ strict_, True] :=
	With[ {imageBase = FileNameJoin[ $defaultTheoryImage]},
		If[ !TrueQ[ $inCore] || getTheoryCounter[] === 0 || dialogYesNo[ 217, {}],
			$inCore = False;	(* '$inCore' is set to 'False' only to ensure that no dialog window appears any more. *)
			If[ !FileExistsQ[ imageBase <> ".mx"] && dialogYesNo[ 262, {imageBase <> ".nb"}],
				
				(* The folowing must be done, since otherwise 'makeTheoryFileName' does not work: *)
				initIO[ False];
				$theoryDirs = {};	(* '$theoryDirs' is set to a proper Core variable later. *)
				
				With[ {log = FileNameJoin[ {$InitialDirectory, Last[ $defaultTheoryImage] <> "-buildlog.txt"}]},
					If[ TrueQ[ buildTheory[ Path[ imageBase <> ".nb", $OperatingSystem], log, True]],
						If[ exportImage[ Automatic, False, True] === $Failed, coreNotification[ {MSG$[ coreMessageToList, 263, {}, "Fatal"]}, True, {}]],
					(*else*)
						coreNotification[ {MSG$[ coreMessageToList, 264, {imageBase <> ".nb", log}, "Fatal"]}, True, {}]
					]
				]
			];
			$inCore = False;
			enterCore[ strict, False]
		]
	]
enterCore[ strict_, False] :=
	If[ !TrueQ[ $inCore] || getTheoryCounter[] === 0 || dialogYesNo[ 217, {}],
		$coreVars = {};
		initIO[ True];
		
		(* parsing and pretty-printing *)
		$typeParsingRules = {};
		$binaryTypeParsingRules = {};
		$typePrintingRules = {};
		$binaryTypePrintingRules = {};
		$typeHeads = Alternatives[ TYPRNG$, Theorema`Knowledge`TPREDRNG$];
		initCoreVar[ $defaultInputSort, {}];
		initCoreVar[ $termSortConstructor, Identity];
		$uncurriedNominalParsingRules = {};
		$curriedDBParsingRules = {};
		$postParsingRules = {};
		$binaryTermParsingRules = {};
		$abbrevTermUnfoldingRules = {};
		$abbrevTypeUnfoldingRules = {};
		initCoreVar[ $binderParser, defaultBinderParser];
		initCoreVar[ $coreFmlPre, {#1, #2, #3, Automatic, (#2&)}&];
		initCoreVar[ $rangeReconstructionF, defaultReconstructRange];
		$abbrevTermFoldingRules = {};
		$abbrevTypeFoldingRules = {};
		$uncurriedNominalPrintingRules = {};
		$curriedNominalPrintingRules = {};
		$curriedDBPrintingRules = {};
		$binaryTermPrintingRules = {};
		$uncurriedNominalInputPrintingRules = {};
		$curriedNominalInputPrintingRules = {};
		$curriedDBInputPrintingRules = {};
		$binaryTermInputPrintingRules = {};
		$condDefaultBinderPrinting = Alternatives[];
		$uncondDefaultBinderPrinting = Alternatives[ TYPED$[ CoreAll$TM, _]];
		$condDefaultBinderInputPrinting = Alternatives[];
		$uncondDefaultBinderInputPrinting = Alternatives[ CoreAll$TM];
		$curriedHeads = Alternatives[];
		registerFullTypeDisplayRules[];
		
		initCoreVar[ $proofLanguageTranslator, defaultPLT];
		initCoreVar[ $theoryLanguageTranslator, defaultTLT];
		registerCustomProofInterpretationRules[];
		registerCustomTheoryInterpretationRules[];
		registerProofMethodRules[];
		registerFormulaVariantRules[];
		$criticalTheoryCommands = {};
		
		initCoreVar[ $theoryDirs, {}];
		initCoreVar[ $initialProofSteps, Hold[ Theorema`CoreProvers`coreInitialProofStep[]]];
		$terminalProofSteps = {};
		$globalFactLists = Association[];
		(* In strict theory mode the only possibility for cyclic arguments is when the formula to be proved is used in
			its own proof. This, however, is ruled out even if '$ignoreScope' is set to 'False' (it actually *must* be
			ruled out by 'inScopeQ', since otherwise proofs going through in exploration mode might not go through in
			strict mode). *)
		$ignoreScope = TrueQ[ strict];	(* no Core variable, because shall not be exported to theory images *)
		$registeredProofNumbers = Association[];
		$theoremIndex = {};
		initCoreVar[ $typedefProofTransformer, Identity];
		initCoreVar[ $typedefFormulaConstructor, defaultTypedefFormulaConstructor];
		initCoreVar[ $typedefThmGenerator, addCoreTypedefThms];
		initCoreVar[ $objectValidity, TYPED$[ CoreEqual$TM, {FUN$TM[ FORM$TM, FUN$TM[ FORM$TM, FORM$TM]]}][ TYPED$[ CoreEqual$TM, {FORM$TM}]][ TYPED$[ CoreEqual$TM, {FORM$TM}]]];
		initCoreVar[ $objectValidityProof, Hold[ INF$[ "refl"]]];
		
		$abbrevTypeDecls = Association[];
		$abbrevConstDecls = Association[];
		
		(* In 'RULEC$'-structures, formulas are grouped together according to the theories they originate from and
			according to their "characteristic constants" (--> see "/Core/Tools/Reasoning.m"), which enables proof
			methods to iterate over them efficiently.
			If further rule collections are added, don't forget to restrict them in '$preProof'. *)
		(* If further levels of nesting are added to '$autoredRules', 'extractRelevantRules' in
			"/CoreProvers/CoreProofMethods/AutoReduce.m" must be changed. *)
		$autoredRules = makeRULEC$[ 1];
		$elimRules = makeRULEC$[ 1];
		$introRules = makeRULEC$[ 1];
		$symRules = makeRULEC$[ 1];
		$transRules = makeRULEC$[ 1];
		(* Attention! '$compRules' and '$congRules' include those formulas that are declared as "comp"/"cong" *and*
			"simp"! Those that are declared as "comp"/"cong" only are not listed here. When changing tags of formulas
			one must ensure that these lists are updated accordingly, e.g. when adding tag "simp" to a formula 'fml'
			that already has tag "cong", 'fml' must be added to '$congRules' rather than '$simpRules'.
			Also note that there might be formulas having tag "simp" but being not included in any of the associations.
			This is because when tagging a formula cell as "simp", the resulting formula gets tag "simp" as well, but
			only its equational forms are added to the simpset. The original formula gets tag "simp" mainly because
			this makes it comparatively easy to properly handle scenarios where this formula is completely removed from
			the context again (or where only its tag "simp" is removed).
			If further levels of nesting are added to the simplification rules, 'extractRelevantRules' in
			"/CoreProvers/CoreProofMethods/Rewrite/Indirect.m" must be changed. *)
		$compRules = makeRULEC$[ 1];
		$congRules = makeRULEC$[ 1];
		$simpRules = makeRULEC$[ 1];
		
		With[ {var = VARIANT$[ #, Theorema`CoreProvers`CoreProofMethods`Private`canonVariant, {}]&},
			initCoreVar[ $autoredRuleVariant, var];
			initCoreVar[ $elimRuleVariant, var];
			initCoreVar[ $introRuleVariant, var];
			initCoreVar[ $symRuleVariant, var];
			initCoreVar[ $transRuleVariant, var];
			initCoreVar[ $simpRuleVariant, VARIANT$[ #, Theorema`CoreProvers`CoreProofMethods`Private`eqVariant, {"rw"}]&]
		];
		
		$theoryMapping = Association[];
		$thmLabelData = Association[];
		$labelKeyMapping = Association[];
		$labelThmMapping = Association[];
		
		Block[ {$ContextPath},
			$inCore = True;
			Quiet[ Check[ Needs[ "Theorema`Interface`InteractivePalette`"], Null]];
			Theorema`Interface`InteractivePalette`$communicatingCells = {"TheoremaCode", "FormalTextInputFormula", "ProofInfo"};
			Theorema`Interface`InteractivePalette`$defaultProgressDisplay =
				With[ {def = Style[ translate[ "selectItem"], Italic, Gray], uneval = "[" <> translate[ "unevalFmlCell"] <> "]"},
					Dynamic[
						Refresh[
							Replace[
								NotebookRead[ SelectedNotebook[]],
								{
									TagBox[ _, TAG$[ t_TYPED$], ___] :>
										Replace[ getFullTypeForDisplay[ t],
											{
												l_List :> DisplayForm[ RowBox[ {"[", RowBox[ Riffle[ Composition[ theoremaBoxes, printType] /@ l, ","]], "]"}]],
												typ_ :> theoremaDisplay[ printType[ typ]]
											}
										],
									Cell[ _, "FormalTextInputFormula", ___, CellID -> cid_, ___] :>
										With[ {k = makeFmlKey[ cid, NotebookFileName[ SelectedNotebook[]]]},
											Replace[ getFormulas[ False][ k],
												{
													fml_FML$ :> printFmlStatus[ fml],
													_ -> uneval
												}
											]
										],
									Cell[ _, "ProofInfo", ___, CellTags -> tags_, ___] :> printPrfStatus[ tags],
									_ -> def
								}
							],
							UpdateInterval -> 0.5
						]
					]
				];
			Theorema`Interface`InteractivePalette`$permanentAbortionExpr =
				Join[ Theorema`Interface`InteractivePalette`$permanentAbortionExpr, Hold[ $evaluationNotebook = Null, $evaluationCell = Null]];
			Theorema`Interface`InteractivePalette`openPalette[];
			Theorema`Provers`$proverSettings =
				{
					initInteractiveCoreProof[ Hold[]],
					executeCoreProofCommands,
					displayCoreProofState,
					displayCoreKnowledge,
					printCoreProofSituation,
					exportCoreProof
				};
			Needs[ "Theorema`CoreProvers`"];
			$tmaFmlPre = coreFmlPre;
			$tmaFmlPost = Identity;
			Theorema`Common`$defaultFormulaCollection = getFormulas[ True];
			Theorema`Common`$retrieveFormulasFunction = retrieveFormulasFromContext;
			(* The reason why we have to use '$codeEvaluationFunction' is that '$PreRead' would be applied to each line
				in a multiline input individually, and this we do not want. *)
			$codeEvaluationFunction = parseTheoremaCodeCell;
			$codeProlog := ($Pre = processTheoremaCell;);
			$codeEpilog := ($Pre =.;);
			$removeFromEnvTest = Function[ k, remFormsFromTheory[ {k}]; False];
			$TMALogFile = Null;
			$HistoryLength = 5;
			restartSession[ strict];
			True
		],
	(*else*)
		False
	]

(* Yet to be tested; normally one should not terminate a Core session anyway ... *)
leaveCore[] :=
	If[ TrueQ[ $inCore] && (getTheoryCounter[] === 0 || dialogYesNo[ 217, {}]),
		$inCore = False;
		With[ {set = Theorema`Interface`ProofScript`getDefaultProverSettings[]},
			If[ ListQ[ set], Theorema`Provers`$proverSettings = set]
		];
		Theorema`Interface`InteractivePalette`$communicatingCells = {};
		Theorema`Interface`InteractivePalette`$defaultProgressDisplay = "";
		Theorema`Interface`InteractivePalette`$permanentAbortionExpr =
			DeleteCases[ Theorema`Interface`InteractivePalette`$permanentAbortionExpr, HoldPattern[ $evaluationNotebook|$evaluationCell = _]];
		$registeredProofNumbers = Association[];
		$tmaCompPre = sequenceFlatten;
        $tmaCompPost = sequenceFlatten;
		$tmaFmlPre = defaultFmlPre&;
		$tmaFmlPost = formulaToComputation;
		Theorema`Common`$defaultFormulaCollection = Hold[ $tmaEnv];
		Theorema`Common`$retrieveFormulasFunction = {}&;
		$codeProlog := ($parseTheoremaQuoted = True;);
		$codeEpilog := ($parseTheoremaQuoted = False;);
		$inputProlog := ($parseTheoremaQuoted = True;);
		$inputEpilog := ($parseTheoremaQuoted = False;);
		$removeFromEnvTest = (True&);
		$HistoryLength = Infinity;
	]

(* Core variables are global variables that influence the behavior of Core and that may be set to different values in
	different theories. Hence, they must know which values they have in which theories.
	Lists to which new elements are added and old elements are removed in different theories (e.g. parsing rules, proof
	language interpretation rules, ...) are *not* counted as Core variables! *)
(* 'initCoreVar' is not public! *)
SetAttributes[ initCoreVar, HoldFirst];
initCoreVar[ sym_Symbol, val_] :=
	(
		If[ !MemberQ[ $coreVars, Hold[ sym]], AppendTo[ $coreVars, Hold[ sym]]];
		sym = COREVAR$[ Association[], val]
	)

(* 'resetCoreVars' is not public! *)
resetCoreVars[ syms_List] :=
	Scan[
		Function[ sym,
			Set @@ Append[ sym, resetCoreVars[ First[ sym]]]
		],
		syms
	]
resetCoreVars[ COREVAR$[ ___, assoc_Association, def_]] :=
	Catch[ COREVAR$[ lookupCurrentValue[ assoc, Throw[ COREVAR$[ assoc, def]]], assoc, def]]
resetCoreVars[ x_] := x

SetAttributes[ lookupCurrentValue, HoldRest];
lookupCurrentValue[ a_Association, def_] :=
	lookupCurrentValue[ Normal[ a], Infinity, def]
lookupCurrentValue[ x_, def_] :=
	lookupCurrentValue[ x, Infinity, def]
lookupCurrentValue[ {}, _, def_] := def
lookupCurrentValue[ {(Rule|RuleDelayed)[ thy_, val_], rest___}, min_, def_] :=
	With[ {dist = getTheoryRootDist[ thy]},
		Which[
			dist === 0,
			val,
		(*else*)
			dist < min,
			lookupCurrentValue[ {rest}, dist, val],
		(*else*)
			True,
			lookupCurrentValue[ {rest}, min, def]
		]
	]
lookupCurrentValue[ x_, _, _] := x

(* 'setCoreVar', 'unsetCoreVar' and 'getCoreVar' can handle Core variables and other variables alike,
	so in case of doubt just use these functions. *)
SetAttributes[ setCoreVar, HoldFirst];
setCoreVar[ sym_Symbol, val_] :=
	setCoreVar[ sym, val, getEvaluationTheoryName[]]
setCoreVar[ sym_Symbol, val_, thy_String] :=
	(sym = setCoreVar[ 0, sym, val, thy])
setCoreVar[ 0, COREVAR$[ v___, assoc_Association, def_], val_, thy_String] :=
	If[ getTheory[] === thy,
		COREVAR$[ val, Append[ assoc, thy -> val], def],
	(*else*)
		resetCoreVars[ COREVAR$[ v, Append[ assoc, thy -> val], def]]
	]
setCoreVar[ 0, _, val_, _] := val

SetAttributes[ unsetCoreVar, HoldFirst];
unsetCoreVar[ sym_Symbol] :=
	unsetCoreVar[ sym, getEvaluationTheoryName[]]
unsetCoreVar[ sym_Symbol, thy_String] :=
	(sym = unsetCoreVar[ 0, sym, thy])
unsetCoreVar[ 0, COREVAR$[ v___, assoc_Association, def_], thy_String] :=
	If[ MemberQ[ Keys[ assoc], thy],
		resetCoreVars[ COREVAR$[ v, KeyDrop[ assoc, thy], def]],
	(*else*)
		COREVAR$[ v, assoc, def]
	]
unsetCoreVar[ 0, x_, _] := x

getCoreVar[ COREVAR$[ val_, _Association, _]] := val
getCoreVar[ COREVAR$[ _Association, def_]] := def
getCoreVar[ x_] := x
getCoreVar[] := $coreVars

getCoreVarAssoc[ COREVAR$[ ___, assoc_Association, ___]] := assoc
getCoreVarAssoc[ _] := $Failed

printFmlStatus[ FML$[ k_, _, l0_String, ___]] :=
	With[ {l = makeLabel[ l0], sep = ", "},
		Replace[ Last[ getFmlEnv[ False][ k]],
			{
				0 :> l <> sep <> translate[ "axm"],
				5 :> l <> sep <> translate[ "thm"],
				11 :> l <> sep <> translate[ "defIllegalType"],
				12 :> l <> sep <> translate[ "defRecursive"],
				13 :> l <> sep <> translate[ "defCyclic"],
				14 :> l <> sep <> translate[ "defOverlapping"],
				15 :> l <> sep <> translate[ "defValid"],
				25 :> l <> sep <> translate[ "axmTrusted"],
				s_ :>
					Which[
						1 <= s <= 4,
						l <> sep <> printPrfStatus[ s],
					(*else*)
						21 <= s <= 24,
						l <> sep <> translate[ "axmTrusted"] <> " (" <> printPrfStatus[ s - 20] <> ")",
					(*else*)
						True,
						l
					]
			}
		]
	]

printPrfStatus[ stat_Integer] :=
	Switch[ stat,
		1, translate[ "prfInvalid"],
		2, translate[ "prfCyclic"],
		3, translate[ "prfIncomplete"],
		4, translate[ "prfSkipped"],
		5, translate[ "prfValid"],
		_, translate[ "prf"]
	]
printPrfStatus[ tags_] :=
	printPrfStatus[ FirstCase[ Flatten[ {tags}], s_String /; StringMatchQ[ s, DigitCharacter..] :> ToExpression[ s], 0]]

(* 'inScopeQ[ sp]' checks whether the given source-position specification 'sp' is currently in scope (w.r.t.
	'$currentScope'); more precisely, it checks whether it does not come *after* '$currentScope', since in case the
	order of the two source-positions cannot be determined the function simply returns 'True'.
	This function is, for instance, used by proof methods which add knowledge to the proof, in order not to add
	"illegal" knowledge. It must be noted, however, that this is merely a service for the authors of proofs: in the
	end, it is solely the ConsistencyChecker that decides whether formalizations have circular dependencies or not!
	This also explains why global variable '$ignoreScope' may safely be set to avoid the whole checking, and why even
	the values of '$proofGoal' and '$currentScope' are allowed to change during a proof.
	Note that many functions for retrieving formulas from the theory context do not need to call 'inScopeQ', because
	functions like 'getFormulas[ True]' only "see" formulas currently in scope anyway. *)
inScopeQ[ f_] :=
	inScopeQ[ f, None]
inScopeQ[ f_, thyScope_] :=
	If[ TrueQ[ $ignoreScope],
		!MatchQ[ getCurrentScope[], {Replace[ f, {FML$[ ky_, __] :> ky, {ky_List, _List} :> ky}], _}], (* 'getCurrentScope[]' might not be a list *)
	(*else*)
		With[ {res = inScopeQ[ f, getCurrentScope[], thyScope]},
			res === Less || (res === Indeterminate && thyScope === None)
		]
	]
inScopeQ[ a_FML$, b_List, thyScope_] :=
	compareTheoryPositions[ {key@a, getOptionalComponent[ a, "pos"]}, b, thyScope]
inScopeQ[ a:{_List, _List}, b_List, thyScope_] :=
	compareTheoryPositions[ a, b, thyScope]
inScopeQ[ a_List, b_List, thyScope_] :=
	inScopeQ[ getFormulas[ True][ a], b, thyScope]
inScopeQ[ _, _, _] :=
	Indeterminate

inScopeTSQ[ f_] :=
	inScopeQ[ f, inTheoryScopeQ]

(* A formal entity originating from "$Core" (without "Source:"-prefix!) is loaded automatically whenever entering
	Core, even if theory "Minimal" is not loaded. Hence, "$Core" is always in scope and moreover entities originating
	from "$Core" do not have to be exported when creating theory images.
	Note that the only entities which might originate from "$Core" are those that are not tracked by "Consistency".
	This, in particular, rules out types, constants, formulas and theorem schemas. *)
inTheoryScopeQ[ "$Core"] = True
inTheoryScopeQ[ thy_] := IntegerQ[ getTheoryRootDist[ thy]]
inTheoryScopeQ[ "$Core", excpt_List] :=
	NoneTrue[ excpt, inTheoryScopeQ]
inTheoryScopeQ[ thy_, {}] :=
	inTheoryScopeQ[ thy]
inTheoryScopeQ[ thy_, excpt_List] :=
	With[ {dist = getTheoryRootDist[ thy]},
		IntegerQ[ dist] && AllTrue[ excpt, dist < getTheoryRootDist[ #] &]
	]

getFormulas[ False][] :=
	getFmlEnv[]
getFormulas[ True][] :=
	Select[ getFmlEnv[], Composition[ inTheoryScopeQ, Last, First, First]]
getFormulas[ scopeOnly_][ k_] :=
	getFmlEnv[ scopeOnly][ k][[2]]

getTheorems[ False][] :=
	getThmEnv[]
getTheorems[ True][] :=
	Select[ getThmEnv[], Composition[ inTheoryScopeQ, Last, First, First]]
getTheorems[ scopeOnly_][ k_] :=
	getThmEnv[ scopeOnly][ k][[2]]

retrieveFormulasFromContext[ Cases, keyPatt_, lblPatt_, num_, evalSrc_, coll:Except[ _List]] :=
	coll /@ retrieveFormulasFromContext[ key, keyPatt, lblPatt, num, evalSrc, coll]
retrieveFormulasFromContext[ Cases, keyPatt_, lblPatt_, num_, _, coll:Except[ _List], test_] :=
	Select[ coll /@ retrieveFormulasFromContext[ key, keyPatt, lblPatt, Infinity, None, coll], test, num]	(* no sense to set 'evalSrc' to anything different from 'None' *)
retrieveFormulasFromContext[ key, keyPatt_, lblPatt_, num_, evalSrc_, coll:Except[ _List]] :=
	retrieveFormulasFromContext[ key, keyPatt, lblPatt, num, evalSrc, coll[]]
retrieveFormulasFromContext[ key, {idPatt_, srcPatt_}, lblPatt_, num0_, evalSrc_, coll_List] :=
	Which[
		IntegerQ[ num0] && StringQ[ evalSrc] && (srcPatt === Blank[] || srcPatt === Blank[ String]),
		(* First search in the current theory, then search in all other theories. *)
		With[ {aux = retrieveFormulasFromContext[ key, {idPatt, evalSrc}, lblPatt, num0, None, coll]},
		With[ {len = Length[ aux]},
			If[ len === num0,
				aux,
			(*else*)
				Join[ aux, retrieveFormulasFromContext[ key, {idPatt, Except[ evalSrc, _String]}, lblPatt, num0 - len, None, coll]]
			]
		]],
	(*else*)
		num0 === 1 || StringQ[ srcPatt],
		Catch[
			Scan[
				Function[ c,
					If[ MatchQ[ c, {{{_, srcPatt}, _}, ___}],
						With[ {ks = Cases[ c, {k:{idPatt, _}, lblPatt} :> k, {1}, num0]},
							If[ StringQ[ srcPatt] || ks =!= {}, Throw[ ks]]
						]
					]
				],
				coll
			];
			{}
		],
	(*else*)
		num0 === Infinity,
		Join @@ Cases[ coll, c:{{{_, srcPatt}, _}, ___} :> Cases[ c, {k:{idPatt, _}, lblPatt} :> k]],
	(*else*)
		True,
		Module[ {num = num0},
			Join @@ Map[
				Function[ c,
					If[ Positive[ num] && MatchQ[ c, {{{_, srcPatt}, _}, ___}],
						With[ {ks = Position[ c, {k:{idPatt, _}, lblPatt} :> k, {1}, num]},
							num -= Length[ ks];
							ks
						],
					(*else*)
						{}
					]
				],
				coll
			]
		]
	]
retrieveFormulasFromContext[ key, keyPatt_, lblPatt_, num_, _, coll:Except[ _List], test_] :=
	Select[ retrieveFormulasFromContext[ key, keyPatt, lblPatt, Infinity, None, coll[]], Composition[ test, coll], num]
retrieveFormulasFromContext[ Position, keyPatt_, lblPatt_, num_, evalSrc_, coll:Except[ _List]] :=
	retrieveFormulasFromContext[ Position, keyPatt, lblPatt, num, evalSrc, coll[]]
retrieveFormulasFromContext[ Position, {idPatt_, srcPatt_}, lblPatt_, num0_, evalSrc_, coll_List] :=
	(* We give separate definitions for 'key' and 'Position', since 'retrieveFormulasFromContext' is a time-critical
		function that should be implemented as efficiently as possible. *)
	Which[
		IntegerQ[ num0] && StringQ[ evalSrc] && (srcPatt === Blank[] || srcPatt === Blank[ String]),
		(* First search in the current theory, then search in all other theories. *)
		With[ {aux = retrieveFormulasFromContext[ Position, {idPatt, evalSrc}, lblPatt, num0, None, coll]},
		With[ {len = Length[ aux]},
			If[ len === num0,
				aux,
			(*else*)
				Join[ aux, retrieveFormulasFromContext[ Position, {idPatt, Except[ evalSrc, _String]}, lblPatt, num0 - len, None, coll]]
			]
		]],
	(*else*)
		num0 === 1 || StringQ[ srcPatt],
		Catch[
			Module[ {i = 0},
				Scan[
					Function[ c,
						++i;
						If[ MatchQ[ c, {{{_, srcPatt}, _}, ___}],
							With[ {js = Position[ c, {{idPatt, _}, lblPatt}, {1}, num0, Heads -> False]},
								If[ StringQ[ srcPatt] || js =!= {}, Throw[ Join[ i, #]& /@ js]]
							]
						]
					],
					coll
				]
			];
			{}
		],
	(*else*)
		num0 === Infinity,
		Join @@ MapIndexed[
			Function[ {c, i},
				If[ MatchQ[ c, {{{_, srcPatt}, _}, ___}],
					With[ {js = Position[ c, {{idPatt, _}, lblPatt}, {1}, Heads -> False]},	(* if one element in 'c' matches 'srcPatt', then all elements do *)
						Join[ i, #]& /@ js
					],
				(*else*)
					{}
				]
			],
			coll
		],
	(*else*)
		True,
		Module[ {num = num0},
			Join @@ MapIndexed[
				Function[ {c, i},
					If[ Positive[ num] && MatchQ[ c, {{{_, srcPatt}, _}, ___}],
						With[ {js = Position[ c, {{idPatt, _}, lblPatt}, {1}, num, Heads -> False]},
							num -= Length[ js];
							Join[ i, #]& /@ js
						],
					(*else*)
						{}
					]
				],
				coll
			]
		]
	]
retrieveFormulasFromContext[ Position, keyPatt_, lblPatt_, num_, _, coll:Except[ _List], test_] :=
	With[ {l = coll[]},
		Select[ retrieveFormulasFromContext[ Position, keyPatt, lblPatt, None, Infinity, l], Composition[ test, coll, First[ l[[#]]]&], num]
	]
retrieveFormulasFromContext[ ___] := {}

(*
coreFmlPreGlob[ form_, glob_] :=
	Which[
		Theorema`Interface`InteractivePalette`cellStyleActivatedQ[ "FormalTextInputFormula"] === False,
		{form, {}, {}},
	(*else*)
		glob === {},
		bindFix[ bindMeta[ form]],
	(*else*)
		True,
		{
			Theorema`Language`Session`Private`applyGlobalDeclaration[
				First[ bindFix[ bindMeta[ form]]],
				glob
			],
			{},
			{}
		}
	]

bindMeta[ term:((h:(CoreImplies$TM|LongRightArrow$TM|Hold[ CoreImplies$TM|LongRightArrow$TM]|
			(TYPED$|MessageName$TM|Hold[ TYPED$|MessageName$TM])[ CoreImplies$TM|LongRightArrow$TM|Hold[ CoreImplies$TM|LongRightArrow$TM], _]))[ A_, G__])] :=
	If[ FreeQ[ A, META$[ _, _Integer]],
		h[ A, bindMeta[ If[ Length[ {G}] === 1, G, h[ G]]]],
	(*else*)
		bindMetaHere[ term]
	]
bindMeta[ term:((h:(CoreImplies$TM|LongRightArrow$TM|Hold[ CoreImplies$TM|LongRightArrow$TM]|
			(TYPED$|MessageName$TM|Hold[ TYPED$|MessageName$TM])[ CoreImplies$TM|LongRightArrow$TM|Hold[ CoreImplies$TM|LongRightArrow$TM], _]))[ A_][ G_])] :=
	If[ FreeQ[ A, META$[ _, _Integer]],
		h[ A][ bindMeta[ G]],
	(*else*)
		bindMetaHere[ term]
	]
bindMeta[ term_] :=
	bindMetaHere[ term]

bindMetaHere[ term_] :=
	With[ {p = Position[ term, META$[ _, _Integer]]},
		If[ p === {},
			term,
		(*else*)
			Replace[ Transpose[ Replace[ GatherBy[ MapThread[ Rule, {p, Extract[ term, p]}], Last], all:{_ -> _[ n_, _], ___} :> {all[[All, 1]], VAR$[ n]}, {1}]],
				{pos_List, vars0_List} :>
					With[ {vars = freshName[ vars0, DeleteDuplicates[ Cases[ term, v_VAR$ :> varName[ v], {0, Infinity}, Heads -> True]]],
							ot = getObjectTruthType[]},
					With[ {tj = If[ ot === FORM$TM, Identity, TYPED$[ TrueQ$TM, {}]]},
						TYPED$[ CoreAll$TM, {ot}][ Lambda$TM[ RNG$[ TYPRNG$[ VAR$[ Theorema`Knowledge`VAR$thesis$TM], ot]],
							TYPED$[ CoreImplies$TM, {}][
								Fold[
									CoreAll$TM[ Lambda$TM[ RNG$[ TYPRNG$[ #2, TVAR$[]]], #1]]&,
									TYPED$[ CoreImplies$TM, {}][ ReplacePart[ term, MapThread[ Rule, {pos, vars}]]][ tj[ VAR$[ Theorema`Knowledge`VAR$thesis$TM]]],
									Reverse[ vars]
								]
							][
								tj[ VAR$[ Theorema`Knowledge`VAR$thesis$TM]]		(* no de Bruijn indices yet *)
							]
						]]
					]]
			]
		]
	]

bindFix[ term_] :=
	With[ {p = Position[ term, FIX$[ _, _Integer]]},
		If[ p === {},
			{term, {}, {}},
		(*else*)
			Replace[ Transpose[ Replace[ GatherBy[ MapThread[ Rule, {p, Extract[ term, p]}], Last], all:{_ -> _[ n_, _], ___} :> {all[[All, 1]], VAR$[ n]}, {1}]],
				{pos_List, vars0_List} :>
					{
						With[ {vars = freshName[ vars0, DeleteDuplicates[ Cases[ term, v_VAR$ :> varName[ v], {0, Infinity}, Heads -> True]]]},
							Fold[
								CoreAll$TM[ Lambda$TM[ RNG$[ TYPRNG$[ #2, TVAR$[]]], #1]]&,
								ReplacePart[ term, MapThread[ Rule, {pos, vars}]],
								Reverse[ vars]
							]
						],
						vars0[[All, 1]],
						{}
					}
			]
		]
	]
*)

coreFmlPre[ Except[ _NotebookObject], cell_][ form_, lbl_, simp_, tags_List, k_, pos_] :=
	Replace[ EvaluationNotebook[],
		{
			nb_NotebookObject :> coreFmlPre[ nb, cell][ form, lbl, simp, tags, k, pos],
			_ -> $Failed
		}
	]
coreFmlPre[ nb_NotebookObject, Except[ _CellObject]][ form_, lbl_, simp_, tags_List, k_, pos_] :=
	Replace[ SelectedCells[ nb],
		{
			{cell_CellObject} :> coreFmlPre[ nb, cell][ form, lbl, simp, tags, k, pos],
			_ -> $Failed
		}
	]
coreFmlPre[ nb_NotebookObject, cell_CellObject][ form_, lbl_, _, tags_List, k:{_, src_String}, pos_] :=
	Which[
		Theorema`Interface`InteractivePalette`cellStyleActivatedQ[ "FormalTextInputFormula"] === False,
		(* TODO: This should actually be checked earlier, before even parsing the formula cell. Similar for global declarations. *)
		Null,
	(*else*)
		setTheory[ src] === $Failed,
		Null,
	(*else*)
		True,
		Replace[
			enterCoreCellEvaluation[ nb, cell, 2, lbl];
			$msgID = None;
			Reap[
				Replace[ formulaToAssumption[ parseCoreTermPre[ form], If[ MemberQ[ tags, "no-norm"], Identity, normalize], False],
					{
						_String|Null|$Failed -> $Failed,
						{ret_, tvar_Association} :>
							Replace[ getCoreVar[ $coreFmlPre][ ret, lbl, tags, k, pos],
								{
									{newForm_, newLbl_, newTags_, kind_, post_} :>
										With[ {fml =
												Join[
													FML$[
														k,
														newForm,
														newLbl
													],
													FML$ @@ Join[
														If[ newTags === {}, {}, {"tags" -> newTags}],
														If[ ListQ[ pos], {"pos" -> pos}, {}]
													]
													(* "arData" and "rwData" cannot be computed here, because 'addFmlToTheory'
														may change the formula (for instance if it is a definition whose head has
														not been declared yet). *)
												]},
											post[
												fml,
												addFmlToTheory[
													fml,
													tvar,
													kind,
													MemberQ[ newTags, "autored"|"autored?"],
													kind === "def",
													True,
													False
												]
											]
										],
									_ -> $Failed
								}
							]
					}
				],
				"Message"
			],
			{res_, msg_List} :>
				(If[ coreNotification[ msg, res === $Failed, {MSG$[ coreMessageToList, 203, {k, lbl}, "Fatal"]}, {nb, cell}],
					If[ getTheoryMode[] =!= True, remFormsFromTheory[ {k}]];
					exitCoreCellEvaluation[ nb, cell, False],
				(*else*)
					exitCoreCellEvaluation[ nb, cell, True];
				];)
				(* Instead of returning 'Null' one could also return '{newForm, newLbl, False, ...}'. Then, the
					formula would be added to '$tmaEnv', too. *)
		]
	]


(* ::Subsection:: *)
(*Import/Export*)

(* Problems with the current implementation of theory imports:

	1. Importing theory images that are extensions of others, like "Main" and "Complex_Main" in Isabelle.
	Namely, assume the following theory dependency graph:
	
		  T
		 / \
	 	A	B
	   /	 \
	Main     Complex_Main
	
	When T imports A and B, the first image found is Main, meaning that the image will be loaded. However, later on
	Complex_Main shall be imported as well, which must now be accomplished by loading its contents individually (or,
	more precisely, those theories not included in Main) --> SLOW!
	One way to overcome such a situation is to import the largest possible theory image, i.e. Complex_Main, in the
	common successor of A and B, i.e. T, *before* importing A and B.
	
	2. Importing a theory into a context that is larger than the context the theory was originally developed in.
	Assume the same situation as above. When T imports Complex_Main at the very beginning, A is imported in a context
	where more theorems, declarations, definitions etc. are known than when A was originally developed. From the purely
	logical point of view this should not do any harm, because if some definition in A contradicts one in Complex_Main,
	then importing A and B in T cannot be accomplished in any way. Still, Complex_Main might interfere with parsing
	terms in A, change global settings of proof methods, etc.
	The same is true when importing B after importing A, even if both A and B depend on the same image.
	
	The best and most elegant solution to both problems would be to implement arbitrary merging of contexts, taking
	care of definitions being non-overlapping and terminating; this requires a considerable effort, though. *)

importTheories[ specs0_List][ False, _, id_, src0_, _, nb_NotebookObject] :=
	With[ {specs1 = Replace[ specs0, STRING$[ s_String] :> s, {1}]},
		Replace[ FirstCase[ specs1, Except[ _String], ""],
			{
				_String :>
					Catch[
						With[ {cur = getTheory[],
								files =
									Map[
										Function[ spec,
											With[ {file = makeTheoryFileName[ spec, False, nb]},
												Which[
													!StringQ[ file],
													coreNotification[ {MSG$[ coreMessageToList, 1085, {spec}, "Fatal"]}, True, {}, {src0, id}];
													Throw[ $Failed],
												(*else*)
													FileExtension[ file] =!= "nb",
													coreNotification[ {MSG$[ coreMessageToList, 247, {file}, "Fatal"]}, True, {}, {src0, id}];
													Throw[ $Failed],
												(*else*)
													True,
													file
												]
											]
										],
										specs1
									]},
							Scan[
								Function[ file,
									If[ getTheoryDeps[ sourceLabel[ file]] === $Failed,
										Which[
											getImageFile[] === None && FileExistsQ[ StringReplace[ file, pre__ ~~ ".nb" ~~ EndOfString :> pre <> ".mx"]],
											Switch[ importImage[ file, True],
												StringReplace[ file, pre__ ~~ ".nb" ~~ EndOfString :> pre <> "-extra2.mx"],
												Null,
											(*else*)
												_String,
												(* The image could not be imported entirely, so the notebook must be evaluated. *)
												If[ evaluateNotebook[ file, Automatic] === $Failed, Throw[ $Failed]],
											(*else*)
												_,
												Throw[ $Failed]
											],
										(*else*)
											evaluateNotebook[ file, Automatic] === $Failed,
											Throw[ $Failed]
										]
									]
								],
								files
							];
							
							setTheory[ cur];
							addTheoryDeps[ sourceLabel /@ files, cur];
							Null
						]
					],
				t_ :> (coreNotification[ {MSG$[ coreMessageToList, 223, {t}, "Fatal"]}, True, {}, {src0, id}]; $Failed)
			}
		]
	]
importTheories[ ___][ True, _, id_, src_, ___] :=
	(coreNotification[ {}, True, 222, {src, id}]; $Failed)

(* 'setTheory' is no theory command! *)
setTheory[ thy_String] :=
	changeTheoryScope[ manipulateContext[ "set thy", thy]]

(* 'getTheory' is no theory command! *)
getTheory[] :=
	Quiet[ Check[ First[ getTheoryScope[]], $Failed]]

(* 'addTheoryDeps' is no theory command! *)
addTheoryDeps[ new_List] :=
	addTheoryDeps[ new, getTheory[]]
addTheoryDeps[ new_List, thy_String] :=
	changeTheoryScope[ manipulateContext[ "change deps", thy, Join[ Lookup[ getTheoryDeps[], thy, {}], new]]]

(* 'remTheoryDeps' is no theory command! *)
remTheoryDeps[ new_List] :=
	remTheoryDeps[ new, getTheory[]]
remTheoryDeps[ new_List, thy_String] :=
	changeTheoryScope[ manipulateContext[ "change deps", thy, DeleteCases[ Lookup[ getTheoryDeps[], thy, {}], Alternatives @@ new]]]

changeTheoryScope[ out_] :=
	(
		If[ TrueQ[ out],
			resetCoreVars[ $coreVars];
			$abbrevConstDecls = resetCoreVars /@ $abbrevConstDecls
		];
		out
	)

(* 'registerTheory' is no theory command! *)
registerTheory[ key_String] :=
	registerTheory[ key, getEvaluationTheoryName[]]
registerTheory[ key_String, thy_String] :=
	AssociateTo[ $theoryMapping, key -> thy]

System`RegisterTheory = registerTheory;


(* ::Subsection:: *)
(*Typedef, Overloading, Locking, Sort*)

typedefOnOff[ oo_][ undo_, _, id_, src_, ___] :=
	Replace[
		Reap[
			If[ oo === undo,
				manipulateContext[ "typedef off"],
			(*else*)
				manipulateContext[ "typedef on"]
			],
			"Message"
		],
		{res_, msg_List} :> (coreNotification[ msg, False, {}, {src, id}]; If[ res === $Failed, $Failed, Null])
	]

overloadingOnOff[ l_List, oo_][ undo_, _, id_, src_, ___] :=
	Replace[
		Reap[
			If[ oo === undo,
				(!MemberQ[ l, "type"] || manipulateContext[ "type overloading off"] =!= $Failed) &&
					(!MemberQ[ l, "const"] || manipulateContext[ "const overloading off"] =!= $Failed),
			(*else*)
				(!MemberQ[ l, "type"] || manipulateContext[ "type overloading on"] =!= $Failed) &&
					(!MemberQ[ l, "const"] || manipulateContext[ "const overloading on"] =!= $Failed)
			],
			"Message"
		],
		{res_, msg_List} :> (coreNotification[ msg, False, {}, {src, id}]; If[ TrueQ[ res], Null, $Failed])
	]

lockConstants[ consts0_List][ undo_, i_, id_, src0_, ___] :=
	With[ {consts = Replace[ consts0, EXPR$[ e_, ___] :> e, {1}]},
	With[ {msg =
			Last[
				Reap[
					If[ TrueQ[ undo],
						Scan[
							manipulateContext[ "rem lock", #]&,
							DeleteCases[ consts, _TYPED$]
						],
					(*else*)
						With[ {src = sourceLabel[ src0]},
						Module[ {j = 0},
							Scan[
								manipulateContext[ "lock", #, {cellIDLabel[ {id, i, ++j}], src}]&,
								DeleteCases[ consts, _TYPED$]
							]
						]]
					],
					"Message"
				]
			]},
		coreNotification[ Join[ msg, If[ MemberQ[ consts, _TYPED$], {MSG$[ coreMessageToList, 220, Cases[ consts, _TYPED$], "Fatal"]}, {}]], False, {}, {src0, id}];
	]]

setDefaultSort[ internal_, sort0_][ undo_, _, id_, src_, ___] :=
	If[ TrueQ[ internal],
		(* We keep function 'setDefaultSort' as it is, because maybe at some point it will be possible to change the
			internal default sort. For now, however, we show a warning. *)
		coreNotification[ {}, True, {MSG$[ coreMessageToList, 248, {}, "Warning"]}, {src, id}];
		Null,
	(*else*)
		With[ {sort =
				If[ TrueQ[ undo],
					{},
				(*else*)
					Replace[ Replace[ sort0, EXPR$[ e_, ___] :> e], {s:(_Set$TM|_List) :> Union[ List @@ s], s_ :> {s}}]
				]},
			setCoreVar[ $defaultInputSort, sort, sourceLabel[ src]];
		]
	]


(* ::Subsection:: *)
(*Truth Judgment*)

setTruthJudgment[ term:EXPR$[ TYPED$[ TrueQ$TM, _], ___]] :=
	setTruthJudgment[ makeSTERM[ term]]
setTruthJudgment[ term:STERM$[ TYPED$[ TrueQ$TM, _], _Association, _Association]][ False, i_, id_, src_, ___] :=
	Replace[
		Reap[
			If[ unsetTruthJudgment[ TrueQ$TM, False] =!= $Failed,
				With[ {k = {cellIDLabel[ {id, i}], sourceLabel[ src]}},
					If[ manipulateContext[ "decl", term, k, True&] =!= $Failed &&
							manipulateContext[ "lock", TrueQ$TM, k] =!= $Failed,
						Null,
					(*else*)
						$Failed
					]
				],
			(*else*)
				$Failed
			],
			"Message"
		],
		{res_, msg_List} :> (coreNotification[ msg, False, {}, {src, id}]; res)
	]
setTruthJudgment[ term0_][ False, i_, id_, src_, cell_CellObject, nb_NotebookObject] :=
	If[ FreeQ[ term0, TrueQ$TM],
		(* We do not set '$strictTypeChecking' to 'True', because 'parseCoreTermPost' could still replace constants by other terms.
			Also, we do not set '$reportUndeclaredConsts' to 'True' either, because warnings shall only be generated when the
			term enters the trusted kernel. *)
		Replace[ makeCoreExpr[ term0, Blank[], 0, {}, {}, Association[], Association[], {}, "gen", normalize, None],	(* MODIFICATION: "gen" was 'None' before. *)
			{
				{{typ_, STERM$[ term_, Association[], tv_Association], _, {}, {}}, {}} :>
					Replace[
						Reap[
							If[ unsetTruthJudgment[ TrueQ$TM, False] =!= $Failed,
								With[ {k = {cellIDLabel[ {id, i}], sourceLabel[ src]}},
									If[ manipulateContext[ "decl", STERM$[ TYPED$[ TrueQ$TM, typ], Association[], tv], k, True&] =!= $Failed &&
											manipulateContext[ "axiom",
												(* It could be that a formula with key 'k' exists already; then it is overwritten. *)
												FML$[ k,
													TYPED$[ CoreEqual$TM, {typ}][ TYPED$[ TrueQ$TM, {}]][ term],
													"TrueQ def",
													"pos" -> getEvaluationPosition[ id, nb]
												],
												tv,
												True
											] =!= $Failed,
										Null,
									(*else*)
										$Failed
									]
								],
							(*else*)
								$Failed
							],
							"Message"
						],
						{res_, msg_List} :> (coreNotification[ msg, False, {}, {src, id}]; res)
					],
				_ :> (coreNotification[ {MSG$[ coreMessageToList, 208, {term0}, "Fatal"]}, False, {}, {src, id}]; $Failed)
			}
		],
	(*else*)
		coreNotification[ {MSG$[ coreMessageToList, 210, {TrueQ$TM, term0}, "Fatal"]}, False, {}, {src, id}];
		$Failed
	]
setTruthJudgment[ _][ True, _, id_, src_, ___] :=
	Replace[ Reap[ unsetTruthJudgment[ TrueQ$TM, True], "Message"],
		{res_, msg_List} :> If[ coreNotification[ msg, res === $Failed, {MSG$[ coreMessageToList, 207, {TrueQ$TM}, "Fatal"]}, {src, id}], $Failed, Null]
	]

(* 'unsetTruthJudgment' returns 'False' if nothing needs to be done, 'True' if the given truth judgment can be
	successfully unset, and '$Failed' if something goes wrong. *)
unsetTruthJudgment[ tj_, undecl_] :=
	Which[
		Lookup[ $abbrevConstDecls, tj, $Failed] =!= $Failed,	(* we check whether 'tj' has been abbreviated *somewhere* *)
		sowMessage[ 231, {tj}, "Fatal"];
		$Failed,
	(*else*)
		True,
		With[ {defs = getConstDefs[ False][ tj]},
			If[ defs === {},
				If[ getLockedConsts[ tj] === $Failed,
					Which[
						!undecl,
						False,
					(*else*)
						getConstDecls[ False][ tj] === $Failed,
						False,
					(*else*)
						manipulateContext[ "rem decl", tj] === $Failed,
						$Failed,
					(*else*)
						True,
						True
					],
				(*else*)
					If[ manipulateContext[ "rem lock", tj] =!= $Failed &&
							(!undecl || getConstDecls[ False][ tj] === $Failed || manipulateContext[ "rem decl", tj] =!= $Failed),
						True,
					(*else*)
						$Failed
					]
				],
			(*else*)
				Catch[
					Scan[
						If[ manipulateContext[ "rem fml", #[[2, 1]]] === $Failed, Throw[ $Failed]]&,
						defs
					];
					If[ manipulateContext[ "rem defs", tj] =!= $Failed && (!undecl || manipulateContext[ "rem decl", tj] =!= $Failed),
						True,
					(*else*)
						$Failed
					]
				]
			]
		]
	]


(* ::Subsection:: *)
(*Type Declarations*)

declareTypes[ types_List][ False, i_, id_, src_, ___] :=
	Replace[
		declareTypesDirect[ Replace[ types, EXPR$[ t_, ___] :> First[ generalizeTypes[ t, {}]], {1}], cellIDLabel[ {id, i, #}]&, sourceLabel[ src], Interactive, {src, id}],
		{
			{_, {}} -> Null,
			_ -> $Failed
		}
	]
declareTypes[ types_List][ True, _, id_, src_, ___] :=
	Replace[ remTypesFromTheory[ Replace[ types, EXPR$[ t_, ___] :> First[ generalizeTypes[ t, {}]], {1}], {src, id}],
		{
			{_, {}} -> Null,
			_ -> $Failed
		}
	]

(* 'types' is supposed to be a list of plain types without sort constraints. *)
declareTypesDirect[ types_, {id_String, src_String}, rest___] :=
	declareTypesDirect[ types, id, src, rest]
declareTypesDirect[ types_, id_, src_String] :=
	declareTypesDirect[ types, id, src, Interactive]
declareTypesDirect[ types_, id_, src_String, overwrite_] :=
	declareTypesDirect[ types, id, src, overwrite, None]
declareTypesDirect[ types_List, id_String, src_String, overwrite_, origin_] :=
	declareTypesDirect[ types, id&, src, overwrite, origin]
declareTypesDirect[ types_List, idF_, src_String, overwrite_, origin_] :=
	With[ {overwriteF = Switch[ overwrite, True, True&, False, False&, _, overwriteTypesDialog]},
	Module[ {i = 0},
		Replace[
			Last[
				Reap[
					Scan[
						If[ getTypeOverlaps[ #, $abbrevTypeDecls] === {},
							With[ {overlapKeys = getTypedefOverlapKeyPatts[ #],
									res = manipulateContext[ "typedecl", #, {idF[ ++i], src}, overwriteF]},
								If[ res =!= $Failed && res =!= False, remFormsFromTheory[ overlapKeys]];
								Sow[ #, res =!= $Failed]
							],
						(*else*)
							sowMessage[ 236, {#}, "Fatal"];
							Sow[ #, False]
						]&,
						types
					],
					{True, False, "Message"}
				]
			],
			{succ_List, fail_List, msg_List} :>
				(
					coreNotification[ msg, False, {}, origin];
					{Join @@ succ, Join @@ fail}
				)
		]
	]]
declareTypesDirect[ typ_, id_, src_String, overwrite_, origin_] :=
	declareTypesDirect[ {typ}, id, src, overwrite, origin]

(* 'types' is supposed to be a list of plain types without sort constraints. *)
remTypesFromTheory[ types_] :=
	remTypesFromTheory[ types, None]
remTypesFromTheory[ types_List, origin_] :=
	Replace[
		Reap[
			Scan[
				With[ {overlapKeys = getTypedefOverlapKeyPatts[ #],
						res = manipulateContext[ "rem type", #]},
					If[ res =!= $Failed && res =!= False, remFormsFromTheory[ overlapKeys]];
					Sow[ #, res =!= $Failed]
				]&,
				types
			],
			{True, False, "Message"}
		],
		{_, {succ_List, fail_List, msg_List}} :>
			(
				coreNotification[ msg, False, {}, origin];
				{Join @@ succ, Join @@ fail}
			)
	]
remTypesFromTheory[ typ_, origin_] :=
	remTypesFromTheory[ {typ}, origin]

getTypedefOverlapKeyPatts[ typ_] :=
	getTypedefOverlapKeyPatts[ typ, {}]
getTypedefOverlapKeyPatts[ (TVAR$|TFIX$)[ ___], _List] :=
	{}
getTypedefOverlapKeyPatts[ typ:((h_Symbol)[ ___]), excpt_List] :=
	With[ {tvarNum = getMaxTVar[ typ] + 1},
		Cases[ getTypeDefs[ False][ h],
			(t_ -> {_, {id_String, src_String}, __}) /; ListQ[ unifyTypes[ typ, shiftTypeVars[ t, tvarNum], None&, {None, {}}, 0]] :>
				With[ {test = StringMatchQ[ #, id <> "." ~~ __]&},
					If[ excpt === {},
						{_String?test, src},
					(*else*)
						{Except[ Alternatives @@ Map[ id <> "." <> # &, excpt], _String?test], src}
					]
				]
		]
	]
getTypedefOverlapKeyPatts[ typ_Symbol, excpt_List] :=
	Cases[ getTypeDefs[ False][ typ],
		(typ -> {_, {id_String, src_String}, __}) :>
			With[ {test = StringMatchQ[ #, id <> "." ~~ __]&},
				If[ excpt === {},
					{_String?test, src},
				(*else*)
					{Except[ Alternatives @@ Map[ id <> "." <> # &, excpt], _String?test], src}
				]
			]
	]
getTypedefOverlapKeyPatts[ _, _] :=
	{}

getTypeOverlaps[ typ_] :=
	getTypeOverlaps[ typ, getTypeDecls]
getTypeOverlaps[ typ_, coll_Association, rest___] :=
	getTypeOverlaps[ typ, (#[[1, 1]] -> Last[ #])& /@ Lookup[ coll, Replace[ typ, {(h_Symbol)[ ___] :> h, Except[ _Symbol] -> $Failed}], {}], rest]
getTypeOverlaps[ _, {}, ___] :=
	{}
getTypeOverlaps[ typ_, decls_List] :=
	getTypeOverlaps[ typ, decls, getMaxTVar[ typ] + 1]
getTypeOverlaps[ typ_, decls_List, tvarNum_Integer] :=
	Position[ decls, (t_ -> _) /; ListQ[ unifyTypes[ typ, shiftTypeVars[ t, tvarNum], None&, {None, {}}, 0]], {1}, Heads -> False]
		(* Unification intentionally happens without sort constraints. The RHS of 't_ -> _' could be both
			a pair and a COREVAR$-structure. *)
getTypeOverlaps[ typ_, coll_, rest___] :=
	getTypeOverlaps[ typ, coll[ Replace[ typ, {(h_Symbol)[ ___] :> h, Except[ _Symbol] -> $Failed}]], rest]


(* ::Subsection:: *)
(*Type Definitions*)

copyType[ typ0_, rt0_, abs_, rep_][ False, i_, id_, src_, cell_CellObject, nb_NotebookObject] :=
	Replace[
		msgID = None;
		Reap[
			With[ {tl = Replace[ {typ0, rt0}, t_EXPR$ :> makeSTERM[ t], {1}]},
				Replace[
					If[ tl === {typ0, rt0},
						tl,
					(*else*)
						First[ generalizeTypes[ makeSingleSTERM[ tl], {}]]	(* type constants must be generalized together *)
					],
					STERM$[ {typ_, rt_}, tfix_Association, tvar_Association] :>		(* 'tfix' should be empty anyway *)
						defineTypeAux[
							typ,
							STERM$[ Lambda$TM[ RNG$[ TYPRNG$[ Null, rt]], getCoreVar[ $objectValidity]], tfix, tvar],
							Join[
								Hold[
									INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$thesis$TM, 0]],
									INF$[ "allE", 1, makeSTERM[ TYPED$[ Theorema`Knowledge`TYPE$TM, {rt}]]],
									INF$[ "impE", 1, 0]
								],
								getCoreVar[ $objectValidityProof],
								Hold[ INF$[ "close"]]
							],
							abs,
							rep,
							{cellIDLabel[ {id, i}], sourceLabel[ src]},
							Automatic,
							getEvaluationPosition[ id, nb],
							Interactive,
							getCoreVar[ $objectValidityProof],
							False
						]
				]
			],
			"Message"
		],
		{ret_, msg_List} :> If[ coreNotification[ msg, ret === $Failed, {}, {src, id}], $Failed, Null]
	]
copyType[ typ_, _][ True, args___] :=
	declareTypes[ {typ}][ True, args]

defineType[ typ0_, pred0_, prf_, abs_, rep_, lbl0_][ False, i_, id_, src_, cell_CellObject, nb_NotebookObject] :=
	Replace[
		msgID = None;
		Reap[
			Catch[
				With[ {ot = getObjectTruthType[], srcLabel = sourceLabel[ src], l0 = makeSingleSTERM[ {makeSTERM[ typ0], makeSTERM[ pred0]}]},
				With[ {sterm =
						If[ MatchQ[ pred0, _EXPR$],
							With[ {p1 = eliminateAnonymousVars[ l0[[1, 2]]]},
								Replace[ parseCoreTermPost[ inferTypes[ STERM$[ First[ p1], l0[[2]], l0[[3]]], Blank[], Last[ p1], 0, {}, {}, {}, l0[[2]], l0[[3]], None], normalize],
									(* ATTENTION! Type variables in 'pred0' must not be lifted, since they must agree with those in 'typ'! *)
									{
										{{FUN$TM[ rt_, ot], STERM$[ p_, newTFix_, newTVar_], _, {}, {}}, _} :>
											STERM$[ {l0[[1, 1]], rt, p}, Join[ l0[[2]], newTFix], newTVar],
										_ :> (sowMessage[ 214, {l0[[1, 1]]}, "Fatal"]; Throw[ $Failed])
									}
								]
							],
						(*else*)
							Replace[ checkType[ l0[[1, 2]], Blank[], {}, {}, {}, l0[[2]], l0[[3]]],
								{
									{FUN$TM[ rt_, ot], _, _, newTFix_Association, newTVar_Association} :>
										STERM$[ {l0[[1, 1]], rt, l0[[1, 2]]}, Join[ l0[[2]], newTFix], newTVar],
									_ :> (sowMessage[ 214, {l0[[1, 1]]}, "Fatal"]; Throw[ $Failed])
								}
							]
						]},
					Replace[ generalizeTypes[ sterm, {}],		(* type constants must be generalized together, and *after* inferring types *)
						{STERM$[ {typ_, rt_, pred_}, Association[], tv_Association], idx_} :>
							If[ MatchQ[ prf, _Hold],
								If[ StringQ[ lbl0], sowMessage[ 219, {lbl0}, "Warning"]];
								defineTypeAux[
									typ,
									STERM$[ pred, Association[], tv],
									getCoreVar[ $typedefProofTransformer][ prf],
									abs,
									rep,
									{cellIDLabel[ {id, i}], srcLabel},
									Automatic,
									getEvaluationPosition[ id, nb],
									Interactive,
									None,
									True
								],
							(*else*)
								(* Look at formulas in the context with the right label only. *)
								With[ {str = StringJoin[ Riffle[ tmaSymbolToString /@ typeToConstrList[ typ], $symbolCombinationString]]},
								With[ {lbl = If[ StringQ[ lbl0], lbl0, str <> " not empty"],
										f = getCoreVar[ $typedefFormulaConstructor][ rt, pred, True]},
								Block[ {$instantiateTypes = True,
										$proofGoal = FML$[ {"", srcLabel}, f, lbl, "pos" -> getEvaluationPosition[ id, nb]],
										$currentScope = {{"", srcLabel}, getEvaluationPosition[ id, nb]}},
										(* The position of the evaluation cell is not exactly the same as the position of the
											new formula cell (if one must be added), but the order of cells is correct. *)
								Module[ {newCellID = False},
								With[ {fkey =
										SelectFirst[ formulaKeys[ lbl, Infinity, None, getFormulas[ True], inScopeQ],
											With[ {f0 = extractFormula[ getFormulas[ False][ #], idx]},		(* argument of 'getFormulas' is irrelevant; 'f0' contains a 'STERM$'-structure *)
											With[ {tunif = If[ f0 === $Failed, $Failed, alphaEquiv[ f, f0[[1, 1]], Association[], Join[ tv, f0[[1, 3]]]]]},
												ListQ[ tunif] && MatchQ[ Last[ tunif], {(_ -> _TVAR$)...}] && DuplicateFreeQ[ tunif[[2, All, 2]]]
											]]&,
											
											If[ TrueQ[ getTheoryMode[]],
												sowMessage[ 221, {typ, lbl}, "Fatal"];
												Throw[ $Failed],
											(*else*)
												newCellID = generateNewCellID[ nb];
												{cellIDLabel[ newCellID], srcLabel}
											]
										],
										ac = assmCount[ f, 0, Infinity]},
									If[ IntegerQ[ newCellID],
										(* add formula to context and adjust position of evaluation cell *)
										$proofGoal = ReplacePart[ $proofGoal, 1 -> fkey];
										$currentScope = ReplacePart[ $currentScope, 1 -> fkey];
										increaseEvaluationPosition[ 1];
										If[ addAxmToTheory[ $proofGoal, tv, False, False, False, True, False] =!= True, Throw[ $Failed]]
									];
									With[ {aux =
											defineTypeAux[
												typ,
												STERM$[ pred, Association[], tv],
												getCoreVar[ $typedefProofTransformer][
													Hold[
														INF$[ "formula", fkey],
														INF$[ "impE", 1, ac],
														INF$[ "close"]
													]
												],
												abs,
												rep,
												{cellIDLabel[ {id, i}], srcLabel},
												Automatic,
												getEvaluationPosition[ id, nb],
												Interactive,
												None,
												False
											]},
										If[ IntegerQ[ newCellID],
											If[ aux === $Failed || aux === False,
												(* remove formula from context and reset position of evaluation cell *)
												increaseEvaluationPosition[ -1];
												remFormsFromTheory[ {fkey}, False],
											(*else*)
												With[ {orig = makeSTERM[ pred0]},
													(* add cell to notebook; use 'orig' for that purpose *)
													SelectionMove[ cell, Before, Cell, AutoScroll -> False];
													insertNewEnv[ "LMA",
														nb,
														"type " <> str <> " is non-empty",
														{
															Cell[
																theoremaInputDisplay[
																	printCoreTerm[
																		getCoreVar[ $typedefFormulaConstructor][ sterm[[1, 2]], orig, False],
																		{{}, {}, sterm[[2]], sterm[[3]]},
																		False
																	]
																],
																"FormalTextInputFormula",
																CellTags -> {lbl},
																ShowCellTags -> False,
																CellFrameLabels -> makeCellFrameLabel[ {lbl}, True],
																CellID -> newCellID
															]
														}
													];
													SelectionMove[ cell, After, Cell]
												]
											]
										];
										aux
									]
								]]]]]
							]
					]
				]]
			],
			"Message"
		],
		{ret_, msg_} :> If[ coreNotification[ msg, False, {}, {src, id}], $Failed, Null]
	]
defineType[ typ_, _, _, _, _, _][ True, args___] :=
	declareTypes[ {typ}][ True, args]

defaultTypedefFormulaConstructor[ rt_, STERM$[ pred_, tfix_, tvar_], types_] :=
	STERM$[ defaultTypedefFormulaConstructor[ rt, pred, types], tfix, tvar]
defaultTypedefFormulaConstructor[ rt_, pred_, types_] :=
	With[ {ot = getObjectTruthType[]},
	With[ {tj = If[ ot === FORM$TM || types =!= True, Identity, TYPED$[ TrueQ$TM, {}]]},
		Replace[
			If[ TrueQ[ types],
				{TYPED$[ CoreAll$TM, {ot}], TYPED$[ CoreAll$TM, {rt}], TYPED$[ CoreImplies$TM, {}]},
			(*else*)
				{CoreAll$TM, CoreAll$TM, CoreImplies$TM}
			],
			{allForm_, allRT_, imp_} :>
				allForm[
					Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$thesis$TM, ot]],
						imp[
							allRT[
								Lambda$TM[ RNG$[ TYPRNG$[ Replace[ pred, {Lambda$TM[ RNG$[ _[ x_, _]], _] :> x, _ -> Null}], rt]],
									imp[ tj[ applyFunction[ pred, VAR$[ 0], False]]][ tj[ VAR$[ 1]]]
								]
							]
						][
							tj[ VAR$[ 0]]
						]
					]
				]
		]
	]]

(* 'typ' is supposed to be a plain type without sort constraints, whereas 'pred' is supposed to be a 'STERM$'-structure. *)
defineTypeAux[ typ_, pred_, prf_] :=
	defineTypeAux[ typ, pred, prf, Automatic]
defineTypeAux[ typ_, pred_, prf_, abs_] :=
	defineTypeAux[ typ, pred, prf, abs, Automatic]
defineTypeAux[ typ_, pred_, prf_, abs_, rep_] :=
	defineTypeAux[ typ, pred, prf, abs, rep, makeFormulaKey[ 0]]
defineTypeAux[ typ_, pred_, prf_, abs_, rep_, k_] :=
	defineTypeAux[ typ, pred, prf, abs, rep, k, Automatic]
defineTypeAux[ typ_, pred_, prf_, abs_, rep_, k_, lbls_] :=
	defineTypeAux[ typ, pred, prf, abs, rep, k, lbls, None]
defineTypeAux[ typ_, pred_, prf_, abs_, rep_, k_, lbls_, pos_] :=
	defineTypeAux[ typ, pred, prf, abs, rep, k, lbls, pos, Interactive]
defineTypeAux[ typ_, pred_, prf_, abs_, rep_, k_, lbls_, pos_, overwrite_] :=
	defineTypeAux[ typ, pred, prf, abs, rep, k, lbls, pos, overwrite, None]
defineTypeAux[ typ_, pred_, prf_, abs_, rep_, k_, lbls_, pos_, overwrite_, predPrf_] :=
	defineTypeAux[ typ, pred, prf, abs, rep, k, lbls, pos, overwrite, predPrf, True]
defineTypeAux[ typ_, pred_, prf_, abs0_, rep0_, k_, lbls0_, pos_, overwrite_, predPrf_, check_] :=
	With[ {base =
			If[ abs0 === Automatic || rep0 === Automatic,
				"Theorema`Knowledge`" <> StringJoin[ Riffle[ tmaSymbolToString /@ typeToConstrList[ typ], $symbolCombinationString]],
			(*else*)
				Null
			]},
	With[ {abs = If[ abs0 === Automatic, Symbol[ base <> $symbolCombinationString <> "abs$TM"], abs0],
			rep = If[ rep0 === Automatic, Symbol[ base <> $symbolCombinationString <> "rep$TM"], rep0]},
	With[ {absName = tmaSymbolToString[ abs], repName = tmaSymbolToString[ rep]},
	With[ {lbls = If[ lbls0 === Automatic, {repName, repName <> " inverse", absName <> " inverse"}, lbls0]},
		defineTypeAux[ typ, pred, prf, abs, rep, k, lbls, pos, overwrite, predPrf, absName, repName, check]
	]]]]
defineTypeAux[ typ0_, pred_STERM$, prf_, abs_, rep_, k:{id_String, src_String}, lbls_List, pos_, overwrite_, predPrf_, absName_String, repName_String, check_] :=
	Which[
		getTypeOverlaps[ typ0, $abbrevTypeDecls] =!= {},
		sowMessage[ 236, {typ0}, "Fatal"];
		$Failed,
	(*else*)
		Lookup[ $abbrevConstDecls, abs, $Failed] =!= $Failed,
		sowMessage[ 231, {abs}, "Fatal"];
		$Failed,
	(*else*)
		Lookup[ $abbrevConstDecls, rep, $Failed] =!= $Failed,
		sowMessage[ 231, {rep}, "Fatal"];
		$Failed,
	(*else*)
		True,
		With[ {overwriteF = Switch[ overwrite, True, True&, False, False&, _, overwriteTypesDialog],
				overlapKeys = getTypedefOverlapKeyPatts[ typ0, {"1", "2", "3"}]},
			Replace[ manipulateContext[ "typedef", typ0, pred, prf, abs, rep, k, lbls, If[ ListQ[ pos], {"pos" -> pos}, {}], overwriteF, check],
				{head_, typ_} :>
					(
						remFormsFromTheory[ overlapKeys];
						Replace[
							Replace[ Replace[ typ, getTypeDefs[ False][ head]],
								{_, {id1_String, src1_String}, pred1_, tvp_List, abs1_, rep1_, _List} :>
									(* 'abs1' and 'rep1' should be the same as 'abs' and 'rep', but just to be sure ... *)
									Replace[ getConstDecls[ False][ abs1],
										{FUN$TM[ rt_, _], num_Integer, _List} :>
											With[ {tvars = Array[ TVAR$, num, 0]},
											With[ {abs2 = TYPED$[ abs1, tvars], rep2 = TYPED$[ rep1, tvars],
													repKey = {id1 <> ".1", src1}, repInvKey = {id1 <> ".2", src1}, absInvKey = {id1 <> ".3", src1}},
												getCoreVar[ $typedefThmGenerator][
													rep1, abs1, rep2, abs2, pred1, typ, rt, tvpToSorts[ pred1, tvp], predPrf,
													repName, absName, id1, src1, pos, repKey, repInvKey, absInvKey
												]
											]]
									]
							],
							ret_List :> (If[ MemberQ[ ret, False], sowMessage[ 215, {typ}, "Warning"]]; ret)
						]
					)
			]
		]
	]

addCoreTypedefThms[ args___] :=
	Join[
		addCoreTypedefInductThms[ args],
		addCoreTypedefCasesThms[ args],
		addCoreTypedefInjectThms[ args]
	]

addCoreTypedefInductThms[ rep0_, abs0_, rep_, abs_, pred_, typ_, rt_, tvar_, predPrf_, repName_String, absName_String, id_String, src_String, pos_List,
		repKey_List, repInvKey_List, absInvKey_List] :=
	With[ {trivPred = If[ MatchQ[ predPrf, _Hold], First, Identity], ot = getObjectTruthType[]},
	With[ {tj = If[ ot === FORM$TM, Identity, TYPED$[ TrueQ$TM, {}]]},
	With[ {predVar0 = If[ trivPred === First, Null, tj[ applyFunction[ pred, VAR$[ 0], False]]],
			posData = If[ ListQ[ pos], "pos" -> pos, Sequence @@ {}]},
		{
			(* "rep induct": '/\P::(rt->FORM) y::rt. (tj[ pred[ y]] ==>) (/\x::typ. P[ rep[ x]]) ==> P[ y]' *)
			addThmToTheory[
				FML$[ {id <> ".4", src},
					TYPED$[ CoreAll$TM, {FUN$TM[ rt, FORM$TM]}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$P$TM, FUN$TM[ rt, FORM$TM]]],
						TYPED$[ CoreAll$TM, {rt}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$y$TM, rt]],
							trivPred[
								TYPED$[ CoreImplies$TM, {}][
									predVar0
								][
									TYPED$[ CoreImplies$TM, {}][
										TYPED$[ CoreAll$TM, {typ}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$x$TM, typ]],
											VAR$[ 2][ rep[ VAR$[ 0]]]
										]]
									][
										VAR$[ 1][ VAR$[ 0]]
									]
								]
							]
						]]
					]],
					repName <> " induct",
					posData
				],
				tvar,
				Join[
					Hold[
						INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$P$TM, 0]],
						INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]
					],
					If[ trivPred === First,
						Hold[],
					(*else*)
						Hold[
							INF$[ "makeImplicit", 1],
							INF$[ "drop", 1]
						]
					],
					Hold[
						Theorema`CoreProvers`instantiateCoreAll[
							{EXPR$[ abs0[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]}
						],
						Theorema`CoreProvers`setupSubgoal[
							EXPR$[ CoreEqual$TM[ rep0[ abs0[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]][ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]
						]
					],
					If[ trivPred === First,
						Join[
							Hold[
								INF$[ "formula", absInvKey],
								INF$[ "allE", 1, makeSTERM[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]],
								INF$[ "impE", 1, 0]
							],
							predPrf
						],
					(*else*)
						Hold[
							INF$[ "makeExplicit", 1],
							INF$[ "formula", absInvKey],
							INF$[ "allE", 1, makeSTERM[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]],
							INF$[ "impE", 1, 1]
						]
					],
					Hold[
						INF$[ "close"],
						INF$[ "makeExplicit", 1],
						(* 'rep[ abs[ y]] === y ==> P[ rep[ abs[ y]]] ==> P[ y]' *)
						INF$[ "replace", 1, {1, 0, 1, 1}],
						INF$[ "impE", 2, 0],
						INF$[ "close"]
					]
				],
				False,
				False,
				False,
				True,
				False
			],
			
			(* "abs induct": '/\P::(typ->FORM) x::typ. (/\y::rt. (tj[ pred[ y]] ==>) P[ abs[ y]]) ==> P[ x]' *)
			addThmToTheory[
				FML$[ {id <> ".5", src},
					TYPED$[ CoreAll$TM, {FUN$TM[ typ, FORM$TM]}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$P$TM, FUN$TM[ typ, FORM$TM]]],
						TYPED$[ CoreAll$TM, {typ}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$x$TM, typ]],
							TYPED$[ CoreImplies$TM, {}][
								TYPED$[ CoreAll$TM, {rt}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$y$TM, rt]],
									trivPred[
										TYPED$[ CoreImplies$TM, {}][
											predVar0
										][
											VAR$[ 2][ abs[ VAR$[ 0]]]
										]
									]
								]]
							][
								VAR$[ 1][ VAR$[ 0]]
							]
						]]
					]],
					absName <> " induct",
					posData
				],
				tvar,
				Join[
					Hold[
						INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$P$TM, 0]],
						INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$x$TM, 0]],
						Theorema`CoreProvers`instantiateCoreAll[
							{EXPR$[ rep0[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]]}
						],
						INF$[ "formula", repInvKey],
						INF$[ "allE", 1, makeSTERM[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]]
					],
					Hold @@ {
						(* 'abs[ rep[ x]] === x ==> ((pred[ rep[ x]] ==>) P[ abs[ rep[ x]]]) ==> P[ x]' *)
						INF$[ "replace", 1, If[ trivPred === First, {1, 0, 1, 1}, {1, 0, 1, 1, 1}]]
					},
					Hold[
						INF$[ "drop", 1],
						INF$[ "impE", 1, 0]
					],
					If[ trivPred === First,
						Hold[],
					(*else*)
						Hold[
							INF$[ "formula", repKey],
							INF$[ "allE", 1, makeSTERM[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]],
							INF$[ "impE", 1, 0]
						]
					],
					Hold[
						INF$[ "close"]
					]
				],
				False,
				False,
				False,
				True,
				False
			]
		}
	]]]

addCoreTypedefCasesThms[ rep0_, abs0_, rep_, abs_, pred_, typ_, rt_, tvar_, predPrf_, repName_String, absName_String, id_String, src_String, pos_List,
		repKey_List, repInvKey_List, absInvKey_List] :=
	With[ {trivPred = If[ MatchQ[ predPrf, _Hold], First, Identity], ot = getObjectTruthType[]},
	With[ {tj = If[ ot === FORM$TM, Identity, TYPED$[ TrueQ$TM, {}]]},
	With[ {predVar0 = If[ trivPred === First, Null, tj[ applyFunction[ pred, VAR$[ 0], False]]],
			posData = If[ ListQ[ pos], "pos" -> pos, Sequence @@ {}]},
		{
			(* "rep cases": '/\P::FORM y::rt. (tj[ pred[ y]] ==>) (/\x::typ. y === rep[ x] ==> P) ==> P' *)
			addThmToTheory[
				FML$[ {id <> ".6", src},
					TYPED$[ CoreAll$TM, {FORM$TM}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$P$TM, FORM$TM]],
						TYPED$[ CoreAll$TM, {rt}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$y$TM, rt]],
							trivPred[
								TYPED$[ CoreImplies$TM, {}][
									predVar0
								][
									TYPED$[ CoreImplies$TM, {}][
										TYPED$[ CoreAll$TM, {typ}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$x$TM, typ]],
											TYPED$[ CoreImplies$TM, {}][
												TYPED$[ CoreEqual$TM, {rt}][ VAR$[1]][ rep[ VAR$[ 0]]]
											][
												VAR$[ 2]
											]]
										]
									][
										VAR$[ 1]
									]
								]
							]
						]]
					]],
					repName <> " cases",
					posData
				],
				tvar,
				Join[
					Hold @@ {
						INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$P$TM, 0]],
						INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$y$TM, 0]],
						Theorema`CoreProvers`instantiateCoreAll[
							{EXPR$[ abs0[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]},
							If[ trivPred === First, 1, 2]
						],
						INF$[ "impE", If[ trivPred === First, 1, 2], 0]
					},
					If[ trivPred === First,
						Hold[],
					(*else*)
						Hold[
							INF$[ "makeImplicit", 1],
							INF$[ "drop", 1]
						]
					],
					Hold[
						Theorema`CoreProvers`setupSubgoal[
							EXPR$[ CoreEqual$TM[ rep0[ abs0[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]][ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]
						]
					],
					If[ trivPred === First,
						Join[
							Hold[
								INF$[ "formula", absInvKey],
								INF$[ "allE", 1, makeSTERM[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]],
								INF$[ "impE", 1, 0]
							],
							predPrf
						],
					(*else*)
						Hold[
							INF$[ "makeExplicit", 1],
							INF$[ "formula", absInvKey],
							INF$[ "allE", 1, makeSTERM[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]],
							INF$[ "impE", 1, 1]
						]
					],
					Hold[
						INF$[ "close"],
						INF$[ "makeExplicit", 1],
						(* 'rep[ abs[ y]] === y ==> y === rep[ abs[ y]]' *)
						INF$[ "replace", 1, {1, 1}],
						INF$[ "refl"],
						INF$[ "close"]
					]
				],
				False,
				False,
				False,
				True,
				False
			],
			
			(* "abs cases": '/\P::FORM x::typ. (/\y::rt. x === abs[ y] ==> (tj[ pred[ y]] ==>) P) ==> P' *)
			addThmToTheory[
				FML$[ {id <> ".7", src},
					TYPED$[ CoreAll$TM, {FORM$TM}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$P$TM, FORM$TM]],
						TYPED$[ CoreAll$TM, {typ}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$x$TM, typ]],
							TYPED$[ CoreImplies$TM, {}][
								TYPED$[ CoreAll$TM, {rt}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$y$TM, rt]],
									TYPED$[ CoreImplies$TM, {}][
										TYPED$[ CoreEqual$TM, {typ}][ VAR$[ 1]][ abs[ VAR$[ 0]]]
									][
										trivPred[
											TYPED$[ CoreImplies$TM, {}][
												predVar0
											][
												VAR$[ 2]
											]
										]
									]
								]]
							][
								VAR$[ 1]
							]
						]]
					]],
					absName <> " cases",
					posData
				],
				tvar,
				Join[
					Hold[
						INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$P$TM, 0]],
						INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$x$TM, 0]],
						Theorema`CoreProvers`instantiateCoreAll[
							{EXPR$[ rep0[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]]}
						],
						INF$[ "impE", 1, 0],
						INF$[ "formula", repInvKey],
						INF$[ "allE", 1, makeSTERM[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]],
						(* 'abs[ rep[ x]] === x ==> x === abs[ rep[ x]]' *)
						INF$[ "replace", 1, {1, 1}],
						INF$[ "refl"]
					],
					If[ trivPred === First,
						Hold[],
					(*else*)
						Hold[
							INF$[ "formula", repKey],
							INF$[ "allE", 1, makeSTERM[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]],
							INF$[ "impE", 1, 0]
						]
					],
					Hold[
						INF$[ "close"]
					]
				],
				False,
				False,
				False,
				True,
				False
			]
		}
	]]]

addCoreTypedefInjectThms[ rep0_, abs0_, rep_, abs_, pred_, typ_, rt_, tvar_, predPrf_, repName_String, absName_String, id_String, src_String, pos_List,
		repKey_List, repInvKey_List, absInvKey_List] :=
	With[ {trivPred = If[ MatchQ[ predPrf, _Hold], First, Identity], ot = getObjectTruthType[]},
	With[ {tj = If[ ot === FORM$TM, Identity, TYPED$[ TrueQ$TM, {}]]},
	With[ {predVar0 = If[ trivPred === First, Null, tj[ applyFunction[ pred, VAR$[ 0], False]]],
			posData = If[ ListQ[ pos], "pos" -> pos, Sequence @@ {}]},
		{
			(* "rep inject": '/\x y::typ. rep[ x] === rep[ y] ==> x === y' *)
			addThmToTheory[
				FML$[ {id <> ".8", src},
					TYPED$[ CoreAll$TM, {typ}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$x$TM, typ]],
						TYPED$[ CoreAll$TM, {typ}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$y$TM, typ]],
							TYPED$[ CoreImplies$TM, {}][
								TYPED$[ CoreEqual$TM, {rt}][ rep[ VAR$[ 1]]][ rep[ VAR$[ 0]]]
							][
								TYPED$[ CoreEqual$TM, {typ}][ VAR$[ 1]][ VAR$[ 0]]
							]
						]]
					]],
					repName <> " inject",
					posData
				],
				tvar,
				Hold[
					INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$x$TM, 0]],
					INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$y$TM, 0]],
					INF$[ "makeImplicit", 1],
					INF$[ "drop", 1],
					Theorema`CoreProvers`setupSubgoal[
						EXPR$[ CoreEqual$TM[ abs0[ rep0[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]]][ abs0[ rep0[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]]]
					],
					INF$[ "makeExplicit", 1],
					(* 'rep[ x] === rep[ y] ==> abs[ rep[ x]] === abs[ rep[ y]]' *)
					INF$[ "replace", 1, {1, 0, 1, 1}],
					INF$[ "refl"],
					INF$[ "close"],
					INF$[ "makeExplicit", 1],
					INF$[ "insert", repInvKey], (* formula is used twice, so we do not apply "formula" *)
					INF$[ "makeExplicit", 1],
					INF$[ "allE", 1, makeSTERM[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]],
					(* 'abs[ rep[ y]] === y ==> abs[ rep[ x]] === abs[ rep[ y]] ==> x === y' *)
					INF$[ "replace", 1, {1, 0, 1, 1}],
					INF$[ "drop", 1],
					INF$[ "makeExplicit", 1],
					INF$[ "allE", 1, makeSTERM[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]],
					(* 'abs[ rep[ x]] === x ==> abs[ rep[ x]] === y ==> x === y' *)
					INF$[ "replace", 1, {1, 0, 1, 0, 1}],
					INF$[ "impE", 2, 0],
					INF$[ "close"]
				],
				False,
				False,
				False,
				True,
				False
			],
			
			(* "abs inject": '/\x y. (tj[ pred[ x]] ==> tj[ pred[ y]] ==>) abs[ x] === abs[ y] ==> x === y' *)
			addThmToTheory[
				FML$[ {id <> ".9", src},
					TYPED$[ CoreAll$TM, {rt}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$x$TM, rt]],
						TYPED$[ CoreAll$TM, {rt}][ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$y$TM, rt]],
							trivPred[
								trivPred[
									TYPED$[ CoreImplies$TM, {}][
										If[ trivPred === First, Null, tj[ applyFunction[ pred, VAR$[ 1], False]]]
									][
										TYPED$[ CoreImplies$TM, {}][
											predVar0
										][
											TYPED$[ CoreImplies$TM, {}][
												TYPED$[ CoreEqual$TM, {typ}][ abs[ VAR$[ 1]]][ abs[ VAR$[ 0]]]
											][
												TYPED$[ CoreEqual$TM, {rt}][ VAR$[ 1]][ VAR$[ 0]]
											]
										]
									]
								]
							]
						]]
					]],
					absName <> " inject",
					posData
				],
				tvar,
				Join[
					Hold[
						INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$x$TM, 0]],
						INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]
					],
					If[ trivPred === First,
						Hold[],
					(*else*)
						Hold[
							INF$[ "makeImplicit", 1],
							INF$[ "drop", 1],
							INF$[ "makeImplicit", 1],
							INF$[ "drop", 1]
						]
					],
					Hold[
						INF$[ "makeImplicit", 1],
						INF$[ "drop", 1],
						Theorema`CoreProvers`setupSubgoal[
							EXPR$[ CoreEqual$TM[ rep0[ abs0[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]]][ rep0[ abs0[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]]]
						],
						INF$[ "makeExplicit", 1],
						(* 'abs[ x] === abs[ y] ==> rep[ abs[ x]] === rep[ abs[ y]]' *)
						INF$[ "replace", 1, {1, 0, 1, 1}],
						INF$[ "refl"],
						INF$[ "close"],
						INF$[ "makeExplicit", 1],
						INF$[ "insert", absInvKey], (* formula is used twice, so we do not apply "formula" *)
						Theorema`CoreProvers`setupSubgoal[
							EXPR$[ CoreEqual$TM[ rep0[ abs0[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]]][ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]]
						]
					],
					If[ trivPred === First,
						Hold[],
					(*else*)
						Hold[
							INF$[ "makeExplicit", 5]
						]
					],
					Hold[
						INF$[ "makeExplicit", 1],
						INF$[ "allE", 1, makeSTERM[ FIX$[ Theorema`Knowledge`VAR$x$TM, 0]]]
					],
					If[ trivPred === First,
						Join[
							Hold[
								INF$[ "impE", 1, 0]
							],
							predPrf
						],
					(*else*)
						Hold[
							INF$[ "impE", 1, 1]
						]
					],
					Hold[
						INF$[ "close"],
						INF$[ "makeExplicit", 1],
						(* 'rep[ abs[ x]] === x ==> rep[ abs[ x]] === rep[ abs[ y]] ==> x === y' *)
						INF$[ "replace", 1, {1, 0, 1, 0, 1}],
						INF$[ "drop", 1],
						Theorema`CoreProvers`setupSubgoal[
							EXPR$[ CoreEqual$TM[ rep0[ abs0[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]][ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]
						]
					],
					If[ trivPred === First,
						Hold[],
					(*else*)
						Hold[
							INF$[ "makeExplicit", 5]
						]
					],
					Hold[
						INF$[ "makeExplicit", 2],
						INF$[ "allE", 1, makeSTERM[ FIX$[ Theorema`Knowledge`VAR$y$TM, 0]]]
					],
					If[ trivPred === First,
						Join[
							Hold[
								INF$[ "impE", 1, 0]
							],
							predPrf
						],
					(*else*)
						Hold[
							INF$[ "impE", 1, 1]
						]
					],
					Hold[
						INF$[ "close"],
						INF$[ "makeExplicit", 1],
						(* 'rep[ abs[ y]] === y ==> x === rep[ abs[ y]] ==> x === y' *)
						INF$[ "replace", 1, {1, 0, 1, 1}],
						INF$[ "impE", 2, 0],
						INF$[ "close"]
					]
				],
				False,
				False,
				False,
				True,
				False
			]
		}
	]]]

typeToConstrList[ _TVAR$|_TFIX$] :=
	{}
typeToConstrList[ h_[ a___]] :=
	Join[ typeToConstrList[ h], Join @@ typeToConstrList /@ {a}]
typeToConstrList[ h_Symbol] :=
	{h}
typeToConstrList[ _] :=
	{}

overwriteTypesDialog[ old_List] :=
	dialogYesNo[ 213, old]


(* ::Subsection:: *)
(*Constant Declarations*)

declareConstants[ consts_List][ False, i_, id_, src_, ___] :=
	(declareConstantsDirect[ Replace[ consts, e_EXPR$ :> First[ generalizeTypes[ makeSTERM[ e], {}]], {1}], cellIDLabel[ {id, i, #}]&, sourceLabel[ src], Interactive, {src, id}];)
declareConstants[ consts_List][ True, _, id_, src_, ___] :=
	(undeclareConstantsDirect[ Replace[ consts, EXPR$[ e_, ___] :> First[ generalizeTypes[ e, {}]], {1}], {src, id}];)

(* 'consts' is supposed to be a list of 'STERM$'-structures. *)
declareConstantsDirect[ consts_, {id_String, src_String}, rest___] :=
	declareConstantsDirect[ consts, id, src, rest]
declareConstantsDirect[ consts_, id_, src_String] :=
	declareConstantsDirect[ consts, id, src, Interactive]
declareConstantsDirect[ consts_, id_, src_String, overwrite_] :=
	declareConstantsDirect[ consts, id, src, overwrite, None]
declareConstantsDirect[ consts_List, id_String, src_String, overwrite_, origin_] :=
	declareConstantsDirect[ consts, id&, src, overwrite, origin]
declareConstantsDirect[ consts_List, idF_, src_String, overwrite_, origin_] :=
	With[ {overwriteF = Switch[ overwrite, True, True&, False, False&, _, overwriteDeclDialog]},
	Module[ {i = 0},
		Replace[
			Last[
				Reap[
					Scan[
						Sow[ #,
							If[ Lookup[ $abbrevConstDecls, #[[1, 1]], $Failed] =!= $Failed,
								sowMessage[ 231, {First[ #]}, "Fatal"];
								False,
							(*else*)
								manipulateContext[ "decl", #, {idF[ ++i], src}, overwriteF] =!= $Failed
							]
						]&,
						Cases[ consts, STERM$[ TYPED$[ Except[ TrueQ$TM], _], _Association, _Association]]
					],
					{True, False, "Message"}
				]
			],
			{succ_List, fail_List, msg_List} :>
				(
					coreNotification[
						Join[
							msg,
							If[ MemberQ[ consts, STERM$[ Except[ _TYPED$], _, _]], {MSG$[ coreMessageToList, 205, Cases[ consts, STERM$[ c:Except[ _TYPED$], _, _] :> c], "Fatal"]}, {}],
							If[ MemberQ[ consts, STERM$[ TYPED$[ TrueQ$TM, _], _, _]], {MSG$[ coreMessageToList, 204, {TrueQ$TM}, "Fatal"]}, {}]
						],
						False,
						{},
						origin
					];
					{Join @@ succ, Join @@ fail}
				)
		]
	]]
declareConstantsDirect[ const_, id_, src_String, overwrite_, origin_] :=
	declareConstantsDirect[ {const}, id, src, overwrite, origin]

(* 'consts' is supposed to be a list of plain constants without sort constraints. *)
undeclareConstantsDirect[ consts_] :=
	undeclareConstantsDirect[ consts, None]
undeclareConstantsDirect[ consts_List, origin_] :=
	Replace[
		Last[
			Reap[
				Scan[
					Sow[ #, manipulateContext[ "rem decl", #] =!= $Failed]&,
					DeleteDuplicates[ DeleteCases[ Replace[ consts, TYPED$[ c_, _] :> c, {1}], TrueQ$TM]]
				],
				{True, False, "Message"}
			]
		],
		{succ_List, fail_List, msg_List} :>
			(
				coreNotification[
					Join[ msg, If[ MemberQ[ consts, TrueQ$TM|TYPED$[ TrueQ$TM, _]], {MSG$[ coreMessageToList, 211, {TrueQ$TM}, "Fatal"]}, {}]],
					False,
					{},
					origin
				];
				{Join @@ succ, Join @@ fail}
			)
	]
undeclareConstantsDirect[ const_, origin_] :=
	undeclareConstantsDirect[ {const}, origin]

overwriteDeclDialog[ const_, typ_] :=
	dialogYesNo[ 216, {const, typ}]


(* ::Subsection:: *)
(*Theorem Schemas*)

(* 'addThmSchemasToTheory' is no theory command! *)
addThmSchemasToTheory[ thms_List] :=
	addThmSchemasToTheory[ thms, getEvaluationTheoryName[]]
addThmSchemasToTheory[ thms0_List, thy_String] :=
	With[ {thms =
			With[ {pos = Position[ thms0, {Automatic, __}, {1}, Heads -> False]},
				If[ pos === {},
					thms0,
				(*else*)
					With[ {cid = CurrentValue[ getEvaluationCell[], CellID]},
						ReplacePart[ thms0, (Append[ #, 1] -> {cellIDLabel[ {cid, First[ #]}], thy})& /@ pos]
					]
				]
			]},
		Replace[ Last[ Reap[ Scan[ addThmSchemaToTheory, thms], {True, False, "Message"}]],
			{succ_List, fail_List, msg_List} :>
				(
					coreNotification[ msg];
					{Join @@ succ, Join @@ fail}
				)
		]
	]

System`AddTheoremSchemas = addThmSchemasToTheory;

addThmSchemaToTheory[ {k_, statement_}] :=
	addThmSchemaToTheory[ {k, statement, Hold[]&, Automatic, Automatic, Automatic, False}]
addThmSchemaToTheory[ {k_, statement_, proof_}] :=
	addThmSchemaToTheory[ {k, statement, proof, Automatic, Automatic, Automatic, False}]
addThmSchemaToTheory[ {k_, statement_, proof_, index_}] :=
	addThmSchemaToTheory[ {k, statement, proof, index, Automatic, Automatic, False}]
addThmSchemaToTheory[ {k_, statement_, proof_, index_, lbl_}] :=
	addThmSchemaToTheory[ {k, statement, proof, index, lbl, Automatic, False}]
addThmSchemaToTheory[ {k:{_, src_String}, statement_, proof_, index0_, lbl_, data_, addToFast_}] :=
	Sow[ k,
		If[ With[ {index = If[ index0 === Automatic, With[ {i = ($theoremIndex = increaseTheoremIndex[ $theoremIndex])}, i&], index0]},
				manipulateContext[ "thm schema", k, lbl, statement, proof, index]
			] === $Failed,
			If[ TrueQ[ addToFast] && StringQ[ lbl], KeyDropFrom[ $labelThmMapping, lbl]];
			False,
		(*else*)
			If[ Length[ retrieveFormulasFromContext[ key, {_String, src}, lbl, 2, None, getTheorems[ True]]] === 2,
				sowMessage[ 206, {lbl}, "Warning"]
			];
			If[ TrueQ[ addToFast] && StringQ[ lbl], AssociateTo[ $labelThmMapping, lbl -> k]];
			If[ Head[ data] === Association || StringQ[ lbl],
				If[ src === getTheory[],
					Which[
						!StringQ[ lbl],
						AssociateTo[ $thmLabelData, k -> {getThmLabel[ k], data}],
					(*else*)
						Head[ data] === Association,
						AssociateTo[ $thmLabelData, k -> {lbl, data}],
					(*else*)
						True,
						AssociateTo[ $thmLabelData, k -> {lbl, getThmData[ k]}]
					],
				(*else*)
					sowMessage[ 261, {}, "Warning"]
				]
			];
			True
		]
	]

remThmSchemasFromTheory[ keyPatts_List] :=
	With[ {keys = Join @@ (retrieveFormulasFromContext[ key, #, _, Infinity, None, getTheorems[ False]]& /@ keyPatts)},
		Replace[
			Reap[
				Scan[
					Sow[ #, manipulateContext[ "rem thm", #] =!= $Failed]&,
					keys
				],
				{True, False, "Message"}
			],
			{_, {succ0_List, fail_List, msg_List}} :>
				With[ {succ = Join @@ succ0},
					coreNotification[ msg];
					If[ succ =!= {}, KeyDropFrom[ $thmLabelData, succ]];
					{succ, Join @@ fail}
				]
		]
	]

getThmData[ THM$[ k_, ___]] :=
	getThmData[ k]
getThmData[ k_] :=
	Last[ Lookup[ $thmLabelData, Key[ k], {Association[]}]]

getThmLabel[ THM$[ k_, ___]] :=
	getThmLabel[ k]
getThmLabel[ k_] :=
	First[ Lookup[ $thmLabelData, Key[ k], {"?"}]]


(* ::Subsection:: *)
(*Formulas, Axioms, Theorems*)

addFmlToTheory[ fml_, tvar_, kind_] :=
	addFmlToTheory[ fml, tvar, kind, False, False, True, True]
addFmlToTheory[ fml_FML$, tvar_, "def", arData_, rwData_, checkDupLbls_, notif_] :=
	addAxmToTheory[ fml, tvar, True, arData, rwData, checkDupLbls, notif]
addFmlToTheory[ fml_FML$, tvar_, "axiom", arData_, rwData_, checkDupLbls_, notif_] :=
	addAxmToTheory[ fml, tvar, False, arData, rwData, checkDupLbls, notif]
addFmlToTheory[ fml_FML$, tvar_, "thm", arData_, rwData_, checkDupLbls_, notif_] :=
	With[ {prf = getProofOfFormula[ fml, Automatic]},
		Switch[ prf,
			_Hold|_List,
			addThmToTheory[ fml, tvar, Join[ getCoreVar[ $initialProofSteps], Hold @@ prf], False, arData, rwData, checkDupLbls, notif],
		(*else*)
			False,
			addAxmToTheory[ fml, tvar, False, arData, rwData, checkDupLbls, notif],
		(*else*)
			_,
			coreNotification[ {}, True, 209];
			False
		]
	]
addFmlToTheory[ fml_FML$, tvar_, prf_Hold, arData_, rwData_, checkDupLbls_, notif_] :=
	addThmToTheory[ fml, tvar, prf, True, arData, rwData, checkDupLbls, notif]
addFmlToTheory[ fml_FML$, tvar_, _, arData_, rwData_, checkDupLbls_, notif_] :=
	If[ TrueQ[ getTheoryMode[]],
		addFmlToTheory[ fml, tvar, "thm", arData, rwData, checkDupLbls, notif],
	(*else*)
		addAxmToTheory[ fml, tvar, False, arData, rwData, checkDupLbls, notif]
	]

addAxmToTheory[ fml_, tvar_Association] :=
	addAxmToTheory[ fml, tvar, False, False, False, True, True]
addAxmToTheory[ fml_FML$, tvar_Association, False, arData_, rwData_, checkDupLbls_, False] :=
	addFmlToTheoryAux[ manipulateContext[ "axiom", #, tvar, False]&, addRuleData[ fml, tvar, arData, rwData, False], checkDupLbls, key@fml]
addAxmToTheory[ fml_FML$, tvar_Association, True, arData_, rwData_, checkDupLbls_, False] :=
	With[ {k = key@fml},
		Replace[
			Replace[ formula@fml,
				{
					TYPED$[ CoreEqual$TM, {typ_}][ TYPED$[ const_, _]][ _] :>
						Which[
							Lookup[ $abbrevConstDecls, const, $Failed] =!= $Failed,
							sowMessage[ 231, {const}, "Warning"];
							Null,
						(*else*)
							getConstDecls[ True][ const] === $Failed,
							If[ manipulateContext[ "decl", STERM$[ TYPED$[ const, typ], Association[], tvar], k, True&] === $Failed,
								$Failed,
							(*else*)
								{const, getTypeInstance[ const, typ, Association[], tvar]}
							],
						(*else*)
							True,
							Null
						],
					_ -> Null
				}
			],
			{
				Null :> addFmlToTheoryAux[ manipulateContext[ "axiom", #, tvar, True]&, addRuleData[ fml, tvar, arData, rwData, False], checkDupLbls, k],
				{const_, {newTVar_, tl_List}} :>
					If[ TrueQ[ addFmlToTheoryAux[
							manipulateContext[ "axiom", #, newTVar, True]&,
							addRuleData[ ReplacePart[ fml, {2, 0, 1, 2} -> tl], newTVar, arData, rwData, False],
							checkDupLbls,
							k
						]],
					(*then*)
						True,
					(*else*)
						manipulateContext[ "rem decl", const];
						False
					],
				_ -> False
			}
		]
	]
addAxmToTheory[ fml_FML$, isDef_, arData_, rwData_, checkDupLbls_, True] :=
	Replace[ Reap[ addAxmToTheory[ fml, isDef, arData, rwData, checkDupLbls, False], "Message"],
		{ret_, msg_List} :> (coreNotification[ msg]; ret)
	]

addThmToTheory[ fml_, tvar_, prf_] :=
	addThmToTheory[ fml, tvar, prf, True]
addThmToTheory[ fml_, tvar_, prf_, check_] :=
	addThmToTheory[ fml, tvar, prf, check, False, False, True, True]
addThmToTheory[ fml_FML$, tvar_, prf_Hold, check_, arData_, rwData_, checkDupLbls_, False] :=
	With[ {k = key@fml},
		addFmlToTheoryAux[
			If[ manipulateContext[ "axiom", #, tvar, False] === $Failed, $Failed, manipulateContext[ "proof", k, prf, check]]&,
			addRuleData[ fml, tvar, arData, rwData, False],
			checkDupLbls,
			k
		]
	]
addThmToTheory[ fml_, tvar_, prf_, check_, arData_, rwData_, checkDupLbls_, True] :=
	Replace[ Reap[ addThmToTheory[ fml, tvar, prf, check, arData, rwData, checkDupLbls, False], "Message"],
		{ret_, msg_List} :> (coreNotification[ msg]; ret)
	]

addFmlToTheoryAux[ act_, fml_FML$] :=
	addFmlToTheoryAux[ act, fml, True]
addFmlToTheoryAux[ act_, fml_FML$, checkDupLbls_] :=
	addFmlToTheoryAux[ act, fml, checkDupLbls, key@fml]
addFmlToTheoryAux[ act_, fml_FML$, checkDupLbls_, k_] :=
	With[ {old = getFormulas[ False][ k]},
		If[ act[ fml] === $Failed,
			False,
		(*else*)
			If[ TrueQ[ checkDupLbls] && Length[ retrieveFormulasFromContext[ key, {_String, Last[ k]}, label@fml, 2, None, getFormulas[ True]]] === 2,
				sowMessage[ 206, {label@fml}, "Warning"]
			];
			updateRuleLists[ fml, old, getOptionalComponent[ fml, "pos"]];
			True
		]
	]

getProofOfFormula[ FML$[ k_, __], num_] :=
	getProofOfFormula[ k, num]
getProofOfFormula[ k_List, Automatic] :=
	With[ {src = sourceFile[ k]},
	With[ {dir = FileNameJoin[ {DirectoryName[ src], FileBaseName[ src]}], cid = extractCellID[ k]},
		Replace[ getRegisteredProofNumber[ k, src, dir, cid, "-nb.nb"],
			num_Integer :> getProofOfFormula[ dir, cid, num]
		]
	]]
getProofOfFormula[ k_List, num_Integer] :=
	With[ {src = sourceFile[ k]},
		getProofOfFormula[ FileNameJoin[ {DirectoryName[ src], FileBaseName[ src]}], extractCellID[ k], num]
	]
getProofOfFormula[ dir_String, cid_String, num_Integer] :=
	Replace[ Quiet[ Check[ Get[ FileNameJoin[ {dir, "p" <> cid <> "-" <> ToString[ num] <> "-nb.nb"}]], $Failed]],
		{
			(* If "ProofScript.m" has not been loaded, the RHS below does not evaluate and hence the proof notebook cannot be parsed. *)
			Notebook[ cells_List, ___] :> Theorema`Interface`ProofScript`cellsToProof[ cells, False],
			_ -> False
		}
	]

remFormsFromTheory[ keyPatts_List] :=
	remFormsFromTheory[ keyPatts, True]
remFormsFromTheory[ keyPatts_List, False] :=
	With[ {keys = Join @@ (retrieveFormulasFromContext[ key, #, _, Infinity, None, getFormulas[ False]]& /@ keyPatts)},
		Replace[
			Reap[
				Scan[
					With[ {fml = getFormulas[ False][ #]},
					With[ {r = manipulateContext[ "rem fml", #] =!= $Failed},
						Sow[ If[ r, fml, #], r]
					]]&,
					keys
				],
				{True, False}
			],
			{_, {succ0_List, fail_List}} :>
				(* 'succ0' contains formulas, whereas 'fail' contains keys *)
				With[ {succ = Join @@ succ0},
				With[ {succK = key /@ succ},
					If[ succ =!= {},
						$tmaEnv = DeleteCases[ $tmaEnv, FML$[ Alternatives @@ succK, ___]];
						Scan[
							updateRuleLists[ $Failed, #, getOptionalComponent[ #, "pos"]]&,
							succ
						];
						Scan[
							(
								If[ MemberQ[ DownValues[ kbSelectProve], #, {0, Infinity}, Heads -> True], Unset[ kbSelectProve[ #]]];
								If[ MemberQ[ DownValues[ kbSelectCompute], #, {0, Infinity}, Heads -> True], Unset[ kbSelectCompute[ #]]]
							)&,
							succK
						]
					];
					{succK, Join @@ fail}
				]]
		]
	]
remFormsFromTheory[ keyPatts_List, True] :=
	Replace[ Reap[ remFormsFromTheory[ keyPatts, False], "Message"],
		{ret_, msg_List} :> (coreNotification[ msg]; ret)
	]


(* ::Subsection:: *)
(*Poved Facts and Rules*)

addFactsToTheory[ l_List, coll_, add_, rest___][ False, i_Integer, id_Integer, src_String, cell_CellObject, nb_NotebookObject] :=
	With[ {srcLabel = sourceLabel[ src]},
		Replace[
			addFactsToTheoryDirect[
				Hold[ Theorema`CoreProvers`addKnowledgeImplicit[ l, Null, False, rest]],
				srcLabel,
				cellIDLabel[ {id, i, #2}]&,
				getEvaluationPosition[ id, nb],
				{src, id},
				updateRuleLists[ getEvaluationPosition[ id, nb]],
				True
			],
			res_List :>
				(
					If[ StringQ[ coll],
						With[ {ky = getFactListKeys[ coll, 1, srcLabel], idLabel = cellIDLabel[ {id, i}]},
							If[ ky === {},
								coreNotification[ {MSG$[ coreMessageToList, 255, {coll}, "Warning"]}, False, {}, {src, id}];
								Null,
							(*else*)
								With[ {res1 = {key[ #], {{idLabel, srcLabel}, getEvaluationPosition[ id, nb]}}& /@ res},
									Which[
										!add,
										AssociateTo[ $globalFactLists, First[ ky] -> res1],
									(*else*)
										res =!= {},
										AssociateTo[ $globalFactLists,
											First[ ky] -> Join[ Select[ Lookup[ $globalFactLists, Key @@ ky, {}], (!MemberQ[ res1, {First[ #], _}])&], res1]
										]
									]
								]
							]
						]
					];
					Null
				)
		]
	]
addFactsToTheory[ ___][ True, _, id_, src_, ___] :=
	(coreNotification[ {}, True, 218, {src, id}]; $Failed)

addFactsToTheoryDirect[ prf_Hold, src_String, baseID_String, offset_Integer, pos_, origin_, overwriteF_, notif_] :=
	addFactsToTheoryDirect[ prf, src, baseID <> "." <> ToString[ offset + #2] &, pos, origin, overwriteF, notif]
addFactsToTheoryDirect[ prf_Hold, src_String, baseID_String, Hold[ offset_], pos_, origin_, overwriteF_, notif_] :=
	addFactsToTheoryDirect[ prf, src, baseID <> "." <> ToString[ ++offset] &, pos, origin, overwriteF, notif]
addFactsToTheoryDirect[ prf_Hold, src_String, idFun_, pos_, {_, id_}, overwriteF_, False] :=
	With[ {new = manipulateContext[ "facts", prf, src, idFun, {cellIDLabel[ id], src}, If[ ListQ[ pos], {"pos" -> pos}, {}], overwriteF]},
		If[ ListQ[ new],
			Scan[
				(
					If[ Length[ retrieveFormulasFromContext[ key, {_String, Last[ key@#]}, label@#, 2, None, getFormulas[ True]]] === 2,
						sowMessage[ 206, {label@#}, "Warning"]
					]
				)&,
				new
			]
		];
		new
	]
addFactsToTheoryDirect[ prf_Hold, src_String, idFun_, pos_, origin_, overwriteF_, True] :=
	Replace[ Reap[ addFactsToTheoryDirect[ prf, src, idFun, pos, origin, overwriteF, False], "Message"],
		{new_, msg_List} :> (coreNotification[ msg, False, {}, origin]; new)
	]

(* The elements of parameter 'cats' can be a individual category-strings, or pairs consisting of a category-string and
	the variant-function that shall be applied, or 'Automatic'. At the moment it is not yet possible to pass
	variant-functions via top-level theory commands. *)
addRulesToTheory[ {}, _][ __] :=
	Null
addRulesToTheory[ cats_List, l_List][ undo_, i_Integer, id_Integer, src_String, cell_CellObject, nb_NotebookObject] :=
	Replace[
		Reap[
			With[ {srcLabel = sourceLabel[ src]},
			With[ {trueSrc = {cellIDLabel[ {id, i}], srcLabel}, pos = getEvaluationPosition[ id, nb]},
			With[ {keys =
					Block[ {$proofGoal = FML$[ trueSrc, Null, Null, "pos" -> pos], $currentScope = {trueSrc, pos}},
						unfoldFactList[ l, None, 250, 251, 252, 253, 254]
					]},
				If[ ListQ[ keys],
					If[ MemberQ[ keys, {_, Except[ srcLabel]}],
						sowMessage[ 260, {}, "Fatal"],
					(*else*)
						Scan[
							addRulesToTheoryAux[ #, keys, trueSrc, pos, undo]&,
							DeleteDuplicatesBy[ Replace[ cats, c:Except[ { _, _}] :> {c, Automatic}, {1}], First]
						]
					]
				]
			]]],
			"Message"
		],
		{_, msg_List} :> (coreNotification[ msg, False, {}, {src, id}];)
	]

addRulesToTheoryAux[ {"autored", var_}, keys_List, trueSrc_, pos_, undo_] :=
	addRulesToTheoryAux[
		"autored",
		"intro",
		Composition[ VARIANT$[ #, Theorema`CoreProvers`CoreProofMethods`Private`arVariant, {}]&, If[ var === Automatic, getCoreVar[ $autoredRuleVariant], var]],
		Hold[ $autoredRules],
		keys,
		trueSrc,
		pos,
		undo
	]
addRulesToTheoryAux[ {"elim", var_}, keys_List, trueSrc_, pos_, undo_] :=
	addRulesToTheoryAux[ "elim", "elim", If[ var === Automatic, getCoreVar[ $elimRuleVariant], var], Hold[ $elimRules], keys, trueSrc, pos, undo]
addRulesToTheoryAux[ {"intro", var_}, keys_List, trueSrc_, pos_, undo_] :=
	addRulesToTheoryAux[ "intro", "intro", If[ var === Automatic, getCoreVar[ $introRuleVariant], var], Hold[ $introRules], keys, trueSrc, pos, undo]
addRulesToTheoryAux[ {"sym", var_}, keys_List, trueSrc_, pos_, undo_] :=
	addRulesToTheoryAux[ "sym", "elim", If[ var === Automatic, getCoreVar[ $symRuleVariant], var], Hold[ $symRules], keys, trueSrc, pos, undo]
addRulesToTheoryAux[ {"trans", var_}, keys_List, trueSrc_, pos_, undo_] :=
	addRulesToTheoryAux[ "trans", "trans", If[ var === Automatic, getCoreVar[ $transRuleVariant], var], Hold[ $transRules], keys, trueSrc, pos, undo]
addRulesToTheoryAux[ {"simp", var_}, keys_List, trueSrc_, pos_, undo_] :=
	addRulesToTheoryAux[
		"simp",
		"simp",
		Composition[ VARIANT$[ #, Theorema`CoreProvers`CoreProofMethods`Private`rwVariant, {}]&, If[ var === Automatic, getCoreVar[ $simpRuleVariant], var]],
		Hold[ Null],
		keys,
		trueSrc,
		pos,
		undo
	]
addRulesToTheoryAux[ {cat_, _}, _List, _, _, _] :=
	(sowMessage[ 249, {cat}, "Warning"]; Null)
addRulesToTheoryAux[ lbl_String, ccat_, var_, coll_Hold, keys_List, trueSrc_, pos_, undo_] :=
	Scan[
		addSingleRuleToTheory[ #, lbl, ccat, var, coll, trueSrc, pos, undo]&,
		keys
	]

addSingleRuleToTheory[ ky:{id_String, src_String}, lbl_String, ccat_, var_, Hold[ coll_], trueSrc0_, pos_, undo_] :=
	With[ {cid = StringDrop[ id, 3], orig = getFormulas[ False][ ky]},
	With[ {obtainedRules = DeleteCases[ getRuleFormulas[ lbl, ky, False, False, False, False], ky], tagsOrig = getOptionalComponent[ orig, "tags"]},
		Catch[
			Scan[
				If[ manipulateContext[ "rem fml", #] === $Failed, Throw[ Null]]&,
				obtainedRules
			];
			With[ {rules = Alternatives @@ Append[ obtainedRules, ky]},
				If[ lbl === "simp",
					(* The rules must be removed from all collections, because of the following possible scenario:
						- Add formula 'f' without any tags.
						- Evaluate "rules (simp) 'f'".
						- Add tag "cong" to 'f'.
						- Evaluate "rules (simp) 'f'" again. *)
					remRulesByPattFromRULEC$[ $congRules, rules];
					remRulesByPattFromRULEC$[ $compRules, rules];
					remRulesByPattFromRULEC$[ $simpRules, rules],
				(*else*)
					remRulesByPattFromRULEC$[ coll, rules]
				]
			];
			If[ undo =!= True,
				With[ {trueSrc = If[ src === Last[ trueSrc0], First[ trueSrc0], trueSrc0],
						lblOrig = orig[[3]],
						coll1 =
							Which[
								lbl =!= "simp",
								Hold[ coll],
							(*else*)
								MemberQ[ tagsOrig, "cong"],
								Hold[ $congRules],
							(*else*)
								MemberQ[ tagsOrig, "comp"],
								Hold[ $compRules],
							(*else*)
								True,
								Hold[ $simpRules]
							]},
					addFactsToTheoryDirect[
						Hold[ Evaluate[ Theorema`CoreProvers`addKnowledgeImplicit[ {var[ ky]}, Null, False, 3, False, None]]],
						src,
						cellIDLabel[ {cid, lbl, #2}]&,	(* The new ID is not composed from the cell issuing this command, but from the ID of the original formula. *)
						pos,
						trueSrc0,
						With[ {tags =
								Which[
									First[ #1] === ky,
									DeleteDuplicates[ Append[ tagsOrig, lbl]],
								(*else*)
									lbl === "simp",
									Append[ Intersection[ tagsOrig, {"comp", "cong"}], lbl],
								(*else*)
									True,
									{lbl}
								]},
							updateRuleList[ coll1, First[ #1], getRuleConsts[ #1[[2]], ccat, True], False];
							If[ First[ #1] === ky,
								Prepend[ DeleteDuplicatesBy[ Prepend[ Drop[ #1, 3], "tags" -> tags], First], lblOrig],
							(*else*)
								Prepend[
									DeleteDuplicatesBy[
										Join[
											FML$[ "tags" -> tags, "trueSrc" -> trueSrc],
											If[ ListQ[ pos], FML$[ "pos" -> pos], FML$[]],
											Drop[ #1, 3]
										],
										First
									],
									With[ {fid = #1[[1, 1]]}, StringReplacePart[ fid, lblOrig <> " " <> lbl <> " ", {1, StringPosition[ fid, "."][[-1, 1]]}]]
								]
							]
						]&,
						False
					]
				]
			]
		]
	]]

unfoldFactList[ l_List, ruleCat_, thmErr_Integer, varErr_Integer, neErr_Integer, posErr_Integer, typErr_Integer] :=
	Catch[
		Replace[
			Reap[
				scanFactSpec[
					l,
					PRFSIT$[ {}, {}, 0, {}, {}, Association[], Association[], 0, Association[]],
					ruleCat,
					False,
					Replace[ #,
						{
							{___, {thm_THM$, _}, ___} :> (sowMessage[ thmErr, {thm}, "Fatal"]; Throw[ $Failed]),
							{___, {var_VARIANT$, _}, ___} :> (sowMessage[ varErr, {var}, "Fatal"]; Throw[ $Failed]),
							{{_List, _}..} :> Sow[ #[[All, 1]], 0]
						}
					]&,
					(sowMessage[ Switch[ #2, 0, neErr, 1, posErr, _, typErr], {#1}, "Fatal"]; Throw[ $Failed])&
				],
				0
			],
			{_, res_List} :> Flatten[ res, 2]
		]
	]


(* ::Subsection:: *)
(*Global Fact Lists*)

addGlobalFactList[ l0_List][ False, i_Integer, id_Integer, src_String, cell_CellObject, nb_NotebookObject] :=
	With[ {l = Replace[ l0, STRING$[ s_String] :> s, {1}]},
		Which[
			l === {},
			Null,
		(*else*)
			MemberQ[ l, Except[ _String]],
			coreNotification[ {}, True, 256, {src, id}];
			$Failed,
		(*else*)
			MemberQ[ l, _?(StringPosition[ #, "::", 1] =!= {} &)],
			coreNotification[ {}, True, 257, {src, id}];
			$Failed,
		(*else*)
			True,
			With[ {srcLabel = sourceLabel[ src], ids = ToString[ id]},
			With[ {dups = Cases[ Keys[ $globalFactLists], {n_String, {{cid_String, s_String}, _}} /; MemberQ[ l, n] && !(s === srcLabel && extractCellID[ cid] === ids)]},
			With[ {dups0 = Select[ dups, #[[2, 1, 2]] === srcLabel &]},
				If[ dups0 === {},
					If[ dups =!= {},
						coreNotification[ {MSG$[ coreMessageToList, 258, {#[[2, 1]], First[ #]}& /@ dups, "Warning"]}, False, {}, {src, id}]
					];
					AssociateTo[ $globalFactLists, MapIndexed[ {#1, {{cellIDLabel[ {id, i, First[ #2]}], srcLabel}, getEvaluationPosition[ id, nb]}} -> {} &, l]];
					Null,
				(*else*)
					coreNotification[ {MSG$[ coreMessageToList, 259, {#[[2, 1]], First[ #]}& /@ dups0, "Fatal"]}, False, {}, {src, id}];
					$Failed
				]
			]]]
		]
	]
addGlobalFactList[ l0_List][ True, _, id_Integer, src_String, ___] :=
	With[ {l = Replace[ l0, STRING$[ s_String] :> s, {1}]},
		Which[
			l === {},
			Null,
		(*else*)
			MemberQ[ l, Except[ _String]],
			coreNotification[ {}, True, 256, {src, id}];
			$Failed,
		(*else*)
			True,
			With[ {keys = DeleteDuplicates[ Join @@ Function[ coll, getFactListKeys[ coll, Infinity]] /@ l]},
				KeyDropFrom[ $globalFactLists, keys];
				Null
			]
		]
	]

(* 'getFactListKeys[ coll, num, evalSrc, test]' returns at most the first 'num' keys in '$globalFactLists' whose names
	are 'coll' and whose position-specifications satisfy 'test'.
	If 'src' is a wildcard-pattern, 'num' is a finite integer and 'evalSrc' is a string, the function first searches
	for keys with source 'evalSrc'. *)
getFactListKeys[ coll_] :=
	getFactListKeys[ coll, 1, None, inScopeTSQ]
getFactListKeys[ coll_, num_] :=
	getFactListKeys[ coll, num, None, inScopeTSQ]
getFactListKeys[ coll_, num_, evalSrc_] :=
	getFactListKeys[ coll, num, evalSrc, inScopeTSQ]
getFactListKeys[ coll_String, num_, evalSrc_, test_] :=
	getFactListKeys[ splitSourceLabel[ coll], num, evalSrc, test]
getFactListKeys[ {src_, coll_String}, num_, evalSrc_String, test_] :=
	If[ IntegerQ[ num] && (src === Blank[] || src === Blank[ String]),
		With[ {aux = getFactListKeys[ {evalSrc, coll}, num, None, test]},
		With[ {len = Length[ aux]},
			If[ len === num,
				aux,
			(*else*)
				Join[ aux, getFactListKeys[ {Except[ evalSrc, _String], coll}, num - len, None, test]]
			]
		]],
	(*else*)
		getFactListKeys[ {src, coll}, num, None, test]
	]
getFactListKeys[ {src_, coll_String}, num_, _, None] :=
	Cases[ Keys[ $globalFactLists], {coll, {{_String, makeSourcePattern[ src]}, _}}, {1}, num]
getFactListKeys[ {src_, coll_String}, num_, _, test_] :=
	Cases[ Keys[ $globalFactLists], {coll, PatternTest[ {{_String, makeSourcePattern[ src]}, _List}, test]}, {1}, num]


(* ::Subsection:: *)
(*Syntax and Notation*)

(* TODO: Remove. *)
typeSyntax[ ___][ ___] :=
	$Failed

(* TODO: Remove. *)
termSyntax[ ___][ ___] :=
	$Failed
(*
termSyntax[ dir_, lhs0_, rhs0_][ False, _, id_, src_, ___] :=
	With[ {lhs = Replace[ lhs0, EXPR$[ e_] :> e], rhs1 = Replace[ rhs0, EXPR$[ e_] :> e]},
		Which[
			MemberQ[ lhs, META$|FIX$|TYPED$|Lambda$TM, {0, Infinity}, Heads -> True],
			coreNotification[ {}, True, {MSG$[ 0, {lhs}, "Fatal"]}, {src, id}];
			$Failed,
		(*else*)
			MemberQ[ rhs1, META$|FIX$|TVAR$|TFIX$, {0, Infinity}, Heads -> True],
			coreNotification[ {}, True, {MSG$[ 0, {rhs1}, "Fatal"]}, {src, id}];
			$Failed,
		(*else*)
			True,
			With[ {rhs = rhs1 //. TYPED$[ c_, t:Except[ _List]] :> With[ {ti = getTypeInstance[ c, t]}, TYPED$[ c, ti] /; ListQ[ ti]]},
				If[ OddQ[ dir],
					(* input *)
					$termInputSyntax = Append[ DeleteCases[ $termInputSyntax, lhs -> _], lhs -> rhs]
				];
				If[ dir >= 2,
					(* output *)
					Replace[ makeCoreExpr[ rhs0, Blank[], 0, {}, {}, {}, {}, None, Identity, None],
						{
							{{typ_, rhs2_, _, _, _, _}, _} :>
								(* to keep things simple, we employ purely syntactic pattern matching *)
								With[ {ac = abstractionCount[ rhs2, 0, Infinity], rhs3 = stripAbstractions[ rhs2]},
								With[ {termVarPatts = Table[ ToExpression[ "x" <> ToString[ i] <> "$$$_"], {i, 1, ac}],
										tvars = DeleteDuplicates[ Cases[ rhs3, _TVAR$|_TFIX$, {0, Infinity}, Heads -> True]]},
								With[ {tvarPattRepl = MapThread[ Rule, {tvars, Table[ ToExpression[ "t" <> ToString[ i] <> "$$$_"], {i, 1, Length[ tvars]}]}]},
								With[ {tvarInstRepl = Replace[ tvarPattRepl, (l_ -> r_Pattern) :> (l -> Take[ Hold @@ r, 1]), {1}]},
								With[ {tvarPatts = Replace[ DeleteDuplicates[ Cases[ typ, _TVAR$|_TFIX$, {0, Infinity}, Heads -> True]], tvarPattRepl, {1}],
										typInst = HoldComplete[ Evaluate[ typ /. tvarInstRepl]] /. Hold[ x_] :> x},
								With[ {lhs1 = ReplacePart[ Hold[ Evaluate[ TYPED$[ lhs, (Hold @@ tvarPatts)[[All, 1]]]]], {1, 2, 0} -> List]},		(* Apply 'lhs' to variables. *)
									$termOutputSyntax =
										Append[ DeleteCases[ $termOutputSyntax, (_ :> TYPED$[ lhs, _])|(_ :> TYPED$[ lhs, _][ ___])],
											RuleDelayed @@ Prepend[ lhs1, If[ ac === 0, rhs3, mapFreeVars[ rhs3, termVarPatts[[#3 + 1 - #2]]&]] /. tvarPattRepl]
										];
									$fullTypeDisplay =
										Append[ DeleteCases[ $fullTypeDisplay, ],
											RuleDelayed @@ Prepend[ typInst, lhs1]
										]
								]]]]]],
							_ :> (coreNotification[ {MSG$[ 0, {}, "Warning"]}, False, {}, {src, id}];)
						}
					]
				];
			]
		]
	]
termSyntax[ dir_, lhs_, rhs_][ True, ___] :=
	Null
*)

typeNotation[ _, _List, {0, _}][ False, ___] :=
	(sowMessage[ 244, {"binder"}, "Fatal"]; $Failed)
typeNotation[ dir_, types0_List, {1, fun0_}][ False, _, _, src_String, ___] :=
	With[ {types = Replace[ types0, EXPR$[ e_, ___] :> e, {1}] /. TVAR$[] -> Blank[],
			fun = Replace[ fun0, EXPR$[ e_, ___] :> e], srcLabel = sourceLabel[ src]},
		Which[
			dir === 2,
			sowMessage[ 238, {"binary"}, "Warning"];
			Null,
		(*else*)
			types === {},
			Null,
		(*else*)
			MemberQ[ {types, fun}, (TVAR$|SVAR$|TFIX$)[ __], {0, Infinity}, Heads -> True],
			sowMessage[ 239, {}, "Fatal"];
			$Failed,
		(*else*)
			True,
			With[ {patt = binaryUnfoldingPatt[ types, Null]},
				remFromConditionalRules[ $binaryTypeParsingRules, srcLabel, patt];
				addToConditionalRules[
					$binaryTypeParsingRules,
					((h:#)[ a_, b_, c__] :> binaryApply[ h, {a, b, c}, fun, False])& /@ types,
					srcLabel,
					patt
				]
			];
		]
	]
typeNotation[ dir_, types0_List, {2, base0_, constr0_, lr_, rev_}][ False, _, _, src_String, ___] :=
	With[ {base = Replace[ base0, EXPR$[ e_, ___] :> e], constr = Replace[ constr0, EXPR$[ e_, ___] :> e],
			types = Replace[ types0, EXPR$[ e_, ___] :> e, {1}], srcLabel = sourceLabel[ src]},
		Which[
			types === {},
			Null,
		(*else*)
			MemberQ[ {types, base, constr}, (TVAR$|SVAR$|TFIX$)[ __], {0, Infinity}, Heads -> True],
			sowMessage[ 239, {}, "Fatal"];
			$Failed,
		(*else*)
			True,
			If[ OddQ[ dir],
				(* input *)
				With[ {types1 = types /. TVAR$[] -> Blank[], argF = If[ Xor[ TrueQ[ rev], lr === Right], Reverse, Identity]},
				With[ {fun = If[ lr === Right, constr[ #2, #1]&, constr], patt = binaryUnfoldingPatt[ types1, Null]},
					remFromConditionalRules[ $binaryTypeParsingRules, srcLabel, patt];
					addToConditionalRules[
						$binaryTypeParsingRules,
						(#[ args___] :> recursiveApply[ argF[ {args}], base, fun])& /@ types1,
						srcLabel,
						patt
					]
				]]
			];
			If[ dir >= 2,
				(* output *)
				With[ {base1 = base /. TVAR$[] -> Blank[], constr1 = constr /. TVAR$[] -> Blank[]},
				With[ {add = If[ Xor[ TrueQ[ rev], lr === Right], Prepend, Append],
						fun = If[ lr === Right, constr1[ #2, #1]&, constr1],
						patt = binaryFoldingPatt[ types, True]},
					
					(* If other notations (e.g. "binary") create printing rules as well, delete them below, too. *)
					remFromConditionalRules[ $binaryTypePrintingRules, srcLabel, patt];
					addToConditionalRules[
						$binaryTypePrintingRules,
						Join[
							(base1 -> #[])& /@ types,
							(fun[ b:Blank[ #], a_] :> add[ b, a])& /@ types
						],
						srcLabel,
						patt
					]
				]]
			];
		]
	]
typeNotation[ dir_, types0_List, {kind_, ___}][ True, _, _, src_String, ___] :=
	With[ {types = Replace[ types0, EXPR$[ e_, ___] :> e, {1}], srcLabel = sourceLabel[ src]},
		If[ types =!= {},
			Switch[ kind,
				0,
				sowMessage[ 244, {"binder"}, "Warning"],
			(*else*)
				1,
				If[ dir === 2,
					sowMessage[ 238, {"binary"}, "Warning"],
				(*else*)
					remFromConditionalRules[ $binaryTypeParsingRules, srcLabel, binaryUnfoldingPatt[ types /. TVAR$[] -> Blank[], Null, Blank[ binaryApply]]]
				],
			(*else*)
				2,
				If[ OddQ[ dir],
					(* input *)
					remFromConditionalRules[ $binaryTypeParsingRules, srcLabel, binaryUnfoldingPatt[ types /. TVAR$[] -> Blank[], Null, Blank[ recursiveApply]]]
				];
				If[ dir >= 2,
					(* output *)
					(* If other notations (e.g. "binary") create printing rules as well, delete them here, too. *)
					remFromConditionalRules[ $binaryTypePrintingRules, srcLabel, binaryFoldingPatt[ types, True]]
				]
			]
		];
	]
typeNotation[ _, _, {}][ ___] :=
	Null

termNotation[ dir_, terms0_List, {0, cond_}][ False, _, _, src_String, ___] :=
	With[ {termsAux = Replace[ terms0, EXPR$[ e_, ___] :> e, {1}], srcLabel = sourceLabel[ src]},
		Which[
			dir === 1,
			sowMessage[ 237, {"binder"}, "Warning"];
			Null,
		(*else*)
			termsAux === {},
			Null,
		(*else*)
			MemberQ[ termsAux, (TVAR$|SVAR$|TFIX$)[ __], {0, Infinity}, Heads -> True],
			sowMessage[ 239, {}, "Fatal"];
			$Failed,
		(*else*)
			True,
			With[ {termsT = DeleteCases[ makeAnonymousTypePattern[ #, False]& /@ termsAux, $Failed],
					terms = termsAux /. (TVAR$|SVAR$)[] -> Blank[]},
				If[ Length[ termsT] < Length[ terms], sowMessage[ 240, {"binder"}, "Warning"]];
				If[ TrueQ[ cond],
					(* Conditional default binders should have declared type '(a_ -> c) -> (a_ -> t) -> s', where 'a_' is the first type variable. *)
					addToConstPattern[ $condDefaultBinderPrinting, termsT, srcLabel];
					addToConstPattern[ $condDefaultBinderInputPrinting, terms, srcLabel],
				(*else*)
					(* Unconditional default binders should have declared type '(a_ -> t) -> s', where 'a_' is the first type variable. *)
					addToConstPattern[ $uncondDefaultBinderPrinting, termsT, srcLabel];
					addToConstPattern[ $uncondDefaultBinderInputPrinting, terms, srcLabel]
				]
			];
		]
	]
termNotation[ dir_, terms0_List, {1, fun0_}][ False, _, _, src_String, ___] :=
	With[ {terms = Replace[ terms0, EXPR$[ e_, ___] :> e, {1}] /. (TVAR$|SVAR$)[] -> Blank[],
			fun = Replace[ fun0, EXPR$[ e_, ___] :> e], srcLabel = sourceLabel[ src]},
		Which[
			dir === 2,
			sowMessage[ 238, {"binary"}, "Warning"];
			Null,
		(*else*)
			terms === {},
			Null,
		(*else*)
			MemberQ[ {terms, fun}, (TVAR$|SVAR$|TFIX$)[ __], {0, Infinity}, Heads -> True],
			sowMessage[ 239, {}, "Fatal"];
			$Failed,
		(*else*)
			MemberQ[ fun, _SVAR$, {0, Infinity}, Heads -> True],
			sowMessage[ 241, {"binary", fun}, "Fatal"];
			$Failed,
		(*else*)
			True,
			With[ {untypedTerms = DeleteCases[ terms, _TYPED$]},
			With[ {patt = binaryUnfoldingPatt[ terms, untypedTerms]},
				remFromConditionalRules[ $binaryTermParsingRules, srcLabel, patt];
				addToConditionalRules[
					$binaryTermParsingRules,
					Join[
						((h:#)[ a:Except[ _RNG$], b_, c__] :> binaryApply[ h, {a, b, c}, fun])& /@ terms,
						((h:TYPED$[ #, _])[ a:Except[ _RNG$], b_, c__] :> binaryApply[ h, {a, b, c}, fun])& /@ untypedTerms
					],
					srcLabel,
					patt
				]
			]];
		]
	]
termNotation[ dir_, terms0_List, {2, base0_, constr0_, lr_, rev_}][ False, _, _, src_String, ___] :=
	With[ {terms = Replace[ terms0, EXPR$[ e_, ___] :> e, {1}],
			base = Replace[ base0, EXPR$[ e_, ___] :> e], constr = Replace[ constr0, EXPR$[ e_, ___] :> e],
			srcLabel = sourceLabel[ src]},
		Which[
			terms === {},
			Null,
		(*else*)
			MemberQ[ {terms, base, constr}, (TVAR$|SVAR$|TFIX$)[ __], {0, Infinity}, Heads -> True],
			sowMessage[ 239, {}, "Fatal"];
			$Failed,
		(*else*)
			MemberQ[ base, _SVAR$, {0, Infinity}, Heads -> True],
			sowMessage[ 241, {"recursive", base}, "Fatal"];
			$Failed,
		(*else*)
			MemberQ[ constr, _SVAR$, {0, Infinity}, Heads -> True],
			sowMessage[ 241, {"recursive", constr}, "Fatal"];
			$Failed,
		(*else*)
			True,
			If[ OddQ[ dir],
				(* input *)
				With[ {terms1 = terms /. (TVAR$|SVAR$)[] -> Blank[]},
				With[ {untypedTerms = DeleteCases[ terms1, _TYPED$]},
				With[ {fun = If[ lr === Right, constr[ #2][ #1]&, constr[ #1][ #2]&],
						argF = If[ Xor[ TrueQ[ rev], lr === Right], Reverse, Identity],
						patt = binaryUnfoldingPatt[ terms1, untypedTerms]},
					remFromConditionalRules[ $binaryTermParsingRules, srcLabel, patt];
					addToConditionalRules[
						$binaryTermParsingRules,
						Join[
							(#[ args___] :> recursiveApply[ argF[ {args}], base, fun])& /@ terms1,
							(TYPED$[ #, t_][ args___] :> recursiveApply[ argF[ {args}], TYPED$[ base, t], fun])& /@ untypedTerms
						],
						srcLabel,
						patt
					]
				]]]
			];
			If[ dir >= 2,
				(* output *)
				With[ {thinnedTerms = Select[ terms, FreeQ[ #, _SVAR$]&]},
				With[ {terms1 = thinnedTerms /. TYPED$[ e_, _] :> e,
						add = If[ Xor[ TrueQ[ rev], lr === Right], Prepend, Append],
						fun = If[ lr === Right, constr[ #2][ #1]&, constr[ #1][ #2]&],
						constrT = makeAnonymousTypePattern[ constr, False]},
				With[ {patt = binaryFoldingPatt[ terms1, False]},
					If[ Length[ thinnedTerms] < Length[ terms], sowMessage[ 242, {"recursive"}, "Warning"]];
					(* If other notations (e.g. "binary") create printing rules as well, delete them here, too. *)
					remFromConditionalRules[ $binaryTermPrintingRules, srcLabel, patt];
					
					If[ constrT === $Failed,
						sowMessage[ 243, {"recursive", constr}, "Warning"],
					(*else*)
						addToConditionalRules[
							$binaryTermPrintingRules,
							With[ {funT = If[ lr === Right, constrT[ #2][ #1]&, constrT[ #1][ #2]&]},
								Replace[ base,
									{
										TYPED$[ _Symbol|_Integer|_Rational|_Real|_Complex|_DirectedInfinity|_String, t_] :>
											Join[
												(base -> TYPED$[ #[], t])& /@ terms1,
												(funT[ TYPED$[ b:Blank[ #], typ_], a_] :> TYPED$[ add[ b, a], typ])& /@ terms1
											],
										_Symbol|_Integer|_Rational|_Real|_Complex|_DirectedInfinity|_String :>
											Join[
												(TYPED$[ base, typ_] :> TYPED$[ #[], typ, base])& /@ terms1,
												(funT[ TYPED$[ b:Blank[ #], typ_, e_], a_] :> TYPED$[ add[ b, a], typ, e])& /@ terms1
											],
										_ :>
											With[ {baseT = makeAnonymousTypePattern[ base, False]},
												If[ baseT === $Failed,
													sowMessage[ 243, {"recursive", base}, "Warning"];
													{},
												(*else*)
													Join[
														(baseT -> #[])& /@ terms1,
														(funT[ b:Blank[ #], a_] :> add[ b, a])& /@ terms1
													]
												]
											]
									}
								]
							],
							srcLabel,
							patt
						]
					];
					With[ {patt2 = binaryFoldingPatt[ thinnedTerms, False]},
						remFromConditionalRules[ $binaryTermInputPrintingRules, srcLabel, patt2];
						addToConditionalRules[
							$binaryTermInputPrintingRules,
							Join[
								(base -> #[])& /@ thinnedTerms,
								(fun[ b:Blank[ #], a_] :> add[ b, a])& /@ thinnedTerms,
								If[ !MatchQ[ base, _TYPED$],
									Join[
										(TYPED$[ base, t_] :> TYPED$[ #[], t])& /@ thinnedTerms,
										(fun[ TYPED$[ b:Blank[ #], t_], a_] :> TYPED$[ add[ b, a], t])& /@ thinnedTerms
											(* On input, the constructor does not get type-annotated, so we do not type-annotate it here either. *)
									]
								]
							],
							srcLabel,
							patt2
						]
					]
				]
			]]];
		]
	]
termNotation[ dir_, terms0_List, {kind_, args___}][ True, _, _, src_String, ___] :=
	With[ {terms = Replace[ terms0, EXPR$[ e_, ___] :> e, {1}], srcLabel = sourceLabel[ src]},
		If[ terms =!= {},
			Switch[ kind,
				0,
				If[ dir === 1,
					sowMessage[ 237, {"binder"}, "Warning"],
				(*else*)
					With[ {termsT = DeleteCases[ makeAnonymousTypePattern[ #, False]& /@ terms, $Failed],
							terms1 = terms /. (TVAR$|SVAR$)[] -> Blank[]},
						If[ Length[ termsT] < Length[ terms], sowMessage[ 240, {"binder"}, "Warning"]];
						If[ {args} === {True},
							remFromConstPattern[ $condDefaultBinderPrinting, termsT, srcLabel];
							remFromConstPattern[ $condDefaultBinderInputPrinting, terms1, srcLabel],
						(*else*)
							remFromConstPattern[ $uncondDefaultBinderPrinting, termsT, srcLabel];
							remFromConstPattern[ $uncondDefaultBinderInputPrinting, terms1, srcLabel]
						]
					]
				],
			(*else*)
				1,
				If[ dir === 2,
					sowMessage[ 238, {"binary"}, "Warning"],
				(*else*)
					With[ {terms1 = terms /. (TVAR$|SVAR$)[] -> Blank[]},
						remFromConditionalRules[ $binaryTermParsingRules, srcLabel, binaryUnfoldingPatt[ terms1, DeleteCases[ terms1, _TYPED$], Blank[ binaryApply]]]
					]
				],
			(*else*)
				2,
				If[ OddQ[ dir],
					(* input *)
					With[ {terms1 = terms /. (TVAR$|SVAR$)[] -> Blank[]},
						remFromConditionalRules[ $binaryTermParsingRules, srcLabel, binaryUnfoldingPatt[ terms1, DeleteCases[ terms1, _TYPED$], Blank[ recursiveApply]]]
					]
				];
				If[ dir >= 2,
					(* output *)
					With[ {thinnedTerms = Select[ terms, FreeQ[ #, _SVAR$]&]},
					With[ {terms1 = thinnedTerms /. TYPED$[ e_, _] :> e},
						If[ Length[ thinnedTerms] < Length[ terms], sowMessage[ 242, {"recursive"}, "Warning"]];
						
						(* If other notations (e.g. "binary") create printing rules as well, delete them here, too. *)
						remFromConditionalRules[ $binaryTermPrintingRules, srcLabel, binaryFoldingPatt[ terms1, False]];
						remFromConditionalRules[ $binaryTermInputPrintingRules, srcLabel, binaryFoldingPatt[ thinnedTerms, False]]
					]]
				]
			]
		];
	]
termNotation[ _, _, {}][ ___] :=
	Null

makeAnonymousTypePattern[ term_, _] :=
	Catch[
		makeAnonymousTypePattern[ term]
	]
makeAnonymousTypePattern[ TYPED$[ c_, t_]] :=
	If[ MatchQ[ c, _Symbol|_Integer|_Rational|_Real|_Complex|_DirectedInfinity|_String],
		TYPED$[ c, t /. _TVAR$ -> Blank[]],
	(*else*)
		Throw[ $Failed]
	]
makeAnonymousTypePattern[ Lambda$TM[ RNG$[ TYPRNG$[ _, t_]], expr_]] :=
	Lambda$TM[ RNG$[ TYPRNG$[ Blank[], t /. _TVAR$ -> Blank[]]], makeAnonymousTypePattern[ expr]]
makeAnonymousTypePattern[ term:(_VAR$|_FIX$|_META$)] :=
	term
makeAnonymousTypePattern[ _SVAR$] :=
	Blank[]
makeAnonymousTypePattern[ term:(_Symbol|_Integer|_Rational|_Real|_Complex|_DirectedInfinity|_String)] :=
	TYPED$[ term, Blank[]]
makeAnonymousTypePattern[ h_[ a___]] :=
	makeAnonymousTypePattern[ h] @@ (makeAnonymousTypePattern /@ {a})

binaryUnfoldingPatt[ terms_List, untypedTerms_] :=
	binaryUnfoldingPatt[ terms, untypedTerms, Blank[]]
binaryUnfoldingPatt[ types_List, Null, rhsPatt_] :=
	With[ {patt = (Alternatives @@ types)|(h_[ _, (Alternatives @@ types)] /; h === Pattern)},
		_[ (patt[ ___])|(h2_[ patt[ ___], _] /; h2 === Condition), rhsPatt]
	]
binaryUnfoldingPatt[ terms_List, untypedTerms_List, rhsPatt_] :=
	With[ {patt1 = (Alternatives @@ terms)|TYPED$[ Alternatives @@ untypedTerms, _]},
	With[ {patt = patt1|(h_[ _, patt1] /; h === Pattern)},
		_[ (patt[ ___])|(h2_[ patt[ ___], _] /; h2 === Condition), rhsPatt]
	]]

binaryFoldingPatt[ types_List, True] :=
	If[ types === {},
		Except[ _],
	(*else*)
		(_[ _, (Alternatives @@ types)[]])|
		(_[ h2_[ _[ _[ _, h_[ x_]], _], _], _] /; (h2 === Condition && h === Blank && MemberQ[ types, x]))|
		(_[ h2_[ _[ _, _[ _, h_[ x_]]], _], _] /; (h2 === Condition && h === Blank && MemberQ[ types, x]))
	]
binaryFoldingPatt[ terms_List, False] :=
	If[ terms === {},
		Except[ _],
	(*else*)
		(_[ _, ((Alternatives @@ terms)[])|TYPED$[ (Alternatives @@ terms)[], ___]])|
		(_[ h2_[ _[ _[ _, h_[ x_]]][ _], _], _] /; (h2 === Condition && h === Blank && MemberQ[ terms, x]))|
		(_[ h2_[ _[ _][ _[ _, h_[ x_]]], _], _] /; (h2 === Condition && h === Blank && MemberQ[ terms, x]))|
		(_[ h2_[ _[ TYPED$[ _[ _, h_[ x_]], ___]][ _], _], _] /; (h2 === Condition && h === Blank && MemberQ[ terms, x]))|
		(_[ h2_[ _[ _][ TYPED$[ _[ _, h_[ x_]], ___]], _], _] /; (h2 === Condition && h === Blank && MemberQ[ terms, x]))
	]


(* ::Subsection:: *)
(*Abbreviations*)

abbrevType[ dir_, lhs0_, rhs0_][ False, i_Integer, id_Integer, src_String, ___] :=
	Replace[
		Reap[
			With[ {lr = First[ generalizeTypes[ Replace[ {lhs0, rhs0}, EXPR$[ t_, ___] :> t, {1}], {}]]},
				(* Note: All the checks below can easily be circumvented by manipulating '$abbrevTypeDecls' etc. directly.
					Nevertheless we keep them, because if a user only makes use of high-level commands like "abbreviation",
					parsing and pretty-printing should behave properly. *)
				Replace[ Replace[ First[ lr], h_Symbol[ __] :> h],
					{
						head_Symbol :>
							Which[
								MemberQ[ getProtectedConsts[], head] || head === FORM$TM || head === FUN$TM || head === STRING$TM,
								sowMessage[ 225, {head}, "Fatal", "abbreviation"];
								$Failed,
							(*else*)
								getTypeOverlaps[ First[ lr], getTypeDefs[ False][ head]] =!= {},
								sowMessage[ 226, {head}, "Fatal", "abbreviation"];
								$Failed,
							(*else*)
								True,
								(* 'rhs' is not checked for being free of 'lhs', because normally type abbreviations in
									'rhs' are unfolded before entering this function. This might lead to unexpected
									behavior in some pathological situations. *)
								Replace[ liftTypeVars[ lr, 0, False],	(* names of type variables must disappear *)
									{{lhs_, rhs_}, tvarNum_} :>
										Replace[
											freeOfUnknownObjects[
												(* 'FORM$TM' and 'FUN$TM' are dummy type constructors, since the original head 'head' must be removed.
													'CoreEqual$TM' merely serves as a dummy, too. *)
												TYPED$[ CoreEqual$TM, {Replace[ lhs, {_Symbol :> FORM$TM, (_Symbol)[ a_, b___] :> Fold[ FUN$TM, a, {b}]}], rhs}],
												getConstDecls[ True],
												getTypeDefs[ True],
												None,
												All
											],
											{
												{{t_}} :> (sowMessage[ 233, {head, t}, "Fatal"]; $Failed),
												{const_} :> (sowMessage[ 234, {head, const}, "Fatal"]; $Failed),
												_ :>
													Replace[ generateSimpleRewriteRule[ "type"][ lhs, rhs, Association[], Hold[ $parsingCtxt]],
														{
															rl:(_RuleDelayed|_Rule) :>
																With[ {abbrevs = Lookup[ $abbrevTypeDecls, head, {}]},
																With[ {overlapPos = getTypeOverlaps[ lhs, (#[[1, 1]] -> Last[ #])& /@ abbrevs, tvarNum]},
																With[ {overlaps = Extract[ abbrevs, overlapPos], srcLabel = sourceLabel[ src]},
																With[ {patt = Alternatives @@ overlaps[[All, 1, 1]]},
																	(* It is not possible to give parsing- and pretty-printing rules separately! *)
																	remAbbrevPPRules[ patt, srcLabel, True, True];
																	With[ {diffCell =
																			Select[
																				overlaps,
																				(MemberQ[ #, HoldPattern[ inTheoryScopeQ[ srcLabel, ___]]] && !sameCellQ[ id, src, Last[ #]])&
																			]},
																		If[ diffCell =!= {}, sowMessage[ 232, {head, diffCell[[All, 2]]}, "Warning"]]
																	];
																	
																	Module[ {newAbbrevs = abbrevs},
																		remFromConditionalRules[ newAbbrevs, srcLabel, overlapPos];
																		addToConditionalRules[
																			newAbbrevs,
																			{lhs -> {cellIDLabel[ {id, i}], srcLabel}},
																			srcLabel,
																			getTypeOverlaps[ lhs, (#[[1, 1]] -> Last[ #])& /@ newAbbrevs, tvarNum]
																				(* Re-computing the positions of overlapping types is certainly not the most
																					efficient way, but we cannot use 'overlapPos' since some rules might
																					have been removed in the meantime. *)
																		];
																		AssociateTo[ $abbrevTypeDecls, head -> newAbbrevs]
																	];
																	
																	addToConditionalRules[
																		$abbrevTypeUnfoldingRules,
																		{rl},
																		srcLabel,
																		abbrevUnfoldingPatt[ patt, True]
																	];
																	If[ dir >= 2,
																		(* output *)
																		Replace[ generateSimpleRewriteRule[ "type"][ rhs, lhs, Association[], Hold[ $printingCtxt]],
																			{
																				rlo:(_RuleDelayed|_Rule) :>
																					addToConditionalRules[
																						$abbrevTypeFoldingRules,
																						{rlo},
																						srcLabel,
																						abbrevFoldingPatt[ patt, True]
																					],
																				_ :> sowMessage[ 230, {head}, "Warning"]
																			}
																		]
																	];
																]]]],
															_ :> (sowMessage[ 229, {head}, "Fatal"]; $Failed)
														}
													]
											}
										]
								]
							],
						expr_ :> (sowMessage[ 235, {expr}, "Fatal"]; $Failed)
					}
				]
			],
			"Message"
		],
		{ret_, msg_List} :> If[ coreNotification[ msg, ret === $Failed, {}, {src, id}], $Failed, Null]
	]
abbrevType[ dir_, lhs0_, _][ True, _, id_, src_, ___] :=
	With[ {lhs = Replace[ lhs0, EXPR$[ t_, ___] :> First[ generalizeTypes[ t, {}]]]},
		Replace[ Replace[ lhs, h_Symbol[ __] :> h],
			{
				head_Symbol :>
					With[ {abbrevs = Lookup[ $abbrevTypeDecls, head, {}], tvarNum = getMaxTVar[ lhs] + 1, srcLabel = sourceLabel[ src]},
					With[ {overlapPos = getTypeOverlaps[ lhs, (#[[1, 1]] -> Last[ #])& /@ abbrevs, tvarNum]},
					With[ {overlaps = Extract[ abbrevs, overlapPos]},
						remAbbrevPPRules[ Alternatives @@ overlaps[[All, 1, 1]], srcLabel, OddQ[ dir], True];
						Module[ {newAbbrevs = abbrevs},
							remFromConditionalRules[ newAbbrevs, srcLabel, overlapPos];
							If[ newAbbrevs === {},
								KeyDropFrom[ $abbrevTypeDecls, head],
							(*else*)
								AssociateTo[ $abbrevTypeDecls, head -> newAbbrevs]
							]
						];
					]]],
				expr_ :> (coreNotification[ {}, True, {MSG$[ coreMessageToList, 235, {expr}, "Fatal"]}, {src, id}]; $Failed)
			}
		]
	]

abbrevTerm[ dir_, lhs0_, rhs0_][ False, i_Integer, id_Integer, src_String, ___] :=
	Replace[
		Reap[
			(* Note: All the checks below can easily be circumvented by manipulating '$abbrevConstDecls' etc. directly.
				Nevertheless we keep them, because if a user only makes use of high-level commands like "abbreviation",
				parsing and pretty-printing should behave properly. *)
			Replace[ extractFreshConstant[ lhs0, True, False, "abbreviation"],
				{c_, ctyp0_, lhs1_} :>
					(* We cannot use 'makeCoreExpr', because abbreviations shall not be unfolded. *)
					With[ {rhs1 = makeSTERM[ rhs0],
							ctyp = getFullTypeNoSorts[ ctyp0, HoldComplete[ c], getAllConstDecls[ False]]},	(* The argument of 'getAllConstDecls' does not matter here. *)
					Replace[
						liftTypeVars[
							makeSingleSTERM[ {ReplacePart[ lhs1, 1 -> TYPED$[ Theorema`Knowledge`TYPE$TM, {If[ ctyp === $Failed, TVAR$[], ctyp]}]], rhs1}],
							0,
							False
						],	(* The constant is replaced by 'TYPE$', since otherwise old declarations of it might interfere with type inference. *)
						{STERM$[ {lhs11_, rhs11_}, tfix0_Association, tvar0_Association], idx_} :>
							Replace[
								parseCoreTermPost[
									inferTypes[ STERM$[ CoreEqual$TM[ lhs11][ rhs11], tfix0, tvar0], FORM$TM, idx, 0, {}, {}, {}, Association[], tvar0, "gen"],
									Identity,
									False
								],
								{
									{{_, STERM$[ TYPED$[ CoreEqual$TM, {typ0_}][ _][ rhs2_], tfix_, tvar_], __}, _} :>	(* 'tfix' should be empty *)
										If[ MemberQ[ rhs2, TYPED$[ c, _], {0, Infinity}, Heads -> True],
											sowMessage[ 228, {c}, "Fatal"];
											$Failed,
										(*else*)
											With[ {rhs = normalize[ Block[ {$parsingCtxt = {tfix, tvar}}, rhs2 /. $abbrevTermUnfoldingRules]]},
												Replace[ freeOfUnknownObjects[ TYPED$[ CoreEqual$TM, {typ0}][ rhs], getConstDecls[ True], getTypeDefs[ True], None, All],
													{
														{{t_}} :> (sowMessage[ 233, {c, t}, "Fatal"]; $Failed),
														{const_} :> (sowMessage[ 234, {c, const}, "Fatal"]; $Failed),
														_ :>
															With[ {typ1 = liftTypeVars[ typ0, 0, False]},	(* type variables in 'typ0' must be lifted again *)
															With[ {tvp =
																	Table[
																		Position[ First[ typ1], TVAR$[ j]] -> Lookup[ tvar, TVAR$[ j], getDefaultSort[]],
																		{j, 0, Last[ typ1] - 1}
																	]},
															With[ {typ2 = Append[ typ1, tvp], typ = Extract[ typ0, tvp[[All, 1, 1]]]},		(* 'typ' is a short type *)
																Replace[ generateSimpleRewriteRule[ "higher-order"][ TYPED$[ c, typ], rhs, tvar, Hold[ $parsingCtxt]],
																	{
																		rl:(_RuleDelayed|_Rule) :>
																				(* It is not possible to give parsing- and pretty-printing rules separately! *)
																			With[ {srcLabel = sourceLabel[ src]},
																			With[ {cv =
																					Replace[ Lookup[ $abbrevConstDecls, c, $Failed],
																						{
																							cv_COREVAR$ :>
																								(
																									Replace[ Lookup[ getCoreVarAssoc[ cv], srcLabel, $Failed],
																										{k_List, _} :>
																											If[ !sameCellQ[ id, src, k],
																												sowMessage[ 232, {c, {k}}, "Warning"]
																											]
																									];
																									remAbbrevPPRules[ c, srcLabel, True, False];
																									cv
																								),
																							_ -> COREVAR$[ Association[], {$Failed, $Failed}]
																						}
																					]},
																				AssociateTo[
																					$abbrevConstDecls,
																					c -> setCoreVar[ 0, cv, {{cellIDLabel[ {id, i}], srcLabel}, typ2}, srcLabel]
																				];
																				addToConditionalRules[
																					$abbrevTermUnfoldingRules,
																					{rl},
																					srcLabel,
																					abbrevUnfoldingPatt[ c, False]
																				];
																				If[ dir >= 2,
																					(* output *)
																					Replace[
																						generateSimpleRewriteRule[ "higher-order"][ rhs, TYPED$[ c, typ], tvar, Hold[ $printingCtxt]],
																						{
																							rlo:(_RuleDelayed|_Rule) :>
																								addToConditionalRules[
																									$abbrevTermFoldingRules,
																									{rlo},
																									srcLabel,
																									abbrevFoldingPatt[ c, False]
																								],
																							_ :> sowMessage[ 230, {c}, "Warning"]
																						}
																					]
																				];
																			]],
																		_ :> (sowMessage[ 229, {c}, "Fatal"]; $Failed)
																	}
																]
															]]]
													}
												]
											]
										],
									_ :> (sowMessage[ 227, First /@ {lhs1, rhs1}, "Fatal", "abbreviation"]; $Failed)
								}
							]
					]]
			],
			"Message"
		],
		{ret_, msg_List} :> If[ coreNotification[ msg, ret === $Failed, {}, {src, id}], $Failed, Null]
	]
abbrevTerm[ dir_, lhs_, _][ True, _, id_, src_, ___] :=
	Replace[ Replace[ lhs, {TYPED$[ c_, _] :> c, EXPR$[ TYPED$[ c_, _], ___] :> c, EXPR$[ c_, ___] :> c}],
		{
			c:(_Symbol|_Integer|_Rational|_Real|_Complex|_DirectedInfinity) :>
				With[ {srcLabel = sourceLabel[ src]},
					remAbbrevPPRules[ c, srcLabel, OddQ[ dir], False];
					Which[
						!MemberQ[ $abbrevTermUnfoldingRules, abbrevUnfoldingPatt[ c, False]],
						KeyDropFrom[ $abbrevConstDecls, c],
					(*else*)
						OddQ[ dir],
						Replace[ Lookup[ $abbrevConstDecls, c, $Failed],
							cv_COREVAR$ :>
								With[ {cvNew = unsetCoreVar[ 0, cv, srcLabel]},
									If[ MatchQ[ getCoreVarAssoc[ cvNew], Association[ __]],
										AssociateTo[ $abbrevConstDecls, c -> cvNew],
									(*else*)
										KeyDropFrom[ $abbrevConstDecls, c]
									]
								]
						]
					];
				],
			expr_ :> (coreNotification[ {}, True, {MSG$[ coreMessageToList, 224, {expr}, "Fatal"]}, {src, id}]; $Failed)
		}
	]

remAbbrevPPRules[ typ_, thy_String, parsing_, True] :=
	(
		remFromConditionalRules[ $abbrevTypeFoldingRules, thy, abbrevFoldingPatt[ typ, True]];
		If[ TrueQ[ parsing],
			remFromConditionalRules[ $abbrevTypeUnfoldingRules, thy, abbrevUnfoldingPatt[ typ, True]]
		];
	)
remAbbrevPPRules[ const_, thy_String, parsing_, False] :=
	(
		remFromConditionalRules[ $abbrevTermFoldingRules, thy, abbrevFoldingPatt[ const, False]];
		If[ TrueQ[ parsing],
			remFromConditionalRules[ $abbrevTermUnfoldingRules, thy, abbrevUnfoldingPatt[ const, False]]
		];
	)

abbrevUnfoldingPatt[ typ_, True] :=
	With[ {patt = typ /. _TVAR$|_TFIX$ -> Blank[]},
		_[ patt|(h_[ patt, _] /; h === Condition), _]
	]
abbrevUnfoldingPatt[ const_, False] := _[ (TYPED$[ const, _])|(h_[ TYPED$[ const, _], _] /; h === Condition), _]

abbrevFoldingPatt[ typ_, True] := _[ _, typ /. _TVAR$|_TFIX$ -> Blank[]]
abbrevFoldingPatt[ const_, False] := _[ _, x_ /; MatchQ[ getHeadAtom[ HoldComplete[ x]], {HoldComplete[ TYPED$[ const, _]], _}]]

(* 'freeOfUnknownObjects[ term, cdecls, tdefs, tfix, tvars]' checks whether term (or list of terms) 'term' contains
	undeclared constants (w.r.t. 'cdecls') or undeclared/undefined types (w.r.t. 'tdefs').
	Type variables and a.b.f. type constants count as known iff they are contained in 'tvars' or 'tfix', respectively.
	The function returns either 'True', '{const}' or '{{typ}}'. *)
freeOfUnknownObjects[ term_, cdecls_, tdefs_, tfix_, tvars_] :=
	With[ {undecl =
			Select[
				DeleteDuplicates[ Cases[ term, TYPED$[ c:Except[ CoreEqual$TM|CoreImplies$TM|CoreAll$TM], _] :> c, {0, Infinity}, Heads -> True]],
				(!ListQ[ cdecls[ #]])&,
				1
			]},
		If[ undecl === {},
			With[ {allTypes = DeleteDuplicates[ Join @@ Cases[ term, (TYPED$|TYPRNG$)[ _, t_] :> If[ ListQ[ t], t, {t}], {0, Infinity}, Heads -> True]]},
				isLegalTypeList[
					allTypes,
					False,
					tdefs,
					Switch[ tvars, _List, MemberQ[ tvars, #]&, None, False&, _, True&],
					Switch[ tfix, _List, MemberQ[ tfix, #]&, None, False&, _, True&],
					Throw[ {{#}}]&,
					Throw[ {{#}}]&
				]
			],
		(*else*)
			undecl
		]
	]

(* TODO: Always check scope? *)
extractFreshConstant[ expr_, checkDecls_, checkAbbrevs_, header_] :=
	With[ {const = makeSTERM[ expr]},
		Replace[ Replace[ First[ const], c:Except[ _TYPED$] :> TYPED$[ c, TVAR$[]]],
			{
				TYPED$[ c:(_Symbol|_Integer|_Rational|_Real|_Complex|_DirectedInfinity), ctyp_] :>
					Which[
						MemberQ[ getProtectedConsts[], c] || getLockedConsts[ c] =!= $Failed,
						If[ StringQ[ header], sowMessage[ 225, {c}, "Fatal", header]];
						$Failed,
					(*else*)
						TrueQ[ checkDecls] && ListQ[ getConstDecls[ False][ c]],
						If[ StringQ[ header], sowMessage[ 226, {c}, "Fatal", header]];
						$Failed,
					(*else*)
						TrueQ[ checkAbbrevs] && Lookup[ $abbrevConstDecls, c, $Failed] =!= $Failed,
						If[ StringQ[ header], sowMessage[ 231, {c}, "Fatal", header]];
						$Failed,
					(*else*)
						True,
						{c, ctyp, const}
					],
				TYPED$[ e_, _] :> (If[ StringQ[ header], sowMessage[ 224, {e}, "Fatal", header]]; $Failed)
			}
		]
	]


(* ::Subsection:: *)
(*Misc*)

updateRuleLists[ pos_][ new_, {_, old_, _}] :=
	(* This definition is used in 'addFactsByProof', so we have to return the label and data of the formula. *)
	(updateRuleLists[ new, old, pos]; Join[ Drop[ new, 2], If[ ListQ[ pos], FML$[ "pos" -> pos], FML$[]]])
updateRuleLists[ $Failed, FML$[ k_, f_, _, opts___], pos_] :=
	(* remove formula *)
	With[ {tags = Replace[ "tags", {opts, _ -> {}}]},
		updateRuleList[ $autoredRules, k, False, MemberQ[ tags, "autored"|"autored?"], Null, f, "intro"];
		updateRuleList[ $elimRules, k, False, MemberQ[ tags, "elim"], Null, f, "elim"];
		updateRuleList[ $introRules, k, False, MemberQ[ tags, "intro"], Null, f, "intro"];
		updateRuleList[ $symRules, k, False, MemberQ[ tags, "sym"], Null, f, "elim"];
		updateRuleList[ $transRules, k, False, MemberQ[ tags, "trans"], Null, f, "trans"];
		If[ MemberQ[ tags, "simp"],
			Block[ {$ignoreScope = True}, addRulesToTheoryAux[ {"simp", Automatic}, {k}, k, pos, True]]
		];
		$globalFactLists = Map[ DeleteCases[ #, {k, _}]&, $globalFactLists];
	]
updateRuleLists[ FML$[ k_, f_, _, opts___], $Failed, pos_] :=
	(* add new formula *)
	With[ {tags = Replace[ "tags", {opts, _ -> {}}]},
		updateRuleList[ $autoredRules, k, MemberQ[ tags, "autored"|"autored?"], False, f, Null, "intro"];
		updateRuleList[ $elimRules, k, MemberQ[ tags, "elim"], False, f, Null, "elim"];
		updateRuleList[ $introRules, k, MemberQ[ tags, "intro"], False, f, Null, "intro"];
		updateRuleList[ $symRules, k, MemberQ[ tags, "sym"], False, f, Null, "elim"];
		updateRuleList[ $transRules, k, MemberQ[ tags, "trans"], False, f, Null, "trans"];
		If[ MemberQ[ tags, "simp"],
			Block[ {$ignoreScope = True}, addRulesToTheoryAux[ {"simp", Automatic}, {k}, k, pos, False]]
		]
	]
updateRuleLists[ FML$[ k_, fn_, _, on___], FML$[ _, fo_, _, oo___], pos_] :=
	(* replace formula *)
	With[ {tn = Replace[ "tags", {on, _ -> {}}], to = Replace[ "tags", {oo, _ -> {}}]},
	With[ {sn = MemberQ[ tn, "simp"], so = MemberQ[ to, "simp"]},
		updateRuleList[ $autoredRules, k, MemberQ[ tn, "autored"|"autored?"], MemberQ[ to, "autored"|"autored?"], fn, fo, "intro"];
		updateRuleList[ $elimRules, k, MemberQ[ tn, "elim"], MemberQ[ to, "elim"], fn, fo, "elim"];
		updateRuleList[ $introRules, k, MemberQ[ tn, "intro"], MemberQ[ to, "intro"], fn, fo, "intro"];
		updateRuleList[ $symRules, k, MemberQ[ tn, "sym"], MemberQ[ to, "sym"], fn, fo, "elim"];
		updateRuleList[ $transRules, k, MemberQ[ tn, "trans"], MemberQ[ to, "trans"], fn, fo, "trans"];
		If[ sn || so,
			Block[ {$ignoreScope = True}, addRulesToTheoryAux[ {"simp", Automatic}, {k}, k, pos, so && !sn]]
		];
		If[ ListQ[ pos],
			(* It is possible that a formula is added to a fact collection in the same place where it is added to the
				theory context, e.g. by "facts".
				Example scenario:
				- "facts "list" = "B": "fml"[OF "A"]"
					The new formula "B" is constructed and immediately added to "list". Hence, the source-position of "B" is the
					same as the source-position of the place where it is added to "list".
				- Re-evaluating the above command re-constructs "B", then removes it from "list" because its source-position
					does not come before the place where it has been added to "list" before, and finally adds it again to "list".
				- Removing the "list ="-part of the above command and then re-evaluating it (in the same cell) removes "B" from
					"list" and does not add it any more. *)
			$globalFactLists =
				Map[
					DeleteCases[ #, {k, sp_} /; With[ {cp = compareTheoryPositions[ sp, {k, pos}, inTheoryScopeQ]}, cp === Less || cp === Equal]]&,
					$globalFactLists
				]
		];
	]]
updateRuleLists[ _, _, _] :=
	Null

SetAttributes[ updateRuleList, HoldFirst];
updateRuleList[ Hold[ coll_], args___] :=
	updateRuleList[ coll, args]
updateRuleList[ coll_, elem_, True, False, ___] :=
	AppendTo[ coll, elem]
updateRuleList[ coll_, elem_, False, True, ___] :=
	(coll = DeleteCases[ coll, elem, {1}, 1])
updateRuleList[ coll_, elem_, cn_List, False] :=
	addRulesToRULEC$[ coll, cn -> elem]
updateRuleList[ coll_, elem_, False, co_List] :=
	remRulesFromRULEC$[ coll, co -> elem]
updateRuleList[ coll_, elem_, cn_List, co_List] :=
	If[ cn === co,
		coll,
	(*else*)
		updateRuleList[ coll, elem, False, co];
		updateRuleList[ coll, elem, cn, False]
	]
updateRuleList[ coll_, elem_, True, False, fn_, _, cat_] :=
	If[ ListQ[ coll],
		updateRuleList[ coll, elem, True, False],
	(*else*)
		updateRuleList[ coll, elem, getRuleConsts[ fn, cat, True], False]
	]
updateRuleList[ coll_, elem_, False, True, _, fo_, cat_] :=
	If[ ListQ[ coll],
		updateRuleList[ coll, elem, False, True],
	(*else*)
		updateRuleList[ coll, elem, False, getRuleConsts[ fo, cat, True]]
	]
updateRuleList[ coll_, elem_, True, True, fn_, fo_, cat_] :=
	If[ ListQ[ coll],
		coll,
	(*else*)
		updateRuleList[ coll, elem, getRuleConsts[ fn, cat, True], getRuleConsts[ fo, cat, True]]
	]
updateRuleList[ coll_, ___] :=
	coll

formulaToAssumption[ $Failed, ___] :=
	$Failed
formulaToAssumption[ f_, norm_, computePos_] :=
	Replace[
		Replace[ liftTypeVars[ makeSTERM[ f], 0],
			(* We do not set '$strictTypeChecking' to 'True', because 'parseCoreTermPost' could still replace constants by other terms. *)
			{f1:STERM$[ _, tfix_, tvar_], idx1_} :> parseCoreTermPost[ inferTypes[ f1, FORM$TM, idx1, 0, {}, {}, {}, Association[], tvar, "gen"], norm]
		],
		{
			{{_, STERM$[ f1_, Association[], tv_Association], _, {}, {}}, {}} :>
				If[ TrueQ[ computePos],
					{f1, makeTypeVarPositions[ f1, tv]},
				(*else*)
					{f1, tv}
				],
			_ -> "typeInference"
		}
	]

increaseTheoremIndex[ i:(_Integer|_Rational|_Real)] := i + 1
increaseTheoremIndex[ {}] := 0
increaseTheoremIndex[ {a_, ___}] := increaseTheoremIndex[ a]

(* 'addRuleData' assumes that the given formula does not have a "TypeVarPositions" entry yet. *)
addRuleData[ fml:FML$[ k_, f_, l_, data___], tvar_, True, rwData_, overwrite_] :=
	addRuleData[
		With[ {rd = Hold[ generateReductionRule[ f, True, True, tvar]]},
			If[ MemberQ[ {data}, "arData" -> _],
				If[ TrueQ[ overwrite],
					FML$[ k, f, l, Sequence @@ DeleteCases[ {data}, "arData" -> _], "arData" -> ReleaseHold[ rd]],
				(*else*)
					fml
				],
			(*else*)
				FML$[ k, f, l, data, "arData" -> ReleaseHold[ rd]]
			]
		],
		tvar,
		False,
		rwData,
		overwrite
	]
addRuleData[ fml:FML$[ k_, f_, l_, data___], tvar_, False, True, overwrite_] :=
	With[ {rd = Hold[ generateEqRewriteRule[ f, True, Association[], tvar]]},
		If[ MemberQ[ {data}, "rwData" -> _],
			If[ TrueQ[ overwrite],
				FML$[ k, f, l, Sequence @@ DeleteCases[ {data}, "rwData" -> _], "rwData" -> ReleaseHold[ rd]],
			(*else*)
				fml
			],
		(*else*)
			FML$[ k, f, l, data, "rwData" -> ReleaseHold[ rd]]
		]
	]
addRuleData[ fml_FML$, _, False, False, _] :=
	fml


(* ::Section:: *)
(*Passive Theory Commands*)

(* ::Subsection:: *)
(*Printing Types*)

printTypes[ types_List, spec_Association][ False, _, id_, src_, ___] :=
	With[ {disp =
			Switch[ Lookup[ spec, "display", 2],
				0,
				Identity,
				1,
				printType,
				_,
				Composition[ If[ $Notebooks, theoremaDisplay, ToString], printType]
			],
			err = ($msgID = None; sowMessage[ 212, {#}, "Fatal"]&)},
		Replace[
			Reap[
				Map[
					If[ isLegalType[ #, getTypeDefs[ True], True&, True&, err, err], disp[ #], $Failed]&,
					Replace[ types, EXPR$[ t_, ___] :> t, {1}]
				],
				"Message"
			],
			{res_List, msg_List} :>
				(
					coreNotification[ msg, False, {}, {src, id}];
					Which[
						Lookup[ spec, "return", False],
						res,
					(*else*)
						isNotebookEvaluation[],
						Null,
					(*else*)
						True,
						Scan[ Print, res]
					]
				)
		]
	]
printTypes[ _List, _Association][ True, ___] :=
	Null


(* ::Subsection:: *)
(*Printing Terms*)

printTerms[ terms_List, spec_Association][ False, _, id_, src_, ___] :=
	With[ {disp =
			Switch[ Lookup[ spec, "display", 2],
				0,
				Identity,
				1,
				{printCoreTerm[ #[[2]], {#[[4]], #[[5]], Association[], Association[]}], printType[ #[[1]]]}&,
				_,
				printDisplayStruct[ TTDISP$ @@ Reverse[ Take[ #, 2]], $Notebooks]&
			]},
		Replace[
			Reap[
				Map[
					Replace[ makeCoreExpr[ #, Blank[], 0, {}, {}, Association[], Association[], {}, None, normalize, Check],
						{
							{r_List, _List} :> disp[ r],
							_ -> $Failed
						}
					]&,
					terms
				],
				"Message"
			],
			{res_List, msg_List} :>
				(
					coreNotification[ msg, False, {}, {src, id}];
					Which[
						Lookup[ spec, "return", False],
						res,
					(*else*)
						isNotebookEvaluation[],
						Null,
					(*else*)
						True,
						Scan[ Print, res]
					]
				)
		]
	]
printTerms[ _List, _Association][ True, ___] :=
	Null


(* ::Subsection:: *)
(*Printing Formulas*)

printForms[ spec_Association, all_][ forms_List, ___][ False, i_, id_, src0_, cell_CellObject, nb_NotebookObject] :=
	With[ {disp =
			Switch[ Lookup[ spec, "display", 2],
				0,
				Identity,
				1,
				Composition[ printCoreTerm, formula],
				_,
				Composition[ If[ $Notebooks, theoremaDisplay, ToString], printCoreTerm, formula]
			],
			src = sourceLabel[ src0]},
		Replace[
			Reap[
				obtainFactsFromProof[
					Hold[ Theorema`CoreProvers`addKnowledgeImplicit[ forms, Null, False, 3, False, None]],
					src,
					cellIDLabel[ {id, i, #2}]&,
					{If[ TrueQ[ all], "", cellIDLabel[ {id, i}]], src},
					If[ TrueQ[ all], {}, {"pos" -> getEvaluationPosition[ id, nb]}]
				],
				"Message"
			],
			{res_, msg_List} :>
				If[ coreNotification[ msg, !ListQ[ res], {}, {src0, id}],
					$Failed,
				(*else*)
					Which[
						Lookup[ spec, "return", False],
						disp /@ First[ res],
					(*else*)
						isNotebookEvaluation[],
						Null,
					(*else*)
						True,
						Scan[ Composition[ Print, disp], First[ res]]
					]
				]
		]
	]
printForms[ _Association, _][ ___][ True, ___] :=
	Null


(* ::Subsection:: *)
(*Simplifying Terms*)

simplifyTerms[ terms_List, printSpec_Association, context_, prf_Hold][ False, i_, id_, src_, cell_CellObject, nb_NotebookObject] :=
	Module[ {head},
	With[ {disp =
			Switch[ Lookup[ printSpec, "display", 2],
				0,
				Identity,
				1,
				{printCoreTerm[ #[[2]], {#[[4]], #[[5]], Association[], Association[]}], printType[ #[[1]]]}&,
				_,
				printDisplayStruct[ TTDISP$ @@ Reverse[ Take[ #, 2]], $Notebooks]&
			]},
		Replace[
			Reap[
				Map[
					Function[ trm0,
						Replace[
							makeCoreExpr[
								If[ TrueQ[ context], trm0, Replace[ trm0, {(h:(STERM$|EXPR$))[ e_, s___] :> h[ head[ e], s], e_ :> head[ e]}]],
								FORM$TM, 0, {}, {}, Association[], Association[], {}, "gen", normalize, Check
							],
							{
								{{_, STERM$[ trm_, _, tvar_], __}, _List} :>
									Replace[
										executeProofCommands[
											prf,
											initiateProof[
												FML$[
													{cellIDLabel[ {id, i}], sourceLabel[ src]},
													trm,
													"",
													"TypeVarPositions" -> makeTypeVarPositions[ trm, tvar],
													"pos" -> getEvaluationPosition[ id, nb]
												],
												{},
												2
											],
											2,
											True
										],
										{
											{PRFSTAT$[ _, open_List, _List, _List, ___], _} :>
												With[ {aux =
														Flatten[
															Replace[ open,
																PRFSIT$[ G_List, _, idx_, fix_, meta_, tfix1_, tvar1_, __] :>
																	Map[
																		FirstCase[ #,
																			TYPED$[ head, FUN$TM[ typ_, _]][ trm1_] :> {typ, STERM$[ trm1, tfix1, tvar1], idx, fix, meta},
																			{FORM$TM, STERM$[ #, tfix1, tvar1], idx, fix, meta},
																			Infinity,
																			Heads -> False
																		]&,
																		G
																	],
																{1}
															],
															1
														]},
													Switch[ aux,
														{},
														Null,
													(*else*)
														{_},
														disp @@ aux,
													(*else*)
														_,
														disp /@ aux
													]
												],
											_ -> $Failed
										}
									],
								_ -> $Failed
							}
						]
					],
					terms
				],
				"Message"
			],
			{res_List, msg_List} :>
				(
					coreNotification[ msg, False, {}, {src, id}];
					Which[
						Lookup[ printSpec, "return", False],
						res,
					(*else*)
						isNotebookEvaluation[],
						Null,
					(*else*)
						True,
						Scan[ Print, res]
					]
				)
		]
	]]
simplifyTerms[ _List, _Association, _, _Hold][ True, ___] :=
	Null


(* ::Section:: *)
(*Further Functions*)

getRuleFormulas[ cat_, includeOrig_, checkOrigPos_, checkNewPos_, scopeOnly_][ spec_] :=
	getRuleFormulas[ cat, spec, includeOrig, checkOrigPos, checkNewPos, scopeOnly]
getRuleFormulas[ cat_, fml:FML$[ {id_String, src_String}, __], includeOrig_, checkOrigPos_, checkNewPos_, scopeOnly_] :=
	Which[
		StringQ[ cat] && !StringMatchQ[ id, "ID" ~~ _ ~~ "core" ~~ (DigitCharacter..)],
		With[ {baseID = id <> "." <> cat <> "."},
		With[ {rules =
				With[ {rules0 = retrieveFormulasFromContext[ Cases, {_String?(StringMatchQ[ #, baseID ~~ (DigitCharacter..)]&), src}, Blank[], Infinity, None, getFormulas[ scopeOnly]]},
					If[ TrueQ[ checkNewPos],
						Select[ rules0, inScopeQ],
					(*else*)
						rules0
					]
				]},
			Which[
				rules =!= {},
				If[ TrueQ[ includeOrig] && (!checkOrigPos || inScopeQ[ fml]),
					Prepend[ rules, fml],
				(*else*)
					rules
				],
			(*else*)
				!checkOrigPos || inScopeQ[ fml],
				{fml},
			(*else*)
				True,
				{}
			]
		]],
	(*else*)
		!checkOrigPos || inScopeQ[ fml],
		{fml},
	(*else*)
		True,
		{}
	]
getRuleFormulas[ cat_, ky:{_String, _String}, includeOrig_, checkOrigPos_, checkNewPos_, scopeOnly_] :=
	key /@ getRuleFormulas[ cat, getFormulas[ True][ ky], includeOrig, checkOrigPos, checkNewPos, scopeOnly]
getRuleFormulas[ cat_, spec_List, includeOrig_, checkOrigPos_, checkNewPos_, scopeOnly_] :=
	Which[
		StringQ[ cat],
		Flatten[ getRuleFormulas[ cat, includeOrig, checkOrigPos, checkNewPos, scopeOnly] /@ spec, 1],
	(*else*)
		!checkOrigPos,
		spec,
	(*else*)
		True,
		Select[ spec, inScopeQ]
	]
getRuleFormulas[ _, _, _, _, _, _] :=
	{}

(* 'getThmInst' returns the same as 'extractFormula', i.e. either '$Failed' or a pair involving a
	'STERM$'-structure. *)
getThmInst[ scopeOnly_, idx_Integer][ THM$[ k_, args___]] :=
	Replace[ getTheorems[ scopeOnly][ k][ args],
		{f_, tvp_List} :> extractFormula[ f, tvp, idx]
	];
getThmInst[ _, _][ _] = $Failed;

generateNewCellID[] :=
	RandomInteger[ {1*^6, 1*^10}]
generateNewCellID[ nb_NotebookObject] :=
	Max[ 1*^6, Max @@ CurrentValue[ Cells[ nb, CellStyle -> {"FormalTextInputFormula", "TheoremaCode", "TheoremaInput"}], CellID]] + 1

End[]

EndPackage[]
