(* ::Package:: *)

(* Theorema
    Copyright (C) 2017 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

BeginPackage[ "Theorema`Core`Tools`Types`"]

Needs[ "Theorema`Common`"]
Needs[ "Theorema`Core`Trusted`"]
Needs[ "Theorema`Core`Tools`"]
Needs[ "Theorema`Language`"]

Begin["`Private`"]

(* ::Section:: *)
(*Misc*)

getAllConstDecls[ scopeOnly_][ c_] :=
	Replace[ getConstDecls[ scopeOnly][ c],
		Except[ _List] :> Last[ getCoreVar[ Lookup[ $abbrevConstDecls, c, {$Failed}]]]
	]

getObjectTruthType[] :=
	Replace[ getConstDecls[ False][ TrueQ$TM],	(* we do not look at abbreviations here *)
		{
			{FUN$TM[ t_, FORM$TM], 0, {}} :> t,
			_ -> FORM$TM
		}
	]

(* 'getHeadAtom' has nothing to do with types, but it is needed below in 'adjustFlexType'. *)
getHeadAtom[ expr_] :=
	getHeadAtom[ expr, 0]
getHeadAtom[ expr:HoldComplete[ _Complex|_Rational|_DirectedInfinity|_VAR$|_FIX$|_META$|_TYPED$|_Lambda$TM], i_] :=
	{expr, i}
getHeadAtom[ HoldComplete[ h_[ ___]], i_] :=
	getHeadAtom[ HoldComplete[ h], i + 1]
getHeadAtom[ expr_HoldComplete, i_] :=
	{expr, i}

resultType[ FUN$TM[ _, t_], n_Integer] :=
	resultType[ t, n + 1]
resultType[ t_, n_Integer] :=
	{t, n}

(* Attention! 'argT' must be in reversed order. *)
makeFunctionType[ argT_, resT_] :=
	Fold[ FUN$TM[ #2, #1]&, resT, argT]

unifyTypeListsRec[ l1_, l2_, tfix_, acc_List] :=
	unifyTypeListsRec[ l1, l2, tfix, acc, $instantiateTypes]
unifyTypeListsRec[ l1_, l2_, tfix_, acc_List, instTypes_] :=
	Which[
		l1 === l2,
		acc,
	(*else*)
		instTypes,
		unifyTypes[ l1, l2, hasSort[ tfix], acc, 0],
	(*else*)
		True,
		$Failed
	]

(* iterative procedures are faster for long lists *)
unifyTypeListsIt[ l1_, l2_, tfix_, acc_List] :=
	unifyTypeListsIt[ l1, l2, tfix, acc, $instantiateTypes]
unifyTypeListsIt[ l1_List, l2_List, tfix_, acc0_List, instTypes_] :=
	Which[
		l1 === l2,
		acc0,
	(*else*)
		instTypes,
		Module[ {acc = acc0},
			Catch[
				MapThread[
					Replace[ unifyTypes[ #1, #2, hasSort[ tfix], acc, 0],
						{
							newAcc_List :> (acc = newAcc;),
							_ :> Throw[ $Failed]
						}
					]&,
					{l1, l2}
				];
				acc
			]
		],
	(*else*)
		True,
		$Failed
	]

(* 'getType[ expr]' returns the full type of the fully type-annotated expression 'expr', without doing any type
	checking or -inference. The function returns '$Failed' if something goes wrong. *)
getType[ term_, var_, fix_, meta_] :=
	getType[ term, var, fix, meta, getConstDecls[ True]]
getType[ HoldComplete[ m_META$], _, _, meta_, _] :=
	With[ {info = Replace[ m, meta]},
		Switch[ info,
			{__},
			First[ info],
			m|{},
			$Failed,
			_,
			info
		]
	]
getType[ HoldComplete[ c_FIX$], _, fix_, _, _] :=
	With[ {t = Replace[ c, fix]},
		If[ t === c,
			$Failed,
		(*else*)
			t
		]
	]
getType[ HoldComplete[ VAR$[ i_Integer]], loose_, _, _, _] :=
	If[ 0 <= i < Length[ loose],
		loose[[i + 1]],
	(*else*)
		$Failed
	]
getType[ HoldComplete[ TYPED$[ term_, typ_]], _, _, _, decls_] := getFullTypeNoSorts[ typ, HoldComplete[ term], decls]
getType[ HoldComplete[ Lambda$TM[ RNG$[ TYPRNG$[ _, t_]], term_]], var_, fix_, meta_, decls_] :=
	With[ {t0 = getType[ HoldComplete[ term], Prepend[ var, t], fix, meta, decls]},
		If[ t0 === $Failed,
			$Failed,
		(*else*)
			FUN$TM[ t, t0]
		]
	]
getType[ HoldComplete[ f_[ a___]], var_, fix_, meta_, decls_] :=
	With[ {t = getType[ HoldComplete[ f], var, fix, meta, decls]},	(* 'getType' also works for multiple arguments *)
		If[ t === $Failed,
			$Failed,
		(*else*)
			stripArgumentTypes[ t, Length[ {a}]]	(* we do not get the type of the arguments, since the expression is assumed to be type-checked already *)
		]
	]
getType[ c:HoldComplete[ _], _, _, _, decls_] :=
	Replace[ getDeclaredType[ c, 0, decls],
		{
			{t_, _, _, Except[ True], ___} :> t,
			_ -> $Failed
		}
	]
getType[ HoldComplete[ term_, _, _], var_, fix_, meta_, decls_] := getType[ HoldComplete[ term], var, fix, meta, decls]
getType[ Hold[ term_], var_, fix_, meta_, decls_] := getType[ HoldComplete[ term], var, fix, meta, decls]
getType[ term_, var_, fix_, meta_, decls_] := getType[ HoldComplete[ term], var, fix, meta, decls]

stripArgumentTypes[ t_, 0] := t
stripArgumentTypes[ FUN$TM[ _, t_], n_Integer] := stripArgumentTypes[ t, n - 1]
stripArgumentTypes[ ___] := $Failed

applyTypeGenSpecRepl[ fixToVar_List, varToFix_List] :=
	applyTypeGenSpecRepl[ fixToVar, varToFix, True]
applyTypeGenSpecRepl[ fixToVar_List, varToFix_List, False][ expr_] :=
	expr /. STERM$[ term_, tfix_, tvar_] :> swapSortAssociations[ fixToVar, varToFix][ STERM$[ replaceTypes[ replaceTypes[ term, fixToVar], varToFix], tfix, tvar]]
applyTypeGenSpecRepl[ fixToVar_List, varToFix_List, True][ STERM$[ term_, tfix_, tvar_]] :=
	swapSortAssociations[ fixToVar, varToFix][ STERM$[ replaceTypes[ replaceTypes[ term, fixToVar], varToFix], tfix, tvar]]
applyTypeGenSpecRepl[ fixToVar_List, varToFix_List, True][ term_] :=
	replaceTypes[ replaceTypes[ term, fixToVar], varToFix]

tvpToSorts[ term_, tvp_List] :=
	With[ {list = DeleteCases[ Map[ extractSingle[ term, #[[1, 1]]] -> Last[ #] &, tvp], _ -> getDefaultSort[]]},
		Association[ list]
	]


(* ::Section:: *)
(*Type Inference*)

(* 'inferTypes[ term, target, idx1, idx2, loose, fix, meta, tfix, tvar, specgen]' type-checks and -annotates 'term'.
	'idx1' is the current type-variable index; it must be strictly greater than all type variables occurring in 'fix',
	'meta' and 'target'!
	'idx2' is a type-variable index specifying which type instantiations to include in the resulting list (see below).
	'loose' is the list of types of the loose bound variables.
	'fix' and 'meta' are the types of the a.b.f. constants and meta variables, repsectively, in the context.
	'tfix' and 'tvar' are the a.b.f. type constants and type variables, respectively, in the context.
	'specgen' specifies whether all type variables shall immediately be turned into fresh type constants ("spec"), or
	whether new a.b.f. type constants shall immediately be generalized ("gen"), or whether no specialization/
	generalization shall happen ('None'; default).
	The function returns
	- '$Failed' if 'expr' is ill-formed (see below) or ill-typed, and
	- '{{type, STERM$[ newTerm, newTfix, newTVar], newIdx, newFix, newMeta}, inst}' otherwise, where
		- 'type' is the primary type of 'TYPED$[ expr, target]',
		- 'newTerm' is the type-annotated expression, i.e. where every constant symbol has its most general type
			attached to it,
		- 'newIdx' is the new type-variable-index, i.e. 'idx' plus the number of fresh type variables that were
			generated during the execution (even if they do not appear in the final result any more).
		- 'newFix' and 'newMeta' are the lists of types of the *new* a.b.f. constants and meta variables, respectively,
			appearing in 'term'.
		- 'newTFix' is the association of sort constraints on the *new* a.b.f. type constants appearing in 'term', i.e.
			those that appear in 'term' but not in 'tfix'. Note that sort constraints on *new* a.b.f. type constants
			(i.e. those neither passed along with 'term' nor included in 'tfix') might be adjusted!
			In contrast to 'checkType', 'inferTypes' returns a.b.f. type constants that are passed along with 'term',
			even if they do not appear in the plain term; the same is true for type variables.
		- 'newTVar' is the new association of type variables; it includes both the variables occurring in 'term' and
			those occurring in 'tvar'. Type variables with index >= 'idx2' that are instantiated during type inference
			are not included in 'newTVar'.
		- 'inst' is a substitution for all type variables with index < 'idx2' that needs to be applied in order for the
			term to type-check with primary type 'type'.
	In order for this function to succeed, 'term' must be well-formed. This means that
	- all 'VAR$', 'FIX$' and 'META$' must be well-formed according to 'isWellFormedAtom',
	- all 'VAR$' must be bound,
	- it must not contain 'TVAR$' and 'TFIX$' outside of types,
	- all annotated types must be well-formed according to function 'isWellFormedType',
	- it must not contain 'SVAR$', 'SEQ$', 'SEQ0$' and 'SEQ1$' at all,
	- it must be fully curried,
	- all variable ranges must be single type ranges, and
	- 'Lambda$TM' is the only binder, i.e. the only function whose first argument is 'RNG$'.
*)
inferTypes[ term_, idx_Integer, args__] :=
	inferTypes[ term, Blank[], idx, args]
inferTypes[ STERM$[ term_, termTFix_Association, termTVar_Association],
			target_, idx1_Integer, idx2_Integer, loose_List, fix_List, meta_List, tfix_Association, tvar_Association, specgen_] :=
	With[ {allGivenTFix = mergeSorts[ tfix, termTFix]},
		If[ MatchQ[ allGivenTFix, _Association],
			Replace[ inferTypesAux[ HoldComplete[ term], fix, meta, allGivenTFix, loose, idx1, {}, {}, Association[], intersectSorts[ tvar, termTVar]],
				{
					{type_, flex_, newTerm0_HoldComplete, subst0_List, newLoose_List, newIdx_Integer, newFix0_List, newMeta0_List, newTFix0_Association, newTVar0_Association} :>
						If[ Length[ newLoose] === Length[ loose],
							Replace[ adjustTerm[ newTerm0, type, False, target, allGivenTFix, newTFix0, newTVar0, {106, {type, target}}],
								{
									{HoldComplete[ res_], t_, tinst_List, newTFix01_Association, newTVar1_Association} :>
										With[ {newTerm1 = replaceTypes[ res, tinst],
												newFix = newFix0 /. tinst,
												newMeta1 = newMeta0[[All, 2]] /. tinst,	(* types only *)
												newTFix1 = Join[ KeyComplement[ {termTFix, tfix}], newTFix01],	(* keys are disjoint *)
												newTVar2 = KeyDrop[ newTVar1, Select[ tinst[[All, 1]], First[ #] >= idx2 &]],
												subst = composeTypeSubsts[ tinst, subst0, idx2, idx2]},
											Replace[
												Switch[ specgen,
													"spec",
													With[ {spec = typeSpecializationReplacement[ {newTerm1, newFix[[All, 2]], newMeta1}, Join[ Keys[ allGivenTFix], Keys[ newTFix01]]]},
														{
															Join[ newTFix1, Association @@ Map[ Last[ #] -> Lookup[ newTVar2, #, getDefaultSort[]] &, spec]],
															KeyDrop[ newTVar2, spec[[All, 1]]],
															spec,
															0
														}
													],
													"gen",
													With[ {gen = typeGeneralizationReplacement[ Keys[ newTFix1], {}, newIdx]},
														{
															Association[],
															Join[
																newTVar2,
																Association @@ DeleteCases[ Map[ (Last[ #] -> Lookup[ newTFix1, First[ #], Default])&, gen], _ -> Default]
															],
															gen,
															Length[ gen]
														}
													],
													_,
													{newTFix1, intersectSorts[ tvar, newTVar2], {}, 0}
												],
												{newTFix_Association, newTVar_Association, repl_List, addIdx_Integer} :>
													With[ {newTerm = replaceTypes[ newTerm1, repl],
															newMeta = With[ {l = Length[ fix] + Length[ newFix]}, MapThread[ #1 -> {#2, l} &, {newMeta0[[All, 1]], newMeta1 /. repl}]]},
														{
															{t /. repl, STERM$[ newTerm, newTFix, newTVar], newIdx + addIdx, newFix /. repl, newMeta},
															composeTypeSubsts[ repl, subst, idx2, Infinity]
														}
													]
											]
										],
									_ -> $Failed
								}
							],
						(*else*)
							$Failed
						],
					_ -> $Failed
				}
			],
		(*else*)
			$Failed
		]
	]
inferTypes[ term_, target_, idx1_Integer, idx2_Integer, loose_List, fix_List, meta_List, tfix_Association, tvar_Association] :=
	inferTypes[ term, target, idx1, idx2, loose, fix, meta, tfix, tvar, None]

(* 'inferTypeAux[ term, ctxtFix, ctxtMeta, ctxtTFix, loose, idx, fix, meta, tfix, tvar]' infers the missing types in
	'term' and annotates the non-VAR$, non-FIX$, non-META$ atoms appearing in it. 'ctxtFix', 'ctxtMeta', 'ctxtTFix' and
	'loose' are the current typing context. 'idx' is the current type-variable index, and 'fix', 'meta', 'tfix' and
	'tvar' are the new a.b.f. term constants, meta variables, a.b.f. type constants and type variables, respectively,
	found so far in 'term'.
	The function either returns '$Failed' or a 10-tuple
	'{typ, flex, newTerm, subst, newLoose, newIdx, newFix, newMeta, newTFix, newTVar}', where
	- 'typ' is the full primary type of 'term',
	- 'flex' indicates whether 'typ' is flexible or not,
	- 'newTerm' is the type-annotated version of 'term',
	- 'subst' is the computed type substitution,
	- 'newLoose' are the new types of the loose bound variables,
	- 'newIdx' is the new type variable index,
	- 'newFix' and 'newMeta' are the new a.b.f. constants and meta variables, respectively, appearing in 'term' but not
		in the context; these lists also include the input 'fix' and 'meta' (but possibly with new types),
	- 'newTFix' and 'newTVar' are the new a.b.f. type constants and type variables, respectively, in 'term'; these
		associations include the input 'tfix' and 'tvar' (but possibly with smaller sorts). *)
inferTypesAux[ HoldComplete[ TYPED$[ term_, type_]], ctxtFix_, ctxtMeta_, ctxtTFix_, loose_, idx_, fix_, meta_, tfix_, tvar_] :=
	With[ {newTFix0 = isWellFormedType[ type, Join[ ctxtTFix, tfix]]},
		If[ ListQ[ newTFix0],
			Replace[ inferTypesAux[ HoldComplete[ term], ctxtFix, ctxtMeta, ctxtTFix, loose, idx, fix, meta, Join[ tfix, Association @@ Map[ (# -> getDefaultSort[])&, newTFix0]], tvar],
				{newType_, flex_, newTerm_, subst_, newLoose_, newIdx_, newFix_, newMeta_, newTFix_, newTVar_} :>
					With[ {res = compareTypes[ HoldComplete[ TYPED$[ term, type]], HoldComplete[ TYPED$[ Null, newType]], hasSortTFix[ ctxtTFix], {{newTFix, newTVar}, {}}, 0]},
						Which[
							MatchQ[ res, {{_Association, _Association}, _List}],
							Join[
								{newType, False, newTerm} /. Last[ res],
								{composeTypeSubsts[ Last[ res], subst, idx, idx]},
								{newLoose, newIdx, newFix, newMeta} /. Last[ res],
								First[ res]
							],
						(*else*)
							ListQ[ type],
							sowMessage[ 107, {type}];
							$Failed,
						(*else*)
							True,
							sowMessage[ 106, {newType, type}];
							If[ TrueQ[ flex],
								With[ {adj = adjustFlexType[ newTerm, type]},
									If[ adj === $Failed,
										$Failed,
									(*else*)
										{type, False, adj, subst, newLoose, newIdx, newFix, newMeta, newTFix, newTVar}
									]
								],
							(*else*)
								$Failed
							]
						]
					]
			],
		(*else*)
			sowMessage[ 102, {type}];
			$Failed
		]
	]
inferTypesAux[ a:HoldComplete[ c:(_Complex|_DirectedInfinity|_Rational)], _, _, _, loose_, idx_, args__, tvar_] :=
	Replace[ getDeclaredType[ a, idx, getAllConstDecls[ True]],
		{
			{full_, short_, newTVar_, flex_, i_} :> {full, flex, HoldComplete[ TYPED$[ c, short]], {}, loose, i, args, Join[ tvar, newTVar]},
			_ :> (sowMessage[ 109, {c}]; $Failed)
		}
	]
inferTypesAux[ v:HoldComplete[ _VAR$], _, _, _, loose_, args__] :=
	If[ isWellFormedAtom[ v],
		With[ {i = v[[1, 1]]},
			If[ i < Length[ loose],
				{loose[[i + 1]], False, v, {}, loose, args},
			(*else*)
				sowMessage[ 104, {First[ v]}];
				$Failed
			]
		],
	(*else*)
		sowMessage[ 103, {v}];
		$Failed
	]
inferTypesAux[ term:HoldComplete[ c_FIX$], ctxtFix_, _, _, loose_, idx_, fix_, args__] :=
	If[ isWellFormedAtom[ term],
		With[ {t0 = Replace[ c, ctxtFix]},
			If[ t0 === c,
				With[ {t1 = Replace[ c, fix]},
					If[ t1 === c,
						{TVAR$[ idx], False, term, {}, loose, idx + 1, Append[ fix, c -> TVAR$[ idx]], args},
					(*else*)
						{t1, False, term, {}, loose, idx, fix, args}
					]
				],
			(*else*)
				{t0, False, term, {}, loose, idx, fix, args}
			]
		],
	(*else*)
		sowMessage[ 103, {c}];
		$Failed
	]
inferTypesAux[ term:HoldComplete[ v_META$], _, ctxtMeta_, _, loose_, idx_, fix_, meta_, args__] :=
	If[ isWellFormedAtom[ term],
		With[ {t0 = Replace[ v, ctxtMeta]},
			If[ ListQ[ t0],	(* 't0' is of the form '{type, constIndex}' *)
				{First[ t0], False, term, {}, loose, idx, fix, meta, args},
			(*else*)
				With[ {t1 = Replace[ v, meta]},
					If[ t1 === v,
						{TVAR$[ idx], False, term, {}, loose, idx + 1, fix, Append[ meta, v -> TVAR$[ idx]], args},
					(*else*)
						{t1, False, term, {}, loose, idx, fix, meta, args}
					]
				]
			]
		],
	(*else*)
		sowMessage[ 103, {v}];
		$Failed
	]
inferTypesAux[ HoldComplete[ Lambda$TM[ RNG$[ TYPRNG$[ var_, typ_]], expr_]], ctxtFix_, ctxtMeta_, ctxtTFix_, loose_, idx_, fix_, meta_, tfix_, tvar_] :=
	If[ MatchQ[ var, _Symbol],
		With[ {newTFix0 = isWellFormedType[ typ, Join[ ctxtTFix, tfix]]},
			If[ ListQ[ newTFix0],
				Replace[
					inferTypesAux[
						HoldComplete[ expr],
						ctxtFix, ctxtMeta, ctxtTFix,
						Prepend[ loose, typ], idx, fix, meta,
						Join[ tfix, Association @@ Map[ (# -> getDefaultSort[])&, newTFix0]],
						tvar
					],
					{newType_, flex_, newTerm0_, subst_, newLoose_, res__} :>
						With[ {varType = First[ newLoose]},
							With[ {newTerm = Prepend[ newTerm0, RNG$[ TYPRNG$[ var, varType]]]},
								{FUN$TM[ varType, newType], False, Lambda$TM @@@ HoldComplete[ newTerm], subst, Rest[ newLoose], res}
							]
						]
				],
			(*else*)
				sowMessage[ 102, {typ}];
				$Failed
			]
		],
	(*else*)
		sowMessage[ 108, {var}];
		$Failed
	]
inferTypesAux[ HoldComplete[ f_[ a_]], ctxtFix_, ctxtMeta_, ctxtTFix_, loose_, idx_, args__] :=
	Replace[ inferTypesAux[ HoldComplete[ f], ctxtFix, ctxtMeta, ctxtTFix, loose, idx, args],
		{
			{$Failed, __} -> $Failed,
			{typeF0_, flexF_, termF_, substF_, resF__} :>
				Replace[ inferTypesAux[ HoldComplete[ a] /. substF, ctxtFix, ctxtMeta, ctxtTFix, resF],
					{typeA_, flexA_, termA_, substA_, newLoose_, newIdx_, newFix_, newMeta_, newTFix0_, newTVar0_} :>
						With[ {typeF = typeF0 /. substA, newF = replaceTypes[ termF, substA], subst0 = composeTypeSubsts[ substA, substF, idx, idx]},
							If[ MatchQ[ typeF, FUN$TM[ _, _]],
								If[ TrueQ[ flexF && !flexA],
									Replace[ unifyTypes[ First[ typeF], typeA, hasSortTFix[ ctxtTFix], {{newTFix0, newTVar0}, {}}, 0],
												(* since both 'typeF' and 'typeA' are full types, we can call 'unifyTypes' directly *)
										{
											{newSortCtxt_List, tsubst_List} :>
												Join[
													{
														Last[ typeF] /. tsubst,
														flexF,
														replaceTypes[ Delete[ newF /@ termA, {1, 0, 0}], tsubst],
														composeTypeSubsts[ tsubst, subst0, idx, idx],
														newLoose /. tsubst,
														newIdx
													},
													{newFix, newMeta} /. tsubst,
													newSortCtxt
												],
											_ :>
												(
													sowMessage[ 105, {typeF, typeA}];
													With[ {adj = adjustFlexType[ newF, FUN$TM[ typeA, TVAR$[ newIdx]]]},
														If[ adj === $Failed,
															$Failed,
														(*else*)
															{
																TVAR$[ newIdx],
																True,
																Delete[ adj /@ termA, {1, 0, 0}],
																subst0,
																newLoose,
																newIdx + 1,
																newFix,
																newMeta,
																newTFix0,
																newTVar0
															}
														]
													]
												)
										}
									],
								(*else*)
									Replace[ adjustTerm[ termA, typeA, Xor[ flexF, flexA], First[ typeF], ctxtTFix, newTFix0, newTVar0, {105, {typeF, typeA}}],
										{newTerm_, newType_, tsubst_, newTFix_, newTVar_} :>
											Join[
												{
													Last[ typeF] /. tsubst,
													flexF,
													replaceTypes[ Delete[ newF /@ newTerm, {1, 0, 0}], tsubst],
													composeTypeSubsts[ tsubst, subst0, idx, idx],
													newLoose /. tsubst,
													newIdx
												},
												{newFix, newMeta} /. tsubst,
												{newTFix, newTVar}
											]
									]
								],
							(*else*)
								Replace[ unifyTypes[ typeF, FUN$TM[ typeA, TVAR$[ newIdx]], hasSortTFix[ ctxtTFix], {{newTFix0, newTVar0}, {}}, 0],
									{
										{newSortCtxt_List, tsubst_List} :>
											Join[
												{
													Replace[ TVAR$[ newIdx], tsubst],
													flexF,
													replaceTypes[ Delete[ newF /@ termA, {1, 0, 0}], tsubst],
													composeTypeSubsts[ tsubst, subst0, idx, idx],
													newLoose /. tsubst,
													newIdx + 1
												},
												{newFix, newMeta} /. tsubst,
												newSortCtxt
											],
										_ :>
											(
												sowMessage[ 105, {typeF, typeA}];
												If[ TrueQ[ flexF && !flexA],
													(* unification failed, but type of 'f' is flexible; dual case does not need to be considered *)
													With[ {adj = adjustFlexType[ newF, FUN$TM[ typeA, TVAR$[ newIdx]]]},
														If[ adj === $Failed,
															$Failed,
														(*else*)
															{
																TVAR$[ newIdx],
																True,
																Delete[ adj /@ termA, {1, 0, 0}],
																subst0,
																newLoose,
																newIdx + 1,
																newFix,
																newMeta,
																newTFix0,
																newTVar0
															}
														]
													]
												]
											)
									}
								]
							]
						]
				]
		}
	]
inferTypesAux[ HoldComplete[ term:(TYPED$|VAR$|FIX$|META$|SVAR$|TVAR$|TFIX$|SEQ$|SEQ0$|SEQ1$|TYPRNG$|RNG$|Lambda$TM)], __] :=
	(
		sowMessage[ 101, {SymbolName[ term]}];
		$Failed
	)
inferTypesAux[ a:HoldComplete[ c_], _, _, _, loose_, idx_, args__, tvar_] :=
	Replace[ getDeclaredType[ a, idx, getAllConstDecls[ True]],
		{
			{full_, short_, newTVar_, flex_, i_} :> {full, flex, HoldComplete[ TYPED$[ c, short]], {}, loose, i, args, Join[ tvar, newTVar]},
			_ :> (sowMessage[ 109, {c}]; $Failed)
		}
	]

composeTypeSubsts[ a_, b_, threshold1_, 0] :=
	composeTypeSubsts[ a, {}, threshold1, Infinity]
composeTypeSubsts[ a_, b_, threshold1_, threshold2_Integer] :=
	composeTypeSubsts[ a, Cases[ b, HoldPattern[ TVAR$[ _?(# < threshold2 &), ___] -> _]], threshold1, Infinity]
composeTypeSubsts[ a_, b_, 0, Infinity] :=
	b /. a
composeTypeSubsts[ a_, b_, threshold1_Integer, Infinity] :=
	Join[ b/. a, Cases[ a, HoldPattern[ TVAR$[ _?(# < threshold1 &), ___] -> _]]]
composeTypeSubsts[ a_, b_, Infinity, Infinity] :=
	Join[ b /. a, a]

(* 'adjustTerm[ term, type, flex, target, ctxtTFix, tfix, tvar, msg]' adjusts the given term 'term' to make its type
	match 'target'. 'type' is the type of 'term', 'flex' specifies whether 'type' is flexible or not, 'target' is the
	target type, and 'msg' is the message that shall be sown if a problem occurs. The function returns '$Failed' if the
	term cannot be adjusted, and the 4-tuple '{newTerm, newType, inst, newTFix, newTVar}' otherwise; 'inst' has not
	been applied to 'newTerm' yet! *)
adjustTerm[ term_HoldComplete, type_, flex_, target_, ctxtTFix_, tfix_, tvar_, msg_] :=
	If[ target === Blank[],
		{term, type, {}, tfix, tvar},
	(*else*)
		With[ {rt1 = resultType[ type, 0], rt2 = resultType[ target, 0], ott = getObjectTruthType[]},
			If[ ott =!= FORM$TM && MatchQ[ rt2, {FORM$TM, _}] && rt1 =!= rt2,
				With[ {target0 = If[ Last[ rt2] === 0, ott, ReplacePart[ target, Table[ 2, {Last[ rt2]}] -> ott]]},
					Replace[ unifyTypes[ target0, type, hasSortTFix[ ctxtTFix], {{tfix, tvar}, {}}, 0],
						{
							{{tf_Association, tv_Association}, tsubst_List} :>
								With[ {adj = insertJudgement[ term, Last[ rt2]]},
									If[ adj === $Failed,
										$Failed,
									(*else*)
										{adj, target /. tsubst, tsubst, tf, tv}
									]
								],
							_ :>
								(
									sowMessage @@ msg;
									If[ TrueQ[ flex],
										(* unification failed, but type of 'term' is flexible. *)
										With[ {adj = insertJudgement[ adjustFlexType[ term, target0], Last[ rt2]]},
											If[ adj === $Failed,
												$Failed,
											(*else*)
												{adj, target, {}, tfix, tvar}
											]
										],
									(*else*)
										$Failed
									]
								)
						}
					]
				],
			(*else*)
				Replace[ unifyTypes[ target, type, hasSortTFix[ ctxtTFix], {{tfix, tvar}, {}}, 0],
					{
						{{tf_Association, tv_Association}, tsubst_List} :> {term, target /. tsubst, tsubst, tf, tv},
						_ :>
							(
								sowMessage @@ msg;
								If[ TrueQ[ flex],
									(* unification failed, but type of 'term' is flexible. *)
									With[ {adj = adjustFlexType[ term, target]},
										If[ adj === $Failed,
											$Failed,
										(*else*)
											{adj, target, {}, tfix, tvar}
										]
									],
								(*else*)
									$Failed
								]
							)
					}
				]
			]
		]
	]

adjustFlexType[ term_HoldComplete, type_] :=
	Replace[ getHeadAtom[ term, 0],
		{
			{HoldComplete[ TYPED$[ h_, typ_List]], i_Integer} :>
				With[ {t = ReplacePart[ getFullTypeNoSorts[ typ, HoldComplete[ h], getAllConstDecls[ True]], {Table[ 2, {i}]} -> type]},
					If[ t === $Failed,
						$Failed,
					(*else*)
						ReplacePart[ term, Join[ {1}, Table[ 0, {i}], {2}] -> t]
					]
				],
			{h:HoldComplete[ TYPED$[ _, _]], i_Integer} :>
				ReplacePart[ term, Join[ {1}, Table[ 0, {i}], {2}, Table[ 2, {i}]] -> type],
			_ -> $Failed
		}
	]

insertJudgement[ term_HoldComplete, 0] :=
	TYPED$[ TrueQ$TM, {}] @@@ HoldComplete[ term]
insertJudgement[ HoldComplete[ Lambda$TM[ r_RNG$, body_]], n_Integer?Positive] :=
	With[ {aux = insertJudgement[ HoldComplete[ body], n - 1]},
		If[ aux === $Failed,
			$Failed,
		(*else*)
			With[ {args = Join[ HoldComplete[ r], aux]},
				Lambda$TM @@@ HoldComplete[ args]
			]
		]
	]
insertJudgement[ ___] :=
	$Failed

(* 'mergeSorts' returns '$Failed' if the given sorts are not compatible. It does not remove entries with default sort.
	Entries that do not occur in all given associations are not treated as having default sort in those where they do
	not occur, but they are considered missing there. *)
mergeSorts[] :=
	Association[]
mergeSorts[ assoc_Association] :=
	assoc
mergeSorts[ assocs__Association] :=
	Catch[ Merge[ {assocs}, mergeSorts]]
mergeSorts[ sorts_List] :=
	If[ SameQ @@ sorts,
		First[ sorts],
	(*else*)
		Throw[ $Failed]
	]

End[]

EndPackage[]