(* Theorema 
    Copyright (C) 2010 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

BeginPackage[ "Theorema`CoreProvers`CoreProofMethods`"]

Needs[ "Theorema`Common`"]
Needs[ "Theorema`Language`"]
Needs[ "LazyList`", FileNameJoin[ {$TheoremaDirectory, "Theorema", "Components", "LazyList.m"}]]
Needs[ "Theorema`Core`Trusted`"]
Needs[ "Theorema`Core`Tools`"]
Needs[ "Theorema`Core`HOUnification`"]
Needs[ "Theorema`CoreProvers`"]

Begin["`Private`"]

EchoProofSit[][ ps_PRFSIT$, ___] :=
	(ps >>> "log.txt"; Hold[])
EchoProofSit[ n_Integer][ PRFSIT$[ G_List, __], ___] :=
	(If[ n === 0, G, G[[n]]] >>> "log.txt"; Hold[])
	

(* ::Section:: *)
(*Messages*)

sowCPMMessage[ text_String] :=
	sowMessage[ MSG$[ text]]
sowCPMMessage[ text_String, header_String] :=
	sowMessage[ MSG$[ text, header]]
sowCPMMessage[ id_Integer, header_String] :=
	sowCPMMessage[ id, {}, header]
sowCPMMessage[ id_Integer, header_String, src_String] :=
	sowCPMMessage[ id, {}, header, src]
sowCPMMessage[ args__] :=
	sowMessage[ MSG$[ cpmMessageToList, args]]


(* ::Section:: *)
(*Auxiliary Functions*)

addToList[ a_, {}, succF_, _, emptySucc_] :=
	If[ TrueQ[ emptySucc],
		succF[ a],
	(*else*)
		a
	]
addToList[ a_, b_, succF_, dupF_, _] :=
	With[ {out = DeleteDuplicates[ Join[ a, b]]},
		If[ Length[ out] < Length[ a] + Length[ b],
			dupF[ out],
		(*else*)
			succF[ out]
		]
	]

remFromList[ a_, {}, succF_, _, emptySucc_] :=
	If[ TrueQ[ emptySucc],
		succF[ a],
	(*else*)
		a
	]
remFromList[ a_, b_, succF_, missingF_, _] :=
	With[ {out = DeleteCases[ a, Alternatives @@ b]},
		If[ Length[ out] + Length[ b] > Length[ a],
			missingF[ out],
		(*else*)
			succF[ out]
		]
	]

preserveFormulaTags[ fml_STERM$, orig_FML$, tfix_Association] :=
	preserveFormulaTags[ fml, tfix, getOptionalComponent[ orig, "tags"], MemberQ[ orig, "arData" -> _], MemberQ[ orig, "rwData" -> _]]
preserveFormulaTags[ fml_STERM$, tfix_Association, tags_List, arData_, rwData_] :=
	Join[
		If[ tags === {}, {}, {"tags" -> tags}],
		If[ TrueQ[ arData], {"arData" -> generateReductionRule[ fml, True, True]}, {}],
		If[ TrueQ[ rwData], {"rwData" -> generateEqRewriteRule[ fml, True, tfix]}, {}]
	]


(* ::Section:: *)
(*Tacticals*)

(* 'evalTac', 'voidTac' and 'failTac' can be used to (re)set global parameters before/after some inferences or proof
	methods are applied. *)
	
SetAttributes[ evalTac, HoldRest];
evalTac[ m_INF$, statements___][ __] :=
	(statements; Hold[ m])
evalTac[ m_, statements___][ args__] :=
	(statements; m[ args])

SetAttributes[ {voidTac, failTac}, HoldAll];

voidTac[ statements___][ __] :=
	(statements; Hold[])

failTac[ statements___][ __] :=
	(statements; $Failed)

orTac[ m___][ __] :=
	lFromList[ Hold /@ Hold[ m]]

(* 'orElseTac' applies the first applicable proof method and completely discards all others. *)
orElseTac[] :=
	failTac[]
orElseTac[ m_, post___][ __] :=
	Hold[ INF$[ "if", m, voidTac[], orElseTac[ post]]]

andTac[ pre___, "!", post___][ __] :=
	Hold[ INF$[ "if", andTac[ pre, "commit"], andTac[ post], failTac[]]]
andTac[ m___][ __] :=
	Hold[ m]

andNewTac[ {}, ___] :=
	voidTac[]
andNewTac[ {m_, post___}][ PRFSIT$[ G_List, __], __] :=
	With[ {num = Length[ G]},
		Hold[ m, andNewTac[ {post}, num]]
	]
andNewTac[ {m_, post___}, numOrig_][ PRFSIT$[ G_List, __], __] :=
	With[ {num = Length[ G]},
	With[ {n = num + 1 - numOrig},
		Switch[ n,
			0,
			Hold[],
			1,
			Hold[ m, andNewTac[ {post}, num]],
			_,
			Hold[ mapAllTac[ andTac[ m, andNewTac[ {post}, num]], n]]
		]
	]]

thenTac[][ __] :=
	Hold[]
thenTac[ m_, rest___][ __] :=
	Hold[ INF$[ "if", m, thenTac[ rest], voidTac[]]]

tryTac[][ __] :=
	Hold[]
tryTac[ m_, rest___][ __] :=
	Hold[ INF$[ "if", m, tryTac[ rest], tryTac[ rest]]]

repeatTac[ m_][ __] :=
	Hold[ INF$[ "if", m, repeatTac[ m], voidTac[]]]

repeatNTac[ _, 0][ __] :=
	Hold[]
repeatNTac[ m_, n_Integer?Positive][ __] :=
	With[ {nn = n - 1},
		Hold[ m, repeatNTac[ m, nn]]
	]

repeatSTac[ rel_, m_][ PRFSIT$[ goals_, __], psNum_, ___] :=
	With[ {len = Length[ goals]},
		Hold[ repeatSTac[ rel, m, {len, psNum}]]
	]
repeatSTac[ rel_, m_, num_][ PRFSIT$[ goals_, __], psNum_, ___] :=
	If[ TrueQ[ rel[ {Length[ goals], psNum}, num]],
		Hold[],
	(*else*)
		Hold[ INF$[ "if", m, repeatSTac[ rel, m, num], voidTac[]]]
	]

repeatNewTac[ m_][ PRFSIT$[ G_List, __], __] :=
	With[ {num = Length[ G]},
		Hold[ INF$[ "if", m, repeatNewTac[ m, num], voidTac[]]]
	]
repeatNewTac[ m_, numOrig_][ PRFSIT$[ G_List, __], __] :=
	With[ {num = Length[ G]},
	With[ {n = num + 1 - numOrig},
	With[ {rec = INF$[ "if", m, repeatNewTac[ m, num], voidTac[]]},
		Switch[ n,
			0,
			Hold[],
			1,
			Hold[ rec],
			_,
			Hold[ mapAllTac[ rec, n]]
		]
	]]]

solveTac[ rel_, m_][ PRFSIT$[ goals_, __], psNum_, ___] :=
	With[ {len = Length[ goals]},
		Hold[ m, solveTac[ rel, len, psNum]]
	]
solveTac[ rel_, a_Integer, b_Integer][ PRFSIT$[ goals_, __], psNum_, ___] :=
	If[ TrueQ[ rel[ {Length[ goals], psNum}, {a, b}]],
		Hold[],
	(*else*)
		$Failed
	]

ifTac[ cond_, true_, false_][ ps_, psNum_, ___] :=
	If[ TrueQ[ cond[ ps, psNum]],
		Hold[ true],
	(*else*)
		Hold[ false]
	]

(* 'mapAllTac[ m, n][ ps, ___]' applies 'm' to the first 'n' subgoals in 'ps', where 'n' can be an integer or 'All' or
	"all"; a negative value refers to all but the last '-n - 1' subgoals (so '-1' is the same as 'All'). The subgoals
	are attacked in reverse order, starting with the 'n'-th and ending with the first.
	This tactical works properly only if 'm' does not rearrange subgoals and closes at most the subgoal it is applied
	to (i.e. the first subgoal in the proof situation 'm' is applied to); 'm' might introduce arbitrarily many new
	subgoals, though, but it is not applied to them repeatedly. *)
mapAllTac[ __][ PRFSIT$[ {}, __], ___] :=
	Hold[]
mapAllTac[ m_, n0:(_Integer?NonNegative|All|"all")][ ps:PRFSIT$[ G_List, __], args___] :=
	With[ {num = Length[ G]},
	With[ {n = If[ n0 === All || n0 === "all", num, n0]},
		If[ n <= num,
			mapAllTac[ m, num, n, n][ ps, args],
		(*else*)
			$Failed
		]
	]]
mapAllTac[ m_, n0_Integer?Negative][ ps:PRFSIT$[ G_List, __], args___] :=
	With[ {num = Length[ G]},
	With[ {n = num + 1 + n0},
		If[ NonNegative[ n],
			mapAllTac[ m, num, n, n][ ps, args],
		(*else*)
			$Failed
		]
	]]
mapAllTac[ m_, numOrig_, i_Integer, n_Integer][ PRFSIT$[ G_List, __], ___] :=
	If[ n === 0,
		Hold[],
	(*else*)
		With[ {num = Length[ G]},
		With[ {ii = i + num - numOrig, nn = n - 1},
			Join[ If[ ii === 1, Hold[], Hold[ INF$[ "rearrangeG", ii, 1]]], Hold[ m, mapAllTac[ m, num, ii, nn]]]
		]]
	]


(* ::Section:: *)
(*Optional Information in Proof Situations*)

(* 'updateAbbrevLists[ facts, terms, otherNew, otherDel][ ps, __]' updates the optional information stored in proof
	situations.
	'facts' is a list of new fact collections, with elements of the form '"name" -> list'. 'list' is supposed to
	be a list of indices, keys, labels, or existing fact collections.
	'terms' is a list of new term abbreviations, i.e. a list of transformation rules. The intended meaning of meta
	variables and type variables in term abbreviations is explained below; 'updateAbbrevLists' automatically computes
	the type variable positions of terms, unless this has been done already.
	'otherNew' and 'otherDel' are lists of other optional parameters that shall be added/updated or deleted,
	respectively. *)
updateAbbrevLists[ facts_, terms_, otherNew_, otherDel_][ ps_PRFSIT$, __] :=
	With[ {newData = updateAbbrevLists[ facts, terms, otherNew, otherDel, ps, Short]},
		Which[
			ListQ[ newData],
			Hold[ INF$[ "psData", newData]],
		(*else*)
			newData === Null,
			Hold[],
		(*else*)
			True,
			$Failed
		]
	]

updateAbbrevLists[ {}, {}, {}, {}, _, Short] :=
	Null
updateAbbrevLists[ {}, {}, {}, {}, ps_PRFSIT$, ___] :=
	getProofSitData[ ps]
updateAbbrevLists[ {}, {}, otherNew_List, otherDel_List, ps_PRFSIT$, ___] :=
	setProofSitData[ ps, otherNew, otherDel]
updateAbbrevLists[ {}, {Hold[ ellipsisAbbrev[ i_Integer]], rest___}, otherNew_List, otherDel_List, ps:PRFSIT$[ _, hyps_List, __], short___] :=
	updateAbbrevLists[
		{},
		Replace[ Quiet[ Check[ hyps[[i]], $Failed]],
			{
				FML$[ _, f_, __] :> Join[ ellipsisAbbrev[ f], {rest}],
				_ -> {rest}
			}
		],
		otherNew,
		otherDel,
		ps,
		short
	]
updateAbbrevLists[ {}, terms_List, otherNew_List, otherDel_List, ps_PRFSIT$, ___] :=
	With[ {termsNew =
			Replace[ terms,
				{
					(a_ -> STERM$[ t_, _, tv_Association]) :> (a -> {t, makeTypeVarPositions[ t, tv]}),
					(a_ -> t:Except[ {_, _}]) :> (a -> {t, makeTypeVarPositions[ t]})
				},
				{1}
			]},
		setProofSitData[ ps,
			Prepend[
				otherNew,
				"termAbbrevs" -> updateData[ getProofSitData[ ps, "termAbbrevs", Association[]], termsNew]
			],
			otherDel
		]
	]
updateAbbrevLists[ facts_List, terms_List, otherNew_List, otherDel_List, ps:PRFSIT$[ _, hyps_List, __], short___] :=
	With[ {factsData = getProofSitData[ ps, "factLists", Association[]]},
	With[ {factsNew =
			Cases[ facts,
				(k_String -> lst_List) :>
					(k ->
						Join @@ Replace[ lst,
							{
								ky:{_, _} :> {ky},
								STRING$[ lbl_String] :> Cases[ hyps, _[ ky_, _, lbl, ___] :> ky],
								coll_String :> Lookup[ factsData, coll, {}],
								i:Except[ 0, _Integer] :> Quiet[ Check[ {hyps[[i, 1]]}, {}]],
								_ -> {}
							},
							{1}
						])
			],
			factsDrop = Cases[ facts, (k_String -> Except[ _List]) :> k]},
		updateAbbrevLists[
			{},
			terms,
			Join[ If[ factsNew === {} === factsDrop, {}, {"factLists" -> updateData[ KeyDrop[ factsData, factsDrop], factsNew]}], otherNew],
			otherDel,
			ps,
			short
		]
	]]


(* ::Section:: *)
(*Initial Proof Steps*)

coreInitialProofStep[][ ps:PRFSIT$[ {G_}, _, _, {}, __], args__] :=
	With[ {fixes0 = getOptionalComponent[ $proofGoal, "fixes"],
			assms = getOptionalComponent[ $proofGoal, "assms"]},
	With[ {n = allCount[ G, 0, Length[ fixes0], True]},
		Which[
			n =!= 0,
			With[ {fixes = Fold[ With[ {i = Max[ -1, Max @@ Cases[ #1, FIX$[ #2, j_] :> j]] + 1}, Append[ #1, FIX$[ #2, i]]]&, {}, Take[ fixes0, n]]},
				Hold @@ Join[
					INF$[ "allI", #]& /@ fixes,
					If[ assms === {}, {}, {coreInitialProofStep[ assms]}]
				]
			],
		(*else*)
			assms === {},
			Hold[],
		(*else*)
			True,
			coreInitialProofStep[ assms][ ps, args]
		]
	]]
coreInitialProofStep[ assms_List][ PRFSIT$[ {G_}, __], __] :=
	With[ {n = assmCount[ G, 0, Length[ assms]]},
		If[ n =!= 0,
			Append[
				Hold @@ Flatten[ Table[ {INF$[ "makeImplicit", 1], INF$[ "drop", 1]}, {n}]],
				MapIndexed[ renameHyps[ {n + 1 - First[ #2], #1, {}, {}}]&, Take[ assms, n]]
			],
		(*else*)
			Hold[]
		]
	]
coreInitialProofStep[][ __] :=
	Hold[]
coreInitialProofStep[ _List][ __] :=
	Hold[]


(* ::Section:: *)
(*Reflection and Atomization*)

(* ::Subsection:: *)
(*Parameters*)

(* The following symbols appear in "Theorema`CoreProvers`CoreProofMethods`"-context. *)
Scan[
	With[ {$$name = #[[1]], $$desc = #[[2]]},
	With[ {$$args = Join[ ToExpression[ "Theorema`CoreProvers`CoreProofMethods`" <> $$name, InputForm, Hold], Hold[ "usage", "English"]]},
		Set @@ ReplacePart[ Hold[ $$args, $$desc], {1, 0} -> MessageName];
	]]&,
	{
		{"$reflectionFunction",
			"$reflectionFunction is a function that is applied whenever terms of the form 'TrueQ[ _]' shall be replaced
			by their meta-logic analogues of type 'FORM', without head 'TrueQ'. It gets the term to be reflected
			(including 'TrueQ'), the short specification of its main position, the proper main position, the
			subposition, and the whole subgoal as arguments and is supposed to return either '$Failed' or a sequence of
			proof commands."},
		{"$atomizationFunction",
			"$atomizationFunction is a function that is applied whenever terms of type 'FORM' shall be replaced by
			their object-logic analogues of the form 'TrueQ[ _]'. It gets the term to be atomized, the short
			specification of its main position, the proper main position, the subposition, and the whole subgoal as
			arguments and is supposed to return either '$Failed' or a sequence of proof commands."}
	}
]

If[ !ValueQ[ $reflectionFunction], $reflectionFunction = $Failed&];
If[ !ValueQ[ $atomizationFunction], $atomizationFunction = $Failed&];

reflectAtomizeTerm[ main_Integer, sub_List, G_, fun_, msg_String] :=
	Which[
		main === 0,
		reflectAtomizeTerm[ {assmCount[ G, 0, Infinity]}, sub, G, fun],
	(*else*)
		Positive[ main],
		If[ assmCount[ G, 0, main] === main,
			reflectAtomizeTerm[ main, sub, G, fun],
		(*else*)
			sowCPMMessage[ 1017, {main}, "Fatal", msg];
			$Failed
		],
	(*else*)
		True,
		With[ {i = assmCount[ G, main + 1, Infinity]},
			If[ Positive[ i],
				reflectAtomizeTerm[ i, sub, G, fun],
			(*else*)
				sowCPMMessage[ 1017, {main}, "Fatal", msg];
				$Failed
			]
		]
	]
reflectAtomizeTerm[ {main_Integer}, sub_List, G_, fun_, msg_String] :=
	If[ Negative[ main],
		With[ {i = assmCount[ G, main, Infinity]},
			If[ NonNegative[ i],
				reflectAtomizeTerm[ {i}, sub, G, fun],
			(*else*)
				sowCPMMessage[ 1017, {{main}}, "Fatal", msg];
				$Failed
			]
		],
	(*else*)
		If[ assmCount[ G, 0, main] === main,
			reflectAtomizeTerm[ {main}, sub, G, fun],
		(*else*)
			sowCPMMessage[ 1017, {{main}}, "Fatal", msg];
			$Failed
		]
	]
reflectAtomizeTerm[ i_Integer, sub_List, G_, fun_] :=
	With[ {pos = assmPosition[ i]},
		fun[ validExtract[ extractSingle[ G, pos], sub], i, pos, sub, G]
	]
reflectAtomizeTerm[ n_List, sub_List, G_, fun_] :=
	With[ {pos = Table[ 1, n]},
		fun[ validExtract[ extractSingle[ G, pos], sub], n, pos, sub, G]
	]


(* ::Subsection:: *)
(*Reflection*)

(* 'reflectTerm[ main, sub][ ps, _, _]' tries to reflect the term at the given position in the first subgoal of 'ps';
	to that end, '$reflectionFunction' is employed.
	'main' is the main position of the term to be reflected; it may be a non-zero integer, referring to the
	corresponding explicit assumption, or '0', referring to the atomic goal, or '{n}', referring to the goal after
	stripping off the first 'n' explicit assumptions (if 'n' is non-negative) or to the goal including the last '-n'
	explicit assumptions (if 'n' is negative).
	'sub' is the subposition of the part specified by 'main'. It must refer to a valid subposition, of course.
	The proof method fails if the given term cannot be reflected. *)
reflectTerm[ main_] :=
	reflectTerm[ main, {}]
reflectTerm[ main_, sub_List][ PRFSIT$[ {G_, ___}, __], ___] :=
	reflectAtomizeTerm[ main, sub, G, reflectTermAux, "reflectTerm"]

reflectTermAux[ $Failed, _, _List, sub_List, _] :=
	(sowCPMMessage[ 1018, {sub}, "Fatal", "reflectTerm"]; $Failed)
reflectTermAux[ term_, main_, pos_List, sub_List, G_] :=
	Replace[ $reflectionFunction[ term, main, pos, sub, G],
		Except[ _Hold] :> (sowCPMMessage[ 191, {term}, "Fatal"]; $Failed)
	]


(* ::Subsection:: *)
(*Atomization*)

(* 'atomizeTerm[ main, sub][ ps, _, _]' tries to atomize the term at the given position in the first subgoal of 'ps';
	to that end, '$atomizationFunction' is employed.
	'main' and 'sub' determine the position of the term to be atomized; they have precisely the same meaning as in
	'reflectTerm'.
	The proof method fails if the given term cannot be atomized. *)
atomizeTerm[ main_] :=
	atomizeTerm[ main, {}]
atomizeTerm[ main_, sub_List][ PRFSIT$[ {G_, ___}, __], ___] :=
	reflectAtomizeTerm[ main, sub, G, atomizeTermAux, "atomizeTerm"]

atomizeTermAux[ $Failed, _, _List, sub_List, _] :=
	(sowCPMMessage[ 1018, {sub}, "Fatal", "atomizeTerm"]; $Failed)
atomizeTermAux[ term_, main_, pos_List, sub_List, G_] :=
	Replace[ $atomizationFunction[ term, main, pos, sub, G],
		Except[ _Hold] :> (sowCPMMessage[ 192, {term}, "Fatal"]; $Failed)
	]


(* ::Section:: *)
(*Manipulating Explicit Assumptions*)

(* ::Subsection:: *)
(*Making Assumptions Explicit*)

(* 'makeExplicitAssms[ assms, sub][ ps, _, _]' makes the hypotheses of 'ps' specified by 'assms' explicit in the first
	'sub' subgoals of 'ps'; 'sub' may be an integer, 'All' or "all"; in the latter cases, the assumptions are added to
	all subgoals. Hypotheses that do not exist make the of the proof method fail completely.
	The elements of 'assms' can be indices, keys or labels. *)
makeExplicitAssms[ assms_List] :=
	makeExplicitAssms[ assms, 1]
makeExplicitAssms[ assms_List, sub_][ PRFSIT$[ G_List, hyps_List, ___], __] :=
	Catch[
		With[ {n = Length[ hyps] + 1, errFun = (sowCPMMessage[ 1, {#}, "Fatal"]; Throw[ $Failed])&},
		With[ {cmd =
				Map[
					INF$[
						"makeExplicit",
						Replace[ #,
							{
								i:Except[ 0, _Integer] :> If[ Negative[ i], n + i, i],
								{i:Except[ 0, _Integer], _} :> If[ Negative[ i], n + i, i],
								k:{_String, _String} :>
									Quiet[ Check[ Position[ hyps, _[ k, __], {1}, 1, Heads -> False][[1, 1]], errFun[ k]]],
								{k:{_String, _String}, _} :>
									Quiet[ Check[ Position[ hyps, _[ k, __], {1}, 1, Heads -> False][[1, 1]], errFun[ k]]],
								STRING$[ l_String] :>
									Quiet[ Check[ formulaPosition[ l, 1, None, hyps][[1, 1]], errFun[ l]]],
								a_ :> errFun[ a]
							}
						]
					]&,
					Reverse[ assms]
				]},
			mapAllInline[ Hold @@ cmd, sub, Length[ G]]
		]]
	]


(* ::Subsection:: *)
(*Rearranging Assumptions*)

(* 'rearrangeAssms[ i, j][ ps, _, mode]' moves the 'i'-th explicit assumption of the first subgoal in 'ps' to the 'j'-th place.
	'i' and 'j' may both be negative, in which case counting starts from the very last assumption.
	Depending on 'mode', the proof method either directly manipulates the proof situation (fast- and trace mode), or employs
	theorem "Rearrange" (kernel mode). *)
rearrangeAssms[ i_Integer, i_] :=
	voidTac[]
rearrangeAssms[ i:Except[ 0, _Integer], j:Except[ 0, _Integer]][ ps:PRFSIT$[ {G_, ___}, __], ___, mode_] :=
	If[ Negative[ i] || Negative[ j],
		With[ {ag = toAssmsGoal[ G, {}, Infinity]},
		With[ {num = Length[ ag]},
		With[ {i0 = If[ Negative[ i], num + i, i], j0 = If[ Negative[ j], num + j, j]},
			rearrangeAssms[ i0, j0, num - 1, ag, ps, mode]
		]]],
	(*else*)
		With[ {ag = toAssmsGoal[ G, {}, Max[ i, j]]},
			rearrangeAssms[ i, j, Length[ ag] - 1, ag, ps, mode]
		]
	]
rearrangeAssms[ i_Integer?Positive, j_Integer?Positive, num_Integer, ag_List, args__] :=
	With[ {M = Max[ i, j]},
		If[ i === j,
			Hold[],
		(*else*)
			rearrangeAssms[ i, j, num, M, ag, args]
		] /; M <= num
	]
rearrangeAssms[ i_Integer, j_Integer, _Integer, __] :=
	(
		sowCPMMessage[ 2, {i, j}, "Fatal"];
		$Failed
	)
rearrangeAssms[ i_Integer, j_Integer, num_Integer, M_Integer, ag_List, _PRFSIT$, 2] :=
	Hold @@ Join[
		{INF$[ "theorem", THM$[ fastThmFromLabel[ "Rearrange"], i, j]]},
		Map[ INF$[ "allE", 1, makeSTERM[ #]]&, If[ M === num, ag, Append[ Take[ ag, M], fromAssmsGoal[ Drop[ ag, M]]]]],	(* all instances type-check *)
		{INF$[ "impE", 1, num + assmCount[ Last[ ag], 0, Infinity]]}	(* here one could 'testIdentityTac[ ...]' *)
	]
rearrangeAssms[ i_Integer, j_Integer, _Integer, _Integer, ag_List, ps_PRFSIT$, _] :=
	{ReplacePart[ ps, {1, 1} -> fromAssmsGoal[ Insert[ Delete[ ag, i], ag[[i]], j]]], {}, {THM$[ fastThmFromLabel[ "Rearrange"], i, j]}, Hold[]}


(* ::Subsection:: *)
(*Duplicating Assumptions*)

(* 'duplicateAssms[ i][ ps, _, mode]' duplicates the explicit assumptions at the given indices, by prepending copies of them
	to the explicit assumptions. Indices may occur several times, in which case the corresponding assumptions are duplicated
	several times. The order of the elements in 'i' is of course respected when duplicating the assumptions, i.e. the first
	element of 'i' generates the first assumption in the new subgoal, and so on.
	Invalid indices cause the proof method to fail instantly. *)
duplicateAssms[ i_Integer] :=
	duplicateAssms[ {i}]
duplicateAssms[ {}] :=
	voidTac[]
duplicateAssms[ i_List][ ps:PRFSIT$[ {G_, ___}, __], _, mode_] :=
	With[ {ag = toAssmsGoal[ G, {}, Infinity]},
	With[ {num = Length[ ag]},
		Catch[
			duplicateAssms[
				Replace[ i,
					{
						j_Integer?Negative :> If[ -j >= num, sowCPMMessage[ 3, {j}]; Throw[ $Failed], num + j],
						j_Integer?Positive :> If[ j >= num, sowCPMMessage[ 3, {j}]; Throw[ $Failed], j],
						j_ :> (sowCPMMessage[ 3, {j}]; Throw[ $Failed])
					},
					{1}
				],
				num - 1,
				ag,
				ps,
				mode
			]
		]
	]]
duplicateAssms[ i_List, num_, ag_List, _PRFSIT$, 2] :=
	With[ {M = Max @@ i},
		Hold @@ Join[
			{INF$[ "theorem", THM$[ fastThmFromLabel[ "Rearrange"], Join[ i, Range[ M]]]]},
			Map[ INF$[ "allE", 1, makeSTERM[ #]]&, If[ M === num, ag, Append[ Take[ ag, M], fromAssmsGoal[ Drop[ ag, M]]]]],	(* all instances type-check *)
			{INF$[ "impE", 1, num + assmCount[ Last[ ag], 0, Infinity]]}	(* here one could 'testIdentityTac[ ...]' *)
		]
	]
duplicateAssms[ i_List, _, ag_List, ps_PRFSIT$, _] :=
	{ReplacePart[ ps, {1, 1} -> fromAssmsGoal[ Join[ Extract[ ag, List /@ i], ag]]], {}, {THM$[ fastThmFromLabel[ "Rearrange"], Join[ i, Range[ Max @@ i]]]}, Hold[]}


(* ::Section:: *)
(*Knowledge Management*)

Get[ FileNameJoin[ {$TheoremaDirectory, "Theorema", "CoreProvers", "CoreProofMethods", "Knowledge.m"}]];


(* ::Section:: *)
(*Subgoals*)

(* ::Subsection:: *)
(*Arrangement*)

(* 'reverseSubgoals[ n][ ps, ___]' reverses the 'n' first subgoals of 'ps'. 'n' can be an integer of 'All' or "all",
	with exactly the same meaning as in 'mapAllTac'.
	In fact, the main purpose of 'reverseSubgoals' is to reverse subgoals before applying 'mapAllTac', in order to
	attack the subgoals in the *right* order (keep in mind that 'mapAllTac' for technical reasons attacks the subgoals
	in reverse order). *)
reverseSubgoals[ -1|0|1] :=
	voidTac[]
reverseSubgoals[ n:(_Integer|All|"all")][ PRFSIT$[ G_List, ctxt__], ___, mode_] :=
	With[ {num = Length[ G]},
	With[ {spec = If[ IntegerQ[ n], n, num]},
		Which[
			-1 <= spec <= 1,
			Hold[],
		(*else*)
			mode === 2,
			Which[
				Negative[ spec],
				With[ {spec1 = num + spec + 1},
					If[ Positive[ spec1],
						Hold @@ Table[ INF$[ "rearrangeG", spec1, num - i], {i, 0, -(spec + 2)}],
					(*else*)
						$Failed
					]
				],
			(*else*)
				spec <= num,
				Hold @@ Table[ INF$[ "rearrangeG", 1, spec - i], {i, 0, spec - 2}],
			(*else*)
				True,
				$Failed
			],
		(*else*)
			True,
			Which[
				spec === num || spec === -num,
				{PRFSIT$[ Reverse[ G], ctxt], {}, {}, Hold[]},
			(*else*)
				Negative[ spec],
				With[ {spec1 = num + spec},
					If[ Positive[ spec1],
						{PRFSIT$[ Join[ Take[ G, spec1], Reverse[ Drop[ G, spec1]]], ctxt], {}, {}, Hold[]},
					(*else*)
						$Failed
					]
				],
			(*else*)
				spec <= num,
				{PRFSIT$[ Join[ Reverse[ Take[ G, spec]], Drop[ G, spec]], ctxt], {}, {}, Hold[]},
			(*else*)
				True,
				$Failed
			]
		]
	]]


(* ::Subsection:: *)
(*Setup*)

(* ::Subsubsection:: *)
(*Top-Level Proof Method*)

(* 'setupSubgoals[ goals][ ps, ___]' sets up a new subgoal consisting of the "conjunction" of the formulas in 'goals'
	by employing basic inference "cut". In case of success, the conjuncts are added individually to the list of
	hypotheses of 'ps'. The proof method is only applicable if none of the formulas in 'goals' contains free variables
	or meta variables. Type variables and new a.b.f type constants that appear in multiple subgoals are treated like
	being distinct; basic inference "cut" entails this.
	The elements of 'goals' are supposed to be either plain formulas (possibly inside 'EXPR$', in which case types
	are inferred rather than checked) or of the form 'LABELED$[ fml, data...]', where 'fml' is the actual formula and
	'data' is additional data for constructing the 'FML$' structures in the hypotheses (e.g. the label, which by
	default is set to "?").
	Note that the order of subgoals in the first proof situation corresponds precisely to the order of the formulas in
	'goals', as does the order of the new formulas in the hypotheses of the second proof situation. *)
setupSubgoals[ {}] :=
	voidTac[]
setupSubgoals[ {goal_}] :=
	setupSubgoal[ goal]	(* more efficient *)
setupSubgoals[ goals0:Except[ {}|{_}, _List]][ ps:PRFSIT$[ _, _, idx_, fix_, _, tfix_Association, tvar_Association, __], args___, mode_] :=
	With[ {goals1 =
			Transpose[
				Replace[ goals0,
					{
						LABELED$[ f_, lbl_String, data___] :> {f, {lbl, data}},
						LABELED$[ f_, _, data___] :> {f, {"?", data}},
						f_ :> {f, {"?"}}
					},
					{1}
				]
			]},
		Replace[ setupSubgoalsAux[ First[ goals1], idx, fix, tfix, tvar, getTermAbbrevList[ ps], "gen", mode =!= 2],
			{goals_STERM$, newIdx_, newTFix1_Association, tsubst_List, newAbbrevs_List} :>
				setupSubgoalsAux[
					goals,
					Last[ goals1],
					newIdx,
					newTFix1,
					tsubst,
					newAbbrevs,
					ps,
					mode
				]
		]
	]

setupSubgoalsAux[ goals0_List, idx0_, fix0_, tfix0_Association, tvar0_Association, abbrevs_, specgen_, check_] :=
	Module[ {subst = {}, idx = idx0, fix = fix0, tfix = Association[], tvar = tvar0, newAbbrevs = {}},
		Catch[
			With[ {goals =
					With[ {param = If[ TrueQ[ check], Replace, None]},
						Map[
							Function[ G,
								Replace[ makeCoreExprPatt[ G, FORM$TM, idx, fix, {}, Join[ tfix, tfix0], tvar, abbrevs, specgen, normalize, param],
									{
										{{_, STERM$[ f0_, tf_Association, tv_Association], i_, {}, {}}, ts_List, {}, abb_} :>
											If[ tf =!= Association[] && MemberQ[ Lookup[ tfix, Keys[ tf], $Failed], Except[ $Failed]],
												sowCPMMessage[ 1001, "Fatal", "setupSubgoals"];
												$Failed,
											(*else*)
												idx = i;
												fix = fix /. ts;
												tfix = Join[ tfix, tf];
												tvar = tv;
												subst = Join[ ts, subst];	(* no need to replace 'subst' w.r.t. 'ts', since type variables are disjoint *)
												newAbbrevs = Join[ abb, newAbbrevs];	(* new abbreviations must come before old ones *)
												f0
											],
										_ :> (sowCPMMessage[ 31, "Fatal"]; Throw[ $Failed])
									}
								]
							],
							goals0
						]
					]},
				With[ {subst0 =
						typeSpecializationReplacement[
							Replace[
								DeleteDuplicates[ Cases[ goals, _FIX$, {0, Infinity}, Heads -> True]],
								fix,
								{1}
							],	(* no need to replace w.r.t. 'subst', since 'subst' has been applied to 'fix' already *)
							Join[ Keys[ tfix0], Keys[ tfix]]
						]},
					If[ subst0 =!= {}, sowCPMMessage[ 1002, "Warning", "setupSubgoals"]];
					(* we must apply both 'subst' and 'subst0' to 'goals', because the goals might share type
						variables (e.g. because of a.b.f. term constants) *)
					{
						applyTypeGenSpecRepl[ {}, subst0][ STERM$[ replaceTypes[ goals, subst], tfix, tvar]],
						idx,
						With[ {list = Map[ Last[ #] -> Lookup[ tvar, First[ #], getDefaultSort[]] &, subst0]},	(* entries with default sort must be kept! *)
							Association[ list]
						],
						Join[ subst /. subst0, subst0],
						applyTypeGenSpecRepl[ {}, subst0, False][ replaceTypes[ DeleteDuplicatesBy[ newAbbrevs, First], subst]]
					}
				]
			]
		]
	]
setupSubgoalsAux[ goals:STERM$[ _, newTFix2_Association, newTVar_Association], data_, newIdx_, newTFix1_Association, tsubst_, newAbbrevs_,
		ps:PRFSIT$[ G_, hyps_, idx_, fix_, meta_, tfix_Association, tvar_Association, fid_, ctxt___], 0] :=
	With[ {len = Length[ First[ goals]]},
	With[ {tvp = getPositions[ goals, _TVAR$],
			tfp = If[ newTFix1 === Association[], {}, getPositions[ goals, Alternatives @@ Keys[ newTFix1]]]},	(* only new a.b.f. type constants may be generalized *)
	With[ {fSpec = specializeTypes[ goals, Keys[ tfix], tvp],
			fGen = generalizeTypes[ goals, {}, newIdx, tfp][[1, 1]]},
		{
			{
				Join[
					PRFSIT$[
						fSpec[[1, 1]],
						hyps,
						0,
						fix /. tsubst,
						{},
						Join[ tfix, newTFix2, Last[ fSpec]],
						Association[],
						fid
					],
					PRFSIT$ @@ updateAbbrevLists[ {"this" -> Null}, newAbbrevs, {}, {"calc", "cases"}, ps]
				],
				With[ {newKeys = Table[ makeFormulaKey[ fid + i], {i, 0, len - 1}]},
					Join[
						PRFSIT$[
							replaceTypes[ G, tsubst],
							Join[
								MapThread[
									FML$[
										#3,
										#1,
										First[ #2],
										"TypeVarPositions" ->
											Cases[
												Join[ tfp, tvp],
												(allPos_List -> v_) :>
													With[ {thinnedPos = Cases[ allPos, {1, #4, p___} :> {p}]},
														(thinnedPos -> Lookup[ If[ MatchQ[ v, _TVAR$], newTVar, newTFix1], v, getDefaultSort[]]) /; thinnedPos =!= {}
													]
											],
										Sequence @@ Rest[ #2]
									]&,
									{fGen, data, newKeys, Range[ len]}
								],
								hyps
							],
							idx,
							fix /. tsubst,
							meta /. tsubst,
							Join[ tfix, newTFix2],
							tvar,
							fid + len
						],
						PRFSIT$ @@ updateAbbrevLists[ {"this" -> newKeys}, Join[ ellipsisAbbrev[ Last[ fGen]], newAbbrevs], {}, {"chainedFacts"}, ps]
					]
				]
			},
			{},
			{},
			Hold[]
		}
	]]]
setupSubgoalsAux[ goals_STERM$, data_, idx_, tfix1_, tsubst_, newAbbrevs_, ps_, 1] :=
	Replace[ setupSubgoalsAux[ goals, data, idx, tfix1, tsubst, newAbbrevs, ps, 0],
		{sits:{_PRFSIT$, PRFSIT$[ _, hyps_List, __]}, assms_List, thms_List, cmd_Hold} :>
			{
				makeANDNODE[ makePRFINFO[ name -> setupSubgoals, used -> {}, generated -> {Take[ hyps, Length[ goals]]}], sits],
				assms,
				thms,
				cmd
			}
	]
setupSubgoalsAux[ goals1:STERM$[ goals_List, tfix_Association, tvar_Association], data_, _, _, tsubst_, newAbbrevs_, ps_, 2] :=
	With[ {len = Length[ goals],
			form =
				ReplacePart[
					goals1,
					1 -> TYPED$[ CoreAll$TM, {FORM$TM}][ Lambda$TM[ RNG$[ TYPRNG$[ Null, FORM$TM]], TYPED$[ CoreImplies$TM, {}][ fromAssmsGoal[ goals, VAR$[ 0]]][ VAR$[ 0]]]]
				]},
			(* The cut formula is '/\C. (goals_1 ==> goals_2 ==> ... ==> goals_n ==> C) ==> C', which corresponds to the conjunction of the 'goal_i'. *)
		Hold @@ Join[
			If[ tsubst === {}, {}, {INF$[ "instT", tsubst]}],
			If[ newAbbrevs === {}, {}, {updateAbbrevLists[ {}, newAbbrevs, {}, {}]}],	(* this is done before "cut", so the new abbreviations appear in both subgoals *)
			{
				INF$[ "cut", form],
				INF$[ "allI", FIX$[ Module[ {C}, C], 0]],
				INF$[ "impE", 1, 0],	(* here one could 'testIdentityTac[ ...]' *)
				updateAbbrevLists[ {"this" -> Null}, {}, {}, {"calc", "cases"}],
				INF$[ "rearrangePS", 2, 1]
			},
			Join @@ Reverse[
				MapThread[
					With[ {elim = assmCount[ #1, 0, Infinity]},
						Join[
							{
								INF$[ "cut", STERM$[ #1, tfix, tvar]],
								INF$[ "makeExplicit", len + 1 - #3],
								instantiateCoreAllGoal[],
								INF$[ "impE", 1, elim],
								INF$[ "impE", #3, elim],
								INF$[ "close"]
							},
							If[ #2 === {"?"}, {}, {Join[ INF$[ "hypData", 1], INF$ @@ #2]}]
						]
					]&,
					{goals, data, Range[ len]}
				]
			],
			{INF$[ "forget", len + 1], updateAbbrevLists[ {"this" -> Range[ len]}, {Hold[ ellipsisAbbrev[ len]]}, {}, {"chainedFacts"}], INF$[ "rearrangePS", 1, 2]}
		]
	]


(* ::Subsubsection:: *)
(*Setting Up Subgoals with "cut"*)

(* 'setupSubgoal[ goal][ ps, ___]' sets up the new subgoal 'goal' using basic inference "cut".
	'goal' must not contain *any* meta variable, the new subgoal is added as a completely separate proof situation and
	'goal' is added to the implicit hypotheses of the current proof situation. Moreover, the types of all polymorphic
	a.b.f. constants in 'goal' are automatically specialized before the subgoal is set up.
	As in 'setupSubgoals', 'goal' can be either a plain formula or of the form 'LABELED$[ fml, data...]'. *)
setupSubgoal[ LABELED$[ goal_, data___]] :=
	setupSubgoal[ goal, data]
setupSubgoal[ goal_, STRING$[ lbl_String], data___] :=
	setupSubgoal[ goal, lbl, data]
setupSubgoal[ goal_] :=
	setupSubgoal[ goal, "?"]
setupSubgoal[ goal_, data___][ ps_PRFSIT$, args___, 1] :=
	Replace[ executeProofCommands[ Hold[ setupSubgoal[ goal, data]], {ps}, 0, False],
		{
			PRFSTAT$[ _, sits:{_PRFSIT$, PRFSIT$[ _, {newGoal_FML$, ___}, __]}, assms_List, thms_List, ___] :>
				{
					makeANDNODE[ makePRFINFO[ name -> setupSubgoal, used -> {}, generated -> {newGoal}], sits],
					{},
					{},
					Hold[]
				},
			_ -> $Failed
		}
	]
setupSubgoal[ goal_, lbl_String, data___][ ps:PRFSIT$[ _, _, idx_, fix_, _, tfix_Association, tvar_Association, __], ___] :=
	Replace[ makeCoreExprPatt[ goal, FORM$TM, idx, fix, {}, tfix, tvar, getTermAbbrevList[ ps], "gen", normalize, None],
		{
			{{_, STERM$[ g0_, Association[], tv_Association], _, {}, {}}, ts_, {}, abb_} :>
				With[ {subst0 =
						typeSpecializationReplacement[
							Replace[ DeleteDuplicates[ Cases[ g0, _FIX$, {0, Infinity}, Heads -> True]], fix, {1}] /. ts,
							Keys[ tfix]
						]},
				With[ {subst = Join[ ts /. subst0, subst0]},	(* "subst0 after ts" *)
					If[ subst0 =!= {}, sowCPMMessage[ 1002, "Warning", "setupSubgoal"]];
					Hold @@ Join[
						If[ subst === {}, {}, {INF$[ "instT", subst]}],
						If[ abb === {}, {}, {updateAbbrevLists[ {}, abb, {}, {}]}],
						{
							INF$[ "cut", applyTypeGenSpecRepl[ {}, subst0][ STERM$[ g0, Association[], tv]]],
							updateAbbrevLists[ {"this" -> Null}, {}, {}, {"calc", "cases"}],
							INF$[ "rearrangePS", 2, 1]
						},
						If[ {lbl, data} === {"?"}, {}, {INF$[ "hypData", 1, lbl, data]}],
						{
							updateAbbrevLists[ {"this" -> {1}}, {Hold[ ellipsisAbbrev[ 1]]}, {}, {"chainedFacts"}],
							INF$[ "rearrangePS", 1, 2]
						}
					]
				]],
			_ :> (sowCPMMessage[ 31, "Fatal"]; $Failed)
		}
	]
setupSubgoal[ goal_, _, data___][ args__] :=
	setupSubgoal[ goal, "?", data][ args]


(* ::Subsubsection:: *)
(*Setting Up Subgoals with "Modus Ponens"*)

(* 'setupSubgoalsMP[ goals, n][ ps, ___]' sets up the new subgoals 'goals' using theorem "Modus Ponens".
	'n' specifies how many of the explicit assumptions of the current subgoal shall become explicit assumptions of the
	new subgoal (negative integers count from the end, so '-1' refers to all assumptions). Summarizing: if 'n' is
	non-negative, the elements of 'goals' become the 'n+1'-th through 'n+Length[goals]' explicit assumptions of the
	current subgoal.
	The 'goals' can be any ground formulas, the new subgoals are prepended to the list of subgoals of the current proof
	situation. *)
setupSubgoalsMP[ goals_] :=
	setupSubgoalsMP[ goals, -1]
setupSubgoalsMP[ {}, _] :=
	voidTac[]
setupSubgoalsMP[ goal:Except[ _List], n_][ args__] :=
	setupSubgoalsMP[ {goal}, n][ args]
setupSubgoalsMP[ goals_List, n0_Integer][ ps:PRFSIT$[ {G_, R___}, hyps_, idx_, fix_, meta_, tfix_, tvar_, fid_, ___], ___, mode_] :=
	With[ {ac = assmCount[ G, 0, Infinity]},
	With[ {n = If[ Negative[ n0], ac + 1 + n0, n0], thm = THM$[ fastThmFromLabel[ "Modus Ponens"], Length[ goals]]},
		If[ 0 <= n <= ac,
			If[ mode === 2,
				Hold @@ {
					INF$[ "theorem", thm],
					instantiateCoreAll[ goals],
					instantiateCoreAllGoal[ 1, n],
					INF$[ "impE", 1, ac - n]	(* here one could 'testIdentityTac[ ...]' *)
				},
			(*else*)
				Replace[ makeCoreExprPattListFast[ goals, FORM$TM, idx, fix, meta, tfix, tvar, getTermAbbrevList[ ps], None, normalize, Replace],
					{newGoals_List, newIdx_, newFix_, newMeta_, newTFix_, newTVar_, newAbbrevs_, tsubst_} :>
						{
							With[ {GNew = replaceTypes[ G, tsubst], p = Table[ 1, {n}]},
								Join[
									PRFSIT$[
										Join[
											Map[ ReplacePart[ GNew, {p} -> #]&, newGoals],
											{ReplacePart[ GNew, {p} -> Fold[ TYPED$[ CoreImplies$TM, {}][ #2][ #1]&, extractSingle[ GNew, p], Reverse[ newGoals]]]},
											replaceTypes[ {R}, tsubst]
										],
										hyps,
										newIdx,
										newFix,
										newMeta,
										newTFix,
										newTVar,
										fid
									],
									PRFSIT$ @@ updateAbbrevLists[ {}, newAbbrevs, {}, {}, ps]
								]
							],
							{},
							{thm},
							Hold[]
						}
				]
			],
		(*else*)
			sowCPMMessage[ 32, {n0}, "Fatal"];
			$Failed
		]
	]]

(* No new a.b.f. term constants, no new meta variables, no instantiations of existing meta variables.
	The function returns the *complete* lists of a.b.f. term/type constants, meta variables, and term abbreviations. *)
makeCoreExprPattListFast[ l_List, typ_, idx_Integer, fix_List, meta_List, tfix_Association, tvar_Association, abbrevs_, specgen_, post_, param_] :=
	makeCoreExprPattListFast[ l, typ, idx, fix, meta, tfix, tvar, abbrevs, specgen, post, param, {}, {}]
makeCoreExprPattListFast[ {}, typ_, idx_Integer, fix_List, meta_List, tfix_Association, tvar_Association, {abbrevs_List, _Hold}, specgen_, post_, param_, out_, tsubst_] :=
	{out, idx, fix, meta, tfix, tvar, abbrevs, tsubst}
makeCoreExprPattListFast[ {expr_, rest___}, typ_, idx_Integer, fix_List, meta_List, tfix_Association, tvar_Association, abbrevs:{abb_List, i_Hold}, specgen_, post_, param_, out_, tsubst_] :=
	Replace[ makeCoreExprPatt[ replaceTypes[ expr, tsubst], typ, idx, fix, meta, tfix, tvar, abbrevs, specgen, post, param],
		{
			{{_, STERM$[ newExpr_, newTFix_, newTVar_], newIdx_, {}, {}}, ts_List, {}, newAbbrevs_} :>
				makeCoreExprPattListFast[
					{rest},
					typ /. ts,
					newIdx,
					fix /. ts,
					meta /. ts,
					Join[ tfix, newTFix],
					newTVar,
					{Join[ abb, newAbbrevs], i},
					specgen,
					post,
					param,
					Append[ replaceTypes[ out, ts], newExpr],
					Join[ tsubst /. ts, ts]
				],
			_ -> $Failed
		}
	]


(* ::Subsection:: *)
(*Selection*)

(* 'selectSubgoal[ goalPatt, assms, m][ ps, ___]' splits off the first subgoal in 'ps' whose goal matches 'goalPatt'
	and which contains all the given explicit assumptions 'assms'; in case of success, these explicit assumptions are
	immediately handled by 'introduceAssms' (which in addition drops 'm' of the remaining explicit assumptions).
	Higher-order unification is employed to find a subgoal in 'ps' that matches 'goalPatt'. Note that 'goalPatt' may
	contain arbitrary meta variables, possibly even fresh ones, and also syntactic variables. *)
selectSubgoal[ goalPatt_, assms_List] :=
	selectSubgoal[ goalPatt, assms, 0]
selectSubgoal[ LABELED$[ expr_, ___], assms_List, m_] :=
	selectSubgoal[ expr, assms, m]
selectSubgoal[ ___][ PRFSIT$[ {}, __], ___] :=
	(sowCPMMessage[ 33, "Fatal"]; $Failed)
selectSubgoal[ goalPatt_, assms_List, m_][ ps:PRFSIT$[ _, _, _, _, meta_, _, tvar_, __], ___] :=
	Replace[ makeCorePattern[ Replace[ goalPatt, PEXPR$[ args___] :> {args}], FORM$TM, ps, "gen", normalize, Max[ 0, Max @@ meta[[All, 1, 2]] + 1]],
		{
			{patts_List, _, _, {}, newMeta_, _, newTVar_Association, ts_List, {midx_, _}} :>
				Hold @@ Join[
					If[ ts === {}, {}, {INF$[ "instT", ts]}],
					If[ newTVar === tvar, {}, {INF$[ "sorts", newTVar]}],
					{selectSubgoal[ patts, replaceTypes[ assms, ts], m, 1, newMeta, midx]}
				],
			_ :> (sowCPMMessage[ 34, "Fatal"]; $Failed)
		}
	]
selectSubgoal[ goalPatts_List, assms_List, m_, n_Integer?Positive, newMeta_List, midx_Integer][ ps:PRFSIT$[ goals_List, __], args___] :=
	With[ {tail = selectSubgoal[ goalPatts, assms, m, n + 1, newMeta, midx]},
		If[ goalPatts === {} && newMeta === {},
			If[ n <= Length[ goals],
				With[ {cmd = Hold @@ Join[ If[ n === 1, {}, {INF$[ "rearrangeG", n, 1]}], {introduceAssms[ assms, m, False, All, False, True]}]},
					LL[ cmd, LL[ Hold[ tail], LL[]]]
				],
			(*else*)
				$Failed
			],
		(*else*)
			Replace[ Quiet[ Check[ goals[[n]], $Failed]],
				G:Except[ $Failed] :>
					With[ {acPatt = Max @@ Map[ assmCount[ #, 0, Infinity]&, goalPatts], acGoal = assmCount[ G, 0, Infinity]},
						If[ acGoal >= acPatt + Length[ assms],
							With[ {cmd = Hold @@ Join[ If[ n === 1, {}, {INF$[ "rearrangeG", n, 1]}], {selectFirstSubgoal[ goalPatts, assms, m, acGoal - acPatt, newMeta, midx]}]},
								LL[ cmd, LL[ Hold[ tail], LL[]]]
							],
						(*else*)
							tail[ ps, args]
						]
					]
			]
		]
	]
(* 'selectSubgoal[ i][ ps, ___]' splits off the 'i'-th subgoal of 'ps'. 'i' may be negative. *)
selectSubgoal[ i_Integer][ PRFSIT$[ G_List, __], ___] :=
	With[ {num = Length[ G]},
	With[ {j = If[ Negative[ i], num + i + 1, i]},
		Which[
			j === 1,
			If[ num === 1,
				Hold[],
			(*else*)
				Hold[ INF$[ "split"]]
			],
		(*else*)
			2 <= j <= num,
			Hold[ INF$[ "rearrangeG", j, 1], INF$[ "split"]],
		(*else*)
			True,
			sowCPMMessage[ 35, {i}, "Fatal"];
			$Failed
		]
	]]

selectFirstSubgoal[ patts_List, assms_List, m_, ac_Integer, meta1_List, midx_Integer][ PRFSIT$[ {G_, ___}, _, _, fix_, meta_, tfix_Association, tvar_Association, __], ___] :=
	Module[ {tag, aux},
	With[ {mPatt = Alternatives @@ meta[[All, 1]], m1Patt = Alternatives @@ meta1[[All, 1]]},
		Sow[ Hold[ clearIterator[ tag], Clear[ aux]], "GC"];
		aux[ arg_] := selectFirstSubgoalPost[ unifierToAbbrevs[ arg, mPatt, m1Patt, midx], assms, m, ac];
		
		With[ {goal = First[ Extract[ G, {Table[ 1, {ac}]}]]},
			lMap[ aux, unify[ Map[ {goal, #}&, patts], Join[ meta, meta1], fix, tfix, tvar, tag, "AllowFlexFlex" -> False]]
		]
	]]

selectFirstSubgoalPost[ {mInst_List, m1Inst_List, abbrevs_List, rem_List, tinst_List, tvar_Association}, assms_List, n_, ac_Integer] :=
	Hold @@ Join[
		{INF$[ "sorts", tvar]},
		If[ tinst === {}, {}, {INF$[ "instT", tinst]}],
		If[ abbrevs === {}, {}, {updateAbbrevLists[ {}, abbrevs, {}, {}]}],
		Cases[ rem, (m_ -> {t_, d_}) :> INF$[ "meta", m, t, d]],
		Cases[ mInst, (m_ -> inst_) :> INF$[ "inst", m -> STERM$[ normalize[ inst], Association[], Association[]]]],
		{introduceAssms[ replaceTypes[ assms, tinst] /. Join[ mInst, m1Inst], n, False, ac, False, True]}
	]
selectFirstSubgoalPost[ ___] := $Failed


(* ::Section:: *)
(*Obtaining Constants*)

(* 'obtainConstants[ consts, fmls][ ps, __]' sets up a new subgoal that later, in the main branch, allows one to obtain
	the given fresh a.b.f. constants and the given formulas.
	If 'consts' is empty, 'setupSubgoals' is applied; otherwise all type variables 'fmls' depends on are automatically
	specialized (this *must* happen, and in fact also happens in Isabelle). Moreover, the proof method is only
	applicable if there is exactly one subgoal. *)
obtainConstants[ {}, fmls_List] :=
	setupSubgoals[ fmls]
obtainConstants[ consts:Except[ {}, _List], fmls0_List][ ps:PRFSIT$[ {_}, _, idx0_, fix_, _, tfix_Association, tvar_Association, __], ___, mode_] :=
	Catch[
		With[ {fmls1 =
				Transpose[
					Replace[ fmls0,
						{
							LABELED$[ f_, lbl_String, data___] :> {f, {lbl, data}},
							LABELED$[ f_, _, data___] :> {f, {"?", data}},
							f_ :> {f, {"?"}}
						},
						{1}
					]
				],
				consts0 =
					With[ {aux = Function[ If[ MemberQ[ fix, #1 -> _], sowCPMMessage[ 41, "Fatal"]; Throw[ $Failed], makeSTERM[ EXPR$[ #1 -> #2, ##3]]]]},
						Replace[ consts,
							{
								EXPR$[ TYPED$[ c_FIX$, t_], sorts___] :> aux[ c, t, sorts],
								EXPR$[ c_FIX$, sorts___] :> aux[ c, TVAR$[], sorts],
								TYPED$[ c_FIX$, t_] :> aux[ c, t],
								c_FIX$ :> aux[ c, TVAR$[]],
								_ :> (sowCPMMessage[ 42, "Fatal"]; Throw[ $Failed])
							},
							{1}
						]
					]},
			If[ MatchQ[ consts0, {___, STERM$[ c_ -> _, _, _], ___, STERM$[ c_ -> _, _, _], ___}],
				sowCPMMessage[ 43, "Fatal"];
				$Failed,
			(*else*)
				Replace[ liftTypeVars[ makeSingleSTERM[ consts0], idx0],
					{STERM$[ consts1_List, constTFix_, constTVar_], idx1_Integer} :>
						With[ {tf = mergeSorts[ tfix, constTFix]},
							If[ tf === $Failed,
								$Failed,
							(*else*)
								Replace[ setupSubgoalsAux[ First[ fmls1], idx1, Join[ fix, consts1], tf, Join[ tvar, constTVar], getTermAbbrevList[ ps], "spec", mode =!= 2],
									{STERM$[ fmls_List, newTFix2_Association, newTVar_Association], newIdx_, newTFix1_Association, tsubst_List, newAbbrevs_List} :>
											(* 'newTVar' should be empty *)
										With[ {consts2 = consts1 /. tsubst},
											obtainConstantsAux[
												STERM$[ obtainConstantsAux[ consts2, fmls], Join[ newTFix1, newTFix2], newTVar],
												consts2,
												fmls,
												Last[ fmls1],
												newIdx,
												tsubst,
												newAbbrevs,
												ps,
												mode
											]
										]
								]
							]
						]
				]
			]
		]
	]

obtainConstantsAux[ consts_List, fmls_List] :=
	With[ {ot = getObjectTruthType[]},
	With[ {num = Length[ consts], judgment = If[ ot === FORM$TM, Identity, TYPED$[ TrueQ$TM, {}]]},
		TYPED$[ CoreAll$TM, {ot}][
			Lambda$TM[
				RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$thesis$TM, ot]],
				TYPED$[ CoreImplies$TM, {}][
					Fold[
						TYPED$[ CoreAll$TM, {Last[ #2]}][ Lambda$TM[ RNG$[ TYPRNG$[ #2[[1, 1]], Last[ #2]]], #1]]&,
						fromAssmsGoal[
							Map[
								ReplacePart[ #,
									Join @@ Cases[
										findAtomPositions[ #, consts[[All, 1]]],	(* MODIFICATION: There was 'Last[ findPositions[ ...]]' before. *)
										(i_Integer -> pos_List) :> Replace[ pos, (p_ -> abs_) :> (p -> VAR$[ abs + num - i]), {1}]
									]
								]&,
								fmls
							],
							judgment[ VAR$[ num]]
						],
						Reverse[ consts]
					]
				][
					judgment[ VAR$[ 0]]
				]
			]
		]
	]]
obtainConstantsAux[ STERM$[ cut0:TYPED$[ CoreAll$TM, {typ_}][ Lambda$TM[ _RNG$, cut_]], newTFix_, newTVar_], consts_, fmls_, data_, newIdx_, tsubst_, newAbbrevs_,
		ps:PRFSIT$[ G_, hyps_, idx_, fix_, meta_, tfix_, tvar_, fid_, ___], 0] :=
	If[ getObjectTruthType[] === FORM$TM,
		With[ {len = Length[ data]},
			{
				{
					Join[
						With[ {thesis = FIX$[ Theorema`Knowledge`VAR$thesis$TM, Max[ 0, Max @@ Cases[ fix, (FIX$[ Theorema`Knowledge`VAR$thesis$TM, i_] -> _) :> i] + 1]]},
							PRFSIT$[
								{mapVars[ cut, (#3 === #2)&, thesis&]},
								hyps,
								0,
								Append[ fix /. tsubst, thesis -> typ],
								{},
								Join[ tfix, newTFix],
								tvar,
								fid
							]
						],
						PRFSIT$ @@ updateAbbrevLists[ {"this" -> Null}, newAbbrevs, {}, {"calc", "cases"}, ps]
					],
					With[ {newKeys = Table[ makeFormulaKey[ fid + i], {i, 0, len - 1}]},
						Join[
							PRFSIT$[
								replaceTypes[ G, tsubst],
								Join[
									MapThread[
										Function[ {f, d, k},
											FML$[
												k,
												f,
												First[ d],
												"TypeVarPositions" -> {},
												Sequence @@ Rest[ d]
											]
										],
										{fmls, data, newKeys}
									],
									hyps
								],
								idx,
								Join[ fix /. tsubst, consts],
								meta /. tsubst,
								Join[ tfix, newTFix],
								tvar,
								fid + len
							],
							PRFSIT$ @@ updateAbbrevLists[ {"this" -> newKeys}, Join[ ellipsisAbbrev[ Last[ fmls]], newAbbrevs], {}, {"chainedFacts"}, ps]
						]
					]
				},
				{},
				{},
				Hold[]
			}
		],
	(*else*)
		(* In that case, matters are far more complicated --> use kernel mode. *)
		obtainConstantsAux[ STERM$[ cut0, newTFix, newTVar], consts, fmls, data, newIdx, tsubst, newAbbrevs, ps, 2]
	]
obtainConstantsAux[ cut_, consts_, fmls_, data_, idx_, tsubst_, abbrevs_, ps_, 1] :=
	Replace[ obtainConstantsAux[ cut, consts, fmls, data, idx, tsubst, abbrevs, ps, 0],
		{sits:{_PRFSIT$, PRFSIT$[ _, hyps_List, __]}, assms_List, thms_List, cmd_Hold} :>
			{
				makeANDNODE[ makePRFINFO[ name -> obtainConstants, used -> {}, generated -> {Take[ hyps, Length[ data]], consts}], sits],
				assms,
				thms,
				cmd
			}
	]
obtainConstantsAux[ cut_STERM$, consts_, _, data_, _, tsubst_, newAbbrevs_, ps:PRFSIT$[ {G_}, _, _, fix_, __], 2] :=
	With[ {len = Length[ data], ac = assmCount[ G, 0, Infinity]},
	With[ {assmsInGoal = If[ getObjectTruthType[] === FORM$TM, ac, 0]},
		Hold @@ Join[
			If[ tsubst === {}, {}, {INF$[ "instT", tsubst]}],
			If[ newAbbrevs === {}, {}, {updateAbbrevLists[ {}, newAbbrevs, {}, {}]}],
			{
				INF$[ "cut", cut],
				INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$thesis$TM, Max[ 0, Max @@ Cases[ fix, (FIX$[ Theorema`Knowledge`VAR$thesis$TM, i_] -> _) :> i] + 1]]],
				updateAbbrevLists[ {"this" -> Null}, {}, {}, {"calc", "cases"}],
				INF$[ "rearrangePS", 2, 1],
				INF$[ "makeExplicit", 1],
				INF$[ "forget", 1],
				instantiateCoreAllGoal[
					1,
					0,
					Hold @@ Join[
						{INF$[ "impE", 1, assmsInGoal]},
						Map[ INF$[ "allI", First[ #]]&, consts],
						Join @@ MapIndexed[
							Prepend[ If[ #1 === {"?"}, {}, {Join[ INF$[ "hypData", 1], INF$ @@ #1]}], INF$[ "makeImplicit", len + ac + 1 - (assmsInGoal + First[ #2])]]&,
							Reverse[ data]
						],
						Table[ Evaluate[ INF$[ "drop", ac + 1 - assmsInGoal]], {len}]
					],
					True
				]
			},
			{updateAbbrevLists[ {"this" -> Range[ len]}, {Hold[ ellipsisAbbrev[ len]]}, {}, {"chainedFacts"}], INF$[ "rearrangePS", 1, 2]}
		]
	]]


(* ::Section:: *)
(*Abbreviations and Definitions*)

(* ::Subsection:: *)
(*Abbreviations*)

(* In a term abbreviation 'short -> {long, tvp}', 'short' is supposed to be an untyped atom, 'long' is supposed to be a
term free of meta variables, and 'tvp' is supposed to be the list of all type variables appearing in 'term', just as in
implicit hypotheses. Type variables appearing in 'term' are treated as in implicit hypotheses, namely as being distinct
from all type variables appearing in the list of subgoals of the proof situation. Polymorphic a.b.f. constants may
still appear in 'long'; then, type inference has to instantiate the type variables by appropriate types before the term
is passed to the inference kernel. *)

(* 'addExplicitTermAbbrev[ abbrev][ ___]' adds a new term abbreviation based on 'abbrev', and replaces old ones if
	necessary. 'abbrev' must be a Core equality between a syntactic variable 'v' and an arbitrary term 't'. 't' must
	type-check and not contain any meta variables. Fresh a.b.f. type constants in 't' are generalized. *)
addExplicitTermAbbrev[ EXPR$[ (CoreEqual$TM|TYPED$[ CoreEqual$TM, _])[ v_SVAR$][ t_], sorts___]][ ps:PRFSIT$[ _, _, idx_, fix_, _, tfix_, tvar_, __], args___] :=
	Replace[ makeCoreExpr[ EXPR$[ t, sorts], Blank[], idx, fix, {}, tfix, tvar, getTermAbbrevList[ ps], "gen", normalize, None],
		{
			{{_, term_STERM$, i_, {}, {}}, {}} :>
				If[ v === SVAR$[],
					Hold[],
				(*else*)
					updateAbbrevLists[ {}, {v -> term}, {}, {}][ ps, args]
				],
			_ :> (sowCPMMessage[ 51, "Fatal"]; $Failed)
		}
	]
addExplicitTermAbbrev[ (CoreEqual$TM|TYPED$[ CoreEqual$TM, _])[ v_SVAR$][ t_]][ ps:PRFSIT$[ _, _, idx_, fix_, _, tfix_, tvar_, __], args___] :=
	Replace[ makeCoreExpr[ t, Blank[], idx, fix, {}, tfix, tvar, {}, None, Identity, Check],
		{
			{{_, term_STERM$, i_, {}, {}}, {}} :>
				If[ v === SVAR$[],
					Hold[],
				(*else*)
					updateAbbrevLists[ {}, {v -> term}, {}, {}][ ps, args]
				],
			_ :> (sowCPMMessage[ 51, "Fatal"]; $Failed)
		}
	]

ellipsisAbbrev[ STERM$[ term_, tf_Association, tv_Association]] :=
	ellipsisAbbrev[ term, tf, tv]
ellipsisAbbrev[ term_] :=
	ellipsisAbbrev[ term, Association[], Association[]]
ellipsisAbbrev[ TYPED$[ TrueQ$TM, typ_][ term_], tf_, tv_] :=
	If[ typ === {},
		ellipsisAbbrev[ term, tf, tv],
	(*else*)
		{}
	]
ellipsisAbbrev[ (Except[ VAR$|TVAR$|FIX$|TFIX$|META$|SVAR$|TYPED$|SEQ0$|SEQ1$|SEQ$|Complex|DirectedInfinity|Rational])[ term_], tf_, tv_] :=
	If[ FreeQ[ term, _META$],
		{Theorema`Knowledge`\[Ellipsis]$TM -> STERM$[ term, tf, tv]},
	(*else*)
		{}
	]
ellipsisAbbrev[ _, _, _] := {}


(* ::Subsection:: *)
(*Definitions*)

(* 'defineConst[ def][ ___]' defines a fresh a.b.f. constant 'c' to be equal to a given term 'term'; the resulting
	equality is added to the hypotheses of the current proof situation. 'c' and 'term' are given by 'def', which must
	be a meta-equality (possibly inside 'EXPR$' and/or 'LABELED$').
	In order for the proof method to succeed, 'const' must be a fresh a.b.f. constant, 'term' must be well-typed and
	free of meta variables, and there must be only one single subgoal. Note that *all* type variables 'term' depends
	on, either explicitly or implicitly via a.b.f. constants, are automatically specialized. *)
defineConst[ LABELED$[ def_, lbl_String, ___]] :=
	defineConst[ def, lbl]
defineConst[ LABELED$[ def_, ___]] :=
	defineConst[ def, "?"]
defineConst[ def_] :=
	defineConst[ def, Automatic]
defineConst[ def_, lbl_][ ps:PRFSIT$[ {G_}, _, idx_, fix_, _, tfix_Association, tvar_, __], ___] :=
	Replace[ makeCoreExprPatt[ def, FORM$TM, idx, fix, {}, tfix, tvar, getTermAbbrevList[ ps], None, normalize, None],
		{
			{{_, STERM$[ TYPED$[ CoreEqual$TM, _][ const_FIX$][ term_], newTFix_Association, newTVar_Association], _, {const_ -> typ_}, {}}, ts_List, {}, abb_} :>
				With[ {fixInTerm = Cases[ term, _FIX$, {0, Infinity}, Heads -> True]},
					If[ MemberQ[ fixInTerm, const],
						sowCPMMessage[ 56, "Fatal"];
						$Failed,
					(*else*)
						With[ {subst = typeSpecializationReplacement[ {term, Replace[ fixInTerm, fix, {1}] /. ts}, Join[ Keys[ tfix], Keys[ newTFix]]]},
							If[ subst =!= {}, sowCPMMessage[ 57, "Warning"]];
							defineConst[
								const,
								applyTypeGenSpecRepl[ {}, subst][ STERM$[ term, newTFix, newTVar]],
								typ /. subst,
								Join[ subst, ts /. subst],
								applyTypeGenSpecRepl[ {}, subst, False][ abb],
								G,
								lbl
							]
						]
					]
				],
			_ :> (sowCPMMessage[ 58, "Fatal"]; $Failed)
		}
	]
defineConst[ const:FIX$[ name_Symbol, n_Integer?NonNegative], STERM$[ term_, tf_, tv_], typ_, tsubst_, abbrev_, G_, lbl_] :=
	Hold @@ Join[
		If[ tsubst === {}, {}, {INF$[ "instT", tsubst]}],
		{
			setupSubgoalsMP[
				(* '/\R. (/\x. (x === term ==> R)) ==> R' *)
				STERM$[
					TYPED$[ CoreAll$TM, {FORM$TM}][ Lambda$TM[ RNG$[ TYPRNG$[ Null, FORM$TM]],
						TYPED$[ CoreImplies$TM, {}][
							TYPED$[ CoreAll$TM, {typ}][ Lambda$TM[ RNG$[ TYPRNG$[ Null, typ]],
								TYPED$[ CoreImplies$TM, {}][
									TYPED$[ CoreEqual$TM, {typ}][ VAR$[ 0]][ term]
								][
									VAR$[ 1]
								]
							]]
						][
							VAR$[ 0]
						]
					]],
					tf,
					tv
				],
				0
			],
			INF$[ "allI", FIX$[ Module[ {R}, R], 0]],
			INF$[ "allE", 1, STERM$[ term, tf, tv]],
			INF$[ "impE", 1, 0],
			INF$[ "refl"],
			instantiateCoreAllGoal[],
			INF$[ "impE", 1, assmCount[ G, 0, Infinity]],
			INF$[ "allI", const],
			INF$[ "makeImplicit", 1],
			INF$[ "drop", 1],
			INF$[ "hypData", 1, Replace[ lbl, {STRING$[ l_String] :> l, Except[ _String] :> tmaSymbolToString[ name] <> If[ n === 0, "", "_" <> ToString[ n]] <> " def"}]],
			updateAbbrevLists[ {"this" -> {1}}, Prepend[ abbrev, Hold[ ellipsisAbbrev[ 1]]], {}, {}]
		}
	]


(* ::Section:: *)
(*Universal Quantifiers*)

Get[ FileNameJoin[ {$TheoremaDirectory, "Theorema", "CoreProvers", "CoreProofMethods", "Universal.m"}]];


(* ::Section:: *)
(*Implications*)

Get[ FileNameJoin[ {$TheoremaDirectory, "Theorema", "CoreProvers", "CoreProofMethods", "Implication.m"}]];


(* ::Section:: *)
(*Formula Formats*)

Get[ FileNameJoin[ {$TheoremaDirectory, "Theorema", "CoreProvers", "CoreProofMethods", "Format.m"}]];


(* ::Section:: *)
(*Calculational Reasoning*)

Get[ FileNameJoin[ {$TheoremaDirectory, "Theorema", "CoreProvers", "CoreProofMethods", "Calculation.m"}]];


(* ::Section:: *)
(*Backchaining*)

Get[ FileNameJoin[ {$TheoremaDirectory, "Theorema", "CoreProvers", "CoreProofMethods", "Backchain.m"}]];


(* ::Section:: *)
(*AutoReduce*)

Get[ FileNameJoin[ {$TheoremaDirectory, "Theorema", "CoreProvers", "CoreProofMethods", "AutoReduce.m"}]];


(* ::Section:: *)
(*Rewriting*)

Get[ FileNameJoin[ {$TheoremaDirectory, "Theorema", "CoreProvers", "CoreProofMethods", "Rewrite", "Common.m"}]];


(* ::Section:: *)
(*Closing Proof Situations*)

(* 'closeByTerminal[ keep][ ps, ___]' tries to close all but the last 'keep' subgoals in 'ps' by terminal proof steps.
	If 'keep' is omitted, the first subgoal is tried to be closed, unless there are no subgoals at all. *)
closeByTerminal[][ PRFSIT$[ G_List, __], __] :=
	If[ G === {},
		Hold[],
	(*else*)
		Hold[ fastFix[ 1], closeByTerminal[ 1, False]]
	]
closeByTerminal[ keep_][ PRFSIT$[ G_List, __], __] :=
	With[ {num = Length[ G] - keep},
		If[ Positive[ num],
			Hold[ fastFix[ num], closeByTerminal[ num, False]],
		(*else*)
			Hold[]
		]
	]
closeByTerminal[ num_Integer, False][ PRFSIT$[ G_, hyps_, idx_, fix_, meta_, tfix_, tvar_, __], __] :=
	With[ {res = tryTerminalProofSteps[ Take[ G, num], hyps, idx, fix, meta, tfix, tvar]},
		If[ ListQ[ res],
			Hold @@ Join @@ (#[ 0]& /@ res),
		(*else*)
			$Failed
		]
	]

(* 'closeByRefl[ intro][ ps, ___]' tries to close the first subgoal in 'ps' by reflexivity, i.e. it tries to instantiate
	meta variables in such a way that basic inference "refl" is applicable. 'intro' is a boolean value specifying
	whether a.b.f. constants shall be introduced for universal quantifiers first.
	Note that the instantiation found (by unification) might well close the current goal, but be infeasible for further
	progress in the proof. *)
closeByRefl[] := closeByRefl[ True]
closeByRefl[ True][ ___] :=
	Hold[ introduceFix[ {RepeatedNull}], closeByRefl[ False]]
closeByRefl[ False][ PRFSIT$[ {G_, ___}, _, _, fix_, meta_, tfix_Association, tvar_Association, __], ___] :=
	Replace[ NestWhile[ First, G, MatchQ[ #, TYPED$[ CoreImplies$TM, _][ _][ _]] &],
		{
			TYPED$[ CoreEqual$TM, _][ a_][ b_] :>
				Module[ {tag, aux},
					Sow[ Hold[ clearIterator[ tag], Clear[ aux]], "GC"];
					With[ {meta0 = Alternatives @@ meta[[All, 1]]},
						aux[ u_] := closeByRefl[ u, meta0]
					];
					
					lMap[ aux, unify[ {{a, b}}, meta, fix, tfix, tvar, tag, "FlexFlexToTerms" -> False]]
				],
			_ -> $Failed
		}
	]
closeByRefl[ {ffEqns_List, rem_List, subst_List, tsubst_List, tvar_Association}, meta0_] :=
	Hold @@ Join[
		{INF$[ "sorts", tvar]},
		If[ tsubst === {}, {}, {INF$[ "instT", tsubst]}],
		Cases[ rem, (m:Except[ meta0] -> {t_, d_}) :> INF$[ "meta", m, t, d]],
		Cases[ subst, (m_ -> inst_) :> INF$[ "inst", m -> STERM$[ normalize[ inst], Association[], Association[]]]],
		Map[
			With[ {typ = flattenTypes[ First[ Replace[ #, rem]], {}]},
				INF$[ "inst", # -> STERM$[ applyCurried[ TYPED$[ Theorema`Knowledge`TYPE$TM, {Last[ typ]}], {}, Reverse[ Most[ typ]]], Association[], Association[]]]
			]&,
			Union @@ Replace[ ffEqns, {{{u_META$, _}, {v_META$, _}, _} :> {u, v}, _ -> {}}, {1}]
		],
		{INF$[ "refl"]}
	]

(* 'closeAllByRefl[ keep, keepMeta][ ps, __]' tries to close all but the last 'keep' remaining subgoals in 'ps' by
	reflexivity. To that end, it first introduces a.b.f. constants for all universal quantifiers and checks whether the
	atomic subgoals are indeed Core equalities. If so, general higher-order unification is employed to solve all these
	equations *simultaneously*, possibly instantiating any but the first 'keepMeta' meta variables in the list. All
	flex-flex equations are immediately solved by constant instantiations. 'closeAllByRefl' is in general more powerful
	than mapping 'closeByRefl' over all subgoals, since it solves the equations simultaneously.
	Note that the proof method does not backtrack over unifiers, but commits itself to the first one. *)
closeAllByRefl[] := closeAllByRefl[ 0, 0]
closeAllByRefl[ keep_] := closeAllByRefl[ keep, 0]
closeAllByRefl[ keep_Integer, keepMeta_Integer][ ps:PRFSIT$[ G_, _, _, fix_, __], args__] :=
	With[ {num = Length[ G] - keep},
		If[ Positive[ num],
			Catch[
				Module[ {allc = 0},
				With[ {data =
						Map[
							Replace[ extractAtomicGoal[ #, 0, 0, Infinity, Infinity],
								{
									{TYPED$[ CoreEqual$TM, {_}][ _][ _], assms_, all_} :> {assms, allc += all; all},
									_ :> Throw[ $Failed]
								}
							]&,
							Take[ G, num]
						]},
				With[ {cmd = closeAllByRefl[ data[[All, 1]], num, keepMeta]},
					If[ allc === 0,
						cmd[ ps, args],
					(*else*)
						Append[ fastFix[ data[[All, 2]], num, fix], cmd]
					]
				]]]
			],
		(*else*)
			Hold[]
		]
	]
closeAllByRefl[ ac_List, close_, keepMeta_][ PRFSIT$[ G_, _, _, fix_, meta_, tfix_Association, tvar_Association, __], __] :=
	Replace[
		unifyFirst[
			MapThread[ With[ {atom = dropAssms[ #1, #2]}, {atom[[0, 1]], atom[[1]]}]&, {Take[ G, close], ac}],
			If[ keepMeta === 0, meta, ReplacePart[ meta, Table[ {k, 2, 2} -> False, {k, 1, keepMeta}]]],
			fix,
			tfix,
			tvar,
			Automatic,
			"FlexFlexToTerms" -> False
		],
		{
			{ffEqns_List, rem_List, subst_List, tsubst_List, tv_Association} :>
				With[ {meta0 = Alternatives @@ meta[[All, 1]]},
					Hold @@ Join[
						{INF$[ "sorts", tv]},
						If[ tsubst === {}, {}, {INF$[ "instT", tsubst]}],
						Cases[ rem, (m:Except[ meta0] -> {t_, d_}) :> INF$[ "meta", m, t, d]],
						Cases[ subst, (m_ -> inst_) :> INF$[ "inst", m -> STERM$[ normalize[ inst], Association[], Association[]]]],
						Map[
							With[ {typ = flattenTypes[ First[ Replace[ #, rem]], {}]},
								INF$[ "inst", # -> STERM$[ applyCurried[ TYPED$[ Theorema`Knowledge`TYPE$TM, {Last[ typ]}], {}, Reverse[ Most[ typ]]], Association[], Association[]]]
							]&,
							Union @@ Replace[ ffEqns, {{{u_META$, _}, {v_META$, _}, _} :> {u, v}, _ -> {}}, {1}]
						],
						Table[ INF$[ "refl"], {close}]
					]
				],
			_ -> $Failed
		}
	]

(* 'closeByAssm[ keepMeta][ ps, ___]' tries to close the first subgoal in 'ps' by backchaining w.r.t. one of its
	explicit assumptions. The atomic conclusion of the goal, as well as the assumptions, may of course involve meta
	variables, which might be instantiated -- as long as they are not among the first 'keepMeta' meta variables in the
	list. Furthermore, the assumptions may be universally quantified, in which case the universal quantifiers are
	automatically instantiated by fresh meta variables. A.b.f. constants are introduced for all universal quantifiers
	in the goal. *)
closeByAssm[] := closeByAssm[ 0]
closeByAssm[ keepMeta_][ ps:PRFSIT$[ {G_, ___}, _, _, fix_, meta_, __], args___] :=
	With[ {maxMeta = Max[ -1, Max @@ meta[[All, 1, 2]]]},
		Replace[ aaCount[ G, 0, 0, Infinity, Infinity],
			{
				{assms_, 0} :> closeByAssm[ 1, assms, maxMeta, keepMeta][ ps, args],
				{assms_, all_} :> Append[ fastFix[ {all}, 1, fix], closeByAssm[ 1, assms, maxMeta, keepMeta]]
			}
		]
	]
closeByAssm[ i_Integer, ac_Integer, max_Integer, keepMeta_][ ps:PRFSIT$[ {G_, ___}, _, idx_, fix_, __], args__] :=
	If[ i <= ac,
		With[ {rule = Extract[ G, assmPosition[ i]],
				tail = closeByAssm[ i + 1, ac, max, keepMeta]},
			Replace[ aaCount[ rule, 0, 0, Infinity, Infinity],
				{assms_, all_} :>
					If[ ac >= assms + i,
						Module[ {tag, aux},
							Sow[ Hold[ clearIterator[ tag], Clear[ aux]], "GC"];
							With[ {numFix = Length[ fix]},
							With[ {cmd =
									Hold @@ Join[
										Table[ INF$[ "meta", META$[ Null, k + max + 1], TVAR$[ k + idx], numFix], {k, 0, all - 1}],
										Table[ INF$[ "allE", i, makeSTERM[ META$[ Null, k]]], {k, max + 1, all + max}],
										{backChainSingle[ i, 1, assms, assms, keepMeta, True, {dropAssms[ G, ac - assms], ac - assms, {}}, tag, aux, Hold[]]}
									]},
								LL[ cmd, LL[ Hold[ tail], LL[]]]
							]]
						],
					(*else*)
						tail[ ps, args]
					]
			]
		],
	(*else*)
		$Failed
	]

(* 'closeByFact[ facts][ ps, __]' tries to close the first subgoal in 'ps' by backchaining it w.r.t. a known fact in
	the implicit hypotheses, or w.r.t. any of the given 'facts' (depending on whether 'facts' is empty or not).
	The proof method works similar as 'closeByAssm', i.e. it introduces a.b.f. constants for all remaining quantifiers
	in the goal and it allows flex-flex equations to remain after backcahining, which may then be closed by something
	like 'closeAllByRefl'.
	Note again that, if 'facts' is empty, only the implicit hypotheses are searched, not the whole environment!
	Elements of 'facts' may still refer to arbitrary knowledge, of course. *)
closeByFact[ facts_List][ ps:PRFSIT$[ {G_, ___}, hyps_, _, fix_, meta_, __], args__] :=
	If[ facts === {},
		With[ {maxMeta = Max[ -1, Max @@ meta[[All, 1, 2]]], num = Length[ hyps]},
			Replace[ aaCount[ G, 0, 0, Infinity, Infinity],
				{
					{assms_, 0} :> closeByFact[ 1, num, assms, maxMeta][ ps, args],
					{assms_, all_} :> Append[ fastFix[ {all}, 1, fix], closeByFact[ 1, num, assms, maxMeta]]
				}
			]
		],
	(*else*)
		Hold[ addKnowledgeImplicit[ facts, "methodFacts", False, 3, True, None], closeByFact[]]
	]
closeByFact[][ ps:PRFSIT$[ {G_, ___}, _, _, fix_, meta_, __], args__] :=
	With[ {maxMeta = Max[ -1, Max @@ meta[[All, 1, 2]]],
			num = -Length[ getProofSitData[ ps, "methodFacts", {}]]},
		Replace[ aaCount[ G, 0, 0, Infinity, Infinity],
			{
				{assms_, 0} :> closeByFact[ num, assms, maxMeta][ ps, args],
				{assms_, all_} :> Append[ fastFix[ {all}, 1, fix], closeByFact[ num, assms, maxMeta]]
			}
		]
	]
(*"methodFacts"*)
closeByFact[ i_Integer, ac_, max_][ ps:PRFSIT$[ _, hyps_, __], args__] :=
	If[ Negative[ i],
		With[ {k = Replace[ getProofSitData[ ps, "methodFacts", {}][[i]], {ky:{_, _}, _} :> ky]},
			closeByFactAux[ Position[ hyps, FML$[ k, __], {1}, 1][[1, 1]], ac, max, closeByFact[ i + 1, ac, max]][ ps, args]
		],
	(*else*)
		$Failed
	]
(*all knowledge*)
closeByFact[ i_Integer, hypNum_, ac_, max_][ args__] :=
	If[ i <= hypNum,
		closeByFactAux[ i, ac, max, closeByFact[ i + 1, hypNum, ac, max]][ args],
	(*else*)
		(* one could also try global knowledge ... *)
		$Failed
	]

closeByFactAux[ i_, ac_, max_, tail_][ ps:PRFSIT$[ {G_, ___}, hyps_, idx_, fix_, __], args__] :=
	With[ {fact = hyps[[i]]},
		Replace[ aaCount[ formula@fact, 0, 0, Infinity, Infinity],
			{assms_, all_} :>
				If[ ac >= assms,
					Module[ {tag, aux},
						Sow[ Hold[ clearIterator[ tag], Clear[ aux]], "GC"];
						With[ {numFix = Length[ fix], numTVars = Length[ getOptionalComponent[ fact, "TypeVarPositions"]]},
						With[ {cmd =
								Hold @@ Join[
									{INF$[ "makeExplicit", i]},
									Table[ INF$[ "meta", META$[ Null, k + max + 1], TVAR$[ k + idx + numTVars], numFix], {k, 0, all - 1}],
									Table[ INF$[ "allE", 1, makeSTERM[ META$[ Null, k]]], {k, max + 1, all + max}],
									{backChainSingle[ 1, 1, assms, ac, 0, True, {G, 1, {}}, tag, aux, Hold[]]}
								]},
							LL[ cmd, LL[ Hold[ tail], LL[]]]
						]]
					],
				(*else*)
					tail[ ps, args]
				]
		]
	]

(* 'closeSubgoals[ keep, keepMeta, tryTerminal][ ps, __]' attempts to close all but the last 'keep' subgoals in 'ps' by
	assumption and reflexivity, possibly instantiating any but the first 'keepMeta' meta variables.
	Note that this proof method does not close the whole proof situation, even if 'keep' is '0'!
	'tryTerminal' indicates whether terminal proof methods shall also be tried, *before* trying assumptions and
	reflexivity. *)
closeSubgoals[ keep_, keepMeta_] :=
	closeSubgoals[ keep, keepMeta, False]
closeSubgoals[ keep_Integer, keepMeta_Integer, tryTerminal_][ PRFSIT$[ G_, __], __] :=
	With[ {num = Length[ G] - keep},
		If[ Positive[ num],
			Hold[
				fastFix[ num],
				Evaluate[ mapAllTac[ If[ TrueQ[ tryTerminal], orTac[ closeByTerminal[ 1, False], closeByAssm[ keepMeta], voidTac[]], tryTac[ closeByAssm[ keepMeta]]], num]],
				closeAllByRefl[ keep, keepMeta]
			],
		(*else*)
			Hold[]
		]
	]

(* 'closeProofSit[ m][ ps, ___]' tries to close the current proof situation by applying the given proof method 'm' to
	all subgoals. The proof method fails if the proof situation as a whole cannot be closed (even if some subgoals can
	be closed). *)
closeProofSit[][ __] :=
	Hold[ fastFix[ All], mapAllTac[ tryTac[ closeByAssm[]], All], closeAllByRefl[], INF$[ "close"]]
closeProofSit[ m_andTac][ __] :=
	Join[ Hold @@ m, Hold[ closeAllByRefl[], INF$[ "close"]]]
closeProofSit[ m_][ __] :=
	Hold[ m, closeAllByRefl[], INF$[ "close"]]

End[]

EndPackage[]
