(* Core proof methods for backchaining. *)

(* ::Subsection:: *)
(*Top-Level Proof Methods*)

(* ::Subsubsection:: *)
(*this*)

coreThis[][ ps:PRFSIT$[ {G_, ___}, _, _, fix_, __], args__] :=
	With[ {ac1 = assmCount[ G, 0, Infinity]},
		Replace[ introFixCommandList[ dropAssms[ G, ac1], fix, False],
			{ac2_Integer, introFix_Hold} :>
				Join[
					introFix,
					Hold @@ {updateAbbrevLists[ {}, {}, {"methodFacts" -> getProofSitData[ ps, "chainedFacts", {}]}, {}], coreThis[ ac1, ac2, Length[ introFix]]}
				]
		]
	]
coreThis[ ac1_Integer, ac2_Integer, gen_Integer][ ps:PRFSIT$[ {G_, R___}, __], __] :=
	With[ {num = -Length[ getProofSitData[ ps, "methodFacts", {}]]},
		If[ num === 0,
			sowCPMMessage[ 1009, "Fatal", "this"];
			$Failed,
		(*else*)
			Module[ {data = {1, 1, 0, 0, True, True, {dropAssms[ G, ac1 + ac2], ac1 + ac2 + 1, {}}, Hold @@ {coreRulePost[ True, ac1, gen, Length[ {R}]]}}},
					(* these are precisely the arguments of 'backChainSingle'; 'elim + ac1 + ac2 + 1' because rules are made explicit *)
				Hold[ coreRuleAux[ data, num]]		(* 'num' is non-positive *)
			]
		]
	]


(* ::Subsubsection:: *)
(*rule*)

(* 'coreRule[ rules]': operationally equivalent to backchaining w.r.t. 'rule[OF chainedFacts]'. Goal is not considered
	atomic. The behavior in Isabelle seems to be slightly different, but I don't know how to explain it. *)
coreRule[ rules_List][ ps:PRFSIT$[ {G_, ___}, _, _, fix_, __], args__] :=
	With[ {cfc = Length[ getProofSitData[ ps, "chainedFacts", {}]], ac1 = assmCount[ G, 0, Infinity]},
		Replace[ introFixCommandList[ dropAssms[ G, ac1], fix, False],
			{ac2_Integer, introFix_Hold} :>
				With[ {base = If[ rules === {}, coreRule[ cfc, ac1, ac2, Length[ introFix], All], coreRule[ cfc, ac1, ac2, Length[ introFix]]]},
				With[ {cmd = If[ cfc === 0, base, internalFactHandler[ base, First, 0]]},
					Join[
						introFix,
						Hold @@ If[ rules === {}, {cmd}, {addKnowledgeImplicit[ rules, "methodFacts", False, 3, True, If[ cfc === 0, "intro", "elim"]], cmd}]
					]
				]]
		]
	]
(*"methodFacts"*)
coreRule[ elim_Integer, ac1_Integer, ac2_Integer, gen_Integer][ ps:PRFSIT$[ {G_, R___}, __], __] :=
	(* eliminate first 'elim' assumptions, trying all rules in "methodFacts" *)
	(* 'G' has 'elim + ac1 + ac2' many explicit assumptions *)
	With[ {num = -Length[ getProofSitData[ ps, "methodFacts", {}]]},
		If[ num === 0,
			sowCPMMessage[ 1010, "Fatal", "rule"];
			$Failed,
		(*else*)
			Module[ {data = {1, elim + 1, 0, 0, True, True, {dropAssms[ G, elim + ac1 + ac2], elim + ac1 + ac2 + 1, {}}, Hold @@ {coreRulePost[ True, ac1, gen, Length[ {R}]]}}},
					(* these are precisely the arguments of 'backChainSingle'; 'elim + ac1 + ac2 + 1' because rules are made explicit *)
				Hold @@ Append[ Table[ instantiateCoreAll[ {Automatic...}, k], {k, 1, elim}], coreRuleAux[ data, num]]	(* 'num' is non-positive *)
			]
		]
	]
(*all rules*)
coreRule[ elim_Integer, ac1_Integer, ac2_Integer, gen_Integer, All][ ps:PRFSIT$[ {G_, R___}, hyps_, __], __] :=
	(* eliminate first 'elim' assumptions, trying all suitable rules in the global knowledge base *)
	Module[ {data = {1, elim + 1, 0, 0, True, True, {dropAssms[ G, elim + ac1 + ac2], elim + ac1 + ac2 + 1, {}}, Hold @@ {coreRulePost[ False, ac1, gen, Length[ {R}]]}}},
			(* these are precisely the arguments of 'backChainSingle'; 'elim + ac1 + ac2 + 1' because rules are made explicit *)
	With[ {hypNum = Length[ hyps]},
		Hold @@ Append[ Table[ instantiateCoreAll[ {Automatic...}, k], {k, 1, elim}], coreRuleAux[ data, 1, hypNum]]
	]]
	
SetAttributes[ coreRuleAux, HoldAll];
(*"methodFacts"*)
coreRuleAux[ data_Symbol, i_Integer][ ps:PRFSIT$[ _, hyps_List, __], __] :=
	If[ Negative[ i],
		With[ {k = Replace[ getProofSitData[ ps, "methodFacts", {}][[i]], {ky:{_, _}, _} :> ky]},
		With[ {j = Position[ hyps, FML$[ k, __], {1}, 1][[1, 1]], ii = i + 1, cmd = backChainSingle @@ data},
			LL[ Hold[ INF$[ "makeExplicit", j], cmd], LL[ Hold[ coreRuleAux[ data, ii]], LL[]]]
		]],
	(*else*)
		$Failed
	]
(*implicit hypotheses*)
coreRuleAux[ data_Symbol, i_Integer, hypNum_Integer][ ps:PRFSIT$[ {G_, ___}, hyps_, __], args__] :=
	With[ {isElim = TrueQ[ data[[2]] > 1]},
	With[ {attrib = If[ isElim, "elim", "intro"]},
		Replace[ Position[ Drop[ hyps, i - 1], _?(MemberQ[ getOptionalComponent[ #, "tags"], attrib]&), {1}, 1, Heads -> False] + i - 1,
			{
				{{j_Integer}} :>
					With[ {ii = j + 1, cmd = backChainSingle @@ data},
						LL[ Hold[ INF$[ "makeExplicit", j], cmd], LL[ Hold[ coreRuleAux[ data, ii, hypNum]], LL[]]]
					],
				_ :>
					Module[ {it = getRuleConsts[ G, attrib, False]},
						coreRuleAux[ data, isElim, it][ ps, args]
					]
			}
		]
	]]
(*formula environment & theorem instances*)
coreRuleAux[ data_Symbol, isElim_, it_Symbol][ ps:PRFSIT$[ {G_, ___}, hyps_, __], args__] :=
	Replace[
		With[ {coll = If[ isElim, extendRULEC$[ $elimRules], extendRULEC$[ $introRules]]},
			(* iterate over 'coll' until it is exhausted or a formula not contained in the hypotheses is reached *)
			NestWhile[
				iterateRULEC$[ it, coll]&,
				Null,
				(# === Null || (ListQ[ #] && MemberQ[ hyps, FML$[ #, __]]))&
			]
		],
		{
			k_List :>
				With[ {cmd = backChainSingle @@ data},
					LL[ Hold[ INF$[ "formula", k], cmd], LL[ Hold[ coreRuleAux[ data, isElim, it]], LL[]]]
				],
			thm_THM$ :>
				With[ {cmd = backChainSingle @@ data},
					LL[ Hold[ INF$[ "theorem", thm], cmd], LL[ Hold[ coreRuleAux[ data, isElim, it]], LL[]]]
				],
			_ :>
				If[ isElim,
					(* After elimination rules, we also try introduction rules. This solves goals like 'from 'A' and 'B' have 'A & B' ..'. *)
					it = getRuleConsts[ G, "intro", False];
					coreRuleAux[ data, False, it][ ps, args],
				(*else*)
					$Failed
				]
		}
	]

coreRulePost[ remove_, ac_Integer, gen_Integer, num_Integer][ ps:PRFSIT$[ G_List, _, _, fix_, __], args__] :=
	With[ {cmd =
			If[ Positive[ gen] && TrueQ[ $generalizeFix],
				With[ {goalNum = Length[ G], gfix = Take[ fix, -gen][[All, 1]]},
					Append[ mapAllInline[ Hold @@ {generalizeFix[ gfix, ac]}, goalNum - num, goalNum], INF$[ "forgetFix", gfix]]
				],
			(*else*)
				Hold[]
			]},
		If[ TrueQ[ remove],
			internalFactHandler[ "methodFacts", andTac @@ cmd, None, 2][ ps, args],
		(*else*)
			cmd
		]
	]


(* ::Subsubsection:: *)
(*intro*)

(* 'coreIntro[ rules, rep]': make chained facts explicit and apply given rules (repeatedly) to the atomic goal without
	eliminating any explicit assumptions. Hence, making facts explicit only ensures that they appear in the subgoals
	emerging from the application of the proof method. 'rep' specifies whether rules shall be applied repeatedly.
	Note that "intro r1 r2" is *not* the same as "intro r1, intro r2", even if indeed first 'r1' is applied and then
	'r2', since the latter makes facts explicit twice (perhaps even in the same subgoal); this is the same in Isabelle. *)
coreIntro[ rules_List] :=
	coreIntro[ rules, True]
coreIntro[ rules_List, rep_][ ps:PRFSIT$[ {G_, ___}, _, _, fix_, __], __] :=
	With[ {cfc = Length[ getProofSitData[ ps, "chainedFacts", {}]], ac1 = assmCount[ G, 0, Infinity]},
		Replace[ introFixCommandList[ dropAssms[ G, ac1], fix, False],
			{ac2_Integer, introFix_Hold} :>
				With[ {base = coreIntro[ cfc + ac1, ac2, Length[ introFix], rep]},
				With[ {cmd = If[ cfc === 0, base, internalFactHandler[ base, First, 0]]},
					Join[
						introFix,
						Hold @@ {
							addKnowledgeImplicit[ rules, "methodFacts", False, 3, True, "intro"],
							cmd,
							internalFactHandler[ "methodFacts", If[ TrueQ[ $generalizeFix && rep], forgetConsts[ Length[ fix]], voidTac[]], None, 2]
						}
					]
				]]
		]
	]
coreIntro[ ac1_Integer, ac2_Integer, numIntro_Integer, rep_][ ps:PRFSIT$[ {G_, ___}, __], args__] :=
	(* try all rules in "methodFacts" *)
	(* 'G' has 'ac1 + ac2' many explicit assumptions *)
	With[ {num = -Length[ getProofSitData[ ps, "methodFacts", {}]]},
		If[ num === 0,
			sowCPMMessage[ 1010, "Fatal", "intro"];
			$Failed,
		(*else*)
			Module[ {data =
						{
							1, 1, 0, 0, True, True,
							{dropAssms[ G, ac1 + ac2], ac1 + ac2 + 1, {}},
							{num, If[ numIntro === 0, {}, {{ac1, numIntro}}]}
						}
					},
				coreIntroAux[ data, num, If[ TrueQ[ rep], Repeated, False]][ ps, args]
			]
		]
	]

(* 'coreIntroAux' is an auxiliary proof method. There are 4 versions of it, with one argument, two arguments, three
	arguments, and four arguments, respectively:
	- The 3-argument version is called by 'coreIntro'. It iterates over the given rules and, as soon as the rules are
		exhausted, generalizes all a.b.f. constants that have been introduced in the course of the proof method. In
		fact, the *last* recursive application takes care of *all* constants.
	- The 2-argument version is passed as parameter 'post' to 'backChainSingle', unless the last argument of the
		3-argument version ('applied') is 'False'; it basically sets up the recursive application of this proof method
		by mapping the 1-argument version of 'coreIntroAux' over all newly emerged subgoals.
	- The 1-argument version is called by the 2-argument version as sketched above. It simply introduces fresh a.b.f.
		constants for all new universal quantifiers, similar as 'coreIntro', and then applies the 4-argument version of
		'coreIntroAux'.
	- The 4-argument version, finally, creates a new 'data' variable with updated 'goal', 'dropGoal' and 'gen' entries
		and then directly applies the 3-argument version again. *)
SetAttributes[ coreIntroAux, HoldFirst];
coreIntroAux[ data_Symbol, origGoals_Integer][ args__] :=
	mapAllTac[ coreIntroAux[ data], -origGoals][ args]	(* 'origGoals' includes the subgoal the proof method was applied to *)
coreIntroAux[ data_Symbol][ PRFSIT$[ {G_, ___}, _, _, fix_, __], __] :=
	With[ {ac1 = assmCount[ G, 0, Infinity]},
		Replace[ introFixCommandList[ dropAssms[ G, ac1], fix, False],
			{ac2_Integer, introFix_Hold} :>
				Append[
					introFix,
					coreIntroAux[ data, ac1, ac2, Length[ introFix]]
				]
		]
	]
coreIntroAux[ data_Symbol, ac1_, ac2_, numIntro_][ ps:PRFSIT$[ {G_, ___}, __], args__] :=
	Replace[ Last[ data],
		{num_Integer, gen_List} :>
			Module[ {dataNew =
						{
							1, 1, 0, 0, True, True,
							{dropAssms[ G, ac1 + ac2], ac1 + ac2 + 1, {}},
							{num, Join[ If[ numIntro === 0, {}, {{ac1, numIntro}}], gen]}
						}
					},
				coreIntroAux[ dataNew, num, True]
			][ ps, args]
	]
coreIntroAux[ data_Symbol, i_Integer, applied_][ ps:PRFSIT$[ G_List, hyps_List, _, fix_, __], __] :=
	Which[
		Negative[ i],
		With[ {k = Replace[ getProofSitData[ ps, "methodFacts", {}][[i]], {ky:{_, _}, _} :> ky],
				numGoals = Length[ G]},
		With[ {j = Position[ hyps, FML$[ k, __], {1}, 1][[1, 1]], ii = i + 1,
				cmd = backChainSingle @@ ReplacePart[ data, -1 -> If[ applied === False, Hold[], Hold[ coreIntroAux[ data, numGoals]]]]},
			LL[ Hold[ INF$[ "makeExplicit", j], cmd], LL[ Hold[ coreIntroAux[ data, ii, applied]], LL[]]]
		]],
	(*else*)
		TrueQ[ applied],
		If[ TrueQ[ $generalizeFix],
			(* post-processing *)
			(* no need to map over subgoals, because 'coreIntroAux' is mapped over new subgoals anyway *)
			Module[ {k = Length[ fix]},
				Hold @@ Replace[
					SplitBy[ data[[-1, -1]], First],
					all:{{ac_, _}, ___} :>
						With[ {n = Plus @@ all[[All, 2]]},
							generalizeFix[ Take[ fix, {k + 1 - n, k}][[All, 1]], k -= n; ac]
						],
					{1}
				]
			],
		(*else*)
			Hold[]
		],
	(*else*)
		True,
		$Failed
	]


(* ::Subsubsection:: *)
(*elim*)

(* 'coreElim[ rules, rep]': like "intro", but in each rule-application eliminate *one* explicit assumption (any one,
	not necessarily the first one and not necessarily using the first rule provided). Fail if this is not possible.
	Operational behavior: First iterate through given rules, and for each rule iterate through assumptions. *)
coreElim[ rules_List] :=
	coreElim[ rules, True]
coreElim[ rules_List, rep_][ ps_PRFSIT$, __] :=
	coreElimInit[ rules, rep, ps, "elim"]
coreElim[ ac_Integer, gen_Integer, rep_, src_String][ ps:PRFSIT$[ {G_, ___}, __], args__] :=
	With[ {num = -Length[ getProofSitData[ ps, "methodFacts", {}]]},
		If[ num === 0,
			sowCPMMessage[ 1010, "Fatal", src];
			$Failed,
		(*else*)
			Module[ {data = {1, 0, 0, 0, True, True, {dropAssms[ G, ac], ac + 1, {}}, {num, ac, gen}}},
				coreElimAux[ data, num, If[ TrueQ[ rep], Repeated, False]][ ps, args]
			]
		]
	]

SetAttributes[ {coreElimAux, coreElimAux2}, HoldFirst];
coreElimAux[ data_Symbol, origGoals_Integer][ args__] :=
	mapAllTac[ coreElimAux[ data], -origGoals][ args]	(* 'origGoals' includes the subgoal the proof method was applied to *)
coreElimAux[ data_Symbol][ PRFSIT$[ {G_, ___}, _, _, fix_, __], __] :=
	With[ {ac1 = assmCount[ G, 0, Infinity]},
		Replace[ introFixCommandList[ dropAssms[ G, ac1], fix, False],
			{ac2_Integer, introFix_Hold} :>
				Append[
					introFix,
					coreElimAux[ data, ac1 + ac2, Length[ introFix]]
				]
		]
	]
coreElimAux[ data_Symbol, ac_, numIntro_Integer][ ps:PRFSIT$[ {G_, ___}, __], args__] :=
	Replace[ Last[ data],
		{num_Integer, _, gen_Integer} :>
			Module[ {dataNew = {1, 0, 0, 0, True, True, {dropAssms[ G, ac], ac + 1, {}}, {num, ac, gen + numIntro}}},
				coreElimAux[ dataNew, num, True]
			][ ps, args]
	]
coreElimAux[ data_Symbol, i_Integer, applied_][ ps:PRFSIT$[ _, hyps_List, _, fix_, __], __] :=
	Which[
		Negative[ i],
		With[ {k = Replace[ getProofSitData[ ps, "methodFacts", {}][[i]], {ky:{_, _}, _} :> ky]},
		With[ {j = Position[ hyps, FML$[ k, __], {1}, 1][[1, 1]], ii = i + 1},
			LL[ Hold[ INF$[ "makeExplicit", j], coreElimAux2[ data, 1, applied]], LL[ Hold[ coreElimAux[ data, ii, applied]], LL[]]]
		]],
	(*else*)
		TrueQ[ applied],
		(* post-processing *)
		With[ {gen = data[[-1, -1]]},
			If[ gen =!= 0 && TrueQ[ $generalizeFix],
				Hold @@ {generalizeFix[ Take[ fix, -gen][[All, 1]], 0]},
			(*else*)
				Hold[]
			]
		],
	(*else*)
		True,
		$Failed
	]

coreElimAux2[ data_Symbol, e_Integer, applied_][ PRFSIT$[ G_, __], __] :=
	If[ e <= data[[-1, 2]],
		With[ {numGoals = Length[ G]},
		With[ {cmd = backChainSingle @@ ReplacePart[ data, {2 -> {e + 1}, -1 -> If[ applied === False, Hold[], Hold[ coreElimAux[ data, numGoals]]]}],
				ee = e + 1},
			LL[ Hold[ cmd], LL[ Hold[ coreElimAux2[ data, ee, applied]], LL[]]]
		]],
	(*else*)
		$Failed
	]

coreElimInit[ rules_List, rep_, ps:PRFSIT$[ {G_, ___}, _, _, fix_, __], src_String] :=
	With[ {cfc = Length[ getProofSitData[ ps, "chainedFacts", {}]], ac1 = assmCount[ G, 0, Infinity]},
		Replace[ introFixCommandList[ dropAssms[ G, ac1], fix, False],
			{ac2_Integer, introFix_Hold} :>
				With[ {ac = cfc + ac1 + ac2},
					If[ ac === 0,
						sowCPMMessage[ 141, "Fatal", src];
						$Failed,
					(*else*)
						With[ {base = coreElim[ ac, Length[ introFix], rep, src]},
						With[ {cmd = If[ cfc === 0, base, internalFactHandler[ base, First, 0]]},
							Join[
								introFix,
								Hold @@ {
									addKnowledgeImplicit[ rules, "methodFacts", False, 3, True, "elim"],
									cmd,
									internalFactHandler[ "methodFacts", If[ TrueQ[ $generalizeFix && rep], forgetConsts[ Length[ fix]], voidTac[]], None, 2]
								}
							]
						]]
					]
				]
		]
	]


(* ::Subsubsection:: *)
(*dest*)

(* 'coreDest[ rules, rep]': like "elim", but first convert given rules into elimination format with parameters
	'n = -2', 'dup = 0' and 'force = True'. *)
coreElim[ rules_List] :=
	coreElim[ rules, True]
coreDest[ rules_List, rep_][ ps_PRFSIT$, __] :=
	coreElimInit[ VARIANT$[ #, eliminationVariant, {True, -2, 0}, "elim"]& /@ rules, rep, ps, "dest"]


(* ::Subsection:: *)
(*Low-Level Proof Methods*)

(* OBSOLETE, kept for testing only *)
(* 'backChain[ elim, explicit, implicit, env, atomR, atomG, instMeta][ ps, _, mode]' backchains the first subgoal in
	'ps' w.r.t. a suitable rule 'r'. 'r' is searched for in the explicit assumptions of the subgoal, in the implicit
	assumptions of 'ps', and even in the whole Theorema environment, depending on the parameters 'explicit', 'implicit'
	and 'env':
	- 'explicit' is either 'All' (try all possible explicit assumptions) or an integer (try all explicit assumptions
		until 'explicit') or a list of integers (try precisely the specified explicit assumptions; could be empty),
	- 'implicit' is 'All' or an integer, with the same meaning as 'explicit', or a list of formula-identifiers (labels,
		keys, indices) with the same meaning as in 'makeExplicitAssms',
	- 'env' is 'All' or a list of formula identifiers.
	In any case: specifications referring to non-existing formulas are simply *ignored*; they do not trigger any
	warning!
	'elim' specifies which explicit assumption to eliminate.
	'atomR' and 'atomG' specify how many premises of the rule, and assumptions of the goal, shall be treated as being
	part of the rule's atomic conclusion and subgoal's atomic goal, respectively.
	'instMeta', finally, specifies whether the proof method is allowed to instantiate term meta variables or not; it is
	'True' by default. The instantiation of type variables is guided by the global parameter '$instantiateTypes'. *)
backChain[ elim_, explicit_, implicit_, env_] :=
	backChain[ elim, explicit, implicit, env, 0, 0, True]
backChain[ elim_, explicit_, implicit_, env_, atomR_] :=
	backChain[ elim, explicit, implicit, env, atomR, 0, True]
backChain[ elim_, explicit_, implicit_, env_, atomR_, atomG_] :=
	backChain[ elim, explicit, implicit, env, atomR, atomG, True]
backChain[ elim_, {}, rest___] := backChain[ elim, 0, rest]
backChain[ elim_, 0, {}, rest___] := backChain[ elim, 0, 0, rest]
backChain[ elim_, 0, 0, {}, rest___] := backChain[ elim, 0, 0, 0, rest]
backChain[ elim0_, explicit_, implicit_, env_, atomR_, atomG0_, instMeta_][ ps:PRFSIT$[ {G_, ___}, __], args___, mode_] :=
	With[ {acGoal = assmCount[ G, 0, Infinity]},
	With[ {atomG =
			Switch[ atomG0,
				All|"all",
				acGoal,
				_Integer?NonNegative,
				If[ acGoal >= atomG0, atomG0, $Failed],
				_,
				$Failed
			]},
		If[ IntegerQ[ atomG],
			With[ {drop = acGoal - atomG,
					elim =
						Switch[ elim0,
							{},
							0,
						(*else*)
							_List,
							DeleteDuplicates[ Replace[ elim0, e_Integer?Negative :> (acGoal + 1 + e), {1}]],
						(*else*)
							_?Negative,
							acGoal + 1 + elim0,
						(*else*)
							_,
							elim0
						]},
				If[ NonNegative[ Min[ elim]] && Max[ elim] <= acGoal,
					With[ {goal = dropAssms[ G, drop]},
						Module[ {data = {1, elim + 1, atomR, atomG, instMeta, True, {goal, drop + 1, {}}, Hold[]}},
							(* these are precisely the arguments of 'backChainSingle'; 'elim + 1' and 'drop + 1' because rules are added *)
							backChainAux[
								data,
								Switch[ explicit,
									All|"all",
									drop,
									_Integer,
									Min[ explicit, drop],
									_List,
									DeleteCases[ explicit, _Integer?(# > drop &)],
									_,
									0
								],
								implicit,
								env
							][ ps, args, mode]
						]
					],
				(*else*)
					sowCPMMessage[ 142, {elim0}, "Fatal", "backChain"];
					$Failed
				]
			],
		(*else*)
			sowCPMMessage[ 1011, {atomG0}, "Fatal", "backChain"];
			$Failed
		]
	]]
	
SetAttributes[ backChainAux, HoldFirst];
(* explicit *)
backChainAux[ data_, explicit_Integer?Positive, implicit_, env_][ ___] :=
	With[ {explicit0 = explicit - 1},
	With[ {tail = LL[ Hold[ backChainAux[ data, explicit0, implicit, env]], LL[]], elim = data[[2]]},
	Module[ {data0 = ReplacePart[ data, {1 -> explicit, 2 -> If[ explicit === 1, elim, 1], {-2, 2} -> data[[-2, 2]] - 1}]},
		LL[ Hold[ backChainSingle @@ data0], tail]
	]]]
backChainAux[ data_, {i_Integer?Positive, r___}, implicit_, env_][ ___] :=
	With[ {explicit0 = {r}},
	With[ {tail0 = backChainAux[ data, explicit0, implicit, env]},
	With[ {tail = LL[ Hold[ tail0], LL[]], elim = data[[2]]},
	Module[ {data0 = ReplacePart[ data, {1 -> i, 2 -> If[ i === 1, elim, 1], {-2, 2} -> data[[-2, 2]] - 1}]},
		LL[ Hold[ backChainSingle @@ data0], tail]
	]]]]
(* implicit *)
backChainAux[ data_, 0, All|"all", env_][ ps:PRFSIT$[ _, hyps_List, ___], args___, mode_] :=
	backChainAux[ data, 0, Length[ hyps], env][ ps, args, mode]
backChainAux[ data_, 0, implicit_Integer?Positive, env_][ ___] :=
	With[ {implicit0 = implicit - 1},
		LL[ Hold[ INF$[ "makeExplicit", implicit], backChainSingle @@ data], LL[ Hold[ backChainAux[ data, 0, implicit0, env]], LL[]]]
	]
backChainAux[ data_, 0, {f_, r___}, env_][ ___] :=
	With[ {tail0 = backChainAux[ data, 0, {r}, env]},
		LL[ Hold[ makeExplicitAssms[ {f}, 1], backChainSingle @@ data], LL[ Hold[ tail0], LL[]]]
	]
(* environment *)
backChainAux[ data_, 0, 0, All|"all"] :=
	backChainAux[ data, 0, 0, (Join @@ getFormulas[ True][])[[All, 1]]]
backChainAux[ _, 0, 0, 0][ ___] :=
	LL[]
backChainAux[ data_, 0, 0, {f_, r___}][ ___] :=
	With[ {tail0 = backChainAux[ data, 0, 0, {r}]},
		(* 'f' must only refer to a single formula! *)
		LL[ Hold[ addKnowledgeExplicit[ {f}, 1, 3, Hold[ backChainSingle @@ data], False, None]], LL[ Hold[ tail0], LL[]]]
	]

(* 'backChainSingle[ i, elim, atomR, atomG, instMeta, flexflex, addEquations, post][ ps, _, mode]' backchains the first
	subgoal in 'ps' w.r.t. its 'i'-th explicit assumption ('i' may be negative, of course).
	'elim' can be an arbitrary integer or a list of non-zero integers specifying which explicit assumptions to
	eliminate; a single integer means that the first 'elim' assumptions shall be eliminated (if 'elim' is negative, all
	but the last '-elim - 1' assumptions are eliminated). ATTENTION! If 'elim' is neither '0' nor '{}', 'i' MUST be '1'
	and 'elim' refers to the assumptions AFTER the first one; for instance, '{1}' means that actually the second
	assumption shall be eliminated.
	'atomR' and 'atomG' are also related to eliminating assumptions, but in contrast to
	'elim' they do not refer to an arbitrary assumption, but rather specify how many of the trailing premises of the
	rule are treated as being part of the rule's atomic conclusion, and how many assumptions of the subgoal are treated
	as being part of the atomic goal; hence, they roughly correspond to the parameter of basic inference rule "impE".
	Note, however, that 'atomR' and 'atomG' may be 'All' or "all", and that 'atomR' may also be 'Automatic' (in which
	case it is set to the same value as 'atomG').
	'flexflex' specifies what to do with flex-flex pairs: if 'True', such pairs are on the one hand added as new
	subgoals, and on the other hand used for making the goal and the conclusion of the rule alpha-equivalent; if
	'False', all unifiers involving flex-flex pairs are simply discarded; if 'Eliminate', all meta-variables in
	flex-flex pairs are immediately instantiated by a constant function.
	'addEquations' is a boolean value specifying whether further equations from other subgoals shall be added to the
	unification-system or not; this could potentially ease the unification problem and lead to fewer flex-flex pairs in
	the solutions.
	'post' is an arbitrary sequence of proof commands inside 'Hold', to be applied after backchaining.
	In contrast to basic inference "impE", this proof method potentially instantiates Core-universally-quantified
	variables in the rule, as well as type- and term meta variables, depending on '$instantiateTypes' and 'instMeta'. *)
backChainSingle[] :=
	backChainSingle[ 1, 0, Automatic, 0, True, True, False, Hold[]]
backChainSingle[ i_Integer] :=
	backChainSingle[ i, 0, Automatic, 0, True, True, False, Hold[]]
backChainSingle[ i_Integer, elim_] :=
	backChainSingle[ i, elim, Automatic, 0, True, True, False, Hold[]]
backChainSingle[ i_Integer, elim_, atomR_] :=
	backChainSingle[ i, elim, atomR, 0, True, True, False, Hold[]]
backChainSingle[ i_Integer, elim_, atomR_, atomG_] :=
	backChainSingle[ i, elim, atomR, atomG, True, True, False, Hold[]]
backChainSingle[ i_Integer, elim_, atomR_, atomG_, instMeta:Except[ _List]] :=
	backChainSingle[ i, elim, atomR, atomG, instMeta, True, False, Hold[]]
backChainSingle[ i_Integer, elim_, atomR_, atomG_, instMeta:Except[ _List], flexflex_] :=
	backChainSingle[ i, elim, atomR, atomG, instMeta, flexflex, False, Hold[]]
backChainSingle[ i_Integer, elim_, atomR_, atomG_, instMeta:Except[ _List], flexflex_, addEquations:Except[ _List]] :=
	backChainSingle[ i, elim, atomR, atomG, instMeta, flexflex, addEquations, Hold[]]
backChainSingle[ i0_Integer, elim0_, atomR0_, atomG0_, instMeta:Except[ _List], flexflex_, addEquations:Except[ _List], post_Hold][
		ps:PRFSIT$[ {G_, R___}, __], args___, mode_] :=
	With[ {acGoal = assmCount[ G, 0, Infinity]},
	With[ {i = If[ Negative[ i0], acGoal + 1 + i0, i0],
			elim =
				Switch[ elim0,
					{},
					0,
				(*else*)
					_List,
					DeleteDuplicates[ Replace[ elim0, e_Integer?Negative :> (acGoal + 1 + e), {1}]],
				(*else*)
					_?Negative,
					acGoal + 1 + elim0,
				(*else*)
					_,
					elim0
				]},
	With[ {minAssmNum = Max[ elim] + i},
		If[ (elim === 0 || i === 1) && Positive[ i] && NonNegative[ Min[ elim]] && minAssmNum <= acGoal,
			With[ {atomG =
					Switch[ atomG0,
						All|"all",
						acGoal - minAssmNum,
						_Integer?NonNegative,
						If[ acGoal >= atomG0 + minAssmNum, atomG0, $Failed],
						_,
						$Failed
					]},
				If[ IntegerQ[ atomG],
					With[ {drop = acGoal - atomG},	(* 'drop' is not zero, since 'i' is positive *)
					With[ {goal = dropAssms[ G, drop]},
						backChainSingle[
							i,
							(* 'elim' is increased by 1 throughout the rest of 'backChainSingle'!
								That means the rule (which must be the first assumption) is now counted, too. *)
							elim + 1,
							atomR0,
							atomG,
							instMeta,
							flexflex,
							{
								goal,
								drop,
								If[ TrueQ[ instMeta && addEquations],
									First[ collectMetaEquations[ {R}, DeleteDuplicates[ Cases[ {Extract[ G, assmPosition[ i]], goal}, _META$, {0, Infinity}, Heads -> True]]]],
								(*else*)
									{}
								]
							},
							post
						][ ps, args, mode]
					]],
				(*else*)
					sowCPMMessage[ 1011, {atomG0}, "Fatal", "backChainSingle"];
					$Failed
				]
			],
		(*else*)
			sowCPMMessage[ 1006, {i0}, "Fatal", "backChainSingle"];
			$Failed
		]
	]]]
backChainSingle[ i_Integer, elim_, atomR_, atomG_Integer, instMeta_, flexflex_, info:{goal_, drop_, equations_List}, post_Hold][
		ps:PRFSIT$[ {G_, R___}, _, _, fix_, meta_, tfix_Association, tvar_Association, __], args___, mode_] :=
	With[ {rule = Extract[ G, assmPosition[ i]]},
		If[ mode === 1,
			Replace[ executeProofCommands[ Hold[ backChainSingle[ i, elim, atomR, atomG, instMeta, flexflex, info, post]], {ps}, 0, False],
				{
					PRFSTAT$[ _, {sit:PRFSIT$[ G0_List, __]}, assms_List, thms_List, ___] :>
						{
							makeANDNODE[ makePRFINFO[ name -> backChainSingle, used -> {{rule}}, generated -> {Drop[ G0, -Length[ {R}]]}], sit],
							assms,
							thms,
							Hold[]
						},
					_ -> $Failed
				}
			],
		(*else*)
			Module[ {tag, aux},
				Sow[ Hold[ clearIterator[ tag], Clear[ aux]], "GC"];
				With[ {mc = If[ TrueQ[ instMeta], 0, Length[ meta]]},
					aux[] := Hold[ instantiateCoreAll[ {Automatic...}, i], backChainSingle[ i, elim, atomR, atomG, mc, flexflex, info, tag, aux, post]]
				];
				
				If[ elim === 1 && MatchQ[ goal, TYPED$[ CoreAll$TM, _][ _]] && MatchQ[ rule, TYPED$[ CoreAll$TM, _][ _]],
					If[ TrueQ[ instMeta],
						With[ {meta0 = Alternatives @@ meta[[All, 1]]},
							aux[ False][ arg_] := backChainSinglePost[ i, elim, 0, flexflex, arg, {0, drop}, meta0, post]
						];
						lInterleave[
							lMap[ aux[ False], unify[ Prepend[ equations, {goal, rule}], meta, fix, tfix, tvar, tag, "FlexFlexToTerms" -> False, "AllowFlexFlex" -> (flexflex =!= False)]],
							Hold[ aux[]],
							True
						],
					(*else*)
						LL[ Hold[ INF$[ "impE", i, atomG]], aux[]]
					],
				(*else*)
					aux[]
				]
			]
		]
	]
backChainSingle[ i_Integer, elim_, atomR0_, atomG_Integer, instMeta_Integer?NonNegative, flexflex_, {goal_, dropGoal_, equations_}, tag_Symbol, aux_Symbol, post_Hold][
		PRFSIT$[ {G_, ___}, _, _, fix_, meta_, tfix_Association, tvar_Association, __], ___] :=
	With[ {rule = Extract[ G, assmPosition[ i]]},
	With[ {acRule = assmCount[ rule, 0, Infinity], m = If[ ListQ[ elim], Length[ elim], elim - 1]},
		If[ IntegerQ[ elim] && dropGoal === elim === acRule + 1,	(* 'dropGoal' counts the rule at the 'i'-th assumption as well *)
			(* the goal and the rule can be fully atomized *)
			(* if 'elim === 1', the result is exactly the same as in the other case below *)
			With[ {meta0 = Alternatives @@ meta[[All, 1]], max = Max[ acRule, atomG + dropGoal - 1]},	(* 'atomG + dropGoal' equals 'assmCount[ G, 0, Infinity]' *)
				(* Using 'max' here could in some rare situations lead to problems (because unification may yield more assumptions than 'max'),
					but it avoids computing the number of explicit assumptions in 'backChainSinglePost'. *)
				aux[ True][ arg_] := backChainSinglePost[ i, 1, max, flexflex, arg, {0, 1}, meta0, post]
			];
			lMap[
				aux[ True],
				unify[
					Join[ {{dropAssms[ G, 1], rule}}, equations],
					If[ instMeta === 0, meta, ReplacePart[ meta, Table[ {k, 2, 2} -> False, {k, 1, instMeta}]]],
					fix,
					tfix,
					tvar,
					tag,
					"FlexFlexToTerms" -> False,
					"AllowFlexFlex" -> (flexflex =!= False)
				]
			],
		(*else*)
			With[ {atomR =
					Switch[ atomR0,
						All|"all",
						acRule - m,
						_Integer?NonNegative,
						If[ acRule >= atomR0 + m, atomR0, $Failed],
						Automatic,
						If[ acRule >= atomG + m, atomG, $Failed],
						_,
						$Failed
					]},
				If[ IntegerQ[ atomR],
					With[ {dropRule = acRule - atomR},	(* 'dropRule' could be zero *)
						(* we must provide a new definition for 'aux' with parameter 'True', because 'ps' might have changed *)
						With[ {meta0 = Alternatives @@ meta[[All, 1]], max = Max[ atomR, atomG]},
							(* Using 'max' here could in some rare situations lead to problems (because unification may yield more assumptions than 'max'),
								but it avoids computing the number of explicit assumptions in 'backChainSinglePost'. *)
							aux[ True][ arg_] := backChainSinglePost[ i, elim, max, flexflex, arg, {dropRule, dropGoal}, meta0, post]
						];
						lMap[
							aux[ True],
							unify[
								Join[
									{{goal, dropAssms[ rule, dropRule]}},
									Switch[ elim,
										1|{},
										{},
									(*else*)
										_List,
										Transpose[ {Extract[ G, assmPosition[ elim]], takeAssms[ rule, {}, Length[ elim]]}],
									(*else*)
										_,
										Transpose[ {Rest[ takeAssms[ G, {}, elim]], takeAssms[ rule, {}, elim - 1]}]
									],
									equations
								],
								If[ instMeta === 0, meta, ReplacePart[ meta, Table[ {k, 2, 2} -> False, {k, 1, instMeta}]]],
								fix,
								tfix,
								tvar,
								tag,
								"FlexFlexToTerms" -> False,
								"AllowFlexFlex" -> (flexflex =!= False)
							]
						]
					],
				(*else*)
					sowCPMMessage[ 143, "Fatal"];
					$Failed
				]
			]
		]
	]]

(* 'backChainSinglePost' is an auxiliary function that constructs a sequence of proof steps that perform backchaining,
	knowing a unifier already. *)
backChainSinglePost[ i_, elim_, atom_, _, {{}, rem_List, subst_List, tsubst_List, tvar_Association}, {dropRule_, _}, meta0_, post_] :=
	Join[
		Hold @@ Join[
			{INF$[ "sorts", tvar]},
			If[ tsubst === {}, {}, {INF$[ "instT", tsubst]}],
			Cases[ rem, (m:Except[ meta0] -> {t_, d_}) :> INF$[ "meta", m, t, d]],
			Cases[ subst, (m_ -> inst_) :> INF$[ "inst", m -> STERM$[ normalize[ inst], Association[], Association[]]]],
			{INF$[ "impE", i, atom]},
			Switch[ elim,
				1|{},
				{},
			(*else*)
				_List,
				{backChainSingleFF[ elim, dropRule - Length[ elim]]},
			(*else*)
				_,
				{backChainSingleFF[ elim, dropRule + 1 - elim]}
			]
		],
		post
	]
backChainSinglePost[ i_, elim_, atom_, Eliminate, {ffEqns_List, rem_List, subst_List, tsubst_List, tvar_Association}, {dropRule_, _}, meta0_, post_] :=
	Join[
		Hold @@ Join[
			{INF$[ "sorts", tvar]},
			If[ tsubst === {}, {}, {INF$[ "instT", tsubst]}],
			Cases[ rem, (m:Except[ meta0] -> {t_, d_}) :> INF$[ "meta", m, t, d]],
			Cases[ subst, (m_ -> inst_) :> INF$[ "inst", m -> STERM$[ normalize[ inst], Association[], Association[]]]],
			Map[
				With[ {typ = flattenTypes[ First[ Replace[ #, rem]], {}]},
					INF$[ "inst", # -> STERM$[ applyCurried[ TYPED$[ Theorema`Knowledge`TYPE$TM, {Last[ typ]}], {}, Reverse[ Most[ typ]]], Association[], Association[]]]
				]&,
				Union @@ Replace[ ffEqns, {{{u_META$, _}, {v_META$, _}, _} :> {u, v}, _ -> {}}, {1}]
			],
			{INF$[ "impE", i, atom]},
			Switch[ elim,
				0|{},
				{},
			(*else*)
				_List,
				{backChainSingleFF[ elim, dropRule - Length[ elim]]},
			(*else*)
				_,
				{backChainSingleFF[ elim, dropRule + 1 - elim]}
			]
		],
		post
	]
backChainSinglePost[ i_, elim_, atom_, True, {ffEqns_List, rem_List, subst_List, tsubst_List, tvar_Association}, drop_, meta0_, post_] :=
	Join[
		Hold @@ Join[
			{INF$[ "sorts", tvar]},
			If[ tsubst === {}, {}, {INF$[ "instT", tsubst]}],
			Cases[ rem, (m:Except[ meta0] -> {t_, d_}) :> INF$[ "meta", m, t, d]],
			Cases[ subst, (m_ -> inst_) :> INF$[ "inst", m -> STERM$[ normalize[ inst], Association[], Association[]]]],
			{backChainSingleFF[ i, elim, atom, ffEqns, drop]}
		],
		post
	]
backChainSinglePost[ ___] :=
	$Failed

(* The binary version of 'backChainSingleFF' is applied after applying "impE". *)
backChainSingleFF[ 0|{}, _] :=
	voidTac[]
backChainSingleFF[ elim0_List, numGoals_Integer][ PRFSIT$[ {G_, ___}, __], ___] :=
	With[ {elim = elim0 - 1},
		backChainSingleElim[ elim, numGoals, Map[ assmCount[ Extract[ G, #], 0, Infinity]&, assmPosition[ elim]]]
	]
backChainSingleFF[ elim_Integer, numGoals_Integer][ PRFSIT$[ {G_, ___}, __], ___] :=
	backChainSingleElim[ elim - 1, numGoals, Map[ assmCount[ #, 0, Infinity]&, takeAssms[ G, {}, elim - 1]]]
(* The following version of 'backChainSingleFF' is applied after instantiating the meta variables, but before applying
	"impE". It handles remaining flex-flex equations, as well as elimination. *)
backChainSingleFF[ i_, elim_, atom_, flexflex_List, {dropRule_Integer, dropGoal_Integer}][ PRFSIT$[ {G_, ___}, _, _, _, meta_, __], ___] :=
	With[ {posRule0 = assmPosition[ i], numElim = If[ ListQ[ elim], Length[ elim], elim - 1]},
	With[ {posRule = Join[ posRule0, Table[ 1, {dropRule}]], posGoal = Table[ 1, {dropGoal}],
			posElimP = Map[ Join[ posRule0, #]&, assmPosition[ Range[ numElim]]],
			posElimA = assmPosition[ If[ ListQ[ elim], elim, Range[ 2, elim]]]},
	With[ {elimA = Extract[ G, posElimA]},
		Replace[
			Reap[
				Block[ {$backChainFlexFlex = flexflex},
					identifyTerms[ Extract[ G, posRule], Extract[ G, posGoal], {}, posRule, Prepend[ posGoal, 1]] &&
					Catch[
						MapThread[
							If[ !identifyTerms[ Extract[ G, #2], #1, {}, #2, #3], Throw[ False]]&,
							{elimA, posElimP, posElimA}
						];
						True
					]
				]
			],
			{
				{True, rep_List} :>
					Join[
						Hold @@ Join[
							Join @@ Replace[
								GatherBy[ Join @@ rep, First],
								all:{j_Integer -> _, ___} :>
									With[ {eqn =
											Replace[ flexflex[[j]],
												{{ah_, aargs_List}, {bh_, bargs_List}, loose_List} :>
													TYPED$[ CoreEqual$TM, {Fold[ FUN$TM[ #2, #1]&, Nest[ Last, First[ Replace[ ah, meta]], Length[ aargs]], loose]}][
														applyCurried[ ah, aargs, loose]][ applyCurried[ bh, bargs, loose]
													]
											]},
									With[ {abs = abstractionCount[ eqn[[0, 1]], 0, Infinity]},
										(* we first construct the term with 'applyCurried' and only then compute the number of abstractions,
											because there might be eta-reductions *)
										Join[
											{
												setupSubgoalsMP[ eqn, dropGoal],	(* 'eqn' becomes the 'dropGoal + 1'-th explicit assumption *)
												INF$[ "rearrangeG", 2, 1]
											},
											If[ abs === 0,
												{},
											(*else*)
												{
													convertBetaEta[
														all[[All, 2]],
														{"beta", Expand, abs},
														False
													]
												}
											],
											With[ {posAbs = Table[ 0, {abs}]},
												Map[
													INF$[ "replace", dropGoal + 1, Join[ Last[ #], posAbs]]&,
													all
												]
											],
											{INF$[ "drop", dropGoal + 1]}
										]
									]],
								{1}
							],
							{INF$[ "impE", i, atom]}
						],
						backChainSingleElim[ elim - 1, dropRule - numElim, Map[ assmCount[ #, 0, Infinity]&, elimA]]
					],
				_ :> (sowCPMMessage[ 1012, "Fatal", "backChainSingle"]; $Failed)
			}
		]
	]]]

(* 'backChainSingleElim' is no proof method but merely an auxiliary function for constructing a sequence of proof commands *)
backChainSingleElim[ 0|{}, _Integer, _List] :=
	Hold[]
backChainSingleElim[ elim_List, numGoals_Integer, ac_List] :=
	Join[
		Hold @@ MapThread[ INF$[ "impE", #1, #2]&, {elim, ac}],
		mapAllInline[ Hold @@ MapIndexed[ INF$[ "drop", #1 + 1 - First[ #2]]&, Sort[ elim]], numGoals, numGoals]
	]
backChainSingleElim[ elim_Integer, numGoals_Integer, ac_List] :=
	Join[
		Hold @@ MapIndexed[ INF$[ "impE", First[ #2], #1]&, ac],
		mapAllInline[ Hold @@ Table[ INF$[ "drop", 1], {elim}], numGoals, numGoals]
	]

identifyTerms[ a:(_VAR$|_FIX$|_META$), b_, _, _, _] :=
	(a === b)
identifyTerms[ _, _VAR$|_FIX$|_META$, _, _, _] :=
	False
identifyTerms[ TYPED$[ a_, _], TYPED$[ b_, _], loose_, pos1_, pos2_] :=
	identifyTerms[ a, b, loose, Append[ pos1, 1], Append[ pos1, 2]]
identifyTerms[ TYPED$[ a_, _], b_, loose_, pos1_, pos2_] :=
	identifyTerms[ a, b, loose, Append[ pos1, 1], pos2]
identifyTerms[ a_, TYPED$[ b_, _], loose_, pos1_, pos2_] :=
	identifyTerms[ a, b, loose, pos1, Append[ pos2, 1]]
identifyTerms[ Lambda$TM[ RNG$[ TYPRNG$[ _, s_]], a_], Lambda$TM[ RNG$[ TYPRNG$[ _, _]], b_], loose_, pos1_, pos2_] :=
	identifyTerms[ a, b, Prepend[ loose, s], Append[ pos1, 2], Append[ pos2, 2]]
identifyTerms[ Lambda$TM[ RNG$[ TYPRNG$[ _, s_]], a_], b_, loose_, pos1_, pos2_] :=
	identifyTerms[ a, mapFreeVars[ b, VAR$[ #3 + 1]&][ VAR$[ 0]], Prepend[ loose, s], Append[ pos1, 2], pos2]
identifyTerms[ a_, Lambda$TM[ RNG$[ TYPRNG$[ _, s_]], b_], loose_, pos1_, pos2_] :=
	identifyTerms[ mapFreeVars[ a, VAR$[ #3 + 1]&][ VAR$[ 0]], b, Prepend[ loose, s], pos1, Append[ pos2, 2]]
identifyTerms[ a:(_[ _]), b_, loose_, pos1_, pos2_] :=
	If[ MatchQ[ b, _[ _]],
		Replace[ {splitHeadArgs[ a], splitHeadArgs[ b]},
			{
				{a0:{_META$, _List}, b0:{_META$, _List}} :>
					With[ {i = findFlexFlexEquation[ a0, b0, loose]},
						(
							If[ Negative[ i],
								Sow[ -i -> pos2],
							(*else*)
								Sow[ i -> pos1]
							];
							True
						) /; IntegerQ[ i]
					],
				{{ah_, aargs_List}, {bh_, bargs_List}} :>
					With[ {la = Length[ aargs]},
						If[ la === Length[ bargs],
							With[ {tab = Table[ 0, {la}]}, identifyTerms[ ah, bh, loose, Join[ pos1, tab], Join[ pos2, tab]]] &&
							Catch[
								MapThread[
									With[ {tab = Table[ 0, {#3}]},
										If[ !identifyTerms[ #1, #2, loose, Join[ pos1, tab, {1}], Join[ pos2, tab, {1}]], Throw[ False]]
									]&,
									{aargs, bargs, Range[ la - 1, 0, -1]}
								];
								True
							],
						(*else*)
							False
						]
					]
			}
		],
	(*else*)
		False
	]
identifyTerms[ a_, b_, _, _, _] :=
	a === b

findFlexFlexEquation[ a_List, b_List, loose_List] :=
	Module[ {i = 0},
		Catch[
			Scan[
				(
					++i;
					Which[
						findFlexFlexEquation[ {a, b, loose}, #],
						Throw[ i],
					(*else*)
						findFlexFlexEquation[ {b, a, loose}, #],
						Throw[ -i]
					]
				)&,
				$backChainFlexFlex
			];
			$Failed
		]
	]
findFlexFlexEquation[ {{a1h_, a1args_List}, {b1h_, b1args_List}, loose1_List}, {{a2h_, a2args_List}, {b2h_, b2args_List}, loose2_List}] :=
	If[ a1h === a2h && b1h === b2h,
		Catch[
			Quiet[ Check[
				Block[ {$backChainContext = {loose1, loose2}},
					MapThread[ If[ !alphaEquivCtxt[ #1, #2, -1], Throw[ False]]&, {a1args, a2args}];
					MapThread[ If[ !alphaEquivCtxt[ #1, #2, -1], Throw[ False]]&, {b1args, b2args}]
				],
				Throw[ False]
			]];
			True
		],
	(*else*)
		False
	]

alphaEquivCtxt[ a:(_FIX$|_META$), b_, _] :=
	a === b
alphaEquivCtxt[ VAR$[ i_Integer?NonNegative], b_, n_] :=
	Replace[ b,
		{
			VAR$[ i] :> (i <= n || $backChainContext[[1, i - n]] === $backChainContext[[2, i - n]]),
			_ -> False
		}
	]
alphaEquivCtxt[ _, _VAR$|_FIX$|_META$, _] :=
	False
alphaEquivCtxt[ x:TYPED$[ a_, s_], y_, n_] :=
	Replace[ y,
		{
			TYPED$[ b_, s] :> alphaEquivCtxt[ a, b, n],
			_ -> False
		}
	]
alphaEquivCtxt[ _, _TYPED$, _] :=
	False
alphaEquivCtxt[ Lambda$TM[ RNG$[ TYPRNG$[ _, s_]], a_], b0_, n_] :=
	Replace[ b0,
		{
			Lambda$TM[ RNG$[ TYPRNG$[ _, s]], b_] :> alphaEquivCtxt[ a, b, n + 1],
			_ -> False
		}
	]
alphaEquivCtxt[ _, _Lambda$TM, _] :=
	False
alphaEquivCtxt[ f_[ a_], b0_, n_] :=
	Replace[ b0,
		{
			g_[ b_] :> alphaEquivCtxt[ f, g, n] && alphaEquivCtxt[ a, b, n],
			_ -> False
		}
	]
alphaEquivCtxt[ a_, b_, _] :=
	a === b

(* yet to be tested *)
collectMetaEquations[ {}, meta_List] :=
	{{}, meta}
collectMetaEquations[ _List, {}] :=
	{{}, {}}
collectMetaEquations[ goals_List, meta0_List] :=
	Module[ {meta = meta0, added = True, addedPos = {}, i},
		{
			Join @@ Last @ Reap[
				While[ added,
					added = False;
					i = 0;
					Scan[
						If[ !MemberQ[ addedPos, ++i],
							With[ {res = isMetaEquation[ #, Hold[ meta], 0]},
								If[ IntegerQ[ res],
									Sow[ {Extract[ #, assmPosition[ res + 1]], dropAssms[ #, res + 1]}, Null];
									added = True;
									AppendTo[ addedPos, i];
									Null
								]
							]
						]&,
						goals
					]
				],
				Null
			],
			meta
		}
	]

isMetaEquation[ TYPED$[ CoreImplies$TM, _][ _][ G_], meta_Hold, n_] :=
	isMetaEquation[ G, meta, n + 1]
isMetaEquation[ TYPED$[ CoreEqual$TM, _][ a0_][ b0_], Hold[ meta_], n_] :=
	With[ {a = stripAbstractions[ a0], b = stripAbstractions[ b0]},
	With[ {ah = getHeadAtom[ HoldComplete[ a]][[1, 1]], bh = getHeadAtom[ HoldComplete[ b]][[1, 1]]},
		If[ MemberQ[ ah, meta] || MemberQ[ bh, meta],
			meta = Join[ meta, DeleteDuplicates[ Cases[ {a, b}, _META$?(!MemberQ[ #, meta] &), {0, Infinity}, Heads -> True]]];
			n,
		(*else*)
			False
		]
	]]
isMetaEquation[ _, _, _] :=
	False
