(* Theorema 
    Copyright (C) 2010 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

BeginPackage[ "Theorema`CoreProvers`HOLProofMethods`"]

Needs[ "Theorema`Common`"]
Needs[ "Theorema`Language`"]
Needs[ "Theorema`HOL`", FileNameJoin[ {$TheoremaDirectory, "Theorema", "CoreTheories", "HOL", "Packages", "init.m"}]]
Needs[ "LazyList`", FileNameJoin[ {$TheoremaDirectory, "Theorema", "Components", "LazyList.m"}]]
Needs[ "Theorema`Core`Trusted`"]
Needs[ "Theorema`Core`Tools`"]
Needs[ "Theorema`Core`HOUnification`"]
Needs[ "Theorema`CoreProvers`"]

(* "HOLProofMethods.m" depends upon the following theorems:
	"TrueI", "\[Not]FalseI", "FalseE", "\[Equal] refl", "\[ForAll] reflect", "\[ForAll]C reflect", "\[Implies] reflect",
	"\[And] reflect", "\[Equal] reflect", "\[Not] reflect", "TrueQ reflect", "\[Exists]I", "\[Exists]CI", "\[Exists]E",
	"\[Exists]CE". *)

Begin["`Private`"]

(* ::Section:: *)
(*Terminal Proof Steps*)

If[ !MemberQ[ $terminalProofSteps, HOLTerminalStep], PrependTo[ $terminalProofSteps, HOLTerminalStep]];

HOLTerminalStep[ G_, hyps_List, _, _, _, _, _] :=
	With[ {res = goalIterator[ G, Placeholder, terminalAssmFun[ G], terminalGoalFun[ G, hyps], $Failed]},
		If[ res === $Failed,
			res,
		(*else*)
			Function[ res] /. Placeholder -> Slot[ 1]
		]
	]

terminalAssmFun[ G_][ TYPED$[ False, {}], i_, True] :=
	Replace[ extractAtomicGoal[ G, 0, 0, Infinity, Infinity],
		{
			{TYPED$[ TrueQ$TM, {}][ _], assms_Integer, all_Integer} :>
				Replace[ makeFalseCmd[ i],
					cmd_Hold :>
						Join[
							If[ Positive[ all], Hold[ fastFix[ 1, all]], Hold[]],
							Hold @@ Table[ INF$[ "drop", i + 1], {assms - i}],
							cmd
						]
				],
			{TYPED$[ CoreEqual$TM, {_}][ _][ _], assms_Integer, all_Integer} :>
				Replace[ makeFalseCmd[ i],
					cmd_Hold :>
						Join[
							If[ Positive[ all], Hold[ fastFix[ 1, all]], Hold[]],
							Hold @@ Table[ INF$[ "drop", i + 1], {assms - i}],
							Hold @@ {atomizeHOLConstant[ CoreEqual$TM, Join[ Table[ 1, {i}], {0, 0}]]},
							cmd
						]
				],
			_ -> $Failed
		}
	]

terminalGoalFun[ _, _][ TYPED$[ True, {}], n_, True] :=
	makeTrueCmd[ False]
terminalGoalFun[ _, _][ TYPED$[ Not$TM, {}][ TYPED$[ False, {}]], n_, True] :=
	makeTrueCmd[ True]

goalIterator[ TYPED$[ CoreImplies$TM, {}][ TYPED$[ TrueQ$TM, {}][ A_]][ G_], n_, assmFun_, goalFun_, final_] :=
	Replace[ assmFun[ A, n + 1, True],
		Except[ _Hold] :> goalIterator[ G, n + 1, assmFun, goalFun, final]
	]
goalIterator[ TYPED$[ CoreImplies$TM, {}][ A_][ G_], n_, assmFun_, goalFun_, final_] :=
	Replace[ assmFun[ A, n + 1, False],
		Except[ _Hold] :> goalIterator[ G, n + 1, assmFun, goalFun, final]
	]
goalIterator[ TYPED$[ TrueQ$TM, {}][ G_], n_, _, goalFun_, final_] :=
	Replace[ goalFun[ G, n, True],
		Except[ _Hold] -> final
	]
goalIterator[ G_, n_, _, goalFun_, final_] :=
	Replace[ goalFun[ G, n, False],
		Except[ _Hold] -> final
	]


(* ::Section:: *)
(*Definitions*)

(* 'defineHOLConst[ def][ ___]' defines a fresh a.b.f. constant 'c' to be equal to a given term 'term'; the resulting
	equality is added to the hypotheses of the current proof situation. 'c' and 'term' are given by 'def', which must
	be a (meta-) equality (possibly universally quantified, possibly inside 'EXPR$' and/or 'LABELED$').
	In order for the proof method to succeed, 'const' must be a fresh a.b.f. constant, 'term' must be well-typed and
	free of meta variables, and there must be only one single subgoal. Note that *all* type variables 'term' depends
	on, either explicitly or implicitly via a.b.f. constants, are automatically specialized. *)
defineHOLConst[ LABELED$[ def_, lbl_String, ___]] :=
	defineHOLConst[ def, lbl]
defineHOLConst[ LABELED$[ def_, _, ___]] :=
	defineHOLConst[ def, "?"]
defineHOLConst[ def_] :=
	defineHOLConst[ def, Automatic]
defineHOLConst[ def0_, lbl_][ ps:PRFSIT$[ {G_}, _, idx_, fix_, _, tfix_Association, tvar_Association, __], ___] :=
	Replace[ makeCoreExprPatt[ def0, FORM$TM, idx, fix, {}, tfix, tvar, getTermAbbrevList[ ps], None, normalize, None],
		{
			{{_, STERM$[ def_, newTFix_, newTVar_], _, {const_ -> typ_}, {}}, ts_List, {}, abb_} :>
				Replace[ translateToCoreEquality[ def, {}, True],
					{
						TYPED$[ CoreEqual$TM, _][ c_][ term_] :>
							If[ c === const,
								With[ {fixInTerm = Cases[ term, _FIX$, {0, Infinity}, Heads -> True]},
									If[ MemberQ[ fixInTerm, const],
										sowHOLMessage[ 1001, "Fatal"];
										$Failed,
									(*else*)
										With[ {subst = typeSpecializationReplacement[ {term, Replace[ fixInTerm, fix, {1}] /. ts}, Join[ Keys[ tfix], Keys[ newTFix]]]},
											If[ subst =!= {}, sowHOLMessage[ 1002, "Warning"]];
											defineConst[
												const,
												applyTypeGenSpecRepl[ {}, subst][ STERM$[ term, newTFix, newTVar]],
												typ /. subst,
												Join[ subst, ts /. subst],
												applyTypeGenSpecRepl[ {}, subst, False][ abb],
												G,
												lbl
											]
										]
									]
								],
							(*else*)
								sowHOLMessage[ 1004, "Fatal"];
								$Failed
							],
						0 :> (sowHOLMessage[ 1, "Fatal"]; $Failed),
						i_Integer :> (sowHOLMessage[ i, "Fatal"]; $Failed)
					}
				],
			_ :> (sowHOLMessage[ 1003, "Fatal"]; $Failed)
		}
	]


(* ::Section:: *)
(*Formula Formats*)

Theorema`CoreProvers`CoreProofMethods`$canonicalFormatConstructor = makeHOLCanonicalFormat;
Theorema`CoreProvers`CoreProofMethods`$equationalFormatConstructor = makeHOLEquationalFormat;

(* 'makeHOLCanonicalFormat' simulates repeated application of the following transformations:
	- '... ==> \all x. P[x]' --> '... ==> /\x. P[x]',
	- '... ==> \all C[x]. P[x]' --> '... ==> /\x. C[x] ==> P[x]',
	- '... ==> A => B' --> '... ==> A ==> B',
	- '... ==> A ==> /\x. P[x]' --> '... ==> /\x. A ==> P[x]'.
	With option "conj" (default 'False') it also transforms
	- '... ==> (A & B) ==> C' --> '... ==> A ==> B ==> C'.
	With option "split" (default 'False') it also transforms
	- '... ==> (A & B)' --> '... ==> A', '... ==> B'.
	With option "equal" (default 'False') it also transforms
	- '... ==> x == y' --> '... ==> x === y'.
	Therefore, it might return more than one result. *)
makeHOLCanonicalFormat[ form_, i_, n_, args_] :=
	With[ {revArgs = Reverse[ args]},
	With[ {conj = getOptValue[ revArgs, "conj", False],
			split = getOptValue[ revArgs, "split", False],
			equal = getOptValue[ revArgs, "equal", False]},
		Cases[ makeHOLCanonicalFormat[ equal, i, n] /@ analyzeHOLStructure[ form, conj, split], {_, _}]
	]]
makeHOLCanonicalFormat[ eq_, i_, n_][ {A_, hist_, loose_, indices_, prems_, premsScope_}] :=
	Which[
		TrueQ[ eq] && MatchQ[ A, TYPED$[ TrueQ$TM, {}][ TYPED$[ Equal$TM, {_}][ _][ _]]],
		makeHOLCanonicalFormat[ False, i, n][ {ReplacePart[ First[ A], {0, 0, 1} -> CoreEqual$TM], Append[ hist, 4], loose, indices, prems, premsScope}],
	(*else*)
		MatchQ[ hist, {-1..., -2...}],
		(* formula is in canonical format already *)
		Null,
	(*else*)
		True,
		Replace[ Theorema`CoreProvers`CoreProofMethods`univToLambdaAdjustDB[ False, A, loose, indices, prems, premsScope],
			{A1_, loose1_, prems1_, {}} :>
				{
					Fold[ TYPED$[ CoreAll$TM, {Last[ #2]}][ Lambda$TM[ RNG$[ #2], #1]] &, fromAssmsGoal[ prems1, A1], Reverse[ loose1]],
					With[ {ll = Length[ loose1]},
						HoldComplete[ getHOLCanonicalFormatProof[ i, n, hist, ll, 0]]
					]
				}
		]
	]

(* 'makeHOLEquationalFormat' simulates repeated application of the following transformations:
	- '... ==> \all x. P[x]]' --> '... ==> /\x. P[x]',
	- '... ==> \all C[x]. P[x]' --> '... ==> /\x. C[x] ==> P[x]',
	- '... ==> A => B' --> '... ==> A ==> B',
	- '... ==> !A' --> '... ==> A === False',
	- '... ==> TrueQ[A]' --> '... ==> A === True',
	- '... ==> A ==> /\x. P[x]' --> '... ==> /\x. A ==> P[x]',
	- '/\x. ... ==> f[x] === g[x]' --> '... ==> f === g' (if 'x' does not occur anywhere else).
	With option "conj" (default 'False') it also transforms
	- '... ==> (A & B) ==> C' --> '... ==> A ==> B ==> C'.
	With option "split" (default 'True') it also transforms
	- '... ==> (A & B)' --> '... ==> A', '... ==> B'.
	With option "equal" (default 'True') it also transforms
	- '... ==> x == y' --> '... ==> x === y'.
	Therefore, it might return more than one result, but all results are equalities.
	If 'univ' is 'False' but there are still universally quantified variables in the end, as many *object-level*
	universal quantifiers are "atomized" as necessary (but note that Core quantifiers cannot be atomized at all).
	For instance, 'A ==> /\x. \all y. P[y] => f[x,y] == ...' becomes
	'A ==> \x. (\all y. P[y] => f[x,y] == ...) === \x. True'.
	Note: If the formula is a non-equational Core formula, no equational format can be obtained! *)
makeHOLEquationalFormat[ form_, i_, n_, univ_, args_] :=
	With[ {revArgs = Reverse[ args]},
	With[ {conj = getOptValue[ revArgs, "conj", False],
			split = getOptValue[ revArgs, "split", True],
			equal = getOptValue[ revArgs, "equal", True]},
		Cases[ makeHOLEquationalFormat[ equal, If[ TrueQ[ univ], Null, form], i, n] /@ analyzeHOLStructure[ form, conj, split], {_, _}]
	]]
makeHOLEquationalFormat[ eq_, form_, i_, n_][ {A_, hist_, loose_, indices_, prems_, premsScope_}] :=
	Switch[ {eq, A},
		{True, TYPED$[ TrueQ$TM, {}][ TYPED$[ Equal$TM, {_}][ _][ _]]},
		makeHOLEquationalFormat[ False, form, i, n][
			{ReplacePart[ First[ A], {0, 0, 1} -> CoreEqual$TM], Append[ hist, 4], loose, indices, prems, premsScope}
		],
	(*else*)
		{_, TYPED$[ TrueQ$TM, {}][ TYPED$[ Not$TM, {}][ _]]},
		makeHOLEquationalFormat[ False, form, i, n][
			{TYPED$[ CoreEqual$TM, {Theorema`Knowledge`BOOL$TM}][ A[[ 1, 1]]][ TYPED$[ False, {}]], Append[ hist, 5], loose, indices, prems, premsScope}
		],
	(*else*)
		{_, TYPED$[ TrueQ$TM, {}][ _]},
		makeHOLEquationalFormat[ False, form, i, n][
			{TYPED$[ CoreEqual$TM, {Theorema`Knowledge`BOOL$TM}][ First[ A]][ TYPED$[ True, {}]], Append[ hist, 6], loose, indices, prems, premsScope}
		],
	(*else*)
		{_, TYPED$[ CoreEqual$TM, {_}][ _][ _]},
		Replace[ Theorema`CoreProvers`CoreProofMethods`univToLambdaAdjustDB[ True, A, loose, indices, prems, premsScope],
			{A1_, loose1_, prems1_, good1_} :>
				Which[
					form =!= Null && loose1 =!= {},
					With[ {firstbad = First[ FirstPosition[ adjustHistoryList[ hist, good1], -1, {0}, {1}, Heads -> False]]},
						If[ firstbad <= 1 || 1 =!= hist[[firstbad - 1]] =!= 2,
							(* "bad" variables are bound by Core quantifiers *)
							$Failed,
						(*else*)
							With[ {hist2 = Take[ hist, firstbad - 1]},
							With[ {allNum = Count[ hist2, -1], premNum = Count[ hist2, -2]},
								makeHOLEquationalFormat[ False, form, i, n][
									{
										TYPED$[ CoreEqual$TM, {Theorema`Knowledge`BOOL$TM}][
											extractSingle[ form,
												Module[ {object = False, noImp = 0},
													Join @@ Replace[ DeleteCases[ hist2, 4|5|6],
														{
															-2 :> If[ Positive[ noImp], --noImp; {}, {1}],
															-1 -> {1, 2},
															1|3 :> If[ object, {}, object = True; {1}],
															2 :> (++noImp; If[ object, {}, object = True; {1}]),
															7 :> (++noImp; {}),
															8 :> If[ object, {0, 1}, object = True; {1, 0, 1}],
															9 :> If[ object, {1}, object = True; {1, 1}]
														},
														{1}
													]
												]
											]
										][ TYPED$[ True, {}]],
										ReplacePart[ hist2, -1 -> 6],
										Take[ loose, allNum],
										Take[ indices, allNum],
										Take[ prems, premNum],
										Take[ premsScope, premNum]
									}
								]
							]]
						]
					],
				(*else*)
					good1 === {} && MatchQ[ hist, {-1..., -2...}],
					Null,
				(*else*)
					True,
					{
						Fold[ TYPED$[ CoreAll$TM, {Last[ #2]}][ Lambda$TM[ RNG$[ #2], #1]] &, fromAssmsGoal[ prems1, A1], Reverse[ loose1]],
						With[ {ll = Length[ loose1], lg = Length[ good1], hist1 = adjustHistoryList[ hist, good1]},
							HoldComplete[ getHOLCanonicalFormatProof[ i, n, hist1, ll, lg]]
						]
					}
				]
		],
	(*else*)
		_,
		$Failed
	]

getOptValue[ revOpts_List, n_String, def_] :=
	With[ {i = First[ FirstPosition[ revOpts, n|{n}, {0}, {1}, Heads -> False]]},
		If[ i === 0,
			def,
		(*else*)
			Quiet[ Check[ revOpts[[i + 1]] =!= "-", True]]
		]
	]

adjustHistoryList[ hist_, good_] :=
	MapAt[ List, hist, Position[ hist, -1, {1}, Heads -> False][[good]]]

(* 'analyzeHOLStructure[ term, andL, andR]' analyzes the top-level universal/implication/conjunction structure of the
	given 'term' and returns a list of 6-tuples '{atom, hist, loose, indices, prems, premsScope}', where
	- 'atom' are the atomic conclusions of 'term', after stripping off all universal quantifiers and implications, and,
		if 'andR' is 'True', branching on conjunctions. Both object- and meta-level connectives are stripped off.
	- 'hist' are lists describing "what happened" in the process of obtaining 'atom' from 'term', i.e. which connective
		was stripped off etc.; see above for the meaning of the entries of 'hist'.
	- 'loose' are the type-ranges of all variables bound by quantifiers that have been stripped off.
	- 'indices' are lists of integers mapping every entry in 'loose' to the number of premises in whose scope the
		respective variables are bound.
	- 'prems' are lists of all premises that have been stripped off.
	- 'premsScope' are lists of integers mapping every entry in 'prems' to the number of quantifiers in whose scope the
		respective premise appears; it is the dual of 'indices'.
	If 'andR' is 'False', the function returns a singleton list. *)
analyzeHOLStructure[ term_, andL_, andR_] :=
	analyzeHOLStructure[ term, {}, {}, {}, {}, {}, 0, 0, andL, andR]
(* Forall, no condition *)
analyzeHOLStructure[ TYPED$[ TrueQ$TM, {}][ TYPED$[ Forall$TM, {typ_}][ Lambda$TM[ _, TYPED$[ True, {}]]][ Lambda$TM[ rng_, A_]]],
		hist_, loose_, indices_, prems_, premsScope_, m_, n_, andL_, andR_] :=
	analyzeHOLStructure[
		TYPED$[ CoreAll$TM, {typ}][ Lambda$TM[ rng, TYPED$[ TrueQ$TM, {}][ A]]],
		Append[ hist, 1], loose, indices, prems, premsScope, m, n, andL, andR
	]
analyzeHOLStructure[ TYPED$[ TrueQ$TM, {}][ TYPED$[ Forall$TM, {typ_}][ Lambda$TM[ rng_, TYPED$[ True, {}]]][ A_]],
		hist_, loose_, indices_, prems_, premsScope_, m_, n_, andL_, andR_] :=
	analyzeHOLStructure[
		TYPED$[ CoreAll$TM, {typ}][ Lambda$TM[ rng, TYPED$[ TrueQ$TM, {}][ etaExpandOne[ A]]]],
		Append[ hist, 1], loose, indices, prems, premsScope, m, n, andL, andR
	]
(* Forall, condition *)
analyzeHOLStructure[ TYPED$[ TrueQ$TM, {}][ TYPED$[ Forall$TM, {typ_}][ Lambda$TM[ _, C_]][ Lambda$TM[ rng_, A_]]],
		hist_, loose_, indices_, prems_, premsScope_, m_, n_, andL_, andR_] :=
	analyzeHOLStructure[
		TYPED$[ CoreAll$TM, {typ}][ Lambda$TM[ rng, TYPED$[ CoreImplies$TM, {}][ TYPED$[ TrueQ$TM, {}][ C]][ TYPED$[ TrueQ$TM, {}][ A]]]],
		Append[ hist, 2], loose, indices, prems, premsScope, m, n, andL, andR
	]
analyzeHOLStructure[ TYPED$[ TrueQ$TM, {}][ TYPED$[ Forall$TM, {typ_}][ C_][ Lambda$TM[ rng_, A_]]],
		hist_, loose_, indices_, prems_, premsScope_, m_, n_, andL_, andR_] :=
	analyzeHOLStructure[
		TYPED$[ CoreAll$TM, {typ}][ Lambda$TM[ rng, TYPED$[ CoreImplies$TM, {}][ TYPED$[ TrueQ$TM, {}][ etaExpandOne[ C]]][ TYPED$[ TrueQ$TM, {}][ A]]]],
		Append[ hist, 2], loose, indices, prems, premsScope, m, n, andL, andR
	]
analyzeHOLStructure[ TYPED$[ TrueQ$TM, {}][ TYPED$[ Forall$TM, {typ_}][ Lambda$TM[ rng_, C_]][ A_]],
		hist_, loose_, indices_, prems_, premsScope_, m_, n_, andL_, andR_] :=
	analyzeHOLStructure[
		TYPED$[ CoreAll$TM, {typ}][ Lambda$TM[ rng, TYPED$[ CoreImplies$TM, {}][ TYPED$[ TrueQ$TM, {}][ C]][ TYPED$[ TrueQ$TM, {}][ etaExpandOne[ A]]]]],
		Append[ hist, 2], loose, indices, prems, premsScope, m, n, andL, andR
	]
analyzeHOLStructure[ TYPED$[ TrueQ$TM, {}][ TYPED$[ Forall$TM, {typ_}][ C_][ A_]],
		hist_, loose_, indices_, prems_, premsScope_, m_, n_, andL_, andR_] :=
	analyzeHOLStructure[
		TYPED$[ CoreAll$TM, {typ}][
			Lambda$TM[
				RNG$[ TYPRNG$[ Null, typ]],
				TYPED$[ CoreImplies$TM, {}][ TYPED$[ TrueQ$TM, {}][ etaExpandOne[ C]]][ TYPED$[ TrueQ$TM, {}][ etaExpandOne[ A]]]
			]
		],
		Append[ hist, 2], loose, indices, prems, premsScope, m, n, andL, andR
	]
(* CoreAll *)
analyzeHOLStructure[ TYPED$[ CoreAll$TM, {_}][ Lambda$TM[ RNG$[ tr_TYPRNG$], A_]],
		hist_, loose_, indices_, prems_, premsScope_, m_, n_, andL_, andR_] :=
	analyzeHOLStructure[
		A,
		Append[ hist, -1], Append[ loose, tr], Append[ indices, n], prems, premsScope, m + 1, n, andL, andR
	]
analyzeHOLStructure[ TYPED$[ CoreAll$TM, {_}][ TYPED$[ CoreEqual$TM, {typ_}][ a_]],
		hist_, loose_, indices_, prems_, premsScope_, m_, n_, andL_, andR_] :=
	analyzeHOLStructure[
		TYPED$[ CoreEqual$TM, {FUN$TM[ typ, typ]}][
			Lambda$TM[ RNG$[ TYPRNG$[ Null, typ]], mapFreeVars[ a, VAR$[ #3 + 1]&]]
		][
			Lambda$TM[ RNG$[ TYPRNG$[ Null, typ]], VAR$[ 0]]
		],
		hist, loose, indices, prems, premsScope, m, n, andL, andR
	]
analyzeHOLStructure[ TYPED$[ CoreAll$TM, {typ_}][ TYPED$[ TrueQ$TM, {}]],
		hist_, loose_, indices_, prems_, premsScope_, m_, n_, andL_, andR_] :=
	analyzeHOLStructure[
		TYPED$[ TrueQ$TM, {}][ VAR$[ 0]],
		Append[ hist, -1], Append[ loose, TYPRNG$[ Null, typ]], Append[ indices, n], prems, premsScope, m + 1, n, andL, andR
	]
(* Implies *)
analyzeHOLStructure[ TYPED$[ TrueQ$TM, {}][ TYPED$[ Implies$TM, {}][ A_][ B_]],
		hist_, loose_, indices_, prems_, premsScope_, m_, n_, andL_, andR_] :=
	analyzeHOLStructure[
		TYPED$[ CoreImplies$TM, {}][ TYPED$[ TrueQ$TM, {}][ A]][ TYPED$[ TrueQ$TM, {}][ B]],
		Append[ hist, 3], loose, indices, prems, premsScope, m, n, andL, andR
	]
(* And, left *)
analyzeHOLStructure[ TYPED$[ CoreImplies$TM, {}][ TYPED$[ TrueQ$TM, {}][ TYPED$[ And$TM, {}][ A_][ B_]]][ C_],
		hist_, loose_, indices_, prems_, premsScope_, m_, n_, True, andR_] :=
	analyzeHOLStructure[
		TYPED$[ CoreImplies$TM, {}][ TYPED$[ TrueQ$TM, {}][ A]][ TYPED$[ CoreImplies$TM, {}][ TYPED$[ TrueQ$TM, {}][ B]][ C]],
		Append[ hist, 7], loose, indices, prems, premsScope, m, n, True, andR
	]
(* And, right *)
analyzeHOLStructure[ TYPED$[ TrueQ$TM, {}][ TYPED$[ And$TM, {}][ A_][ B_]],
		hist_, loose_, indices_, prems_, premsScope_, m_, n_, andL_, True] :=
	Join[
		analyzeHOLStructure[
			TYPED$[ TrueQ$TM, {}][ A],
			Append[ hist, 8], loose, indices, prems, premsScope, m, n, andL, True
		],
		analyzeHOLStructure[
			TYPED$[ TrueQ$TM, {}][ B],
			Append[ hist, 9], loose, indices, prems, premsScope, m, n, andL, True
		]
	]
(* CoreImplies *)
analyzeHOLStructure[ TYPED$[ CoreImplies$TM, {}][ A_][ B_],
		hist_, loose_, indices_, prems_, premsScope_, m_, n_, andL_, andR_] :=
	analyzeHOLStructure[
		B,
		Append[ hist, -2], loose, indices, Append[ prems, A], Append[ premsScope, m], m, n + 1, andL, andR
	]
(* atom *)
analyzeHOLStructure[ A_, hist_, loose_, indices_, prems_, premsScope_, _, _, _, _] :=
	{{A, hist, loose, indices, prems, premsScope}}

etaExpandOne[ expr_] :=
	mapFreeVars[ expr, VAR$[ #3 + 1]&][ VAR$[ 0]]

(* Meaning of entries in 'hist':
-1: CoreAll
-2: CoreImplies
 1: Forall, no condition
 2: Forall, condition
 3: Implies
 4: Equal
 5: Not
 6: TrueQ
 7: And, left
 8: And, right, first
 9: And, right, last *)
getHOLCanonicalFormatProof[ i0_Integer, n0_Integer, hist_, allNum_, goodNum_] :=
	Module[ {X, Y, x = -1, y = -1, p = 0, i = i0, n = n0},
		Hold @@ Join[
			(* goal *)
			Table[ INF$[ "allI", FIX$[ X, k]], {k, 0, allNum - 1}],
			If[ goodNum === 0, {}, {INF$[ "ext", goodNum]}],
			Table[ INF$[ "allI", FIX$[ Y, k]], {k, 0, goodNum - 1}],
			
			(* assm *)
			Replace[ hist,
				{
					-1 :> INF$[ "allE", i, makeSTERM[ FIX$[ X, ++x]]],
					{-1} :> INF$[ "allE", i, makeSTERM[ FIX$[ Y, ++y]]],
					-2 :> (++p; Sequence @@ {}),
					h:(8|9) :>
						With[ {i1 = i, p1 = p, n1 = n},
							i += n;
							p = n = 0;
							Sequence @@ Join[
								{obtainConjuncts[ i1, p1, n1]},
								Table[ INF$[ "drop", i1 + n1], {p1}],
								{INF$[ "drop", i1 + n1 + 9 - h]}
							]
						],
					h_ :>
						reflectHOLConstant[
							Switch[ h,
								1, Forall$TM,
								2, {Forall$TM},
								3, Implies$TM,
								4, Equal$TM,
								5, Not$TM,
								6, TrueQ$TM,
								7, And$TM,
								_, $Failed
							],
							Join[ Table[ 1, {i - 1}], {0, 1}, Table[ 1, {p}]]
						]
				},
				{1}
			],
			
			{INF$[ "impE", i, p]}
		]
	]

(* 'obtainConjuncts[ i, p, n]' is an auxiliary proof method applicable to subgoals of the form
	'A1 ==> ... ==> A(i-1) ==> (C1 ==> ... ==> Cp ==> P & Q) ==> B1 ==> ... ==> Bn ==> C1 ==> ... ==> Cp ==> thesis',
	which are replaced by
	'A1 ==> ... ==> A(i-1) ==> B1 ==> ... ==> Bn ==> C1 ==> Cp ==> P ==> Q ==> thesis'.
	Note, however, that the form of the subgoal is not checked! *)
obtainConjuncts[ i_, p_, n_][ PRFSIT$[ {G_, ___}, __], __] :=
	With[ {conj = Extract[ G, Join[ Table[ 1, {i - 1}], {0, 1}, Table[ 1, {p}]]]},
	With[ {thesis = dropAssms[ G, i + p + n]},
	With[ {cmd =
			(* 'A1 ==> ... ==> A(i-1) ==> B1 ==> ... ==> Bn ==> C1 ==> Cp ==> P & Q ==> thesis' *)
			makeFactsExplicitInline[
				{STRING$[ Lookup[ $theoryMapping, "$HOL", $Failed], "\[And] reflect"]},
				1,
				Hold @@ {
					INF$[ "allE", 1, makeSTERM[ conj[[1, 0, 1]]]],
					INF$[ "allE", 1, makeSTERM[ conj[[1, 1]]]],
					INF$[ "allE", 1, makeSTERM[ thesis]],
					INF$[ "replace", 1, Table[ 1, {i + n + p}]],
					INF$[ "drop", 1]
				},
				False
			]},
		If[ !(p === 0 === n) && MatchQ[ cmd, _Hold],
			Join[
				Hold @@ {
					setupSubgoalsMP[ conj, i + p + n],
					INF$[ "impE", i, p],
					INF$[ "drop", i]
				},
				cmd
			],
		(*else*)
			cmd
		]
	]]]


(* ::Section:: *)
(*Reflection and Atomization*)

Theorema`CoreProvers`CoreProofMethods`$reflectionFunction = HOLReflectionFunction;
Theorema`CoreProvers`CoreProofMethods`$atomizationFunction = HOLAtomizationFunction;

(* Note that 'HOLReflectionFunction' does not make use of "TrueQ reflect" and "\[And] reflect". *)
HOLReflectionFunction[ TYPED$[ TrueQ$TM, {}][ TYPED$[ Forall$TM, {_}][ Lambda$TM[ _RNG$, TYPED$[ True, {}]]][ term_]], _, pos_List, sub_List, _] :=
	replaceHOLConstant[ "\[ForAll] reflect", Join[ pos, sub], {term}]
HOLReflectionFunction[ TYPED$[ TrueQ$TM, {}][ TYPED$[ Forall$TM, {_}][ cond_][ term_]], _, pos_List, sub_List, _] :=
	replaceHOLConstant[ "\[ForAll]C reflect", Join[ pos, sub], {cond, term}]
HOLReflectionFunction[ TYPED$[ TrueQ$TM, {}][ TYPED$[ Implies$TM, {}][ A_][ B_]], _, pos_List, sub_List, _] :=
	replaceHOLConstant[ "\[Implies] reflect", Join[ pos, sub], {A, B}]
HOLReflectionFunction[ TYPED$[ TrueQ$TM, {}][ TYPED$[ Not$TM, {}][ A_]], _, pos_List, sub_List, _] :=
	replaceHOLConstant[ "\[Not] reflect", Join[ pos, sub], {A}]
HOLReflectionFunction[ TYPED$[ TrueQ$TM, {}][ TYPED$[ Equal$TM, {_}][ A_][ B_]], _, pos_List, sub_List, _] :=
	replaceHOLConstant[ "\[Equal] reflect", Join[ pos, sub], {A, B}]

HOLAtomizationFunction[ TYPED$[ CoreEqual$TM, {_}][ _][ _], _, pos_List, sub_List, _] :=
	replaceHOLConstant[ "CoreEqual atomize", Join[ pos, sub, {0, 0}], {}]
HOLAtomizationFunction[ TYPED$[ TrueQ$TM, {}][ _], _, _List, _List, _] :=
	Hold[]

(* 'reflectHOLConstant[ const, pos]' is an auxiliary proof method that replaces the given HOL constant 'const' at the
	given position 'pos' by its Core analogue. 'pos' is assumed to refer to a valid position of the first subgoal of
	the current proof situation, where the corresponding subexpression is not in the scope of a binder and of the form
	'TrueQ[ ...]'; this is not checked, though!
	'const' can be one of 'Forall$TM' (no condition), '{Forall$TM}' (condition), 'Implies$TM', 'Not$TM', 'TrueQ$TM',
	'Equal$TM' or 'And$TM'; in the latter case, the conjunction is assumed to appear as the antecedent of a Core
	implication. *)
reflectHOLConstant[ Forall$TM, pos_List][ PRFSIT$[ {G_, ___}, __], __] :=
	replaceHOLConstant[ "\[ForAll] reflect", pos, {{1, 1}}, G]
reflectHOLConstant[ {Forall$TM}, pos_List][ PRFSIT$[ {G_, ___}, __], __] :=
	replaceHOLConstant[ "\[ForAll]C reflect", pos, {{1, 0, 1}, {1, 1}}, G]
reflectHOLConstant[ Implies$TM, pos_List][ PRFSIT$[ {G_, ___}, __], __] :=
	replaceHOLConstant[ "\[Implies] reflect", pos, {{1, 0, 1}, {1, 1}}, G]
reflectHOLConstant[ Not$TM, pos_List][ PRFSIT$[ {G_, ___}, __], __] :=
	replaceHOLConstant[ "\[Not] reflect", pos, {{1, 1}}, G]
reflectHOLConstant[ TrueQ$TM, pos_List][ PRFSIT$[ {G_, ___}, __], __] :=
	replaceHOLConstant[ "TrueQ reflect", pos, {{1}}, G]
reflectHOLConstant[ Equal$TM, pos_List][ PRFSIT$[ {G_, ___}, __], __] :=
	replaceHOLConstant[ "\[Equal] reflect", pos, {{1, 0, 1}, {1, 1}}, G]
reflectHOLConstant[ And$TM, pos_List][ PRFSIT$[ {G_, ___}, __], __] :=
	replaceHOLConstant[ "\[And] reflect", pos, {{0, 1, 1, 0, 1}, {0, 1, 1, 1}, {1}}, G]

(* 'atomizeHOLConstant[ const, pos]' is the inverse operation of 'reflectHOLConstant'. Note that 'pos' only refers to
	the constant itself, without any arguments!
	ATTENTION! Core constants can only be atomized if their type instances do not require type variables to be
	instantiated by types involving 'FORM'. *)
atomizeHOLConstant[ CoreEqual$TM, pos_List][ __] :=
	replaceHOLConstant[ "CoreEqual atomize", pos, {}]

replaceHOLConstant[ fact_String, pos_List, instpos_List, G_] :=
	replaceHOLConstant[ fact, pos, Extract[ G, Join[ pos, #]& /@ instpos]]
replaceHOLConstant[ fact_String, pos_List, inst_List] :=
	makeFactsExplicitInline[
		{STRING$[ Lookup[ $theoryMapping, "$HOL", $Failed], fact]},
		1,
		Hold @@ Join[
			Map[ INF$[ "allE", 1, makeSTERM[ #]]&, inst],
			{INF$[ "replace", 1, Prepend[ pos, 1]], INF$[ "drop", 1]}
		],
		False
	]


(* ::Section:: *)
(*Rewriting*)

Theorema`CoreProvers`CoreProofMethods`$rewritingMaster = HOLRewritingMaster;

SetAttributes[ HOLRewritingMaster, HoldFirst];
HOLRewritingMaster[ _, G_, Null|0, _, hyps_, _, _] :=
	goalIterator[ G, 0, terminalAssmFun[ G], terminalGoalFun[ G, hyps], Continue]
HOLRewritingMaster[ _, G_, n_Integer, _, hyps_, _, _] :=
	goalIterator[ dropAssms[ G, n], n, terminalAssmFun[ G], terminalGoalFun[ G, hyps], Continue]
HOLRewritingMaster[ nForms_, G_, {n_Integer}, newmeta_, hyps_, tfix_, tvar_] :=
	Replace[ dropAssms[ G, n],
		{
			a_META$ :>
				HOLRewritingMaster[ nForms, False, a, TYPED$[ TrueQ$TM, {}][ TYPED$[ True, {}]], n],
			TYPED$[ CoreEqual$TM, {_}][ a_][ b_] :>
				Which[
					MatchQ[ a, _META$] && MemberQ[ newmeta, a -> _],
					Theorema`CoreProvers`CoreProofMethods`defaultRewritingMaster[ nForms, b, a],
				(*else*)
					MatchQ[ b, _META$] && MemberQ[ newmeta, b -> _],
					Theorema`CoreProvers`CoreProofMethods`defaultRewritingMaster[ nForms, a, b],
				(*else*)
					True,
					Replace[ alphaEquiv[ a, b, tfix, tvar],
						{
							{_Association, _List} :>
								Hold[ INF$[ "refl"]],
							_ -> Continue
						}
					]
				],
			TYPED$[ TrueQ$TM, {}][ x_] :>
				Replace[ x,
					{
						a_META$ :>
							(* Instantiate 'a' by 'True', because such goals are typically left when proving conditions
								of certain congruence rules: if they become 'True == X', they are further rewritten into 'X'. *)
							HOLRewritingMaster[ nForms, False, a, TYPED$[ True, {}], n],
						TYPED$[ Not$TM, {}][ a_META$] :>
							HOLRewritingMaster[ nForms, False, a, TYPED$[ False, {}], n],
						TYPED$[ Equal$TM, {_}][ a_][ b_] :>
							Which[
								MatchQ[ a, _META$] && MemberQ[ newmeta, a -> _],
								HOLRewritingMaster[ nForms, True, b, a, n],
							(*else*)
								MatchQ[ b, _META$] && MemberQ[ newmeta, b -> _],
								HOLRewritingMaster[ nForms, True, a, b, n],
							(*else*)
								True,
								Replace[ alphaEquiv[ a, b, tfix, tvar],
									{
										{newTVar_Association, tunif_List} :>
											Replace[ makeReflCmd[ replaceTypes[ a, tunif]],
												{
													cmd_Hold :>
														Join[
															Hold[ INF$[ "sorts", newTVar]],
															If[ tunif === {}, Hold[], Hold[ INF$[ "instT", tunif]]],
															cmd
														],
													_ -> Continue
												}
											],
										_ -> Continue
									}
								]
							],
						_ :>
							Replace[ terminalGoalFun[ G, hyps][ x, n, True],
								Except[ _Hold] -> Continue
							]
					}
				],
			y_ :>
				Replace[ terminalGoalFun[ G, hyps][ y, n, False],
					Except[ _Hold] -> Continue
				]
		}
	]
HOLRewritingMaster[ nForms_, True, t_, X_META$, n_] :=
	If[ t === X,
		Replace[ makeReflCmd[ t],
			Except[ _Hold] -> Continue
		],
	(*else*)
		Theorema`CoreProvers`CoreProofMethods`addToNormalForms[ nForms, t];
		Replace[ makeReflCmd[ t],
			{
				cmd_Hold :>
					Join[
						Hold[ INF$[ "inst", X -> STERM$[ t, Association[], Association[]]]],
						cmd
					],
				_ -> Continue
			}
		]
	]
HOLRewritingMaster[ _, False, X_META$, inst_, n_] :=
	Replace[ makeTrueCmd[ inst === TYPED$[ False, {}]],
		{
			cmd_Hold :>
				Join[
					Hold[ INF$[ "inst", X -> STERM$[ inst, Association[], Association[]]]],
					cmd
				],
			_ -> Continue
		}
	]

makeTrueCmd[ notFalseI_] :=
	makeFactsExplicitInline[ {STRING$[ Lookup[ $theoryMapping, "$HOL", $Failed], If[ TrueQ[ notFalseI], "\[Not]FalseI", "TrueI"]]}, 1, Hold[ INF$[ "impE", 1, 0]], False]

makeFalseCmd[ i_] :=
	With[ {i1 = i + 1},		(* '+1' because of 'TrueQ'. *)
		makeFactsExplicitInline[ {STRING$[ Lookup[ $theoryMapping, "$HOL", $Failed], "FalseE"]}, 1, Hold[ instantiateCoreAllGoal[ 1, i1], INF$[ "impE", 1, 1]], False]
	]

makeReflCmd[ t_] :=
	makeFactsExplicitInline[ {STRING$[ Lookup[ $theoryMapping, "$HOL", $Failed], "\[Equal] refl"]}, 1, Hold[ INF$[ "allE", 1, makeSTERM[ t]], INF$[ "impE", 1, 0]], False]


(* ::Section:: *)
(*Propositional Reasoning*)

(* The proof methods implemented in this section are of interest mainly for other (semi-)automatic proof tools. *)

(* 'proveDisjunct[ i, n][ ps, ___]' proves the 'i'-th disjunct in the first subgoal of 'ps', assuming that there are in
	total 'n' disjuncts (which are aligned to the right). If 'n' is '1', the goal does not need to be a disjunction.
	In any case, the disjunction is proved constructively, i.e. without assuming the negations of the other disjuncts.
	Examples:
	- With 'i=2', 'n=3' and the goal being 'A || (B || C)', 'B' is proved.
	- With 'i=2', 'n=2' and the goal being again 'A || (B || C)', 'B || C' is proved. That's the reason why 'n' must be
		passed as a parameter. *)
proveDisjunct[ i_Integer?Positive, n_Integer?Positive][ PRFSIT$[ {G_, ___}, __], ___] :=
	Which[
		n < i,
		$Failed,
	(*else*)
		n === 1,
		Hold[],
	(*else*)
		True,
		With[ {disjI1Lbl = "\[Or]I1", disjI2Lbl = "\[Or]I2", holThy = Lookup[ $theoryMapping, "$HOL", $Failed]},
		With[ {disjI1Key = If[ n > i, fastKeyFromLabel[ {holThy, disjI1Lbl}], Null],
				disjI2Key = If[ i > 1, fastKeyFromLabel[ {holThy, disjI2Lbl}], Null]},
			Which[
				disjI1Key === $Failed,
				sowHOLMessage[ 11, {STRING$[ disjI1Lbl]}, "Error"];
				$Failed,
			(*else*)
				disjI2Key === $Failed,
				sowHOLMessage[ 11, {STRING$[ disjI2Lbl]}, "Error"];
				$Failed,
			(*else*)
				True,
				Module[ {disj = First[ stripAssms[ G]]},
					Hold @@ Join[
						If[ i > 1,
							Join[
								{INF$[ "insert", disjI2Key]},
								Join @@ Table[
									{
										INF$[ "makeExplicit", 1],
										INF$[ "allE", 1, makeSTERM[ disj[[0, 1]]]],
										INF$[ "allE", 1, makeSTERM[ disj = First[ disj]]],
										INF$[ "impE", 1, 0]
									},
									{i - 1}
								],
								{INF$[ "forget", 1]}
							],
						(*else*)
							{}
						],
						If[ i === n,
							{},
						(*else*)
							{
								INF$[ "formula", disjI1Key],
								INF$[ "allE", 1, makeSTERM[ disj[[0, 1]]]],
								INF$[ "allE", 1, makeSTERM[ First[ disj]]],
								INF$[ "impE", 1, 0]
							}
						]
					]
				]
			]
		]]
	]

(* 'proveConjunction[ cmds][ ps, ___]' proves the first 'Length[ cmds]' conjuncts of the first subgoal of 'ps'; the
	conjuncts must be aligned to the right. Each conjunct is proved using the respective proof given in 'cmds', and
	each such proof must close the corresponding subgoal without closing other subgoals or introducing new ones. *)
proveConjunction[ {}][ ___] := Hold[]
proveConjunction[ {cmd_Hold}][ ___] := cmd
proveConjunction[ cmds_List][ PRFSIT$[ {G_, ___}, __], ___] :=
	With[ {conjILbl = "\[And]I", holThy = Lookup[ $theoryMapping, "$HOL", $Failed]},
	With[ {conjIKey = fastKeyFromLabel[ {holThy, conjILbl}]},
		If[ conjIKey === $Failed,
			sowHOLMessage[ 11, {STRING$[ conjILbl]}, "Error"];
			$Failed,
		(*else*)
			Module[ {conj = First[ stripAssms[ G]]},
				Join[
					Hold @@ Join @@ (
						Join[
							Hold @@ {
								INF$[ "formula", conjIKey],
								INF$[ "allE", 1, makeSTERM[ conj[[0, 1]]]],
								INF$[ "allE", 1, makeSTERM[ conj = First[ conj]]],
								INF$[ "impE", 1, 0]
							},
							#
						]& /@ Most[ cmds]
					),
					Last[ cmds]
				]
			]
		]
	]]

(* TODO: Implement corresponding proof methods for elimination/destruction. *)


(* ::Section:: *)
(*Existential Quantifiers*)

(* 'introduceExists[ inst][ ps, ___]' instantiates the existentially quantified goal in the first subgoal of 'ps' by
	'inst'. 'inst' is a supposed to be a list whose elements are
	- 'Automatic' or "$", in which case a the quantifier is instantiated by a fresh meta variable, or
	- an arbitrary term, possibly inside 'EXPR$'; if it is inside 'EXPR$', types are inferred, otherwise it is simply
	 	type-checked.
	The last element of 'inst' may be one of the above possibilities wrapped inside 'RepeatedNull', in which case
	all remaining quantifiers are instantiated in the respective way.
	Note that quantifiers cannot be skipped! *)
introduceExists[ {}][ ___] := Hold[]
introduceExists[ inst_List][ ps:PRFSIT$[ {G_, ___}, __], ___, mode_] :=
	Replace[ stripAssms[ G, 0],
		{
			{TYPED$[ TrueQ$TM, {}][ A_], n_Integer} :>
				Replace[ makeInstantiation[ inst, A, Exists$TM, True, False, ps, mode],
					{
						{inst1_List, _?Negative, newIdx_Integer, newMeta_List, newTFix_Association, newTVar_Association, abbrevs_List, tsubst_List} :>
							With[ {cmd =
									If[ inst1 === {},
										{},
									(*else*)
										With[ {trivList = Flatten[ Last[ Reap[ trivConds[ A, Exists$TM, Length[ inst1]], 0]]]},
										With[ {hasTriv = MemberQ[ trivList, True], hasNontriv = MemberQ[ trivList, False],
												exILbl = "\[Exists]I", exCILbl = "\[Exists]CI",
												holThy = Lookup[ $theoryMapping, "$HOL", $Failed]},
										With[ {exIKey = If[ hasTriv, fastKeyFromLabel[ {holThy, exILbl}], Null],
												exCIKey = If[ hasNontriv, fastKeyFromLabel[ {holThy, exCILbl}], Null]},
											Which[
												exIKey === $Failed,
												sowHOLMessage[ 11, {exILbl}, "Fatal"];
												$Failed,
											(*else*)
												exCIKey === $Failed,
												sowHOLMessage[ 11, {exCILbl}, "Fatal"];
												$Failed,
											(*else*)
												True,
												Join[
													If[ ListQ[ exIKey], {INF$[ "insert", exIKey]}, {}],
													If[ ListQ[ exCIKey], {INF$[ "insert", exCIKey]}, {}],
													{introduceExists[ inst1, trivList, Table[ 1, {n + 1}], newTFix, Which[ !hasTriv, 0, hasNontriv, 2, True, 1]]}
												]
											]
										]]]
									]},
								If[ ListQ[ cmd],
									Hold @@ Join[
										If[ tsubst === {}, {}, {INF$[ "instT", tsubst]}],
										Replace[ newMeta, (m_ -> {t_, d_}) :> INF$[ "meta", m, t, d], {1}],
										If[ abbrevs === {}, {}, {updateAbbrevLists[ {}, abbrevs, {}, {}]}],
										{INF$[ "sorts", newTVar]},
										cmd
									],
								(*else*)
									$Failed
								]
							],
						_ -> $Failed
					}
				],
			_ :> If[ inst === {} || MatchQ[ inst, {h_[ _] /; h === RepeatedNull}], Hold[], sowHOLMessage[ 1011, "Fatal"]; $Failed]
		}
	]
introduceExists[ {term_, terms___}, {triv_, trivs___}, pos_List, newTFix_, exIInd_][ PRFSIT$[ {G_, ___}, __], ___] :=
	With[ {q = Extract[ G, pos]},
		Hold @@ Join[
			{
				INF$[ "makeExplicit", If[ TrueQ[ triv], exIInd, 1]],
				INF$[ "allE", 1, makeSTERM[ First[ q]]]
			},
			If[ TrueQ[ triv], {}, {INF$[ "allE", 1, makeSTERM[ q[[0, 1]]]]}],
			{
				INF$[ "allE", 1, STERM$[ term, newTFix, Association[]]],
				INF$[ "impE", 1, 0],
				introduceExists[ {terms}, {trivs}, pos, newTFix, exIInd]
			}
		]
	]
introduceExists[ {}, _, _, _, exIInd_][ ___] :=
	Hold @@ Table[ INF$[ "forget", 1], {Max[ 1, exIInd]}]

(* 'eliminateExists[ fixes, i][ ps, ___]' eliminates the outermost 'Length[ fixes]' existential quantifiers in the
	'i'-th assumption of the first subgoal of 'ps' by introducing the a.b.f. constants 'fixes'. 'fixes' is a supposed
	to be a list whose elements are pairwise distinct, fresh 'FIX$'-structures (possibly inside 'EXPR$'); this is not
	checked, though! 'i' may of course be negative. Note that the instantiated formula may not be the 'i'-th assumption
	any more, because each quantifier with non-trivial condition adds the instance of the condition *before* the
	instance of the main body.
	The proof method fails if too many constants are provided. *)
eliminateExists[ {}, _Integer][ ___] := Hold[]
eliminateExists[ fixes0_List, i0_Integer][ ps:PRFSIT$[ {G_, ___}, __], ___, mode_] :=
	With[ {fixes = Replace[ fixes0, EXPR$[ c_, ___] :> c, {1}], assmNum = assmCount[ G, 0, Infinity]},
	With[ {i = If[ Negative[ i0], i0 + assmNum, i0]},
		Which[
			i < 1 || assmNum < i,
			sowHOLMessage[ 1012, {i0}, "Fatal"];
			$Failed,
		(*else*)
			MemberQ[ fixes, Except[ _FIX$]],
			sowHOLMessage[ 1013, "Fatal"];
			$Failed,
		(*else*)
			True,
			With[ {assmPos = assmPosition[ i]},
				Replace[ Extract[ G, assmPos],
					{
						TYPED$[ TrueQ$TM, {}][ A_] :>
							Replace[ Reap[ trivConds[ A, Exists$TM, Length[ fixes]], 0],
								{
									{True, trivList0_List} :>
										With[ {trivList = Flatten[ trivList0]},
										With[ {hasTriv = MemberQ[ trivList, True], hasNontriv = MemberQ[ trivList, False],
												exELbl = "\[Exists]E", exCELbl = "\[Exists]CE",
												holThy = Lookup[ $theoryMapping, "$HOL", $Failed]},
										With[ {exEKey = If[ hasTriv, fastKeyFromLabel[ {holThy, exELbl}], Null],
												exCEKey = If[ hasNontriv, fastKeyFromLabel[ {holThy, exCELbl}], Null]},
											Which[
												exEKey === $Failed,
												sowHOLMessage[ 11, {exELbl}, "Fatal"];
												$Failed,
											(*else*)
												exCEKey === $Failed,
												sowHOLMessage[ 11, {exCELbl}, "Fatal"];
												$Failed,
											(*else*)
												True,
												Hold @@ Join[
													If[ ListQ[ exEKey], {INF$[ "insert", exEKey]}, {}],
													If[ ListQ[ exCEKey], {INF$[ "insert", exCEKey]}, {}],
													{
														eliminateExists[
															fixes,
															trivList,
															i,
															assmNum,
															Append[ assmPos, 1],	(* because of 'TrueQ' *)
															extractSingle[ G, Table[ 1, {assmNum + 1}]],	(* 'assmNum + 1' because of 'TrueQ' *)
															Which[ !hasTriv, 0, hasNontriv, 2, True, 1]
														]
													}
												]
											]
										]]],
									_ :> (sowHOLMessage[ 1014, "Fatal"]; $Failed)
								}
							],
						_ :> (sowHOLMessage[ 1015, "Fatal"]; $Failed)
					}
				]
			]
		]
	]]
eliminateExists[ {const_FIX$, consts___}, {triv_, trivs___}, i_Integer, assmNum_Integer, pos_List, R_, exEInd_Integer][ PRFSIT$[ {G_, ___}, __], ___] :=
	With[ {q = Extract[ G, pos]},
		Hold @@ Join[
			{INF$[ "makeExplicit", If[ TrueQ[ triv], exEInd, 1]]},
			If[ TrueQ[ triv], {}, {INF$[ "allE", 1, makeSTERM[ q[[0, 1]]]]}],
			{
				INF$[ "allE", 1, makeSTERM[ First[ q]]],
				INF$[ "allE", 1, makeSTERM[ R]],
				INF$[ "impE", 1, 0],
				INF$[ "impE", i, 0],
				INF$[ "drop", i],
				INF$[ "allI", const]
			},
			If[ i < assmNum,
				Table[ Theorema`CoreProvers`rearrangeAssms[ -1, i], {If[ TrueQ[ triv], 1, 2]}],
			(*else*)
				{}
			],
			{
				If[ TrueQ[ triv],
					eliminateExists[ {consts}, {trivs}, i, assmNum, pos, R, exEInd],
				(*else*)
					eliminateExists[ {consts}, {trivs}, i + 1, assmNum + 1, Join[ Drop[ pos, -3], {1, 0, 1, 1}], R, exEInd]
				]
			}
		]
	]
eliminateExists[ {}, _, _, _, _, _, exEInd_][ ___] :=
	Hold @@ Table[ INF$[ "forget", 1], {Max[ 1, exEInd]}]

trivConds[ Lambda$TM[ _, A_], head_, n_?Positive] :=
	trivConds[ A, head, n]
trivConds[ TYPED$[ head_, _][ cond_][ A_], head_, n_?Positive] :=
	(Sow[ MatchQ[ cond, Lambda$TM[ _, TYPED$[ True, {}]]], 0]; trivConds[ A, head, n - 1])
trivConds[ TYPED$[ head_, _][ cond_], head_, _?Positive] :=
	(Sow[ MatchQ[ cond, Lambda$TM[ _, TYPED$[ True, {}]]], 0]; True)
trivConds[ _, _, n_] := (n === 0 || !IntegerQ[ n])


(* ::Section:: *)
(*Default Method*)

defaultHOLMethod[][ ps_PRFSIT$, __] :=
	Hold[]

End[]

EndPackage[]
