(* ::Package:: *)

(* Theorema
    Copyright (C) 2017 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

BeginPackage[ "Theorema`HOL`"]

Needs[ "Theorema`Common`"]
Needs[ "Theorema`Language`"]
Needs[ "Theorema`Core`Trusted`"]
Needs[ "Theorema`Core`Tools`"]

With[ {lang = "English"},
	MessageName[ parseWhereConstruct, "usage", lang] = "";
	MessageName[ processWhereConstruct, "usage", lang] = "";
];

Begin["`Inductive`"]	(* Instead of "Private" we call the local context "Inductive", but it has the same meaning. *)

(* ::Section:: *)
(*Theory Commands*)

(* Further definitions may be added incrementally. *)
processWhereConstruct[ "inductive"] = defineInductive;

defineInductive[ consts0_List, intros0_List, opts_List][ False, i_, id_, src_, cell_CellObject, nb_NotebookObject] :=
	Replace[
		Reap[
			If[ Length[ consts0] === 1,
				(* Although 'consts0' must contain only one single element, we formulate much of the code below more
					generally for arbitrary lists. *)
				Replace[ {parseWhereConstruct[ consts0, intros0, id, src, "inductive"], parseInductiveOpts[ opts]},
					{{consts_List, intros1_List, fix_List, params0_List, tvar_Association, declPos_List, introsData_List}, {monos_List, prefix0_}} :>
						With[ {flatTypes = Composition[ flattenTypes, Last] /@ fix, paramNum = Length[ params0]},
						With[ {params = getIndPredParams[ intros1, fix, params0, paramNum]},
							Which[
								!ListQ[ params],
								sowHOLMessage[ 109, {}, "Fatal"];
								$Failed,
							(*else*)
								checkListSowMessage[ flatTypes, Last[ #] === Theorema`Knowledge`BOOL$TM &, First, 110, consts, "Fatal"],
								$Failed,
							(*else*)
								checkListSowMessage[ intros1, checkIndPredOccs[ fix], All, 111, {}, "Fatal"],
								$Failed,
							(*else*)
								True,
								With[ {intros = canonicalizeRule /@ intros1},
									Which[
										checkListSowMessage[
											intros,
											MatchQ[ First[ #], TYPED$[ TrueQ$TM, {}][ x_] /; MemberQ[ fix, getHeadAtom[ HoldComplete[ x], 0][[1, 1]] -> _]] &,
											First, 112, {}, "Fatal"
										],
										$Failed,
									(*else*)
										checkListSowMessage[ intros, Function[ rl, AllTrue[ Last[ rl], MatchQ[ #, TYPED$[ TrueQ$TM, {}][ _]] &]], First, 113, {}, "Fatal"],
										$Failed,
									(*else*)
										True,
										If[ Length[ declPos] < Length[ consts],
											sowHOLMessage[ 104,
												MapThread[ {#1, Last[ #2]}&, {Delete[ consts, List /@ declPos], Delete[ fix, List /@ declPos]}],
												"Info",
												"inductive"
											]
										];
										
										(* From here on we rely on the fact that there is only one constant. *)
										With[ {predTypOrig = fix[[1, 2]], pred = First[ consts], introNum = Length[ intros]},
										With[ {predTyp = stripArgumentTypes[ predTypOrig, paramNum]},
										With[ {srcLabel = sourceLabel[ src], prefix = If[ StringQ[ prefix0], prefix0, tmaSymbolToString[ pred]]},
											If[ (declPos === {} ||
													Last[ declareConstantsDirect[
														{STERM$[ TYPED$[ pred, predTypOrig], Association[], tvar]},
														cellIDLabel[ {id, i, 1}],
														srcLabel,
														True
													]] === {}),
											(*then*)
												With[ {argTyps = Most[ flattenTypes[ predTyp]]},
												With[ {disj =
														If[ intros === {},
															TYPED$[ False, {}],
														(*else*)
															Fold[
																TYPED$[ Or$TM, {}][ #2][ #1]&,
																conjFromIntro[ fix[[1, 1]], argTyps][ Last[ intros]],
																conjFromIntro[ fix[[1, 1]], argTyps] /@ Reverse[ Most[ intros]]
															]
														]},
												With[ {fun0 = replaceConstsByVars[ Fold[ Lambda$TM[ RNG$[ TYPRNG$[ Null, #2]], #1]&, disj, argTyps], Join[ params, fix], 0],
														predTypInst = Last[ getTypeInstance[ pred, predTypOrig, Association[], tvar]]},
												With[ {predTyped = TYPED$[ pred, predTypInst],
														fun1 = Lambda$TM[ RNG$[ TYPRNG$[ If[ MatchQ[ pred, _Symbol], pred, Null], predTyp]], fun0],
														argNum = Length[ argTyps]},
													
													undefineInductiveDirect[ pred, prefix, False, i, id, srcLabel];
														(* To remove old formulas from the context, because in the current invocation of this
															command there might be fewer introduction rules than in previous invocations. *)
														
													If[ TrueQ[ addAxmToTheory[
															FML$[
																{cellIDLabel[ {id, i, 2}], srcLabel},
																TYPED$[ CoreEqual$TM, {predTypOrig}][ predTyped][
																	bindParams[ TYPED$[ Theorema`Knowledge`LFP$TM, {predTyp}][ fun1], params, Lambda$TM]
																],
																prefix <> " def",
																"pos" -> getEvaluationPosition[ id, nb]
															],
															tvar, True, False, True, False, False
														]] &&
														TrueQ[ addThmToTheory[
															FML$[
																{cellIDLabel[ {id, i, 3}], srcLabel},
																bindParams[
																	TYPED$[ TrueQ$TM, {}][ TYPED$[ Theorema`Knowledge`isMono$TM, {predTyp}][ fun1]],
																	params,
																	CoreAll$TM
																],
																prefix <> " mono",
																"pos" -> Append[ getEvaluationPosition[ id, nb], 0]
															],
															tvar, Hold[], False, False, False, False, False	(* TODO: Prove. *)
														]] &&
														TrueQ[ addThmToTheory[
															FML$[
																{cellIDLabel[ {id, i, 4}], srcLabel},
																TYPED$[ CoreEqual$TM, {predTypOrig}][ predTyped][
																	bindParams[
																		mapFreeVars[ fun0,
																			If[ #3 === #2,
																				applyCurried[ predTyped, Table[ VAR$[ j], {j, #3 + paramNum - 1, #3, -1}], {}],
																			(*else*)
																				VAR$[ #3 - 1]
																			]&
																		],
																		params,
																		Lambda$TM
																	]
																],
																prefix <> " simp",
																"pos" -> Append[ getEvaluationPosition[ id, nb], introNum + 1]
															],
															tvar,
															indSimpProof[
																paramNum,
																argNum,
																predTyped,
																bindParams[ fun1, params, Lambda$TM],
																i,
																id,
																srcLabel
															],
															False, False, False, True, False
														]] &&
														AllTrue[
															MapThread[
																Function[ {rl, d, j},
																	{
																		FML$[
																			{cellIDLabel[ {id, i, j + 4}], srcLabel},
																			bindParams[
																				replaceConstsByVars[
																					Theorema`CoreProvers`CoreProofMethods`reconstructFormula[
																						fromAssmsGoal[ rl[[3]], rl[[1]]],
																						rl[[2]],
																						{},
																						{}
																					] /. fix[[1, 1]] -> predTyped,
																					params,
																					0
																				],
																				params,
																				CoreAll$TM
																			],
																			Replace[ First[ d], Automatic :> (prefix <> " intro " <> ToString[ j])],
																			"pos" -> Append[ getEvaluationPosition[ id, nb], j + 1],
																			Sequence @@ Rest[ d]
																		],
																		indIntroProof[ paramNum, argNum, Length[ rl[[2]]], Length[ rl[[3]]], j, introNum, i, id, srcLabel]
																	}
																],
																{intros, introsData, Range[ introNum]}
															],
															TrueQ[ addThmToTheory[ First[ #], tvar, Last[ #], False, False, False, True, False]]&
														] &&
														TrueQ[ addThmToTheory[
															FML$[
																{cellIDLabel[ {id, i, introNum + 5}], srcLabel},
																bindParams[
																	replaceConstsByVars[
																		Fold[
																			TYPED$[ CoreAll$TM, {#2}][ Lambda$TM[ RNG$[ TYPRNG$[ Null, #2]], #1]]&,
																			TYPED$[ CoreAll$TM, {Theorema`Knowledge`BOOL$TM}][
																				Lambda$TM[
																					RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$P$TM, Theorema`Knowledge`BOOL$TM]],
																					fromAssmsGoal[
																						Prepend[
																							elimFromIntro[ fix[[1, 1]], predTyped, argTyps] /@ intros,
																							TYPED$[ TrueQ$TM, {}][
																								applyCurried[ predTyped, Table[ VAR$[ j], {j, paramNum + Length[ argTyps], 1, -1}], {}]
																							]
																						],
																						TYPED$[ TrueQ$TM, {}][ VAR$[ 0]]
																					]
																				]
																			],
																			Reverse[ argTyps]
																		],
																		params,
																		0
																	],
																	params,
																	CoreAll$TM
																],
																prefix <> " cases",
																"pos" -> Append[ getEvaluationPosition[ id, nb], introNum + 2]
															],
															tvar, Hold[], False, False, False, True, False
															(* TODO: Prove. When proving this theorem, rearrange the explicit assumptions first so that the inductive predicate
																that is unfolded becomes the last assumption (speeds up elimination of disjunctions and existentials). *)
														]] &&
														TrueQ[ addThmToTheory[
															FML$[
																{cellIDLabel[ {id, i, introNum + 6}], srcLabel},
																bindParams[
																	replaceConstsByVars[
																		Fold[
																			TYPED$[ CoreAll$TM, {#2}][ Lambda$TM[ RNG$[ TYPRNG$[ Null, #2]], #1]]&,
																			TYPED$[ CoreAll$TM, {predTyp}][
																				Lambda$TM[
																					RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$P$TM, predTyp]],
																					fromAssmsGoal[
																						Prepend[
																							inductFromIntro[ fix[[1, 1]], predTyped, argTyps] /@ intros,
																							TYPED$[ TrueQ$TM, {}][
																								applyCurried[ predTyped, Table[ VAR$[ j], {j, paramNum + Length[ argTyps], 1, -1}], {}]
																							]
																						],
																						TYPED$[ TrueQ$TM, {}][ applyCurried[ VAR$[ 0], Table[ VAR$[ j], {j, Length[ argTyps], 1, -1}], {}]]
																					]
																				]
																			],
																			Reverse[ argTyps]
																		],
																		params,
																		0
																	],
																	params,
																	CoreAll$TM
																],
																prefix <> " induct",
																"pos" -> Append[ getEvaluationPosition[ id, nb], introNum + 3]
															],
															tvar, Hold[], False, False, False, True, False	(* TODO: Prove. *)
														]],
													(*then*)
														addFactsToList[
															prefix <> " intros",
															Table[ {cellIDLabel[ {id, i, j + 4}], srcLabel}, {j, 1, introNum}],
															{cellIDLabel[ {id, i, introNum + 7}], srcLabel},
															MapAt[ (# - 1)&, getEvaluationPosition[ id, nb], -1],
																(* The position of the list itself must be *before* that of the cell,
																	otherwise it does not get deleted when undoing this command. *)
															Append[ getEvaluationPosition[ id, nb], introNum + 4],
															True
														];
														Null,
													(*else*)
														undefineInductiveDirect[ pred, prefix, declPos =!= {}, i, id, srcLabel];
														$Failed
													]
												]]]],
											(*else*)
												undefineInductiveDirect[ pred, prefix, declPos =!= {}, i, id, srcLabel];
												$Failed
											]
										]]]
									]
								]
							]
						]]
				],
			(*else*)
				sowHOLMessage[ 108, {}, "Fatal"];
				$Failed
			],
			"Message"
		],
		{out_, msg_List} :> (coreNotification[ msg, False, {}, {src, id}]; out)
	]
defineInductive[ consts0_List, _List, opts_List][ True, i_, id_, src_, _CellObject, _NotebookObject] :=
	Replace[
		Reap[
			If[ Length[ consts0] === 1,
				With[ {consts = First[ extractFreshConstant[ #, False, True, "inductive"]]& /@ consts0},
					Replace[ parseInductiveOpts[ opts],
						{_, prefix_} :>
							With[ {pred = First[ consts]},
							With[ {ky = Lookup[ getConstDecls[], pred, {$Failed, $Failed}][[2]]},
								undefineInductiveDirect[
									pred,
									If[ prefix === Automatic, tmaSymbolToString[ pred], prefix],
									!ListQ[ ky] || (sameCellQ[ id, src, ky] && AllTrue[ getConstDefs[ False][ pred], sameCellQ[ id, src, #[[2, 1]]]&]),
									i,
									id,
									sourceLabel[ src]
								];
								Null
							]]
					]
				],
			(*else*)
				sowHOLMessage[ 108, {}, "Fatal"];
				$Failed
			],
			"Message"
		],
		{out_, msg_List} :> (coreNotification[ msg, False, {}, {src, id}]; out)
	]

parseInductiveOpts[ opts_List] :=
	Replace[ parseOptsAsAssociation[ "prefix"|"monos", Rest] @@ opts,
		{
			assoc_Association :>
				With[ {monos = Lookup[ assoc, "monos", {}]},
					Replace[ Lookup[ assoc, "prefix", {"$"}],
						{
							{"$"|"_"} -> {monos, Automatic},
							{STRING$[ s_String]} :> {monos, s},
							_ :> (sowHOLMessage[ 105, {}, "Fatal"]; $Failed)
						}
					]
				],
			Missing[ k_] :> (sowHOLMessage[ 106, {k}, "Fatal"]; $Failed),
			_ :> (sowHOLMessage[ 107, {}, "Fatal"]; $Failed)
		}
	]

undefineInductiveDirect[ pred_, prefix_String, undecl_, i_Integer, id_Integer, srcLabel_String] :=
	With[ {idPatt = cellIDLabel[ {id, i}] ~~ __,
			keys = getFactListKeys[ prefix <> " intros", Infinity]},
		remFormsFromTheory[ {{_String?(StringMatchQ[ #, idPatt]&), srcLabel}}];
		(* 'remFormsFromTheory' takes care of removing the introduction rules from the fact list. *)
		KeyDropFrom[ $globalFactLists, Select[ keys, (Lookup[ $globalFactLists, Key[ #], {}] === {})&]];
		If[ TrueQ[ undecl] && ListQ[ getConstDecls[ False][ pred]], undeclareConstantsDirect[ {pred}]]
	]

(* 'canonicalizeRule[ rule]' simply moves all 'CoreAll' in 'rule' to the front and returns the triple
	'{consequent, loose, prems}', where 'loose' is the list of 'TYPRNG$'-es of the universally quantified variables. *)
canonicalizeRule[ rule_] :=
	Replace[ Reap[ Theorema`CoreProvers`CoreProofMethods`analyzeStructure[ rule, 0, 0, Infinity, True], {0, 1, 2, 3}],
		{{A_, _}, {loose_, indices_, prems_, premsScope_}} :>
			Most[ Theorema`CoreProvers`CoreProofMethods`univToLambdaAdjustDB[ False, A, Join @@ loose, Join @@ indices, Join @@ prems, Join @@ premsScope]]
	]

(* 'getIndPredParams[ term, fix, params]' checks whether all occurrences of 'fix' in 'term' are applied to the same
	permutation of 'params'. In case of success, it returns this permutation of 'params' if there is at least one
	occurrence of 'fix', or 'True' if there is no occurrence of 'fix'. In case of failure, it returns '$Failed'. *)
getIndPredParams[ term_, fix_, params_] :=
	getIndPredParams[ term, fix, params, Length[ params]]
getIndPredParams[ _, _, {}, _] := {}
getIndPredParams[ _, {}, _, _] := True
getIndPredParams[ term_, fix_List, params_List, paramNum_Integer] :=
	With[ {pos = Position[ term, Alternatives @@ fix[[All, 1]], {0, Infinity}, Heads -> True]},
		Which[
			pos === {},
			True,
		(*else*)
			AllTrue[ pos, (Length[ #] >= paramNum && MatchQ[ Take[ #, -paramNum], {0...}])&],
			With[ {args = Last[ splitHeadArgs[ extractSingle[ term, Drop[ First[ pos], -paramNum]]]]},
				If[ DuplicateFreeQ[ args] && AllTrue[ args, MemberQ[ params, # -> _]&],
					With[ {patt = applyCurried[ Blank[], args, {}]},
						If[ TrueQ[ AllTrue[ Rest[ pos], MatchQ[ extractSingle[ term, Drop[ #, -paramNum]], patt]&]],
							Thread[ args -> Replace[ args, params, {1}]],
						(*else*)
							$Failed
						]
					],
				(*else*)
					$Failed
				]
			],
		(*else*)
			True,
			$Failed
		]
	]

(* 'checkIndPredOccs[ fix][ term]' checks whether no element of 'fix' occurs in an argument of an element of 'fix'. *)
checkIndPredOccs[ {}][ _] := True
checkIndPredOccs[ fix_List][ term_] :=
	With[ {pos = Position[ term, Alternatives @@ fix[[All, 1]], {0, Infinity}, Heads -> True]},
		AllTrue[
			Cases[ pos, {___, 0}],
			Function[ p0,
				With[ {p = Append[ deleteTrailing[ p0, 0], BlankNullSequence[]]},
					!MemberQ[ pos, p?(# =!= p0 &)]
				]
			]
		]
	]

conjFromIntro[ fix_FIX$, typs_List][ {TYPED$[ TrueQ$TM, {}][ A_], loose_List, prems_List}] :=
	With[ {typNum = Length[ typs], looseNum = Length[ loose], args0 = Last[ splitHeadArgs[ A]]},
	With[ {args = Take[ args0, -typNum]},
	With[ {newPrems =
			Join[
				MapThread[ TYPED$[ Equal$TM, {#2}][ VAR$[ #3]][ #1]&, {args, typs, Range[ looseNum + typNum - 1, looseNum, -1]}],
				First /@ prems
			]},
	With[ {conj =
			If[ newPrems === {},
				TYPED$[ True, {}],
			(*else*)
				Fold[ TYPED$[ And$TM, {}][ #2][ #1]&, Last[ newPrems], Reverse[ Most[ newPrems]]] /. applyCurried[ fix, Table[ Blank[], {Length[ args0] - typNum}], {}] -> fix
			]},
		Fold[ TYPED$[ Exists$TM, {Last[ #2]}][ Lambda$TM[ RNG$[ #2], TYPED$[ True, {}]]][ Lambda$TM[ RNG$[ #2], #1]]&, conj, Reverse[ loose]]
	]]]]

elimFromIntro[ fix_FIX$, predTyped_, typs_List][ {TYPED$[ TrueQ$TM, {}][ A_], loose_List, prems_List}] :=
	With[ {typNum = Length[ typs], looseNum = Length[ loose]},
	With[ {args = Take[ Last[ splitHeadArgs[ A]], -typNum]},
	With[ {newPrems =
			Join[
				MapThread[ TYPED$[ TrueQ$TM, {}][ TYPED$[ Equal$TM, {#2}][ VAR$[ #3]][ #1]]&, {args, typs, Range[ looseNum + typNum, looseNum + 1, -1]}],
																							(* The index must be increased by '1', because of the bound variable 'R::BOOL'. *)
				prems
			]},
	With[ {imp = fromAssmsGoal[ newPrems /. fix -> predTyped, TYPED$[ TrueQ$TM, {}][ VAR$[ looseNum]]]},
		Fold[ TYPED$[ CoreAll$TM, {Last[ #2]}][ Lambda$TM[ RNG$[ #2], #1]]&, imp, Reverse[ loose]]
	]]]]

(* Replace every occurrence of 'fix[ params][ args]' in each premise by 'pred[ params][ args] && R[ args]',
	where 'R' is a universally quantified variable. Special case: if 'fix[ params][ args]' occurs on the outermost
	level, replace it by the two premises 'pred[ params][ args]' and 'R[ args]'.
	Also replace the atomic consequent 'fix[ params][ args]' by 'R[ args]'. *)
inductFromIntro[ fix_FIX$, predTyped_, typs_List][ {TYPED$[ TrueQ$TM, {}][ A_], loose_List, prems_List}] :=
	With[ {typNum = Length[ typs], looseNum = Length[ loose]},
	With[ {paramNum = Length[ Last[ splitHeadArgs[ A]]] - typNum, zeros1 = Table[ 0, {typNum}]},
	With[ {zeros2 = Join[ zeros1, Table[ 0, {paramNum}]]},
	With[ {newPrems =
			Map[
				Function[ prem,
					If[ getHeadAtom[ HoldComplete @@ {First[ prem]}, 0][[1, 1]] === fix,
						Sequence @@ {
							ReplacePart[ prem, Prepend[ zeros2, 1] -> predTyped],
							ReplacePart[ prem, Prepend[ zeros1, 1] -> VAR$[ looseNum]]
						},
					(*else*)
						MapAt[
							Function[ term,
								With[ {argNum1 = Length[ Last[ splitHeadArgs[ term]]] - paramNum},
								With[ {term1 =
										If[ argNum1 === typNum,
											term,
										(*else*)
											(* 'term' has less than 'typNum' arguments --> eta expand *)
											applyCurried[ mapFreeVars[ term, VAR$[ #3 + typNum - argNum1]&], Table[ VAR$[ j], {j, typNum - argNum1 - 1, 0, -1}], {}]
										]},
								With[ {conj = TYPED$[ And$TM, {}][ ReplacePart[ term1, zeros2 -> predTyped]][ ReplacePart[ term1, zeros1 -> VAR$[ looseNum + typNum - argNum1]]]},
									Fold[ Lambda$TM[ RNG$[ TYPRNG$[ Null, #2]], #1]&, conj, Reverse[ Drop[ typs, argNum1]]]
								]]]
							],
							prem,
							deleteTrailing[ 0] /@ Position[ prem, fix]
						]
					]
				],
				prems
			]},
	With[ {imp = fromAssmsGoal[ newPrems, TYPED$[ TrueQ$TM, {}][ ReplacePart[ A, zeros1 -> VAR$[ looseNum]]]]},
		Fold[ TYPED$[ CoreAll$TM, {Last[ #2]}][ Lambda$TM[ RNG$[ #2], #1]]&, imp, Reverse[ loose]]
	]]]]]

replaceConstsByVars[ term_, {}, _] := term
replaceConstsByVars[ term_, consts_List, offset_Integer] :=
	With[ {constPos = Position[ term, First[ #], {0, Infinity}, Heads -> True]& /@ consts, num = Length[ consts]},
		ReplacePart[ term, Join @@ MapIndexed[ Function[ {cp, i}, (# -> VAR$[ countAbstractionsFromPos[ #, term] + num + offset - First[ i]])& /@ cp], constPos]]
	]

bindParams[ term_, params_List, kind_] :=
	Fold[
		Switch[ kind,
			CoreAll$TM,
			TYPED$[ CoreAll$TM, {Last[ #2]}][ Lambda$TM[ RNG$[ TYPRNG$[ #2[[1, 1]], Last[ #2]]], #1]]&,
			_,
			Lambda$TM[ RNG$[ TYPRNG$[ #2[[1, 1]], Last[ #2]]], #1]&
		],
		term,
		Reverse[ params]
	]

deleteTrailing[ patt_][ l_] :=
	deleteTrailing[ l, patt]
deleteTrailing[ l_, patt_] :=
	With[ {len = Length[ l]},
		Take[ l, len + 1 - First[ FirstPosition[ Reverse[ l], Except[ patt], {len + 1}, {1}, Heads -> False]]]
	]

(*UNUSED*)
(* 'constToSymbolString[ c]' converts the given constant 'c' into a string that only contains characters which can be
	used in names of identifiers (i.e. letters, letter-like characters, and digits). Note however that the first
	character in the result might be a digit, i.e. do not call 'Symbol' directly on the result, but prepend some
	prefix first! The conversion is injective. *)
constToSymbolString[ c_Symbol] :=
	SymbolName[ c]
constToSymbolString[ c_Integer] :=
	If[ Negative[ c],
		(* "ng" is appended rather than prepended, to ensure that the resulting string starts with a digit character
			and thus does not overlap with names of symbols. *)
		constToSymbolString[ -c] <> "ng",
	(*else*)
		ToString[ c, InputForm]
	]
constToSymbolString[ c_Rational] :=
	If[ Negative[ c],
		constToSymbolString[ -c] <> "ng",
	(*else*)
		constToSymbolString[ Numerator[ c]] <> "dv" <> constToSymbolString[ Denominator[ c]]
	]
constToSymbolString[ c_Real] :=
	If[ Negative[ c],
		constToSymbolString[ -c] <> "ng",
	(*else*)
		StringReplace[ ToString[ c, InputForm], {"." -> "d", "`" -> "p", "*^-" -> "ee", "*^" -> "e", " " -> ""}]
	]
constToSymbolString[ c_Complex] :=
	With[ {re = Re[ c]},
		If[ re == 0,
			constToSymbolString[ Im[ c]] <> "I",
		(*else*)
			constToSymbolString[ Re[ c]] <> "cx" <> constToSymbolString[ Im[ c]] <> "I"
		]
	]
constToSymbolString[ DirectedInfinity[]] :=
	"0\[Infinity]"
constToSymbolString[ DirectedInfinity[ c_]] :=
	constToSymbolString[ c] <> "\[Infinity]"

checkListSowMessage[ l_List, test_, m:(All|First), msgID_, args_List, rest___] :=
	With[ {p = First /@ Position[ l, _?(Composition[ Not, test]), {1}, If[ m === First, 1, Infinity], Heads -> False]},
		If[ p === {},
			False,
		(*else*)
			sowHOLMessage[ msgID, Prepend[ args, p], rest];
			True
		]
	]

(* 'addFactsToList[ name, facts, ky, listPos, factPos, add]' adds the given facts to the global fact list of name
	'name'. If the fact list does not yet exist, it is generated; otherwise, 'add' specifies whether the facts shall be
	added or whether they shall replace the existing facts.
	'facts' must be a list of keys or 'THM$'-structures. 'ky', 'listPos' and 'factPos' are the source information of the
	cell that calls this function.
	Note that the fact list is generated even if 'facts' is empty! *)
addFactsToList[ name_String, facts0_List, ky_, listPos_, factPos_, add_] :=
	With[ {flKey = getFactListKeys[ name, 1],
			facts = {#, {ky, factPos}}& /@ facts0},
		If[ flKey === {},
			AssociateTo[ $globalFactLists, {name, {ky, listPos}} -> facts],
		(*else*)
			Which[
				add =!= True,
				AssociateTo[ $globalFactLists, First[ flKey] -> facts],
			(*else*)
				facts === {},
				$globalFactLists,
			(*else*)
				True,
				AssociateTo[ $globalFactLists,
					First[ flKey] -> Join[ Select[ Lookup[ $globalFactLists, Key @@ First[ flKey], {}], (!MemberQ[ facts0, First[ #]])&], facts]
				]
			]
		]
	]


(* ::Subsection:: *)
(*Proofs*)

completeLatticeProof[ argNum_Integer] :=
	With[ {latticeBoolLbl = "CompleteLatticeC BOOL", latticeFunLbl = "CompleteLatticeC FUN", inductiveThy = Lookup[ $theoryMapping, "$Inductive", $Failed]},
	With[ {latticeBoolKey = fastKeyFromLabel[ {inductiveThy, latticeBoolLbl}],
			latticeFunKey = If[ Positive[ argNum], fastKeyFromLabel[ {inductiveThy, latticeFunLbl}], Null]},
		Which[
			latticeBoolKey === $Failed,
			sowHOLMessage[ 11, {STRING$[ latticeBoolLbl]}, "Error"];
			$Failed,
		(*else*)
			latticeFunKey === $Failed,
			sowHOLMessage[ 11, {STRING$[ latticeFunLbl]}, "Error"];
			$Failed,
		(*else*)
			True,
			Join[
				If[ Positive[ argNum],
					Join[
						Hold[ INF$[ "insert", latticeFunKey]],
						Hold @@ Join @@ Table[ {INF$[ "makeExplicit", 1], INF$[ "impE", 1, 0]}, {argNum}]
					]
				],
				Hold[
					INF$[ "formula", latticeBoolKey],
					INF$[ "impE", 1, 0]
				]
			]
		]
	]]

indSimpProof[ paramNum_Integer, argNum_Integer, predTyped_, fun_, i_Integer, id_Integer, srcLabel_String] :=
	With[ {fixpointLbl = "LFP fixpoint unfold", inductiveThy = Lookup[ $theoryMapping, "$Inductive", $Failed]},
	With[ {defKey = {cellIDLabel[ {id, i, 2}], srcLabel}, monoKey = {cellIDLabel[ {id, i, 3}], srcLabel},
			fixpointKey = fastKeyFromLbl[ {inductiveThy, fixpointLbl}],
			tcProof = completeLatticeProof[ argNum]},
		Which[
			fixpointKey === $Failed,
			sowHOLMessage[ 11, {STRING$[ fixpointLbl]}, "Error"];
			$Failed,
		(*else*)
			tcProof === $Failed,
			$Failed,
		(*else*)
			True,
			Module[ {X},
			With[ {paramFixes = Table[ FIX$[ X, j], {j, 0, paramNum - 1}]},
				Join[
					If[ Positive[ paramNum],
						Hold @@ Join[
							{INF$[ "ext", paramNum]},
							INF$[ "allI", #]& /@ paramFixes
						],
					(*else*)
						Hold[]
					],
					Hold @@ {
						INF$[ "formula", fixpointKey],
						Theorema`CoreProvers`instantiateCoreAll[
							{
								EXPR$[ applyCurried[ predTyped, paramFixes, {}]],
								EXPR$[ applyCurried[ fun, paramFixes, {}]]
							}
						],
						INF$[ "impE", 1, 0]
					},
						
					(* type-class constraint *)
					tcProof,
					
					Hold @@ {	
						(* equality *)
						INF$[ "formula", defKey],
						INF$[ "replace", 1, Join[ {1, 0, 1}, Table[ 0, {paramNum}]]],
						INF$[ "refl"],
						
						(* monotonicity *)
						INF$[ "formula", monoKey]
					},
					Hold @@ (INF$[ "allE", 1, makeSTERM[ #]]& /@ paramFixes),
					Hold[
						INF$[ "impE", 1, 0],
						
						INF$[ "close"]
					]
				]
			]]
		]
	]]

indIntroProof[ paramNum_Integer, argNum_Integer, varNum_Integer, premNum_Integer, j_Integer, introNum_Integer, i_Integer, id_Integer, srcLabel_String] :=
	With[ {reflLbl = "\[Equal] refl", holThy = Lookup[ $theoryMapping, "$HOL", $Failed]},
	With[ {simpKey = {cellIDLabel[ {id, i, 4}], srcLabel},
			reflKey = If[ argNum === 0, Null, fastKeyFromLabel[ {holThy, reflLbl}]]},
		If[ reflKey === $Failed,
			sowHOLMessage[ 11, {STRING$[ reflLbl]}, "Error"];
			$Failed,
		(*else*)
			Module[ {X},
			With[ {paramFixes = Table[ FIX$[ X, k], {k, 0, paramNum - 1}],
					varFixes = Table[ FIX$[ X, k], {k, paramNum, paramNum + varNum - 1}],
					rhsPos = Table[ 1, {premNum + 3}]},	(* 'premNum + 3', because of additional assumption and 'TrueQ'. *)
				Hold @@ Join[
					INF$[ "allI", #]& /@ paramFixes,
					INF$[ "allI", #]& /@ varFixes,
					{
						INF$[ "formula", simpKey],
						INF$[ "replace", 1, Join[ Table[ 1, {premNum + 2}], Table[ 0, {paramNum + argNum}]]],	(* 'premNum + 2', because of additional assumption and 'TrueQ'. *)
						INF$[ "drop", 1]
					},
					If[ introNum === 1, {}, {Theorema`CoreProvers`proveDisjunct[ j, introNum]}],
					If[ varNum === 0, {}, {Theorema`CoreProvers`introduceExists[ varFixes]}],
					If[ argNum === 0, {}, {INF$[ "insert", reflKey]}],
					{
						Theorema`CoreProvers`proveConjunction[
							Join[
								Table[ Hold[ INF$[ "makeExplicit", 1], closeByReflHOL[ rhsPos]], {argNum}],
								Table[ Hold @@ {INF$[ "impE", k, 0]}, {k, 1, premNum}]
							]
						],
						INF$[ "close"]
					}
				]
			]]
		]
	]]

closeByReflHOL[ termPos_List][ PRFSIT$[ {G_, ___}, __], ___] :=
	Hold @@ {INF$[ "allE", 1, makeSTERM[ extractSingle[ G, termPos]]], INF$[ "impE", 1, 0]}


(* ::Section:: *)
(*Theory-Level "<consts> where <fmls>"-Constructs*)

parseWhereConstruct[ consts0_List, fmls0_List, id_, src_, msgHeader_] :=
	With[ {consts = extractFreshConstant[ #, False, True, msgHeader]& /@ consts0},
		If[ MemberQ[ consts, Except[ {_, _, _STERM$}]],
			$Failed,
		(*else*)
			With[ {fmls1 =
					If[ fmls0 === {},
						{{}, {}},
					(*else*)
						Transpose[
							Replace[ fmls0,
								{
									LABELED$[ f_, lbl_String, data___] :> {f, {lbl, data}},
									LABELED$[ f_, _, data___] :> {f, {Automatic, data}},
									f_ :> {f, {Automatic}}
								},
								{1}
							]
						]
					]},
			Module[ {X, fmls = First[ fmls1], i = 0,
					idx = Max[ Max @@ Join[ First /@ Join @@ (Keys /@ consts[[All, 3, -1]]), Cases[ consts[[All, 2]], TVAR$[ k_Integer, ___] :> k, Infinity, Heads -> True]] + 1, 0],
					tfix = Association[], tvar = Association[]},
				Catch[
					Replace[
						Reap[
							Replace[
								consts,
								{c_, ctyp0_, STERM$[ _, tf_, tv_]} :>
									With[ {ky = Lookup[ getConstDecls[], c, {$Failed, $Failed}][[2]]},
										tfix = intersectSorts[ tfix, tf];
										tvar = intersectSorts[ tvar, tv];
										FIX$[ X, ++i] ->
											If[ ListQ[ ky] && !(sameCellQ[ id, src, ky] && AllTrue[ getConstDefs[ False][ c], sameCellQ[ id, src, #[[2, 1]]]&]),
												Which[
													MatchQ[ ctyp0, _TVAR$],
													If[ StringQ[ msgHeader], sowHOLMessage[ 101, {c}, "Fatal", msgHeader]];
													Throw[ $Failed],
												(*else*)
													FreeQ[ ctyp0, Except[ TVAR$[], _TVAR$]],
													With[ {ctyp1 = ctyp0 /. TVAR$[] :> TVAR$[ idx++]},
													With[ {dtyp = getDeclaredType[ HoldComplete[ c], idx]},
														Replace[ unifyTypes[ First[ dtyp], ctyp1, hasSort[ tfix], {Join[ tvar, dtyp[[3]]], {}}, 0],
															{
																{tvar0_Association, ts_List} :>
																	With[ {ctyp2 = ctyp1 /. ts},
																		idx = Last[ dtyp];
																		tvar = tvar0;
																		fmls =
																			replaceAllNoTypes[
																				fmls,
																				{
																					c -> FIX$[ X, i],
																					(h:(MessageName$TM|TYPED$|$typeHeads))[ c, t_, r___] :>
																						h[ If[ ListQ[ matchType[ ctyp2, t]], FIX$[ X, i], c], t, r]
																				}
																			];
																		ctyp2
																	],
																_ :>
																	(
																		If[ StringQ[ msgHeader], sowHOLMessage[ 102, {c, ctyp0}, "Fatal", msgHeader]];
																		Throw[ $Failed]
																	)
															}
														]
													]],
												(*else*)
													True,
													If[ StringQ[ msgHeader], sowHOLMessage[ 103, {c, ctyp0}, "Fatal", msgHeader]];
													Throw[ $Failed]
												],
											(*else*)
												fmls = replaceAllNoTypes[ fmls, c -> FIX$[ X, i]];
												Sow[ i, "declPos"];
												ctyp0 /. TVAR$[] :> TVAR$[ idx++]
											]
									],
								{1}
							],
							"declPos"
						],
						{fix_List, declPos_List} :>
							Replace[
								makeCoreForms[ fmls, idx, fix, {}, tfix, tvar, {}, None, normalize, None],
								{
									{{_, STERM$[ fmls2_List, tfix1_Association, tvar1_Association], idx1_Integer, newFix_List, {}}, ts_List} :>
										With[ {fix1 = fix /. ts},
										With[ {gen = typeGeneralizationReplacement[ {fmls2, fix1, newFix}, {}, idx1]},
											{
												consts[[All, 1]],
												fmls2 /. gen,
												fix1 /. gen,
												newFix /. gen,
												Join[
													tvar1,
													Association @@ DeleteCases[ Map[ (Last[ #] -> Lookup[ tfix1, First[ #], Lookup[ tfix, First[ #], Default]])&, gen], _ -> Default]
												],	(* This association might contain type variables that do not appear anywhere anymore. *)
												Flatten[ declPos],
												Last[ fmls1]
											}
											(* The generalization of a.b.f. type constants must happen here, because otherwise those constants
												appearing in 'fix' are not generalized. *)
										]],
									_ -> $Failed
								}
							]
					]
				]
			]]
		]
	]

replaceAllNoTypes[ term_, rule:(_Rule|_RuleDelayed)] :=
	replaceAllNoTypes[ term, {rule}]
replaceAllNoTypes[ term_, {}] := term
replaceAllNoTypes[ term_, rules_List] :=
	Module[ {rules1},
		rules1 = Append[ rules, (h:(MessageName$TM|TYPED$|$typeHeads))[ x_, r___] :> h[ replaceAllNoTypes[ x, Hold[ rules1]], r]];
		replaceAllNoTypes[ term, Hold[ rules1]]
	]
replaceAllNoTypes[ term_, Hold[ rules_]] := term /. rules


(* ::Section:: *)
(*Theory Language*)

With[ {srcLabel = sourceLabel[ "$Inductive"]},

System`RegisterTheory[ "$Inductive", srcLabel];

If[ FreeQ[ $theoryCommandSeparators, "inductive"],
	(* TODO: Move to notebook. *)
	addCustomTheoryInterpretationRules[
		{
			{Hold[ pre___, "where"], fmls_Hold, post_} :> {HOLTheoryLanguageTranslator[ "where", Hold[ pre], fmls], post},
			{Hold[ "inductive"], Hold[ args___], Hold[ "where"]} :> {Hold[], Hold[ "inductive", args, "where"]}
		},
		{},
		{"where", "inductive"},
		sourceLabel[ "$Inductive"]
	];
	
	addCriticalTheoryCommands[ {defineInductive -> "inductive"}, srcLabel]
];

HOLTheoryLanguageTranslator[ "where", Hold[ keyword_String, consts___], Hold[ intros___, opts:({Except[ "is", _String], ___}...)]] :=
	Replace[ makeFmlList[ False][ intros],
		l0_List :>
			With[ {cmd = processWhereConstruct[ keyword]},
				If[ Head[ cmd] === processWhereConstruct,
					$Failed,
				(*else*)
					Hold @@ {cmd[ parseCoreTermPre /@ DeleteCases[ {consts}, "and"], l0, {opts}]}
				]
			]
	];


(* ::Section:: *)
(*Theory*)

$currentID = 100;

declareConstantsDirect[
	STERM$[
		TYPED$[ #,
			FUN$TM[ TVAR$[ 0], FUN$TM[ TVAR$[ 0], Theorema`Knowledge`BOOL$TM]]
		],
		Association[],
		Association[ TVAR$[ 0] -> {FORM$TM}]
	]& /@ {Theorema`Language`LessEqual$TM},	(* a_ -> a_ -> BOOL *)
	makeFormulaKey[ ++$currentID, "$Inductive"]
];

declareConstantsDirect[
	STERM$[
		TYPED$[ #,
			FUN$TM[ FUN$TM[ TVAR$[ 0], TVAR$[ 0]], Theorema`Knowledge`BOOL$TM]
		],
		Association[],
		Association[ TVAR$[ 0] -> {FORM$TM}]
	]& /@ {Theorema`Knowledge`isMono$TM},	(* (a_ -> a_) -> BOOL *)
	makeFormulaKey[ ++$currentID, "$Inductive"]
];

declareConstantsDirect[
	STERM$[
		TYPED$[ #,
			FUN$TM[ TVAR$[ 0], Theorema`Knowledge`BOOL$TM]
		],
		Association[],
		Association[ TVAR$[ 0] -> {FORM$TM}]
	]& /@ {Theorema`Knowledge`CompleteLatticeC$TM},	(* a_ -> BOOL *)
	makeFormulaKey[ ++$currentID, "$Inductive"]
];

declareConstantsDirect[
	STERM$[
		TYPED$[ #,
			FUN$TM[ FUN$TM[ TVAR$[ 0], TVAR$[ 0]], TVAR$[ 0]]
		],
		Association[],
		Association[ TVAR$[ 0] -> {FORM$TM}]
	]& /@ {Theorema`Knowledge`LFP$TM},	(* (a_ -> a_) -> a_ *)
	makeFormulaKey[ ++$currentID, "$Inductive"]
];

If[ selectFormulas[ {srcLabel, "CompleteLatticeC BOOL"}, 1] === {},
	Reap[
		fastKeyFromLabel[ {srcLabel, "CompleteLatticeC BOOL"}] = {cellIDLabel[ ++$currentID], srcLabel};
		fastKeyFromLabel[ {srcLabel, "CompleteLatticeC FUN"}] = {cellIDLabel[ ++$currentID], srcLabel};
		
		(* CompleteLatticeC[ TYPE::BOOL] *)
		Replace[
			Block[ {$postParsingRules = {}},
				formulaToAssumption[
					STERM$[
						TrueQ$TM[
							Theorema`Knowledge`CompleteLatticeC$TM[
								TYPED$[ Theorema`Knowledge`TYPE$TM, {Theorema`Knowledge`BOOL$TM}]
							]
						],
						Association[],
						Association[]
					],
					Identity,
					False
				]
			],
			{f_, tv_Association} :>
				addAxmToTheory[
					FML$[ fastKeyFromLabel[ {srcLabel, "CompleteLatticeC BOOL"}], f, "CompleteLatticeC BOOL"],
					tv,
					False,
					False,
					False,
					False,
					False
				]
		];
		
		(* CompleteLatticeC[ TYPE::b_] ==> CompleteLatticeC[ TYPE::(a_ -> b_)] *)
		Replace[
			Block[ {$postParsingRules = {}},
				formulaToAssumption[
					STERM$[
						CoreImplies$TM[
							TrueQ$TM[
								Theorema`Knowledge`CompleteLatticeC$TM[
									TYPED$[ Theorema`Knowledge`TYPE$TM, {TVAR$[ 1]}]
								]
							]
						][
							TrueQ$TM[
								Theorema`Knowledge`CompleteLatticeC$TM[
									TYPED$[ Theorema`Knowledge`TYPE$TM, {FUN$TM[ TVAR$[ 0], TVAR$[ 1]]}]
								]
							]
						],
						Association[],
						Association[ TVAR$[ 0] -> {FORM$TM}, TVAR$[ 1] -> {FORM$TM}]
					],
					Identity,
					False
				]
			],
			{f_, tv_Association} :>
				addAxmToTheory[
					FML$[ fastKeyFromLabel[ {srcLabel, "CompleteLatticeC FUN"}], f, "CompleteLatticeC FUN"],
					tv,
					False,
					False,
					False,
					False,
					False
				]
		],
		
		"Message"
	]
];

If[ selectFormulas[ {srcLabel, "LFP fixpoint unfold"}, 1] === {},
	Reap[
		fastKeyFromLabel[ {srcLabel, "LFP fixpoint unfold"}] = {cellIDLabel[ ++$currentID], srcLabel};
		
		(* /\h::a_ f::(a_->a_). CompleteLatticeC[ TYPE::a_] ==> h === LFP[ f] ==> isMono[ f] ==> h === f[ h] *)
		Replace[
			Block[ {$postParsingRules = {}},
				formulaToAssumption[
					STERM$[
						CoreAll$TM[ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$h$TM, TVAR$[ 0]]],
							CoreAll$TM[ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$f$TM, FUN$TM[ TVAR$[ 0], TVAR$[ 0]]]],
								CoreImplies$TM[
									TrueQ$TM[ Theorema`Knowledge`CompleteLatticeC$TM[ TYPED$[ Theorema`Knowledge`TYPE$TM, {TVAR$[ 0]}]]]
								][
									CoreImplies$TM[
										CoreEqual$TM[ VAR$[ 1]][ Theorema`Knowledge`LFP$TM[ VAR$[ 0]]]
									][
										CoreImplies$TM[
											TrueQ$TM[ Theorema`Knowledge`isMono$TM[ VAR$[ 0]]]
										][
											CoreEqual$TM[ VAR$[ 1]][ VAR$[ 0][ VAR$[ 1]]]
										]
									]
								]
							]]
						]],
						Association[],
						Association[ TVAR$[ 0] -> {FORM$TM}]
					],
					Identity,
					False
				]
			],
			{f_, tv_Association} :>
				addAxmToTheory[
					FML$[ fastKeyFromLabel[ {srcLabel, "LFP fixpoint unfold"}], f, "LFP fixpoint unfold"],
					tv,
					False,
					False,
					False,
					False,
					False
				]
		],
		
		"Message"
	]
];


(* ::Subsection:: *)
(*Testing Only*)

declareTypesDirect[ Theorema`Knowledge`NAT$TM, cellIDLabel[ "core200"], srcLabel, True];

declareConstantsDirect[
	STERM$[
		TYPED$[ #,
			FUN$TM[ TVAR$[ 0], FUN$TM[ TVAR$[ 0], TVAR$[ 0]]]
		],
		Association[],
		Association[ TVAR$[ 0] -> {FORM$TM}]
	]& /@ {Plus$TM, Times$TM, Subtract$TM, Divide$TM},	(* a_ -> a_ -> a_ *)
	makeFormulaKey[ 27, "$Inductive"]
];
declareConstantsDirect[
	STERM$[
		TYPED$[ #,
			FUN$TM[ TVAR$[ 0], TVAR$[ 0]]
		],
		Association[],
		Association[ TVAR$[ 0] -> {FORM$TM}]
	]& /@ {Minus$TM},	(* a_ -> a_ *)
	makeFormulaKey[ 28, "$Inductive"]
];
declareConstantsDirect[
	STERM$[
		TYPED$[ #,
			TVAR$[ 0]
		],
		Association[],
		Association[ TVAR$[ 0] -> {FORM$TM}]
	]& /@ {0},	(* a_ *)
	makeFormulaKey[ 301, "$Inductive"]
];

termNotation[ 1, {Plus$TM, Times$TM, Power$TM}, {1, Right}][ False, 0, 0, "$Inductive"];
termNotation[ 1, {Subtract$TM, Divide$TM}, {1, Left}][ False, 0, 0, "$Inductive"];

setTheory[ sourceLabel[ NotebookFileName[]]];
addTheoryDeps[ {srcLabel}];

] (* With *)

End[]

EndPackage[]
