(* ::Package:: *)

(* Theorema
    Copyright (C) 2017 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

BeginPackage[ "Theorema`Minimal`"]

Needs[ "Theorema`Common`"]
Needs[ "Theorema`Language`"]
Needs[ "Theorema`Core`Trusted`"]
Needs[ "Theorema`Core`Tools`"]

Begin["`Private`"]


(* ::Section:: *)
(*Theorem Schemas*)

(* ::Subsection:: *)
(*Rearranging Explicit Assumptions*)

rearrangeStatement[ i_Integer?Positive, j_Integer?Positive] :=
	{
		With[ {M = Max[ i, j]},
			Nest[
				TYPED$[ CoreAll$TM, {FORM$TM}][ Lambda$TM[ RNG$[ TYPRNG$[ Null, FORM$TM]], #]]&,
				TYPED$[ CoreImplies$TM, {}][
					Fold[ TYPED$[ CoreImplies$TM, {}][ VAR$[ #2]][ #1]&, VAR$[ 0], Insert[ Delete[ Range[ M], M + 1 - i], M + 1 - i, M + 1 - j]]
				][
					Fold[ TYPED$[ CoreImplies$TM, {}][ VAR$[ #2]][ #1]&, VAR$[ 0], Range[ M]]
				],
				M + 1
			]
		],
		{}
	}

rearrangeProof[ i_Integer?Positive, j_Integer?Positive] :=
	Hold @@ Join[
		Table[ INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$A$TM, k]], {k, 0, Max[ i, j]}],
		{INF$[ "impE", 1, If[ i === j, i, 0]]},
		Which[
			i === j,
			{},
		(*else*)
			i < j,
			Join[
				Table[
					INF$[ "impE", k, 0],
					{k, 1, i - 1}
				],
				Table[
					INF$[ "impE", k, 0],
					{k, i + 1, j}
				],
				{INF$[ "impE", i, 0]}
			],
		(*else*)
			True,
			Join[
				Table[
					INF$[ "impE", k, 0],
					{k, 1, j - 1}
				],
				{INF$[ "impE", i, 0]},
				Table[
					INF$[ "impE", k, 0],
					{k, j, i - 1}
				]
			]
		],
		{INF$[ "close"]}
	]

rearrangeStatement[ l:{___Integer?Positive}] :=
	{
		With[ {M = Max @@ l + 1},
			Nest[
				TYPED$[ CoreAll$TM, {FORM$TM}][ Lambda$TM[ RNG$[ TYPRNG$[ Null, FORM$TM]], #]]&,
				TYPED$[ CoreImplies$TM, {}][
					Fold[ TYPED$[ CoreImplies$TM, {}][ VAR$[ M - #2]][ #1]&, VAR$[ 0], Reverse[ l]]
				][
					Fold[ TYPED$[ CoreImplies$TM, {}][ VAR$[ #2]][ #1]&, VAR$[ 0], Range[ M - 1]]
				],
				M
			]
		],
		{}
	}

rearrangeProof[ l_List] :=
	With[ {M = Max @@ l},
		Hold @@ Join[
			Table[ INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$A$TM, k]], {k, 0, M}],
			{INF$[ "impE", 1, 0]},
			Map[
				INF$[ "impE", #, 0]&,
				l
			],
			{INF$[ "close"]}
		]
	]


(* ::Subsection:: *)
(*Modus Ponens*)

mpStatement[ Forward][] = mpStatement[ Forward][ 1];
mpProof[ Forward][] = mpProof[ Forward][ 1];

(* /\A1 ... An thesis. A1 ==> ... ==> An ==> (A1 ==> ... ==> An ==> thesis) ==> thesis *)
mpStatement[ Forward][ n_Integer?Positive] :=
	{
		Nest[
			TYPED$[ CoreAll$TM, {FORM$TM}][ Lambda$TM[ RNG$[ TYPRNG$[ Null, FORM$TM]], #]]&,
			TYPED$[ CoreAll$TM, {FORM$TM}][
				Lambda$TM[
					RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$thesis$TM, FORM$TM]],
					With[ {varList = Array[ VAR$, n], fun = TYPED$[ CoreImplies$TM, {}][ #2][ #1]&},
						Fold[ fun, TYPED$[ CoreImplies$TM, {}][ Fold[ fun, VAR$[ 0], varList]][ VAR$[ 0]], varList]
					]
				]
			],
			n
		],
		{}
	}

mpProof[ Forward][ n_Integer] :=
	Hold @@ Join[
		Table[ INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$X$TM, i]], {i, 0, n}],
		{INF$[ "impE", n + 1, 0]},
		Table[ INF$[ "impE", i, 0], {i, 1, n}],
		{INF$[ "close"]}
	]

(* /\P Q. (Q ==> P) ==> Q ==> P *)
mpStatement[ Backward][] =
	Block[ {$postParsingRules = {}},
		formulaToAssumption[
			CoreAll$TM[ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$P$TM, FORM$TM]],
				CoreAll$TM[ Lambda$TM[ RNG$[ TYPRNG$[ Theorema`Knowledge`VAR$Q$TM, FORM$TM]],
					CoreImplies$TM[ CoreImplies$TM[ VAR$[ 0]][ VAR$[ 1]]][ CoreImplies$TM[ VAR$[ 0]][ VAR$[ 1]]]
				]]
			]],
			Identity,
			True
		]
	]

mpProof[ Backward][] =
	Hold[
		INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$P$TM, 0]],
		INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$Q$TM, 0]],
		INF$[ "impE", 1, 1],
		INF$[ "close"]
	]

(* "GMP" means "generalized Modus Ponens". Depending on the values of 'i' and 'l', the theorems roughly look like
	'/\A1 ... An C, P1 ... Pm, R1 ... Rk, G.
		((P1 ==> ... ==> Pm ==> C) ==> R1 ==> ... ==> Rk ==> (P1 ==> ... Pi ==> A1) ==> ... ==> (Pj ==> ... ==> Pm ==> Al) ==> G) ==>
		((A1 ==> ... ==> An ==> C) ==> R1 ==> ... ==> Rk ==> (P1 ==> ... Pi ==> A1) ==> ... ==> (Pj ==> ... ==> Pm ==> Al) ==> G)'.
	'i' determines the position of the main implication among the assumptions.
	'l' is supposed to be a list consisting of elements that are either pairs of positive- and non-negative integers, or
	'Skip'.
	The variables in "GMP" are bound in the following order:
	- first come the constituents of the original implication, in the right order,
	- then come the premises of the assumptions that are used for discharging original premises, in the right order,
	- then come all remaining assumptions,
	- then comes the remaining goal. *)
gmpStatement[ i_Integer?Positive, l_List] :=
	{
		With[ {m = Max[ i, Max @@ Cases[ l, {j_, _} :> j]],
				origNum = Length[ l] + 1, premNum = Plus @@ Cases[ l, {_, j_} :> j]},
		With[ {remNum = m - origNum + Count[ l, Skip]},
		With[ {varNum = origNum + premNum + remNum + 1},
		With[ {goalPatt =
				Module[ {rem = remNum + 1, premTot = 0},
					Table[
						If[ k === i,
							Null,
						(*else*)
							Replace[ Position[ l, {k, _} , {1}, 1],
								{
									{{j_}} :>
										With[ {prem = l[[j, 2]]},
											premTot += prem;
											Fold[ TYPED$[ CoreImplies$TM, {}][ #2][ #1]&, VAR$[ varNum - j], Array[ VAR$, prem, varNum - origNum - premTot]]
										],
									_ :> VAR$[ --rem]
								}
							]
						],
						{k, 1, m}
					]
				]},
			Nest[
				TYPED$[ CoreAll$TM, {FORM$TM}][ Lambda$TM[ RNG$[ TYPRNG$[ Null, FORM$TM]], #]]&,
				TYPED$[ CoreImplies$TM, {}][
					fromAssmsGoal[
						ReplacePart[
							goalPatt,
							i -> Fold[
									TYPED$[ CoreImplies$TM, {}][ VAR$[ #2]][ #1]&,
									VAR$[ varNum - origNum],
									Reverse[
										MapIndexed[
											Replace[ #1,
												{
													{_, 0} -> Sequence[],
													{a_, b_} :> With[ {base = remNum + Plus @@ Cases[ l, {a0_, b0_} /; a0 > a :> b0]}, Sequence @@ Range[ b + base, base + 1, -1]],
													Skip -> varNum - First[ #2]
												}
											]&,
											l
										]
									]
								]
						],
						VAR$[ 0]
					]
				][
					fromAssmsGoal[
						ReplacePart[
							goalPatt,
							i -> Fold[
									TYPED$[ CoreImplies$TM, {}][ #2][ #1]&,
									VAR$[ varNum - origNum],
									Array[ VAR$, origNum - 1, varNum - origNum + 1]
								]
						],
						VAR$[ 0]
					]
				],
				varNum
			]
		]]]],
		{}
	}

gmpProof[ i_Integer, l_List] :=
	With[ {m = Max[ i, Max @@ Cases[ l, {j_, _} :> j]],
				origNum = Length[ l] + 1, premNum = Plus @@ Cases[ l, {_, j_} :> j]},
	With[ {remNum = m - origNum + Count[ l, Skip]},
	With[ {varNum = origNum + premNum + remNum + 1},
		Hold @@ Join[
			Table[ INF$[ "allI", FIX$[ Theorema`Knowledge`VAR$X$TM, k]], {k, 0, varNum - 1}],
			{INF$[ "impE", 1, 0]},
			
			(* proof of the original assumptions coming before the subject implication *)
			Table[ INF$[ "impE", k, Replace[ Cases[ l, {k, p_} :> p, {1}, 1], {{p_} :> p, _ -> 0}]], {k, 1, i - 1}],
			
			(* proof of the new implication *)
			{INF$[ "impE", i, 0]},
			Module[ {counter = m - 1},
				Replace[ l,
					{
						Skip :> INF$[ "impE", ++counter, 0],
						{a_, b_} :> Sequence @@ Prepend[ Table[ INF$[ "impE", counter++, 0], {b}], INF$[ "impE", If[ a > i, a - 1, a], 0]]
					},
					{1}
				]
			],
			
			(* proof of the original assumptions coming after the subject implication *)
			Table[ INF$[ "impE", k, Replace[ Cases[ l, {k, p_} :> p, {1}, 1], {{p_} :> p, _ -> 0}]], {k, i + 1, m}],
			
			{INF$[ "close"]}
		]
	]]]


(* ::Section:: *)
(*Proof Methods*)

Theorema`CoreProvers`loadProvers[ "CoreProofMethods"];

End[]

EndPackage[]
