(* ::Package:: *)

(* Theorema
    Copyright (C) 2017 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

BeginPackage[ "Theorema`Interface`ProofScript`"];

Needs[ "Theorema`Common`"]
If[ Theorema`Common`$privateSymbolsLoaded =!= True, Get[ FileNameJoin[ {Theorema`$TheoremaDirectory, "Theorema", "System", "Common.m"}]]]	(* TODO: Change when publish. *)
Needs[ "Theorema`Language`"]
Needs[ "SequentialEvaluation`", FileNameJoin[ {Theorema`$TheoremaDirectory, "Theorema", "Components", "SequentialEvaluation.m"}]]

With[ {lang = "English"},
	MessageName[ addInterpretationRules, "usage", lang] = "";
	MessageName[ cellsToProof, "usage", lang] = "";
	MessageName[ getDefaultProverSettings, "usage", lang] = "";
	MessageName[ interpretBoxes, "usage", lang] = "";
	MessageName[ makeButtonLabel, "usage", lang] = "";
	MessageName[ prove, "usage", lang] = "";
	
	MessageName[ $commandInterpretationRules, "usage", lang] = "";
	MessageName[ $commandSeparators, "usage", lang] = "";
	MessageName[ $commandPreSplit, "usage", lang] = "";
	MessageName[ $formulaQuotingOperators, "usage", lang] = "";
]

Begin[ "`Private`"];


(* ::Section:: *)
(*Public*)

getDefaultProverSettings[] :=
	{
		initInteractiveProveCall,	(* initialization function:		goal_ * assms_List * interactive:(True|False) -> {prfObj_, data_Hold, msg_List} *)
		execProofCommands,			(* execution function:			cmd_Hold * poIn_ * interactive:(True|False) -> {poOut_, msg_List} *)
		displayProofSituation,		(* display function:			pending_Hold -> id_Integer -> boxes_ *)
		Null&,						(* knowledge display function:	-> _ *)
		printCurrentStatus,			(* printing function:			-> cells_List *)
		exportProof					(* exportation function:		goal_ * assms_List * prfObj_|failureIndex_ * time_ * num_Integer * reg:(True|False) * flag_Integer -> _ *)
									(* 'flag' is
										- '0' if function is called after checking proof in mode '0',
										- '1' if function is called when user exports proof manually,
										- '2' if function is called after replaying proof in mode '2'. *)
	}

If[ !ListQ[ Theorema`Provers`$proverSettings], Theorema`Provers`$proverSettings = getDefaultProverSettings[]];
If[ !ValueQ[ $running], $running = False];
If[ !ValueQ[ $currentSEID], $currentSEID = 0];

(* OBSOLETE *)
openProof[ spec_, Number -> num_Integer] :=
	openProof[ spec, num]
openProof[ spec_] :=
	openProof[ spec, 1]
openProof[ spec_, num_Integer?NonNegative] :=
	Replace[ selectFormulas[ spec, 1],
		{
			{fml_} :> openProof[ fml, num],
			_ :> notificationStyled[ "Error", False][ translate[ "fmlNotFound"], spec]
		}
	]
openProof[ fml_FML$, num_Integer?NonNegative] :=
	With[ {src = sourceFile[ fml]},
	With[ {file = FileNameJoin[ {DirectoryName[ src], FileBaseName[ src], "p" <> extractCellID[ fml] <> "-" <> ToString[ num] <> "-nb.nb"}]},
		If[ FileExistsQ[ file],
			NotebookOpen[ file],
		(*else*)
			notificationStyled[ "Error", False][ translate[ "prfNotFound"], num]
		]
	]]

prove[ spec_, rest___] :=
	Replace[ selectFormulas[ spec, 1],
		{
			{fml_} :> prove[ fml, rest],
			_ :> (notificationStyled[ "Error", False][ translate[ "goalNotFound"], spec]; $Failed)
		}
	]
prove[ goal_, opts___?OptionQ] :=
	prove[ goal, {}, opts]
prove[ goal_, assms_List, opts___?OptionQ] :=
	prove[ goal, assms, 0, opts]
prove[ goal_FML$, assms_List, mode_, opts___?OptionQ] :=
	With[ {origin =
			Replace[ Trigger,
				{
					opts,
					Trigger :> {EvaluationNotebook[], EvaluationCell[]}
				}
			]},
		Replace[
			getProofNumber[ key@goal, Replace[ Number, {opts, Number -> 1}], Automatic],
			{
				{num_Integer?NonNegative, reg_} :>
					prove[ goal, assms, mode, num, reg,
						Replace[ {Notebook, Default},
							{
								opts,
								Notebook :> FileNameJoin[ {createPerNotebookDirectory[ sourceFile[ goal]], "p" <> extractCellID[ goal] <> "-" <> ToString[ num] <> "-nb.nb"}],
								Default -> None
							},
							{1}
						],
						origin
					],
				num_ :> (notificationStyled[ "Error", origin, False][ translate[ "prfNumberInvalid"], num]; $Failed)
			}
		]
	]
prove[ goal_, assms_List, mode_, num_Integer, reg_, {nb_String, default_}, origin_List] :=
	prove[
		goal,
		assms,
		mode,
		num,
		reg,
		Replace[ Quiet[ Check[ Get[ nb], $Failed]],
			{
				Notebook[ cells_List, ___] :> cells,
				_ :>
					Replace[ Quiet[ Check[ Get[ default], $Failed]],
						{
							Notebook[ cells_List, ___] :> (CopyFile[ default, nb]; cells),
							_ -> {}
						}
					]
			}
		],
		nb,
		origin
	]
prove[ goal_, assms_List, mode_, num_Integer, reg_, cells_List, file_, origin_List] :=
	prove[
		goal,
		Join @@ Map[
			Replace[ selectFormulas[ #, 1],
				{} :> (notificationStyled[ "Warning", origin, False][ translate[ "assmNotFound"], #]; {})
			]&,
			assms
		],
		mode, num, reg, cells, file, origin, Theorema`Provers`$proverSettings
	]
prove[ goal_, assms_List, 2, num_, reg_, cells_List, _, origin_List, settings_List] :=
	replayProof[ goal, assms, cellsToProof[ cells, False], num, reg, origin, settings]
prove[ goal:FML$[ _, _, lbl_String, ___], assms_List, mode_, num_, reg_, cells_List, file_, origin:{nb_NotebookObject, cell_CellObject}, {init_, exec_, disp_, dispK_, print_, expfn_}] :=
	If[ TrueQ[ $running],
		Theorema`Interface`InteractivePalette`exitCellEvaluation[ nb, cell, False];
		Null,
	(*else*)
			(* For circumventing the strange behavior where sometimes 'NotebookDynamicExpression' wrongly calls 'closeSE'. *)
		With[ {ps0 = init[ goal, assms, True], $id = $currentSEID++},
		With[ {ps = First[ ps0]},
			Scan[ notificationStyled[ origin, False], Last[ ps0]];
			If[ ps === $Failed,
				Theorema`Interface`InteractivePalette`exitCellEvaluation[ nb, cell, False];
				Null,
			(*else*)
				$displayKnowledge = dispK;
				$printProofSituation = print;
				$running = True;
				$proofExportationFunction = Function[ po, expfn[ goal, assms, po, Indeterminate, num, reg, 1]];
				If[ !MemberQ[ Theorema`Interface`InteractivePalette`$permanentAbortionExpr, HoldPattern[ $running = _]],
					Theorema`Interface`InteractivePalette`$permanentAbortionExpr = Join[ Theorema`Interface`InteractivePalette`$permanentAbortionExpr, Hold[ $running = False]]
				];
				With[ {cmd =
						If[ mode === 0,
							Replace[ cellsToProof[ cells, True], {$Failed :> (notificationStyled[ "Error", origin, False][ translate[ "prfNotebookSyntaxError"]]; {}), c_List :> {c}}],
						(*else*)
							{}
						]},
				With[ {out =
						If[ cmd === {},
							initiateSE,
						(*else*)
							Theorema`Interface`InteractivePalette`$enterInteractive = False;
							Theorema`Interface`InteractivePalette`addAbortionExprs[ Hold[ closeSE[ $id, False]]];
							Theorema`Interface`InteractivePalette`$currentProofLength = 0;
							Theorema`Interface`InteractivePalette`$currentProofStep = 0;
							Theorema`Interface`InteractivePalette`enterCellEvaluation[ nb, cell, 3, lbl, num, reg];
							runSE
						][
							Join[ Hold[ $TMAproofObject -> ps, $pendingCommand -> Hold[], $lastCmd -> {}], ps0[[2]]],
							Sequence @@ cmd,
							$id,
							"DisplayFunction" -> disp[ Hold[ $pendingCommand =!= Hold[]]],
							"ParsingFunction" -> parseInput,
							"ExecutionFunction" -> executeStep[ exec],
							"ExecutionFunction1" -> executeStep[ exec, {}],
							"StepFunction" ->
								Function[ {step, len},
									Theorema`Interface`InteractivePalette`$currentProofStep = step;
									Theorema`Interface`InteractivePalette`$currentProofLength = len;
								],
							"EnterInteractive" -> Hold[ Theorema`Interface`InteractivePalette`$enterInteractive],
							"EnterInteractiveFunction" ->
								Function[ {reason, step, time},
									Theorema`Interface`InteractivePalette`dropAbortionExprs[ All];
									Theorema`Interface`InteractivePalette`abortEvaluation[ False];
									If[ reason === 2, expfn[ goal, assms, step, time, num, reg, 0]]
								],
							"Postprocessing" ->
								Function[ {id, time},
									Replace[ expfn[ goal, assms, $TMAproofObject, time, num, reg, 0],
										out:Except[ $Failed] :> (Theorema`Interface`InteractivePalette`dropAbortionExprs[ -1]; out)
									]
								],
							"ClosingFunction" -> (($running = False)&),
							"MainCellStyle" -> $cmdStyle,
							"MinorCellStyle" -> {$fmlStyle},
							"Input" -> cells,
							"InputWindowOptions" ->
								{
									StyleDefinitions -> FileNameJoin[ {$TheoremaDirectory, "Theorema", "Interface", "Templates", "Proof Script.nb"}],
									WindowSelected -> True,
									WindowTitle -> "Proof of " <> makeLabel[ lbl],
									NotebookFileName -> file
								},
							"StatusWindowOptions" ->
								{
									StyleDefinitions -> makeColoredStylesheet[ "Proof State"],
									WindowSize -> $StatusSize,
									WindowMargins -> {{Automatic, First[ $StatusPos]}, {Automatic, Last[ $StatusPos]}}
								}
						]},
					Theorema`Interface`InteractivePalette`exitCellEvaluation[ nb, cell, True];
					out
				]]
			]
		]]
	]


(* ::Section:: *)
(*Private*)

(* ::Subsection:: *)
(*Stylesheets*)

$cmdStyle = "ProofText";
$fmlStyle = "FormalTextInputFormula";
$StatusSize = {600, 700};
$StatusPos = {500, 0};	(* the initial position of the status window, relative to the upper right corner of the screen *)
$FixBox = FormBox[ StyleBox[ "\[SelectionPlaceholder]", "ExpressionABF"], TheoremaForm];
$MetaBox = FormBox[ StyleBox[ "\[SelectionPlaceholder]", "ExpressionMeta"], TheoremaForm];
$SyntacticBox = FormBox[ StyleBox[ "\[SelectionPlaceholder]", "ExpressionSyntactic"], TheoremaForm];

(* If the template changes, or the input header changes, or further formula buttons are added to the GUI, just call
	'updateStylesheet[]' once to update the stylesheet of proof scripts.
	Changing the color scheme in the GUI does not have any effect on proof notebooks (yet).
	The reason for fixing the stylesheet is to keep the sizes of proof notebooks as small as possible; they can simply
	refer to the fixed stylesheet-notebook and do not have to include all the style definitions in-line.
	Actually, they would not even need to refer to the fixed stylesheet: the style definitions are automatically set to
	the stylesheet whenever an interactive proof is initiated. *)
updateStylesheet[] :=
	updateStylesheet[
		FileNameJoin[ {$TheoremaDirectory, "Theorema", "Interface", "Templates", "Proof Script-Template.nb"}],
		FileNameJoin[ {$TheoremaDirectory, "Theorema", "Interface", "Templates", "Proof Script.nb"}],
		$TheoremaColorScheme
	]
updateStylesheet[ fileIn_String, fileOut_String, color_] :=
	With[ {tmp = NotebookOpen[ fileIn, Visible -> False]},
	With[{styles = NotebookGet[ tmp]},
		NotebookClose[ tmp];
		With[ {alias =
				Join[
					Cases[ Flatten[ Transpose[ Theorema`Interface`GUI`Private`allFormulae][[2]], 2],
						{n_String, i_Integer} :>
							With[ {bd = Theorema`Interface`GUI`Private`langButtonData[ n]},
								If[ i === 1,
									bd[[4]] -> RowBox[ {TagBox[ "(", "AutoParentheses"], bd[[2]], TagBox[ ")", "AutoParentheses"]}],
								(*else*)
									bd[[4]] -> bd[[2]]
								]
							]
					],
					{
						"(" -> TagBox[ "(", "AutoParentheses"],
						")" -> TagBox[ ")", "AutoParentheses"],
						"tma" -> RowBox[ {"\[OpenCurlyQuote]", "\[SelectionPlaceholder]", "\[CloseCurlyQuote]"}],
						"fix" -> $FixBox,
						"meta" -> $MetaBox,
						"svar" -> $SyntacticBox
					}
				]},
			NotebookSave[
				styles /. Append[ Table[ CMYKColor @@ IntegerDigits[ i, 2, 4] -> TMAcolor[ i, color], {i, 0, 15}], (InputAliases -> {}) -> (InputAliases -> alias)]
					/. Cell[ StyleData[ All, "Working"]] :> Cell[ StyleData[ All, "Working"], DockedCells -> inputHeader[ $displayKnowledge, $printProofSituation]],
				fileOut
			]
		]
	]]

inputHeader[ showKnowledge_, internalStatus_] :=
	{
		Cell[
			BoxData[
				ToBoxes[
					Row[
						{
							Button[ "FIX",
								FrontEndExecute[
									{
										NotebookApply[
											InputNotebook[],
											$FixBox,
											Placeholder
										]
									}
								],
								Appearance -> "Frameless"
							],
							Button[ "META",
								FrontEndExecute[
									{
										NotebookApply[
											InputNotebook[],
											$MetaBox,
											Placeholder
										]
									}
								],
								Appearance -> "Frameless"
							],
							Button[ "SVAR",
								FrontEndExecute[
									{
										NotebookApply[
											InputNotebook[],
											$SyntacticBox,
											Placeholder
										]
									}
								],
								Appearance -> "Frameless"
							],
							Button[ "Available Knowledge",
								showKnowledge[],
								Appearance -> "Frameless"
							],
							Button[ "Internal Status",
								CreateDocument[
									With[ {cells = internalStatus[]},
										Join[
											If[ ListQ[ cells], cells, {}],
											{CellGroup[
												Join[ {TextCell[ "Last/Pending Commands", "Subsubsection"]}, $lastCmd, If[ MatchQ[ $pendingCommand, _Hold], {$pendingCommand}, {}]],
												1
											]}
										]
									],
									WindowTitle -> "Current Status"
								],
								Appearance -> "Frameless"
							],
							Button[ "Proof Object",
								CreateDocument[
									{$TMAproofObject},
									WindowTitle -> "Proof Object"
								],
								Appearance -> "Frameless"
							],
							Button[ "Export Proof",
								(
									Replace[ $proofExportationFunction[ $TMAproofObject],
										{
											Null :> MessageDialog[ "\[Checkmark] " <> translate[ "exportSuccess"]],
											res_String :> MessageDialog[ "\[WarningSign] " <> translate[ "exportFailure"] <> ": " <> res],
											_ :> MessageDialog[ "\[WarningSign] " <> translate[ "exportFailure"] <> "."]
										}
									];
								),
								Appearance -> "Frameless"
							]
						},
						Spacer[ 5]
					]
				]
			],
			"Text",
			Background -> Theorema`TMAcolor[ 1],
			Deployed -> True
		]
	}


(* ::Subsection:: *)
(*Interactive Proving*)

getProofNumber[ k_List, {num_, reg_}, _] :=
	getProofNumber[ k, num, reg]
getProofNumber[ k_List, num_Integer, reg_] :=
	{num, If[ reg === Automatic, With[ {regNum = getRegisteredProofNumber[ k]}, regNum === num || (regNum === False && num === 1)], reg]}
getProofNumber[ k_List, "\[FivePointedStar]", reg_] :=
	{
		Replace[ getRegisteredProofNumber[ k],
			Except[ _Integer] -> 1
		],
		If[ reg === Automatic, True, reg]
	}
getProofNumber[ k_List, "-", reg_] :=
	With[ {prefix = "p" <> extractCellID[ k] <> "-", suffix = "-nb.nb", src = sourceFile[ k]},
		Replace[
			Max[
				Map[
					ToExpression[ StringReplace[ #, __ ~~ prefix ~~ d:(DigitCharacter ..) ~~ suffix :> d]] &,
					FileNames[ prefix ~~ (DigitCharacter..) ~~ suffix, {FileNameJoin[ {DirectoryName[ src], FileBaseName[ src]}]}]
				]
			],
			{
				num_Integer :> {num + 1, If[ reg === Automatic, False, reg]},
				_ :> {1, If[ reg === Automatic, True, reg]}
			}
		]
	]
getProofNumber[ _, _, _] :=
	$Failed

makeButtonLabel[ FML$[ _, _, l_, ___, "origForm" -> orig_, ___], i_Integer, $id_] :=
	Tooltip[ makeButtonLabel[ l, i, $id], theoremaDisplay[ orig]]
makeButtonLabel[ FML$[ _, _, l_, ___], i_Integer, $id_] :=
	makeButtonLabel[ l, i, $id]
makeButtonLabel[ l_String, 0, $id_] :=
	makeButtonLabel[ makeLabel[ l], "goal", $id]
makeButtonLabel[ l_String, _, $id_] :=
	makeButtonLabel[ makeLabel[ l], "\"\<" <> l <> "\>\"", $id]
makeButtonLabel[ _, 0, $id_] :=
	makeButtonLabel[ "?", "goal", $id]
makeButtonLabel[ _, i_Integer?Positive, $id_] :=
	makeButtonLabel[ "?", ToString[ i], $id]
makeButtonLabel[ l_String, p_String, $id_] :=
	Button[ l, NotebookApply[ SequentialEvaluation`Private`$input[$id], RowBox[ {p, " "}]], Appearance -> "Frameless"]

parseInput[ _, $cmdStyle, (BoxData|RowBox)[ {}]] :=
	Hold[]
parseInput[ $id_, $cmdStyle, RowBox[ {"\[OpenCurlyQuote]", box_, "\[CloseCurlyQuote]"}]] :=
	parseInput[ $id, $fmlStyle, BoxData[ box]]
parseInput[ $id_, $cmdStyle, RowBox[ {open_, box_, close_}]] :=
	With[ {quotingOp = Replace[ {open, close}, $formulaQuotingOperators]},
		parseInput[ $id, $fmlStyle, BoxData[ quotingOp[ box]]] /; !ListQ[ quotingOp]
	]
parseInput[ $id_, $cmdStyle, box:FormBox[ _, TheoremaForm]] :=
	parseInput[ $id, $fmlStyle, BoxData[ box]]
parseInput[ $id_, $cmdStyle, box:TagBox[ _, "FIX$"|"META$"|"SVAR$"]] :=
	parseInput[ $id, $fmlStyle, BoxData[ box]]
parseInput[ $id_, $cmdStyle, InterpretationBox[ FormBox[ _, TheoremaForm], expr_]] :=
	Hold[ EXPR$[ expr]]
parseInput[ $id_, $cmdStyle, TagBox[ box_, DisplayForm]] :=
	parseInput[ $id, $cmdStyle, box]
parseInput[ $id_, $cmdStyle, SubscriptBox[ box_, sub_]] :=
	Replace[ {parseInput[ $id, $cmdStyle, box], parseInput[ $id, $cmdStyle, sub]},
		{
			{Hold[ a_], Hold[ b_]} :> Hold[ Subscript[ a, b]],
			_ -> $Failed
		}
	]
parseInput[ $id_, $cmdStyle, SuperscriptBox[ box_, sup_]] :=
	Replace[ {parseInput[ $id, $cmdStyle, box], parseInput[ $id, $cmdStyle, sup]},
		{
			{Hold[ a_], Hold[ b_]} :> Hold[ Superscript[ a, b]],
			_ -> $Failed
		}
	]
parseInput[ _, $cmdStyle, RowBox[ {"(*", ___}]] :=
	Hold[]
parseInput[ $id_, $cmdStyle, RowBox[ {left:("("|"{"|"\[LeftAngleBracket]"), args___, right:(")"|"]"|"}"|"\[RightAngleBracket]"|"\[RightDoubleBracket]")}]] :=
	Replace[ parseInput[ $id, $cmdStyle, RowBox[ {args}]],
		{
			out_Hold :> ReplacePart[ Hold[ Boxes[ left, out, right]], {1, 2, 0} -> List],
			_ -> $Failed
		}
	]
parseInput[ $id_, $cmdStyle, RowBox[ {pre___, left:("["|"\[LeftDoubleBracket]"), args___, right:(")"|"]"|"}"|"\[RightAngleBracket]"|"\[RightDoubleBracket]")}]] :=
	Replace[ {Switch[ {pre}, {}, Hold[], {_}, parseInput[ $id, $cmdStyle, pre], _, parseInput[ $id, $cmdStyle, RowBox[ {pre}]]], parseInput[ $id, $cmdStyle, RowBox[ {args}]]},
		{
			{out1_Hold, out2_Hold} :> Join[ out1, ReplacePart[ Hold[ Boxes[ left, out2, right]], {1, 2, 0} -> List]],
			_ -> $Failed
		}
	]
parseInput[ $id_, $cmdStyle, RowBox[ {sgn:("-"|"+"), arg_}]] :=
	With[ {out = parseInput[ $id, $cmdStyle, arg]},
		If[ MatchQ[ out, Hold[ _Integer|_Rational|_Real]],
			If[ sgn === "-",
				Hold @@ {-First[ out]},
			(*else*)
				out
			],
		(*else*)
			Prepend[ out, sgn]
		]
	]
parseInput[ $id_, $cmdStyle, (BoxData|RowBox)[ boxes_List]] :=
	With[ {split = splitList[ boxes, _String?(StringMatchQ[ #, Whitespace]&)]},
		If[ Length[ split] === 1,
			Hold @@ (Join @@ Map[ parseInput[ $id, $cmdStyle, #]&, First[ split]]),
		(*else*)
			Hold @@ (Join @@ Map[ parseInput[ $id, $cmdStyle, Replace[ #, {{x_} :> x, x_List :> RowBox[ x]}]]&, split])
		]
	]
parseInput[ $id_, $cmdStyle, BoxData[ box_]] :=
	parseInput[ $id, $cmdStyle, box]
parseInput[ _, $cmdStyle, s_String] :=
	Which[
		s === "True",
		Hold[ True],
		
		s === "False",
		Hold[ False],
		
		StringMatchQ[ s, "\"" ~~ ___ ~~ "\""],
		Hold @@ {STRING$[ ToExpression[ s]]},

		StringMatchQ[ s, NumberString],
		Hold @@ {ToExpression[ s]},

		True,
		Hold[ s]
	]
parseInput[ _, $fmlStyle, box_] :=
	Block[ {$delimiters = Hold[]},
		Join[
			Replace[ parseFormula[ removeRedundantBoxes[ box]],
				{
					LABELED$[ fml:Except[ _EXPR$], data__] :> LABELED$[ EXPR$[ fml], data],
					fml:Except[ _EXPR$] :> EXPR$[ fml]
				},
				{1}
			],
			$delimiters
		]
	]
parseInput[ ___] :=
	$Failed

splitList[ l_List, patt_] :=
	Join @@ Last[ Reap[ splitList[ l, patt, {}]]]
splitList[ {}, _, acc_List] :=
	(
		If[ acc =!= {}, Sow[ acc]];
		Null
	)
splitList[ {f_, r___}, patt_, acc_List] :=
	If[ MatchQ[ f, patt],
		If[ acc =!= {}, Sow[ acc]];
		splitList[ {r}, patt, {}],
	(*else*)
		splitList[ {r}, patt, Append[ acc, f]]
	]

parseFormula[ {box_}] :=
	parseFormula[ BoxData[ box]]
parseFormula[ {pre__, ";"}] :=
	(
		AppendTo[ $delimiters, ";"];
		parseFormula[ {pre}]
	)
parseFormula[ {first_, ",", rest__}] :=
	Join[ parseFormula[ BoxData[ first]], parseFormula[ {rest}]]
parseFormula[ BoxData[ RowBox[ l:{___, _String?(StringMatchQ[ #, Whitespace]&), ___}]]] :=
	parseFormula[ BoxData[ RowBox[ DeleteCases[ l, _String?(StringMatchQ[ #, Whitespace]&)]]]]
parseFormula[ BoxData[ FormBox[ box:StyleBox[ _, "ExpressionABF"|"ExpressionMeta"|"ExpressionSyntactic"], TheoremaForm]]] :=
	parseFormula[ BoxData[ makeFMSBox[ box, Null]]]
(* TagBoxes are used in Theorema code cells, since StyleBoxes are automatically removed (by Mathematica) when evaluating a cell. *)
parseFormula[ BoxData[ TagBox[ v_, t:("FIX$"|"META$"|"SVAR$")]]] :=
	parseFormula[ BoxData[ makeFMSBox[ v, t]]]
parseFormula[ BoxData[ StyleBox[ box_, _]]] :=
	parseFormula[ BoxData[ box]]
parseFormula[ BoxData[ RowBox[ {"\[LeftDoubleBracket]", args___, "\[RightDoubleBracket]"}]]] :=
	parseFormula[ BoxData[ RowBox[ {"Theorema`Language`SEQ$", "[", args, "]"}]]]
parseFormula[ BoxData[ RowBox[ l:{_, PatternSequence[ ",", _]..., ";"}]]] :=
	(
		AppendTo[ $delimiters, ";"];
		parseFormula[ Most[ l]]
	)
parseFormula[ BoxData[ RowBox[ l:{_, PatternSequence[ ",", _]...}]]] :=
	parseFormula[ l]
parseFormula[ BoxData[ boxes_List]] :=
	Join @@ Map[ Composition[ parseFormula, BoxData], DeleteCases[ boxes, _String?(StringMatchQ[ #, Whitespace]&)]]
parseFormula[ BoxData[ RowBox[ {name_String, ":", form_}]]] :=
	With[ {n = Quiet[ Check[ ToExpression[ name, InputForm, Hold], $Failed]],
			f = Quiet[ Check[ parseFormula[ form], Hold[ $Failed]]]},
		With[ {f0 = First[ f], n0 = First[ n]},
			Hold[ LABELED$[ f0, n0]]
		] /; (MatchQ[ n, Hold[ _String]] && MatchQ[ f, Hold[ Except[ _LABELED$|$Failed]]])
	]
parseFormula[ box_] :=
	makeTmaExpressionFromBoxes[
		box //. {FormBox[ bx_, TheoremaForm] :> bx, TagBox[ bx_, DisplayForm] :> bx, TagBox[ InterpretationBox[ _, bx_], "Boxes"] :> bx} /.
		    {
				bx:StyleBox[ _, "ExpressionABF"|"ExpressionMeta"|"ExpressionSyntactic"] :> makeFMSBox[ bx, Null],
				TagBox[ v_, t:("FIX$"|"META$"|"SVAR$")] :> makeFMSBox[ v, t]
			},
		Identity,
		Identity
	]

makeFMSBox[ StyleBox[ v_, "ExpressionABF"], _] :=
	makeFMSBox[ v, "FIX$"]
makeFMSBox[ StyleBox[ v_, "ExpressionMeta"], _] :=
	makeFMSBox[ v, "META$"]
makeFMSBox[ StyleBox[ v_, "ExpressionSyntactic"], _] :=
	makeFMSBox[ v, "SVAR$"]
makeFMSBox[ RowBox[ {"(", v_, ")"}], head_] :=
	makeFMSBox[ v, head]
makeFMSBox[ RowBox[ {v_, " "..., ",", " "..., i_String}], head_] :=
	RowBox[ {head, "[", RowBox[ {addVarPrefix[ v, head], ",", i}], "]"}]
makeFMSBox[ SubscriptBox[ RowBox[ {"(", v_, ")"}], i_], head_] :=
	RowBox[ {head, "[", RowBox[ {addVarPrefix[ v, head], ",", i}], "]"}]
makeFMSBox[ SubscriptBox[ v_, i_], head_] :=
	RowBox[ {head, "[", RowBox[ {addVarPrefix[ v, head], ",", i}], "]"}]
makeFMSBox[ v_, head_] :=
	If[ head === "SVAR$",
		RowBox[ {head, "[", addVarPrefix[ v, head], "]"}],
	(*else*)
		RowBox[ {head, "[", RowBox[ {addVarPrefix[ v, head], ",", "0"}], "]"}]
	]

addVarPrefix[ s_String, "META$"|"SVAR$"] :=
	With[ {s0 = If[ StringMatchQ[ s, __ ~~ "$TM"], s, s <> "$TM"]},
		(* this must happen here, since 'freshNames' ignores meta variables *)
		Which[
			StringMatchQ[ s0, "Theorema`Knowledge`VAR$" ~~ __],
			s,
		(*else*)
			StringMatchQ[ s0, "Theorema`Knowledge`" ~~ __],
			StringInsert[ s0, "VAR$", 20],
		(*else*)
			StringMatchQ[ s0, "VAR$" ~~ __],
			"Theorema`Knowledge`" <> s0,
		(*else*)
			StringMatchQ[ s0, "Theorema`Language`" ~~ __],
			"Theorema`Knowledge`VAR$" <> StringDrop[ s0, 19],
		(*else*)
			True,
			"Theorema`Knowledge`VAR$" <> s0
		]
	]
addVarPrefix[ s_String, _] :=
	If[ StringMatchQ[ s, "VAR$" ~~ __],
		s,
	(*else*)
		"VAR$" <> s
	]
addVarPrefix[ RowBox[ {f_, rest___}], head_] :=
	RowBox[ {addVarPrefix[ f, head], rest}]

cellsToProof[ cells_List, True] :=
	Quiet[
		Catch[
			interpretCommandStream[
				Cases[ cells, Cell[ c_, s:($cmdStyle|$fmlStyle), opts___] :> {parseInput[ 0, s, c], s, Replace[ CellID, {opts}]}],	(* in fact, the cell ID is not needed ... *)
				False
			],
			$Failed
		]
	]
cellsToProof[ cells_List, ___] :=
	Quiet[
		Catch[
			interpretCommandStream[
				Hold @@ (Join @@ Cases[ cells, Cell[ c_, s:($cmdStyle|$fmlStyle), ___] :> parseInput[ 0, s, c]]),
				False
			],
			$Failed
		]
	]

interpretBoxes[ boxes_] :=
	Quiet[
		Catch[
			interpretCommandStream[ parseInput[ 0, $cmdStyle, boxes], False],
			$Failed
		]
	]

executeStep[ exec_][ $id_, cmd1_Hold, cmd2_List] :=
	With[ {c0 = Reap[ interpretCommandStream[ Join[ cmd1, Hold @@ (Join @@ cmd2)], True], "Message"]},
		executeStep[ exec, Switch[ c0, {_, {{___String}}}, c0[[2, 1]], _, {}]][ $id, First[ c0]]
	]
executeStep[ exec_, msg0_List][ _, cmd_Hold] :=
	With[ {res = exec[ cmd, $TMAproofObject, True]},
	With[ {new = First[ res], msg = Join[ msg0, Last[ res]]},
		Switch[ new,
			_Integer,
			Join[ {$Failed}, msg, {"Inference " <> ToString[ new] <> " is not applicable."}],

			_String,
			Join[ {$Failed}, msg, {new}],
			
			$Failed,
			If[ msg === {}, $Failed, Prepend[ msg, $Failed]],

			_,
			$TMAproofObject = new;
			If[ msg === {}, Null, msg]
		]
	]]
executeStep[ _, msg_List][ _, cmd:{__Integer}] :=
	Join[ {$Failed}, msg, {"Command(s) " <> (StringJoin @@ Riffle[ ToString /@ cmd, ", "]) <> " cannot be interpreted"}]
executeStep[ _, msg_List][ _, _] :=
	Join[ {$Failed}, msg, {"Input cannot be parsed"}]

(* 'replayProof' is essentially the same as 'checkProof' in "/Core/Trusted/Consistency.m". *)
replayProof[ goal:FML$[ _, _, lbl_, ___], assms_List, proof:(_List|_Hold), num_, reg_, origin:{nb_NotebookObject, cell_CellObject}, {init_, exec_, _, _, _, expfn_}] :=
	(
		Theorema`Interface`InteractivePalette`enterCellEvaluation[ nb, cell, 3, lbl, num, reg];
		With[ {startTime = SessionTime[], ps0 = init[ goal, assms, False]},
		With[ {ps1 = First[ ps0]},
			If[ ps1 === $Failed,
				Scan[ notificationStyled[ origin, False], Last[ ps0]];
				Theorema`Interface`InteractivePalette`exitCellEvaluation[ nb, cell, False];
				Last[ ps0],
			(*else*)
				With[ {msg = Last[ ps0]},
					If[ msg =!= {}, PutAppend @@ Append[ msg, "log.txt"]]
				];
				With[ {out =
						Module[ {i = 1, po = ps1},
							ReleaseHold[ Set @@@ ps0[[2]]];
							Catch[
								Scan[
									With[ {res = exec[ Hold[ #], po, False]},
									With[ {msg = Last[ res]},
										If[ msg =!= {}, PutAppend @@ Append[ msg, "log.txt"]];
										(* Messages are written to the file immediately, such that one can inspect the messages generated so far even if the proof is aborted. *)
										po = First[ res];
										If[ po === $Failed,
											Throw[ i],
										(*else*)
											++i
										]
									]]&,
									List @@ proof
								];
								po
							]
						]},
					Which[
						IntegerQ[ num] && NonNegative[ num],
						Theorema`Interface`InteractivePalette`exitCellEvaluation[ nb, cell, True];
						expfn[ goal, assms, out, SessionTime[] - startTime, num, reg, 2],
					(*else*)
						IntegerQ[ out],
						notificationStyled[ "Fatal", origin, False][ translate[ "prfError"], makeLabel[ lbl], out];
						Theorema`Interface`InteractivePalette`exitCellEvaluation[ nb, cell, False];
						out,
					(*else*)
						True,
						Theorema`Interface`InteractivePalette`exitCellEvaluation[ nb, cell, True];
						True
					]
				]
			]
		]]
	)
replayProof[ ___] := 0


(* ::Subsubsection::Closed:: *)
(*Interpreting Command Streams*)

interpretCommandStream[ cmd_List, ___] :=
	Catch[
		Module[ {pre = Null, preIndex = 0, main, sep = False},
		With[ {c =
				Join @@ Last[
					Reap[
						Scan[
							If[ sep,
								sep = False;
								With[ {post = First[ #]},
									Replace[ Replace[ {pre, main, post}, $commandInterpretationRules],
										{
											{c:Except[ $Failed], new:Except[ $Failed]} :>
												(
													If[ !MatchQ[ c, _[]], Sow[ {c, preIndex}, "cmd"]];
													pre = new;
													If[ !MatchQ[ new, _Hold] || new === post, preIndex = Last[ #]]
												),
											_ :> Throw[ $Failed]
										}
									]
								],
							(*else*)
								sep = True;
								main = #
							]&,
							splitCommandStream[ cmd, $commandSeparators]
						];
						While[ MatchQ[ pre, _Hold],
							Replace[ Replace[ {pre, Hold[], Null}, $commandInterpretationRules],
								{
									{c:Except[ $Failed], new:Except[ $Failed|pre]} :>
										(
											If[ !MatchQ[ c, _[]], Sow[ {c, preIndex}, "cmd"]];
											pre = new;
										),
									_ :> Throw[ $Failed]
								}
							]
						],
						"cmd"
					]
				]},
			Replace[ SplitBy[ c, Last],
				all:{{_, id_}, ___} :> {Hold @@ Join @@ all[[All, 1]], id},
				{1}
			]
		]]
	]
interpretCommandStream[ cmd_Hold] :=
	interpretCommandStream[ cmd, True]
interpretCommandStream[ cmd_Hold, interactive_] :=
	With[ {split = splitCommandStream[ Join[ If[ MatchQ[ $pendingCommand, _Hold], $pendingCommand, Hold[]], cmd], $commandSeparators]},
	With[ {fun = If[ TrueQ[ interactive], (Sow[ #1, "cmd"]; Sow[ #2 -> #1, "map"])&, Sow[ #1, "cmd"]&]},
	Module[ {pre = Null, main, sep = False},
		Replace[
			Reap[
				Catch[
					Scan[
						If[ sep,
							sep = False;
							Replace[ Replace[ {pre, main, #}, $commandInterpretationRules],
								{
									{c:Except[ $Failed], new:Except[ $Failed]} :>
										(
											If[ !MatchQ[ c, _[]], fun[ c, {pre, main, #}]];
											pre = new
										),
									_ :>
										If[ TrueQ[ interactive] && # === Null,
											If[ MatchQ[ pre, _Hold],
												Throw[ Join[ pre, main]],
											(*else*)
												Throw[ main]
											],
										(*else*)
											Throw[ False]
										]
								}
							],
						(*else*)
							sep = True;
							main = #
						]&,
						split
					];
					While[ MatchQ[ pre, _Hold],
						Replace[ Replace[ {pre, Hold[], Null}, $commandInterpretationRules],
							{
								{c:Except[ $Failed], new:Except[ $Failed|pre]} :>
									(
										If[ !MatchQ[ c, _[]], fun[ c, {pre, Hold[], Null}]];
										pre = new;
									),
								_ :> If[ TrueQ[ interactive], Throw[ pre], Throw[ False]]
							}
						]
					];
					Hold[]
				],
				{"cmd", "map"}
			],
			{
				{p_Hold, {c_List, m_List}} :>
					(
						$pendingCommand = p;
						$lastCmd = Join @@ m;
						Hold @@ Join @@ Join @@ c
					),
				_ -> $Failed
			}
		]
	]]]

splitCommandStream[ stream_List, patt_] :=
	Join @@ Last[
		Reap[
			Module[ {i, pending = Hold[], counter = 0},
				Scan[
					Function[ {cell},
						Replace[ cell,
							{s_Hold, style_, _} :>
								With[ {strm = Join[ pending, s]},
									If[ style === $cmdStyle, ++counter];
									i = 1;
									Scan[
										With[ {j = First[ #]},
											Sow[ Take[ strm, {i, j - 1}]];
											Sow[ {Take[ strm, {j, j}], counter}];
											i = j + 1
										]&,
										Position[ strm, patt, {1}]
									];
									pending = Drop[ strm, i - 1]
								]
						]
					],
					$commandPreSplit[ stream]
				];
				If[ pending =!= Hold[],
					Sow[ pending];
					Sow[ {Null, counter}]
				]
			]
		]
	]
splitCommandStream[ stream0_Hold, patt_] :=
	With[ {stream = $commandPreSplit[ stream0]},
		Join @@ Last[
			Reap[
				With[ {p = Position[ stream, patt, {1}]},
				Module[ {i = 1},
					Scan[
						With[ {j = First[ #]},
							Sow[ Take[ stream, {i, j - 1}]];
							Sow[ Take[ stream, {j, j}]];
							i = j + 1
						]&,
						p
					];
					If[ !MatchQ[ p, {___, {Length[ stream]}}],
						Sow[ Drop[ stream, i - 1]];
						Sow[ Null]
					]
				]]
			]
		]
	]


(* ::Subsection:: *)
(*Default Prover Setup*)

initInteractiveProveCall[ goal_FML$, assms_List, _] :=
	With[ {customRules = Theorema`Provers`customInterpretationRules[ Theorema`Interface`GUI`Private`$selectedRuleSet],
			customSeps = Theorema`Provers`customCommandSeparators[ Theorema`Interface`GUI`Private`$selectedRuleSet]},
		$commandInterpretationRules =
			Join[
				{
					{Hold[ "Skip"], Hold[], post_} :> {{{"Skip"}}, post},
					{Hold[ "Auto"], Hold[], post_} :> {{{"Auto", False, True}}, post},
					{Hold[ "Auto1"], Hold[], post_} :> {{{"Auto", False, False}}, post},
					{Hold[ "Rename"], Hold[ old__, "to", new:(STRING$[ _String]..)], post_Hold} :> {{{"Rename", {old}, First /@ {new}}}, post},
					{Hold[ "#"], Hold[ b_Integer], post_} :> {{{"Choose", b}}, post},
					{Hold[ "!"], Hold[ forms__], post_} :> {{{"Activate", {forms}}}, post},
					{Hold[ ">"], Hold[ m_String, args___], post_Hold} :> {{{"Apply", {m, args}}}, post},
					{Hold[ ">>"], Hold[ m_String, args___], post_Hold} :> {{{"Activate", {"this"}}, {"Apply", {m, args}}}, post},
					{Hold[ ">>>"], Hold[ m_String, args___], post_Hold} :> {{{"Activate", {"this", "goal"}}, {"Apply", {m, args}}}, post}
				},
				If[ ListQ[ customRules], customRules, {}],
				{
					{Hold[ ";"]|Null, Hold[], post_} :> {{}, post}
				}
			];
		$commandSeparators = Alternatives @@ Join[ {"Skip", "Auto", "Auto1", "Rename", "#", "!", ">", ">>", ">>>", ";"}, If[ ListQ[ customSeps], customSeps, {}]];
		$commandPreSplit = Function[ stream, stream //. Boxes[ "(", l_List, ")"] :> l];
		$formulaQuotingOperators = {};
		$formulaLabel = 0;	(* must be set explicitly *)
		$TMAcheckSuccess = True;
		$rewriteRules = {};
		$generated = {};
		Clear[ $TMAproofNotebook, $TMAproofTree];
		With[ {res =
				execProofCommands[
					Hold[ {"Auto", True, False}],
					{
						Theorema`Provers`Common`Private`makeInitialProofObject[
							goal,
							assms,
							{
								Theorema`Interface`GUI`Private`$selectedRuleSet,
								Map[ # -> Theorema`Interface`GUI`Private`ruleActive[ #]&, Theorema`Interface`GUI`Private`$allRules],
								Map[ # -> rulePriority[ #]&, Theorema`Interface`GUI`Private`$allRules]
							},
							"interactive"
						],
						{{2}},
						True
					},
					False
				]},
		With[ {po = First[ res]},
			If[ MatchQ[ po, {_PRFOBJ$, _, _}],
				{
					po,	(* '$TMAproofObject' is a triple '{po, pos, trust}'. *)
					Hold[
						$formulaLabel -> 0,
						Theorema`Interface`GUI`Private`$addProofKB -> {},
						$goalActivated -> True,
						$knowledgeActivated -> Null,
						$chooseBranch -> Null
					],
					Last[ res]
				},
			(*else*)
				{$Failed, Hold[], Last[ res]}
			]
		]]
	]

execProofCommands[ cmd_Hold, po_, interactive_] :=
	MapAt[ Map[ messageToString, #]&, ReplacePart[ Reap[ execProofCommands[ cmd, po, interactive, {1, 1}], "Message"], {2, 0} -> Join], {2}]
execProofCommands[ Hold[], po_List, _, _] :=
	po
execProofCommands[ Hold[ m_List, ___], po:{_, {}, _}, interactive_, {i_, _}] :=
	Which[
		MatchQ[ m, {"Auto", True, _}],
		po,
		
		interactive,
		i,
		
		True,
		$Failed
	]
execProofCommands[ Hold[ {"Skip"}, rest___], {po_, {_, pos___}, _}, interactive_, ij_] :=
	(
		$goalActivated = True;
		$knowledgeActivated = Null;
		execProofCommands[ Hold[ rest], {po, {pos}, False}, interactive, ij]
	)
execProofCommands[ Hold[ m_List, rest___], {po_PRFOBJ$, pos:{selPSpos_List, ___}, trust_}, interactive_, {i_, j_}] :=
	(
		(* We use '$TMAproofObject' instead of 'po' below, because some inference rules (e.g. 'RRPaddKnowledge') might change it globally.
			However, inference rules must not change the global shape of the proof object, i.e. the changes must not affect 'pos'. *)
		$TMAproofObject = po;
		With[ {selPS = Extract[ po, selPSpos]},
		With[ {new = execProofCommands[ m, selPS, interactive]},
		With[ {newNode = chooseBranch[ First[ new], selPS], newCmd = Last[ new]},
			If[ newNode === $Failed,
				$TMAproofObject = {$TMAproofObject, pos, trust};
				If[ interactive, i, $Failed],
			(*else*)
				$TMAproofObject =
					Which[
						MatchQ[ newNode, _PRFSIT$],
						{ReplacePart[ $TMAproofObject, selPSpos -> newNode], pos, trust},
					(*else*)
						isProofNode[ newNode],
						Theorema`Provers`Common`Private`$lastVisitedNode = id@newNode;
						{propagateProofValues[ ReplacePart[ $TMAproofObject, selPSpos -> newNode]], Join[ getNewRelevantSits[ newNode, selPSpos], Rest[ pos]], trust},
					(*else*)
						True,
						{$TMAproofObject, pos, trust}
					];
				execProofCommands[
					Join[ newCmd, Hold[ rest]],
					$TMAproofObject,
					interactive,
					With[ {j0 = j + Length[ newCmd]},
						If[ j0 === 1, {i + 1, 1}, {i, j0 - 1}]
					]
				]
			]
		]]]
	)
execProofCommands[ {"Choose", b_}, _PRFSIT$, _] :=
	If[ b === 0,
		Sow[ {"Error", "0 is an invalid branch specification."}, "Message"];
		{$Failed, $Failed},
	(*else*)
		$chooseBranch = b;
		{Null, Hold[]}
	]
execProofCommands[ {"Rename", orig_List, new:{__String}}, ps_PRFSIT$, interactive_] :=
	With[ {keys = getFormulaKeys[ orig, ps]},
		If[ Length[ keys] === Length[ new],
			$goalActivated = True;
			$knowledgeActivated = Null;
			With[ {nullPos = Position[ keys, Null, {1}, Heads -> False]},
				{
					ps /. MapThread[ (FML$[ #1, f_, _, r___] :> FML$[ #1, f, #2, r]) &, {Delete[ keys, nullPos], Delete[ new, nullPos]}],
					Hold[]
				}
			],
		(*else*)
			Sow[ {"Error", "The numbers of formulas and labels do not agree."}, "Message"];
			{$Failed, $Failed}
		]
	]
execProofCommands[ {"Auto", allowVoid_, rec_}, ps_PRFSIT$, interactive_] :=
	With[ {threshold0 = getOptionalComponent[ ps, "$FullInteractiveStrat$threshold"],
			proposedRules0 = getOptionalComponent[ ps, "proposedRules"]},
	With[ {allRules =
				Theorema`Provers`Strategies`Private`getRulesPartitionedFilter[
					ps,
					If[ MatchQ[ proposedRules0, {___List}], proposedRules0, {}],
					"term"|"levelSat1"|"levelSat2",
					If[ MatchQ[ threshold0, {__}], threshold0, Theorema`Provers`Strategies`Private`$defaultThresholds]
				]},
		Catch[
			$applyAllPossibleInferences = False;
			Module[ {proposedRules = First[ allRules], newPS = ps},
				While[ proposedRules =!= {},
					With[ {r = First[ proposedRules], rl = proposedRules[[1, 1]]},
						proposedRules = Rest[ proposedRules];
						If[ !MemberQ[ r, "Repeated"],
							newPS = Replace[ newPS, ("proposedRules" -> _List) -> ("proposedRules" -> proposedRules), {1}]
						];
						With[ {newNodes = Theorema`Provers`Strategies`Private`applyAllRulesOnce[ newPS, DeleteCases[ {inferenceRule[ rl]}, _inferenceRule]]},
							Which[
								newNodes === {},
								If[ MemberQ[ r, "Repeated"],
									newPS = Replace[ newPS, ("proposedRules" -> _List) -> ("proposedRules" -> proposedRules), {1}]
								],
							(*else*)
								TrueQ[ rec],
								Throw[ {First[ newNodes], Hold[ {"Auto", True, True}]}],
							(*else*)
								True,
								$goalActivated = True;
								$knowledgeActivated = Null;
								Throw[ {First[ newNodes], Hold[]}]
							]
						]
					]
				]
			];
			
			If[ Length[ allRules] > 2,
				Scan[
					With[ {newNodes = Theorema`Provers`Strategies`Private`applyAllRulesOnce[ ps, #]},
						If[ newNodes =!= {},
							If[ TrueQ[ rec],
								Throw[ {First[ newNodes], Hold[ {"Auto", True, True}]}],
							(*else*)
								$goalActivated = True;
								$knowledgeActivated = Null;
								Throw[ {First[ newNodes], Hold[]}]
							]
						]
					]&,
					Take[ allRules, {2, -2}]
				]
			];
			
			$goalActivated = True;
			$knowledgeActivated = Null;
			If[ TrueQ[ allowVoid], {Null, Hold[]}, {$Failed, $Failed}]
		]
	]]
execProofCommands[ {"Activate", ids_List}, ps_PRFSIT$, interactive_] :=
	(
		If[ MemberQ[ ids, "All"],
			$goalActivated = True;
			$knowledgeActivated = Null,
		(*else*)
			$knowledgeActivated = DeleteDuplicates[ DeleteCases[ getFormulaKeys[ ids, ps], Null]];
			With[ {gk = key@(goal@ps)},
				If[ MemberQ[ $knowledgeActivated, gk],
					$knowledgeActivated = DeleteCases[ $knowledgeActivated, gk];
					$goalActivated = True,
				(*else*)
					$goalActivated = False
				]
			]
		];
		{Null, Hold[]}
	)
execProofCommands[ {"Apply", {m_String, args___}}, ps_PRFSIT$, interactive_] :=
	With[ {ruleName =
			With[ {tmp = Quiet[ Check[ Replace[ m, Theorema`Provers`customTranslationTable[ First[ Cases[ ps, (rules -> r_Hold) :> r, {1}, 1]]]], m]]},
				If[ StringQ[ tmp], Symbol[ "Theorema`Provers`" <> m], tmp]
			]},
	With[ {rule = inferenceRule[ ruleName]},
		If[ MatchQ[ rule, _inferenceRule],
			Sow[ {"Error", "Rule \"" <> m <> "\" not found."}, "Message"];
			{$Failed, $Failed},
		(*else*)
			$applyAllPossibleInferences = True;
			Theorema`Provers`$inferenceRuleParams = {args};
			Theorema`Provers`$allowInteraction = interactive;
			With[ {nodes1 =
					Replace[
						DeleteCases[ ReplaceList[ ps, rule], $Failed],
						{
							Theorema`Provers`Common`Private`ORNODE$[ Theorema`Provers`Common`Private`PRFINFO$[ proveAlternatives, ___], nodes__, pending] :> nodes,
							{nodes___} :> nodes
						},
						{1}
					]},
			With[ {nodes2 =
					DeleteCases[ Map[ propagateProofValues,
										Cases[ nodes1, (_Theorema`Provers`Common`Private`ANDNODE$|_Theorema`Provers`Common`Private`ORNODE$|_Theorema`Provers`Common`Private`TERMINALNODE$)]],
								_[ ___, failed|disproved]
					]},
			With[ {nodes3 =
					With[ {tmp = Cases[ nodes2, _[ ___, proved], {1}, 1]},
						If[ tmp === {},
							nodes2,
						(*else*)
							tmp
						]
					]},
				$goalActivated = True;
				$knowledgeActivated = Null;
				{nodes3, Hold[]}
			]]]
		]
	]]
execProofCommands[ _List, _, _] :=
	{$Failed, $Failed}

chooseBranch[ $Failed|{}, _] :=
	$Failed
chooseBranch[ Null, _] :=
	Null	(* do not reset '$chooseBranch' *)
chooseBranch[ branches_List, ps_PRFSIT$] :=
	Which[
		IntegerQ[ $chooseBranch],
		With[ {i = $chooseBranch},
			$chooseBranch = Null;
			Quiet[
				Check[
					branches[[i]],
					
					Sow[ {"Error", "Branch " <> ToString[ i] <> " does not exist."}, "Message"];
					$Failed
				]
			]
		],
	(*else*)
		MatchQ[ branches, {_}],
		First[ branches],
	(*else*)
		True,
		Module[ {dummy = False},
		With[ {rNB = Theorema`Provers`Strategies`Private`ruleNB[ {formula@goal@ps, Map[ formula, kb@ps]}, branches, Hold[ dummy]]},
			If[ rNB === $Failed,
				$Failed,
			(*else*)
				With[ {tmp = Theorema`Provers`Strategies`Private`FISDialogInput[ rNB, WindowTitle -> translate[ "chooseSteps"]]},
					If[ IntegerQ[ tmp] && 1 <= tmp <= Length[ branches],
						branches[[tmp]],
					(*else*)
						$Failed
					]
				]
			]
		]]
	]
chooseBranch[ branch_, ps_] :=
	chooseBranch[ {branch}, ps]

getFormulaKeys[ ids_List, ps_PRFSIT$] :=
	With[ {assms = kb@ps, gl = goal@ps},
	With[ {goalKey = key@gl, goalLabel = label@gl, newAssms = getOptionalComponent[ ps, "newAssms"]},
		Replace[
			ids,
			{
				0|"goal"|STRING$[ goalLabel] -> goalKey,
				"this"|{"this"} :>
				Sequence @@ newAssms,
				{"this", i_Integer} :>
				Quiet[ Check[ newAssms[[i]], Sow[ {"Warning", "New assumption " <> ToString[ i] <> " does not exist."}, "Message"]; Null]],
				{"this", i_Integer, j_Integer} :>
				Quiet[ Check[ Sequence @@ newAssms[[i;;j]], Sow[ {"Warning", "New assumptions " <> ToString[ i] <> " through " <> ToString[ j] <> " do not exist."}, "Message"]; Null]],
				j_Integer :>
				Quiet[ Check[ key@(assms[[j]]), Sow[ {"Warning", "Assumption " <> ToString[ j] <> " does not exist."}, "Message"]; Null]],
				STRING$[ l_String] :>
				Quiet[ Check[ First[ Cases[ assms, _[ k_, _, l, ___] :> k, {1}, 1]], Sow[ {"Warning", "Formula \"" <> l <> "\" does not exist."}, "Message"]; Null]],
				EXPR$[ expr_] :>
					If[ alphaEquivalent[ expr, formula@gl],
						goalKey,
					(*else*)
						Quiet[ Check[ First[ Cases[ assms, _[ k_, f_, __] /; alphaEquivalent[ expr, f] :> k, {1}, 1]], Sow[ {"Warning", "Literal formula not found."}, "Message"]; Null]]
					]
			},
			{1}
		]
	]]

messageToString[ msg_String] := msg
messageToString[ {"Warning", msg_String}] := "WARNING: " <> msg
messageToString[ {"Error", msg_String}] := "ERROR: " <> msg
messageToString[ {typ_String, msg_String}] := typ <> ": " <> msg

displayProofSituation[ Hold[ pending_]][ $id_] :=
	With[ {pos = $TMAproofObject[[2]]},
		Switch[ pos,
			{},
			displayProofSituation[ $id, Null, Null, 0, pending],
			
			{__},
			With[ {ps = Extract[ First[ $TMAproofObject], First[ pos]]},
				displayProofSituation[ $id, kb@ps, goal@ps, Length[ pos], pending]
			],
			
			_,
			$Failed
		]
	]
displayProofSituation[ $id_, _, _, 0, _] :=
	With[ {statSize = CurrentValue[ SequentialEvaluation`Private`$stateWindow[$id], "WindowSize"]},
		Pane[
			TextCell[ "\[Checkmark] No open proof situations remaining", "Text"],
			Scrollbars -> {False, Automatic},
			AppearanceElements -> None,
			ImageSize -> {Full, Last[ statSize] - 90}
		]
	]
displayProofSituation[ $id_, kb_List, g_FML$, n_Integer, pending_] :=
	With[ {statSize = CurrentValue[ SequentialEvaluation`Private`$stateWindow[$id], "WindowSize"]},
		Column[
			{
				Pane[
					Grid[ Join[
							{{DisplayForm @ selectableExpressionBox[ formula@g], makeButtonLabel[ g, 0, $id]}},
							If[ kb === {},
								{{
									TextCell[ "No knowledge available", Italic, Gray],
									Row[ {}]
								}},
							(*else*)
								MapIndexed[
									Function[ {k, index},
										{DisplayForm @ selectableExpressionBox[ formula@k], makeButtonLabel[ k, First[ index], $id]}
									],
									kb
								]
							]
						],
						ColumnAlignments -> {Left, Right},
						RowLines -> {True, False},
						Spacings -> {Automatic, {1, 2}}
					],
					Scrollbars -> {False, Automatic},
					AppearanceElements -> None,
					ImageSize -> {Full, Last[ statSize] - 90}
				],
				TextCell[ ToString[ n] <> " subgoal" <> If[ n === 1, "", "s"] <> If[ TrueQ[ pending], "; command pending", ""], "Text"]
			},
			Left,
			2,
			ImageSize -> {Full, Last[ statSize] - 20}
		]
	]

printCurrentStatus[] :=
	With[ {pos = $TMAproofObject[[2]]},
		If[ MatchQ[ pos, {Except[ {}], ___}],
			{CellGroup[ {TextCell[ "Proof Situation", "Subsubsection"], Extract[ First[ $TMAproofObject], First[ pos]]}, 1]},
		(*else*)
			{}
		]
	]

exportProof[ goal_, assms_List, {po_, pos_List, _}, time_, num_Integer, _, flag_] :=
	If[ flag === 2 || pos === {},
		Theorema`Interface`GUI`Private`$selectedProofGoal = goal;
		Theorema`Interface`GUI`Private`$selectedProofKB = DeleteDuplicates[ Join[ assms, Theorema`Interface`GUI`Private`$addProofKB]];
		Scan[ With[ {fkey = key[ #]}, kbSelectProve[ fkey] = True]&, Theorema`Interface`GUI`Private`$addProofKB];
		$TMAproofObject = postprocessProofObject[ po, goal, Theorema`Interface`GUI`Private`$selectedProofKB];
		Theorema`Interface`GUI`Private`printProveInfo[
			Map[ {key[#], label[#]}&, Theorema`Interface`GUI`Private`$selectedProofKB],
			proofValue[ po],
			time,
			0,
			{False, False, False},
			num
		];
		Clear[ $TMAproofObject];
		ClearSystemCache[],
	(*else*)
		$Failed
	]
exportProof[ _, _List, _Integer, _, _Integer, _, _] :=
	$Failed

postprocessProofObject[ po_, goal_, assms_List] :=
	ReplacePart[ po, {1, 3, 1} -> Prepend[ assms, goal]]


(* ::Section:: *)
(*translate*)

(* TODO: Change when publish. *)
With[ {lang = "English"},

Theorema`Common`translate[ "assmNotFound", lang] = "Assumption \"`1`\" not found.";
Theorema`Common`translate[ "exportFailure", lang] = "Unable to export proof";
Theorema`Common`translate[ "exportSuccess", lang] = "Proof was exported successfully.";
Theorema`Common`translate[ "fmlNotFound", lang] = "Formula \"`1`\" not found.";
Theorema`Common`translate[ "goalNotFound", lang] = "Goal \"`1`\" not found.";
Theorema`Common`translate[ "prfError", lang] = "Command `2` in proof of formula `1` is not applicable.";
Theorema`Common`translate[ "prfNotebookSyntaxError", lang] = "Proof notebook cannot be parsed.";
Theorema`Common`translate[ "prfNotFound", lang] = "Proof `1` not found.";
Theorema`Common`translate[ "prfNumberInvalid", lang] = "Proof number `1` is invalid.";

]

End[]

EndPackage[]
