(* Package *)

BeginPackage["Theorema`Common`"]

With[ {Theorema`lang = "English"},
	MessageName[ EchoFile, "usage", Theorema`lang] = "'EchoFile' is similar to 'Echo', but directs the output to file \"log.txt\". It is used for debugging only.";
	MessageName[ EchoFileHold, "usage", Theorema`lang] = "'EchoFileHold' is the combination of 'EchoFile' and 'EchoHold'. It is used for debugging only.";
	MessageName[ EchoHold, "usage", Theorema`lang] = "'EchoHold' is similar to 'Echo', but prints its argument in unevaluated form. It is used for debugging only.";
	
	MessageName[ addToData, "usage", Theorema`lang] = "";
	MessageName[ extractSingle, "usage", Theorema`lang] = "";
	MessageName[ transposeEmpty, "usage", Theorema`lang] = "";
	MessageName[ updateData, "usage", Theorema`lang] = "";
	
	MessageName[ getNewRelevantSits, "usage", Theorema`lang] = "";
	MessageName[ sameCellQ, "usage", Theorema`lang] = "";
	
	MessageName[ compareTheoryPositions, "usage", Theorema`lang] = "";
	MessageName[ extractGoalInfoFromProofCell, "usage", Theorema`lang] = "";
	MessageName[ formulaKeys, "usage", Theorema`lang] = "";
	MessageName[ formulaPosition, "usage", Theorema`lang] = "";
	MessageName[ formulaFromKey, "usage", Theorema`lang] = "";
	MessageName[ getRegisteredProofNumber, "usage", Theorema`lang] = "";
	MessageName[ getRegisteredProofNumbersOfFile, "usage", Theorema`lang] = "";
	MessageName[ makeSourcePattern, "usage", Theorema`lang] = "";
	MessageName[ notebookFromFile, "usage", Theorema`lang] = "";
	MessageName[ selectFormulas, "usage", Theorema`lang] = "";
	MessageName[ splitSourceLabel, "usage", Theorema`lang] = "";
	
	MessageName[ makeLinkFunction, "usage", Theorema`lang] = "";
	MessageName[ msgHeaderToList, "usage", Theorema`lang] = "";
	MessageName[ notificationStyled, "usage", Theorema`lang] = "";
	MessageName[ tmaStringToSymbol, "usage", Theorema`lang] = "";
	MessageName[ tmaSymbolToString, "usage", Theorema`lang] = "";
	
	(* The following datatypes are used to distinguish literal facts from labeled facts and from strings. For instance, the expression
		'EXPR$[ "abc"]' represents the literal fact "abc", whereas the expression 'STRING$[ "abc"]' represents the string (label) "abc".
		Certain functions and proof methods can deal with arguments of either kind, so they need to know what is meant. *)
	MessageName[ EXPR$, "usage", Theorema`lang] = "EXPR$ is an auxiliary datatype used to represent literal facts. Used also in \"ProofScript.m\".";
	MessageName[ LABELED$, "usage", Theorema`lang] = "LABELED$ is an auxiliary datatype used to represent labeled facts. Used also in \"ProofScript.m\".";
	MessageName[ PEXPR$, "usage", Theorema`lang] = "PEXPR$ is a data structure containing an expression as well as an arbitrary number of pattern-expressions
													that shall be matched against the expressions. It is mainly used in proofs.";
	MessageName[ STRING$, "usage", Theorema`lang] = "STRING$ is an auxiliary datatype used to represent strings. Used also in \"ProofScript.m\".";
	MessageName[ VARIANT$, "lang", Theorema`lang] = "VARIANT$ is an auxiliary datatype used to represent variants of known facts.";
	
	MessageName[ $defaultFormulaCollection, "usage", Theorema`lang] = "";
	MessageName[ $registeredProofNumbers, "usage", Theorema`lang] = "";
	MessageName[ $retrieveFormulasFunction, "usage", Theorema`lang] = "";
	MessageName[ $privateSymbolsLoaded, "usage", Theorema`lang] = "";
	MessageName[ $symbolCombinationString, "usage", Theorema`lang] = "";
];

Begin["`Private`"]

If[ !ValueQ[ $defaultFormulaCollection], $defaultFormulaCollection = Hold[ $tmaEnv]];
If[ !ValueQ[ $registeredProofNumbers], $registeredProofNumbers = Association[]];
If[ !ValueQ[ $retrieveFormulasFunction], $retrieveFormulasFunction = {}&];
If[ !ValueQ[ $symbolCombinationString], $symbolCombinationString = "\[Breve]"];

SetAttributes[ extractSingle, HoldRest];
extractSingle[ expr_, pos_] :=
	If[ pos === {}, expr, Extract[ expr, pos]]
extractSingle[ expr_, pos_, def_] :=
	Quiet[ Check[ extractSingle[ expr, pos], def]]

transposeEmpty[ {}, n_Integer] :=
	Table[ {}, {n}]
transposeEmpty[ l_List, _] :=
	Transpose[ l]

(* neither 'Append' nor 'Join' works in all 4 cases *)
updateData[ a_List, b_List] :=
	Normal[ Join[ Association[ a], Association[ b]]]
updateData[ a_List, b_Association] :=
	Normal[ Join[ Association[ a], b]]
updateData[ a_List, b:(_Rule|_RuleDelayed)] :=
	Replace[ First[ FirstPosition[ a, _[ First[ b], _], {0}, {1}, Heads -> False]],
		{
			0 :> Append[ a, b],
			i_Integer :> ReplacePart[ a, i -> b]
		}
	]
updateData[ a_Association, b_List] :=
	Append[ a, b]
updateData[ a_Association, b_Association] :=
	Join[ a, b]
updateData[ a_Association, b:(_Rule|_RuleDelayed)] :=
	Append[ a, b]

SetAttributes[ addToData, HoldFirst];
addToData[ data_] :=
	data
addToData[ data_, {}|Association[]] :=
	data
addToData[ data_, new_] :=
	If[ ListQ[ data],
		data = updateData[ data, new],
	(*else*)
		AssociateTo[ data, new]	(* works in all 3 cases *)
	]

getNewRelevantSits[ node_, pos_List] :=
	With[ {psPos = Position[ node, _PRFSIT$],
			irrelevantNodes =
				Alternatives @@ (Append[ #, __]& /@ Position[ node, (Theorema`Provers`Common`Private`ANDNODE$|Theorema`Provers`Common`Private`ORNODE$)[ ___, Except[ pending]]])},
		If[ irrelevantNodes === Alternatives[],
			Join[ pos, #]& /@ psPos,
		(*else*)
			Cases[ psPos, p:Except[ irrelevantNodes] :> Join[ pos, p]]
		]
	]

sameCellQ[ {id1_String, src1_}, {id2_String, src2_}] :=
	src1 === src2 && extractCellID[ id1] === extractCellID[ id2]
sameCellQ[ id1_Integer, src1_String, {id2_String, src2_}] :=
	sourceLabel[ src1] === src2 && ToString[ id1] === extractCellID[ id2]

EchoFile[ expr_] := (expr >>> "log.txt"; expr)
EchoFile[ expr_, label_] := EchoFile[ expr, label, Identity]
EchoFile[ expr_, label_, fun_] := (PutAppend[ label, fun[ expr], "log.txt"]; expr)

SetAttributes[ EchoFileHold, HoldFirst];
EchoFileHold[ expr_] := (Hold[ expr] >>> "log.txt"; expr)
EchoFileHold[ expr_, label_] := EchoFileHold[ expr, label, Identity]
EchoFileHold[ expr_, label_, fun_] := (PutAppend[ label, fun[ Hold[ expr]], "log.txt"]; expr)

SetAttributes[ EchoHold, HoldFirst];
EchoHold[ expr_, rest___] := (Echo[ Hold[ expr], rest]; expr)


(* ::Section:: *)
(*Formulas*)

selectFormulas = retrieveFormulas[ Cases];
formulaKeys = retrieveFormulas[ key];
formulaPosition = retrieveFormulas[ Position];

formulaFromKey[ k_] :=
	formulaFromKey[ k, ReleaseHold[ $defaultFormulaCollection]]
formulaFromKey[ k_, coll:{___FML$}] :=
	Cases[ coll, FML$[ k, __], {1}, 1]
formulaFromKey[ k_, coll_] :=
	$retrieveFormulasFunction[ Cases, k, Blank[], 1, None, coll]

With[ {$sourcePrefix = "Source", $labelDelimiter = "::"},
	retrieveFormulas[ op_][ spec_] := retrieveFormulas[ op][ spec, Infinity, None, ReleaseHold[ $defaultFormulaCollection]];
	retrieveFormulas[ op_][ spec_, num_] := retrieveFormulas[ op][ spec, num, None, ReleaseHold[ $defaultFormulaCollection]];
	retrieveFormulas[ op_][ spec_, num_, evalSrc_] := retrieveFormulas[ op][ spec, num, evalSrc, ReleaseHold[ $defaultFormulaCollection]];
	retrieveFormulas[ key][ {src_, label_String}, num_, _, coll:{___FML$}] :=
		Cases[ coll, FML$[ k:{_String, makeSourcePattern[ src]}, _, label, ___] :> k, {1}, num, Heads -> False];
	retrieveFormulas[ key][ {src_, label_String}, num_, _, coll:{___FML$}, test_] :=
		Cases[ coll, f:FML$[ k:{_String, makeSourcePattern[ src]}, _, label, ___] :> k /; test[ f], {1}, num, Heads -> False];
	retrieveFormulas[ op_][ {src_, label_String}, num_, _, coll:{___FML$}] :=
		op[ coll, FML$[ {_String, makeSourcePattern[ src]}, _, label, ___], {1}, num, Heads -> False];
	retrieveFormulas[ op_][ {src_, label_String}, num_, _, coll:{___FML$}, test_] :=
		op[ coll, f:FML$[ {_String, makeSourcePattern[ src]}, _, label, ___] /; test[ f], {1}, num, Heads -> False];
	retrieveFormulas[ op_][ {src_, label_String}, num_, evalSrc_, coll_, test___] :=
		$retrieveFormulasFunction[ op, {_String, makeSourcePattern[ src]}, label, num, evalSrc, coll, test];
	retrieveFormulas[ op_][ spec_String, args___] :=
		retrieveFormulas[ op][ splitSourceLabel[ spec], args];
	
	splitSourceLabel[ spec_String] :=
		splitSourceLabel[ spec, Blank[]];
	splitSourceLabel[ spec_String, current_] :=
		Replace[ StringSplit[ spec, $labelDelimiter, 2],
			{
				{src_String, lbl_String} :> {If[ src === "Any", Blank[], src], lbl},
				_ :> {current, spec}
			}
		];
	
	makeSourcePattern[ src_] :=
		If[ !StringQ[ src],
			Blank[ String],
		(*else*)
			With[ {source = If[ StringMatchQ[ src, __ ~~ (".m"|".nb")], src, src <> ".nb"]},
			With[ {patt = If[ StringMatchQ[ source, $sourcePrefix ~~ __], source, $sourcePrefix ~~ __ ~~ source]},
				PatternTest[ Blank[ String], StringMatchQ[ #, patt]&]
			]]
		]
];

compareTheoryPositions[ {akey_, apos_}, {bkey_, bpos_}, thyScope_] :=
	Which[
		akey === bkey,
		Equal,
	(*else*)
		Last[ akey] === Last[ bkey],
		If[ apos === {} || bpos === {},
			Indeterminate,
		(*else*)
			comparePositions[ apos, bpos]
		],
	(*else*)
		TrueQ[ thyScope[ Last[ akey]]],
		Less,
	(*else*)
		True,
		Indeterminate
	]

comparePositions[ x_List, y_List] :=
	With[ {lx = Length[ x], ly = Length[ y]},
		Which[
			lx < ly,
			Catch[ MapThread[ Which[ #1 < #2, Throw[ Less], #2 < #1, Throw[ Greater], True, Null]&, {x, Take[ y, lx]}]; Less],
		(*else*)
			ly < lx,
			Catch[ MapThread[ Which[ #1 < #2, Throw[ Less], #2 < #1, Throw[ Greater], True, Null]&, {Take[ x, ly], y}]; Greater],
		(*else*)
			True,
			Catch[ MapThread[ Which[ #1 < #2, Throw[ Less], #2 < #1, Throw[ Greater], True, Null]&, {x, y}]; Equal]
		]
	]
comparePositions[ _, _] := Indeterminate

getRegisteredProofNumbersOfFile[ src_String] :=
	getRegisteredProofNumbersOfFile[ src, True, FileNameJoin[ {DirectoryName[ src], FileBaseName[ src]}]]
getRegisteredProofNumbersOfFile[ src_String, autoload_] :=
	getRegisteredProofNumbersOfFile[ src, autoload]
getRegisteredProofNumbersOfFile[ src_String, autoload_, dir_String] :=
	(* 'src' is the full file name of a notebook, without prefix "Source:". *)
	Replace[ Lookup[ $registeredProofNumbers, src, $Failed],
		Except[ _Association] :>
			Replace[ Quiet[ Check[ Get[ FileNameJoin[ {dir, "reg.m"}]], $Failed]],
				{
					assoc_Association :>
						(
							If[ TrueQ[ autoload], AssociateTo[ $registeredProofNumbers, src -> assoc]];
							assoc
						),
					_ -> $Failed
				}
			]
	]

getRegisteredProofNumber[ arg_] :=
	getRegisteredProofNumber[ arg, "-nb.nb"]
getRegisteredProofNumber[ FML$[ k_, __], filePatt_String] :=
	getRegisteredProofNumber[ k, filePatt]
getRegisteredProofNumber[ k_List, filePatt_String] :=
	With[ {src = sourceFile[ k]},
		getRegisteredProofNumber[ k, src, FileNameJoin[ {DirectoryName[ src], FileBaseName[ src]}], extractCellID[ k], filePatt]
	]
getRegisteredProofNumber[ k_List, src_String, dir_String, cid_String, filePatt_String] :=
	Replace[ getRegisteredProofNumbersOfFile[ src, True, dir],
		{
			assoc_Association :>
				Lookup[ assoc, Key[ k], getRegisteredProofNumber[ dir, cid, filePatt]],
			_ :>
				(
					AssociateTo[ $registeredProofNumbers, src -> Association[]];
					getRegisteredProofNumber[ dir, cid, filePatt]
				)
		}
	]
getRegisteredProofNumber[ dir_String, cid_String, filePatt_String] :=
	With[ {prefix = "p" <> cid <> "-"},
		Replace[
			Min[
				Map[
					ToExpression[ StringReplace[ #, __ ~~ prefix ~~ d:(DigitCharacter ..) ~~ filePatt :> d]] &,
					FileNames[ prefix ~~ (DigitCharacter..) ~~ filePatt, {dir}]
				]
			],
			Infinity -> False
		]
	]

notebookFromFile[ file_String] :=
	FirstCase[ Notebooks[],
		nb_NotebookObject /; Quiet[ Check[ NotebookFileName[ nb] === file, False]] :> {nb, False},
		If[ FileExistsQ[ file],
			{NotebookOpen[ file, Visible -> False], True},
		(*else*)
			{$Failed, False}
		]
	]

extractGoalInfoFromProofCell[] :=
	Replace[ ButtonNotebook[],
		{
			nb_NotebookObject :>
				(
					SelectionMove[ nb, All, ButtonCell];
					Replace[ NotebookRead[ nb],
						{
							Cell[ _, __, CellTags -> tags_, ___] :> {nb, extractGoalInfoFromProofCell[ Flatten[ {tags}]]},
							_ -> $Failed
						}
					]
				),
			_ -> $Failed
		}
	]
extractGoalInfoFromProofCell[ cell_CellObject] :=
	extractGoalInfoFromProofCell[ Flatten[ {CurrentValue[ cell, CellTags]}]]
extractGoalInfoFromProofCell[ tags_List] :=
	FirstCase[ tags,
		s_String /; StringMatchQ[ s, "Proof|ID" ~~ __ ~~ "-" ~~ (DigitCharacter..)] :>
			Replace[ StringSplit[ StringDrop[ s, 6], "-"],
				{
					{id_String, num_String} :> {id, num},
					_ -> $Failed
				}
			],
		$Failed
	]


(* ::Section:: *)
(*Notifications*)

notificationStyled[] :=
	notificationStyled[ "", None, True]
notificationStyled[ header_String] :=
	notificationStyled[ header, None, True]
notificationStyled[ printLine:(True|False)] :=
	notificationStyled[ "", None, printLine]
notificationStyled[ src_List] :=
	notificationStyled[ "", src, True]
notificationStyled[ header_String, src_List] :=
	notificationStyled[ header, src, True]
notificationStyled[ header_String, printLine:(True|False)] :=
	notificationStyled[ header, None, printLine]
notificationStyled[ src_List, printLine_] :=
	notificationStyled[ "", src, printLine]
notificationStyled[ header_String, src_, printLine_][ msg0__] :=
	If[ TrueQ[ $Notebooks],
		With[ {opts =
				If[ TrueQ[ printLine],
					{CellLabel -> ToString[ StringForm[ translate[ "duringEval"], $Line]], ShowCellLabel -> True},
				(*else*)
					{}
				],
				msg =
					Replace[ msgHeaderToList[ header, True],
						{
							{} :> StringForm[ msg0],
							l_List :> Row[ Append[ l, StringForm[ msg0]], Spacer[ 2]]
						}
					]},
			printToConsole[
				Cell[ BoxData[ ToBoxes[ makeLinkFunction[ src][ msg]]], "Print", Sequence @@ opts]
			]
		]
	]

msgHeaderToList[ header_String, boxes_] :=
	Which[
		header === "" || header === "Message",
		{},
	(*else*)
		boxes,
		{Style[ translate[ header],
			FontWeight -> Bold,
			FontColor -> Switch[ header, "Error"|"Fatal", Red, "Warning", Darker[ Orange, .2], "Info", Blue, _, Darker[ Gray]]
		]},
	(*else*)
		True,
		{translate[ header] <> ":"}
	]

makeLinkFunction[ {a_, b_}] :=
	makeLinkFunction[ a, b]
makeLinkFunction[ nb_NotebookObject, b_] :=
	makeLinkFunction[ Quiet[ Check[ NotebookFileName[ nb], $Failed]], b]
makeLinkFunction[ file_String, cell_CellObject] :=
	makeLinkFunction[ file, CurrentValue[ cell, CellID]]
makeLinkFunction[ file_String, cid_Integer] :=
	With[ {link = fmlHyperlink[ ">>", file, cid, CellID]},
		Replace[ #,
			{
				Row[ l_List, opts___] :> Row[ Append[ l, link], opts],
				bx_ :> Row[ {bx, link}, Spacer[ 2]]
			}
		]&
	]
makeLinkFunction[ __] :=
	Identity

tmaSymbolToString[ s_Symbol] := tmaSymbolToString[ SymbolName[ s]]
tmaSymbolToString[ s:(_Integer|_Rational|_Real|_Complex|_DirectedInfinity)] := ToString[ s]
tmaSymbolToString[ s_String] :=
	With[ {s0 = If[ StringMatchQ[ s, "VAR$" ~~ __], StringDrop[ s, 4], s]},
		Which[
			StringMatchQ[ s0, __ ~~ "$TM"],
			StringDrop[ s0, -3],
		(*else*)
			StringMatchQ[ s0, __ ~~ "$"],
			StringDrop[ s0, -1],
		(*else*)
			True,
			s0
		]
	]

tmaStringToSymbol[ s_String] :=
	Block[ {$ContextPath = Join[ {"Theorema`Language`"}, $TheoremaArchives, $ContextPath], $Context = "Theorema`Knowledge`"},
		With[ {s0 = If[ StringMatchQ[ s, __ ~~ ("$TM"|"$")], s, s <> "$TM"]},
			Symbol[ s0]
		]
	]

(* TODO: Change when publish. *)
With[ {lang = "English"},

Theorema`Common`translate[ "Error", lang] = "Error";
Theorema`Common`translate[ "Fatal", lang] = "ERROR";
Theorema`Common`translate[ "Info", lang] = "Info";
Theorema`Common`translate[ "Warning", lang] = "Warning";

]

$privateSymbolsLoaded = True;

End[]

EndPackage[]
