(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 11.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       158,          7]
NotebookDataLength[   1042239,      27594]
NotebookOptionsPosition[    736019,      21242]
NotebookOutlinePosition[    935617,      25503]
CellTagsIndexPosition[    917465,      25042]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["Set Theory", "Title",
 CellChangeTimes->{{3.730284631132609*^9, 3.7302846334784923`*^9}},
 CellID->365833305,ExpressionUUID->"a9b3133e-7cc9-46ff-ac44-1507e25d0b1a"],

Cell["\<\
Alexander Maletzky, RISC, JKU Linz
April 29, 2019\
\>", "Subsubtitle",
 CellChangeTimes->{{3.730284649535728*^9, 3.730284665107551*^9}, 
   3.7303567434581404`*^9, {3.7305445491584578`*^9, 3.7305445492327547`*^9}, 
   3.73061136235915*^9, 3.7314888052585883`*^9, {3.765008681199839*^9, 
   3.765008686006995*^9}, 3.765092226456935*^9, {3.7651801353252172`*^9, 
   3.765180135508604*^9}, {3.765264455711342*^9, 3.765264455751687*^9}, 
   3.7655099311729727`*^9},
 CellID->644169820,ExpressionUUID->"0ef2b046-6bef-49bb-a223-5aad136b4a98"],

Cell[TextData[{
 StyleBox["Note",
  FontWeight->"Bold"],
 ": Have a look at the \
\[OpenCurlyDoubleQuote]Examples\[CloseCurlyDoubleQuote] notebook in the \
\[OpenCurlyDoubleQuote]Documentation/Core+HOL+Interactive/\
\[CloseCurlyDoubleQuote] subdirectory of your Theorema installation."
}], "Text",
 CellChangeTimes->{{3.76525164545562*^9, 3.765251670607235*^9}, {
  3.7652517304623203`*^9, 3.765251756438222*^9}},
 CellID->1791917602,ExpressionUUID->"b5abd81b-7369-46a0-9261-4dcb0df35cf8"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"$RecursionLimit", "=", "2048"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"Needs", "[", "\"\<Theorema`Core`\>\"", "]"}]}], "Input",
 CellChangeTimes->{{3.730284673537689*^9, 3.730284681046693*^9}, {
  3.765251629248996*^9, 3.765251630369158*^9}},
 CellLabel->"In[1]:=",
 CellID->265914558,ExpressionUUID->"4f1f2d8e-c48d-4274-9cfe-b57e907c12d8"],

Cell[BoxData[
 RowBox[{"import", " ", "\"\<HOL\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.730284703443159*^9, 3.7302847103610373`*^9}},
 CellLabel->"In[3]:=",
 CellID->441817464,ExpressionUUID->"4127ed3e-f512-414b-ac69-880b74601b7f"],

Cell[BoxData[
 RowBox[{"decl", " ", 
  RowBox[{"\[OpenCurlyQuote]", 
   RowBox[{"Minus", "::", 
    RowBox[{"(", 
     RowBox[{"a_", "\[Rule]", "a_"}], ")"}]}], "\[CloseCurlyQuote]"}], " ", 
  RowBox[{"\[OpenCurlyQuote]", 
   RowBox[{"Plus", "::", 
    RowBox[{"(", 
     RowBox[{"a_", "\[Rule]", 
      RowBox[{"a_", "\[Rule]", "a_"}]}], ")"}]}], "\[CloseCurlyQuote]"}], " ", 
  RowBox[{"\[OpenCurlyQuote]", 
   RowBox[{"Subtract", "::", 
    RowBox[{"(", 
     RowBox[{"a_", "\[Rule]", 
      RowBox[{"a_", "\[Rule]", "a_"}]}], ")"}]}], 
   "\[CloseCurlyQuote]"}]}]], "TheoremaCode",
 CellChangeTimes->{{3.7302860101421704`*^9, 3.730286054364209*^9}},
 CellLabel->"In[4]:=",
 CellID->38326797,ExpressionUUID->"2c3959e9-507d-4a08-8e71-abcb54f1b1bb"],

Cell["\<\
This theory is modeled after \[OpenCurlyDoubleQuote]Set.thy\
\[CloseCurlyDoubleQuote] of the Isabelle/HOL proof assistant.\
\>", "Text",
 CellChangeTimes->{{3.765182148939528*^9, 3.76518217820324*^9}},
 CellID->537343700,ExpressionUUID->"03254130-9804-400f-8000-122d1989abfe"],

Cell[CellGroupData[{

Cell["Basic Definitions", "Section",
 CellChangeTimes->{{3.7302850572213745`*^9, 3.7302850600546775`*^9}},
 CellID->224383646,ExpressionUUID->"a5436c35-1f68-402a-962c-c5c418d1fe70"],

Cell[BoxData[
 RowBox[{
  RowBox[{"typedef", " ", 
   RowBox[{"\[OpenCurlyQuote]", 
    RowBox[{"SET", "[", "a_", "]"}], "\[CloseCurlyQuote]"}]}], " ", ":=", " ", 
  RowBox[{
   RowBox[{"\[OpenCurlyQuote]", 
    RowBox[{"a_", " ", "\[Rule]", " ", "BOOL"}], "\[CloseCurlyQuote]"}], " ", 
   RowBox[{"(", 
    RowBox[{"morphisms", " ", 
     RowBox[{"\[OpenCurlyQuote]", "Collect", "\[CloseCurlyQuote]"}], " ", 
     RowBox[{"\[OpenCurlyQuote]", "ReverseElement", "\[CloseCurlyQuote]"}]}], 
    ")"}]}]}]], "TheoremaCode",
 CellChangeTimes->{{3.7302850712655325`*^9, 3.730285093154998*^9}, {
  3.7302851743448153`*^9, 3.7302851848538723`*^9}, {3.73028534436603*^9, 
  3.730285355007896*^9}, {3.7302854001737967`*^9, 3.730285434313673*^9}},
 CellLabel->"In[5]:=",
 CellID->1023665245,ExpressionUUID->"4611e6d7-af3c-474e-aabd-e9db75f190d5"],

Cell["", "OpenEnvironment",
 CellID->639499678,ExpressionUUID->"24d6f0fa-57c1-4851-ba7e-68d350fd2d4d"],

Cell[CellGroupData[{

Cell["Definition (Basic Operations on Sets)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.730285566275353*^9, 3.7302855705411243`*^9}},
 CellID->102108795,ExpressionUUID->"ecc598b7-1902-44cd-8a91-8d65c055822a"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "A"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", "\[Element]", "A"}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a:=b"], 
    RowBox[{"A", "\[ReverseElement]", "a"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Element] def)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7302855822017612`*^9, 3.730285608760703*^9}},
 CellTags->"\[Element] def",
 CellLabel->"In[6]:=",
 CellID->464571198,ExpressionUUID->"e62578c8-e0a6-4737-b48a-89439b421a97"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a:=b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", "a"], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"a", "\[Element]", "A"}], "\[Implies]", 
        RowBox[{"a", "\[Element]", "B"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[SubsetEqual] def)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730285654866276*^9, 3.730285705779932*^9}},
 CellTags->"\[SubsetEqual] def",
 CellLabel->"In[7]:=",
 CellID->86239678,ExpressionUUID->"962bc855-b87d-497f-828c-d9a1f6dcc296"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[Subset]", "B"}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a:=b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"A", "\[SubsetEqual]", "B"}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[Not]", 
        RowBox[{"B", "\[SubsetEqual]", "A"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Subset] def)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7302857439875154`*^9, 3.730285789161359*^9}},
 CellTags->"\[Subset] def",
 CellLabel->"In[8]:=",
 CellID->169597205,ExpressionUUID->"8386411c-09c9-4758-8f4c-116b52c47684"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"Pow", "[", "A", "]"}], ":=", 
    RowBox[{"Collect", "[", 
     RowBox[{
      UnderscriptBox["\[Lambda]", "B"], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"B", "\[SubsetEqual]", "A"}], 
       TagBox[")",
        "AutoParentheses"]}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Pow def)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303568308654103`*^9, 3.7303568774483604`*^9}},
 CellTags->"Pow def",
 CellLabel->"In[9]:=",
 CellID->15464031,ExpressionUUID->"903ff798-45d2-49ee-9483-412b4271e487"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[Intersection]", "B"}], ":=", 
    RowBox[{"Collect", "[", 
     RowBox[{
      UnderscriptBox["\[Lambda]", "x"], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"x", "\[Element]", "A"}], "\[And]", 
        RowBox[{"x", "\[Element]", "B"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection] def)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7302858139054527`*^9, 3.730285887069356*^9}},
 CellTags->"\[Intersection] def",
 CellLabel->"In[10]:=",
 CellID->755145795,ExpressionUUID->"bab1ffcc-3be2-40cc-8c31-516831c21a46"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[Union]", "B"}], ":=", 
    RowBox[{"Collect", "[", 
     RowBox[{
      UnderscriptBox["\[Lambda]", "x"], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"x", "\[Element]", "A"}], "\[Or]", 
        RowBox[{"x", "\[Element]", "B"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union] def)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7302859017413387`*^9, 3.7302859069762526`*^9}},
 CellTags->"\[Union] def",
 CellLabel->"In[11]:=",
 CellID->157538211,ExpressionUUID->"331436ef-1768-43c6-a0e6-dd4797dde600"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"-", "A"}], ":=", 
    RowBox[{"Collect", "[", 
     RowBox[{
      UnderscriptBox["\[Lambda]", "x"], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[Not]", 
        RowBox[{"x", "\[Element]", "A"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Minus set def)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7302860676691875`*^9, 3.730286094526515*^9}},
 CellTags->"Minus set def",
 CellLabel->"In[12]:=",
 CellID->375270853,ExpressionUUID->"839c14a1-7e40-49b4-866c-82bfae447357"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "-", "B"}], ":=", 
    RowBox[{"A", "\[Intersection]", 
     RowBox[{"(", 
      RowBox[{"-", "B"}], ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(- set def)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7302861763479967`*^9, 3.7302862030708103`*^9}, {
  3.730286243643827*^9, 3.73028624980927*^9}},
 CellTags->"- set def",
 CellLabel->"In[13]:=",
 CellID->559936356,ExpressionUUID->"6a8c6292-cb26-4cf3-9299-3dca65b6f0ad"],

Cell[BoxData[
 RowBox[{"UNIV", ":=", 
  RowBox[{"Collect", "[", 
   RowBox[{
    UnderscriptBox["\[Lambda]", "x"], "True"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(UNIV def)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730286287021017*^9, 3.730286305742425*^9}},
 CellTags->"UNIV def",
 CellLabel->"In[14]:=",
 CellID->426543191,ExpressionUUID->"1bad0884-8923-430a-9e48-8ee7cdd1b0ec"],

Cell[BoxData[
 RowBox[{"EMPTY", ":=", 
  RowBox[{"Collect", "[", 
   RowBox[{
    UnderscriptBox["\[Lambda]", "x"], "False"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(EMPTY def)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7302863257264557`*^9, 3.730286336918466*^9}},
 CellTags->"EMPTY def",
 CellLabel->"In[15]:=",
 CellID->217880718,ExpressionUUID->"dce4d230-1c21-4c60-8667-799c9b8fd815"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "A"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"Insert", "[", 
     RowBox[{"a", ",", "A"}], "]"}], ":=", 
    RowBox[{"Collect", "[", 
     RowBox[{
      UnderscriptBox["\[Lambda]", "x"], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"x", "\[Equal]", "a"}], "\[Or]", 
        RowBox[{"x", "\[Element]", "A"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Insert def)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730286381460184*^9, 3.73028643873982*^9}},
 CellTags->"Insert def",
 CellLabel->"In[16]:=",
 CellID->394538268,ExpressionUUID->"ce7cf6be-50c3-49f4-8235-e3df79addb62"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "A"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"f", "@", "A"}], ":=", 
    RowBox[{"Collect", "[", 
     RowBox[{
      UnderscriptBox["\[Lambda]", "x"], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox["\[Exists]", 
         RowBox[{"y", "\[Element]", "A"}]], 
        RowBox[{"x", "\[Equal]", 
         RowBox[{"f", "[", "y", "]"}]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(@ def)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765098215013502*^9, 3.765098261629326*^9}, {
  3.765098375059608*^9, 3.765098395531435*^9}, {3.7650984698262453`*^9, 
  3.765098477346416*^9}},
 CellTags->"@ def",
 CellLabel->"In[17]:=",
 CellID->469719290,ExpressionUUID->"e6744ab5-9386-45ac-a300-eaa9ae40a920"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"P", ",", "f"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"SetOf", "[", 
     RowBox[{"P", ",", "f"}], "]"}], ":=", 
    RowBox[{"Collect", "[", 
     RowBox[{
      UnderscriptBox["\[Lambda]", "x"], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox["\[Exists]", 
         RowBox[{"P", "[", "u", "]"}]], 
        RowBox[{"x", "\[Equal]", 
         RowBox[{"f", "[", "u", "]"}]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(SetOf def)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7302883338568983`*^9, 3.730288419431864*^9}},
 CellTags->"SetOf def",
 CellLabel->"In[18]:=",
 CellID->312695955,ExpressionUUID->"0cad2b9d-2d0a-4eb0-9090-3c81a65f8eca"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->556256564,ExpressionUUID->"7a41264a-2f82-4561-acc1-3b324a8f2168"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->227314424,ExpressionUUID->"0cf713e8-dfe9-403e-ba40-61f0e48a1abf"],

Cell[BoxData[
 RowBox[{
  RowBox[{"notation", " ", 
   RowBox[{"\[OpenCurlyQuote]", "Element", "\[CloseCurlyQuote]"}], " ", 
   RowBox[{"\[OpenCurlyQuote]", "Subset", "\[CloseCurlyQuote]"}], " ", 
   RowBox[{"\[OpenCurlyQuote]", "SubsetEqual", "\[CloseCurlyQuote]"}]}], ":", 
  " ", 
  RowBox[{"binary", " ", 
   RowBox[{
   "\[OpenCurlyQuote]", "And", "\[CloseCurlyQuote]"}]}]}]], "TheoremaCode",
 CellChangeTimes->{3.730546532002593*^9},
 CellLabel->"In[19]:=",
 CellID->1335265679,ExpressionUUID->"1fbd38a4-1162-45e5-b5af-ec2820303f0a"],

Cell[BoxData[
 RowBox[{
  RowBox[{"notation", " ", 
   RowBox[{"\[OpenCurlyQuote]", "SetOf", "\[CloseCurlyQuote]"}]}], ":", " ", 
  "binder"}]], "TheoremaCode",
 CellChangeTimes->{3.730546539018053*^9},
 CellLabel->"In[20]:=",
 CellID->428073695,ExpressionUUID->"8b63823c-e860-4837-8842-556b21924da3"],

Cell["\<\
We must add the \[OpenCurlyDoubleQuote]$TM\[CloseCurlyDoubleQuote]-suffix \
below, because standard-Theorema translates plain \[OpenCurlyDoubleQuote]Set\
\[CloseCurlyDoubleQuote] into \[OpenCurlyDoubleQuote]Assign\
\[CloseCurlyDoubleQuote] ...\
\>", "Text",
 CellChangeTimes->{{3.730547541035254*^9, 3.73054756753965*^9}},
 CellID->1333340164,
 ExpressionUUID -> "422b6fbe-f7f9-48af-a839-b22e536716c8"],

Cell[BoxData[
 RowBox[{
  RowBox[{"notation", " ", 
   RowBox[{"\[OpenCurlyQuote]", "Set$TM", "\[CloseCurlyQuote]"}]}], ":", " ", 
  RowBox[{"recursive", " ", 
   RowBox[{"\[OpenCurlyQuote]", "EMPTY", "\[CloseCurlyQuote]"}], " ", 
   RowBox[{"\[OpenCurlyQuote]", "Insert", "\[CloseCurlyQuote]"}], " ", 
   "right"}]}]], "TheoremaCode",
 CellChangeTimes->{{3.730286489404684*^9, 3.730286515636716*^9}, {
   3.730286607849*^9, 3.7302866463424544`*^9}, 3.7302867501951456`*^9, {
   3.730288377622142*^9, 3.730288377903418*^9}, {3.730288446455617*^9, 
   3.7302884490808644`*^9}, {3.73054640685299*^9, 3.730546409653335*^9}, {
   3.730546537617404*^9, 3.73054654063332*^9}, {3.730547186103766*^9, 
   3.73054721823201*^9}, {3.73054743705905*^9, 3.730547451394845*^9}},
 CellLabel->"In[21]:=",
 CellID->811751068,ExpressionUUID->"98ee92b3-488a-4383-af32-2881e63b7b60"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"abbreviation", " ", 
   RowBox[{"\[OpenCurlyQuote]", "NotElement", "\[CloseCurlyQuote]"}]}], " ", 
  "\[Equilibrium]", " ", 
  RowBox[{"\[OpenCurlyQuote]", 
   RowBox[{
    UnderscriptBox["\[Lambda]", 
     RowBox[{"a", ",", "A"}]], 
    RowBox[{"(", 
     RowBox[{"\[Not]", 
      RowBox[{"a", "\[Element]", "A"}]}], ")"}]}], 
   "\[CloseCurlyQuote]"}]}], "\n", 
 RowBox[{
  RowBox[{"abbreviation", " ", 
   RowBox[{"\[OpenCurlyQuote]", "NotSubsetEqual", "\[CloseCurlyQuote]"}]}], 
  " ", "\[Equilibrium]", " ", 
  RowBox[{"\[OpenCurlyQuote]", 
   RowBox[{
    UnderscriptBox["\[Lambda]", 
     RowBox[{"A", ",", "B"}]], 
    RowBox[{"(", 
     RowBox[{"\[Not]", 
      RowBox[{"A", "\[SubsetEqual]", "B"}]}], ")"}]}], 
   "\[CloseCurlyQuote]"}]}], "\n", 
 RowBox[{
  RowBox[{"abbreviation", " ", 
   RowBox[{"\[OpenCurlyQuote]", "NotSubset", "\[CloseCurlyQuote]"}]}], " ", 
  "\[Equilibrium]", " ", 
  RowBox[{"\[OpenCurlyQuote]", 
   RowBox[{
    UnderscriptBox["\[Lambda]", 
     RowBox[{"A", ",", "B"}]], 
    RowBox[{"(", 
     RowBox[{"\[Not]", 
      RowBox[{"A", "\[Subset]", "B"}]}], ")"}]}], 
   "\[CloseCurlyQuote]"}]}], "\n", 
 RowBox[{
  RowBox[{"abbreviation", " ", 
   RowBox[{"\[OpenCurlyQuote]", "SupersetEqual", "\[CloseCurlyQuote]"}]}], 
  " ", "\[RightVector]", " ", 
  RowBox[{"\[OpenCurlyQuote]", 
   RowBox[{
    UnderscriptBox["\[Lambda]", 
     RowBox[{"A", ",", "B"}]], 
    RowBox[{"(", 
     RowBox[{"B", "\[SubsetEqual]", "A"}], ")"}]}], 
   "\[CloseCurlyQuote]"}]}], "\n", 
 RowBox[{
  RowBox[{"abbreviation", " ", 
   RowBox[{"\[OpenCurlyQuote]", "Superset", "\[CloseCurlyQuote]"}]}], " ", 
  "\[RightVector]", " ", 
  RowBox[{"\[OpenCurlyQuote]", 
   RowBox[{
    UnderscriptBox["\[Lambda]", 
     RowBox[{"A", ",", "B"}]], 
    RowBox[{"(", 
     RowBox[{"B", "\[Subset]", "A"}], ")"}]}], "\[CloseCurlyQuote]"}]}], "\n", 
 RowBox[{
  RowBox[{"abbreviation", " ", 
   RowBox[{"\[OpenCurlyQuote]", "NotSuperset", "\[CloseCurlyQuote]"}]}], " ", 
  "\[RightVector]", " ", 
  RowBox[{"\[OpenCurlyQuote]", 
   RowBox[{
    UnderscriptBox["\[Lambda]", 
     RowBox[{"A", ",", "B"}]], 
    RowBox[{"(", 
     RowBox[{"\[Not]", 
      RowBox[{"B", "\[SubsetEqual]", "A"}]}], ")"}]}], 
   "\[CloseCurlyQuote]"}]}], "\n", 
 RowBox[{
  RowBox[{"abbreviation", " ", 
   RowBox[{"\[OpenCurlyQuote]", "NotSuperset", "\[CloseCurlyQuote]"}]}], " ", 
  "\[RightVector]", " ", 
  RowBox[{"\[OpenCurlyQuote]", 
   RowBox[{
    UnderscriptBox["\[Lambda]", 
     RowBox[{"A", ",", "B"}]], 
    RowBox[{"(", 
     RowBox[{"\[Not]", 
      RowBox[{"B", "\[Subset]", "A"}]}], ")"}]}], 
   "\[CloseCurlyQuote]"}]}], "\n", 
 RowBox[{
  RowBox[{"abbreviation", " ", 
   RowBox[{"\[OpenCurlyQuote]", "Backslash", "\[CloseCurlyQuote]"}]}], " ", 
  "\[RightVector]", " ", 
  RowBox[{"\[OpenCurlyQuote]", 
   RowBox[{"Subtract", "::", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"SET", "[", "a_", "]"}], "\[Rule]", 
      RowBox[{
       RowBox[{"SET", "[", "a_", "]"}], "\[Rule]", 
       RowBox[{"SET", "[", "a_", "]"}]}]}], ")"}]}], 
   "\[CloseCurlyQuote]"}]}]}], "TheoremaCode",
 CellChangeTimes->{{3.7302866584656057`*^9, 3.7302866930396214`*^9}, {
  3.7302867938933744`*^9, 3.7302869047246013`*^9}, {3.7302869455659094`*^9, 
  3.73028696550924*^9}, {3.7302870512985144`*^9, 3.730287107476242*^9}, {
  3.7303589705735435`*^9, 3.7303590163076553`*^9}, {3.73054478673948*^9, 
  3.7305447879634457`*^9}},
 CellLabel->"In[22]:=",
 CellID->501147791,ExpressionUUID->"75b4e402-a521-479a-86f4-2c7e87e93ace"],

Cell[CellGroupData[{

Cell["Parsing and Printing", "Subsection",
 CellChangeTimes->{{3.730544904018721*^9, 3.730544908114811*^9}},
 CellID->1880070792,ExpressionUUID->"675abfcf-7eb5-49a1-9cc1-153cbc76f1d7"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"BeginCore", "[", " ", "\"\<Theorema`HOL`\>\"", "]"}], "\n"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"addParsingRules", "[", " ", 
    RowBox[{"\"\<type\>\"", ",", "\n", "\t", 
     RowBox[{"Set$TM", " ", "\[Rule]", " ", "Theorema`Knowledge`SET$TM"}], 
     ",", "\n", "\t", "All", ",", "\n", "\t", "Back"}], "\n", "]"}], ";"}], 
  "\n"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"addToConstPattern", "[", " ", 
    RowBox[{"$typeHeads", ",", " ", "Theorema`Knowledge`TSETRNG$"}], "]"}], 
   ";"}], "\n"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"addPrintingRules", "[", " ", 
    RowBox[{"\"\<type\>\"", ",", "\n", "\t", 
     RowBox[{"Theorema`Knowledge`SET$TM", " ", "\[Rule]", " ", "Set$TM"}], 
     ",", "\n", "\t", "All", ",", "\n", "\t", "Back"}], "\n", "]"}], ";"}], 
  "\n"}], "\n", 
 RowBox[{"EndCore", "[", "]"}]}], "Code",
 CellChangeTimes->{{3.7305449289693747`*^9, 3.730544959916444*^9}, {
  3.73054499101587*^9, 3.7305450074849052`*^9}, {3.730545241439549*^9, 
  3.7305452779274197`*^9}, {3.73054533831301*^9, 3.730545363619904*^9}, {
  3.73054670604279*^9, 3.730546732587851*^9}},
 CellLabel->"In[23]:=",
 CellID->875851793,ExpressionUUID->"8d1d744e-1bd0-41cd-b05e-282671d25f26"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Basic Operations", "Section",
 CellChangeTimes->{{3.7302892481024427`*^9, 3.730289254105051*^9}},
 CellID->72433663,ExpressionUUID->"3e61c363-b8c6-4f55-aaf9-b8cae1db42f3"],

Cell[CellGroupData[{

Cell["Collect", "Subsection",
 CellChangeTimes->{{3.765018663362029*^9, 3.76501866584226*^9}},
 CellID->1814441028,ExpressionUUID->"dfc54ba8-8641-4008-add0-a24265bb944d"],

Cell["", "OpenEnvironment",
 CellID->639519359,ExpressionUUID->"43482684-a9f1-43ad-b7fc-f9a839feb641"],

Cell[CellGroupData[{

Cell["Lemma (Properties of Collect)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.765018680314287*^9, 3.765018684177918*^9}},
 CellID->828673570,ExpressionUUID->"b375ea09-9d2e-4e37-b575-cc7fcb241585"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"x", ",", "P"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"x", "\[Element]", 
      RowBox[{"Collect", "[", "P", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{"P", "[", "x", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Element]Collect)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765018687602034*^9, 3.765018713097361*^9}},
 CellTags->"\[Element]Collect",
 CellLabel->"In[28]:=",
 CellID->270104195,ExpressionUUID->"346c1ec5-297b-426b-a764-6d4ceb56a4e8"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"Collect", "[", 
     RowBox[{
      UnderscriptBox["\[Lambda]", "x"], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"x", "\[Element]", "A"}], 
       TagBox[")",
        "AutoParentheses"]}]}], "]"}], "\[Equal]", "A"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Collect ident)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765100884460264*^9, 3.765100911812031*^9}},
 CellTags->"Collect ident",
 CellLabel->"In[29]:=",
 CellID->893774303,ExpressionUUID->"cdd465a6-9272-4aa3-980d-ab6053ec12e3"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1190941598,ExpressionUUID->"edd89845-6d3c-4630-b07f-d96f82425765"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->441174219,ExpressionUUID->"703bd2cc-6bf2-43bd-b8fc-cee7d0754941"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"b2db075d-77b1-438a-9a21-948ece49b50d"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Element]Collect)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Element]Collect)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["270104195", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:270104195", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:36, 0.058319 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:270104195-1", "3"},
 CellLabel->"In[30]:=",
 CellID->1363674537,ExpressionUUID->"5aad6e8a-a738-407b-bc9f-9ca36dc727e9"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"31bdb609-b45a-4950-9747-8c0268ab0625"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Collect ident)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(Collect ident)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["893774303", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:893774303", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:37, 0.044785 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:893774303-1", "3"},
 CellLabel->"In[31]:=",
 CellID->1532689697,ExpressionUUID->"26847eea-8373-4283-8a9d-d1e9558f9cde"],

Cell[BoxData[
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", "\"\<\[Element]Collect\>\"", " ", 
  "\"\<Collect ident\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.765018751307069*^9, 3.765018756913117*^9}, {
  3.765100928988778*^9, 3.765100932443769*^9}},
 CellLabel->"In[32]:=",
 CellID->864640627,ExpressionUUID->"1a27c0b9-a50b-4961-8848-e4d073b0d107"]
}, Closed]],

Cell[CellGroupData[{

Cell["Subsets", "Subsection",
 CellChangeTimes->{{3.730289264936702*^9, 3.730289272970025*^9}, {
  3.76501768224648*^9, 3.7650176823982973`*^9}},
 CellID->58512089,ExpressionUUID->"b06a8a2d-d680-4e9e-afa8-212a5e7d816e"],

Cell["", "OpenEnvironment",
 CellID->179550781,ExpressionUUID->"e4fa0d71-e146-423a-8659-d1459de1cdaa"],

Cell[CellGroupData[{

Cell["Lemma (Properties of \[SubsetEqual])", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.7303521519811893`*^9, 3.730352156150138*^9}},
 CellID->268601953,ExpressionUUID->"dc7f7121-0d67-41cc-b367-efa704e052f9"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      UnderscriptBox["\[And]", "a"], 
      RowBox[{
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"a", "\[Element]", "A"}], 
        TagBox[")",
         "AutoParentheses"]}], "\[LongRightArrow]", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"a", "\[Element]", "B"}], 
        TagBox[")",
         "AutoParentheses"]}]}]}], ")"}], "\[LongRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", "\[SubsetEqual]", "B"}], 
     TagBox[")",
      "AutoParentheses"]}]}], ")"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(\[SubsetEqual]I)", "FrameLabel"], Evaluator -> Automatic,
           Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "intro"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730352158957178*^9, 3.7303522070671053`*^9}},
 CellTags->{"\[SubsetEqual]I", "intro"},
 CellLabel->"In[33]:=",
 CellID->334402131,ExpressionUUID->"7031efb8-33bc-4ea6-bb2f-20619acc517e"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "a"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"a", "\[Element]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"a", "\[Element]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(\[SubsetEqual]D)", "FrameLabel"], Evaluator -> Automatic,
           Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "trans, intro"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730352276804055*^9, 3.73035231822104*^9}},
 CellTags->{"\[SubsetEqual]D", "trans", "intro"},
 CellLabel->"In[34]:=",
 CellID->591515539,ExpressionUUID->"fb900f0f-ac9a-44d1-979c-f15fb5f5be35"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "a"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"a", "\[Element]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"a", "\[Element]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(\[SubsetEqual]D reverse)", "FrameLabel"], Evaluator -> 
          Automatic, Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "trans, intro"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303523430051394`*^9, 3.7303523641923275`*^9}},
 CellTags->{"\[SubsetEqual]D reverse", "trans", "intro"},
 CellLabel->"In[35]:=",
 CellID->50542765,ExpressionUUID->"e51dde54-c9d6-46ad-abea-f80296524d5c"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x", ",", "P"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"x", "\[NotElement]", "A"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", "P"}], ")"}], 
   "\[LongRightArrow]", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"x", "\[Element]", "A"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"x", "\[Element]", "B"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", "P"}], ")"}], 
   "\[LongRightArrow]", "P"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(\[SubsetEqual]E classical)", "FrameLabel"], Evaluator -> 
          Automatic, Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "elim"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730352429823666*^9, 3.730352540653541*^9}, {
  3.730352723591702*^9, 3.7303527371733665`*^9}},
 CellTags->{"\[SubsetEqual]E classical", "elim"},
 CellLabel->"In[36]:=",
 CellID->389797744,ExpressionUUID->"fb93893a-e64d-4dfe-ae4e-00ff561e6191"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", "\[SubsetEqual]", "B"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{"a", "\[Element]", "A"}]], 
      RowBox[{"a", "\[Element]", "B"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[SubsetEqual] alt)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303527861601715`*^9, 3.730352807839546*^9}, 
   3.7303531683803854`*^9, {3.7303532117487345`*^9, 3.7303532168024073`*^9}},
 CellTags->"\[SubsetEqual] alt",
 CellLabel->"In[37]:=",
 CellID->223583403,ExpressionUUID->"b00cb866-8d95-456e-b04f-0dc3663f9abf"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[NotElement]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[NotElement]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(contraposition \[SubsetEqual]D)", "FrameLabel"], 
         Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303532400107384`*^9, 3.730353284887144*^9}},
 CellTags->"contraposition \[SubsetEqual]D",
 CellLabel->"In[38]:=",
 CellID->324850200,ExpressionUUID->"81ae2f05-9998-4704-978b-a1ea11dc084a"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{"A", "\[SubsetEqual]", "A"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[SubsetEqual] refl)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730353355361144*^9, 3.7303533625066557`*^9}},
 CellTags->"\[SubsetEqual] refl",
 CellLabel->"In[39]:=",
 CellID->232739861,ExpressionUUID->"47260c40-dd6d-4c60-8569-2246238f2cef"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"B", "\[SubsetEqual]", "C"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "C"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(\[SubsetEqual] trans)", "FrameLabel"], Evaluator -> 
          Automatic, Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "trans"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303533804543943`*^9, 3.7303534103958693`*^9}},
 CellTags->{"\[SubsetEqual] trans", "trans"},
 CellLabel->"In[40]:=",
 CellID->373774609,ExpressionUUID->"d95e141f-73bd-461e-b3e2-1457b1c31792"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->58488055,ExpressionUUID->"ed87274a-1db5-4b7c-b265-0710a3a11d35"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->666809684,ExpressionUUID->"3bf15b2d-c8b2-428f-8ece-a9d38046fe20"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"316356eb-a284-4a7d-a7e4-ee1247b284a3"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[SubsetEqual]I)\"", "FormReference", StripOnInput -> False],
             True -> StyleBox[
             "\"(\[SubsetEqual]I)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["334402131", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:334402131", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:38, 0.116705 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:334402131-1", "5"},
 CellLabel->"In[41]:=",
 CellID->624024459,ExpressionUUID->"300c87b6-2c78-4afd-9a22-f6e65b245b94"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"b295c11c-94d0-4be6-b3aa-302425de4d95"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[SubsetEqual]D)\"", "FormReference", StripOnInput -> False],
             True -> StyleBox[
             "\"(\[SubsetEqual]D)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["591515539", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:591515539", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:38, 0.142877 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:591515539-1", "5"},
 CellLabel->"In[42]:=",
 CellID->1850605281,ExpressionUUID->"c1bf422f-8799-415e-8d2a-f931451d75c6"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"d9373cd6-ce36-40ec-acd3-522085b36f4a"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[SubsetEqual]D reverse)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[SubsetEqual]D reverse)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["50542765", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:50542765", CellTags]], Appearance -> 
          None, Evaluator -> Automatic, Method -> "Preemptive"]]], 
       " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:39, 0.023384 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:50542765-1", "5"},
 CellLabel->"In[43]:=",
 CellID->940041033,ExpressionUUID->"fbf77b68-1b7c-497c-86af-dfad1033b410"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"4afb812d-3a6b-43ce-8927-23dff9b4d42d"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[SubsetEqual]E classical)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[SubsetEqual]E classical)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["389797744", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:389797744", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:39, 0.083955 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:389797744-1", "5"},
 CellLabel->"In[44]:=",
 CellID->1780566989,ExpressionUUID->"b5a607eb-73fe-4dbb-a4e6-6a0da3ab04fd"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"c0470af1-eab7-425b-a571-2cb84d8ec64d"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[SubsetEqual] alt)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[SubsetEqual] alt)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["223583403", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:223583403", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:39, 0.072305 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:223583403-1", "5"},
 CellLabel->"In[45]:=",
 CellID->1680503574,ExpressionUUID->"a6e2338e-aaf5-4f15-b6ec-71a25b011cce"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"05b45364-422d-4c70-ada0-e2f7b9f97b90"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(contraposition \[SubsetEqual]D)\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(contraposition \[SubsetEqual]D)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["324850200", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:324850200", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:39, 0.063589 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:324850200-1", "5"},
 CellLabel->"In[46]:=",
 CellID->623913484,ExpressionUUID->"576a4a46-e65f-47df-8867-6961c8be3b2d"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"4f4e21e4-03cd-4156-8197-72b1bf19fc27"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[SubsetEqual] refl)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[SubsetEqual] refl)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["232739861", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:232739861", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:40, 0.020233 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:232739861-1", "5"},
 CellLabel->"In[47]:=",
 CellID->1074729877,ExpressionUUID->"5f2f6951-acfb-40e5-85f6-8302f0285cf1"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"a590f93a-3549-4c5e-8b78-5fd02d101a7c"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[SubsetEqual] trans)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[SubsetEqual] trans)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["373774609", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:373774609", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:40, 0.059067 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:373774609-1", "5"},
 CellLabel->"In[48]:=",
 CellID->1891328660,ExpressionUUID->"d361de80-4f01-4878-9ebd-2f690ccfb113"],

Cell[BoxData[
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", 
  "\"\<\[SubsetEqual] refl\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.76501769634216*^9, 3.76501770835034*^9}},
 CellLabel->"In[49]:=",
 CellID->755107797,ExpressionUUID->"a9b217e6-c68f-4e1b-bb54-a0e9c95b2309"]
}, Closed]],

Cell[CellGroupData[{

Cell["Equality", "Subsection",
 CellChangeTimes->{{3.730289331665319*^9, 3.730289333133805*^9}},
 CellID->445026059,ExpressionUUID->"843f838a-acb8-41d2-a70e-2bf4ffcec857"],

Cell["", "OpenEnvironment",
 CellID->9238772,ExpressionUUID->"d5b300c3-9cb8-4aae-beb0-db6795feeb60"],

Cell[CellGroupData[{

Cell["Lemma (Properties of \[Equal] for Sets)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.7303556518613925`*^9, 3.7303556573879533`*^9}},
 CellID->172029412,ExpressionUUID->"47f6c4c3-2db8-4595-9230-fbe83022565a"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", "\[Equal]", "B"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", "x"], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"x", "\[Element]", "A"}], 
         TagBox[")",
          "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"x", "\[Element]", "B"}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Equal] iff SET)", "FrameLabel"], Evaluator -> Automatic,
          Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303556601669245`*^9, 3.730355697860181*^9}},
 CellTags->"\[Equal] iff SET",
 CellLabel->"In[50]:=",
 CellID->333794707,ExpressionUUID->"cc3b44c7-7696-48b2-9a11-9b0e02cbd0de"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"B", "\[SubsetEqual]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[Equal]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(\[SubsetEqual] antisym)", "FrameLabel"], Evaluator -> 
          Automatic, Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "intro"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730355755711995*^9, 3.730355783383793*^9}},
 CellTags->{"\[SubsetEqual] antisym", "intro"},
 CellLabel->"In[51]:=",
 CellID->833708774,ExpressionUUID->"ef16daca-313c-4d86-a78a-d1cac485fb01"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", "\[Equal]", "B"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"A", "\[SubsetEqual]", "B"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"B", "\[SubsetEqual]", "A"}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Equal] iff \[SubsetEqual])", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7650849914743633`*^9, 3.7650850163780212`*^9}},
 CellTags->"\[Equal] iff \[SubsetEqual]",
 CellLabel->"In[52]:=",
 CellID->1768369467,ExpressionUUID->"734498b0-c5ef-4a95-a0a8-2bf22258d021"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", 
    RowBox[{"thesis", "::", "FORM"}]}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[Equal]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"A", "\[SubsetEqual]", "B"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"B", "\[SubsetEqual]", "A"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", "thesis"}], ")"}], 
   "\[LongRightArrow]", "thesis"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Equal]E SET)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730355836508546*^9, 3.7303558782279243`*^9}},
 CellTags->"\[Equal]E SET",
 CellLabel->"In[53]:=",
 CellID->129158904,ExpressionUUID->"39d27beb-c3fa-4672-8c21-115b260ef59e"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x", ",", "P"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[Equal]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"x", "\[Element]", "A"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"x", "\[Element]", "B"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", "P"}], ")"}], 
   "\[LongRightArrow]", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"x", "\[NotElement]", "A"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"x", "\[NotElement]", "B"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", "P"}], ")"}], 
   "\[LongRightArrow]", "P"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(\[Equal]E classical SET)", "FrameLabel"], Evaluator -> 
          Automatic, Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "elim"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730355915911871*^9, 3.7303559807525*^9}},
 CellTags->{"\[Equal]E classical SET", "elim"},
 CellLabel->"In[54]:=",
 CellID->187730850,ExpressionUUID->"499b7ea4-c1ec-4a3c-93aa-52ab71002cec"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->790548136,ExpressionUUID->"57e6dcfc-053d-422d-a477-1e83166d1b37"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->146581488,ExpressionUUID->"438c5a65-ac66-4fe8-b287-df4ac79616c0"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"2be9a068-24c9-4392-97b8-4ed989f6bea4"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Equal] iff SET)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Equal] iff SET)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["333794707", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:333794707", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:42, 0.305111 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:333794707-1", "3"},
 CellLabel->"In[55]:=",
 CellID->1506794386,ExpressionUUID->"d29d2a48-174c-424d-be8c-88ec8b4d64fe"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"f9ba4602-eff1-413e-80a8-341f1689afa5"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[SubsetEqual] antisym)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[SubsetEqual] antisym)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["833708774", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:833708774", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:42, 0.180891 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:833708774-1", "3"},
 CellLabel->"In[56]:=",
 CellID->74925535,ExpressionUUID->"5027dd32-9d55-40fe-a936-08172c79cad1"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"81d744c6-86ef-412c-a8a4-ef5d38b72bb8"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Equal] iff \[SubsetEqual])\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(\[Equal] iff \[SubsetEqual])\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1768369467", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1768369467", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:44, 0.970624 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1768369467-1", "3"},
 CellLabel->"In[57]:=",
 CellID->1488100858,ExpressionUUID->"5f4442fc-9b36-4593-b740-1806f36d41e7"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"28ad224a-1a5f-45dd-a353-9b871f4fee3c"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Equal]E SET)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[Equal]E SET)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["129158904", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:129158904", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:44, 0.087912 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:129158904-1", "3"},
 CellLabel->"In[58]:=",
 CellID->1301219521,ExpressionUUID->"46efcca5-5269-4858-be47-88d1f3ecc77b"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"b319ecb8-7387-4e2b-a4cb-28777da6d078"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Equal]E classical SET)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Equal]E classical SET)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["187730850", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:187730850", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:45, 0.192009 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:187730850-1", "3"},
 CellLabel->"In[59]:=",
 CellID->815869071,ExpressionUUID->"e3d8ee28-9559-4c6b-b535-c0f73713378a"]
}, Closed]],

Cell[CellGroupData[{

Cell["Proper Subsets", "Subsection",
 CellChangeTimes->{{3.765009454533329*^9, 3.76500946341322*^9}},
 CellID->65075678,ExpressionUUID->"5c871da2-25b4-4e48-ac4a-37581c088cfe"],

Cell["", "OpenEnvironment",
 CellID->171209090,ExpressionUUID->"4188a6f5-1e35-4b9c-b1fe-18e85a2cf358"],

Cell[CellGroupData[{

Cell["Lemma (Properties of \[Subset])", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.730353565394315*^9, 3.7303535689581966`*^9}},
 CellID->513227992,ExpressionUUID->"75bb5aed-dd93-47b0-ae4f-be3c6a0b5f1a"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[NotEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[Subset]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(\[Subset]I)", "FrameLabel"], Evaluator -> Automatic, 
          Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "intro"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303535749873147`*^9, 3.7303536014909005`*^9}},
 CellTags->{"\[Subset]I", "intro"},
 CellLabel->"In[60]:=",
 CellID->32648321,ExpressionUUID->"31eeffd9-8416-4535-8aef-65a137cf0890"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "a"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[Subset]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"a", "\[Element]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"a", "\[Element]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(\[Subset]D)", "FrameLabel"], Evaluator -> Automatic, 
          Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "trans"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730354249632192*^9, 3.7303542755060167`*^9}},
 CellTags->{"\[Subset]D", "trans"},
 CellLabel->"In[61]:=",
 CellID->457314425,ExpressionUUID->"b0ed0cc7-bc57-483e-a835-36a12066a7d1"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", 
    RowBox[{"thesis", "::", "FORM"}]}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[Subset]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"A", "\[SubsetEqual]", "B"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"B", "\[NotSubsetEqual]", "A"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", "thesis"}], ")"}], 
   "\[LongRightArrow]", "thesis"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(\[Subset]E)", "FrameLabel"], Evaluator -> Automatic, 
          Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "elim"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730353640183064*^9, 3.7303536993137875`*^9}},
 CellTags->{"\[Subset]E", "elim"},
 CellLabel->"In[62]:=",
 CellID->9845002,ExpressionUUID->"c1ed4c0b-f24a-4d7a-9cb7-e956d9316e61"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", "\[Subset]", "B"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"A", "\[SubsetEqual]", "B"}], "\[And]", 
      RowBox[{"A", "\[NotEqual]", "B"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Subset] alt)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730354050807024*^9, 3.7303540778980293`*^9}},
 CellTags->"\[Subset] alt",
 CellLabel->"In[63]:=",
 CellID->688844702,ExpressionUUID->"dbd4d924-a7ac-4077-824e-c31ad364a13b"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[Subset]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Subset] implies \[SubsetEqual])", "FrameLabel"], 
         Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303541115737696`*^9, 3.7303541325044236`*^9}},
 CellTags->"\[Subset] implies \[SubsetEqual]",
 CellLabel->"In[64]:=",
 CellID->53899447,ExpressionUUID->"80efbdd3-5c09-4513-8814-fe0da4d53969"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[Subset]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"B", "\[SubsetEqual]", "C"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[Subset]", "C"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(\[Subset]\[SubsetEqual] trans)", "FrameLabel"], 
          Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "trans"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303543921550026`*^9, 3.7303544056249104`*^9}},
 CellTags->{"\[Subset]\[SubsetEqual] trans", "trans"},
 CellLabel->"In[65]:=",
 CellID->214375017,ExpressionUUID->"9525b07f-5513-4080-98ed-049a5ac6c399"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"B", "\[Subset]", "C"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[Subset]", "C"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(\[SubsetEqual]\[Subset] trans)", "FrameLabel"], 
          Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "trans"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730354427370691*^9, 3.7303544311932435`*^9}},
 CellTags->{"\[SubsetEqual]\[Subset] trans", "trans"},
 CellLabel->"In[66]:=",
 CellID->705380086,ExpressionUUID->"c48fc25a-8ea6-4254-bb22-ab5d44b79f90"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[Subset]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"B", "\[Subset]", "C"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[Subset]", "C"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(\[Subset] trans)", "FrameLabel"], Evaluator -> Automatic,
           Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "trans"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303543396353703`*^9, 3.7303543661322255`*^9}},
 CellTags->{"\[Subset] trans", "trans"},
 CellLabel->"In[67]:=",
 CellID->122404649,ExpressionUUID->"ab797c4b-7e7e-4267-afe1-61013ada1bc0"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Not]", 
    RowBox[{"A", "\[Subset]", "A"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Subset] irrefl)", "FrameLabel"], Evaluator -> Automatic,
          Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765011343501699*^9, 3.7650113575259457`*^9}},
 CellTags->"\[Subset] irrefl",
 CellLabel->"In[68]:=",
 CellID->490505039,ExpressionUUID->"8eacd303-9562-4ec0-b90b-56283033ded7"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->27085985,ExpressionUUID->"67c188f5-2a05-4fc3-9299-224de209fdf7"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->716136528,ExpressionUUID->"36f8c886-d198-48a0-99e7-3309c33bff47"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"8cf5356c-fb92-4a74-805d-8b9b8f181a04"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Subset]I)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[Subset]I)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["32648321", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:32648321", CellTags]], Appearance -> 
          None, Evaluator -> Automatic, Method -> "Preemptive"]]], 
       " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:46, 0.171530 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:32648321-1", "3"},
 CellLabel->"In[69]:=",
 CellID->925901741,ExpressionUUID->"6e659942-0c6d-405c-9b68-372d2c14a0ed"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"213d6b34-a4f7-47c3-a5bc-9a2b37df34e0"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Subset]D)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[Subset]D)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["457314425", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:457314425", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:46, 0.115736 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:457314425-1", "3"},
 CellLabel->"In[70]:=",
 CellID->1854094453,ExpressionUUID->"c887437d-b8f9-4f4a-abc2-4c844d627714"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"153a94d7-ce3b-41dc-8c38-24b26702408e"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Subset]E)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[Subset]E)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["9845002", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:9845002", CellTags]], Appearance -> 
          None, Evaluator -> Automatic, Method -> "Preemptive"]]], 
       " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:46, 0.162016 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:9845002-1", "3"},
 CellLabel->"In[71]:=",
 CellID->666993796,ExpressionUUID->"9d0f6fb7-026a-419b-8d81-21560f62f383"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"d00af29a-5293-4848-bd89-c3b08cc4ef0e"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Subset] alt)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[Subset] alt)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["688844702", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:688844702", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:47, 0.330516 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:688844702-1", "3"},
 CellLabel->"In[72]:=",
 CellID->894137823,ExpressionUUID->"5f1c3502-0db1-4202-8b55-474b927d5261"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"31020c36-dca2-48e0-a834-2dac3f95353b"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Subset] implies \[SubsetEqual])\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(\[Subset] implies \[SubsetEqual])\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["53899447", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:53899447", CellTags]], Appearance -> 
          None, Evaluator -> Automatic, Method -> "Preemptive"]]], 
       " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:47, 0.041804 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:53899447-1", "3"},
 CellLabel->"In[73]:=",
 CellID->1854942837,ExpressionUUID->"c6989024-1dd7-4cbd-bf7d-4edaa6f98d71"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"f3eba502-b211-4304-88e5-f68e94fa1388"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Subset]\[SubsetEqual] trans)\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(\[Subset]\[SubsetEqual] trans)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["214375017", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:214375017", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:48, 0.236895 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:214375017-1", "3"},
 CellLabel->"In[74]:=",
 CellID->187880605,ExpressionUUID->"441a9d35-8e30-4162-9e21-8cf3471755aa"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"0e789166-89a2-45b5-8396-35679128c56d"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[SubsetEqual]\[Subset] trans)\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(\[SubsetEqual]\[Subset] trans)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["705380086", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:705380086", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:48, 0.251268 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:705380086-1", "3"},
 CellLabel->"In[75]:=",
 CellID->719967367,ExpressionUUID->"5ba5cdc6-46d4-4ad2-94e9-7d5f31bc3fe9"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"42b3252c-5cd7-4039-be77-d25a81f1f212"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Subset] trans)\"", "FormReference", StripOnInput -> False],
             True -> StyleBox[
             "\"(\[Subset] trans)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["122404649", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:122404649", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:49, 0.030652 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:122404649-1", "3"},
 CellLabel->"In[76]:=",
 CellID->28123779,ExpressionUUID->"76b67bb5-6cbd-4d7e-8eed-928b19df89d2"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"3ea324ce-c799-4f7d-99c0-672aeee83d35"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Subset] irrefl)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Subset] irrefl)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["490505039", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:490505039", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:49, 0.092175 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:490505039-1", "3"},
 CellLabel->"In[77]:=",
 CellID->1482758530,ExpressionUUID->"2887fda8-1b65-4980-bb82-685424955125"]
}, Closed]],

Cell[CellGroupData[{

Cell["Empty Set", "Subsection",
 CellChangeTimes->{{3.730289344796713*^9, 3.7302893467507815`*^9}},
 CellID->398691178,ExpressionUUID->"319ae35b-f354-4f8d-894c-8478b4686747"],

Cell["", "OpenEnvironment",
 CellID->22140062,ExpressionUUID->"6e852d7e-d449-4a55-9c75-03282f176ba7"],

Cell[CellGroupData[{

Cell["Lemma (Properties of the Empty Set)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.73035605587866*^9, 3.730356062729162*^9}},
 CellID->225411204,ExpressionUUID->"0f914343-0bd7-494c-be74-1d7419de28d4"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "x"], 
  RowBox[{"x", "\[NotElement]", 
   RowBox[{"{", "}"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[NotElement]{})", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.73035606582572*^9, 3.7303560748100014`*^9}},
 CellTags->"\[NotElement]{}",
 CellLabel->"In[78]:=",
 CellID->109474291,ExpressionUUID->"f40e860d-7927-4260-b558-399c721a0a03"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"x", ",", "P"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", 
     RowBox[{"{", "}"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   "P"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["({}E)", "FrameLabel"], Evaluator -> Automatic, Appearance -> 
          None, ButtonFunction :> Theorema`Common`editCellTags[]], "elim"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303561491283245`*^9, 3.7303561678499956`*^9}, {
  3.765017341180232*^9, 3.7650173448667793`*^9}},
 CellTags->{"{}E", "elim"},
 CellLabel->"In[79]:=",
 CellID->54906548,ExpressionUUID->"5e2a16f1-59da-4a82-8948-50266063e236"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   RowBox[{"{", "}"}], "\[SubsetEqual]", "A"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["({}\[SubsetEqual])", "FrameLabel"], Evaluator -> Automatic,
          Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303562617124276`*^9, 3.730356269378311*^9}},
 CellTags->"{}\[SubsetEqual]",
 CellLabel->"In[80]:=",
 CellID->289979004,ExpressionUUID->"c1739f2e-64f5-45b3-bfc9-a17967b39339"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     UnderscriptBox["\[And]", "x"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"x", "\[Element]", "A"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[LongRightArrow]", "False"}]}], ")"}], 
   "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[Equal]", 
     RowBox[{"{", "}"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Equal]{} I)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730356322533823*^9, 3.7303563607458706`*^9}},
 CellTags->"\[Equal]{} I",
 CellLabel->"In[81]:=",
 CellID->828225892,ExpressionUUID->"26118ba7-c5c4-4498-bf72-00252a9af324"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Not]", 
    RowBox[{"A", "\[Subset]", 
     RowBox[{"{", "}"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Not]\[Subset]{})", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765179787233038*^9, 3.765179795144704*^9}},
 CellTags->"\[Not]\[Subset]{}",
 CellLabel->"In[82]:=",
 CellID->1987360397,ExpressionUUID->"9cd78d1f-85a0-4723-897c-277cb4c9e703"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->40835667,ExpressionUUID->"843f2252-85c3-42c4-9fdb-f3591cb11305"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->27978791,ExpressionUUID->"e50fc8f6-c3fe-4b13-a863-8b649fb65e90"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"8e501afa-7659-40c4-a720-9efa0a262bae"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[NotElement]{})\"", "FormReference", StripOnInput -> False],
             True -> StyleBox[
             "\"(\[NotElement]{})\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["109474291", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:109474291", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:50, 0.069420 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:109474291-1", "3"},
 CellLabel->"In[83]:=",
 CellID->1750098064,ExpressionUUID->"ae3cd3a9-b24b-48b2-8218-78763782452d"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"3c72c428-2609-40fe-ac68-85f5e02ad6aa"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"({}E)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"({}E)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["54906548", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:54906548", CellTags]], Appearance -> 
          None, Evaluator -> Automatic, Method -> "Preemptive"]]], 
       " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:50, 0.046693 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:54906548-1", "3"},
 CellLabel->"In[84]:=",
 CellID->1933668764,ExpressionUUID->"7c42a337-6cab-46e3-bd77-c0a11197a7dc"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"bf08d3ee-c3ff-41ae-b5ab-51831c70bcb5"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"({}\[SubsetEqual])\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"({}\[SubsetEqual])\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["289979004", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:289979004", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:51, 0.027990 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:289979004-1", "3"},
 CellLabel->"In[85]:=",
 CellID->2138873264,ExpressionUUID->"6f47db7d-2333-492d-87ac-73a3393a968e"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"76be9390-d383-40b3-aa3c-7fa00a7fbc44"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Equal]{} I)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[Equal]{} I)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["828225892", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:828225892", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:51, 0.136354 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:828225892-1", "3"},
 CellLabel->"In[86]:=",
 CellID->157843443,ExpressionUUID->"f789d536-f6cf-4bca-84d6-4532e0bed010"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"dfe546f0-950a-494c-a17f-448f7c7d7740"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Not]\[Subset]{})\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Not]\[Subset]{})\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1987360397", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1987360397", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:51, 0.083215 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1987360397-1", "3"},
 CellLabel->"In[87]:=",
 CellID->1029679455,ExpressionUUID->"3cb753de-8224-4302-aee9-4b257cf1136a"],

Cell[BoxData[
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", "\"\<\[NotElement]{}\>\"", " ", 
  "\"\<{}\[SubsetEqual]\>\"", " ", 
  "\"\<\[Not]\[Subset]{}\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.7303561165186214`*^9, 3.730356128810933*^9}, {
  3.7303562816703367`*^9, 3.730356308844104*^9}, {3.765179810225616*^9, 
  3.765179814056596*^9}},
 CellLabel->"In[88]:=",
 CellID->435688659,ExpressionUUID->"ea74a46f-344b-4e6a-946f-391f2dba6215"]
}, Closed]],

Cell[CellGroupData[{

Cell["Universal Set", "Subsection",
 CellChangeTimes->{{3.7302893515653043`*^9, 3.730289353628006*^9}},
 CellID->243229164,ExpressionUUID->"c4569365-eae7-43ae-9d6b-96c63302ed0c"],

Cell["", "OpenEnvironment",
 CellID->418476274,ExpressionUUID->"676f0161-cbcd-44b4-ac0d-3474620a07f7"],

Cell[CellGroupData[{

Cell["Lemma (Properties of the Universal Set)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.7303564935962696`*^9, 3.730356499412635*^9}},
 CellID->102519102,ExpressionUUID->"4b3b91bf-0c36-4782-bf4b-5136344ebe69"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "x"], 
  RowBox[{"x", "\[Element]", "UNIV"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(UNIV I)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303565021467314`*^9, 3.730356508219396*^9}},
 CellTags->"UNIV I",
 CellLabel->"In[89]:=",
 CellID->313591036,ExpressionUUID->"84774292-9b5d-4170-8a68-008bf0e627b8"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{"A", "\[SubsetEqual]", "UNIV"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[SubsetEqual]UNIV)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730356585618762*^9, 3.7303565934973955`*^9}},
 CellTags->"\[SubsetEqual]UNIV",
 CellLabel->"In[90]:=",
 CellID->223522958,ExpressionUUID->"5a1c8aed-f913-4370-96ff-617b5a19afed"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     UnderscriptBox["\[And]", "x"], 
     RowBox[{"x", "\[Element]", "A"}]}], ")"}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[Equal]", "UNIV"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Equal]UNIV I)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303566516200066`*^9, 3.730356670013482*^9}},
 CellTags->"\[Equal]UNIV I",
 CellLabel->"In[91]:=",
 CellID->256313502,ExpressionUUID->"5edf972c-0fd1-45e3-886f-b3b8b285efab"],

Cell[BoxData[
 RowBox[{"UNIV", "\[NotEqual]", 
  RowBox[{"{", "}"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(UNIV\[NotEqual]{})", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303566971052666`*^9, 3.7303567020005045`*^9}},
 CellTags->"UNIV\[NotEqual]{}",
 CellLabel->"In[92]:=",
 CellID->358676608,ExpressionUUID->"eb06d768-e9ca-4788-ad46-4aa781f6599b"],

Cell[BoxData[
 RowBox[{
  RowBox[{"{", "}"}], "\[NotEqual]", "UNIV"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["({}\[NotEqual]UNIV)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.73035671645079*^9, 3.7303567194063654`*^9}},
 CellTags->"{}\[NotEqual]UNIV",
 CellLabel->"In[93]:=",
 CellID->13969540,ExpressionUUID->"f150aa79-de04-4fd6-8a60-7995eeec44d4"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->62064389,ExpressionUUID->"3b694a22-d12f-435d-ac03-bbcd673b2ff2"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->208900752,ExpressionUUID->"a462eee5-f1a6-4424-9fc6-527262d5e68c"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"a9b40c7f-d78d-4cc4-8ed0-82fb7d7cc154"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(UNIV I)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(UNIV I)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["313591036", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:313591036", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:52, 0.063759 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:313591036-1", "3"},
 CellLabel->"In[94]:=",
 CellID->1173747547,ExpressionUUID->"f472fc82-802b-410b-868c-d07e017b0268"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"77dc8247-0ae6-41b7-b66f-83f01aca1558"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[SubsetEqual]UNIV)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[SubsetEqual]UNIV)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["223522958", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:223522958", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:53, 0.079804 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:223522958-1", "3"},
 CellLabel->"In[95]:=",
 CellID->2010038796,ExpressionUUID->"184057ce-28ca-4f86-bd4e-79e86a17715d"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"357effc5-98e0-460f-ba0d-0c374659c143"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Equal]UNIV I)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[Equal]UNIV I)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["256313502", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:256313502", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:53, 0.147719 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:256313502-1", "3"},
 CellLabel->"In[96]:=",
 CellID->741200714,ExpressionUUID->"de3eb27b-5b7d-4ae2-988d-b2653d4d8d31"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"0de550ca-21a7-4bec-a8e4-2de545c61393"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(UNIV\[NotEqual]{})\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(UNIV\[NotEqual]{})\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["358676608", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:358676608", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:54, 0.069793 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:358676608-1", "3"},
 CellLabel->"In[97]:=",
 CellID->1104785395,ExpressionUUID->"824b928c-cb7f-4776-8b3e-3335d33de91a"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"6f615a28-0f93-4369-9c01-fafee8ed7a1c"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"({}\[NotEqual]UNIV)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"({}\[NotEqual]UNIV)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["13969540", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:13969540", CellTags]], Appearance -> 
          None, Evaluator -> Automatic, Method -> "Preemptive"]]], 
       " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:54, 0.016550 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:13969540-1", "3"},
 CellLabel->"In[98]:=",
 CellID->400553415,ExpressionUUID->"49775606-3bad-4639-ad88-ce4f5704386f"],

Cell[BoxData[
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", "\"\<UNIV I\>\"", " ", 
  "\"\<\[SubsetEqual]UNIV\>\"", " ", "\"\<UNIV\[NotEqual]{}\>\"", " ", 
  "\"\<{}\[NotEqual]UNIV\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.7303561165186214`*^9, 3.730356128810933*^9}, {
  3.7303562816703367`*^9, 3.730356308844104*^9}, {3.7303565362228093`*^9, 
  3.7303565400519514`*^9}, {3.7303566107425056`*^9, 3.7303566150434475`*^9}, {
  3.7650177578779182`*^9, 3.7650177715018597`*^9}},
 CellLabel->"In[99]:=",
 CellID->904482657,ExpressionUUID->"e1462291-f7cd-49ae-ac8d-6c511fa5abc1"]
}, Closed]],

Cell[CellGroupData[{

Cell["Powerset", "Subsection",
 CellChangeTimes->{{3.7302893602402*^9, 3.730289372386496*^9}},
 CellID->683848815,ExpressionUUID->"0f6ebc42-5c52-420e-a539-869385485218"],

Cell["", "OpenEnvironment",
 CellID->578346443,ExpressionUUID->"14fa329f-4d46-4367-a4e8-abe24002157e"],

Cell[CellGroupData[{

Cell["Lemma (Properties of Pow)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.7303569198146944`*^9, 3.730356923581992*^9}},
 CellID->167038458,ExpressionUUID->"3e1294c2-57fa-432b-8aca-09f3e21dba79"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", "\[Element]", 
      RowBox[{"Pow", "[", "B", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", "\[SubsetEqual]", "B"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Element]Pow)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303569410522623`*^9, 3.730356966310022*^9}},
 CellTags->"\[Element]Pow",
 CellLabel->"In[100]:=",
 CellID->91383453,ExpressionUUID->"2d5efc8e-b5dc-4fd3-9859-b98a6c77774b"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[Element]", 
     RowBox[{"Pow", "[", "B", "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(PowI)", "FrameLabel"], Evaluator -> Automatic, 
          Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "intro"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730357023815112*^9, 3.7303570451655054`*^9}},
 CellTags->{"PowI", "intro"},
 CellLabel->"In[101]:=",
 CellID->1005450222,ExpressionUUID->"d3c585f0-0af8-4b77-b8f2-4a2e80d5ec5a"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[Element]", 
     RowBox[{"Pow", "[", "B", "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(PowD)", "FrameLabel"], Evaluator -> Automatic, Appearance -> 
         None, ButtonFunction :> Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.73035705514585*^9, 3.730357075884611*^9}},
 CellTags->"PowD",
 CellLabel->"In[102]:=",
 CellID->71998340,ExpressionUUID->"134a7df0-6a03-483f-8b3c-9ed6ba820c3f"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   RowBox[{"{", "}"}], "\[Element]", 
   RowBox[{"Pow", "[", "A", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["({}\[Element]Pow)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730357093769213*^9, 3.7303571013591146`*^9}},
 CellTags->"{}\[Element]Pow",
 CellLabel->"In[103]:=",
 CellID->135248918,ExpressionUUID->"0039201d-a8f0-4273-ac3a-5b47720c154c"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{"A", "\[Element]", 
   RowBox[{"Pow", "[", "A", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(self\[Element]Pow)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730357121148073*^9, 3.7303571234214525`*^9}},
 CellTags->"self\[Element]Pow",
 CellLabel->"In[104]:=",
 CellID->365430671,ExpressionUUID->"3975308c-87da-4dbf-bd5e-7f174b4273f7"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   RowBox[{"Pow", "[", "A", "]"}], "\[NotEqual]", 
   RowBox[{"{", "}"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Pow\[NotEqual]{})", "FrameLabel"], Evaluator -> Automatic,
          Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303571763352337`*^9, 3.7303571848134637`*^9}},
 CellTags->"Pow\[NotEqual]{}",
 CellLabel->"In[105]:=",
 CellID->761522039,ExpressionUUID->"725fd266-ab88-47ba-adf8-00acdeb98c1b"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->175636950,ExpressionUUID->"8066a26b-19d3-45ab-959d-de9d5da23820"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->316659783,ExpressionUUID->"8946ff36-3df7-4ac4-ad26-ce1ef17ad359"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"f321aa08-774b-42eb-895d-cb49bddbd4c0"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Element]Pow)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[Element]Pow)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["91383453", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:91383453", CellTags]], Appearance -> 
          None, Evaluator -> Automatic, Method -> "Preemptive"]]], 
       " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:55, 0.057330 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:91383453-1", "3"},
 CellLabel->"In[106]:=",
 CellID->1669010862,ExpressionUUID->"135e6c9f-6b5f-45fe-bca0-d1d8524dfc01"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"b0fde61a-02ae-4182-a0bc-d5aa64cfa0ad"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(PowI)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(PowI)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1005450222", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1005450222", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:55, 0.078594 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1005450222-1", "3"},
 CellLabel->"In[107]:=",
 CellID->969791405,ExpressionUUID->"2eff29d4-750b-432e-b353-66e38bf0fded"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"61478380-eb67-4d72-aada-6010360cdefd"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(PowD)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(PowD)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["71998340", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:71998340", CellTags]], Appearance -> 
          None, Evaluator -> Automatic, Method -> "Preemptive"]]], 
       " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:56, 0.079963 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:71998340-1", "3"},
 CellLabel->"In[108]:=",
 CellID->1724899203,ExpressionUUID->"e66761b0-9d29-4441-b48e-3fce10402d48"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"86b8a276-82db-46f1-9982-d286439088f3"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"({}\[Element]Pow)\"", "FormReference", StripOnInput -> False],
             True -> StyleBox[
             "\"({}\[Element]Pow)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["135248918", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:135248918", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:56, 0.045896 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:135248918-1", "3"},
 CellLabel->"In[109]:=",
 CellID->417457554,ExpressionUUID->"e9f8d063-f619-4ed9-811c-81d57ff90446"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"7fa1e3bc-9fcb-402b-94d4-5e58008b2604"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(self\[Element]Pow)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(self\[Element]Pow)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["365430671", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:365430671", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:57, 0.047720 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:365430671-1", "3"},
 CellLabel->"In[110]:=",
 CellID->1046907733,ExpressionUUID->"88881ba8-d043-4491-afa7-e2b45acd577c"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"206fd92b-da02-4132-a54d-d5c7d2ffb22c"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Pow\[NotEqual]{})\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(Pow\[NotEqual]{})\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["761522039", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:761522039", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:57, 0.097895 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:761522039-1", "3"},
 CellLabel->"In[111]:=",
 CellID->1384782616,ExpressionUUID->"7af83b7e-0a64-42d4-9cfa-8fb1f1515d6b"],

Cell[BoxData[
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", "\"\<\[Element]Pow\>\"", " ", 
  "\"\<{}\[Element]Pow\>\"", " ", "\"\<self\[Element]Pow\>\"", " ", 
  "\"\<Pow\[NotEqual]{}\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.7303561165186214`*^9, 3.730356128810933*^9}, {
  3.7303562816703367`*^9, 3.730356308844104*^9}, {3.7303565362228093`*^9, 
  3.7303565400519514`*^9}, {3.7303566107425056`*^9, 3.7303566150434475`*^9}, {
  3.7303569752673616`*^9, 3.7303569764848404`*^9}, {3.765017813108994*^9, 
  3.7650178396450243`*^9}, {3.765018991886299*^9, 3.7650189945505123`*^9}},
 CellLabel->"In[112]:=",
 CellID->404830536,ExpressionUUID->"1859160d-8dbb-44dd-b244-cb10bb4d705c"]
}, Closed]],

Cell[CellGroupData[{

Cell["Complement", "Subsection",
 CellChangeTimes->{{3.7302893694000974`*^9, 3.730289371057134*^9}},
 CellID->84239928,ExpressionUUID->"4adeb3cb-5bc8-4193-8ec2-a95f87a70931"],

Cell["", "OpenEnvironment",
 CellID->564974621,ExpressionUUID->"9364deea-7691-419d-b60e-955bdaa9477c"],

Cell[CellGroupData[{

Cell["Lemma (Properties of the Complement)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.73035726495271*^9, 3.73035727214167*^9}},
 CellID->203867575,ExpressionUUID->"564c50c6-2fb4-4244-8eb4-ce407adb7a92"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "x"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"x", "\[Element]", 
      RowBox[{"-", "A"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"x", "\[NotElement]", "A"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Element]complement)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303572761364393`*^9, 3.7303573112375274`*^9}},
 CellTags->"\[Element]complement",
 CellLabel->"In[113]:=",
 CellID->499422188,ExpressionUUID->"7d81673e-75fc-41aa-b4b7-569230380565"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "x"}]], 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"x", "\[Element]", "A"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", "False"}], ")"}], 
   "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", 
     RowBox[{"-", "A"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(complement I)", "FrameLabel"], Evaluator -> Automatic, 
          Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "intro"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303573470216007`*^9, 3.7303573783274546`*^9}},
 CellTags->{"complement I", "intro"},
 CellLabel->"In[114]:=",
 CellID->319529804,ExpressionUUID->"d27356cd-910b-4e27-a702-8f3e1f38e9d3"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "x"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", 
     RowBox[{"-", "A"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[NotElement]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(complement D)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730357440878751*^9, 3.730357466048712*^9}},
 CellTags->"complement D",
 CellLabel->"In[115]:=",
 CellID->7165872,ExpressionUUID->"0bda5580-c651-4792-9b44-3829608ed193"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->18980456,ExpressionUUID->"9d9b35e8-0ba4-4422-975e-8ed6dca8e0af"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->203717016,ExpressionUUID->"b3edc74f-edee-44e9-9fde-57bec6435130"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"e403e993-e5e9-4b95-bc15-d959b6e67bbe"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Element]complement)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Element]complement)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["499422188", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:499422188", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:58, 0.051748 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:499422188-1", "3"},
 CellLabel->"In[116]:=",
 CellID->742874104,ExpressionUUID->"4bb986a8-28da-4979-8914-a7bf05134176"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"01d94537-d1ea-49af-96b4-29b3555c2d98"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(complement I)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(complement I)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["319529804", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:319529804", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:59, 0.112078 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:319529804-1", "3"},
 CellLabel->"In[117]:=",
 CellID->692678955,ExpressionUUID->"6b7cdde2-f7e1-4895-96bd-f5dba2fa343f"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"0f580cfc-91b1-40ce-b3e3-41b79022d7c8"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(complement D)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(complement D)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["7165872", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:7165872", CellTags]], Appearance -> 
          None, Evaluator -> Automatic, Method -> "Preemptive"]]], 
       " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 08:59:59, 0.079293 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:7165872-1", "3"},
 CellLabel->"In[118]:=",
 CellID->214206169,ExpressionUUID->"1b5041f8-10cf-4c78-8caf-cee02595fafa"],

Cell[BoxData[
 RowBox[{
  RowBox[{"facts", " ", "\"\<complement E\>\""}], ":", " ", 
  RowBox[{"\"\<complement D\>\"", "[", "elimination", "]"}]}]], "TheoremaCode",
 CellChangeTimes->{{3.730357497226617*^9, 3.730357538442931*^9}},
 CellLabel->"In[119]:=",
 CellID->19405246,ExpressionUUID->"30ba7cb1-88f9-4fd1-afb5-8439bab1d1f0"],

Cell[BoxData[
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", 
  "\"\<\[Element]complement\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.7303561165186214`*^9, 3.730356128810933*^9}, {
  3.7303562816703367`*^9, 3.730356308844104*^9}, {3.7303565362228093`*^9, 
  3.7303565400519514`*^9}, {3.7303566107425056`*^9, 3.7303566150434475`*^9}, {
  3.7303569752673616`*^9, 3.7303569764848404`*^9}, {3.7303573339410152`*^9, 
  3.7303573351247845`*^9}, {3.765019192459543*^9, 3.765019194996132*^9}},
 CellLabel->"In[120]:=",
 CellID->131828019,ExpressionUUID->"22060e06-cda4-491b-b33a-1575f31a4956"]
}, Closed]],

Cell[CellGroupData[{

Cell["Intersection", "Subsection",
 CellChangeTimes->{{3.730289379498002*^9, 3.7302893813274345`*^9}},
 CellID->466527033,ExpressionUUID->"d344c0ef-7e76-40e8-b6fb-0ed849f63b7a"],

Cell["", "OpenEnvironment",
 CellID->465207003,ExpressionUUID->"76ce294e-31e6-41fa-94dc-d1b5f8ac0836"],

Cell[CellGroupData[{

Cell["Lemma (Properties of \[Intersection])", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.730357632836278*^9, 3.730357636833165*^9}},
 CellID->373067187,ExpressionUUID->"87893071-e583-408a-8705-aec67b67f9af"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"x", "\[Element]", 
      RowBox[{"A", "\[Intersection]", "B"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"x", "\[Element]", "A"}], "\[And]", 
      RowBox[{"x", "\[Element]", "B"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Element]\[Intersection])", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303576396179795`*^9, 3.7303576722513847`*^9}},
 CellTags->"\[Element]\[Intersection]",
 CellLabel->"In[121]:=",
 CellID->299832540,ExpressionUUID->"faf9caf8-e1c5-4ad6-a348-5e25e999ed9e"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", 
     RowBox[{"A", "\[Intersection]", "B"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(\[Intersection]I)", "FrameLabel"], Evaluator -> 
          Automatic, Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "intro"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303577117159915`*^9, 3.7303577386049376`*^9}},
 CellTags->{"\[Intersection]I", "intro"},
 CellLabel->"In[122]:=",
 CellID->742723078,ExpressionUUID->"f282f462-c6f2-4c42-93f9-8add211dc534"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", 
     RowBox[{"A", "\[Intersection]", "B"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection]D1)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730357753413187*^9, 3.7303577661134195`*^9}},
 CellTags->"\[Intersection]D1",
 CellLabel->"In[123]:=",
 CellID->196903271,ExpressionUUID->"728e5702-a968-4311-b864-1dd1f476124d"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", 
     RowBox[{"A", "\[Intersection]", "B"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection]D2)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730357784414705*^9, 3.7303577867671013`*^9}},
 CellTags->"\[Intersection]D2",
 CellLabel->"In[124]:=",
 CellID->214165647,ExpressionUUID->"2c742309-e67a-46f8-923c-ec686d1342d2"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x", ",", 
    RowBox[{"thesis", "::", "FORM"}]}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", 
     RowBox[{"A", "\[Intersection]", "B"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"x", "\[Element]", "A"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"x", "\[Element]", "B"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", "thesis"}], ")"}], 
   "\[LongRightArrow]", "thesis"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(\[Intersection]E)", "FrameLabel"], Evaluator -> 
          Automatic, Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "elim"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730357811582844*^9, 3.7303578512574606`*^9}},
 CellTags->{"\[Intersection]E", "elim"},
 CellLabel->"In[125]:=",
 CellID->31355863,ExpressionUUID->"d24a4225-19d0-415c-b284-7fe9384810ea"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->940647718,ExpressionUUID->"52002887-2767-48d7-8efc-8b61dc177c28"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->323591903,ExpressionUUID->"692f6ffe-50f7-4dac-8332-1ecce010a4bc"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"f9de8bb1-1d8b-47b6-96a4-6a7eb2b5cf86"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Element]\[Intersection])\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Element]\[Intersection])\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["299832540", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:299832540", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:00, 0.063302 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:299832540-1", "3"},
 CellLabel->"In[126]:=",
 CellID->520490995,ExpressionUUID->"06ccfced-9f01-4ea1-b866-1dc0ce6111d4"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"7a64d813-487d-4274-bf90-ccb0f7ed048e"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection]I)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Intersection]I)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["742723078", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:742723078", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:01, 0.138541 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:742723078-1", "3"},
 CellLabel->"In[127]:=",
 CellID->1206322670,ExpressionUUID->"b655e7ec-ba70-4141-b8ce-f4c7c9313eb5"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"ae65ee7c-3edb-422f-b6fc-3dadbcec8741"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection]D1)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Intersection]D1)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["196903271", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:196903271", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:01, 0.116653 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:196903271-1", "3"},
 CellLabel->"In[128]:=",
 CellID->1182674228,ExpressionUUID->"999e3753-e878-46ae-816d-c1c14791130d"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"696ad2cc-ebfd-4b51-ba50-48937244733d"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection]D2)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Intersection]D2)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["214165647", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:214165647", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:02, 0.133158 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:214165647-1", "3"},
 CellLabel->"In[129]:=",
 CellID->1685360901,ExpressionUUID->"263b3d20-5c8b-4d42-8cd7-bb5ecf6bba6c"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"c28ef374-de39-4214-93b6-24a5eafd669e"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection]E)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Intersection]E)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["31355863", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:31355863", CellTags]], Appearance -> 
          None, Evaluator -> Automatic, Method -> "Preemptive"]]], 
       " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:02, 0.079372 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:31355863-1", "3"},
 CellLabel->"In[130]:=",
 CellID->634954845,ExpressionUUID->"e8ca3e41-bda4-4b23-8e4b-61f126bed335"],

Cell[BoxData[
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", 
  "\"\<\[Element]\[Intersection]\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.7303561165186214`*^9, 3.730356128810933*^9}, {
  3.7303562816703367`*^9, 3.730356308844104*^9}, {3.7303565362228093`*^9, 
  3.7303565400519514`*^9}, {3.7303566107425056`*^9, 3.7303566150434475`*^9}, {
  3.7303569752673616`*^9, 3.7303569764848404`*^9}, {3.7303573339410152`*^9, 
  3.7303573351247845`*^9}, {3.730357693835204*^9, 3.73035769520304*^9}, {
  3.765019400617015*^9, 3.7650194025053062`*^9}},
 CellLabel->"In[131]:=",
 CellID->558707431,ExpressionUUID->"5a8634c2-f223-4970-b9d3-0023b971effd"],

Cell["", "OpenEnvironment",
 CellID->154478456,ExpressionUUID->"b8b32025-d4d0-4408-9a9b-adf9aee41347"],

Cell[CellGroupData[{

Cell["Lemma (Lattice Properties of \[Intersection])", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.73035540831593*^9, 3.7303554141677246`*^9}},
 CellID->121705945,ExpressionUUID->"5994a4e0-47c8-4b41-8340-de158fdf537a"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   RowBox[{"A", "\[Intersection]", "B"}], "\[SubsetEqual]", 
   "A"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection] lower1)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303551567910347`*^9, 3.730355171980382*^9}},
 CellTags->"\[Intersection] lower1",
 CellLabel->"In[132]:=",
 CellID->330976770,ExpressionUUID->"5a07b132-a168-47df-8a0a-d4cbc2bdf2be"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   RowBox[{"A", "\[Intersection]", "B"}], "\[SubsetEqual]", 
   "B"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection] lower2)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303552091176047`*^9, 3.7303552174177895`*^9}},
 CellTags->"\[Intersection] lower2",
 CellLabel->"In[133]:=",
 CellID->567877347,ExpressionUUID->"69552c22-c7e7-4e84-ae6d-0c027f7b7ca1"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"C", "\[SubsetEqual]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"C", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"C", "\[SubsetEqual]", 
     RowBox[{"A", "\[Intersection]", "B"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection] greatest)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730355236262454*^9, 3.7303552734114885`*^9}},
 CellTags->"\[Intersection] greatest",
 CellLabel->"In[134]:=",
 CellID->370774399,ExpressionUUID->"da054ea7-7898-424b-8f20-ba26dcda115c"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->181179688,ExpressionUUID->"8ad37a03-2479-40ce-8bb8-0b77f545807f"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->916921185,ExpressionUUID->"a1a3e511-7907-45f5-a063-0dffd7b3714c"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"a80de1ac-9b08-4219-aded-7ff6c1432453"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection] lower1)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Intersection] lower1)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["330976770", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:330976770", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:03, 0.045815 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:330976770-1", "3"},
 CellLabel->"In[135]:=",
 CellID->1901417663,ExpressionUUID->"c9408df1-52d0-43ba-9c83-f86df76887a4"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"f8e73c77-28e4-42c3-b96c-3f9d5da76a5b"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection] lower2)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Intersection] lower2)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["567877347", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:567877347", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:03, 0.057037 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:567877347-1", "3"},
 CellLabel->"In[136]:=",
 CellID->1604414518,ExpressionUUID->"cd544126-02a6-4a0d-b5ab-040100796126"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"c7b8f466-1353-4293-9aa2-d25809d369b6"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection] greatest)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Intersection] greatest)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["370774399", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:370774399", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:04, 0.139008 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:370774399-1", "3"},
 CellLabel->"In[137]:=",
 CellID->1352341786,ExpressionUUID->"067b7963-94d6-4a20-aae6-bc77ea005db6"],

Cell[BoxData[
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", "\"\<\[Intersection] lower1\>\"", " ", 
  "\"\<\[Intersection] lower2\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.7650194342430143`*^9, 3.7650194519445744`*^9}},
 CellLabel->"In[138]:=",
 CellID->563438415,ExpressionUUID->"ee7d02d6-2f91-43c3-b10a-841d0491cb48"]
}, Closed]],

Cell[CellGroupData[{

Cell["Union", "Subsection",
 CellChangeTimes->{{3.7302893856560397`*^9, 3.7302893863904696`*^9}},
 CellID->392728112,ExpressionUUID->"eec10edc-69fe-43f8-a984-5fd292e58f17"],

Cell["", "OpenEnvironment",
 CellID->32649250,ExpressionUUID->"3b04699b-4a98-40f5-8d1d-c23bc26329a4"],

Cell[CellGroupData[{

Cell["Lemma (Properties of \[Union])", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.7303579710231347`*^9, 3.7303579742693915`*^9}},
 CellID->473756208,ExpressionUUID->"641b9c20-4123-4a68-818a-9a5a4b3ee1b6"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"x", "\[Element]", 
      RowBox[{"A", "\[Union]", "B"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"x", "\[Element]", "A"}], "\[Or]", 
      RowBox[{"x", "\[Element]", "B"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Element]\[Union])", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730357977174007*^9, 3.730358000924909*^9}},
 CellTags->"\[Element]\[Union]",
 CellLabel->"In[139]:=",
 CellID->24302859,ExpressionUUID->"f52d2f46-1098-465b-90bd-12324ecbac1f"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", 
     RowBox[{"A", "\[Union]", "B"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union]I1)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730358028638996*^9, 3.730358057991708*^9}},
 CellTags->"\[Union]I1",
 CellLabel->"In[140]:=",
 CellID->204712299,ExpressionUUID->"5ea72cf8-f892-430b-956b-10c727c945bf"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", 
     RowBox[{"A", "\[Union]", "B"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union]I2)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303580695249805`*^9, 3.7303580717505183`*^9}},
 CellTags->"\[Union]I2",
 CellLabel->"In[141]:=",
 CellID->118362674,ExpressionUUID->"55fd758c-2ba1-4a9c-a211-2ba0089a2f44"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x"}]], 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"x", "\[NotElement]", "B"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"x", "\[Element]", "A"}], 
      TagBox[")",
       "AutoParentheses"]}]}], ")"}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", 
     RowBox[{"A", "\[Union]", "B"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(\[Union]I classical)", "FrameLabel"], Evaluator -> 
          Automatic, Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "intro"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303580863755255`*^9, 3.7303581256837287`*^9}},
 CellTags->{"\[Union]I classical", "intro"},
 CellLabel->"In[142]:=",
 CellID->71686966,ExpressionUUID->"3dacd0ce-b618-4ccf-8cc9-6fc2a7468b22"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x", ",", "P"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", 
     RowBox[{"A", "\[Union]", "B"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"x", "\[Element]", "A"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", "P"}], ")"}], 
   "\[LongRightArrow]", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"x", "\[Element]", "B"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", "P"}], ")"}], 
   "\[LongRightArrow]", "P"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(\[Union]E)", "FrameLabel"], Evaluator -> Automatic, 
          Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "elim"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730358162072847*^9, 3.7303582088000135`*^9}, {
  3.765083106436204*^9, 3.765083112546167*^9}},
 CellTags->{"\[Union]E", "elim"},
 CellLabel->"In[143]:=",
 CellID->734373493,ExpressionUUID->"cc6aa25c-83a1-4025-8cd9-4fe49291cfe4"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->364554349,ExpressionUUID->"a4411408-efb7-4cf4-83a2-657117340369"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->91957176,ExpressionUUID->"365c834d-b98f-42f8-8f01-89f88fb7b941"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"97452220-41b4-4b9e-b864-1b73a0bd45ba"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Element]\[Union])\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Element]\[Union])\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["24302859", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:24302859", CellTags]], Appearance -> 
          None, Evaluator -> Automatic, Method -> "Preemptive"]]], 
       " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:05, 0.078047 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:24302859-1", "3"},
 CellLabel->"In[144]:=",
 CellID->910429301,ExpressionUUID->"5d01d301-0853-41ec-8b5b-f654db10d92f"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"567eaff8-357b-4529-b838-22821ef29b5d"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union]I1)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[Union]I1)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["204712299", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:204712299", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:05, 0.104256 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:204712299-1", "3"},
 CellLabel->"In[145]:=",
 CellID->2106560278,ExpressionUUID->"58cda68d-cdf0-4074-9690-d94a7ad8e87b"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"36729887-7535-4424-a50f-042de15c10ed"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union]I2)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[Union]I2)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["118362674", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:118362674", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:06, 0.123081 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:118362674-1", "3"},
 CellLabel->"In[146]:=",
 CellID->430856106,ExpressionUUID->"5ab74a01-dcb4-4410-adbb-2e6f4bc3d887"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"a6f4fc69-3979-4a7b-a344-0cbb3de85bcc"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union]I classical)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Union]I classical)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["71686966", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:71686966", CellTags]], Appearance -> 
          None, Evaluator -> Automatic, Method -> "Preemptive"]]], 
       " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:06, 0.168252 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:71686966-1", "3"},
 CellLabel->"In[147]:=",
 CellID->634618502,ExpressionUUID->"c438dae3-6253-4324-8a41-6d2ae06ea6ec"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"d8414e25-6215-4062-96fe-41a077288ca4"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union]E)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(\[Union]E)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["734373493", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:734373493", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:07, 0.213730 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:734373493-1", "3"},
 CellLabel->"In[148]:=",
 CellID->1988227689,ExpressionUUID->"7192f13d-fafc-45a0-9eba-e5628635aa2e"],

Cell[BoxData[
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", 
  "\"\<\[Element]\[Union]\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.7303561165186214`*^9, 3.730356128810933*^9}, {
  3.7303562816703367`*^9, 3.730356308844104*^9}, {3.7303565362228093`*^9, 
  3.7303565400519514`*^9}, {3.7303566107425056`*^9, 3.7303566150434475`*^9}, {
  3.7303569752673616`*^9, 3.7303569764848404`*^9}, {3.7303573339410152`*^9, 
  3.7303573351247845`*^9}, {3.730357693835204*^9, 3.73035769520304*^9}, {
  3.7303580201129065`*^9, 3.7303580204571686`*^9}, {3.76501949079985*^9, 
  3.765019493192212*^9}},
 CellLabel->"In[149]:=",
 CellID->575866964,ExpressionUUID->"b498f909-0860-4ebc-9ce5-d68efdf15213"],

Cell["", "OpenEnvironment",
 CellID->913765167,ExpressionUUID->"ed69799d-0791-49c4-ae25-a6159e3bbe56"],

Cell[CellGroupData[{

Cell["Lemma (Lattice Properties of \[Union])", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.7303554165512853`*^9, 3.7303554205597715`*^9}},
 CellID->861897061,ExpressionUUID->"0851ea46-50e5-45a0-a430-e7263c195598"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{"A", "\[SubsetEqual]", 
   RowBox[{"A", "\[Union]", "B"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union] upper1)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303553098728037`*^9, 3.730355318682676*^9}},
 CellTags->"\[Union] upper1",
 CellLabel->"In[150]:=",
 CellID->776879610,ExpressionUUID->"23354d54-d353-451c-96d2-7d23f93c1ab9"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{"B", "\[SubsetEqual]", 
   RowBox[{"A", "\[Union]", "B"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union] upper2)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730355336318226*^9, 3.730355338345626*^9}},
 CellTags->"\[Union] upper2",
 CellLabel->"In[151]:=",
 CellID->463603332,ExpressionUUID->"fd46a9f7-6c14-40c4-a4e9-9c816ccbf6cd"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "C"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"B", "\[SubsetEqual]", "C"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"A", "\[Union]", "B"}], "\[SubsetEqual]", "C"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union] least)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303553484382143`*^9, 3.7303553830035524`*^9}},
 CellTags->"\[Union] least",
 CellLabel->"In[152]:=",
 CellID->672172372,ExpressionUUID->"2c6fd105-c6b3-4c91-930d-cdf36722c20d"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->384387934,ExpressionUUID->"8a5c25f6-6ffc-44e4-8865-e8c47b4e9639"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->137981057,ExpressionUUID->"429a5b27-6a34-4289-a130-7628cd5d781c"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"befc947e-25df-402e-9620-5642822c2b80"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union] upper1)\"", "FormReference", StripOnInput -> False],
             True -> StyleBox[
             "\"(\[Union] upper1)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["776879610", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:776879610", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:08, 0.039500 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:776879610-1", "3"},
 CellLabel->"In[153]:=",
 CellID->1308387989,ExpressionUUID->"4273e686-36d1-479a-869b-b04747afdf48"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"b4fa1b5b-d15e-4ff0-92c6-d854b343f38b"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union] upper2)\"", "FormReference", StripOnInput -> False],
             True -> StyleBox[
             "\"(\[Union] upper2)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["463603332", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:463603332", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:08, 0.046949 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:463603332-1", "3"},
 CellLabel->"In[154]:=",
 CellID->283700494,ExpressionUUID->"a0289919-612a-43d3-83be-5e58d75618f3"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"959cba23-f890-412f-8c51-0ccbee574eae"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union] least)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[Union] least)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["672172372", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:672172372", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:09, 0.128022 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:672172372-1", "3"},
 CellLabel->"In[155]:=",
 CellID->563063133,ExpressionUUID->"bacbec27-f6b0-4574-94e4-76900e1b7a43"],

Cell[BoxData[
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", "\"\<\[Union] upper1\>\"", " ", 
  "\"\<\[Union] upper2\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.7650195042342*^9, 3.765019514183869*^9}},
 CellLabel->"In[156]:=",
 CellID->853725528,ExpressionUUID->"8f1d7af7-f61b-48b9-b3bb-f3ee60948a33"]
}, Closed]],

Cell[CellGroupData[{

Cell["Difference", "Subsection",
 CellChangeTimes->{{3.7302893987550087`*^9, 3.73028940056837*^9}},
 CellID->614708801,ExpressionUUID->"4ca4939a-241f-4eb0-8fd4-8da4587bd5c2"],

Cell["", "OpenEnvironment",
 CellID->850854527,ExpressionUUID->"66a30973-35c5-465b-808e-ef3b010f0f47"],

Cell[CellGroupData[{

Cell["Lemma (Properties of - for Sets)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.7303584233775473`*^9, 3.7303584286269064`*^9}},
 CellID->180283027,ExpressionUUID->"f05615d2-d878-40ce-a534-a569ea814ae7"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "-", "B"}], "\[Equal]", 
    RowBox[{"Collect", "[", 
     RowBox[{
      UnderscriptBox["\[Lambda]", "x"], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"x", "\[Element]", "A"}], "\[And]", 
        RowBox[{"x", "\[NotElement]", "B"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(- alt SET)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730358760691406*^9, 3.7303588026053505`*^9}},
 CellTags->"- alt SET",
 CellLabel->"In[157]:=",
 CellID->540115919,ExpressionUUID->"e2f0a575-b7f7-478c-9cc1-7d579996f383"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"x", "\[Element]", 
      RowBox[{"A", "-", "B"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"x", "\[Element]", "A"}], "\[And]", 
      RowBox[{"x", "\[NotElement]", "B"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Element]-)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730358431684712*^9, 3.7303584566973686`*^9}},
 CellTags->"\[Element]-",
 CellLabel->"In[158]:=",
 CellID->201090499,ExpressionUUID->"dbbc2d0c-ae52-42a9-a1a3-6ac77bbec4ef"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[NotElement]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", 
     RowBox[{"A", "-", "B"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(-I SET)", "FrameLabel"], Evaluator -> Automatic, 
          Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "intro"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303584847609076`*^9, 3.730358511345359*^9}},
 CellTags->{"-I SET", "intro"},
 CellLabel->"In[159]:=",
 CellID->17112112,ExpressionUUID->"f5cee395-7010-4ff0-ba10-f12db0c70eb9"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", 
     RowBox[{"A", "-", "B"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(-D1 SET)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.73035853446835*^9, 3.7303585536558914`*^9}},
 CellTags->"-D1 SET",
 CellLabel->"In[160]:=",
 CellID->590579378,ExpressionUUID->"62a66027-3b18-408d-9e80-b3a57cfe1bcf"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x", ",", "P"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", 
     RowBox[{"A", "-", "B"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   "P"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(-D2 SET)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303585661076937`*^9, 3.730358571826558*^9}, {
  3.73035861215505*^9, 3.73035862198454*^9}, {3.765084131503669*^9, 
  3.765084134021306*^9}},
 CellTags->"-D2 SET",
 CellLabel->"In[161]:=",
 CellID->984086,ExpressionUUID->"b24c8d74-3d7a-402c-958b-25766a575785"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x", ",", "P"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", 
     RowBox[{"A", "-", "B"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"x", "\[Element]", "A"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"x", "\[NotElement]", "B"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", "P"}], ")"}], 
   "\[LongRightArrow]", "P"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(-E SET)", "FrameLabel"], Evaluator -> Automatic, 
          Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "elim"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730358636014018*^9, 3.7303586613309984`*^9}, {
  3.76508413691964*^9, 3.7650841405091667`*^9}, {3.765084260693514*^9, 
  3.765084282403314*^9}},
 CellTags->{"-E SET", "elim"},
 CellLabel->"In[162]:=",
 CellID->128535443,ExpressionUUID->"a15f38a2-1024-4357-b3ae-3d4405f34edf"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"-", "A"}], "\[Equal]", 
    RowBox[{"UNIV", "-", "A"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(complement \[Equal] UNIV-)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730358693284683*^9, 3.7303587173024464`*^9}},
 CellTags->"complement \[Equal] UNIV-",
 CellLabel->"In[163]:=",
 CellID->679084152,ExpressionUUID->"25eb9645-ea9d-42ae-8802-e919a21ef9e4"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->353082906,ExpressionUUID->"dee1df3a-a638-4d09-87d2-d08a6bcbe94c"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->760389402,ExpressionUUID->"88ba20c7-9519-47e9-ab87-0b43883e097d"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"ceb95bbd-0728-475d-8822-eca4c9a7862b"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(- alt SET)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(- alt SET)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["540115919", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:540115919", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:10, 0.199983 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:540115919-1", "3"},
 CellLabel->"In[164]:=",
 CellID->2034290086,ExpressionUUID->"413f783b-5cb5-476c-8d17-7d845255998a"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"f173f29d-c16b-4f36-b3cc-187d37ab7b1f"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Element]-)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[Element]-)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["201090499", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:201090499", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:10, 0.082876 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:201090499-1", "3"},
 CellLabel->"In[165]:=",
 CellID->1196091729,ExpressionUUID->"0e3f881b-3ec2-4218-8f49-710e98423e7e"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"7472123b-79be-41d1-941e-b31e3fdf60c0"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(-I SET)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(-I SET)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["17112112", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:17112112", CellTags]], Appearance -> 
          None, Evaluator -> Automatic, Method -> "Preemptive"]]], 
       " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:11, 0.143934 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:17112112-1", "3"},
 CellLabel->"In[166]:=",
 CellID->1855793195,ExpressionUUID->"fae8ecb6-0379-4df0-8561-e4c239de4b21"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"08c966a4-159a-4be4-b05b-8e8e2dc0e7bc"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(-D1 SET)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(-D1 SET)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["590579378", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:590579378", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:11, 0.115020 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:590579378-1", "3"},
 CellLabel->"In[167]:=",
 CellID->87366359,ExpressionUUID->"ff98ca43-6f95-4017-83fe-5e58556b956d"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"19473811-5c9d-44b4-9a2f-f98f265f1409"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(-D2 SET)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(-D2 SET)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["984086", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:984086", CellTags]], Appearance -> 
          None, Evaluator -> Automatic, Method -> "Preemptive"]]], 
       " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:12, 0.135169 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:984086-1", "3"},
 CellLabel->"In[168]:=",
 CellID->738044893,ExpressionUUID->"497b9d54-4fde-4ed0-af44-f8d1fb40d21a"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"6b9c1b24-a701-44e7-bbb0-b125237c2e5a"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(-E SET)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(-E SET)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["128535443", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:128535443", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:12, 0.192189 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:128535443-1", "3"},
 CellLabel->"In[169]:=",
 CellID->187234111,ExpressionUUID->"9ab0c91e-bd51-4993-aa37-27e47b2db8cf"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"349f6524-0dbb-42f2-b4a5-1366f5810c79"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(complement \[Equal] UNIV-)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(complement \[Equal] UNIV-)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["679084152", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:679084152", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:13, 0.209293 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:679084152-1", "3"},
 CellLabel->"In[170]:=",
 CellID->887782668,ExpressionUUID->"526b809c-7826-439b-ac09-9eaa35adbac7"],

Cell[BoxData[
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", "\"\<\[Element]-\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.7303561165186214`*^9, 3.730356128810933*^9}, {
  3.7303562816703367`*^9, 3.730356308844104*^9}, {3.7303565362228093`*^9, 
  3.7303565400519514`*^9}, {3.7303566107425056`*^9, 3.7303566150434475`*^9}, {
  3.7303569752673616`*^9, 3.7303569764848404`*^9}, {3.7303573339410152`*^9, 
  3.7303573351247845`*^9}, {3.730357693835204*^9, 3.73035769520304*^9}, {
  3.7303580201129065`*^9, 3.7303580204571686`*^9}, {3.730358474481353*^9, 
  3.730358477628085*^9}, {3.765019545201787*^9, 3.765019547167543*^9}},
 CellLabel->"In[171]:=",
 CellID->876677958,ExpressionUUID->"2c17d1a6-6bc5-408a-a897-98c78635ecd3"],

Cell["", "OpenEnvironment",
 CellID->147803878,ExpressionUUID->"cb496f36-e22c-4a14-8eb8-d92f7bfa2169"],

Cell[CellGroupData[{

Cell["Lemma (- and \[SubsetEqual])", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.730355557614715*^9, 3.7303555742846184`*^9}},
 CellID->685751960,ExpressionUUID->"3a73a610-399d-4dcc-a7b1-e3e7b30b481b"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   RowBox[{"A", "-", "B"}], "\[SubsetEqual]", 
   "A"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(-\[SubsetEqual])", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730355444624297*^9, 3.7303554532007313`*^9}},
 CellTags->"-\[SubsetEqual]",
 CellLabel->"In[172]:=",
 CellID->917574097,ExpressionUUID->"d07f39c7-4b49-4a87-bb3c-ecff0503e8a1"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"A", "-", "B"}], "\[SubsetEqual]", "C"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", "\[SubsetEqual]", 
      RowBox[{"B", "\[Union]", "C"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(-\[SubsetEqual] iff)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303554814504013`*^9, 3.730355515463417*^9}},
 CellTags->"-\[SubsetEqual] iff",
 CellLabel->"In[173]:=",
 CellID->34104065,ExpressionUUID->"138bd892-f1bc-49e8-a2ee-0d8190289c18"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[Subset]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     UnderscriptBox["\[Exists]", "x"], 
     RowBox[{"x", "\[Element]", 
      RowBox[{"B", "-", "A"}]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Subset] implies \[Exists]\[Element])", "FrameLabel"], 
         Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730354519273844*^9, 3.7303545502285037`*^9}},
 CellTags->"\[Subset] implies \[Exists]\[Element]",
 CellLabel->"In[174]:=",
 CellID->168218769,ExpressionUUID->"5f3dad52-4b63-4783-80a8-4673a24fc263"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->114851357,ExpressionUUID->"97e9c61e-e45f-487a-9870-0ac5d2ba6006"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1543820,ExpressionUUID->"f85aad04-0ff9-48d5-bea1-8e828e74391f"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"e06f49b0-196a-4942-b1c1-3cea0f0dc3d1"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(-\[SubsetEqual])\"", "FormReference", StripOnInput -> False],
             True -> StyleBox[
             "\"(-\[SubsetEqual])\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["917574097", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:917574097", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:14, 0.031607 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:917574097-1", "3"},
 CellLabel->"In[175]:=",
 CellID->1547795340,ExpressionUUID->"94ee77de-dbad-4d4e-97d6-1755cd58f119"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"7b252496-2a15-4d04-9377-a458f87f20d9"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(-\[SubsetEqual] iff)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(-\[SubsetEqual] iff)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["34104065", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:34104065", CellTags]], Appearance -> 
          None, Evaluator -> Automatic, Method -> "Preemptive"]]], 
       " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:15, 0.359943 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:34104065-1", "3"},
 CellLabel->"In[176]:=",
 CellID->838116264,ExpressionUUID->"19482a26-cb3e-4e1f-af6d-49608b2ed75c"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"c2abc790-5c13-42f4-9a5e-d10320d5fa2e"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Subset] implies \[Exists]\[Element])\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(\[Subset] implies \[Exists]\[Element])\"", 
              "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["168218769", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:168218769", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:15, 0.307442 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:168218769-1", "3"},
 CellLabel->"In[177]:=",
 CellID->1909735428,ExpressionUUID->"9127f290-b8cf-4129-98e6-27563e725e5a"],

Cell[BoxData[
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", 
  "\"\<-\[SubsetEqual]\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.7650195823139973`*^9, 3.765019586023138*^9}, {
  3.765084876955523*^9, 3.765084908587023*^9}},
 CellLabel->"In[178]:=",
 CellID->1973892905,ExpressionUUID->"43ad83ae-f229-40f6-816b-725713df2e53"]
}, Closed]],

Cell[CellGroupData[{

Cell["Insertion", "Subsection",
 CellChangeTimes->{{3.7302894111826043`*^9, 3.73028943893574*^9}, {
  3.7302895535532703`*^9, 3.730289567976989*^9}, {3.765095101315874*^9, 
  3.765095101748146*^9}},
 CellID->468103083,ExpressionUUID->"b2d4dae3-8fcd-4813-89b9-e475fa132b8a"],

Cell["", "OpenEnvironment",
 CellID->441329419,ExpressionUUID->"87eab77c-8cfa-4157-807e-4e76d59c7276"],

Cell[CellGroupData[{

Cell["Lemma (Properties of Insert)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.765087696480769*^9, 3.765087699976699*^9}},
 CellID->113481011,ExpressionUUID->"65e5f855-cb62-45c6-af2e-1b5b1b8a6956"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"x", ",", "a", ",", "A"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"x", "\[Element]", 
      RowBox[{"Insert", "[", 
       RowBox[{"a", ",", "A"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"x", "\[Equal]", "a"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[Or]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"x", "\[Element]", "A"}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Element]Insert)", "FrameLabel"], Evaluator -> Automatic,
          Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765019642261903*^9, 3.765019677223444*^9}},
 CellTags->"\[Element]Insert",
 CellLabel->"In[179]:=",
 CellID->833542610,ExpressionUUID->"87cdeabf-648d-464d-b035-5d143d98c87f"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "A"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", "\[Element]", 
    RowBox[{"Insert", "[", 
     RowBox[{"a", ",", "A"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(InsertI1)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7650877980952272`*^9, 3.7650878031992693`*^9}, {
  3.7650915132186747`*^9, 3.7650915264013453`*^9}},
 CellTags->"InsertI1",
 CellLabel->"In[180]:=",
 CellID->265974781,ExpressionUUID->"3dfc2ea8-5b58-4bed-bc5a-781776a474ee"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"x", ",", "a", ",", "A"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Equal]", "a"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", 
     RowBox[{"Insert", "[", 
      RowBox[{"a", ",", "A"}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(InsertI1')", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7650915965572166`*^9, 3.7650916324319344`*^9}},
 CellTags->"InsertI1'",
 CellLabel->"In[181]:=",
 CellID->369554344,ExpressionUUID->"a53e2216-ee84-4b9a-9367-f3854ab8dbba"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"x", ",", "a", ",", "A"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", 
     RowBox[{"Insert", "[", 
      RowBox[{"a", ",", "A"}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(InsertI2)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765091646309437*^9, 3.7650916505597057`*^9}},
 CellTags->"InsertI2",
 CellLabel->"In[182]:=",
 CellID->276809760,ExpressionUUID->"eee79b47-d910-49b8-b643-3d87423df22a"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"x", ",", "a", ",", "A"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"x", "\[NotElement]", "A"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"x", "\[Equal]", "a"}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", 
     RowBox[{"Insert", "[", 
      RowBox[{"a", ",", "A"}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(InsertI classical)", "FrameLabel"], Evaluator -> 
          Automatic, Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "intro"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7650932444280252`*^9, 3.765093282915406*^9}},
 CellTags->{"InsertI classical", "intro"},
 CellLabel->"In[183]:=",
 CellID->794800661,ExpressionUUID->"94305d11-47e3-4dc3-a7b2-5f685d28a398"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"x", ",", "a", ",", "A", ",", "P"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", 
     RowBox[{"Insert", "[", 
      RowBox[{"a", ",", "A"}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"x", "\[Equal]", "a"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", "P"}], ")"}], 
   "\[LongRightArrow]", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"x", "\[Element]", "A"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[LongRightArrow]", "P"}], ")"}], 
   "\[LongRightArrow]", "P"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(InsertE)", "FrameLabel"], Evaluator -> Automatic, 
          Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "elim"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765091669087371*^9, 3.765091719798806*^9}},
 CellTags->{"InsertE", "elim"},
 CellLabel->"In[184]:=",
 CellID->752025544,ExpressionUUID->"e669061b-1a1d-4dbf-970c-ca9f87768bd7"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "A"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"Insert", "[", 
     RowBox[{"a", ",", "A"}], "]"}], "\[Equal]", 
    RowBox[{
     RowBox[{"Collect", "[", 
      RowBox[{
       UnderscriptBox["\[Lambda]", "x"], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"x", "\[Equal]", "a"}], 
        TagBox[")",
         "AutoParentheses"]}]}], "]"}], "\[Union]", "A"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Insert alt)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730358247452802*^9, 3.730358283248752*^9}},
 CellTags->"Insert alt",
 CellLabel->"In[185]:=",
 CellID->60042161,ExpressionUUID->"2c22bdf3-e000-413b-9d7a-df4eabf07af9"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->418443738,ExpressionUUID->"c6618d7d-8e37-46d8-809b-07a6244df8a0"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->162133543,ExpressionUUID->"5cd13aa9-2f40-41ee-af41-e9247a2ba5b2"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"f9774bcc-c815-4354-a124-8ae55d975d42"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Element]Insert)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Element]Insert)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["833542610", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:833542610", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:16, 0.078526 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:833542610-1", "3"},
 CellLabel->"In[186]:=",
 CellID->1603093206,ExpressionUUID->"69efc2d1-78ff-494c-9afc-e15c84128c04"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"f94d9a86-6bd3-4aa8-8863-cfc50ec2c181"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(InsertI1)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(InsertI1)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["265974781", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:265974781", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:17, 0.067993 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:265974781-1", "3"},
 CellLabel->"In[187]:=",
 CellID->1283043907,ExpressionUUID->"e9466e7d-028a-4117-ac46-65b9a92ddf8e"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"37a9d7ad-730b-4492-9008-d6deccbe28a3"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(InsertI1')\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(InsertI1')\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["369554344", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:369554344", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:17, 0.109500 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:369554344-1", "3"},
 CellLabel->"In[188]:=",
 CellID->111935385,ExpressionUUID->"9dffdc4b-0a67-4869-8254-6e3cf0381dfa"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"e8492af9-e8ec-4c8a-a310-0c8428826013"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(InsertI2)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(InsertI2)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["276809760", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:276809760", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:18, 0.119893 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:276809760-1", "3"},
 CellLabel->"In[189]:=",
 CellID->591244542,ExpressionUUID->"53cab25a-3a17-4a43-97f3-cc2b778e43fc"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"e7cbaf7c-28b6-4bdb-b716-feb1be763fbf"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(InsertI classical)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(InsertI classical)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["794800661", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:794800661", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:18, 0.072350 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:794800661-1", "3"},
 CellLabel->"In[190]:=",
 CellID->1050155476,ExpressionUUID->"893df1bb-a5ff-4f0d-a86c-b9a2a51b065b"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"3241e753-c155-48d8-999c-e66d35489d0c"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(InsertE)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(InsertE)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["752025544", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:752025544", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:19, 0.275283 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:752025544-1", "3"},
 CellLabel->"In[191]:=",
 CellID->1183293999,ExpressionUUID->"8908cda5-38e0-4ff0-806e-c21baafb2a94"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"db63303b-2670-4fb1-9b20-57c887510987"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Insert alt)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(Insert alt)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["60042161", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:60042161", CellTags]], Appearance -> 
          None, Evaluator -> Automatic, Method -> "Preemptive"]]], 
       " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:19, 0.189398 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:60042161-1", "3"},
 CellLabel->"In[192]:=",
 CellID->807514983,ExpressionUUID->"acf349b9-ba5d-40be-9cc8-8cf74e4537e3"],

Cell[BoxData[
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", 
  "\"\<\[Element]Insert\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.765019702792677*^9, 3.7650197328610888`*^9}, 
   3.76509184508486*^9},
 CellLabel->"In[193]:=",
 CellID->2044667126,ExpressionUUID->"cb9b2ff7-9b03-4f49-a797-f9f6c3e9e115"],

Cell["", "OpenEnvironment",
 CellID->840329746,ExpressionUUID->"c5bf61ce-25c6-4a30-a6b9-924af40e911f"],

Cell[CellGroupData[{

Cell["Lemma (Insert and \[SubsetEqual])", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.76509180195173*^9, 3.7650918073015137`*^9}, {
  3.765092115817635*^9, 3.7650921166816673`*^9}},
 CellID->955019610,ExpressionUUID->"c0587a64-5e71-4530-bace-5373a518f661"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "a"}]], 
  RowBox[{"A", "\[SubsetEqual]", 
   RowBox[{"Insert", "[", 
    RowBox[{"a", ",", "A"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[SubsetEqual]Insert I)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.73035479088305*^9, 3.730354801100807*^9}, {
  3.765091779151457*^9, 3.765091779861944*^9}},
 CellTags->"\[SubsetEqual]Insert I",
 CellLabel->"In[194]:=",
 CellID->499325579,ExpressionUUID->"1378f62a-f8a1-4388-a3e7-d01a5c35792f"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", 
     RowBox[{"Insert", "[", 
      RowBox[{"x", ",", "B"}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[SubsetEqual]Insert I2)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.730354848320573*^9, 3.730354874359871*^9}},
 CellTags->"\[SubsetEqual]Insert I2",
 CellLabel->"In[195]:=",
 CellID->89181006,ExpressionUUID->"27164a92-0285-4493-bacb-7c74cb461d52"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[NotElement]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"A", "\[SubsetEqual]", 
       RowBox[{"Insert", "[", 
        RowBox[{"x", ",", "B"}], "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"A", "\[SubsetEqual]", "B"}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[SubsetEqual]Insert)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7303549025761228`*^9, 3.730354906649744*^9}, {
  3.7303549521306863`*^9, 3.7303549936932034`*^9}},
 CellTags->"\[SubsetEqual]Insert",
 CellLabel->"In[196]:=",
 CellID->17294591,ExpressionUUID->"fe809dfb-5875-4485-ae29-7091cd6e4d62"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", "\[SubsetEqual]", 
      RowBox[{"Insert", "[", 
       RowBox[{"x", ",", "B"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Piecewise]", GridBox[{
        {
         RowBox[{
          RowBox[{"A", "-", 
           RowBox[{"{", "x", "}"}]}], "\[SubsetEqual]", "B"}], 
         "\[DoubleLeftArrow]", 
         RowBox[{"x", "\[Element]", "A"}]},
        {
         RowBox[{"A", "\[SubsetEqual]", "B"}], "\[DoubleLeftArrow]", "True"}
       }]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[SubsetEqual]Insert iff)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7650936344548607`*^9, 3.7650936937503347`*^9}},
 CellTags->"\[SubsetEqual]Insert iff",
 CellLabel->"In[197]:=",
 CellID->1726053781,ExpressionUUID->"ad82da62-4b0a-490c-8806-1851e7b37e9b"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "x"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"Insert", "[", 
       RowBox[{"x", ",", "B"}], "]"}], "\[SubsetEqual]", "A"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"x", "\[Element]", "A"}], "\[And]", 
      RowBox[{"B", "\[SubsetEqual]", "A"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Insert\[SubsetEqual] iff)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765095510167485*^9, 3.7650955431031647`*^9}},
 CellTags->"Insert\[SubsetEqual] iff",
 CellLabel->"In[198]:=",
 CellID->140408834,ExpressionUUID->"23802000-c362-4a0b-afa6-0ac37540f0d2"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1197911587,ExpressionUUID->"605767b6-42d2-40b2-bbbf-59adf70a6090"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->408625003,ExpressionUUID->"13f5ce1e-b466-42c4-b2a1-1a8e64b5d2db"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"cb520f18-b0fd-4d1d-8b17-abe7b697dfa9"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[SubsetEqual]Insert I)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[SubsetEqual]Insert I)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["499325579", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:499325579", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:20, 0.048499 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:499325579-1", "3"},
 CellLabel->"In[199]:=",
 CellID->1989694051,ExpressionUUID->"033295b6-03ea-4743-b02f-9a029d18a564"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"b28a8e67-3439-4a2c-9142-3df803e7e1dd"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[SubsetEqual]Insert I2)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[SubsetEqual]Insert I2)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["89181006", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:89181006", CellTags]], Appearance -> 
          None, Evaluator -> Automatic, Method -> "Preemptive"]]], 
       " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:21, 0.047565 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:89181006-1", "3"},
 CellLabel->"In[200]:=",
 CellID->169658549,ExpressionUUID->"74457cd9-9415-4a5b-a6ff-c4224aff06d1"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"60ad7e57-6fb1-4d7d-b358-a337001133dd"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[SubsetEqual]Insert)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[SubsetEqual]Insert)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["17294591", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:17294591", CellTags]], Appearance -> 
          None, Evaluator -> Automatic, Method -> "Preemptive"]]], 
       " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:21, 0.211836 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:17294591-1", "3"},
 CellLabel->"In[201]:=",
 CellID->94829629,ExpressionUUID->"539ab388-aaaf-402a-84e8-d34f4c88e778"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"2da03a98-3096-43b1-b6c3-9143b8bf5262"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[SubsetEqual]Insert iff)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[SubsetEqual]Insert iff)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1726053781", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1726053781", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:23, 1.391469 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1726053781-1", "3"},
 CellLabel->"In[202]:=",
 CellID->1321149923,ExpressionUUID->"a019316d-6449-4592-a608-6892d07b6388"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"13184ac0-4325-43e9-a002-b904c4e58bb2"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Insert\[SubsetEqual] iff)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(Insert\[SubsetEqual] iff)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["140408834", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:140408834", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:23, 0.404799 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:140408834-1", "3"},
 CellLabel->"In[203]:=",
 CellID->662871545,ExpressionUUID->"72ddb486-2189-446a-b4f0-dccfaf5bf1dc"],

Cell[BoxData[
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", "\"\<\[SubsetEqual]Insert I\>\"", " ", 
  "\"\<Insert\[SubsetEqual] iff\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.76509184106358*^9, 3.765091843205068*^9}, {
  3.765095564455844*^9, 3.765095569270553*^9}},
 CellLabel->"In[204]:=",
 CellID->781604907,ExpressionUUID->"313ba586-bc6d-4b7d-91aa-122f9e5bdcfe"]
}, Closed]],

Cell[CellGroupData[{

Cell["Singletons", "Subsection",
 CellChangeTimes->{{3.765095116403945*^9, 3.765095117987976*^9}},
 CellID->1045098640,ExpressionUUID->"4cb24e1a-2154-43e3-b47a-fb2ec281cec9"],

Cell["", "OpenEnvironment",
 CellID->1232065355,ExpressionUUID->"ecb922b7-b44e-49bc-aa68-afb8aafac54e"],

Cell[CellGroupData[{

Cell["Lemma (Properties of Singletons)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.7650952004590197`*^9, 3.765095204187005*^9}},
 CellID->1042012196,ExpressionUUID->"678aade1-04d8-4c3d-b135-cd7641fc0403"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "a"], 
  RowBox[{"a", "\[Element]", 
   RowBox[{"{", "a", "}"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(singleton I)", "FrameLabel"], Evaluator -> Automatic, 
          Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "intro"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765095206723041*^9, 3.765095212939561*^9}},
 CellTags->{"singleton I", "intro"},
 CellLabel->"In[205]:=",
 CellID->1706541997,ExpressionUUID->"154c747d-9b88-42a4-8a21-26f3dd0c0a24"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"x", ",", "a"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Element]", 
     RowBox[{"{", "a", "}"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"x", "\[Equal]", "a"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(singleton D)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7650952251228857`*^9, 3.765095246458469*^9}},
 CellTags->"singleton D",
 CellLabel->"In[206]:=",
 CellID->1946555487,ExpressionUUID->"390adf86-f8b8-4097-9da9-ab6551101d71"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "b"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"{", "a", "}"}], "\[Equal]", 
     RowBox[{"{", "b", "}"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"a", "\[Equal]", "b"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(singleton inject)", "FrameLabel"], Evaluator -> Automatic,
          Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765095339617463*^9, 3.7650953582576847`*^9}},
 CellTags->"singleton inject",
 CellLabel->"In[207]:=",
 CellID->1642372132,ExpressionUUID->"cb45ee28-969d-4085-b7bf-1d4e9c010512"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->776643782,ExpressionUUID->"08a1df6f-26c7-4bce-bc36-fc4843beb8c6"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1880661924,ExpressionUUID->"fe979dd2-f120-4cb7-a66a-354d90db237a"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"b5106d7b-9095-410f-a4d7-42f2d9662904"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(singleton I)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(singleton I)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1706541997", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1706541997", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:24, 0.016490 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1706541997-1", "3"},
 CellLabel->"In[208]:=",
 CellID->1610428281,ExpressionUUID->"24b89c28-f6f2-4a2d-8f6c-d7faa64ba567"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"348455a7-7415-4df2-98c3-e33de3bade7a"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(singleton D)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(singleton D)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1946555487", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1946555487", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:25, 0.133871 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1946555487-1", "3"},
 CellLabel->"In[209]:=",
 CellID->44899502,ExpressionUUID->"c797b61f-6a99-4d82-9375-e86eb679b1f6"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"b589f37e-f55b-4666-b3b3-5d79ef1a29b3"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(singleton inject)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(singleton inject)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1642372132", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1642372132", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:26, 0.209722 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1642372132-1", "3"},
 CellLabel->"In[210]:=",
 CellID->259575914,ExpressionUUID->"2d77d2ad-50ea-45a7-8a70-0483b95c475c"],

Cell[BoxData[
 RowBox[{
  RowBox[{"facts", " ", "\"\<singleton E\>\""}], ":", " ", 
  RowBox[{"\"\<singleton D\>\"", "[", "elimination", "]"}]}]], "TheoremaCode",
 CellChangeTimes->{{3.765095275548415*^9, 3.765095300609846*^9}},
 CellLabel->"In[211]:=",
 CellID->1158961922,ExpressionUUID->"8e8464d3-eac7-4d6a-ad01-5226f6a46e5a"],

Cell["", "OpenEnvironment",
 CellID->1073790972,ExpressionUUID->"29362c51-68af-4f8a-9f81-27252ecddf5a"],

Cell[CellGroupData[{

Cell["Lemma (Singletons and \[SubsetEqual])", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.7650953946755943`*^9, 3.765095398185135*^9}},
 CellID->1080577362,ExpressionUUID->"bf79ca0b-9ea0-44ab-857b-ef58982a7927"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "a"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", "\[SubsetEqual]", 
      RowBox[{"{", "a", "}"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"A", "\[Equal]", 
        RowBox[{"{", "}"}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[Or]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"A", "\[Equal]", 
        RowBox[{"{", "a", "}"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[SubsetEqual] singleton)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765095401264717*^9, 3.7650954376484013`*^9}},
 CellTags->"\[SubsetEqual] singleton",
 CellLabel->"In[212]:=",
 CellID->574835025,ExpressionUUID->"ba3c94b4-891e-4275-8bbf-58fc308ba014"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "A"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"{", "a", "}"}], "\[SubsetEqual]", "A"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", "\[Element]", "A"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(singleton \[SubsetEqual])", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765095466855846*^9, 3.76509548584792*^9}},
 CellTags->"singleton \[SubsetEqual]",
 CellLabel->"In[213]:=",
 CellID->316263559,ExpressionUUID->"987ead35-8676-46b5-86a2-3f01a4952119"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->636142242,ExpressionUUID->"139e9921-f326-44c0-a744-c2fd7ef29831"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->903528505,ExpressionUUID->"bf48eb20-6c77-4107-b527-21098333b892"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"70945569-3721-43a8-acb1-a23e8da99392"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[SubsetEqual] singleton)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[SubsetEqual] singleton)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["574835025", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:574835025", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:27, 0.465143 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:574835025-1", "3"},
 CellLabel->"In[214]:=",
 CellID->991759232,ExpressionUUID->"b866f190-0698-4884-9f66-30d4006e9fd2"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"9a0fc187-e943-4f16-bfe8-ed5cafb86d5c"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(singleton \[SubsetEqual])\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(singleton \[SubsetEqual])\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["316263559", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:316263559", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:27, 0.087211 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:316263559-1", "3"},
 CellLabel->"In[215]:=",
 CellID->811999371,ExpressionUUID->"61354163-ddd7-4786-856c-489f55e87f87"]
}, Closed]],

Cell[CellGroupData[{

Cell["Image under Functions", "Subsection",
 CellChangeTimes->{{3.730289446687948*^9, 3.730289451564494*^9}},
 CellID->633366498,ExpressionUUID->"42bf86ae-570f-4237-bfc5-18eeb00d39a2"],

Cell["", "OpenEnvironment",
 CellID->38110315,ExpressionUUID->"e7046d20-e426-433f-ae66-ffea17fe847c"],

Cell[CellGroupData[{

Cell["Lemma (Properties of @)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.765098525764984*^9, 3.765098528034169*^9}},
 CellID->805541198,ExpressionUUID->"1ba0e1b6-a89f-4a7b-8466-b6c7139908fa"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"b", ",", "a", ",", "f", ",", "A"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"b", "\[Equal]", 
     RowBox[{"f", "[", "a", "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"a", "\[Element]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"b", "\[Element]", 
     RowBox[{"f", "@", "A"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(@I eq)", "FrameLabel"], Evaluator -> Automatic, 
          Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "intro"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7650985395295563`*^9, 3.765098615168549*^9}},
 CellTags->{"@I eq", "intro"},
 CellLabel->"In[216]:=",
 CellID->1595981918,ExpressionUUID->"81f4ee41-7d1b-4661-86c1-cd51f6a030b1"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"b", ",", "a", ",", "f", ",", "A"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"a", "\[Element]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"b", "\[Equal]", 
     RowBox[{"f", "[", "a", "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"b", "\[Element]", 
     RowBox[{"f", "@", "A"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(@I eq reverse)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765098673971974*^9, 3.765098690863871*^9}},
 CellTags->"@I eq reverse",
 CellLabel->"In[217]:=",
 CellID->796685651,ExpressionUUID->"31466d48-898a-47c6-9519-350697bfe016"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "f", ",", "A"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"a", "\[Element]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"f", "[", "a", "]"}], "\[Element]", 
     RowBox[{"f", "@", "A"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(@I)", "FrameLabel"], Evaluator -> Automatic, Appearance -> 
         None, ButtonFunction :> Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765098706090934*^9, 3.765098716007629*^9}},
 CellTags->"@I",
 CellLabel->"In[218]:=",
 CellID->1022035164,ExpressionUUID->"0ce71d43-56f3-43ae-aa1a-139f021caf78"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"b", ",", "f", ",", "A", ",", "P"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"b", "\[Element]", 
     RowBox[{"f", "@", "A"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{"(", 
    RowBox[{
     UnderscriptBox["\[And]", "a"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"a", "\[Element]", "A"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[LongRightArrow]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"b", "\[Equal]", 
        RowBox[{"f", "[", "a", "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[LongRightArrow]", "P"}]}], ")"}], 
   "\[LongRightArrow]", "P"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(@E)", "FrameLabel"], Evaluator -> Automatic, Appearance -> 
          None, ButtonFunction :> Theorema`Common`editCellTags[]], "elim"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765098741782901*^9, 3.765098816374379*^9}},
 CellTags->{"@E", "elim"},
 CellLabel->"In[219]:=",
 CellID->114344294,ExpressionUUID->"27b53510-2aa4-42f2-ac57-bb62701ff6aa"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"b", ",", "f", ",", "A"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"b", "\[Element]", 
      RowBox[{"f", "@", "A"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{"a", "\[Element]", "A"}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"b", "\[Equal]", 
        RowBox[{"f", "[", "a", "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Element]@)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7650990267311163`*^9, 3.765099048483471*^9}},
 CellTags->"\[Element]@",
 CellLabel->"In[220]:=",
 CellID->1795699175,ExpressionUUID->"87637558-3a97-43a1-b39c-38a749176477"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "g", ",", "A", ",", "B"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[Equal]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{"(", 
    RowBox[{
     UnderscriptBox["\[And]", "x"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"x", "\[Element]", "B"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[LongRightArrow]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"f", "[", "x", "]"}], "\[Equal]", 
        RowBox[{"g", "[", "x", "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}]}], ")"}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"f", "@", "A"}], "\[Equal]", 
     RowBox[{"g", "@", "B"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(@ cong)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651002163487997`*^9, 3.7651002701162357`*^9}},
 CellTags->"@ cong",
 CellLabel->"In[221]:=",
 CellID->724997673,ExpressionUUID->"6066f4bb-70bd-49b6-9c7f-ad807e7e6782"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "A"}]], 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     UnderscriptBox["\[And]", "a"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"a", "\[Element]", "A"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[LongRightArrow]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"f", "[", "a", "]"}], "\[Equal]", "a"}], 
       TagBox[")",
        "AutoParentheses"]}]}]}], ")"}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"f", "@", "A"}], "\[Equal]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(@ ident I)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765100540816618*^9, 3.7651005821838913`*^9}},
 CellTags->"@ ident I",
 CellLabel->"In[222]:=",
 CellID->803636397,ExpressionUUID->"90ad9d86-50dd-4746-b1e9-0a5d82ed98c4"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->369196157,ExpressionUUID->"a27a8bf6-b62b-4690-9fa2-88cba8d59229"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->717570174,ExpressionUUID->"2add8c0f-f5fa-47f5-8a8e-55d3e9d8d11c"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"119f0fae-bbcc-4694-a15f-b0d57a604b96"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(@I eq)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(@I eq)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1595981918", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1595981918", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:28, 0.197740 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1595981918-1", "3"},
 CellLabel->"In[223]:=",
 CellID->638235692,ExpressionUUID->"b5119a56-9b6b-4dc9-baca-4b038f13e49d"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"ff173885-4f48-413b-b88c-36e27652824c"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(@I eq reverse)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(@I eq reverse)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["796685651", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:796685651", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:28, 0.030825 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:796685651-1", "3"},
 CellLabel->"In[224]:=",
 CellID->1222338081,ExpressionUUID->"8db8df19-fdf9-4fa7-b82c-0b2e4f83d460"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"81661931-3a9c-4508-ad7a-39348c5d8b66"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(@I)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(@I)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1022035164", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1022035164", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:29, 0.032559 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1022035164-1", "3"},
 CellLabel->"In[225]:=",
 CellID->1120566427,ExpressionUUID->"6d754d5f-fbc5-4f0b-9d25-e0f40502431f"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"5245a1e9-4c21-4d10-8e84-cb61cf18efe5"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(@E)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(@E)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["114344294", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:114344294", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:29, 0.208705 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:114344294-1", "3"},
 CellLabel->"In[226]:=",
 CellID->1155434313,ExpressionUUID->"0d43d638-054b-4e7b-a14f-d856c779b0cc"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"5619a409-8f86-4bce-86c3-6fc5d6c018e2"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Element]@)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[Element]@)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1795699175", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1795699175", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:30, 0.084494 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1795699175-1", "3"},
 CellLabel->"In[227]:=",
 CellID->1210186748,ExpressionUUID->"fc5a6e81-1d94-4b7a-9f62-d18cc426bf95"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"e8f6d7d7-73ed-4e8d-97de-219253672ad3"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(@ cong)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(@ cong)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["724997673", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:724997673", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:31, 0.511371 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:724997673-1", "3"},
 CellLabel->"In[228]:=",
 CellID->637413599,ExpressionUUID->"c828c5b9-b548-4ffe-a01b-69ead7885f9f"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"098bc725-b0e9-467f-8e0e-61fc36f84635"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(@ ident I)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(@ ident I)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["803636397", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:803636397", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:31, 0.344042 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:803636397-1", "3"},
 CellLabel->"In[229]:=",
 CellID->1659057562,ExpressionUUID->"ed4bf583-b592-4cce-934f-774b532b3b39"],

Cell["", "OpenEnvironment",
 CellID->1195143136,ExpressionUUID->"98413215-ec07-4604-a4d8-fec106943722"],

Cell[CellGroupData[{

Cell["Lemma (@ and \[SubsetEqual])", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.76509963751607*^9, 3.765099639595798*^9}},
 CellID->1012321715,ExpressionUUID->"fb5b4c3b-c7ca-4788-a384-5e7a4b4dacc1"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "A", ",", "B"}]], 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     UnderscriptBox["\[And]", "a"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"a", "\[Element]", "A"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[LongRightArrow]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"f", "[", "a", "]"}], "\[Element]", "B"}], 
       TagBox[")",
        "AutoParentheses"]}]}]}], ")"}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"f", "@", "A"}], "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(@\[SubsetEqual] I)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765099644691783*^9, 3.765099709243663*^9}},
 CellTags->"@\[SubsetEqual] I",
 CellLabel->"In[230]:=",
 CellID->1322094901,ExpressionUUID->"5cc6be16-8334-4ca9-b65d-a945b6a3bac8"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "A", ",", "B", ",", "P"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"B", "\[SubsetEqual]", 
     RowBox[{"f", "@", "A"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{"(", 
    RowBox[{
     UnderscriptBox["\[And]", "C"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"C", "\[SubsetEqual]", "A"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[LongRightArrow]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"B", "\[Equal]", 
        RowBox[{"f", "@", "C"}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[LongRightArrow]", "P"}]}], ")"}], 
   "\[LongRightArrow]", "P"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[SubsetEqual]@ E)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765099746577128*^9, 3.7650998328338947`*^9}},
 CellTags->"\[SubsetEqual]@ E",
 CellLabel->"In[231]:=",
 CellID->1691291058,ExpressionUUID->"b58a06f0-0329-40ed-ba74-3f6c05ed1672"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "A", ",", "B"}]], 
  RowBox[{
   RowBox[{"f", "@", 
    RowBox[{"(", 
     RowBox[{"A", "\[Intersection]", "B"}], ")"}]}], "\[SubsetEqual]", 
   RowBox[{
    RowBox[{"f", "@", "A"}], "\[Intersection]", 
    RowBox[{"f", "@", "B"}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(@\[Intersection])", "FrameLabel"], Evaluator -> Automatic,
          Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765100299555586*^9, 3.765100325387549*^9}},
 CellTags->"@\[Intersection]",
 CellLabel->"In[232]:=",
 CellID->1830553618,ExpressionUUID->"270a4307-6358-4635-9c32-2bef59493519"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "A", ",", "B"}]], 
  RowBox[{
   RowBox[{
    RowBox[{"f", "@", "A"}], "-", 
    RowBox[{"f", "@", "B"}]}], "\[SubsetEqual]", 
   RowBox[{"f", "@", 
    RowBox[{"(", 
     RowBox[{"A", "-", "B"}], ")"}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(@-)", "FrameLabel"], Evaluator -> Automatic, Appearance -> 
         None, ButtonFunction :> Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651003632190638`*^9, 3.7651003766752996`*^9}},
 CellTags->"@-",
 CellLabel->"In[233]:=",
 CellID->398205279,ExpressionUUID->"b21e6dec-dd9c-482a-aafe-5ea92234662d"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2000684334,ExpressionUUID->"f837b76c-8d58-40cc-95d3-42a9bbaed8b8"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->460819985,ExpressionUUID->"6740f01c-0997-4fc5-8b37-c64b52955f49"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"3c4a9d41-371d-4fa2-816e-c4b2ad2c3574"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(@\[SubsetEqual] I)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(@\[SubsetEqual] I)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1322094901", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1322094901", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:32, 0.102456 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1322094901-1", "3"},
 CellLabel->"In[234]:=",
 CellID->892892992,ExpressionUUID->"57da981f-f479-4764-8da0-d6a7a3c080ec"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"35414e2f-8d80-43c9-8d8f-0a1691dd26be"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[SubsetEqual]@ E)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[SubsetEqual]@ E)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1691291058", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1691291058", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:33, 0.955089 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1691291058-1", "3"},
 CellLabel->"In[235]:=",
 CellID->903150912,ExpressionUUID->"7588b18e-aa27-40ac-8f3f-f9e9ae729362"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"28406223-6bda-4604-8502-2292e3830242"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(@\[Intersection])\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(@\[Intersection])\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1830553618", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1830553618", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:34, 0.265301 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1830553618-1", "3"},
 CellLabel->"In[236]:=",
 CellID->1214598633,ExpressionUUID->"85d59c5c-1a6e-4476-8410-ecd019893f54"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"7322f339-d702-4e15-9185-49785a6aa793"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(@-)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(@-)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["398205279", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:398205279", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:34, 0.474826 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:398205279-1", "3"},
 CellLabel->"In[237]:=",
 CellID->1641967303,ExpressionUUID->"16659fac-0e9b-4219-a757-d5db7b81bc81"]
}, Closed]],

Cell[CellGroupData[{

Cell["Range", "Subsection",
 CellChangeTimes->{{3.7651014645968723`*^9, 3.765101465196953*^9}},
 CellID->487407862,ExpressionUUID->"0ee84eed-3054-4410-8186-99c14e7a5127"],

Cell[BoxData[
 RowBox[{
  RowBox[{"abbreviation", " ", 
   RowBox[{"\[OpenCurlyQuote]", "Range", "\[CloseCurlyQuote]"}]}], " ", 
  "\[Equilibrium]", " ", 
  RowBox[{"\[OpenCurlyQuote]", 
   RowBox[{
    UnderscriptBox["\[Lambda]", "f"], 
    RowBox[{"(", 
     RowBox[{"f", "@", "UNIV"}], ")"}]}], 
   "\[CloseCurlyQuote]"}]}]], "TheoremaCode",
 CellChangeTimes->{{3.765101473580922*^9, 3.765101508477399*^9}},
 CellLabel->"In[238]:=",
 CellID->492828318,ExpressionUUID->"1d7b6ce8-4f66-43d1-a7c5-bbedde1496c0"],

Cell["", "OpenEnvironment",
 CellID->20339922,ExpressionUUID->"ada86cdc-f85f-4fc3-9bce-913768b66ba6"],

Cell[CellGroupData[{

Cell["Lemma (Properties of Range)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.7651051165087976`*^9, 3.7651051194323397`*^9}},
 CellID->1961526212,ExpressionUUID->"3c5e1ff0-a3f1-429d-94af-2405d5cdad7a"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"b", ",", "x", ",", "f"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"b", "\[Equal]", 
     RowBox[{"f", "[", "x", "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"b", "\[Element]", 
     RowBox[{"Range", "[", "f", "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(RangeI eq)", "FrameLabel"], Evaluator -> Automatic, 
          Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "intro"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651051217602577`*^9, 3.765105174559478*^9}},
 CellTags->{"RangeI eq", "intro"},
 CellLabel->"In[239]:=",
 CellID->686090881,ExpressionUUID->"65d206a7-f149-48f0-8dc8-445143bf53a0"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "x"}]], 
  RowBox[{
   RowBox[{"f", "[", "x", "]"}], "\[Element]", 
   RowBox[{"Range", "[", "f", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(RangeI)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765105167739294*^9, 3.765105187599813*^9}},
 CellTags->"RangeI",
 CellLabel->"In[240]:=",
 CellID->1169658618,ExpressionUUID->"4f9de9e6-85d8-491a-9fd8-f47ebc444eae"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"b", ",", "f", ",", "P"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"b", "\[Element]", 
     RowBox[{"Range", "[", "f", "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{"(", 
    RowBox[{
     UnderscriptBox["\[And]", "x"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"b", "\[Equal]", 
        RowBox[{"f", "[", "x", "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[LongRightArrow]", "P"}]}], ")"}], 
   "\[LongRightArrow]", "P"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(RangeE)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765105197742157*^9, 3.765105215399426*^9}},
 CellTags->"RangeE",
 CellLabel->"In[241]:=",
 CellID->1129746621,ExpressionUUID->"a0d58dda-dbda-43ce-b842-8bf69646e1dc"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "a", ",", "x"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"Range", "[", "f", "]"}], "\[Equal]", 
     RowBox[{"{", "a", "}"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"f", "[", "x", "]"}], "\[Equal]", "a"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Range singleton D)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765105229262662*^9, 3.7651052847339907`*^9}},
 CellTags->"Range singleton D",
 CellLabel->"In[242]:=",
 CellID->1805361717,ExpressionUUID->"35257160-6070-44f3-b494-ab56075ee0e6"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->730000772,ExpressionUUID->"7b7fc611-b97e-459b-b46f-5cbc8b7a53c9"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2127657008,ExpressionUUID->"b2bfdac4-916a-4690-9d8c-92eaeeb272a6"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"9cf3e602-ec9e-4ffa-8b42-40b19c17815a"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(RangeI eq)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(RangeI eq)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["686090881", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:686090881", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:35, 0.016936 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:686090881-1", "3"},
 CellLabel->"In[243]:=",
 CellID->1329697308,ExpressionUUID->"6e01bd25-a1a6-443d-b697-ce5e66313c66"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"fc370f79-7e21-47ed-8ecb-38df0366e24e"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(RangeI)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(RangeI)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1169658618", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1169658618", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:35, 0.014509 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1169658618-1", "3"},
 CellLabel->"In[244]:=",
 CellID->65583140,ExpressionUUID->"8e614f57-fd84-4b20-aada-099cb8f5fd99"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"69c5a1f2-45fa-4d1f-984e-764f978fa4ab"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(RangeE)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(RangeE)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1129746621", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1129746621", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:35, 0.031267 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1129746621-1", "3"},
 CellLabel->"In[245]:=",
 CellID->1205185338,ExpressionUUID->"ec01b3f4-f18b-473a-adaa-b64c4b1794c1"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"96c5750d-2790-473d-8cb0-e72beff94180"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Range singleton D)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(Range singleton D)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1805361717", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1805361717", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:35, 0.172052 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1805361717-1", "3"},
 CellLabel->"In[246]:=",
 CellID->774665428,ExpressionUUID->"796ac993-5d77-429a-8ec4-94a6354c467e"]
}, Closed]],

Cell[CellGroupData[{

Cell["Set Comprehension", "Subsection",
 CellChangeTimes->{{3.765098955323839*^9, 3.7650989599324303`*^9}},
 CellID->312522199,ExpressionUUID->"1bf74fab-e60d-4c22-8562-ea637d43a537"],

Cell["", "OpenEnvironment",
 CellID->1020901598,ExpressionUUID->"7efc5894-c3fc-4ee4-85d6-d1111ba346ed"],

Cell[CellGroupData[{

Cell["Lemma (Properties of Set Comprehension)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.765100662398856*^9, 3.765100668574829*^9}},
 CellID->1926779225,ExpressionUUID->"1ef8f7c3-8722-45c0-aadd-88cabf8fbe16"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "P", ",", "a"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", "\[Element]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"f", "[", "x", "]"}], 
        UnderscriptBox["|", "x"], 
        RowBox[{"P", "[", "x", "]"}]}], "}"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{"P", "[", "x", "]"}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"a", "\[Equal]", 
        RowBox[{"f", "[", "x", "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Element]SetOf)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765100767893818*^9, 3.765100829693288*^9}},
 CellTags->"\[Element]SetOf",
 CellLabel->"In[247]:=",
 CellID->718424401,ExpressionUUID->"7a7bc97e-48fb-447b-ba94-7c245728934d"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "P", ",", "a", ",", "x"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"a", "\[Equal]", 
     RowBox[{"f", "[", "x", "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{"P", "[", "x", "]"}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"a", "\[Element]", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"f", "[", "y", "]"}], 
       UnderscriptBox["|", "y"], 
       RowBox[{"P", "[", "y", "]"}]}], "}"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(SetOf I)", "FrameLabel"], Evaluator -> Automatic, 
          Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "intro"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651011343053293`*^9, 3.765101193488928*^9}},
 CellTags->{"SetOf I", "intro"},
 CellLabel->"In[248]:=",
 CellID->783246906,ExpressionUUID->"ba2298fa-7bbb-45c0-b5bb-c841cdf4f95a"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "P", ",", "a", ",", "R"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"a", "\[Element]", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"f", "[", "y", "]"}], 
       UnderscriptBox["|", "y"], 
       RowBox[{"P", "[", "y", "]"}]}], "}"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{"(", 
    RowBox[{
     UnderscriptBox["\[And]", "x"], 
     RowBox[{
      RowBox[{"P", "[", "x", "]"}], "\[LongRightArrow]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"a", "\[Equal]", 
        RowBox[{"f", "[", "x", "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[LongRightArrow]", "R"}]}], ")"}], 
   "\[LongRightArrow]", "R"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        TooltipBox[
         ButtonBox[
          StyleBox["(SetOf E)", "FrameLabel"], Evaluator -> Automatic, 
          Appearance -> None, ButtonFunction :> 
          Theorema`Common`editCellTags[]], "elim"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.76510119958442*^9, 3.765101253655827*^9}},
 CellTags->{"SetOf E", "elim"},
 CellLabel->"In[249]:=",
 CellID->1984656332,ExpressionUUID->"6eb18b5a-43f0-4bf0-91cb-fa1b2deac1c6"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->523083548,ExpressionUUID->"7b4d5aff-6d8b-4198-bec5-ce161dfa81f5"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->663433186,ExpressionUUID->"98f8d90f-f7e3-46c9-9808-6cb2ecad7f44"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"74d5f8a2-6df7-48d9-8346-ec80e82fe654"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Element]SetOf)\"", "FormReference", StripOnInput -> False],
             True -> StyleBox[
             "\"(\[Element]SetOf)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["718424401", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:718424401", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:36, 0.050319 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:718424401-1", "3"},
 CellLabel->"In[250]:=",
 CellID->961492024,ExpressionUUID->"62e40651-f209-4327-b387-490f4c46bb3d"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"b8d107bd-d10b-4f87-b705-45f032f23947"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(SetOf I)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(SetOf I)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["783246906", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:783246906", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:36, 0.175135 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:783246906-1", "3"},
 CellLabel->"In[251]:=",
 CellID->1483483416,ExpressionUUID->"8842b459-89de-4cca-be91-b8c2e9a5746a"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"f904acd2-883f-4bba-ae6f-2e8e717914b5"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(SetOf E)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(SetOf E)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1984656332", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1984656332", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:36, 0.173411 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1984656332-1", "3"},
 CellLabel->"In[252]:=",
 CellID->1243944656,ExpressionUUID->"3432589c-d2fc-4a23-a0c8-b28e516663cd"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Further Operations and Lemmas", "Section",
 CellChangeTimes->{{3.730289494970537*^9, 3.7302894996907377`*^9}},
 CellID->176566893,ExpressionUUID->"f4b65ffd-44d6-457b-82f5-f4dd34cd4a9d"],

Cell[CellGroupData[{

Cell["Simplification Rules", "Subsection",
 CellChangeTimes->{{3.730289505006777*^9, 3.730289517480152*^9}},
 CellID->2439801,ExpressionUUID->"7d55af62-2424-4ec7-a94f-63e121930d20"],

Cell["", "OpenEnvironment",
 CellID->339955184,ExpressionUUID->"4bbf9cbd-c25f-4f71-8364-536ea8b87204"],

Cell[CellGroupData[{

Cell["Lemma (Simplification Rules for Set Comprehension)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.765178869612247*^9, 3.765178880028008*^9}, {
  3.765179377574573*^9, 3.765179381757831*^9}},
 CellID->1466695304,ExpressionUUID->"81de3f1e-7371-4d87-ae62-c696956b807f"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "P"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"x", 
      UnderscriptBox["|", "x"], 
      RowBox[{"P", "[", "x", "]"}]}], "}"}], "\[Equal]", 
    RowBox[{"Collect", "[", "P", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(SetOf\[Equal]Collect)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765100749491791*^9, 3.7651007539421673`*^9}},
 CellTags->"SetOf\[Equal]Collect",
 CellLabel->"In[253]:=",
 CellID->314660242,ExpressionUUID->"03aff8c6-ce79-4896-9c4b-ff5194aa4795"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "P"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"x", 
      UnderscriptBox["|", "x"], "P"}], "}"}], "\[Equal]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Piecewise]", GridBox[{
        {"UNIV", "\[DoubleLeftArrow]", "P"},
        {
         RowBox[{"{", "}"}], "\[DoubleLeftArrow]", "True"}
       }]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(SetOf const)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651788972519913`*^9, 3.7651789244756813`*^9}, {
  3.765179394877523*^9, 3.765179401645612*^9}},
 CellTags->"SetOf const",
 CellLabel->"In[254]:=",
 CellID->222426804,ExpressionUUID->"43821a10-5e3f-462a-b9d8-7e55cb502399"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "P"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"x", 
        UnderscriptBox["|", "x"], 
        RowBox[{"P", "[", "x", "]"}]}], "}"}], "\[Equal]", 
      RowBox[{"{", "}"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", "x"], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[Not]", 
        RowBox[{"P", "[", "x", "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(SetOf\[Equal]{} iff)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651789488193913`*^9, 3.765178972290831*^9}, {
  3.7651794198591433`*^9, 3.7651794255815277`*^9}},
 CellTags->"SetOf\[Equal]{} iff",
 CellLabel->"In[255]:=",
 CellID->1237361888,ExpressionUUID->"7a67e7ee-094b-4cec-89ae-429266833c16"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "P"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"x", 
      UnderscriptBox["|", "x"], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[Not]", 
        RowBox[{"P", "[", "x", "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], "}"}], "\[Equal]", 
    RowBox[{"-", 
     RowBox[{"{", 
      RowBox[{"x", 
       UnderscriptBox["|", "x"], 
       RowBox[{"P", "[", "x", "]"}]}], "}"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(SetOf\[Not])", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.7651793589226847`*^9, {3.765179448250449*^9, 3.7651794562050877`*^9}},
 CellTags->"SetOf\[Not]",
 CellLabel->"In[256]:=",
 CellID->165006998,ExpressionUUID->"bc81c816-825b-4560-8ed5-0038c9d1cc89"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "P", ",", "Q"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"f", "[", "x", "]"}], 
      UnderscriptBox["|", "x"], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"P", "[", "x", "]"}], "\[Or]", 
        RowBox[{"Q", "[", "x", "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], "}"}], "\[Equal]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"f", "[", "x", "]"}], 
        UnderscriptBox["|", "x"], 
        RowBox[{"P", "[", "x", "]"}]}], "}"}], "\[Union]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"f", "[", "x", "]"}], 
        UnderscriptBox["|", "x"], 
        RowBox[{"Q", "[", "x", "]"}]}], "}"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(SetOf\[Or])", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651790656979647`*^9, 3.765179096906539*^9}, {
  3.765179473057473*^9, 3.765179529172599*^9}},
 CellTags->"SetOf\[Or]",
 CellLabel->"In[257]:=",
 CellID->1855260115,ExpressionUUID->"3e4ae205-39dd-4613-a34d-75a3ca68670a"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"P", ",", "Q"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"x", 
      UnderscriptBox["|", "x"], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"P", "[", "x", "]"}], "\[Implies]", 
        RowBox[{"Q", "[", "x", "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], "}"}], "\[Equal]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"-", 
       RowBox[{"{", 
        RowBox[{"x", 
         UnderscriptBox["|", "x"], 
         RowBox[{"P", "[", "x", "]"}]}], "}"}]}], "\[Union]", 
      RowBox[{"{", 
       RowBox[{"x", 
        UnderscriptBox["|", "x"], 
        RowBox[{"Q", "[", "x", "]"}]}], "}"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(SetOf\[Implies])", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765179121357572*^9, 3.765179135936781*^9}, {
  3.7651795427387657`*^9, 3.765179547908103*^9}, {3.765179617794088*^9, 
  3.76517962331536*^9}},
 CellTags->"SetOf\[Implies]",
 CellLabel->"In[258]:=",
 CellID->985026514,ExpressionUUID->"ca2c40f5-3c9f-4d06-818b-6843e4fc238f"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"P", ",", "Q"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"x", 
      UnderscriptBox["|", "x"], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"P", "[", "x", "]"}], "\[And]", 
        RowBox[{"Q", "[", "x", "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], "}"}], "\[Equal]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"x", 
        UnderscriptBox["|", "x"], 
        RowBox[{"P", "[", "x", "]"}]}], "}"}], "\[Intersection]", 
      RowBox[{"{", 
       RowBox[{"x", 
        UnderscriptBox["|", "x"], 
        RowBox[{"Q", "[", "x", "]"}]}], "}"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Collect\[And])", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765179165076293*^9, 3.765179171656591*^9}, {
  3.7651795589220324`*^9, 3.765179582420102*^9}, {3.7651804591148376`*^9, 
  3.7651804685206137`*^9}, {3.765180740110579*^9, 3.765180745293524*^9}},
 CellTags->"Collect\[And]",
 CellLabel->"In[259]:=",
 CellID->636400335,ExpressionUUID->"e7181290-0067-4cd3-bfdd-00152da0d246"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"P", ",", "Q"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"x", 
        UnderscriptBox["|", "x"], 
        RowBox[{"P", "[", "x", "]"}]}], "}"}], "\[SubsetEqual]", 
      RowBox[{"{", 
       RowBox[{"x", 
        UnderscriptBox["|", "x"], 
        RowBox[{"Q", "[", "x", "]"}]}], "}"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{"P", "[", "x", "]"}]], 
      RowBox[{"Q", "[", "x", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(SetOf mono iff)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651792047401953`*^9, 3.765179234655545*^9}, {
  3.765179627916378*^9, 3.7651796562745123`*^9}},
 CellTags->"SetOf mono iff",
 CellLabel->"In[260]:=",
 CellID->859120159,ExpressionUUID->"8003e81d-2011-440e-b291-7344a40018ad"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2284427,ExpressionUUID->"a43204a5-e280-4589-9668-2b3b6887ef3e"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2066769854,ExpressionUUID->"1f713404-7561-4a92-94ec-d3b14a2bd4b0"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"474a98d2-3a18-4fd9-8582-219036e3a041"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(SetOf\[Equal]Collect)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(SetOf\[Equal]Collect)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["314660242", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:314660242", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:37, 0.145023 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:314660242-1", "3"},
 CellLabel->"In[261]:=",
 CellID->462247040,ExpressionUUID->"90ce6af8-8154-479f-a67a-eb82f1369496"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"91afbf6c-b72e-4297-b433-225f6505c3cd"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(SetOf const)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(SetOf const)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["222426804", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:222426804", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:38, 0.642879 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:222426804-1", "3"},
 CellLabel->"In[262]:=",
 CellID->1447843469,ExpressionUUID->"23123a44-70d0-4b7e-af9d-66586f2c7c3b"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"ac14e6b6-003e-427c-a976-3c5f4069de50"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(SetOf\[Equal]{} iff)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(SetOf\[Equal]{} iff)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1237361888", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1237361888", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:38, 0.178330 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1237361888-1", "3"},
 CellLabel->"In[263]:=",
 CellID->779088903,ExpressionUUID->"216bcc67-94ae-452d-9f9d-fb1577639bae"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"f8c7d2f5-ea89-4834-8897-f41fa88f5af4"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(SetOf\[Not])\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(SetOf\[Not])\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["165006998", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:165006998", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:39, 0.191282 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:165006998-1", "3"},
 CellLabel->"In[264]:=",
 CellID->1390818865,ExpressionUUID->"5156a1b2-c4e6-447c-8758-161115f33ced"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"e265d191-6c84-4464-9086-d879020d4abb"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(SetOf\[Or])\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(SetOf\[Or])\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1855260115", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1855260115", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:39, 0.476374 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1855260115-1", "3"},
 CellLabel->"In[265]:=",
 CellID->859972040,ExpressionUUID->"627a6140-2e49-4960-8e4a-725e778a5b7a"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"4bf89faa-10ea-4d91-a246-af006053bb80"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(SetOf\[Implies])\"", "FormReference", StripOnInput -> False],
             True -> StyleBox[
             "\"(SetOf\[Implies])\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["985026514", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:985026514", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:40, 0.323752 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:985026514-1", "3"},
 CellLabel->"In[266]:=",
 CellID->2146486548,ExpressionUUID->"d35e54e6-1cd4-4c97-a6a7-67a76bf2b9ca"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"76ce6c39-f80a-4fa7-8e34-43e26beb9abc"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Collect\[And])\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(Collect\[And])\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["636400335", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:636400335", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:40, 0.246729 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:636400335-1", "3"},
 CellLabel->"In[267]:=",
 CellID->1733058477,ExpressionUUID->"3f1308e2-2bf7-43a6-b603-e554f2db72a9"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"cffd3b2d-33f1-4b35-b06f-1574df818f84"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(SetOf mono iff)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(SetOf mono iff)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["859120159", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:859120159", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:40, 0.204757 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:859120159-1", "3"},
 CellLabel->"In[268]:=",
 CellID->1491492338,ExpressionUUID->"43fc8102-da0a-4d32-8477-1d0fb499a635"],

Cell[BoxData[
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", "\"\<SetOf const\>\"", " ", 
  "\"\<SetOf\[Equal]{} iff\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.7651797432333403`*^9, 3.765179756337021*^9}},
 CellLabel->"In[269]:=",
 CellID->549704930,ExpressionUUID->"8ba5107c-90ea-4a93-8483-b37b37bd82a7"],

Cell["", "OpenEnvironment",
 CellID->1157303599,ExpressionUUID->"36e3f8b0-332b-4db3-b8d3-0dec5ffafd3b"],

Cell[CellGroupData[{

Cell["Lemma (Simplification Rules for {})", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.76518103930553*^9, 3.7651810513549767`*^9}, {
  3.7651812337346373`*^9, 3.765181234494952*^9}},
 CellID->2127892673,ExpressionUUID->"d310918b-c9e9-4184-9dad-829213bfcc3c"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", "\[SubsetEqual]", 
      RowBox[{"{", "}"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", "\[Equal]", 
      RowBox[{"{", "}"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[SubsetEqual]{})", "FrameLabel"], Evaluator -> Automatic,
          Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651810697452917`*^9, 3.7651810851289997`*^9}},
 CellTags->"\[SubsetEqual]{}",
 CellLabel->"In[270]:=",
 CellID->485185733,ExpressionUUID->"3055a66b-2e7f-420c-b9ad-a698d5a357e9"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", "x"], 
      RowBox[{"x", "\[NotElement]", "A"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", "\[Equal]", 
      RowBox[{"{", "}"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[ForAll]\[NotElement] conv)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765181236583219*^9, 3.765181252230797*^9}},
 CellTags->"\[ForAll]\[NotElement] conv",
 CellLabel->"In[271]:=",
 CellID->775376237,ExpressionUUID->"6cb5ed56-c05b-44f8-91f8-042d5b96b3ca"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", "x"], 
      RowBox[{"x", "\[Element]", "A"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", "\[NotEqual]", 
      RowBox[{"{", "}"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Exists]\[Element] conv)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651812612762423`*^9, 3.765181269830577*^9}},
 CellTags->"\[Exists]\[Element] conv",
 CellLabel->"In[272]:=",
 CellID->978065715,ExpressionUUID->"c11276ab-818d-47e0-906b-fa4426b2d132"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->965328407,ExpressionUUID->"f5257999-7982-4354-b4df-cb208c4d9524"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2063758297,ExpressionUUID->"b35b36d4-bd17-448f-8b9d-11f16caf926c"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"731cd64e-7c49-440f-8fa1-29e0271932c9"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[SubsetEqual]{})\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[SubsetEqual]{})\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["485185733", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:485185733", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:41, 0.093715 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:485185733-1", "3"},
 CellLabel->"In[273]:=",
 CellID->851152370,ExpressionUUID->"533d16e0-ce09-469a-92e6-88dd693c1a07"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"486ff52a-b856-4513-af6b-9ac9de86c2f4"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[ForAll]\[NotElement] conv)\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(\[ForAll]\[NotElement] conv)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["775376237", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:775376237", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:42, 0.150500 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:775376237-1", "3"},
 CellLabel->"In[274]:=",
 CellID->497842656,ExpressionUUID->"d64e40de-107c-465e-a04b-e0d1316fc389"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"115de8ec-504f-48e2-8547-83bfab704ee4"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Exists]\[Element] conv)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Exists]\[Element] conv)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["978065715", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:978065715", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:42, 0.111690 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:978065715-1", "3"},
 CellLabel->"In[275]:=",
 CellID->118302286,ExpressionUUID->"3dea71aa-c69b-4215-9524-803e9c59e9f2"],

Cell[BoxData[
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", "\"\<\[SubsetEqual]{}\>\"", " ", 
  "\"\<\[ForAll]\[NotElement] conv\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.76518132158987*^9, 3.765181332213958*^9}, {
  3.765181378045506*^9, 3.765181382421444*^9}},
 CellLabel->"In[276]:=",
 CellID->1615646849,ExpressionUUID->"a44bb603-ab15-474b-97cc-c13669158c15"],

Cell["", "OpenEnvironment",
 CellID->159553041,ExpressionUUID->"25e40008-4442-464c-be2b-75e1dcfb5b1f"],

Cell[CellGroupData[{

Cell["Lemma (Simplification Rules for Insert)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.7650950171257963`*^9, 3.7650950235410633`*^9}},
 CellID->1812937043,ExpressionUUID->"755f2645-02c2-414b-b25d-98e22fddc29f"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "x"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"Insert", "[", 
     RowBox[{"x", ",", "UNIV"}], "]"}], "\[Equal]", "UNIV"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Insert UNIV)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765095028061399*^9, 3.765095038949696*^9}},
 CellTags->"Insert UNIV",
 CellLabel->"In[277]:=",
 CellID->988950467,ExpressionUUID->"2d432965-27dc-4fc0-99e8-0b89f9dcb74f"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "A"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"Insert", "[", 
     RowBox[{"a", ",", "A"}], "]"}], "\[Equal]", 
    RowBox[{
     RowBox[{"{", "a", "}"}], "\[Union]", "A"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Insert\[Equal]\[Union])", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765181594482644*^9, 3.765181606394395*^9}},
 CellTags->"Insert\[Equal]\[Union]",
 CellLabel->"In[278]:=",
 CellID->1780572558,ExpressionUUID->"55070231-9ff3-4351-bf0a-aab3ba4c1427"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "A"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"Insert", "[", 
     RowBox[{"a", ",", "A"}], "]"}], "\[NotEqual]", 
    RowBox[{"{", "}"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Insert\[NotEqual]{})", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765181640557193*^9, 3.765181656753914*^9}},
 CellTags->"Insert\[NotEqual]{}",
 CellLabel->"In[279]:=",
 CellID->1795020389,ExpressionUUID->"7464ffb2-77ec-46ff-a00f-bf067e26bbb5"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "A"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"a", "\[Element]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"Insert", "[", 
      RowBox[{"a", ",", "A"}], "]"}], "\[Equal]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Insert absorb)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651816854654837`*^9, 3.7651817021371403`*^9}},
 CellTags->"Insert absorb",
 CellLabel->"In[280]:=",
 CellID->2125656893,ExpressionUUID->"7472c75c-9564-488e-935f-bd257382c5ec"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "A"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"Insert", "[", 
     RowBox[{"a", ",", 
      RowBox[{"Insert", "[", 
       RowBox[{"a", ",", "A"}], "]"}]}], "]"}], "\[Equal]", 
    RowBox[{"Insert", "[", 
     RowBox[{"a", ",", "A"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Insert absorb2)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765181735907152*^9, 3.765181746492466*^9}},
 CellTags->"Insert absorb2",
 CellLabel->"In[281]:=",
 CellID->742456464,ExpressionUUID->"f27f74b0-b865-4980-83e7-8f44091d7719"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "b", ",", "A"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"Insert", "[", 
     RowBox[{"a", ",", 
      RowBox[{"Insert", "[", 
       RowBox[{"b", ",", "A"}], "]"}]}], "]"}], "\[Equal]", 
    RowBox[{"Insert", "[", 
     RowBox[{"b", ",", 
      RowBox[{"Insert", "[", 
       RowBox[{"a", ",", "A"}], "]"}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Insert comm)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651817646133947`*^9, 3.765181774776276*^9}},
 CellTags->"Insert comm",
 CellLabel->"In[282]:=",
 CellID->1038355606,ExpressionUUID->"fd79fcdc-631d-4c21-9d76-e467595a23ef"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "P"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"Insert", "[", 
     RowBox[{"a", ",", 
      RowBox[{"{", 
       RowBox[{"x", 
        UnderscriptBox["|", "x"], 
        RowBox[{"P", "[", "x", "]"}]}], "}"}]}], "]"}], "\[Equal]", 
    RowBox[{"{", 
     RowBox[{"x", 
      UnderscriptBox["|", "x"], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"x", "\[NotEqual]", "a"}], "\[Implies]", 
        RowBox[{"P", "[", "x", "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], "}"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Insert SetOf)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765181901782824*^9, 3.7651819455983467`*^9}},
 CellTags->"Insert SetOf",
 CellLabel->"In[283]:=",
 CellID->657964489,ExpressionUUID->"34779372-bb66-44fe-b2b6-9f99f0d10da0"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"Insert", "[", 
       RowBox[{"a", ",", "A"}], "]"}], "\[Intersection]", 
      RowBox[{"Insert", "[", 
       RowBox[{"a", ",", "B"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Equal]", 
    RowBox[{"Insert", "[", 
     RowBox[{"a", ",", 
      RowBox[{"A", "\[Intersection]", "B"}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Insert\[Intersection]Insert)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765181971453848*^9, 3.7651819958214283`*^9}},
 CellTags->"Insert\[Intersection]Insert",
 CellLabel->"In[284]:=",
 CellID->1602428504,ExpressionUUID->"b767ae43-32fd-4d40-9226-48275463435c"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       RowBox[{"Insert", "[", 
        RowBox[{"a", ",", "A"}], "]"}], "\[Intersection]", "B"}], "\[Equal]", 
      RowBox[{"{", "}"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"a", "\[NotElement]", "B"}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"A", "\[Intersection]", "B"}], "\[Equal]", 
        RowBox[{"{", "}"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Insert disjoint)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765182027141282*^9, 3.7651820689646378`*^9}},
 CellTags->"Insert disjoint",
 CellLabel->"In[285]:=",
 CellID->621646007,ExpressionUUID->"be8df8a3-7c41-4241-8b84-f2d34ca0f3fb"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"B", "\[Intersection]", 
       RowBox[{"Insert", "[", 
        RowBox[{"a", ",", "A"}], "]"}]}], "\[Equal]", 
      RowBox[{"{", "}"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"a", "\[NotElement]", "B"}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"A", "\[Intersection]", "B"}], "\[Equal]", 
        RowBox[{"{", "}"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(disjoint Insert)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651820989122257`*^9, 3.7651821061644917`*^9}},
 CellTags->"disjoint Insert",
 CellLabel->"In[286]:=",
 CellID->656967390,ExpressionUUID->"c9610130-bb90-4d52-a0f4-3b30a68d9d28"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1511772922,ExpressionUUID->"03e57289-00f9-42c5-a6ac-c9e4cdaf7111"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1940050514,ExpressionUUID->"2e8b4d52-ee48-4749-82ff-cb4b4c371c49"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"44be3c20-4437-4525-afa1-b971bacf668d"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Insert UNIV)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(Insert UNIV)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["988950467", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:988950467", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:43, 0.182847 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:988950467-1", "3"},
 CellLabel->"In[287]:=",
 CellID->207220836,ExpressionUUID->"711e6551-2491-4f23-a69e-6ba95e9ee3c5"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"6ac25e2b-2b01-4e82-90dd-b0a8f62f0e87"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Insert\[Equal]\[Union])\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(Insert\[Equal]\[Union])\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1780572558", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1780572558", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:44, 0.220724 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1780572558-1", "3"},
 CellLabel->"In[288]:=",
 CellID->709620784,ExpressionUUID->"ce02b660-2efc-48d5-a410-83f64d3cb0a4"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"949ac89a-566b-49b4-bfd7-e6189bde8cc9"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Insert\[NotEqual]{})\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(Insert\[NotEqual]{})\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1795020389", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1795020389", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:44, 0.257027 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1795020389-1", "3"},
 CellLabel->"In[289]:=",
 CellID->201322071,ExpressionUUID->"a593fce4-239a-48ba-a634-0803bc3ce5f5"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"b5d8c121-cc90-4659-90dc-18832942f34c"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Insert absorb)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(Insert absorb)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["2125656893", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:2125656893", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:44, 0.150124 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:2125656893-1", "3"},
 CellLabel->"In[290]:=",
 CellID->1679122200,ExpressionUUID->"1c164155-3908-47ea-80ff-28bb818b47e3"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"ed03ada7-b804-48a6-b812-409f9ab5dfca"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Insert absorb2)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(Insert absorb2)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["742456464", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:742456464", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:44, 0.026682 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:742456464-1", "3"},
 CellLabel->"In[291]:=",
 CellID->292466109,ExpressionUUID->"fdfb52c1-1f7b-42b0-8e1f-bfb7551b6086"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"60ae8234-8bb4-48f7-a5aa-c0f247115835"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Insert comm)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(Insert comm)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1038355606", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1038355606", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:45, 0.265749 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1038355606-1", "3"},
 CellLabel->"In[292]:=",
 CellID->600513012,ExpressionUUID->"0017ff7e-6593-4c7d-a1d7-1bfa18dabc62"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"a3870010-190d-40c9-b0f2-013515d04644"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Insert SetOf)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(Insert SetOf)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["657964489", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:657964489", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:45, 0.260678 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:657964489-1", "3"},
 CellLabel->"In[293]:=",
 CellID->1781110406,ExpressionUUID->"864a6093-7b93-41ad-b279-fcdce277baa9"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"4e35ac8e-6017-4a07-aa04-ef6471f62c0a"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Insert\[Intersection]Insert)\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(Insert\[Intersection]Insert)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1602428504", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1602428504", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:46, 0.306358 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1602428504-1", "3"},
 CellLabel->"In[294]:=",
 CellID->1228710860,ExpressionUUID->"a1dd04b2-c9d9-4f5c-a60f-43df42997afa"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"42abeb47-6f70-4efa-a5cf-956d5c2cd8ef"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Insert disjoint)\"", "FormReference", StripOnInput -> False],
             True -> StyleBox[
             "\"(Insert disjoint)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["621646007", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:621646007", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:47, 0.910278 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:621646007-1", "3"},
 CellLabel->"In[295]:=",
 CellID->1551927440,ExpressionUUID->"7df4f81d-c45a-43a2-a483-44ffc18f92a3"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"7adbc33d-5bf6-48ea-8c8a-0e8e30eb56ac"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(disjoint Insert)\"", "FormReference", StripOnInput -> False],
             True -> StyleBox[
             "\"(disjoint Insert)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["656967390", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:656967390", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:48, 0.776984 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:656967390-1", "3"},
 CellLabel->"In[296]:=",
 CellID->1616332482,ExpressionUUID->"5a84cad7-a764-4fb6-b914-1258f0a32f54"],

Cell[BoxData[
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", "\"\<Insert UNIV\>\"", " ", 
  "\"\<Insert\[NotEqual]{}\>\"", " ", "\"\<Insert absorb2\>\"", " ", 
  "\"\<Insert\[Intersection]Insert\>\"", " ", "\"\<Insert disjoint\>\"", " ", 
  "\"\<disjoint Insert\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.765095075460599*^9, 3.7650950825249023`*^9}, {
  3.7651817117611933`*^9, 3.7651817181450453`*^9}, {3.765181754336514*^9, 
  3.7651817575846367`*^9}, {3.765182005973281*^9, 3.765182011469049*^9}, {
  3.765182115812243*^9, 3.765182124284017*^9}},
 CellLabel->"In[297]:=",
 CellID->695666758,ExpressionUUID->"927fb3a2-4e6b-4a39-bce2-8bad25de40e1"],

Cell["", "OpenEnvironment",
 CellID->1749328760,ExpressionUUID->"6615bd85-1b25-42bb-ab9b-eff19f9ef5d9"],

Cell[CellGroupData[{

Cell["Lemma (Simplification Rules for \[Intersection])", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.765183644553309*^9, 3.765183652329114*^9}},
 CellID->388765566,ExpressionUUID->"daddb5e7-e3b2-45a5-992b-b90b3f4110cc"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[Intersection]", "A"}], "\[Equal]", "A"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection] absorb)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765183664217329*^9, 3.7651836724567633`*^9}},
 CellTags->"\[Intersection] absorb",
 CellLabel->"In[298]:=",
 CellID->125816224,ExpressionUUID->"fc55d8fa-8b90-4254-897c-2a6575a0f480"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[Intersection]", 
     RowBox[{"(", 
      RowBox[{"A", "\[Intersection]", "B"}], ")"}]}], "\[Equal]", 
    RowBox[{"A", "\[Intersection]", "B"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection] absorb')", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.76518370536052*^9, 3.765183721072303*^9}},
 CellTags->"\[Intersection] absorb'",
 CellLabel->"In[299]:=",
 CellID->848809596,ExpressionUUID->"78e30461-db5c-440a-a771-68e1af989dd5"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[Intersection]", "B"}], "\[Equal]", 
    RowBox[{"B", "\[Intersection]", "A"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection] comm)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765183747007793*^9, 3.765183759760017*^9}},
 CellTags->"\[Intersection] comm",
 CellLabel->"In[300]:=",
 CellID->147215930,ExpressionUUID->"ed2c3178-573c-441d-b883-c094a6cdf7fc"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[Intersection]", 
     RowBox[{"(", 
      RowBox[{"B", "\[Intersection]", "C"}], ")"}]}], "\[Equal]", 
    RowBox[{"B", "\[Intersection]", 
     RowBox[{"(", 
      RowBox[{"A", "\[Intersection]", "C"}], ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection] comm')", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765183780979999*^9, 3.765183798671495*^9}},
 CellTags->"\[Intersection] comm'",
 CellLabel->"In[301]:=",
 CellID->902394792,ExpressionUUID->"9303ebaf-4bff-47f8-b699-5c2016271cf7"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"A", "\[Intersection]", "B"}], ")"}], "\[Intersection]", "C"}], 
    "\[Equal]", 
    RowBox[{"A", "\[Intersection]", 
     RowBox[{"(", 
      RowBox[{"B", "\[Intersection]", "C"}], ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection] assoc)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765183812371831*^9, 3.765183831758554*^9}},
 CellTags->"\[Intersection] assoc",
 CellLabel->"In[302]:=",
 CellID->568250332,ExpressionUUID->"201ed866-e06f-4435-b662-35ed14257a22"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"B", "\[SubsetEqual]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"A", "\[Intersection]", "B"}], "\[Equal]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection] absorb1)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765183881309949*^9, 3.76518392971786*^9}},
 CellTags->"\[Intersection] absorb1",
 CellLabel->"In[303]:=",
 CellID->1826338130,ExpressionUUID->"58f867f3-5e00-4674-b9e0-286a263e5f89"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"A", "\[Intersection]", "B"}], "\[Equal]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection] absorb2)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651839123230658`*^9, 3.76518393490971*^9}},
 CellTags->"\[Intersection] absorb2",
 CellLabel->"In[304]:=",
 CellID->1588902597,ExpressionUUID->"9457e9e2-d7e8-4e67-a0fb-022e6c713401"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[Intersection]", 
     RowBox[{"{", "}"}]}], "\[Equal]", 
    RowBox[{"{", "}"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection]{})", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765183954501279*^9, 3.765183963749325*^9}},
 CellTags->"\[Intersection]{}",
 CellLabel->"In[305]:=",
 CellID->1253019475,ExpressionUUID->"ba8fd0f4-f2cd-4892-b0e0-8f45db7178ec"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"{", "}"}], "\[Intersection]", "A"}], "\[Equal]", 
    RowBox[{"{", "}"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["({}\[Intersection])", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765183981585886*^9, 3.765183986549073*^9}},
 CellTags->"{}\[Intersection]",
 CellLabel->"In[306]:=",
 CellID->1274251332,ExpressionUUID->"a5ac7dd9-350f-4831-9b39-0ac63d3d6e99"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"A", "\[Intersection]", "B"}], "\[Equal]", 
      RowBox[{"{", "}"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", "\[SubsetEqual]", 
      RowBox[{"-", "B"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(disjoint iff \[SubsetEqual]-)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765184021916527*^9, 3.7651840398686132`*^9}},
 CellTags->"disjoint iff \[SubsetEqual]-",
 CellLabel->"In[307]:=",
 CellID->903302990,ExpressionUUID->"6135b89a-9bc7-4d84-b9ab-fd95519b0153"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"A", "\[Intersection]", "B"}], "\[Equal]", 
      RowBox[{"{", "}"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{"x", "\[Element]", "A"}]], 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{"y", "\[Element]", "B"}]], 
       RowBox[{"x", "\[NotEqual]", "y"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(disjoint iff \[NotEqual])", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765184076166401*^9, 3.765184099483757*^9}},
 CellTags->"disjoint iff \[NotEqual]",
 CellLabel->"In[308]:=",
 CellID->920640376,ExpressionUUID->"76d4823b-68ee-4ac8-be6a-79c20b3ff6fb"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[Intersection]", "UNIV"}], "\[Equal]", "A"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection]UNIV)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651841234644527`*^9, 3.7651841281472387`*^9}},
 CellTags->"\[Intersection]UNIV",
 CellLabel->"In[309]:=",
 CellID->1023611974,ExpressionUUID->"7810a2e8-7dec-41dc-8f02-f32b02736408"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"UNIV", "\[Intersection]", "A"}], "\[Equal]", "A"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(UNIV\[Intersection])", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651841397830353`*^9, 3.765184144123046*^9}},
 CellTags->"UNIV\[Intersection]",
 CellLabel->"In[310]:=",
 CellID->1195769099,ExpressionUUID->"187c8311-a111-480e-87b9-378d25295ce1"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[Intersection]", 
     RowBox[{"(", 
      RowBox[{"B", "\[Union]", "C"}], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"A", "\[Intersection]", "B"}], ")"}], "\[Union]", 
     RowBox[{"(", 
      RowBox[{"A", "\[Intersection]", "C"}], ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection]\[Union] dist left)", "FrameLabel"], 
         Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651842261539803`*^9, 3.765184258065668*^9}},
 CellTags->"\[Intersection]\[Union] dist left",
 CellLabel->"In[311]:=",
 CellID->56424478,ExpressionUUID->"aec71586-6dd1-456e-8894-29dcd9143ebd"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"B", "\[Union]", "C"}], ")"}], "\[Intersection]", "A"}], 
    "\[Equal]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"B", "\[Intersection]", "A"}], ")"}], "\[Union]", 
     RowBox[{"(", 
      RowBox[{"C", "\[Intersection]", "A"}], ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection]\[Union] dist right)", "FrameLabel"], 
         Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765184284341588*^9, 3.7651842970253468`*^9}},
 CellTags->"\[Intersection]\[Union] dist right",
 CellLabel->"In[312]:=",
 CellID->1624124987,ExpressionUUID->"252a8781-d3e3-43ef-8e2d-c57e47347aba"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"A", "\[Intersection]", "B"}], "\[Equal]", "UNIV"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"A", "\[Equal]", "UNIV"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"B", "\[Equal]", "UNIV"}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection]\[Equal]UNIV)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651843255606747`*^9, 3.765184346024975*^9}},
 CellTags->"\[Intersection]\[Equal]UNIV",
 CellLabel->"In[313]:=",
 CellID->158694025,ExpressionUUID->"c2d8d38f-e73a-4c20-90e9-10e81ae9316a"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"C", "\[SubsetEqual]", 
      RowBox[{"A", "\[Intersection]", "B"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"C", "\[SubsetEqual]", "A"}], "\[And]", 
      RowBox[{"C", "\[SubsetEqual]", "B"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[SubsetEqual]\[Intersection])", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765184380352116*^9, 3.765184406903822*^9}},
 CellTags->"\[SubsetEqual]\[Intersection]",
 CellLabel->"In[314]:=",
 CellID->1837534884,ExpressionUUID->"b2e42c5d-5b00-41ba-bc88-ffaa02e6c6f4"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"Insert", "[", 
      RowBox[{"a", ",", "A"}], "]"}], "\[Intersection]", "B"}], "\[Equal]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Piecewise]", GridBox[{
        {
         RowBox[{"Insert", "[", 
          RowBox[{"a", ",", 
           RowBox[{"A", "\[Intersection]", "B"}]}], "]"}], 
         "\[DoubleLeftArrow]", 
         RowBox[{"a", "\[Element]", "B"}]},
        {
         RowBox[{"A", "\[Intersection]", "B"}], "\[DoubleLeftArrow]", "True"}
       }]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Insert\[Intersection])", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765186788058325*^9, 3.765186836009894*^9}},
 CellTags->"Insert\[Intersection]",
 CellLabel->"In[315]:=",
 CellID->229432372,ExpressionUUID->"37035f2d-8be1-4d9b-b346-2538b15706a2"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"B", "\[Intersection]", 
     RowBox[{"Insert", "[", 
      RowBox[{"a", ",", "A"}], "]"}]}], "\[Equal]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Piecewise]", GridBox[{
        {
         RowBox[{"Insert", "[", 
          RowBox[{"a", ",", 
           RowBox[{"B", "\[Intersection]", "A"}]}], "]"}], 
         "\[DoubleLeftArrow]", 
         RowBox[{"a", "\[Element]", "B"}]},
        {
         RowBox[{"B", "\[Intersection]", "A"}], "\[DoubleLeftArrow]", "True"}
       }]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection]Insert)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765186872526266*^9, 3.7651868874512444`*^9}},
 CellTags->"\[Intersection]Insert",
 CellLabel->"In[316]:=",
 CellID->2031161567,ExpressionUUID->"5baff602-e322-4acf-a322-433104f4e54a"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->139188812,ExpressionUUID->"22e8a15a-a1da-466c-b92f-7672a0e79078"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->595986403,ExpressionUUID->"d5447440-d4bb-4b59-9c83-6f01b5bea9fb"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"2a52ac56-3919-4fb1-8f3e-1a090454575a"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection] absorb)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Intersection] absorb)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["125816224", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:125816224", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:49, 0.096686 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:125816224-1", "3"},
 CellLabel->"In[317]:=",
 CellID->1628613977,ExpressionUUID->"cf2a0b05-947e-4e08-b681-dbb9bb955bbf"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"44e3ab16-7280-439e-9b1e-a00fd2c33672"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection] absorb')\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Intersection] absorb')\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["848809596", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:848809596", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:49, 0.206446 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:848809596-1", "3"},
 CellLabel->"In[318]:=",
 CellID->2030772045,ExpressionUUID->"36472be4-f0d7-4752-bf1f-5e56ab4cfee6"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"17a11496-ffc8-4a05-9521-80a4a9c92b33"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection] comm)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Intersection] comm)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["147215930", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:147215930", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:50, 0.160375 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:147215930-1", "3"},
 CellLabel->"In[319]:=",
 CellID->548669029,ExpressionUUID->"e0bc9de7-8d38-449f-8c01-166784414f07"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"0d330535-4e4c-43e0-9db3-f312240c0d22"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection] comm')\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Intersection] comm')\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["902394792", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:902394792", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:50, 0.243459 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:902394792-1", "3"},
 CellLabel->"In[320]:=",
 CellID->1223279036,ExpressionUUID->"0c6b49b0-74e6-4b95-a486-249347bf00e4"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"c1882ce3-a06e-4768-897e-e18db253e961"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection] assoc)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Intersection] assoc)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["568250332", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:568250332", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:50, 0.267246 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:568250332-1", "3"},
 CellLabel->"In[321]:=",
 CellID->868040367,ExpressionUUID->"c9ae36de-51e6-4ab1-af37-a9c0f0d09f5f"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"5b761f59-983a-4f7b-b106-3431ba3e0e64"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection] absorb1)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Intersection] absorb1)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1826338130", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1826338130", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:51, 0.069268 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1826338130-1", "3"},
 CellLabel->"In[322]:=",
 CellID->679262902,ExpressionUUID->"263a06fe-fc74-4a56-8db9-29590e7b58ff"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"97bb20fc-63cf-49df-ad17-cd31557898a4"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection] absorb2)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Intersection] absorb2)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1588902597", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1588902597", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:51, 0.083182 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1588902597-1", "3"},
 CellLabel->"In[323]:=",
 CellID->1210453596,ExpressionUUID->"e1bfc212-a53d-43e0-b061-4f254f379381"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"b05f2c16-3391-47ce-b677-a69d6801ab25"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection]{})\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Intersection]{})\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1253019475", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1253019475", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:51, 0.157359 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1253019475-1", "3"},
 CellLabel->"In[324]:=",
 CellID->1968362112,ExpressionUUID->"1a5c661b-1557-4af5-b166-96572bd68d8c"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"e29dd0fa-6b8b-4a3e-a130-2cd89687504b"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"({}\[Intersection])\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"({}\[Intersection])\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1274251332", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1274251332", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:51, 0.117650 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1274251332-1", "3"},
 CellLabel->"In[325]:=",
 CellID->2131093972,ExpressionUUID->"3cc072c7-64ff-40a0-8749-2ddcf91e538c"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"a3231ea5-10e2-4530-bf1f-ef506e3e5d32"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(disjoint iff \[SubsetEqual]-)\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(disjoint iff \[SubsetEqual]-)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["903302990", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:903302990", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:52, 0.308800 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:903302990-1", "3"},
 CellLabel->"In[326]:=",
 CellID->72270293,ExpressionUUID->"8a174da6-9259-4456-a888-bca8c161fadd"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"d6c18da1-2ccf-4ea6-9d8f-5ea3b7fa104d"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(disjoint iff \[NotEqual])\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(disjoint iff \[NotEqual])\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["920640376", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:920640376", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:52, 0.409096 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:920640376-1", "3"},
 CellLabel->"In[327]:=",
 CellID->321131777,ExpressionUUID->"677ec8e8-79c5-4849-ab27-fd71ffc1c532"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"17dcdb9d-9d18-4edc-b67d-3c11f751a751"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection]UNIV)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Intersection]UNIV)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1023611974", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1023611974", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:53, 0.123480 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1023611974-1", "3"},
 CellLabel->"In[328]:=",
 CellID->1877569687,ExpressionUUID->"777b4649-e951-4bda-9678-a8b639463f5b"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"32c8c4c3-410f-439f-90a9-b4298ca26d4a"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(UNIV\[Intersection])\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(UNIV\[Intersection])\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1195769099", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1195769099", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:53, 0.137262 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1195769099-1", "3"},
 CellLabel->"In[329]:=",
 CellID->1618369204,ExpressionUUID->"8338a43c-ea58-4e7d-85ec-4622d938afa3"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"63fc5a4e-bbd8-43ec-a861-f92ca707be2d"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection]\[Union] dist left)\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(\[Intersection]\[Union] dist left)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["56424478", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:56424478", CellTags]], Appearance -> 
          None, Evaluator -> Automatic, Method -> "Preemptive"]]], 
       " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:53, 0.370566 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:56424478-1", "3"},
 CellLabel->"In[330]:=",
 CellID->400332023,ExpressionUUID->"f1380f81-fa44-4763-b4e5-077e4ac37ebd"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"5b35eba5-23b2-47b0-9d51-04750277a02f"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection]\[Union] dist right)\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(\[Intersection]\[Union] dist right)\"", "FormReferenceHover",
               StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1624124987", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1624124987", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:54, 0.357490 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1624124987-1", "3"},
 CellLabel->"In[331]:=",
 CellID->1734538043,ExpressionUUID->"f946aa9b-4bc5-4b6b-a6d6-94285e3a76b4"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"981fcc61-8e93-400f-af22-1bfca1c018de"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection]\[Equal]UNIV)\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(\[Intersection]\[Equal]UNIV)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["158694025", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:158694025", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:54, 0.533958 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:158694025-1", "3"},
 CellLabel->"In[332]:=",
 CellID->1315875576,ExpressionUUID->"edb892ab-bd7f-474d-8831-ba95ca7c0306"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"7e92123b-5b26-40e0-9197-0cb3bd193543"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[SubsetEqual]\[Intersection])\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(\[SubsetEqual]\[Intersection])\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1837534884", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1837534884", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:55, 0.125267 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1837534884-1", "3"},
 CellLabel->"In[333]:=",
 CellID->978925165,ExpressionUUID->"49d48a11-4604-4d84-b391-5e5974a2445a"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"e27090c9-353e-4356-ab6c-154c55ba92b3"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Insert\[Intersection])\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(Insert\[Intersection])\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["229432372", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:229432372", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:57, 2.149579 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:229432372-1", "3"},
 CellLabel->"In[334]:=",
 CellID->387372574,ExpressionUUID->"4ac23cf1-4b24-4d31-90fa-4a0775d734ae"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"d49daba0-4b14-4010-b94a-412512f0811e"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection]Insert)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Intersection]Insert)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["2031161567", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:2031161567", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:57, 0.195878 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:2031161567-1", "3"},
 CellLabel->"In[335]:=",
 CellID->895002178,ExpressionUUID->"812c3dd1-7e84-453c-b984-6de8b7f632e4"],

Cell[BoxData[
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", "\"\<\[Intersection] absorb\>\"", " ", 
  "\"\<\[Intersection]{}\>\"", " ", "\"\<{}\[Intersection]\>\"", " ", 
  "\"\<\[Intersection]UNIV\>\"", " ", "\"\<UNIV\[Intersection]\>\"", " ", 
  "\"\<\[Intersection]\[Equal]UNIV\>\"", " ", 
  "\"\<\[SubsetEqual]\[Intersection]\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.765183683808597*^9, 3.7651836922882843`*^9}, {
  3.765183972701*^9, 3.765183998069016*^9}, {3.765184196930562*^9, 
  3.7651842076745367`*^9}, {3.7651843556803017`*^9, 3.7651843597686567`*^9}, {
  3.765184417383593*^9, 3.765184421983671*^9}},
 CellLabel->"In[336]:=",
 CellID->701072721,ExpressionUUID->"5de92f59-1f70-45b1-9192-629fc3c2efab"],

Cell["", "OpenEnvironment",
 CellID->1941586236,ExpressionUUID->"c18afc6d-00fa-4f34-983d-08f8d5fb3b70"],

Cell[CellGroupData[{

Cell["Lemma (Simplification Rules for \[Union])", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.76518588114952*^9, 3.765185886445302*^9}},
 CellID->1625693420,ExpressionUUID->"23182386-327e-4081-ae82-c7469e1c2fe7"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[Union]", "A"}], "\[Equal]", "A"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union] absorb)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765185903637224*^9, 3.7651859114608517`*^9}},
 CellTags->"\[Union] absorb",
 CellLabel->"In[337]:=",
 CellID->510851566,ExpressionUUID->"1dd930a7-a4a4-4184-a990-a1221b81347e"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[Union]", 
     RowBox[{"(", 
      RowBox[{"A", "\[Union]", "B"}], ")"}]}], "\[Equal]", 
    RowBox[{"A", "\[Union]", "B"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union] absorb')", "FrameLabel"], Evaluator -> Automatic,
          Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765185947617241*^9, 3.765185954548667*^9}},
 CellTags->"\[Union] absorb'",
 CellLabel->"In[338]:=",
 CellID->961990306,ExpressionUUID->"30dd6ca7-e166-4968-b230-3695f88edcbe"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[Union]", "B"}], "\[Equal]", 
    RowBox[{"B", "\[Union]", "A"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union] comm)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765185985740016*^9, 3.7651859896999483`*^9}},
 CellTags->"\[Union] comm",
 CellLabel->"In[339]:=",
 CellID->2140682664,ExpressionUUID->"fffc7ebe-1903-470d-a423-01e5da9bdb37"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[Union]", 
     RowBox[{"(", 
      RowBox[{"B", "\[Union]", "C"}], ")"}]}], "\[Equal]", 
    RowBox[{"B", "\[Union]", 
     RowBox[{"(", 
      RowBox[{"A", "\[Union]", "C"}], ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union] comm')", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765186004728304*^9, 3.7651860136915293`*^9}},
 CellTags->"\[Union] comm'",
 CellLabel->"In[340]:=",
 CellID->1655518259,ExpressionUUID->"63cba1dd-7f2a-42ee-bd5b-dd610f3b54b3"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"A", "\[Union]", "B"}], ")"}], "\[Union]", "C"}], "\[Equal]", 
    RowBox[{"A", "\[Union]", 
     RowBox[{"(", 
      RowBox[{"B", "\[Union]", "C"}], ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union] assoc)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765186027160635*^9, 3.765186036699725*^9}},
 CellTags->"\[Union] assoc",
 CellLabel->"In[341]:=",
 CellID->945600630,ExpressionUUID->"e8e6d705-631c-4c9d-aa97-fefca4115ead"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"A", "\[Union]", "B"}], "\[Equal]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union] absorb1)", "FrameLabel"], Evaluator -> Automatic,
          Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765186051367449*^9, 3.765186067275219*^9}},
 CellTags->"\[Union] absorb1",
 CellLabel->"In[342]:=",
 CellID->837689454,ExpressionUUID->"ad5beaab-93b9-48c6-b2c2-a8f8c1ff258d"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"B", "\[SubsetEqual]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"A", "\[Union]", "B"}], "\[Equal]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union] absorb2)", "FrameLabel"], Evaluator -> Automatic,
          Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651860769280567`*^9, 3.76518608203504*^9}},
 CellTags->"\[Union] absorb2",
 CellLabel->"In[343]:=",
 CellID->1727788553,ExpressionUUID->"38deb9e2-550b-48aa-8740-428d4796982e"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[Union]", 
     RowBox[{"{", "}"}]}], "\[Equal]", "A"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union]{})", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765186100592808*^9, 3.7651861043867598`*^9}},
 CellTags->"\[Union]{}",
 CellLabel->"In[344]:=",
 CellID->985635433,ExpressionUUID->"d6414077-67ef-4ac9-8836-f56a0138097e"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"{", "}"}], "\[Union]", "A"}], "\[Equal]", "A"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["({}\[Union])", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651861130388727`*^9, 3.765186116698339*^9}},
 CellTags->"{}\[Union]",
 CellLabel->"In[345]:=",
 CellID->1158821232,ExpressionUUID->"f0351117-b62d-4593-804a-5846ee59e902"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[Union]", "UNIV"}], "\[Equal]", "UNIV"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union]UNIV)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651861901727953`*^9, 3.765186194953978*^9}},
 CellTags->"\[Union]UNIV",
 CellLabel->"In[346]:=",
 CellID->1958752872,ExpressionUUID->"cc76f1fe-6b2e-417e-b83a-4a1e85689882"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"UNIV", "\[Union]", "A"}], "\[Equal]", "UNIV"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(UNIV\[Union])", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765186205933498*^9, 3.7651862098894653`*^9}},
 CellTags->"UNIV\[Union]",
 CellLabel->"In[347]:=",
 CellID->715721472,ExpressionUUID->"2c1fe514-2923-4862-ae06-ebdba0500fae"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"Insert", "[", 
      RowBox[{"a", ",", "A"}], "]"}], "\[Union]", "B"}], "\[Equal]", 
    RowBox[{"Insert", "[", 
     RowBox[{"a", ",", 
      RowBox[{"A", "\[Union]", "B"}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Insert\[Union])", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765186353047626*^9, 3.765186374919368*^9}},
 CellTags->"Insert\[Union]",
 CellLabel->"In[348]:=",
 CellID->727949230,ExpressionUUID->"88ad4619-c3ec-44e6-9cc9-c15783ae50f8"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"B", "\[Union]", 
     RowBox[{"Insert", "[", 
      RowBox[{"a", ",", "A"}], "]"}]}], "\[Equal]", 
    RowBox[{"Insert", "[", 
     RowBox[{"a", ",", 
      RowBox[{"B", "\[Union]", "A"}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union]Insert)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.76518638476193*^9, 3.7651863944952307`*^9}},
 CellTags->"\[Union]Insert",
 CellLabel->"In[349]:=",
 CellID->1429638428,ExpressionUUID->"1108794c-416f-48fd-94ec-fb8fcba1ca2e"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[Union]", 
     RowBox[{"(", 
      RowBox[{"B", "\[Intersection]", "C"}], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"A", "\[Union]", "B"}], ")"}], "\[Intersection]", 
     RowBox[{"(", 
      RowBox[{"A", "\[Union]", "C"}], ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union]\[Intersection] dist left)", "FrameLabel"], 
         Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765186453587562*^9, 3.7651864656622353`*^9}},
 CellTags->"\[Union]\[Intersection] dist left",
 CellLabel->"In[350]:=",
 CellID->1116053495,ExpressionUUID->"08d29689-ef28-4a06-890c-f9306ee30fd1"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"B", "\[Intersection]", "C"}], ")"}], "\[Union]", "A"}], 
    "\[Equal]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"B", "\[Union]", "A"}], ")"}], "\[Intersection]", 
     RowBox[{"(", 
      RowBox[{"C", "\[Union]", "A"}], ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union]\[Intersection] dist right)", "FrameLabel"], 
         Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.76518647992992*^9, 3.765186490224497*^9}},
 CellTags->"\[Union]\[Intersection] dist right",
 CellLabel->"In[351]:=",
 CellID->1097000890,ExpressionUUID->"b2ce6988-93f6-4266-a5cd-26dc6dbf5711"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"A", "\[Union]", "B"}], "\[Equal]", 
      RowBox[{"{", "}"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"A", "\[Equal]", 
        RowBox[{"{", "}"}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"B", "\[Equal]", 
        RowBox[{"{", "}"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union]\[Equal]{})", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651865134199667`*^9, 3.765186522461587*^9}},
 CellTags->"\[Union]\[Equal]{}",
 CellLabel->"In[352]:=",
 CellID->1016692823,ExpressionUUID->"36426bb0-b926-47c4-aeb4-4508b486267d"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"A", "\[Union]", "B"}], "\[SubsetEqual]", "C"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"A", "\[SubsetEqual]", "C"}], "\[And]", 
      RowBox[{"B", "\[SubsetEqual]", "C"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union]\[SubsetEqual])", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651865564031*^9, 3.765186571957157*^9}},
 CellTags->"\[Union]\[SubsetEqual]",
 CellLabel->"In[353]:=",
 CellID->997515848,ExpressionUUID->"c2879c31-4a67-4636-b191-393fcad6e26d"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1899020357,ExpressionUUID->"828c525d-3bea-47cc-aeab-fe08a6d1713b"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->172320552,ExpressionUUID->"d3d9e47a-fe79-4339-be10-f2548eae4b65"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"3a793dfa-e3b4-4773-abf2-4cb981744cbd"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union] absorb)\"", "FormReference", StripOnInput -> False],
             True -> StyleBox[
             "\"(\[Union] absorb)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["510851566", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:510851566", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:58, 0.087870 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:510851566-1", "3"},
 CellLabel->"In[354]:=",
 CellID->365686177,ExpressionUUID->"eb249535-8173-40d6-bbdb-2c2aae352b07"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"c213907b-504e-4025-b422-4f57a8bd80eb"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union] absorb')\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Union] absorb')\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["961990306", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:961990306", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:59, 0.233095 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:961990306-1", "3"},
 CellLabel->"In[355]:=",
 CellID->1321288291,ExpressionUUID->"f296832d-3220-4ad0-9d80-345f79408b1f"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"bdf2c475-9d20-4e63-8090-bfd9765d0c36"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union] comm)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[Union] comm)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["2140682664", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:2140682664", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:59, 0.180058 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:2140682664-1", "3"},
 CellLabel->"In[356]:=",
 CellID->678754593,ExpressionUUID->"e2b5c4d8-ca86-4739-8add-11abf118bb58"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"800f1a9b-3714-4e94-9520-313035aabb08"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union] comm')\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[Union] comm')\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1655518259", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1655518259", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:00:59, 0.245865 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1655518259-1", "3"},
 CellLabel->"In[357]:=",
 CellID->1904743954,ExpressionUUID->"5592b9fd-2693-411b-becf-a4a020ad57be"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"b8bfa81c-48a2-494f-894d-e4db4821c938"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union] assoc)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[Union] assoc)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["945600630", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:945600630", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:00, 0.256402 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:945600630-1", "3"},
 CellLabel->"In[358]:=",
 CellID->1060575105,ExpressionUUID->"3fc23a3b-63ed-4e0f-9760-2469f4d0179b"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"5946e055-8f01-4f42-9ed1-88c9f314e2e3"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union] absorb1)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Union] absorb1)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["837689454", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:837689454", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:00, 0.058074 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:837689454-1", "3"},
 CellLabel->"In[359]:=",
 CellID->363045913,ExpressionUUID->"a45e0c3a-1b21-41cc-887a-dad678f3e9fe"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"557f6418-636a-4de2-9dd2-07fb59d1f6d6"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union] absorb2)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Union] absorb2)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1727788553", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1727788553", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:00, 0.080552 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1727788553-1", "3"},
 CellLabel->"In[360]:=",
 CellID->56107910,ExpressionUUID->"ccf3a6e2-23b5-4cd1-a76c-3ffdf84cc204"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"8bcfec0e-eea2-4e3a-bcac-42951e2a9634"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union]{})\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[Union]{})\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["985635433", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:985635433", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:01, 0.126436 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:985635433-1", "3"},
 CellLabel->"In[361]:=",
 CellID->1189519127,ExpressionUUID->"e1d8893c-5ca6-4930-940f-940100ee72b4"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"f6548f48-1fc1-4348-894c-963fb19b725f"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"({}\[Union])\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"({}\[Union])\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1158821232", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1158821232", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:01, 0.119133 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1158821232-1", "3"},
 CellLabel->"In[362]:=",
 CellID->772575899,ExpressionUUID->"84574fea-7e8d-4451-afb5-958ee98ab045"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"1a29eedc-7900-4722-a2d5-21da26010718"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union]UNIV)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[Union]UNIV)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1958752872", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1958752872", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:01, 0.141392 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1958752872-1", "3"},
 CellLabel->"In[363]:=",
 CellID->1073197046,ExpressionUUID->"57de62df-9c50-4468-8a9b-27db3a02f067"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"b107c659-9e75-4e59-a3d9-457f276ec703"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(UNIV\[Union])\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(UNIV\[Union])\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["715721472", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:715721472", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:01, 0.125364 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:715721472-1", "3"},
 CellLabel->"In[364]:=",
 CellID->645311516,ExpressionUUID->"091f7b25-0bd4-44cb-ae0f-90617fc64236"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"75f4f0b5-24d9-44d3-bacc-3e7f3ee27316"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Insert\[Union])\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(Insert\[Union])\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["727949230", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:727949230", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:02, 0.274502 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:727949230-1", "3"},
 CellLabel->"In[365]:=",
 CellID->1612445395,ExpressionUUID->"f96d0823-3864-4900-af03-5095e0ae6fd2"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"a9df54ea-d36a-45dd-ac0f-1a151da7786b"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union]Insert)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[Union]Insert)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1429638428", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1429638428", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:02, 0.245241 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1429638428-1", "3"},
 CellLabel->"In[366]:=",
 CellID->788463628,ExpressionUUID->"e1b4be1f-45b6-489f-ae2b-a23d03f34073"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"c25d68f7-bcda-4266-b457-b21fa5843c57"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union]\[Intersection] dist left)\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(\[Union]\[Intersection] dist left)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1116053495", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1116053495", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:02, 0.356722 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1116053495-1", "3"},
 CellLabel->"In[367]:=",
 CellID->826786679,ExpressionUUID->"10ae46d4-043b-4352-a8ed-04644b597001"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"75fb54a4-02c7-462c-9edf-74de0b5bbecb"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union]\[Intersection] dist right)\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(\[Union]\[Intersection] dist right)\"", "FormReferenceHover",
               StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1097000890", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1097000890", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:03, 0.333737 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1097000890-1", "3"},
 CellLabel->"In[368]:=",
 CellID->1930524663,ExpressionUUID->"5e24b08f-bb6e-4cbf-80ba-ba0b65e7b37f"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"f81a9a68-0c63-4653-96bd-aa046d4ab5a3"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union]\[Equal]{})\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Union]\[Equal]{})\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1016692823", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1016692823", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:04, 0.577148 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1016692823-1", "3"},
 CellLabel->"In[369]:=",
 CellID->1014581777,ExpressionUUID->"4e864b7c-43bd-41bd-b35b-e80644e4ba7c"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"f6ca9abd-ac71-4e55-9122-5b54956269aa"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union]\[SubsetEqual])\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Union]\[SubsetEqual])\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["997515848", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:997515848", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:04, 0.171414 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:997515848-1", "3"},
 CellLabel->"In[370]:=",
 CellID->273440836,ExpressionUUID->"5bcae50a-1782-48b2-bce6-0bda8d2274c8"],

Cell[BoxData[
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", "\"\<\[Union] absorb\>\"", " ", 
  "\"\<\[Union]{}\>\"", " ", "\"\<{}\[Union]\>\"", " ", 
  "\"\<\[Union]UNIV\>\"", " ", "\"\<UNIV\[Union]\>\"", " ", 
  "\"\<Insert\[Union]\>\"", " ", "\"\<\[Union]Insert\>\"", " ", 
  "\"\<\[Union]\[Equal]{}\>\"", " ", 
  "\"\<\[Union]\[SubsetEqual]\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.765185894245191*^9, 3.7651859235328712`*^9}, {
  3.7651862370094147`*^9, 3.765186248489472*^9}, {3.7651864291262703`*^9, 
  3.765186435871029*^9}, {3.765186531333171*^9, 3.76518653484554*^9}, {
  3.7651865809007*^9, 3.7651865839650173`*^9}},
 CellLabel->"In[371]:=",
 CellID->631533417,ExpressionUUID->"c4aaa5c8-a5cb-4e9a-b9c5-af3f0a10c650"],

Cell["", "OpenEnvironment",
 CellID->1347919889,ExpressionUUID->"33c844af-b832-438b-bdfb-0ed6b8d72678"],

Cell[CellGroupData[{

Cell["Lemma (Simplification Rules for Complement)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.7651891362215347`*^9, 3.765189144812894*^9}},
 CellID->2078327963,ExpressionUUID->"4ba8a9ba-0141-45de-abc8-2e3ff7352704"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[Intersection]", 
     RowBox[{"-", "A"}]}], "\[Equal]", 
    RowBox[{"{", "}"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(disjoint complement)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651891529891768`*^9, 3.765189165196509*^9}},
 CellTags->"disjoint complement",
 CellLabel->"In[372]:=",
 CellID->36426955,ExpressionUUID->"14526c5d-7e1a-491c-a18d-b2adee8c4fd1"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"-", "A"}], "\[Intersection]", "A"}], "\[Equal]", 
    RowBox[{"{", "}"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(complement disjoint)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765189207823638*^9, 3.765189210684174*^9}},
 CellTags->"complement disjoint",
 CellLabel->"In[373]:=",
 CellID->1941170909,ExpressionUUID->"0f6a4806-b037-4cb5-86be-f9e02d910c80"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[Union]", 
     RowBox[{"-", "A"}]}], "\[Equal]", "UNIV"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(partition complement)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765189259551177*^9, 3.765189266108029*^9}},
 CellTags->"partition complement",
 CellLabel->"In[374]:=",
 CellID->980585996,ExpressionUUID->"cba2301e-f964-4dfa-9ffc-d9fc085df72c"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"-", "A"}], "\[Union]", "A"}], "\[Equal]", "UNIV"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(complement partition)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765189252985173*^9, 3.765189272315453*^9}},
 CellTags->"complement partition",
 CellLabel->"In[375]:=",
 CellID->1324045296,ExpressionUUID->"26e2dded-cd7a-432f-9c23-b65e3a2865a1"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", "::", 
    RowBox[{"{", "\[Alpha]_", "}"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"-", 
     RowBox[{"(", 
      RowBox[{"-", "A"}], ")"}]}], "\[Equal]", "A"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(complement complement)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765189315458901*^9, 3.765189330931262*^9}},
 CellTags->"complement complement",
 CellLabel->"In[376]:=",
 CellID->1636712813,ExpressionUUID->"d99264a8-25ec-40e3-a87a-33a8d7399ef1"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"Minus", "[", 
     RowBox[{"A", "\[Union]", "B"}], "]"}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "A"}], "\[Intersection]", 
     RowBox[{"-", "B"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(complement \[Union])", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765189359794273*^9, 3.765189373482491*^9}, {
  3.765190157419442*^9, 3.765190160736587*^9}},
 CellTags->"complement \[Union]",
 CellLabel->"In[377]:=",
 CellID->1334137643,ExpressionUUID->"b414a419-93c9-4969-bb70-aa9e72fb3466"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"Minus", "[", 
     RowBox[{"A", "\[Intersection]", "B"}], "]"}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "A"}], "\[Union]", 
     RowBox[{"-", "B"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(complement \[Intersection])", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.76518940328664*^9, 3.765189407473631*^9}, {
  3.765190165755007*^9, 3.765190169048582*^9}},
 CellTags->"complement \[Intersection]",
 CellLabel->"In[378]:=",
 CellID->1036693907,ExpressionUUID->"b67ff51c-86bc-4021-948a-f367e9f183e7"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "A"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"-", 
     RowBox[{"Insert", "[", 
      RowBox[{"a", ",", "A"}], "]"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "A"}], "-", 
     RowBox[{"{", "a", "}"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(complement Insert)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765189692174295*^9, 3.7651897141663*^9}, 
   3.765190351047364*^9},
 CellTags->"complement Insert",
 CellLabel->"In[379]:=",
 CellID->204929684,ExpressionUUID->"2ca9d86e-03aa-4873-a4f5-ea04b00662b0"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", "\[SubsetEqual]", 
      RowBox[{"-", "A"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", "\[Equal]", 
      RowBox[{"{", "}"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[SubsetEqual]- self)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765189448481133*^9, 3.7651894624336777`*^9}},
 CellTags->"\[SubsetEqual]- self",
 CellLabel->"In[380]:=",
 CellID->849790927,ExpressionUUID->"a2f250c1-a6f8-48b1-917d-4f4472a06ad4"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", "UNIV"}], "\[Equal]", 
  RowBox[{"{", "}"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(complement UNIV)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651895222730303`*^9, 3.765189525928699*^9}},
 CellTags->"complement UNIV",
 CellLabel->"In[381]:=",
 CellID->1609622809,ExpressionUUID->"521382bc-26ea-430e-92b8-33377f2e1ed9"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   RowBox[{"{", "}"}]}], "\[Equal]", "UNIV"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(complement {})", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765189543760069*^9, 3.765189547176091*^9}},
 CellTags->"complement {}",
 CellLabel->"In[382]:=",
 CellID->1579837433,ExpressionUUID->"d6188e01-f675-4a8a-b808-d55954fe886e"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"-", "A"}], "\[SubsetEqual]", 
      RowBox[{"-", "B"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"B", "\[SubsetEqual]", "A"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(complement \[SubsetEqual] complement)", "FrameLabel"], 
         Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765189577615748*^9, 3.765189594783944*^9}},
 CellTags->"complement \[SubsetEqual] complement",
 CellLabel->"In[383]:=",
 CellID->2039310054,ExpressionUUID->"d6e1db89-6011-4f20-b209-e333d98038c8"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", 
    RowBox[{"B", "::", 
     RowBox[{"{", "\[Alpha]_", "}"}]}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"-", "A"}], "\[Equal]", 
      RowBox[{"-", "B"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"B", "\[Equal]", "A"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(complement \[Equal] complement)", "FrameLabel"], 
         Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765189623329789*^9, 3.765189628151423*^9}, {
  3.765191345339937*^9, 3.765191348842288*^9}},
 CellTags->"complement \[Equal] complement",
 CellLabel->"In[384]:=",
 CellID->1867561751,ExpressionUUID->"ba85e675-30b3-4fb0-b8c4-fd034abf0994"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2026674483,ExpressionUUID->"74c9067f-78fa-4dbb-9151-be9dc9d5ecee"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->263470190,ExpressionUUID->"95396ec1-e8c6-46ac-924c-a6db80e2509f"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"ab6eea9b-fcd2-4dc5-b1b7-2421c90e1053"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(disjoint complement)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(disjoint complement)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["36426955", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:36426955", CellTags]], Appearance -> 
          None, Evaluator -> Automatic, Method -> "Preemptive"]]], 
       " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:05, 0.171435 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:36426955-1", "3"},
 CellLabel->"In[385]:=",
 CellID->1087560852,ExpressionUUID->"335661a6-db39-426e-a2d5-3a88bbf119ee"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"f81e7327-56ac-4856-8804-78a802e55ae9"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(complement disjoint)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(complement disjoint)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1941170909", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1941170909", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:05, 0.128891 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1941170909-1", "3"},
 CellLabel->"In[386]:=",
 CellID->769645528,ExpressionUUID->"790c583f-b1fd-4cae-b4c7-3b7f8318ee81"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"836f4205-f318-4676-9410-ea63051f65c3"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(partition complement)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(partition complement)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["980585996", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:980585996", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:06, 0.142985 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:980585996-1", "3"},
 CellLabel->"In[387]:=",
 CellID->1868666534,ExpressionUUID->"978d16d3-2a8d-48c1-8119-e556b594fdbe"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"9c6b8dde-8046-469c-8b37-f8b19d734aaa"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(complement partition)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(complement partition)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1324045296", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1324045296", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:06, 0.147058 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1324045296-1", "3"},
 CellLabel->"In[388]:=",
 CellID->1667315755,ExpressionUUID->"55ba8ff4-e771-4565-ad2a-a535645579da"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"f3d269b9-450f-4c4c-99c2-10fcef920cd3"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(complement complement)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(complement complement)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1636712813", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1636712813", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:06, 0.116342 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1636712813-1", "3"},
 CellLabel->"In[389]:=",
 CellID->193380618,ExpressionUUID->"4270ef55-e968-4e30-88f5-189f15367a47"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"2f7b0412-de55-4d69-81f7-b81f6ff0e3f7"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(complement \[Union])\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(complement \[Union])\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1334137643", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1334137643", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:07, 0.228329 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1334137643-1", "3"},
 CellLabel->"In[390]:=",
 CellID->95672954,ExpressionUUID->"ff8108f1-b1f1-4f1c-be7e-36f13ea484f5"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"48cbc849-16d2-4cce-a389-f46a569d336f"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(complement \[Intersection])\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(complement \[Intersection])\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1036693907", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1036693907", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:07, 0.233987 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1036693907-1", "3"},
 CellLabel->"In[391]:=",
 CellID->197559568,ExpressionUUID->"6c8fd24a-5787-423d-a5a0-9ee8249e79b1"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"2c01ac37-bf7b-4096-9574-68331007d6c2"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(complement Insert)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(complement Insert)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["204929684", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:204929684", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:07, 0.447025 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:204929684-1", "3"},
 CellLabel->"In[392]:=",
 CellID->459265611,ExpressionUUID->"c77a6221-7b61-4579-83d1-c9ad096f840c"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"f6d04030-6c4c-4264-8a22-20c7decb14f2"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[SubsetEqual]- self)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[SubsetEqual]- self)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["849790927", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:849790927", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:08, 0.112992 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:849790927-1", "3"},
 CellLabel->"In[393]:=",
 CellID->1062590287,ExpressionUUID->"fcbb79cd-640b-433d-aca6-6f01bc0cd323"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"d49525b7-f546-408e-a352-b1e5ef439f91"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(complement UNIV)\"", "FormReference", StripOnInput -> False],
             True -> StyleBox[
             "\"(complement UNIV)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1609622809", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1609622809", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:08, 0.141140 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1609622809-1", "3"},
 CellLabel->"In[394]:=",
 CellID->1841417850,ExpressionUUID->"03e8795e-cd22-4b6b-b60f-6a5c21b1baf5"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"b33f421b-e129-4869-8ff9-145c6720a946"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(complement {})\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(complement {})\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1579837433", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1579837433", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:08, 0.136639 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1579837433-1", "3"},
 CellLabel->"In[395]:=",
 CellID->1890841345,ExpressionUUID->"fc5be3c0-cfbd-484d-840c-b987b55aab11"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"270a799e-cbf1-4d89-a96d-8b78a21ca268"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(complement \[SubsetEqual] complement)\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(complement \[SubsetEqual] complement)\"", 
              "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["2039310054", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:2039310054", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:09, 0.326219 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:2039310054-1", "3"},
 CellLabel->"In[396]:=",
 CellID->1409452885,ExpressionUUID->"070c9e7a-2ffe-4eed-a409-8fe374d0f970"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"bc0ab5aa-36d0-4d5a-ad39-7ff8851eb0d8"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(complement \[Equal] complement)\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(complement \[Equal] complement)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1867561751", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1867561751", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:09, 0.153396 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1867561751-1", "3"},
 CellLabel->"In[397]:=",
 CellID->1823633609,ExpressionUUID->"3255cca4-1d25-482e-bef3-695eb3e5fa77"],

Cell[BoxData[{
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", "\"\<disjoint complement\>\"", " ", 
  "\"\<complement disjoint\>\"", " ", "\"\<partition complement\>\"", " ", 
  "\"\<complement partition\>\"", " ", 
  "\"\<complement complement\>\""}], "\n", 
 RowBox[{"\t", 
  RowBox[{
  "\"\<complement \[Union]\>\"", " ", "\"\<complement \[Intersection]\>\"", 
   " ", "\"\<complement UNIV\>\"", " ", "\"\<complement {}\>\"", " ", 
   "\"\<complement \[SubsetEqual] complement\>\"", " ", 
   "\"\<complement \[Equal] complement\>\""}]}]}], "TheoremaCode",
 CellChangeTimes->{{3.7651892282839727`*^9, 3.765189240051874*^9}, {
  3.765189289459175*^9, 3.765189301531516*^9}, {3.7651893431595993`*^9, 
  3.765189347146743*^9}, {3.765189392137981*^9, 3.7651894215617323`*^9}, {
  3.765189536416259*^9, 3.765189562192239*^9}, {3.765189603711293*^9, 
  3.765189606567621*^9}, {3.765189636814789*^9, 3.765189671190587*^9}},
 CellLabel->"In[398]:=",
 CellID->1273666900,ExpressionUUID->"c86e651e-48af-4cbc-9361-9d33d054749b"],

Cell["", "OpenEnvironment",
 CellID->1686577370,ExpressionUUID->"d002514c-7838-4dd5-8a4f-d36656ebd8b4"],

Cell[CellGroupData[{

Cell["Lemma (Simplification Rules for -)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.765086606725917*^9, 3.76508661526989*^9}},
 CellID->85132320,ExpressionUUID->"3e915c3f-6136-4bc8-b8eb-74392ebc93b9"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "-", "B"}], "\[Equal]", 
    RowBox[{"A", "\[Intersection]", 
     RowBox[{"Minus", "[", "B", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(- alt SET 2)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.76519194318637*^9, 3.765191955075014*^9}},
 CellTags->"- alt SET 2",
 CellLabel->"In[399]:=",
 CellID->889927681,ExpressionUUID->"6623863d-220c-43b7-b56a-d97b10f1a117"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"A", "-", "B"}], "\[Equal]", 
      RowBox[{"{", "}"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", "\[SubsetEqual]", "B"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(-\[Equal]{} iff)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765084829724365*^9, 3.765084856724058*^9}},
 CellTags->"-\[Equal]{} iff",
 CellLabel->"In[400]:=",
 CellID->1181933579,ExpressionUUID->"60d9d2ad-a5de-4e53-8980-466c84af87c2"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "-", "A"}], "\[Equal]", 
    RowBox[{"{", "}"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(- self SET)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651920001934958`*^9, 3.765192008946186*^9}},
 CellTags->"- self SET",
 CellLabel->"In[401]:=",
 CellID->1313006718,ExpressionUUID->"6781d634-3dc0-42fb-8e42-aff33a4f6cec"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", 
    RowBox[{"B", "::", 
     RowBox[{"{", "\[Alpha]_", "}"}]}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"A", "-", "B"}], ")"}], "-", "B"}], "\[Equal]", 
    RowBox[{"A", "-", "B"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(- idem SET)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765192046201532*^9, 3.765192057569693*^9}, {
  3.76526470314293*^9, 3.765264705684794*^9}},
 CellTags->"- idem SET",
 CellLabel->"In[402]:=",
 CellID->405798501,ExpressionUUID->"e34deac0-2ee0-4195-b922-322864967d47"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"A", "\[Intersection]", "B"}], "\[Equal]", 
     RowBox[{"{", "}"}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"A", "-", "B"}], "\[Equal]", "A"}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(- triv SET)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765192082504879*^9, 3.765192109136672*^9}},
 CellTags->"- triv SET",
 CellLabel->"In[403]:=",
 CellID->142114837,ExpressionUUID->"ed752dac-cb8f-4130-b775-aa2d387899a4"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "-", 
     RowBox[{"{", "}"}]}], "\[Equal]", "A"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(-{})", "FrameLabel"], Evaluator -> Automatic, Appearance -> 
         None, ButtonFunction :> Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765192136728239*^9, 3.765192143848362*^9}},
 CellTags->"-{}",
 CellLabel->"In[404]:=",
 CellID->655509949,ExpressionUUID->"f814e4e0-d5fd-4f2d-8049-0a0cc5ce82a9"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"{", "}"}], "-", "A"}], "\[Equal]", 
    RowBox[{"{", "}"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["({}-)", "FrameLabel"], Evaluator -> Automatic, Appearance -> 
         None, ButtonFunction :> Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651921587402353`*^9, 3.765192164784378*^9}},
 CellTags->"{}-",
 CellLabel->"In[405]:=",
 CellID->149156198,ExpressionUUID->"b1c87120-9de9-411e-b7fb-51ab4e109a6f"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "-", "UNIV"}], "\[Equal]", 
    RowBox[{"{", "}"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(-UNIV)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765192177944933*^9, 3.765192183264306*^9}},
 CellTags->"-UNIV",
 CellLabel->"In[406]:=",
 CellID->1551567722,ExpressionUUID->"c3b9274b-a806-42e8-a112-b152145efc29"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"Insert", "[", 
      RowBox[{"a", ",", "A"}], "]"}], "-", "B"}], "\[Equal]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Piecewise]", GridBox[{
        {
         RowBox[{"A", "-", "B"}], "\[DoubleLeftArrow]", 
         RowBox[{"a", "\[Element]", "B"}]},
        {
         RowBox[{"Insert", "[", 
          RowBox[{"a", ",", 
           RowBox[{"A", "-", "B"}]}], "]"}], "\[DoubleLeftArrow]", "True"}
       }]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Insert-)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651923277261057`*^9, 3.76519235957423*^9}},
 CellTags->"Insert-",
 CellLabel->"In[407]:=",
 CellID->1618171348,ExpressionUUID->"42e35235-185b-4a52-99ed-8e7cf0803c40"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"a", ",", "A"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"Insert", "[", 
     RowBox[{"a", ",", 
      RowBox[{"A", "-", 
       RowBox[{"{", "a", "}"}]}]}], "]"}], "\[Equal]", 
    RowBox[{"Insert", "[", 
     RowBox[{"a", ",", "A"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Insert- single)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651923892053823`*^9, 3.765192403917317*^9}},
 CellTags->"Insert- single",
 CellLabel->"In[408]:=",
 CellID->263157914,ExpressionUUID->"a4ab1dc9-a2bf-463a-8074-9e4f2ca3cbf3"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[Intersection]", 
     RowBox[{"(", 
      RowBox[{"B", "-", "A"}], ")"}]}], "\[Equal]", 
    RowBox[{"{", "}"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(- disjoint)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651924414364567`*^9, 3.765192453884831*^9}},
 CellTags->"- disjoint",
 CellLabel->"In[409]:=",
 CellID->586501678,ExpressionUUID->"50650e03-a3f8-46d0-b2ff-9a8d0cd54ee2"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "\[Union]", 
     RowBox[{"(", 
      RowBox[{"B", "-", "A"}], ")"}]}], "\[Equal]", 
    RowBox[{"A", "\[Union]", "B"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union]- cancel)", "FrameLabel"], Evaluator -> Automatic,
          Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651929156229973`*^9, 3.765192929758738*^9}, {
  3.7651929714782457`*^9, 3.765192972678772*^9}},
 CellTags->"\[Union]- cancel",
 CellLabel->"In[410]:=",
 CellID->1211775432,ExpressionUUID->"bffe28f1-05a1-4afd-a4bd-cc179788a0cf"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"B", "-", "A"}], ")"}], "\[Union]", "A"}], "\[Equal]", 
    RowBox[{"B", "\[Union]", "A"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(-\[Union] cancel)", "FrameLabel"], Evaluator -> Automatic,
          Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.76519295646084*^9, 3.7651929760625753`*^9}},
 CellTags->"-\[Union] cancel",
 CellLabel->"In[411]:=",
 CellID->1437446913,ExpressionUUID->"4b9cee39-401e-4340-85f0-5cc10f31ca56"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "-", 
     RowBox[{"(", 
      RowBox[{"B", "\[Union]", "C"}], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"A", "-", "B"}], ")"}], "\[Intersection]", 
     RowBox[{"(", 
      RowBox[{"A", "-", "C"}], ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(-\[Union])", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765193014742264*^9, 3.765193046029872*^9}},
 CellTags->"-\[Union]",
 CellLabel->"In[412]:=",
 CellID->1743954910,ExpressionUUID->"1364d8ff-9078-426a-83d8-d37785793f2d"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "-", 
     RowBox[{"(", 
      RowBox[{"B", "\[Intersection]", "C"}], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"A", "-", "B"}], ")"}], "\[Union]", 
     RowBox[{"(", 
      RowBox[{"A", "-", "C"}], ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(-\[Intersection])", "FrameLabel"], Evaluator -> Automatic,
          Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765193060162784*^9, 3.7651930650135183`*^9}},
 CellTags->"-\[Intersection]",
 CellLabel->"In[413]:=",
 CellID->732837784,ExpressionUUID->"14da9918-2877-431c-9676-774a594407e6"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "-", 
     RowBox[{"(", 
      RowBox[{"A", "-", "B"}], ")"}]}], "\[Equal]", 
    RowBox[{"A", "\[Intersection]", "B"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(-- \[Equal] \[Intersection])", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651931017090883`*^9, 3.765193120196815*^9}},
 CellTags->"-- \[Equal] \[Intersection]",
 CellLabel->"In[414]:=",
 CellID->520649721,ExpressionUUID->"3ac7b94b-778f-462d-a022-df58ba35c477"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"A", "\[Union]", "B"}], ")"}], "-", "C"}], "\[Equal]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"A", "-", "C"}], ")"}], "\[Union]", 
     RowBox[{"(", 
      RowBox[{"B", "-", "C"}], ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union]-)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765266145003284*^9, 3.7652661914104757`*^9}},
 CellTags->"\[Union]-",
 CellLabel->"In[415]:=",
 CellID->1959934043,ExpressionUUID->"19d77f44-c8da-4204-b8b0-44e14aee704c"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"A", "\[Intersection]", "B"}], ")"}], "-", "C"}], "\[Equal]", 
    RowBox[{"A", "\[Intersection]", 
     RowBox[{"(", 
      RowBox[{"B", "-", "C"}], ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection]-)", "FrameLabel"], Evaluator -> Automatic,
          Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765266217749906*^9, 3.765266234057909*^9}},
 CellTags->"\[Intersection]-",
 CellLabel->"In[416]:=",
 CellID->1023545665,ExpressionUUID->"a1bb30b1-8de9-47b4-8158-a68f8ab50910"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"C", "\[Intersection]", 
     RowBox[{"(", 
      RowBox[{"A", "-", "B"}], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"C", "\[Intersection]", "A"}], ")"}], "-", 
     RowBox[{"(", 
      RowBox[{"C", "\[Intersection]", "B"}], ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection]- dist left)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7652662478016644`*^9, 3.765266275914125*^9}},
 CellTags->"\[Intersection]- dist left",
 CellLabel->"In[417]:=",
 CellID->1036748287,ExpressionUUID->"83e3520e-f1ff-4eac-a5ba-79ed51ea4e65"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"A", "-", "B"}], ")"}], "\[Intersection]", "C"}], "\[Equal]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"A", "\[Intersection]", "C"}], ")"}], "-", 
     RowBox[{"(", 
      RowBox[{"B", "\[Intersection]", "C"}], ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection]- dist right)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765266327236949*^9, 3.765266341400494*^9}},
 CellTags->"\[Intersection]- dist right",
 CellLabel->"In[418]:=",
 CellID->654257124,ExpressionUUID->"9c74b7bb-8043-4f3f-adef-5cedca906c9b"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"A", "-", 
     RowBox[{"(", 
      RowBox[{"-", "B"}], ")"}]}], "\[Equal]", 
    RowBox[{"A", "\[Intersection]", "B"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(- complement)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765266366456448*^9, 3.765266377280231*^9}},
 CellTags->"- complement",
 CellLabel->"In[419]:=",
 CellID->788943865,ExpressionUUID->"ac1ce070-9cd7-4b47-9574-487965c33172"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"-", 
     RowBox[{"(", 
      RowBox[{"A", "-", "B"}], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "A"}], "\[Union]", "B"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(complement -)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765266403407852*^9, 3.765266419455702*^9}},
 CellTags->"complement -",
 CellLabel->"In[420]:=",
 CellID->811668796,ExpressionUUID->"f380d7e6-5d16-4293-becd-33ad9bfa3087"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1477814126,ExpressionUUID->"51f15a8e-2618-4e52-a726-990229061e99"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1940925515,ExpressionUUID->"693560ca-d387-49be-b62c-5b7317c69a34"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"320b709a-8d37-4639-9089-1e5ad680b4c0"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(- alt SET 2)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(- alt SET 2)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["889927681", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:889927681", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:11, 0.157938 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:889927681-1", "3"},
 CellLabel->"In[421]:=",
 CellID->989686867,ExpressionUUID->"41412e05-3d1b-47af-9a2c-69b7da569934"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"f8949e5d-70ba-4340-82cd-4ac617791e5a"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(-\[Equal]{} iff)\"", "FormReference", StripOnInput -> False],
             True -> StyleBox[
             "\"(-\[Equal]{} iff)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1181933579", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1181933579", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:11, 0.302931 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1181933579-1", "3"},
 CellLabel->"In[422]:=",
 CellID->14631120,ExpressionUUID->"e718483a-9533-47fb-b2b3-21662a495fe5"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"d8b1e3ed-7b79-408b-8dde-673ab8ad83fb"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(- self SET)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(- self SET)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1313006718", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1313006718", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:11, 0.113160 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1313006718-1", "3"},
 CellLabel->"In[423]:=",
 CellID->1424176457,ExpressionUUID->"c083b88a-b5b4-4213-8c12-363f6380f03a"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"f51eb634-f49e-40d2-b2c2-2055550dcec7"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(- idem SET)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(- idem SET)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["405798501", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:405798501", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:12, 0.233013 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:405798501-1", "3"},
 CellLabel->"In[424]:=",
 CellID->1289581872,ExpressionUUID->"8dfe2f53-ea61-4e3d-9764-e5a54581d749"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"225cf157-f793-4456-9e6f-7cc789505233"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(- triv SET)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(- triv SET)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["142114837", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:142114837", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:12, 0.194301 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:142114837-1", "3"},
 CellLabel->"In[425]:=",
 CellID->1757475874,ExpressionUUID->"15000b41-243b-47fb-9432-88147364a061"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"ae09ab58-1ce2-43d9-ad13-c9ae9da93b77"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(-{})\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(-{})\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["655509949", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:655509949", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:12, 0.152730 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:655509949-1", "3"},
 CellLabel->"In[426]:=",
 CellID->939658652,ExpressionUUID->"ba81f3be-43a1-4405-a13e-20392ccc153a"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"c4b33c5b-8181-44ed-8dc8-07400ce5dcb8"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"({}-)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"({}-)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["149156198", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:149156198", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:12, 0.134711 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:149156198-1", "3"},
 CellLabel->"In[427]:=",
 CellID->125409990,ExpressionUUID->"a5e8bedb-4a86-49a2-916f-d64eed66bbf2"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"b0501457-72bd-45a3-8248-dd334c9d40b8"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(-UNIV)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(-UNIV)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1551567722", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1551567722", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:13, 0.074636 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1551567722-1", "3"},
 CellLabel->"In[428]:=",
 CellID->128243725,ExpressionUUID->"81eeefa4-874f-43cf-8a1a-bf67eab02b68"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"5e3d3889-7968-4d8d-8eb5-a84e5228fa1b"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Insert-)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(Insert-)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1618171348", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1618171348", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:14, 0.840703 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1618171348-1", "3"},
 CellLabel->"In[429]:=",
 CellID->1212022066,ExpressionUUID->"27fb32d0-54c1-4430-9891-7032e6c9d015"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"d3b19a3a-2f61-42da-9d5f-1fab1f7913f1"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Insert- single)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(Insert- single)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["263157914", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:263157914", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:14, 0.525809 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:263157914-1", "3"},
 CellLabel->"In[430]:=",
 CellID->1307442480,ExpressionUUID->"0d2dd380-ba5d-4928-b373-142c2857c2c0"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"4ff3e436-9303-4ce8-84e5-1630a5a4602e"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(- disjoint)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(- disjoint)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["586501678", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:586501678", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:15, 0.430107 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:586501678-1", "3"},
 CellLabel->"In[431]:=",
 CellID->1257980868,ExpressionUUID->"1310c805-4a87-4c5e-91d3-1155f73282e3"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"139f51ab-b034-4ebf-92f6-4fd755a287a7"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union]- cancel)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Union]- cancel)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1211775432", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1211775432", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:15, 0.161077 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1211775432-1", "3"},
 CellLabel->"In[432]:=",
 CellID->1583175827,ExpressionUUID->"f7d7c577-0e4f-43f2-aab9-f884ae948b4c"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"becec79e-9147-4bed-8519-1a49c88a5383"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(-\[Union] cancel)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(-\[Union] cancel)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1437446913", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1437446913", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:16, 0.162637 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1437446913-1", "3"},
 CellLabel->"In[433]:=",
 CellID->2050049889,ExpressionUUID->"06df1467-a1fc-4be5-b103-26e4a9577222"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"9ff05e7c-d05c-4926-b793-3a7f1e6517f8"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(-\[Union])\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(-\[Union])\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1743954910", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1743954910", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:16, 0.611745 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1743954910-1", "3"},
 CellLabel->"In[434]:=",
 CellID->1362760740,ExpressionUUID->"9ecc279b-a1b3-4c56-a8fa-e29147a8ea06"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"7860967c-1e49-4208-9d69-c4329d5eaa01"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(-\[Intersection])\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(-\[Intersection])\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["732837784", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:732837784", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:17, 0.370733 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:732837784-1", "3"},
 CellLabel->"In[435]:=",
 CellID->1924200546,ExpressionUUID->"45cfb94c-9228-4f72-a7c6-3adc38573e3f"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"ed4c3358-1d5f-4102-8604-16c38929177c"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(-- \[Equal] \[Intersection])\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(-- \[Equal] \[Intersection])\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["520649721", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:520649721", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:17, 0.354427 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:520649721-1", "3"},
 CellLabel->"In[436]:=",
 CellID->1370228285,ExpressionUUID->"d7f0bdda-8a9a-44b8-8d62-f6a17e32fa14"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"d20835e0-9904-4a89-9a4c-7f2e01f6adef"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union]-)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(\[Union]-)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1959934043", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1959934043", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:18, 0.208070 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1959934043-1", "3"},
 CellLabel->"In[437]:=",
 CellID->1728602517,ExpressionUUID->"dbca4d7d-fc2b-4f4e-9403-9ed1fcdfcd97"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"438e6f90-fa28-404e-aad0-8044cab76e60"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection]-)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Intersection]-)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1023545665", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1023545665", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:18, 0.128802 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1023545665-1", "3"},
 CellLabel->"In[438]:=",
 CellID->124203596,ExpressionUUID->"330a0d43-a3eb-4628-9ac8-18c5d5c04d74"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"4a3e1cab-5a8b-46cd-8627-f10007176f77"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection]- dist left)\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(\[Intersection]- dist left)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1036748287", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1036748287", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:19, 0.856206 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1036748287-1", "3"},
 CellLabel->"In[439]:=",
 CellID->1046065391,ExpressionUUID->"e55012c4-27e1-41ae-920d-8392fb2fc039"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"b8ee3f6f-a2ab-4ea6-ad7d-b3b774aba64c"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection]- dist right)\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(\[Intersection]- dist right)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["654257124", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:654257124", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:19, 0.183813 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:654257124-1", "3"},
 CellLabel->"In[440]:=",
 CellID->190015761,ExpressionUUID->"75c51d93-2b93-476d-8df7-56ee4ba944f9"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"4b567daf-f270-4681-90e5-713acfed65f3"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(- complement)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(- complement)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["788943865", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:788943865", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:20, 0.213565 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:788943865-1", "3"},
 CellLabel->"In[441]:=",
 CellID->1182421826,ExpressionUUID->"bfb4263d-1542-4461-b96b-3cb80e17ef47"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"ca9f9f30-9aec-49b9-b94c-2ebfc2ed82c7"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(complement -)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(complement -)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["811668796", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:811668796", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:20, 0.211406 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:811668796-1", "3"},
 CellLabel->"In[442]:=",
 CellID->1010368971,ExpressionUUID->"7126e94e-5680-4178-a002-f91132148a3e"],

Cell[BoxData[{
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", "\"\<-\[Equal]{} iff\>\"", " ", 
  "\"\<- self SET\>\"", " ", "\"\<- idem SET\>\"", " ", "\"\<-{}\>\"", " ", 
  "\"\<{}-\>\"", " ", "\"\<-UNIV\>\"", " ", "\"\<Insert- single\>\"", " ", 
  "\"\<- disjoint\>\"", " ", "\"\<\[Union]- cancel\>\"", " ", 
  "\"\<-\[Union] cancel\>\""}], "\n", 
 RowBox[{"\t", 
  RowBox[{
  "\"\<- complement\>\"", " ", "\"\<complement -\>\""}]}]}], "TheoremaCode",
 CellChangeTimes->{{3.76519196908232*^9, 3.765191987434304*^9}, {
   3.765192028705634*^9, 3.765192031633821*^9}, {3.765192066009153*^9, 
   3.765192069065486*^9}, {3.7651921522562237`*^9, 3.765192192343857*^9}, {
   3.765192412452772*^9, 3.765192416029148*^9}, {3.765192463972396*^9, 
   3.7651924860840273`*^9}, {3.765192937990459*^9, 3.7651929411671543`*^9}, {
   3.765192986357958*^9, 3.765192989606606*^9}, {3.765193141499928*^9, 
   3.7651931584602213`*^9}, {3.765266386271846*^9, 3.765266390383877*^9}, {
   3.765266429655456*^9, 3.765266432791431*^9}, {3.765266492286653*^9, 
   3.7652664965986443`*^9}, 3.765267494027377*^9},
 CellLabel->"In[443]:=",
 CellID->673690670,ExpressionUUID->"7e78a0f2-fb38-42a7-95d1-866199b8bd64"],

Cell["", "OpenEnvironment",
 CellID->1252732680,ExpressionUUID->"ac145a3d-05f8-4b42-b9f9-1c1e2108743f"],

Cell[CellGroupData[{

Cell["Lemma (Simplification Rules for Pow)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.765267505306003*^9, 3.765267511825913*^9}},
 CellID->1075678771,ExpressionUUID->"176ffabc-7757-47f4-8501-78507d8a2ca1"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Pow", "[", 
   RowBox[{"{", "}"}], "]"}], "\[Equal]", 
  RowBox[{"{", 
   RowBox[{"{", "}"}], "}"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Pow {})", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765267538353818*^9, 3.765267546665798*^9}},
 CellTags->"Pow {}",
 CellLabel->"In[444]:=",
 CellID->525687623,ExpressionUUID->"5be69f99-7725-46d4-bb6b-dbc97f3b3ab6"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Pow", "[", "UNIV", "]"}], "\[Equal]", 
  "UNIV"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Pow UNIV)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765267698183619*^9, 3.765267703063799*^9}},
 CellTags->"Pow UNIV",
 CellLabel->"In[445]:=",
 CellID->537376125,ExpressionUUID->"94af11de-54d0-4a76-a0f9-9d7f0fa4a28f"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"Pow", "[", "A", "]"}], "\[Union]", 
     RowBox[{"Pow", "[", "B", "]"}]}], "\[SubsetEqual]", 
    RowBox[{"Pow", "[", 
     RowBox[{"A", "\[Union]", "B"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union]Pow \[SubsetEqual])", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765267728767654*^9, 3.765267749543336*^9}},
 CellTags->"\[Union]Pow \[SubsetEqual]",
 CellLabel->"In[446]:=",
 CellID->1322774384,ExpressionUUID->"fd9a1629-3d23-485d-a17d-020a9e92a3d4"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"Pow", "[", 
     RowBox[{"A", "\[Intersection]", "B"}], "]"}], "\[Equal]", 
    RowBox[{
     RowBox[{"Pow", "[", "A", "]"}], "\[Intersection]", 
     RowBox[{"Pow", "[", "B", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Pow\[Intersection])", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765267772447011*^9, 3.765267787350893*^9}},
 CellTags->"Pow\[Intersection]",
 CellLabel->"In[447]:=",
 CellID->552587340,ExpressionUUID->"334b0879-21a1-493b-a548-ff7fa6aad20a"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1029449578,ExpressionUUID->"d90dd32d-9143-4137-8c32-c73557871bc7"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2047020259,ExpressionUUID->"9af747af-3746-4b40-af63-c3cbec04840d"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"66b35079-d5f9-4211-8896-ee1c858be462"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Pow {})\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(Pow {})\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["525687623", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:525687623", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:21, 0.195863 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:525687623-1", "3"},
 CellLabel->"In[448]:=",
 CellID->118494995,ExpressionUUID->"60c20774-a0c4-41e0-8583-3cd982a256e3"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"fedfdef0-4b96-47bd-b51e-cb7c6160be83"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Pow UNIV)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(Pow UNIV)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["537376125", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:537376125", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:22, 0.160707 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:537376125-1", "3"},
 CellLabel->"In[449]:=",
 CellID->998603711,ExpressionUUID->"e80da33b-c1dd-4198-bc6a-2a3ca4d75937"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"80a591ae-84ae-49ec-993d-4e771e1456c9"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union]Pow \[SubsetEqual])\"", "FormReference", 
              StripOnInput -> False], True -> 
            StyleBox[
             "\"(\[Union]Pow \[SubsetEqual])\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1322774384", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1322774384", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:22, 0.220822 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1322774384-1", "3"},
 CellLabel->"In[450]:=",
 CellID->1961996116,ExpressionUUID->"1cdbcc0c-c7a2-4a23-ad73-491bc0d57579"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"b8f5e7dd-76d5-4768-9138-1d2b2dec7265"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Pow\[Intersection])\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(Pow\[Intersection])\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["552587340", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:552587340", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:22, 0.210907 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:552587340-1", "3"},
 CellLabel->"In[451]:=",
 CellID->1186950056,ExpressionUUID->"362f630c-4228-4cb1-b302-51e8b7bdf329"],

Cell[BoxData[
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", "\"\<Pow {}\>\"", " ", "\"\<Pow UNIV\>\"", 
  " ", "\"\<Pow\[Intersection]\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.765267516369892*^9, 3.765267518522086*^9}, {
   3.765267558145348*^9, 3.7652675606976023`*^9}, {3.765267632776712*^9, 
   3.7652676362806664`*^9}, {3.7652677116476173`*^9, 3.765267714111672*^9}, {
   3.7652677974865723`*^9, 3.7652678004227877`*^9}, 3.765268008492535*^9},
 CellLabel->"In[452]:=",
 CellID->1895915908,ExpressionUUID->"94ec2d25-14f2-4619-add2-d57a3be773f6"],

Cell["", "OpenEnvironment",
 CellID->1424840277,ExpressionUUID->"a42f4f35-009b-4e43-8cdb-d23c7c2628a3"],

Cell[CellGroupData[{

Cell["Lemma (Simplification Rules for @)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.7650995197213697`*^9, 3.765099526773362*^9}},
 CellID->579114060,ExpressionUUID->"b9fec212-daaa-46f8-a94b-39fa9c10f439"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "A", ",", "B"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"f", "@", 
     RowBox[{"(", 
      RowBox[{"A", "\[Union]", "B"}], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{"f", "@", "A"}], "\[Union]", 
     RowBox[{"f", "@", "B"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(@\[Union])", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765099531941185*^9, 3.765099550932879*^9}},
 CellTags->"@\[Union]",
 CellLabel->"In[453]:=",
 CellID->814265295,ExpressionUUID->"27842332-1915-4319-aa16-d1927fa59d6a"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       UnderscriptBox["\[Lambda]", "x"], "x"}], ")"}], "@", "A"}], "\[Equal]",
     "A"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(@ ident)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7650995871644897`*^9, 3.7650995896207857`*^9}, {
  3.7650999788316097`*^9, 3.7650999931434526`*^9}},
 CellTags->"@ ident",
 CellLabel->"In[454]:=",
 CellID->1283984586,ExpressionUUID->"ebf795ec-f3cd-4609-a17d-968049102958"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "f"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"f", "@", 
     RowBox[{"{", "}"}]}], "\[Equal]", 
    RowBox[{"{", "}"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(@{})", "FrameLabel"], Evaluator -> Automatic, Appearance -> 
         None, ButtonFunction :> Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651000137911367`*^9, 3.765100021935444*^9}},
 CellTags->"@{}",
 CellLabel->"In[455]:=",
 CellID->1257505636,ExpressionUUID->"c6a43110-bc9f-4894-aa96-288c7209b020"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "a", ",", "A"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"f", "@", 
     RowBox[{"Insert", "[", 
      RowBox[{"a", ",", "A"}], "]"}]}], "\[Equal]", 
    RowBox[{"Insert", "[", 
     RowBox[{
      RowBox[{"f", "[", "a", "]"}], ",", 
      RowBox[{"f", "@", "A"}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(@Insert)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765100031574847*^9, 3.765100054079034*^9}},
 CellTags->"@Insert",
 CellLabel->"In[456]:=",
 CellID->565842574,ExpressionUUID->"d15f846d-e95a-4275-90e4-d985073cea8e"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "A"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"f", "@", "A"}], "\[Equal]", 
      RowBox[{"{", "}"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", "\[Equal]", 
      RowBox[{"{", "}"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(@\[Equal]{} iff)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765100086422225*^9, 3.7651001032545033`*^9}},
 CellTags->"@\[Equal]{} iff",
 CellLabel->"In[457]:=",
 CellID->685115154,ExpressionUUID->"4c3aa816-2f51-4a8c-8b3d-a002c2ecd252"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "A", ",", "P"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{"x", "\[Element]", 
        RowBox[{"f", "@", "A"}]}]], 
      RowBox[{"P", "[", "x", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{"a", "\[Element]", "A"}]], 
      RowBox[{"P", "[", 
       RowBox[{"f", "[", "a", "]"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[ForAll]@)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651004035306387`*^9, 3.765100437394388*^9}},
 CellTags->"\[ForAll]@",
 CellLabel->"In[458]:=",
 CellID->1570027835,ExpressionUUID->"66f7bb38-30cc-49de-bc0e-de89ae07f96c"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "A", ",", "P"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{"x", "\[Element]", 
        RowBox[{"f", "@", "A"}]}]], 
      RowBox[{"P", "[", "x", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{"a", "\[Element]", "A"}]], 
      RowBox[{"P", "[", 
       RowBox[{"f", "[", "a", "]"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Exists]@)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765100468501602*^9, 3.765100473153729*^9}},
 CellTags->"\[Exists]@",
 CellLabel->"In[459]:=",
 CellID->771344911,ExpressionUUID->"8b19b12b-215f-4771-b7ad-e947689e6e14"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "P"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"f", "[", "x", "]"}], 
      UnderscriptBox["|", "x"], 
      RowBox[{"P", "[", "x", "]"}]}], "}"}], "\[Equal]", 
    RowBox[{"f", "@", 
     RowBox[{"Collect", "[", "P", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(SetOf\[Equal]@Collect)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7651006759911203`*^9, 3.765100742950398*^9}},
 CellTags->"SetOf\[Equal]@Collect",
 CellLabel->"In[460]:=",
 CellID->1251957142,ExpressionUUID->"79ebb868-37c7-4860-8e0b-65f7121ee750"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "A"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"f", "[", "x", "]"}], 
      UnderscriptBox["|", "x"], 
      RowBox[{"x", "\[Element]", "A"}]}], "}"}], "\[Equal]", 
    RowBox[{"f", "@", "A"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(SetOf\[Equal]@)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765100847623657*^9, 3.765100857356839*^9}},
 CellTags->"SetOf\[Equal]@",
 CellLabel->"In[461]:=",
 CellID->2088273463,ExpressionUUID->"04de836f-2f8e-4c32-b3ee-1f22e4061854"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1593529551,ExpressionUUID->"283ba5a0-14a8-496b-82f9-5a0772fad864"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->182572638,ExpressionUUID->"ce63b603-f078-4331-aa75-0494586678ae"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"09bee950-675c-4eb7-9b5c-51d3aaf580b7"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(@\[Union])\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(@\[Union])\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["814265295", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:814265295", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:24, 0.582684 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:814265295-1", "3"},
 CellLabel->"In[462]:=",
 CellID->1805064061,ExpressionUUID->"4d04f4ae-48b6-4184-a3cd-f21b71cb61b2"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"0d940f26-8c71-4be8-954e-4890802d4304"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(@ ident)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(@ ident)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1283984586", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1283984586", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:24, 0.124954 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1283984586-1", "3"},
 CellLabel->"In[463]:=",
 CellID->203354138,ExpressionUUID->"268be10f-dd81-47de-b132-ddea9a7f3a82"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"1f5719a8-0990-403d-96f9-540528ba7ef6"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(@{})\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(@{})\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1257505636", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1257505636", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:24, 0.171999 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1257505636-1", "3"},
 CellLabel->"In[464]:=",
 CellID->1997254040,ExpressionUUID->"c60cafa1-dc19-4490-a5d0-a50cbd0e6903"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"d3e0ff1e-43a3-485c-a4de-9333926ed329"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(@Insert)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(@Insert)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["565842574", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:565842574", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:25, 0.690976 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:565842574-1", "3"},
 CellLabel->"In[465]:=",
 CellID->477474021,ExpressionUUID->"243af554-932a-474a-a189-73c259821534"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"fe2af352-a1a3-44f7-9b06-c2e485aa2825"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(@\[Equal]{} iff)\"", "FormReference", StripOnInput -> False],
             True -> StyleBox[
             "\"(@\[Equal]{} iff)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["685115154", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:685115154", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:25, 0.259005 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:685115154-1", "3"},
 CellLabel->"In[466]:=",
 CellID->1626158995,ExpressionUUID->"231ca40f-e980-4b69-b7ca-c100eb04e88e"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"8213da70-da30-49e1-ad3f-71ba2fb45dcc"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[ForAll]@)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[ForAll]@)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1570027835", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1570027835", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:26, 0.254684 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1570027835-1", "3"},
 CellLabel->"In[467]:=",
 CellID->1776911464,ExpressionUUID->"e7c56a2e-bc98-4cf0-80e9-3bac79998304"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"14979f6e-7aff-4e28-9dae-3b78ff52153f"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Exists]@)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[Exists]@)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["771344911", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:771344911", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:27, 0.610300 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:771344911-1", "3"},
 CellLabel->"In[468]:=",
 CellID->141629000,ExpressionUUID->"860e2886-3792-4877-8a4c-304213f7d561"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"e7373358-a065-483f-8aac-89fb14130176"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(SetOf\[Equal]@Collect)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(SetOf\[Equal]@Collect)\"", "FormReferenceHover", 
              StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1251957142", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1251957142", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:27, 0.212299 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1251957142-1", "3"},
 CellLabel->"In[469]:=",
 CellID->150377480,ExpressionUUID->"cecad2e4-4de8-4466-95cb-7b7894b699c6"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"22503a60-290f-4546-9907-abaa6fea955c"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(SetOf\[Equal]@)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(SetOf\[Equal]@)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["2088273463", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:2088273463", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:01:27, 0.063998 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:2088273463-1", "3"},
 CellLabel->"In[470]:=",
 CellID->1818369679,ExpressionUUID->"b6d3e804-b5ea-4af4-92d5-df3642a09ce3"],

Cell[BoxData[
 RowBox[{"rules", " ", 
  RowBox[{"(", "simp", ")"}], " ", "\"\<@ ident\>\"", " ", "\"\<@{}\>\"", 
  " ", "\"\<@\[Equal]{} iff\>\"", " ", "\"\<\[ForAll]@\>\"", " ", 
  "\"\<\[Exists]@\>\""}]], "TheoremaCode",
 CellChangeTimes->{{3.765268479670273*^9, 3.765268537349618*^9}},
 CellLabel->"In[471]:=",
 CellID->1437902574,ExpressionUUID->"5a20411d-76e9-4691-abcb-bdc2ec43704d"]
}, Closed]],

Cell[CellGroupData[{

Cell["Monotonicity", "Subsection",
 CellChangeTimes->{{3.7302895247483892`*^9, 3.7302895273118153`*^9}},
 CellID->397026471,ExpressionUUID->"d997d30d-e899-432a-997a-4295f73a9abf"],

Cell["", "OpenEnvironment",
 CellID->1898688307,ExpressionUUID->"601b9d42-41d9-4abd-82cb-f93913ac8545"],

Cell[CellGroupData[{

Cell["Lemma (Monotonicity)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.76510137683836*^9, 3.7651013802140837`*^9}},
 CellID->619323678,ExpressionUUID->"bc300775-4c6c-4e35-88da-615cf64925c8"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "A", ",", "B"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"f", "@", "A"}], "\[SubsetEqual]", 
     RowBox[{"f", "@", "B"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(@ mono)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765099873555023*^9, 3.765099894400775*^9}},
 CellTags->"@ mono",
 CellLabel->"In[472]:=",
 CellID->570770860,ExpressionUUID->"1f0cd3b1-4c97-4af8-889c-d482a16463ad"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"Pow", "[", "A", "]"}], "\[SubsetEqual]", 
     RowBox[{"Pow", "[", "B", "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Pow mono)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765271169269067*^9, 3.765271192076838*^9}},
 CellTags->"Pow mono",
 CellLabel->"In[473]:=",
 CellID->1108679522,ExpressionUUID->"dc492e7f-ab0b-45d7-8e5f-894206d553d3"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "a"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"Insert", "[", 
      RowBox[{"a", ",", "A"}], "]"}], "\[SubsetEqual]", 
     RowBox[{"Insert", "[", 
      RowBox[{"a", ",", "B"}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Insert mono)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765271203596819*^9, 3.765271229668449*^9}},
 CellTags->"Insert mono",
 CellLabel->"In[474]:=",
 CellID->788559876,ExpressionUUID->"94e58813-71c0-439e-b734-f656ba867cb1"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"-", "B"}], "\[SubsetEqual]", 
     RowBox[{"-", "A"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(complement antimono)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765271950419279*^9, 3.765271979283286*^9}},
 CellTags->"complement antimono",
 CellLabel->"In[475]:=",
 CellID->604821491,ExpressionUUID->"e9f9e5f9-d013-4783-8a5e-8ae725783603"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C", ",", "D"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"C", "\[SubsetEqual]", "D"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"A", "\[Union]", "C"}], "\[SubsetEqual]", 
     RowBox[{"B", "\[Union]", "D"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Union] mono)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765272017826434*^9, 3.7652720385783253`*^9}},
 CellTags->"\[Union] mono",
 CellLabel->"In[476]:=",
 CellID->1716105510,ExpressionUUID->"f8f62021-ce16-4148-97bc-a826f1bd212d"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C", ",", "D"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"C", "\[SubsetEqual]", "D"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"A", "\[Intersection]", "C"}], "\[SubsetEqual]", 
     RowBox[{"B", "\[Intersection]", "D"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(\[Intersection] mono)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.7652722542713013`*^9, 3.765272287655199*^9}},
 CellTags->"\[Intersection] mono",
 CellLabel->"In[477]:=",
 CellID->1849777423,ExpressionUUID->"e21c8cc3-a299-4b52-a555-2f06ebaa8254"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"A", ",", "B", ",", "C", ",", "D"}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"A", "\[SubsetEqual]", "B"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"D", "\[SubsetEqual]", "C"}], 
    TagBox[")",
     "AutoParentheses"]}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"A", "-", "C"}], "\[SubsetEqual]", 
     RowBox[{"B", "-", "D"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(- mono)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765510099084961*^9, 3.765510107119927*^9}},
 CellTags->"- mono",
 CellLabel->"In[478]:=",
 CellID->1569569101,ExpressionUUID->"efb2b44d-7bb0-461d-a404-c6b1978ae33a"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", 
   RowBox[{"f", ",", "g", ",", "P", ",", "Q"}]], 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     UnderscriptBox["\[And]", "x"], 
     RowBox[{
      RowBox[{"P", "[", "x", "]"}], "\[LongRightArrow]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox["\[Exists]", 
         RowBox[{"Q", "[", "y", "]"}]], 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          RowBox[{"f", "[", "x", "]"}], "\[Equal]", 
          RowBox[{"g", "[", "y", "]"}]}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}]}], ")"}], "\[LongRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"f", "[", "x", "]"}], 
       UnderscriptBox["|", "x"], 
       RowBox[{"P", "[", "x", "]"}]}], "}"}], "\[SubsetEqual]", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"g", "[", "x", "]"}], 
       UnderscriptBox["|", "x"], 
       RowBox[{"Q", "[", "x", "]"}]}], "}"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(SetOf mono)", "FrameLabel"], Evaluator -> Automatic, 
         Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.765180805820352*^9, 3.765180898731248*^9}},
 CellTags->"SetOf mono",
 CellLabel->"In[479]:=",
 CellID->172277084,ExpressionUUID->"1be2d71c-9305-4560-9d81-fafc180c5b2b"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->833259476,ExpressionUUID->"4fa1ee10-87ae-491c-9c41-de19ffabf718"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->786833653,ExpressionUUID->"83c10aeb-6402-4a2a-b9fe-80469e4d7fc4"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"9a540c0f-d53c-4d75-94c0-a69f40951450"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(@ mono)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(@ mono)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["570770860", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:570770860", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:02:13, 0.211839 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:570770860-1", "3"},
 CellLabel->"In[480]:=",
 CellID->602566564,ExpressionUUID->"d06f131e-6326-4e72-a0bc-30541cdf0aa3"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"629980a3-3fc1-45dd-8f28-ab79ee0ee3ef"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Pow mono)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(Pow mono)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1108679522", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1108679522", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:02:14, 0.128162 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1108679522-1", "3"},
 CellLabel->"In[481]:=",
 CellID->1093017718,ExpressionUUID->"619955c1-53c3-41d6-b1c8-daf0572378fc"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"fa75d247-e335-4664-bccd-3592d1dfc7a4"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Insert mono)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(Insert mono)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["788559876", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:788559876", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:02:14, 0.210764 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:788559876-1", "3"},
 CellLabel->"In[482]:=",
 CellID->1966110247,ExpressionUUID->"6344d294-e9d4-420c-a9cb-bcd22dfbe228"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"6955774f-f367-4c5a-8ee3-6553d1080260"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(complement antimono)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(complement antimono)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["604821491", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:604821491", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:02:15, 0.071898 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:604821491-1", "3"},
 CellLabel->"In[483]:=",
 CellID->429588115,ExpressionUUID->"5334709a-119e-480c-b8ba-516e9e54b89b"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"2acdfb55-9f87-44a1-a349-00190aa87669"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Union] mono)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(\[Union] mono)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1716105510", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1716105510", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:02:16, 0.158510 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1716105510-1", "3"},
 CellLabel->"In[484]:=",
 CellID->2002110999,ExpressionUUID->"4c1795f9-7bd4-48d4-ad29-b0688fdc3008"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"edbdb900-5fa1-4805-832f-b833236d6433"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(\[Intersection] mono)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(\[Intersection] mono)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1849777423", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1849777423", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:02:17, 0.209498 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1849777423-1", "3"},
 CellLabel->"In[485]:=",
 CellID->1702528487,ExpressionUUID->"8dd5ad08-d868-4d2e-ad57-6ee353a4c58b"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"57ee5349-6f9b-419d-9fde-e2144195bb29"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(- mono)\"", "FormReference", StripOnInput -> False], True -> 
            StyleBox[
             "\"(- mono)\"", "FormReferenceHover", StripOnInput -> False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["1569569101", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:1569569101", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:05:33, 0.184612 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:1569569101-1", "3"},
 CellLabel->"In[538]:=",
 CellID->1571614296,ExpressionUUID->"f5c12280-d2b6-42bc-8fbb-e7baab4c0376"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"1dfcf928-207c-4fd0-b310-21851d08f5b3"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(SetOf mono)\"", "FormReference", StripOnInput -> False], 
            True -> StyleBox[
             "\"(SetOf mono)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["172277084", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:172277084", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:02:17, 0.281007 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:172277084-1", "3"},
 CellLabel->"In[486]:=",
 CellID->665117787,ExpressionUUID->"f177dd05-a638-4859-bf62-074cc8860f82"]
}, Closed]],

Cell[CellGroupData[{

Cell["Cantor\[CloseCurlyQuote]s Paradox", "Subsection",
 CellChangeTimes->{{3.731486856575044*^9, 3.731486860605254*^9}},
 CellID->19992184,ExpressionUUID->"4333526b-cfc3-4819-9486-4332bd86473c"],

Cell["", "OpenEnvironment",
 CellID->318971740,ExpressionUUID->"0e335947-1101-49d4-8136-3b5403b8318a"],

Cell[CellGroupData[{

Cell["Lemma (Cantor\[CloseCurlyQuote]s Paradox)", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.7314868677120743`*^9, 3.7314868704851513`*^9}},
 CellID->28949169,ExpressionUUID->"7cf961ec-f8c5-4433-9b54-1fa29b06c3e0"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[NotExists]", 
   RowBox[{"f", "::", 
    RowBox[{"(", 
     RowBox[{"a_", "\[Rule]", 
      RowBox[{"{", "a_", "}"}]}], ")"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", "A"], 
   RowBox[{
    UnderscriptBox["\[Exists]", "x"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"A", "\[Equal]", 
      RowBox[{"f", "[", "x", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Cantor's paradox)", "FrameLabel"], Evaluator -> Automatic,
          Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.731486877323251*^9, 3.731486880922626*^9}, {
   3.7314869351533213`*^9, 3.7314869669511747`*^9}, 3.731487004896861*^9, {
   3.7314885757317495`*^9, 3.7314886088035145`*^9}, {3.731489852452713*^9, 
   3.7314898603538847`*^9}, {3.7315635767032194`*^9, 
   3.7315636283630934`*^9}, {3.7315638976545773`*^9, 3.7315639114556704`*^9}},
 CellTags->"Cantor's paradox",
 CellLabel->"In[539]:=",
 CellID->628580588,ExpressionUUID->"a0cf4926-7497-4691-ac9b-596d4394ea91"],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[And]", "A"], 
  RowBox[{
   UnderscriptBox["\[NotExists]", "f"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", "B"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"B", "\[SubsetEqual]", "A"}], "\[Implies]", 
      RowBox[{
       UnderscriptBox["\[Exists]", 
        RowBox[{"x", "\[Element]", "A"}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"B", "\[Equal]", 
         RowBox[{"f", "[", "x", "]"}]}], 
        TagBox[")",
         "AutoParentheses"]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        ButtonBox[
         StyleBox["(Cantor's paradox 2)", "FrameLabel"], Evaluator -> 
         Automatic, Appearance -> None, ButtonFunction :> 
         Theorema`Common`editCellTags[]], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[]]}]]]}, {
   None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.73156388404119*^9, {3.731564143185074*^9, 3.731564148530254*^9}, {
   3.7315652888244834`*^9, 3.731565291699853*^9}, {3.765106219797222*^9, 
   3.765106220451517*^9}},
 CellTags->"Cantor's paradox 2",
 CellLabel->"In[540]:=",
 CellID->289824169,ExpressionUUID->"c1f91a4d-9e0e-4fc8-b926-adf6ecfaa815"],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->373064460,ExpressionUUID->"8c847f23-e35a-47ff-85a3-50f610dea31d"]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->626869507,ExpressionUUID->"ce4d5482-79dc-43ac-ab6b-69d6eb56b85e"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"8e65139b-d3ad-440f-b12c-b18ea5f1ceb9"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Cantor's paradox)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(Cantor's paradox)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["628580588", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:628580588", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:05:51, 0.615792 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:628580588-1", "3"},
 CellLabel->"In[541]:=",
 CellID->202346487,ExpressionUUID->"dfe4d76c-665e-46e9-a153-3fd07c91d051"],

Cell[TextData[Cell[BoxData[
 ButtonBox[
  StyleBox["\<\"Show proof\"\>",
   StripOnInput->False,
   FontVariations->{"Underline"->True}],
  Appearance->None,
  ButtonFunction:>Theorema`Core`Tools`Interactive`Private`displayCoreProof[],
  Evaluator->Automatic,
  ImageSize->Automatic,
  Method->
   "Queued"]],ExpressionUUID->"36099243-9851-4484-8a5c-95abb68d7dfb"]], \
"ProofInfo",
 ShowCellBracket->Automatic,
 Evaluatable->True,
 CellEvaluationFunction->
  Theorema`Core`Tools`Interactive`Private`evaluateProofCell,
 CellFrameLabels->{{
    Cell[
     TextData[{"\[CheckmarkedBox] ", "\[FivePointedStar]", " Proof of ", 
       Cell[
        BoxData[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> StyleBox[
             "\"(Cantor's paradox 2)\"", "FormReference", StripOnInput -> 
              False], True -> 
            StyleBox[
             "\"(Cantor's paradox 2)\"", "FormReferenceHover", StripOnInput -> 
              False]}, 
           Dynamic[
            CurrentValue["MouseOver"]], ImageSize -> Automatic, FrameMargins -> 
           0], ButtonFunction :> 
          If[Theorema`Common`selectCells["289824169", CellID] === $Failed, 
            Theorema`Common`selectCells["ID:289824169", CellTags]], 
          Appearance -> None, Evaluator -> Automatic, Method -> 
          "Preemptive"]]], " \[NumberSign]1  ", 
       StyleBox[
       "[Mon 29 Apr 2019 09:05:53, 1.124147 sec]", FontSize -> Small]}]], 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Core`Tools`Interactive`Private`removeCoreProof[]]]]}, {
   None, None}},
 CellTags->{"Proof|ID:289824169-1", "4"},
 CellLabel->"In[542]:=",
 CellID->383070079,ExpressionUUID->"64679309-62a4-49fc-8eec-bdb7212bfd48"],

Cell[TextData[{
 "Theorem ",
 ButtonBox["(Cantor\[CloseCurlyQuote]s paradox)",
  BaseStyle->"Hyperlink",
  ButtonData->"Cantor's paradox"],
 " can be obtained directly from ",
 ButtonBox["(Cantor\[CloseCurlyQuote]s paradox 2)",
  BaseStyle->"Hyperlink",
  ButtonData->"Cantor's paradox 2"],
 ":"
}], "Text",
 CellChangeTimes->{{3.765106435752122*^9, 3.7651064902668457`*^9}, {
  3.7651065231366243`*^9, 3.765106529855898*^9}},
 CellID->1219761075,ExpressionUUID->"45f9ec30-3522-4e1a-abe9-001e7f7cf017"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"fml", " ", 
  RowBox[{"\"\<Cantor's paradox 2\>\"", "[", 
   RowBox[{
    RowBox[{"of", " ", 
     RowBox[{"\[OpenCurlyQuote]", "UNIV", "\[CloseCurlyQuote]"}]}], ",", " ", 
    RowBox[{"simplified", " ", 
     RowBox[{"only", ":", " ", 
      RowBox[{
      "\"\<\[SubsetEqual]UNIV\>\"", " ", "\"\<UNIV I\>\"", " ", 
       "\"\<True\[Implies]\>\""}]}]}]}], "]"}]}]], "TheoremaCode",
 CellChangeTimes->{{3.7651061719395247`*^9, 3.765106188459938*^9}, {
  3.765106230893236*^9, 3.765106288626602*^9}},
 CellLabel->"In[543]:=",
 CellID->1767667081,ExpressionUUID->"8c1e3ac1-5f29-4015-995f-0d13a830c93e"],

Cell[BoxData[
 TagBox[
  FormBox[
   RowBox[{
    UnderscriptBox[
     TagBox["\[NotExists]",
      Theorema`Language`TAG$[
       Theorema`Language`TYPED$[Theorema`Language`NotExists$TM, {
         Theorema`Language`FUN$TM[
          Theorema`Language`TVAR$[0], 
          Theorema`Knowledge`SET$TM[
           Theorema`Language`TVAR$[0]]]}]]], 
     TagBox[
      StyleBox["f", "ExpressionVariable"],
      Theorema`Language`TAG$[
       Theorema`Language`TYPED$[
        Theorema`Language`FUN$TM[
         Theorema`Language`TVAR$[0], 
         Theorema`Knowledge`SET$TM[
          Theorema`Language`TVAR$[0]]]]]]], 
    RowBox[{
     UnderscriptBox[
      TagBox["\[ForAll]",
       Theorema`Language`TAG$[
        Theorema`Language`TYPED$[Theorema`Language`Forall$TM, {
          Theorema`Knowledge`SET$TM[
           Theorema`Language`TVAR$[0]]}]]], 
      TagBox[
       StyleBox["B", "ExpressionVariable"],
       Theorema`Language`TAG$[
        Theorema`Language`TYPED$[
         Theorema`Knowledge`SET$TM[
          Theorema`Language`TVAR$[0]]]]]], 
     RowBox[{
      UnderscriptBox[
       TagBox["\[Exists]",
        Theorema`Language`TAG$[
         Theorema`Language`TYPED$[Theorema`Language`Exists$TM, {
           Theorema`Language`TVAR$[0]}]]], 
       TagBox[
        StyleBox["x", "ExpressionVariable"],
        Theorema`Language`TAG$[
         Theorema`Language`TYPED$[
          Theorema`Language`TVAR$[0]]]]], 
      RowBox[{
       TagBox[
        StyleBox["B", "ExpressionVariable"],
        Theorema`Language`TAG$[
         Theorema`Language`TYPED$[
          Theorema`Knowledge`SET$TM[
           Theorema`Language`TVAR$[0]]]]], 
       TagBox["\[Equal]",
        Theorema`Language`TAG$[
         Theorema`Language`TYPED$[Theorema`Language`Equal$TM, {
           Theorema`Knowledge`SET$TM[
            Theorema`Language`TVAR$[0]]}]],
        SyntaxForm->"x\[Equal]y"], 
       RowBox[{
        TagBox[
         StyleBox["f", "ExpressionVariable"],
         Theorema`Language`TAG$[
          Theorema`Language`TYPED$[
           Theorema`Language`FUN$TM[
            Theorema`Language`TVAR$[0], 
            Theorema`Knowledge`SET$TM[
             Theorema`Language`TVAR$[0]]]]]], "[", 
        TagBox[
         StyleBox["x", "ExpressionVariable"],
         Theorema`Language`TAG$[
          Theorema`Language`TYPED$[
           Theorema`Language`TVAR$[0]]]], "]"}]}]}]}]}],
   TheoremaForm],
  DisplayForm]], "Print",
 CellChangeTimes->{{3.765106227589764*^9, 3.76510628999975*^9}, 
   3.7655103538573723`*^9},
 CellLabel->"During evaluation of In[543]:=",
 CellID->1500264271,ExpressionUUID->"d47b3a55-0bb4-4693-b519-7da41c2bf2a7"]
}, Open  ]]
}, Closed]]
}, Closed]]
}, Open  ]]
},
WindowSize->{1258, 1017},
WindowMargins->{{3, Automatic}, {3, Automatic}},
CounterAssignments->{{"TheoremaFormulaCounter" -> 0}, {
  "NotebookFileName" -> 
   "/home/amaletzk/WolframWorkspaces/Base/Theorema/Theorema/CoreTheories/HOL/\
Sets.nb"}},
FrontEndVersion->"11.3 for Linux x86 (64-bit) (March 6, 2018)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Theorema"}, "TheoremaCore.nb"]]], 
   Cell["Modify the definitions below to change the default appearance of all \
cells in a given style. \nMake modifications to any definition using commands \
in the Format menu.", "Text"], 
   Cell[
    CellGroupData[{
      Cell["Notebook Options Settings", "Section"], 
      Cell[
      "The options defined for the style below will be used at the Notebook \
level.", "Text"], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Notebook"], InputAliases -> {"and" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[And]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "or" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", "\[Or]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "not" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Not]", "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "impl" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[Implies]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "equiv" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[DoubleLongLeftRightArrow]", 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "eq" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[Equal]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], ":equiv" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox[
                  
                  RowBox[{
                   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                    "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                  "a:=b"], "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], ":eq" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", ":=", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "andn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[And]", "\[Piecewise]", 
                 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}, 
                  ColumnAlignments -> Left]}], 
               TagBox[")", "AutoParentheses"]}], "orn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Or]", "\[Piecewise]", 
                 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}, 
                  ColumnAlignments -> Left]}], 
               TagBox[")", "AutoParentheses"]}], "equivn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[DoubleLeftRightArrow]", "\[Piecewise]", 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}]}], 
               TagBox[")", "AutoParentheses"]}], "cdist" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Piecewise]", 
                 
                 GridBox[{{
                   "\[SelectionPlaceholder]", "\[DoubleLeftArrow]", 
                    "\[Placeholder]"}, {
                   "\[Placeholder]", "\[DoubleLeftArrow]", 
                    "\[Placeholder]"}}]}], 
               TagBox[")", "AutoParentheses"]}], "far" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", "\[Placeholder]"], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "exr" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["\[Exists]", "\[Placeholder]"], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "farc" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", "\[Placeholder]"], 
                  "\[Placeholder]"], "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "exrc" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[Exists]", "\[Placeholder]"], 
                  "\[Placeholder]"], "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "let" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["let", 
                  RowBox[{"\[Placeholder]", "=", "\[Placeholder]"}]], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "app" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:293a", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "prep" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:293b", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "join" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:22c8", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "elemT" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:22ff", Identity, SyntaxForm -> "a+b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "subop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "supop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SuperscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "subsupop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubsuperscriptBox[
               "\[SelectionPlaceholder]", "\[Placeholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "oop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               OverscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "uoop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               UnderoverscriptBox[
               "\[SelectionPlaceholder]", "\[Placeholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "osop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox[
                OverscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
                "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "op" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               TagBox["\[SelectionPlaceholder]", Identity], 
               TagBox[")", "AutoParentheses"]}], "seq" -> RowBox[{
               TagBox["\[VerticalEllipsis]", Identity, SyntaxForm -> "("], 
               "\[SelectionPlaceholder]", 
               TagBox["\[VerticalEllipsis]", Identity, SyntaxForm -> ")"]}], 
            "()" -> RowBox[{
               TagBox["\:fd3e", Identity, SyntaxForm -> "("], 
               "\[SelectionPlaceholder]", 
               TagBox["\:fd3f", Identity, SyntaxForm -> ")"]}], "(())" -> 
            RowBox[{
               TagBox["\:2e28", Identity, SyntaxForm -> "("], 
               "\[SelectionPlaceholder]", 
               TagBox["\:2e29", Identity, SyntaxForm -> ")"]}], "(|)" -> 
            RowBox[{
               TagBox["\:2987", Identity, SyntaxForm -> "("], 
               "\[SelectionPlaceholder]", 
               TagBox["\:2988", Identity, SyntaxForm -> ")"]}], "[]" -> 
            RowBox[{
               TagBox["\:e114", Identity, SyntaxForm -> "("], 
               "\[SelectionPlaceholder]", 
               TagBox["\:e115", Identity, SyntaxForm -> ")"]}], "[[]]" -> 
            RowBox[{
               TagBox["\:27e6", Identity, SyntaxForm -> "("], 
               "\[SelectionPlaceholder]", 
               TagBox["\:27e7", Identity, SyntaxForm -> ")"]}], "[|]" -> 
            RowBox[{
               TagBox["\:27ec", Identity, SyntaxForm -> "("], 
               "\[SelectionPlaceholder]", 
               TagBox["\:27ed", Identity, SyntaxForm -> ")"]}], "{}" -> 
            RowBox[{
               TagBox["\:e117", Identity, SyntaxForm -> "("], 
               "\[SelectionPlaceholder]", 
               TagBox["\:e118", Identity, SyntaxForm -> ")"]}], "{{}}" -> 
            RowBox[{
               TagBox["\:2983", Identity, SyntaxForm -> "("], 
               "\[SelectionPlaceholder]", 
               TagBox["\:2984", Identity, SyntaxForm -> ")"]}], "<>" -> 
            RowBox[{
               TagBox["\:27e8", Identity, SyntaxForm -> "("], 
               "\[SelectionPlaceholder]", 
               TagBox["\:27e9", Identity, SyntaxForm -> ")"]}], "<<>>" -> 
            RowBox[{
               TagBox["\:27ea", Identity, SyntaxForm -> "("], 
               "\[SelectionPlaceholder]", 
               TagBox["\:27eb", Identity, SyntaxForm -> ")"]}], "<|>" -> 
            RowBox[{
               TagBox["\:2989", Identity, SyntaxForm -> "("], 
               "\[SelectionPlaceholder]", 
               TagBox["\:298a", Identity, SyntaxForm -> ")"]}], "<.>" -> 
            RowBox[{
               TagBox["\:2991", Identity, SyntaxForm -> "("], 
               "\[SelectionPlaceholder]", 
               TagBox["\:2992", Identity, SyntaxForm -> ")"]}], "<c>" -> 
            RowBox[{
               TagBox["\:29fc", Identity, SyntaxForm -> "("], 
               "\[SelectionPlaceholder]", 
               TagBox["\:29fd", Identity, SyntaxForm -> ")"]}], "(" -> 
            TagBox["(", "AutoParentheses"], ")" -> 
            TagBox[")", "AutoParentheses"]}, Background -> 
          RGBColor[0.957, 0.96, 0.97]], 
         Cell[
          StyleData["Notebook", "Printout"], Background -> None]}, Open]]}, 
     Open]], 
   Cell[
    CellGroupData[{
      Cell["Style Environment Names", "Section"], 
      Cell[
       StyleData[All, "Working"], DockedCells -> {
         Cell[
         "Theorema 2.0", "SBO", FontTracking -> 8, 
          CellMargins -> {{0, 0}, {0, 0}}, 
          CellFrameMargins -> {{0, 50}, {0, 0}}, FontSize -> 28, FontSlant -> 
          "Plain", FontColor -> RGBColor[0.29, 0.46, 0.6], TextAlignment -> 
          Right, Background -> RGBColor[0.36, 0.54, 0.69]]}], 
      Cell[
       StyleData[
       All, "Layout", StyleDefinitions -> StyleData[All, "Working"]], 
       MenuSortingValue -> 1200]}, Open]], 
   Cell[
    CellGroupData[{
      Cell["Styles for Title and Section Cells", "Section"], 
      Cell[
       CellGroupData[{
         Cell["Title", "Subsection"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Title"], CellMargins -> {{27, 27}, {10, 30}}, 
             TextAlignment -> Center, MenuSortingValue -> 1100, 
             MenuCommandKey -> "1", FontFamily -> "Bitstream Charter", 
             FontSize -> 36, FontWeight -> "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Title", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{6, 10}, {5, 30}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subtitle"], CellMargins -> {{27, 27}, {20, 2}}, 
             TextAlignment -> Center, MenuSortingValue -> 1150, FontFamily -> 
             "Bitstream Charter", FontSize -> 24, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subtitle", "Printout"], 
             CellMargins -> {{6, 10}, {5, 0}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubtitle"], CellMargins -> {{27, 27}, {8, 2}}, 
             TextAlignment -> Center, MenuSortingValue -> 1200, FontFamily -> 
             "Bitstream Charter", FontSize -> 16, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{6, 10}, {0, 0}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Author"], CellMargins -> {{27, 27}, {18, 20}}, 
             TextAlignment -> Center, MenuSortingValue -> 1200, FontFamily -> 
             "Bitstream Charter", FontSize -> 16, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{6, 10}, {0, 0}}]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Sectioning", "Subsection"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Section"], ShowGroupOpener -> True, MenuSortingValue -> 
             1250, MenuCommandKey -> "4", FontFamily -> "Bitstream Charter", 
             FontSize -> 20, FontWeight -> "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.3, 0.3, 0.3]], 
            Cell[
             StyleData["Section", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{2, 0}, {7, 22}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsection"], CellDingbat -> None, ShowGroupOpener -> 
             True, CellMargins -> {{27, Inherited}, {8, 12}}, 
             MenuSortingValue -> 1300, MenuCommandKey -> "5", FontFamily -> 
             "Bitstream Charter", FontSize -> 14, FontWeight -> "Bold", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.3, 0.3, 0.3]], 
            Cell[
             StyleData["Subsection", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{9, 0}, {7, 22}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsection"], CellDingbat -> None, ShowGroupOpener -> 
             True, CellMargins -> {{27, Inherited}, {2, 10}}, 
             MenuSortingValue -> 1350, MenuCommandKey -> "6", FontFamily -> 
             "Bitstream Charter", FontSize -> 12, FontWeight -> "Bold", 
             FontSlant -> "Italic", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.4, 0.4, 0.4]], 
            Cell[
             StyleData["Subsubsection", "Printout"], ShowGroupOpener -> False,
              CellMargins -> {{35, 0}, {7, 14}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsubsection"], CellDingbat -> None, 
             ShowGroupOpener -> True, 
             CellMargins -> {{27, Inherited}, {Inherited, Inherited}}, 
             FontFamily -> "Bitstream Charter", FontSize -> 11, FontWeight -> 
             "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.4, 0.4, 0.4]], 
            Cell[
             StyleData["Subsubsubsection", "Printout"], 
             CellMargins -> {{50, 0}, {7, 14}}]}, Closed]]}, Open]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell["Styles for Publication\[Hyphen]ready Documents", "Section"], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Author"], CellMargins -> {{27, 30}, {20, 5}}, 
          MenuSortingValue -> 1900, FontFamily -> "Helvetica", FontSize -> 14,
           FontSlant -> "Italic"], 
         Cell[
          StyleData["Author", "Printout"], CellMargins -> {{6, 10}, {0, 10}}, 
          FontSize -> 10]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Abstract"], CellFrame -> {{0, 0}, {0.5, 0.5}}, 
          CellMargins -> {{27, 140}, {20, 20}}, LineSpacing -> {0.9, 0}, 
          MenuSortingValue -> 2050, FontFamily -> "Times", FontSize -> 12], 
         Cell[
          StyleData["Abstract", "Printout"], Hyphenation -> True, 
          LineSpacing -> {1, 2}, FontSize -> 10]}, Closed]]}, Closed]], 
   Cell[
    CellGroupData[{
      Cell["Text", "Section"], 
      Cell[
       StyleData["Text"], CellMargins -> {{27, 27}, {Inherited, Inherited}}, 
       TextJustification -> 1., ParagraphIndent -> 15, FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       StyleData["EnvironmentText"], ShowCellBracket -> Automatic, 
       CellMargins -> {{82, Inherited}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       TextJustification -> 1., ParagraphSpacing -> {0.3, 0.}, 
       MenuSortingValue -> 1480, MenuCommandKey -> "T", FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Item"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{81, 10}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          CellFrameLabelMargins -> 4, CounterIncrements -> "Item", 
          CounterAssignments -> {{"Subitem", 0}, {"Subsubitem", 0}}, 
          MenuSortingValue -> 1600, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Item", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["ItemParagraph"], CellMargins -> {{81, 10}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          DefaultNewCellStyle -> "Item", MenuSortingValue -> 1600, FontFamily -> 
          "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", 
          FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["ItemParagraph", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Subitem"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{105, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", 
          CounterIncrements -> "Subitem", 
          CounterAssignments -> {{"Subsubitem", 0}}, MenuSortingValue -> 1610,
           FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Subitem", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubitemParagraph"], CellMargins -> {{105, 12}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subitem", 
          MenuSortingValue -> 1610, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubitemParagraph", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Subsubitem"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{129, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subsubitem", 
          CounterIncrements -> "Subsubitem", MenuSortingValue -> 1620, 
          FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Subsubitem", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubsubitemParagraph"], 
          CellMargins -> {{129, 12}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subsubitem", 
          CounterIncrements -> "Subsubitem", MenuSortingValue -> 1625, 
          FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubsubitemParagraph", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["ItemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{81, 10}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          CellFrameLabelMargins -> 4, CounterIncrements -> "ItemNumbered", 
          CounterAssignments -> {{"SubitemNumbered", 0}, {
            "SubsubitemNumbered", 0}}, MenuSortingValue -> 1630, FontFamily -> 
          "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", 
          FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["ItemNumbered", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubitemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], ".", 
              CounterBox["SubitemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{105, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", 
          CounterIncrements -> "SubitemNumbered", 
          CounterAssignments -> {{"SubsubitemNumbered", 0}}, MenuSortingValue -> 
          1640, FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubitemNumbered", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubsubitemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], ".", 
              CounterBox["SubitemNumbered"], ".", 
              CounterBox["SubsubitemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{129, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> 
          "SubitemNumbered", CounterIncrements -> "SubsubitemNumbered", 
          MenuSortingValue -> 1645, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubsubitemNumbered", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]]}, Closed]], 
   Cell[
    CellGroupData[{
      Cell["Theorema", "Section"], 
      Cell[
       StyleData["GlobalDeclaration"], CellFrame -> {{0, 2}, {0, 0}}, 
       ShowCellBracket -> Automatic, CellMargins -> {{76, 3}, {2, 2}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       InitializationCell -> True, CellFrameMargins -> 10, CellFrameColor -> 
       RGBColor[0.73, 0.65, 0.61], CellFrameLabelMargins -> {{33, 7}, {1, 1}},
        ShowAutoStyles -> False, MenuSortingValue -> 1441, MenuCommandKey -> 
       "G", FontSize -> 14], 
      Cell[
       StyleData["OpenEnvironment"], Editable -> False, Selectable -> False, 
       CellFrame -> {{False, False}, {True, False}}, ShowCellBracket -> True, 
       CellMargins -> {{76, 6}, {-4, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 2, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       MenuSortingValue -> 1485], 
      Cell[
       StyleData["EnvironmentHeader"], 
       CellFrame -> {{False, False}, {False, False}}, 
       CellMargins -> {{76, Inherited}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 6, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       ShowAutoStyles -> False, MenuSortingValue -> 1460, FontSize -> 14, 
       FontWeight -> "Bold", FontVariations -> {"CapsType" -> "SmallCaps"}, 
       Background -> RGBColor[0.89, 0.8, 0.69]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["FormalTextInputFormula"], ShowCellBracket -> Automatic, 
          CellMargins -> {{96, 10}, {2, 2}}, 
          CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
          InitializationCell -> True, CellFrameLabelMargins -> 53, 
          ShowAutoStyles -> False, LimitsPositioningTokens -> {}, 
          MenuSortingValue -> 1470, MenuCommandKey -> "F", FontWeight -> 
          "Bold", FontColor -> RGBColor[0.25, 0.19, 0.2]], 
         Cell[
          StyleData["FormalTextInputFormula", "Layout"], Editable -> False], 
         Cell[
          StyleData["FormalTextInputFormula", "Presentation"], Editable -> 
          False]}, Open]], 
      Cell[
       StyleData["EndEnvironmentMarker"], 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       TextAlignment -> Right, MenuSortingValue -> 1480], 
      Cell[
       StyleData["CloseEnvironment"], Editable -> False, Selectable -> False, 
       CellFrame -> {{False, False}, {False, True}}, ShowCellBracket -> True, 
       CellMargins -> {{76, 6}, {Inherited, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 2, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       MenuSortingValue -> 1485], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["TheoremaInput"], CellFrame -> {{1.5, 0}, {0, 0}}, 
          CellFrameMargins -> {{8, 8}, {3, 4}}, CellFrameColor -> 
          RGBColor[0.486, 0.33, 0.32], ShowAutoStyles -> False, 
          LimitsPositioningTokens -> {}, MenuSortingValue -> 1500, 
          MenuCommandKey -> "I"], 
         Cell[
          StyleData["TheoremaInput", "Layout"], Editable -> False], 
         Cell[
          StyleData["TheoremaInput", "Presentation"], Editable -> False]}, 
        Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["TheoremaCode"], CellFrame -> {{1.5, 0}, {0, 0}}, 
          CellFrameMargins -> {{8, 8}, {3, 4}}, CellFrameColor -> 
          RGBColor[0.486, 0.33, 0.32], ShowAutoStyles -> False, 
          LimitsPositioningTokens -> {}, MenuSortingValue -> 1455, 
          MenuCommandKey -> "C"], 
         Cell[
          StyleData["TheoremaCode", "Layout"], Editable -> False], 
         Cell[
          StyleData["TheoremaCode", "Presentation"], Editable -> False]}, 
        Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Computation"], CellFrame -> {{1.5, 0}, {0, 0}}, 
          ShowCellBracket -> Automatic, CellMargins -> {{76, 10}, {0, 7}}, 
          CellBracketOptions -> {"Color" -> RGBColor[0.23, 0.25, 0.34]}, 
          CellFrameMargins -> {{8, 8}, {3, 4}}, CellFrameColor -> 
          RGBColor[0.486, 0.33, 0.32], ShowAutoStyles -> False, 
          MenuSortingValue -> 1510, FontWeight -> "Bold", FontColor -> 
          RGBColor[0.23, 0.25, 0.34], Background -> 
          RGBColor[0.88, 0.88, 0.92]], 
         Cell[
          StyleData["Computation", "Layout"], Editable -> False], 
         Cell[
          StyleData["Computation", "Presentation"], Editable -> False]}, 
        Open]], 
      Cell[
       StyleData["ComputationResult"], CellFrame -> {{1.5, 0}, {0, 0}}, 
       ShowCellBracket -> Automatic, CellMargins -> {{76, 10}, {0, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.23, 0.25, 0.34]}, 
       CellFrameMargins -> {{8, 8}, {4, 3}}, CellFrameColor -> 
       RGBColor[0.486, 0.33, 0.32], StyleMenuListing -> None, FontColor -> 
       RGBColor[0.23, 0.25, 0.34], Background -> RGBColor[0.88, 0.88, 0.92]], 
      Cell[
       StyleData["ComputationInfo"], ShowCellBracket -> False, 
       CellMargins -> {{120, 10}, {10, 2}}, StyleMenuListing -> None, 
       FontFamily -> $CellContext`Arial], 
      Cell[
       StyleData["ComputationInfoBody"], ShowCellBracket -> False, 
       CellMargins -> {{144, 10}, {10, 2}}, StyleMenuListing -> None], 
      Cell[
       StyleData["ProofInfo"], ShowCellBracket -> False, 
       CellMargins -> {{76, 10}, {6, 6}}, CellFrameLabelMargins -> 15, 
       StyleMenuListing -> None, FontFamily -> "Arial"], 
      Cell[
       StyleData["ProofInfoBody"], ShowCellBracket -> False, 
       CellMargins -> {{100, 10}, {0, 0}}, StyleMenuListing -> None], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["FrameLabel"], LanguageCategory -> None, StyleMenuListing -> 
          None, FontFamily -> "Times New Roman", FontSize -> 10, FontWeight -> 
          "Plain", FontSlant -> "Italic", FontTracking -> "Plain", 
          FontVariations -> {
           "Outline" -> False, "Shadow" -> False, "StrikeThrough" -> False, 
            "Underline" -> False}, FontColor -> RGBColor[0.55, 0.44, 0.42], 
          Background -> None, FontVariations -> {"Masked" -> False}], 
         Cell[
          StyleData["FrameLabel", "Printout"]]}, Open]], 
      Cell[
       StyleData["OpenArchive"], CellFrame -> {{False, False}, {False, True}},
        ShowCellBracket -> Automatic, CellMargins -> {{70, 70}, {0, 10}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellGroupingRules -> {"SectionGrouping", 20}, CellFrameMargins -> 3, 
       CellFrameColor -> RGBColor[0.4, 0.4, 0.4], CellFrameLabelMargins -> 1, 
       DefaultFormatType -> DefaultInputFormatType, ShowAutoStyles -> False, 
       MenuSortingValue -> 1210, FontWeight -> Bold, Background -> 
       RGBColor[0.73, 0.74, 0.84]], 
      Cell[
       StyleData["ArchiveInfo"], ShowCellBracket -> Automatic, 
       CellMargins -> {{70, 70}, {0, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 5, CellFrameLabelMargins -> 3, ShowCellTags -> 
       False, ShowAutoStyles -> False, MenuSortingValue -> 1220, FontWeight -> 
       "Bold", Background -> RGBColor[0.88, 0.88, 0.92]], 
      Cell[
       StyleData["CloseArchive"], 
       CellFrame -> {{False, False}, {True, False}}, ShowCellBracket -> 
       Automatic, CellMargins -> {{0, 0}, {5, 10}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellGroupingRules -> {"SectionGrouping", 25}, CellFrameMargins -> 2, 
       CellFrameColor -> RGBColor[0.4, 0.4, 0.4], TextAlignment -> Center, 
       MenuSortingValue -> 1230, FontFamily -> "Arial Black", FontSize -> 24, 
       FontWeight -> "Plain", FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}, 
       Background -> RGBColor[0.73, 0.65, 0.61]], 
      Cell[
       StyleData["IncludeArchive"], CellFrame -> {{4, 0}, {False, False}}, 
       ShowCellBracket -> Automatic, 
       CellMargins -> {{81, 100}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameColor -> RGBColor[0.55, 0.44, 0.42], 
       CellFrameLabels -> {{"\[RightSkeleton]", None}, {None, None}}, 
       CellFrameLabelMargins -> 25, MenuSortingValue -> 1445], 
      Cell[
       StyleData["ExpressionVariable"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       StyleData["ExpressionABF"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       StyleData["ExpressionMeta"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["AutoParentheses"], Editable -> False, SpanMaxSize -> 3, 
          FontOpacity -> 0.15], 
         Cell[
          StyleData["AutoParentheses", "Layout"], FontOpacity -> 0], 
         Cell[
          StyleData["AutoParentheses", "SlideShow"], FontOpacity -> 0]}, 
        Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Emphasized"], MenuCommandKey -> "E", FontSlant -> 
          "Italic"], 
         Cell[
          StyleData["Emphasized", "SlideShow"], FontSlant -> "Plain", 
          FontColor -> RGBColor[0.36, 0.54, 0.69]]}, Open]], 
      Cell[
       StyleData["GenericButton"], FontFamily -> "Bitstream Charter", 
       FontWeight -> "Plain", FontSlant -> "Italic", FontTracking -> "Plain", 
       FontVariations -> {
        "Outline" -> False, "Shadow" -> False, "StrikeThrough" -> False, 
         "Underline" -> False}, FontColor -> RGBColor[0.23, 0.25, 0.34], 
       FontVariations -> {"Masked" -> False}], 
      Cell[
       StyleData["TheoremaForm"], "TwoByteSyntaxCharacterAutoReplacement" -> 
       True, LineSpacing -> {1.25, 0, 2}, SingleLetterItalics -> True, 
       StyleMenuListing -> None, FontFamily -> "Courier"], 
      Cell[
       StyleData["ComponentActive"], FontFamily -> "Times New Roman"], 
      Cell[
       StyleData["ComponentInactive"], FontFamily -> "Times New Roman", 
       FontOpacity -> 0.4], 
      Cell[
       StyleData["ComponentEmpty"], FontFamily -> "Times New Roman", 
       FontVariations -> {"StrikeThrough" -> True}], 
      Cell[
       StyleData["TabView"], FontFamily -> "Times New Roman", FontSize -> 10], 
      Cell[
       StyleData["TabViewLabel"], FontFamily -> "Times New Roman", FontSize -> 
       10, FontWeight -> "Bold"], 
      Cell[
       StyleData["FormReference", StyleDefinitions -> StyleData["Text"]], 
       FontFamily -> "Times New Roman", FontSize -> 12, FontWeight -> "Plain",
        FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       StyleData["FormReferenceHover", StyleDefinitions -> StyleData["Text"]],
        FontFamily -> "Times New Roman", FontSize -> 12, FontWeight -> 
       "Plain", FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> True}, 
       FontColor -> RGBColor[0.404, 0.43, 0.545]], 
      Cell[
       StyleData["DisplayFormula"], CellMargins -> {{36, 6}, {7, 7}}, 
       ConversionRules :> {
        "HTML" -> {{"", ""}, {"<p class=\"Equation\">", "</p>"}}}, 
       CellHorizontalScrolling -> True, DefaultFormatType -> TraditionalForm, 
       InputAutoReplacements -> {}, TextAlignment -> Center, FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}, 
       ButtonBoxOptions -> {Appearance -> {Automatic, None}}]}, Open]]}, 
  Visible -> False, FrontEndVersion -> 
  "11.3 for Linux x86 (64-bit) (March 6, 2018)", StyleDefinitions -> 
  "StylesheetFormatting.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "\[Element] def"->{
  Cell[5080, 134, 1138, 35, 50, "FormalTextInputFormula",ExpressionUUID->"e62578c8-e0a6-4737-b48a-89439b421a97",
   CellTags->"\[Element] def",
   CellID->464571198]},
 "\[SubsetEqual] def"->{
  Cell[6221, 171, 1492, 49, 54, "FormalTextInputFormula",ExpressionUUID->"962bc855-b87d-497f-828c-d9a1f6dcc296",
   CellTags->"\[SubsetEqual] def",
   CellID->86239678]},
 "\[Subset] def"->{
  Cell[7716, 222, 1451, 48, 50, "FormalTextInputFormula",ExpressionUUID->"8386411c-09c9-4758-8f4c-116b52c47684",
   CellTags->"\[Subset] def",
   CellID->169597205]},
 "Pow def"->{
  Cell[9170, 272, 1148, 36, 51, "FormalTextInputFormula",ExpressionUUID->"903ff798-45d2-49ee-9483-412b4271e487",
   CellTags->"Pow def",
   CellID->15464031]},
 "\[Intersection] def"->{
  Cell[10321, 310, 1271, 39, 54, "FormalTextInputFormula",ExpressionUUID->"bab1ffcc-3be2-40cc-8c31-516831c21a46",
   CellTags->"\[Intersection] def",
   CellID->755145795]},
 "\[Union] def"->{
  Cell[11595, 351, 1251, 39, 54, "FormalTextInputFormula",ExpressionUUID->"331436ef-1768-43c6-a0e6-dd4797dde600",
   CellTags->"\[Union] def",
   CellID->157538211]},
 "Minus set def"->{
  Cell[12849, 392, 1173, 37, 51, "FormalTextInputFormula",ExpressionUUID->"839c14a1-7e40-49b4-866c-82bfae447357",
   CellTags->"Minus set def",
   CellID->375270853]},
 "- set def"->{
  Cell[14025, 431, 1059, 32, 50, "FormalTextInputFormula",ExpressionUUID->"6a8c6292-cb26-4cf3-9299-3dca65b6f0ad",
   CellTags->"- set def",
   CellID->559936356]},
 "UNIV def"->{
  Cell[15087, 465, 825, 23, 51, "FormalTextInputFormula",ExpressionUUID->"1bad0884-8923-430a-9e48-8ee7cdd1b0ec",
   CellTags->"UNIV def",
   CellID->426543191]},
 "EMPTY def"->{
  Cell[15915, 490, 831, 23, 51, "FormalTextInputFormula",ExpressionUUID->"dce4d230-1c21-4c60-8667-799c9b8fd815",
   CellTags->"EMPTY def",
   CellID->217880718]},
 "Insert def"->{
  Cell[16749, 515, 1269, 40, 54, "FormalTextInputFormula",ExpressionUUID->"ce7cf6be-50c3-49f4-8235-e3df79addb62",
   CellTags->"Insert def",
   CellID->394538268]},
 "@ def"->{
  Cell[18021, 557, 1389, 43, 58, "FormalTextInputFormula",ExpressionUUID->"e6744ab5-9386-45ac-a300-eaa9ae40a920",
   CellTags->"@ def",
   CellID->469719290]},
 "SetOf def"->{
  Cell[19413, 602, 1330, 42, 59, "FormalTextInputFormula",ExpressionUUID->"0cad2b9d-2d0a-4eb0-9090-3c81a65f8eca",
   CellTags->"SetOf def",
   CellID->312695955]},
 "\[Element]Collect"->{
  Cell[29094, 872, 1142, 35, 50, "FormalTextInputFormula",ExpressionUUID->"346c1ec5-297b-426b-a764-6d4ceb56a4e8",
   CellTags->"\[Element]Collect",
   CellID->270104195]},
 "Collect ident"->{
  Cell[30239, 909, 1128, 35, 51, "FormalTextInputFormula",ExpressionUUID->"cdd465a6-9272-4aa3-980d-ab6053ec12e3",
   CellTags->"Collect ident",
   CellID->893774303]},
 "Proof|ID:270104195-1"->{
  Cell[31612, 953, 1810, 47, 29, "ProofInfo",ExpressionUUID->"5aad6e8a-a738-407b-bc9f-9ca36dc727e9",
   Evaluatable->True,
   CellTags->{"Proof|ID:270104195-1", "3"},
   CellID->1363674537]},
 "3"->{
  Cell[31612, 953, 1810, 47, 29, "ProofInfo",ExpressionUUID->"5aad6e8a-a738-407b-bc9f-9ca36dc727e9",
   Evaluatable->True,
   CellTags->{"Proof|ID:270104195-1", "3"},
   CellID->1363674537],
  Cell[33425, 1002, 1787, 46, 29, "ProofInfo",ExpressionUUID->"26847eea-8373-4283-8a9d-d1e9558f9cde",
   Evaluatable->True,
   CellTags->{"Proof|ID:893774303-1", "3"},
   CellID->1532689697],
  Cell[71130, 2124, 1808, 47, 29, "ProofInfo",ExpressionUUID->"d29d2a48-174c-424d-be8c-88ec8b4d64fe",
   Evaluatable->True,
   CellTags->{"Proof|ID:333794707-1", "3"},
   CellID->1506794386],
  Cell[72941, 2173, 1818, 47, 29, "ProofInfo",ExpressionUUID->"5027dd32-9d55-40fe-a936-08172c79cad1",
   Evaluatable->True,
   CellTags->{"Proof|ID:833708774-1", "3"},
   CellID->74925535],
  Cell[74762, 2222, 1833, 47, 29, "ProofInfo",ExpressionUUID->"5f4442fc-9b36-4593-b740-1806f36d41e7",
   Evaluatable->True,
   CellTags->{"Proof|ID:1768369467-1", "3"},
   CellID->1488100858],
  Cell[76598, 2271, 1787, 46, 29, "ProofInfo",ExpressionUUID->"46efcca5-5269-4858-be47-88d1f3ecc77b",
   Evaluatable->True,
   CellTags->{"Proof|ID:129158904-1", "3"},
   CellID->1301219521],
  Cell[78388, 2319, 1821, 47, 29, "ProofInfo",ExpressionUUID->"e3d8ee28-9559-4c6b-b535-c0f73713378a",
   Evaluatable->True,
   CellTags->{"Proof|ID:187730850-1", "3"},
   CellID->815869071],
  Cell[92627, 2768, 1774, 46, 29, "ProofInfo",ExpressionUUID->"6e659942-0c6d-405c-9b68-372d2c14a0ed",
   Evaluatable->True,
   CellTags->{"Proof|ID:32648321-1", "3"},
   CellID->925901741],
  Cell[94404, 2816, 1781, 46, 29, "ProofInfo",ExpressionUUID->"c887437d-b8f9-4f4a-abc2-4c844d627714",
   Evaluatable->True,
   CellTags->{"Proof|ID:457314425-1", "3"},
   CellID->1854094453],
  Cell[96188, 2864, 1771, 46, 29, "ProofInfo",ExpressionUUID->"9d0f6fb7-026a-419b-8d81-21560f62f383",
   Evaluatable->True,
   CellTags->{"Proof|ID:9845002-1", "3"},
   CellID->666993796],
  Cell[97962, 2912, 1786, 46, 29, "ProofInfo",ExpressionUUID->"5f1c3502-0db1-4202-8b55-474b927d5261",
   Evaluatable->True,
   CellTags->{"Proof|ID:688844702-1", "3"},
   CellID->894137823],
  Cell[99751, 2960, 1834, 47, 29, "ProofInfo",ExpressionUUID->"c6989024-1dd7-4cbd-bf7d-4edaa6f98d71",
   Evaluatable->True,
   CellTags->{"Proof|ID:53899447-1", "3"},
   CellID->1854942837],
  Cell[101588, 3009, 1833, 47, 29, "ProofInfo",ExpressionUUID->"441a9d35-8e30-4162-9e21-8cf3471755aa",
   Evaluatable->True,
   CellTags->{"Proof|ID:214375017-1", "3"},
   CellID->187880605],
  Cell[103424, 3058, 1833, 47, 29, "ProofInfo",ExpressionUUID->"5ba5cdc6-46d4-4ad2-94e9-7d5f31bc3fe9",
   Evaluatable->True,
   CellTags->{"Proof|ID:705380086-1", "3"},
   CellID->719967367],
  Cell[105260, 3107, 1789, 46, 29, "ProofInfo",ExpressionUUID->"76b67bb5-6cbd-4d7e-8eed-928b19df89d2",
   Evaluatable->True,
   CellTags->{"Proof|ID:122404649-1", "3"},
   CellID->28123779],
  Cell[107052, 3155, 1808, 47, 29, "ProofInfo",ExpressionUUID->"2887fda8-1b65-4980-bb82-685424955125",
   Evaluatable->True,
   CellTags->{"Proof|ID:490505039-1", "3"},
   CellID->1482758530],
  Cell[114767, 3387, 1791, 46, 29, "ProofInfo",ExpressionUUID->"ae3cd3a9-b24b-48b2-8218-78763782452d",
   Evaluatable->True,
   CellTags->{"Proof|ID:109474291-1", "3"},
   CellID->1750098064],
  Cell[116561, 3435, 1746, 45, 29, "ProofInfo",ExpressionUUID->"7c42a337-6cab-46e3-bd77-c0a11197a7dc",
   Evaluatable->True,
   CellTags->{"Proof|ID:54906548-1", "3"},
   CellID->1933668764],
  Cell[118310, 3482, 1808, 47, 29, "ProofInfo",ExpressionUUID->"6f47db7d-2333-492d-87ac-73a3393a968e",
   Evaluatable->True,
   CellTags->{"Proof|ID:289979004-1", "3"},
   CellID->2138873264],
  Cell[120121, 3531, 1784, 46, 29, "ProofInfo",ExpressionUUID->"f789d536-f6cf-4bca-84d6-4532e0bed010",
   Evaluatable->True,
   CellTags->{"Proof|ID:828225892-1", "3"},
   CellID->157843443],
  Cell[121908, 3579, 1813, 47, 29, "ProofInfo",ExpressionUUID->"3cb753de-8224-4302-aee9-4b257cf1136a",
   Evaluatable->True,
   CellTags->{"Proof|ID:1987360397-1", "3"},
   CellID->1029679455],
  Cell[129434, 3792, 1758, 45, 29, "ProofInfo",ExpressionUUID->"f472fc82-802b-410b-868c-d07e017b0268",
   Evaluatable->True,
   CellTags->{"Proof|ID:313591036-1", "3"},
   CellID->1173747547],
  Cell[131195, 3839, 1812, 47, 29, "ProofInfo",ExpressionUUID->"184057ce-28ca-4f86-bd4e-79e86a17715d",
   Evaluatable->True,
   CellTags->{"Proof|ID:223522958-1", "3"},
   CellID->2010038796],
  Cell[133010, 3888, 1788, 46, 29, "ProofInfo",ExpressionUUID->"de3eb27b-5b7d-4ae2-988d-b2653d4d8d31",
   Evaluatable->True,
   CellTags->{"Proof|ID:256313502-1", "3"},
   CellID->741200714],
  Cell[134801, 3936, 1810, 47, 29, "ProofInfo",ExpressionUUID->"824b928c-cb7f-4776-8b3e-3335d33de91a",
   Evaluatable->True,
   CellTags->{"Proof|ID:358676608-1", "3"},
   CellID->1104785395],
  Cell[136614, 3985, 1803, 47, 29, "ProofInfo",ExpressionUUID->"49775606-3bad-4639-ad88-ce4f5704386f",
   Evaluatable->True,
   CellTags->{"Proof|ID:13969540-1", "3"},
   CellID->400553415],
  Cell[146081, 4266, 1782, 46, 29, "ProofInfo",ExpressionUUID->"135e6c9f-6b5f-45fe-bca0-d1d8524dfc01",
   Evaluatable->True,
   CellTags->{"Proof|ID:91383453-1", "3"},
   CellID->1669010862],
  Cell[147866, 4314, 1757, 45, 29, "ProofInfo",ExpressionUUID->"2eff29d4-750b-432e-b353-66e38bf0fded",
   Evaluatable->True,
   CellTags->{"Proof|ID:1005450222-1", "3"},
   CellID->969791405],
  Cell[149626, 4361, 1749, 45, 29, "ProofInfo",ExpressionUUID->"e66761b0-9d29-4441-b48e-3fce10402d48",
   Evaluatable->True,
   CellTags->{"Proof|ID:71998340-1", "3"},
   CellID->1724899203],
  Cell[151378, 4408, 1791, 46, 29, "ProofInfo",ExpressionUUID->"e9f8d063-f619-4ed9-811c-81d57ff90446",
   Evaluatable->True,
   CellTags->{"Proof|ID:135248918-1", "3"},
   CellID->417457554],
  Cell[153172, 4456, 1811, 47, 29, "ProofInfo",ExpressionUUID->"88881ba8-d043-4491-afa7-e2b45acd577c",
   Evaluatable->True,
   CellTags->{"Proof|ID:365430671-1", "3"},
   CellID->1046907733],
  Cell[154986, 4505, 1809, 47, 29, "ProofInfo",ExpressionUUID->"7af83b7e-0a64-42d4-9cfa-8fb1f1515d6b",
   Evaluatable->True,
   CellTags->{"Proof|ID:761522039-1", "3"},
   CellID->1384782616],
  Cell[162127, 4717, 1816, 47, 29, "ProofInfo",ExpressionUUID->"4bb986a8-28da-4979-8914-a7bf05134176",
   Evaluatable->True,
   CellTags->{"Proof|ID:499422188-1", "3"},
   CellID->742874104],
  Cell[163946, 4766, 1785, 46, 29, "ProofInfo",ExpressionUUID->"6b7cdde2-f7e1-4895-96bd-f5dba2fa343f",
   Evaluatable->True,
   CellTags->{"Proof|ID:319529804-1", "3"},
   CellID->692678955],
  Cell[165734, 4814, 1776, 46, 29, "ProofInfo",ExpressionUUID->"1b5041f8-10cf-4c78-8caf-cee02595fafa",
   Evaluatable->True,
   CellTags->{"Proof|ID:7165872-1", "3"},
   CellID->214206169],
  Cell[176015, 5122, 1826, 47, 29, "ProofInfo",ExpressionUUID->"06ccfced-9f01-4ea1-b866-1dc0ce6111d4",
   Evaluatable->True,
   CellTags->{"Proof|ID:299832540-1", "3"},
   CellID->520490995],
  Cell[177844, 5171, 1809, 47, 29, "ProofInfo",ExpressionUUID->"b655e7ec-ba70-4141-b8ce-f4c7c9313eb5",
   Evaluatable->True,
   CellTags->{"Proof|ID:742723078-1", "3"},
   CellID->1206322670],
  Cell[179656, 5220, 1811, 47, 29, "ProofInfo",ExpressionUUID->"999e3753-e878-46ae-816d-c1c14791130d",
   Evaluatable->True,
   CellTags->{"Proof|ID:196903271-1", "3"},
   CellID->1182674228],
  Cell[181470, 5269, 1811, 47, 29, "ProofInfo",ExpressionUUID->"263b3d20-5c8b-4d42-8cd7-bb5ecf6bba6c",
   Evaluatable->True,
   CellTags->{"Proof|ID:214165647-1", "3"},
   CellID->1685360901],
  Cell[183284, 5318, 1802, 47, 29, "ProofInfo",ExpressionUUID->"e8ca3e41-bda4-4b23-8e4b-61f126bed335",
   Evaluatable->True,
   CellTags->{"Proof|ID:31355863-1", "3"},
   CellID->634954845],
  Cell[189684, 5498, 1821, 47, 29, "ProofInfo",ExpressionUUID->"c9408df1-52d0-43ba-9c83-f86df76887a4",
   Evaluatable->True,
   CellTags->{"Proof|ID:330976770-1", "3"},
   CellID->1901417663],
  Cell[191508, 5547, 1821, 47, 29, "ProofInfo",ExpressionUUID->"cd544126-02a6-4a0d-b5ab-040100796126",
   Evaluatable->True,
   CellTags->{"Proof|ID:567877347-1", "3"},
   CellID->1604414518],
  Cell[193332, 5596, 1825, 47, 29, "ProofInfo",ExpressionUUID->"067b7963-94d6-4a20-aae6-bc77ea005db6",
   Evaluatable->True,
   CellTags->{"Proof|ID:370774399-1", "3"},
   CellID->1352341786],
  Cell[203106, 5898, 1806, 47, 29, "ProofInfo",ExpressionUUID->"5d01d301-0853-41ec-8b5b-f654db10d92f",
   Evaluatable->True,
   CellTags->{"Proof|ID:24302859-1", "3"},
   CellID->910429301],
  Cell[204915, 5947, 1782, 46, 29, "ProofInfo",ExpressionUUID->"58cda68d-cdf0-4074-9690-d94a7ad8e87b",
   Evaluatable->True,
   CellTags->{"Proof|ID:204712299-1", "3"},
   CellID->2106560278],
  Cell[206700, 5995, 1781, 46, 29, "ProofInfo",ExpressionUUID->"5ab74a01-dcb4-4410-adbb-2e6f4bc3d887",
   Evaluatable->True,
   CellTags->{"Proof|ID:118362674-1", "3"},
   CellID->430856106],
  Cell[208484, 6043, 1808, 47, 29, "ProofInfo",ExpressionUUID->"c438dae3-6253-4324-8a41-6d2ae06ea6ec",
   Evaluatable->True,
   CellTags->{"Proof|ID:71686966-1", "3"},
   CellID->634618502],
  Cell[210295, 6092, 1765, 45, 29, "ProofInfo",ExpressionUUID->"7192f13d-fafc-45a0-9eba-e5628635aa2e",
   Evaluatable->True,
   CellTags->{"Proof|ID:734373493-1", "3"},
   CellID->1988227689],
  Cell[216618, 6269, 1792, 46, 29, "ProofInfo",ExpressionUUID->"4273e686-36d1-479a-869b-b04747afdf48",
   Evaluatable->True,
   CellTags->{"Proof|ID:776879610-1", "3"},
   CellID->1308387989],
  Cell[218413, 6317, 1791, 46, 29, "ProofInfo",ExpressionUUID->"a0289919-612a-43d3-83be-5e58d75618f3",
   Evaluatable->True,
   CellTags->{"Proof|ID:463603332-1", "3"},
   CellID->283700494],
  Cell[220207, 6365, 1789, 46, 29, "ProofInfo",ExpressionUUID->"bacbec27-f6b0-4574-94e4-76900e1b7a43",
   Evaluatable->True,
   CellTags->{"Proof|ID:672172372-1", "3"},
   CellID->563063133],
  Cell[232109, 6736, 1765, 45, 29, "ProofInfo",ExpressionUUID->"413f783b-5cb5-476c-8d17-7d845255998a",
   Evaluatable->True,
   CellTags->{"Proof|ID:540115919-1", "3"},
   CellID->2034290086],
  Cell[233877, 6783, 1784, 46, 29, "ProofInfo",ExpressionUUID->"0e3f881b-3ec2-4218-8f49-710e98423e7e",
   Evaluatable->True,
   CellTags->{"Proof|ID:201090499-1", "3"},
   CellID->1196091729],
  Cell[235664, 6831, 1753, 45, 29, "ProofInfo",ExpressionUUID->"fae8ecb6-0379-4df0-8561-e4c239de4b21",
   Evaluatable->True,
   CellTags->{"Proof|ID:17112112-1", "3"},
   CellID->1855793195],
  Cell[237420, 6878, 1759, 45, 29, "ProofInfo",ExpressionUUID->"ff98ca43-6f95-4017-83fe-5e58556b956d",
   Evaluatable->True,
   CellTags->{"Proof|ID:590579378-1", "3"},
   CellID->87366359],
  Cell[239182, 6925, 1748, 45, 29, "ProofInfo",ExpressionUUID->"497b9d54-4fde-4ed0-af44-f8d1fb40d21a",
   Evaluatable->True,
   CellTags->{"Proof|ID:984086-1", "3"},
   CellID->738044893],
  Cell[240933, 6972, 1758, 45, 29, "ProofInfo",ExpressionUUID->"9ab0c91e-bd51-4993-aa37-27e47b2db8cf",
   Evaluatable->True,
   CellTags->{"Proof|ID:128535443-1", "3"},
   CellID->187234111],
  Cell[242694, 7019, 1826, 47, 29, "ProofInfo",ExpressionUUID->"526b809c-7826-439b-ac09-9eaa35adbac7",
   Evaluatable->True,
   CellTags->{"Proof|ID:679084152-1", "3"},
   CellID->887782668],
  Cell[249451, 7212, 1792, 46, 29, "ProofInfo",ExpressionUUID->"94ee77de-dbad-4d4e-97d6-1755cd58f119",
   Evaluatable->True,
   CellTags->{"Proof|ID:917574097-1", "3"},
   CellID->1547795340],
  Cell[251246, 7260, 1808, 47, 29, "ProofInfo",ExpressionUUID->"19482a26-cb3e-4e1f-af6d-49608b2ed75c",
   Evaluatable->True,
   CellTags->{"Proof|ID:34104065-1", "3"},
   CellID->838116264],
  Cell[253057, 7309, 1851, 47, 29, "ProofInfo",ExpressionUUID->"9127f290-b8cf-4129-98e6-27563e725e5a",
   Evaluatable->True,
   CellTags->{"Proof|ID:168218769-1", "3"},
   CellID->1909735428],
  Cell[265606, 7705, 1809, 47, 29, "ProofInfo",ExpressionUUID->"69efc2d1-78ff-494c-9afc-e15c84128c04",
   Evaluatable->True,
   CellTags->{"Proof|ID:833542610-1", "3"},
   CellID->1603093206],
  Cell[267418, 7754, 1763, 45, 29, "ProofInfo",ExpressionUUID->"e9466e7d-028a-4117-ac46-65b9a92ddf8e",
   Evaluatable->True,
   CellTags->{"Proof|ID:265974781-1", "3"},
   CellID->1283043907],
  Cell[269184, 7801, 1764, 45, 29, "ProofInfo",ExpressionUUID->"9dffdc4b-0a67-4869-8254-6e3cf0381dfa",
   Evaluatable->True,
   CellTags->{"Proof|ID:369554344-1", "3"},
   CellID->111935385],
  Cell[270951, 7848, 1762, 45, 29, "ProofInfo",ExpressionUUID->"53cab25a-3a17-4a43-97f3-cc2b778e43fc",
   Evaluatable->True,
   CellTags->{"Proof|ID:276809760-1", "3"},
   CellID->591244542],
  Cell[272716, 7895, 1811, 47, 29, "ProofInfo",ExpressionUUID->"893df1bb-a5ff-4f0d-a86c-b9a2a51b065b",
   Evaluatable->True,
   CellTags->{"Proof|ID:794800661-1", "3"},
   CellID->1050155476],
  Cell[274530, 7944, 1761, 45, 29, "ProofInfo",ExpressionUUID->"8908cda5-38e0-4ff0-806e-c21baafb2a94",
   Evaluatable->True,
   CellTags->{"Proof|ID:752025544-1", "3"},
   CellID->1183293999],
  Cell[276294, 7991, 1775, 46, 29, "ProofInfo",ExpressionUUID->"acf349b9-ba5d-40be-9cc8-8cf74e4537e3",
   Evaluatable->True,
   CellTags->{"Proof|ID:60042161-1", "3"},
   CellID->807514983],
  Cell[285945, 8285, 1821, 47, 29, "ProofInfo",ExpressionUUID->"033295b6-03ea-4743-b02f-9a029d18a564",
   Evaluatable->True,
   CellTags->{"Proof|ID:499325579-1", "3"},
   CellID->1989694051],
  Cell[287769, 8334, 1816, 47, 29, "ProofInfo",ExpressionUUID->"74457cd9-9415-4a5b-a6ff-c4224aff06d1",
   Evaluatable->True,
   CellTags->{"Proof|ID:89181006-1", "3"},
   CellID->169658549],
  Cell[289588, 8383, 1809, 47, 29, "ProofInfo",ExpressionUUID->"539ab388-aaaf-402a-84e8-d34f4c88e778",
   Evaluatable->True,
   CellTags->{"Proof|ID:17294591-1", "3"},
   CellID->94829629],
  Cell[291400, 8432, 1828, 47, 29, "ProofInfo",ExpressionUUID->"a019316d-6449-4592-a608-6892d07b6388",
   Evaluatable->True,
   CellTags->{"Proof|ID:1726053781-1", "3"},
   CellID->1321149923],
  Cell[293231, 8481, 1824, 47, 29, "ProofInfo",ExpressionUUID->"72ddb486-2189-446a-b4f0-dccfaf5bf1dc",
   Evaluatable->True,
   CellTags->{"Proof|ID:140408834-1", "3"},
   CellID->662871545],
  Cell[299598, 8669, 1787, 46, 29, "ProofInfo",ExpressionUUID->"24b89c28-f6f2-4a2d-8f6c-d7faa64ba567",
   Evaluatable->True,
   CellTags->{"Proof|ID:1706541997-1", "3"},
   CellID->1610428281],
  Cell[301388, 8717, 1785, 46, 29, "ProofInfo",ExpressionUUID->"c797b61f-6a99-4d82-9375-e86eb679b1f6",
   Evaluatable->True,
   CellTags->{"Proof|ID:1946555487-1", "3"},
   CellID->44899502],
  Cell[303176, 8765, 1811, 47, 29, "ProofInfo",ExpressionUUID->"2d77d2ad-50ea-45a7-8a70-0483b95c475c",
   Evaluatable->True,
   CellTags->{"Proof|ID:1642372132-1", "3"},
   CellID->259575914],
  Cell[309009, 8943, 1824, 47, 29, "ProofInfo",ExpressionUUID->"b866f190-0698-4884-9f66-30d4006e9fd2",
   Evaluatable->True,
   CellTags->{"Proof|ID:574835025-1", "3"},
   CellID->991759232],
  Cell[310836, 8992, 1824, 47, 29, "ProofInfo",ExpressionUUID->"61354163-ddd7-4786-856c-489f55e87f87",
   Evaluatable->True,
   CellTags->{"Proof|ID:316263559-1", "3"},
   CellID->811999371],
  Cell[323599, 9402, 1759, 45, 29, "ProofInfo",ExpressionUUID->"b5119a56-9b6b-4dc9-baca-4b038f13e49d",
   Evaluatable->True,
   CellTags->{"Proof|ID:1595981918-1", "3"},
   CellID->638235692],
  Cell[325361, 9449, 1788, 46, 29, "ProofInfo",ExpressionUUID->"8db8df19-fdf9-4fa7-b82c-0b2e4f83d460",
   Evaluatable->True,
   CellTags->{"Proof|ID:796685651-1", "3"},
   CellID->1222338081],
  Cell[327152, 9497, 1754, 45, 29, "ProofInfo",ExpressionUUID->"6d754d5f-fbc5-4f0b-9d25-e0f40502431f",
   Evaluatable->True,
   CellTags->{"Proof|ID:1022035164-1", "3"},
   CellID->1120566427],
  Cell[328909, 9544, 1751, 45, 29, "ProofInfo",ExpressionUUID->"0d43d638-054b-4e7b-a14f-d856c779b0cc",
   Evaluatable->True,
   CellTags->{"Proof|ID:114344294-1", "3"},
   CellID->1155434313],
  Cell[330663, 9591, 1787, 46, 29, "ProofInfo",ExpressionUUID->"fc5a6e81-1d94-4b7a-9f62-d18cc426bf95",
   Evaluatable->True,
   CellTags->{"Proof|ID:1795699175-1", "3"},
   CellID->1210186748],
  Cell[332453, 9639, 1758, 45, 29, "ProofInfo",ExpressionUUID->"c828c5b9-b548-4ffe-a01b-69ead7885f9f",
   Evaluatable->True,
   CellTags->{"Proof|ID:724997673-1", "3"},
   CellID->637413599],
  Cell[334214, 9686, 1765, 45, 29, "ProofInfo",ExpressionUUID->"ed4bf583-b592-4cce-934f-774b532b3b39",
   Evaluatable->True,
   CellTags->{"Proof|ID:803636397-1", "3"},
   CellID->1659057562],
  Cell[341753, 9912, 1813, 47, 29, "ProofInfo",ExpressionUUID->"57da981f-f479-4764-8da0-d6a7a3c080ec",
   Evaluatable->True,
   CellTags->{"Proof|ID:1322094901-1", "3"},
   CellID->892892992],
  Cell[343569, 9961, 1813, 47, 29, "ProofInfo",ExpressionUUID->"7588b18e-aa27-40ac-8f3f-f9e9ae729362",
   Evaluatable->True,
   CellTags->{"Proof|ID:1691291058-1", "3"},
   CellID->903150912],
  Cell[345385, 10010, 1812, 47, 29, "ProofInfo",ExpressionUUID->"85d59c5c-1a6e-4476-8410-ecd019893f54",
   Evaluatable->True,
   CellTags->{"Proof|ID:1830553618-1", "3"},
   CellID->1214598633],
  Cell[347200, 10059, 1751, 45, 29, "ProofInfo",ExpressionUUID->"16659fac-0e9b-4219-a757-d5db7b81bc81",
   Evaluatable->True,
   CellTags->{"Proof|ID:398205279-1", "3"},
   CellID->1641967303],
  Cell[355083, 10298, 1765, 45, 29, "ProofInfo",ExpressionUUID->"6e01bd25-a1a6-443d-b697-ce5e66313c66",
   Evaluatable->True,
   CellTags->{"Proof|ID:686090881-1", "3"},
   CellID->1329697308],
  Cell[356851, 10345, 1760, 45, 29, "ProofInfo",ExpressionUUID->"8e614f57-fd84-4b20-aada-099cb8f5fd99",
   Evaluatable->True,
   CellTags->{"Proof|ID:1169658618-1", "3"},
   CellID->65583140],
  Cell[358614, 10392, 1762, 45, 29, "ProofInfo",ExpressionUUID->"ec01b3f4-f18b-473a-adaa-b64c4b1794c1",
   Evaluatable->True,
   CellTags->{"Proof|ID:1129746621-1", "3"},
   CellID->1205185338],
  Cell[360379, 10439, 1813, 47, 29, "ProofInfo",ExpressionUUID->"796ac993-5d77-429a-8ec4-94a6354c467e",
   Evaluatable->True,
   CellTags->{"Proof|ID:1805361717-1", "3"},
   CellID->774665428],
  Cell[367772, 10666, 1791, 46, 29, "ProofInfo",ExpressionUUID->"62e40651-f209-4327-b387-490f4c46bb3d",
   Evaluatable->True,
   CellTags->{"Proof|ID:718424401-1", "3"},
   CellID->961492024],
  Cell[369566, 10714, 1761, 45, 29, "ProofInfo",ExpressionUUID->"8842b459-89de-4cca-be91-b8c2e9a5746a",
   Evaluatable->True,
   CellTags->{"Proof|ID:783246906-1", "3"},
   CellID->1483483416],
  Cell[371330, 10761, 1764, 45, 29, "ProofInfo",ExpressionUUID->"3432589c-d2fc-4a23-a0c8-b28e516663cd",
   Evaluatable->True,
   CellTags->{"Proof|ID:1984656332-1", "3"},
   CellID->1243944656],
  Cell[386423, 11243, 1816, 47, 29, "ProofInfo",ExpressionUUID->"90ce6af8-8154-479f-a67a-eb82f1369496",
   Evaluatable->True,
   CellTags->{"Proof|ID:314660242-1", "3"},
   CellID->462247040],
  Cell[388242, 11292, 1784, 46, 29, "ProofInfo",ExpressionUUID->"23123a44-70d0-4b7e-af9d-66586f2c7c3b",
   Evaluatable->True,
   CellTags->{"Proof|ID:222426804-1", "3"},
   CellID->1447843469],
  Cell[390029, 11340, 1817, 47, 29, "ProofInfo",ExpressionUUID->"216bcc67-94ae-452d-9f9d-fb1577639bae",
   Evaluatable->True,
   CellTags->{"Proof|ID:1237361888-1", "3"},
   CellID->779088903],
  Cell[391849, 11389, 1784, 46, 29, "ProofInfo",ExpressionUUID->"5156a1b2-c4e6-447c-8758-161115f33ced",
   Evaluatable->True,
   CellTags->{"Proof|ID:165006998-1", "3"},
   CellID->1390818865],
  Cell[393636, 11437, 1784, 46, 29, "ProofInfo",ExpressionUUID->"627a6140-2e49-4960-8e4a-725e778a5b7a",
   Evaluatable->True,
   CellTags->{"Proof|ID:1855260115-1", "3"},
   CellID->859972040],
  Cell[395423, 11485, 1792, 46, 29, "ProofInfo",ExpressionUUID->"d35e54e6-1cd4-4c97-a6a7-67a76bf2b9ca",
   Evaluatable->True,
   CellTags->{"Proof|ID:985026514-1", "3"},
   CellID->2146486548],
  Cell[397218, 11533, 1788, 46, 29, "ProofInfo",ExpressionUUID->"3f1308e2-2bf7-43a6-b603-e554f2db72a9",
   Evaluatable->True,
   CellTags->{"Proof|ID:636400335-1", "3"},
   CellID->1733058477],
  Cell[399009, 11581, 1790, 46, 29, "ProofInfo",ExpressionUUID->"43fc8102-da0a-4d32-8477-1d0fb499a635",
   Evaluatable->True,
   CellTags->{"Proof|ID:859120159-1", "3"},
   CellID->1491492338],
  Cell[405817, 11789, 1808, 47, 29, "ProofInfo",ExpressionUUID->"533d16e0-ce09-469a-92e6-88dd693c1a07",
   Evaluatable->True,
   CellTags->{"Proof|ID:485185733-1", "3"},
   CellID->851152370],
  Cell[407628, 11838, 1830, 47, 29, "ProofInfo",ExpressionUUID->"d64e40de-107c-465e-a04b-e0d1316fc389",
   Evaluatable->True,
   CellTags->{"Proof|ID:775376237-1", "3"},
   CellID->497842656],
  Cell[409461, 11887, 1824, 47, 29, "ProofInfo",ExpressionUUID->"3dea71aa-c69b-4215-9524-803e9c59e9f2",
   Evaluatable->True,
   CellTags->{"Proof|ID:978065715-1", "3"},
   CellID->118302286],
  Cell[424874, 12369, 1783, 46, 29, "ProofInfo",ExpressionUUID->"711e6551-2491-4f23-a69e-6ba95e9ee3c5",
   Evaluatable->True,
   CellTags->{"Proof|ID:988950467-1", "3"},
   CellID->207220836],
  Cell[426660, 12417, 1823, 47, 29, "ProofInfo",ExpressionUUID->"ce02b660-2efc-48d5-a410-83f64d3cb0a4",
   Evaluatable->True,
   CellTags->{"Proof|ID:1780572558-1", "3"},
   CellID->709620784],
  Cell[428486, 12466, 1817, 47, 29, "ProofInfo",ExpressionUUID->"a593fce4-239a-48ba-a634-0803bc3ce5f5",
   Evaluatable->True,
   CellTags->{"Proof|ID:1795020389-1", "3"},
   CellID->201322071],
  Cell[430306, 12515, 1791, 46, 29, "ProofInfo",ExpressionUUID->"1c164155-3908-47ea-80ff-28bb818b47e3",
   Evaluatable->True,
   CellTags->{"Proof|ID:2125656893-1", "3"},
   CellID->1679122200],
  Cell[432100, 12563, 1789, 46, 29, "ProofInfo",ExpressionUUID->"fdfb52c1-1f7b-42b0-8e1f-bfb7551b6086",
   Evaluatable->True,
   CellTags->{"Proof|ID:742456464-1", "3"},
   CellID->292466109],
  Cell[433892, 12611, 1786, 46, 29, "ProofInfo",ExpressionUUID->"0017ff7e-6593-4c7d-a1d7-1bfa18dabc62",
   Evaluatable->True,
   CellTags->{"Proof|ID:1038355606-1", "3"},
   CellID->600513012],
  Cell[435681, 12659, 1786, 46, 29, "ProofInfo",ExpressionUUID->"864a6093-7b93-41ad-b279-fcdce277baa9",
   Evaluatable->True,
   CellTags->{"Proof|ID:657964489-1", "3"},
   CellID->1781110406],
  Cell[437470, 12707, 1834, 47, 29, "ProofInfo",ExpressionUUID->"a1dd04b2-c9d9-4f5c-a60f-43df42997afa",
   Evaluatable->True,
   CellTags->{"Proof|ID:1602428504-1", "3"},
   CellID->1228710860],
  Cell[439307, 12756, 1792, 46, 29, "ProofInfo",ExpressionUUID->"7df4f81d-c45a-43a2-a483-44ffc18f92a3",
   Evaluatable->True,
   CellTags->{"Proof|ID:621646007-1", "3"},
   CellID->1551927440],
  Cell[441102, 12804, 1792, 46, 29, "ProofInfo",ExpressionUUID->"5a84cad7-a764-4fb6-b914-1258f0a32f54",
   Evaluatable->True,
   CellTags->{"Proof|ID:656967390-1", "3"},
   CellID->1616332482],
  Cell[467098, 13606, 1821, 47, 29, "ProofInfo",ExpressionUUID->"cf2a0b05-947e-4e08-b681-dbb9bb955bbf",
   Evaluatable->True,
   CellTags->{"Proof|ID:125816224-1", "3"},
   CellID->1628613977],
  Cell[468922, 13655, 1823, 47, 29, "ProofInfo",ExpressionUUID->"36472be4-f0d7-4752-bf1f-5e56ab4cfee6",
   Evaluatable->True,
   CellTags->{"Proof|ID:848809596-1", "3"},
   CellID->2030772045],
  Cell[470748, 13704, 1816, 47, 29, "ProofInfo",ExpressionUUID->"e0bc9de7-8d38-449f-8c01-166784414f07",
   Evaluatable->True,
   CellTags->{"Proof|ID:147215930-1", "3"},
   CellID->548669029],
  Cell[472567, 13753, 1819, 47, 29, "ProofInfo",ExpressionUUID->"0c6b49b0-74e6-4b95-a486-249347bf00e4",
   Evaluatable->True,
   CellTags->{"Proof|ID:902394792-1", "3"},
   CellID->1223279036],
  Cell[474389, 13802, 1818, 47, 29, "ProofInfo",ExpressionUUID->"c9ae36de-51e6-4ab1-af37-a9c0f0d09f5f",
   Evaluatable->True,
   CellTags->{"Proof|ID:568250332-1", "3"},
   CellID->868040367],
  Cell[476210, 13851, 1825, 47, 29, "ProofInfo",ExpressionUUID->"263a06fe-fc74-4a56-8db9-29590e7b58ff",
   Evaluatable->True,
   CellTags->{"Proof|ID:1826338130-1", "3"},
   CellID->679262902],
  Cell[478038, 13900, 1826, 47, 29, "ProofInfo",ExpressionUUID->"e1bfc212-a53d-43e0-b061-4f254f379381",
   Evaluatable->True,
   CellTags->{"Proof|ID:1588902597-1", "3"},
   CellID->1210453596],
  Cell[479867, 13949, 1814, 47, 29, "ProofInfo",ExpressionUUID->"1a5c661b-1557-4af5-b166-96572bd68d8c",
   Evaluatable->True,
   CellTags->{"Proof|ID:1253019475-1", "3"},
   CellID->1968362112],
  Cell[481684, 13998, 1814, 47, 29, "ProofInfo",ExpressionUUID->"3cc072c7-64ff-40a0-8749-2ddcf91e538c",
   Evaluatable->True,
   CellTags->{"Proof|ID:1274251332-1", "3"},
   CellID->2131093972],
  Cell[483501, 14047, 1831, 47, 29, "ProofInfo",ExpressionUUID->"8a174da6-9259-4456-a888-bca8c161fadd",
   Evaluatable->True,
   CellTags->{"Proof|ID:903302990-1", "3"},
   CellID->72270293],
  Cell[485335, 14096, 1824, 47, 29, "ProofInfo",ExpressionUUID->"677ec8e8-79c5-4849-ab27-fd71ffc1c532",
   Evaluatable->True,
   CellTags->{"Proof|ID:920640376-1", "3"},
   CellID->321131777],
  Cell[487162, 14145, 1818, 47, 29, "ProofInfo",ExpressionUUID->"777b4649-e951-4bda-9678-a8b639463f5b",
   Evaluatable->True,
   CellTags->{"Proof|ID:1023611974-1", "3"},
   CellID->1877569687],
  Cell[488983, 14194, 1818, 47, 29, "ProofInfo",ExpressionUUID->"8338a43c-ea58-4e7d-85ec-4622d938afa3",
   Evaluatable->True,
   CellTags->{"Proof|ID:1195769099-1", "3"},
   CellID->1618369204],
  Cell[490804, 14243, 1836, 47, 29, "ProofInfo",ExpressionUUID->"f1380f81-fa44-4763-b4e5-077e4ac37ebd",
   Evaluatable->True,
   CellTags->{"Proof|ID:56424478-1", "3"},
   CellID->400332023],
  Cell[492643, 14292, 1848, 47, 29, "ProofInfo",ExpressionUUID->"f946aa9b-4bc5-4b6b-a6d6-94285e3a76b4",
   Evaluatable->True,
   CellTags->{"Proof|ID:1624124987-1", "3"},
   CellID->1734538043],
  Cell[494494, 14341, 1831, 47, 29, "ProofInfo",ExpressionUUID->"edb892ab-bd7f-474d-8831-ba95ca7c0306",
   Evaluatable->True,
   CellTags->{"Proof|ID:158694025-1", "3"},
   CellID->1315875576],
  Cell[496328, 14390, 1837, 47, 29, "ProofInfo",ExpressionUUID->"49d48a11-4604-4d84-b391-5e5974a2445a",
   Evaluatable->True,
   CellTags->{"Proof|ID:1837534884-1", "3"},
   CellID->978925165],
  Cell[498168, 14439, 1818, 47, 29, "ProofInfo",ExpressionUUID->"4ac23cf1-4b24-4d31-90fa-4a0775d734ae",
   Evaluatable->True,
   CellTags->{"Proof|ID:229432372-1", "3"},
   CellID->387372574],
  Cell[499989, 14488, 1821, 47, 29, "ProofInfo",ExpressionUUID->"812c3dd1-7e84-453c-b984-6de8b7f632e4",
   Evaluatable->True,
   CellTags->{"Proof|ID:2031161567-1", "3"},
   CellID->895002178],
  Cell[522187, 15175, 1791, 46, 29, "ProofInfo",ExpressionUUID->"eb249535-8173-40d6-bbdb-2c2aae352b07",
   Evaluatable->True,
   CellTags->{"Proof|ID:510851566-1", "3"},
   CellID->365686177],
  Cell[523981, 15223, 1809, 47, 29, "ProofInfo",ExpressionUUID->"f296832d-3220-4ad0-9d80-345f79408b1f",
   Evaluatable->True,
   CellTags->{"Proof|ID:961990306-1", "3"},
   CellID->1321288291],
  Cell[525793, 15272, 1790, 46, 29, "ProofInfo",ExpressionUUID->"e2b5c4d8-ca86-4739-8add-11abf118bb58",
   Evaluatable->True,
   CellTags->{"Proof|ID:2140682664-1", "3"},
   CellID->678754593],
  Cell[527586, 15320, 1793, 46, 29, "ProofInfo",ExpressionUUID->"5592b9fd-2693-411b-becf-a4a020ad57be",
   Evaluatable->True,
   CellTags->{"Proof|ID:1655518259-1", "3"},
   CellID->1904743954],
  Cell[529382, 15368, 1790, 46, 29, "ProofInfo",ExpressionUUID->"3fc23a3b-63ed-4e0f-9760-2469f4d0179b",
   Evaluatable->True,
   CellTags->{"Proof|ID:945600630-1", "3"},
   CellID->1060575105],
  Cell[531175, 15416, 1808, 47, 29, "ProofInfo",ExpressionUUID->"a45e0c3a-1b21-41cc-887a-dad678f3e9fe",
   Evaluatable->True,
   CellTags->{"Proof|ID:837689454-1", "3"},
   CellID->363045913],
  Cell[532986, 15465, 1810, 47, 29, "ProofInfo",ExpressionUUID->"ccf3a6e2-23b5-4cd1-a76c-3ffdf84cc204",
   Evaluatable->True,
   CellTags->{"Proof|ID:1727788553-1", "3"},
   CellID->56107910],
  Cell[534799, 15514, 1782, 46, 29, "ProofInfo",ExpressionUUID->"e1d8893c-5ca6-4930-940f-940100ee72b4",
   Evaluatable->True,
   CellTags->{"Proof|ID:985635433-1", "3"},
   CellID->1189519127],
  Cell[536584, 15562, 1784, 46, 29, "ProofInfo",ExpressionUUID->"84574fea-7e8d-4451-afb5-958ee98ab045",
   Evaluatable->True,
   CellTags->{"Proof|ID:1158821232-1", "3"},
   CellID->772575899],
  Cell[538371, 15610, 1789, 46, 29, "ProofInfo",ExpressionUUID->"57de62df-9c50-4468-8a9b-27db3a02f067",
   Evaluatable->True,
   CellTags->{"Proof|ID:1958752872-1", "3"},
   CellID->1073197046],
  Cell[540163, 15658, 1785, 46, 29, "ProofInfo",ExpressionUUID->"091f7b25-0bd4-44cb-ae0f-90617fc64236",
   Evaluatable->True,
   CellTags->{"Proof|ID:715721472-1", "3"},
   CellID->645311516],
  Cell[541951, 15706, 1790, 46, 29, "ProofInfo",ExpressionUUID->"f96d0823-3864-4900-af03-5095e0ae6fd2",
   Evaluatable->True,
   CellTags->{"Proof|ID:727949230-1", "3"},
   CellID->1612445395],
  Cell[543744, 15754, 1792, 46, 29, "ProofInfo",ExpressionUUID->"e1b4be1f-45b6-489f-ae2b-a23d03f34073",
   Evaluatable->True,
   CellTags->{"Proof|ID:1429638428-1", "3"},
   CellID->788463628],
  Cell[545539, 15802, 1845, 47, 29, "ProofInfo",ExpressionUUID->"10ae46d4-043b-4352-a8ed-04644b597001",
   Evaluatable->True,
   CellTags->{"Proof|ID:1116053495-1", "3"},
   CellID->826786679],
  Cell[547387, 15851, 1848, 47, 29, "ProofInfo",ExpressionUUID->"5e24b08f-bb6e-4cbf-80ba-ba0b65e7b37f",
   Evaluatable->True,
   CellTags->{"Proof|ID:1097000890-1", "3"},
   CellID->1930524663],
  Cell[549238, 15900, 1816, 47, 29, "ProofInfo",ExpressionUUID->"4e864b7c-43bd-41bd-b35b-e80644e4ba7c",
   Evaluatable->True,
   CellTags->{"Proof|ID:1016692823-1", "3"},
   CellID->1014581777],
  Cell[551057, 15949, 1820, 47, 29, "ProofInfo",ExpressionUUID->"5bcae50a-1782-48b2-bce6-0bda8d2274c8",
   Evaluatable->True,
   CellTags->{"Proof|ID:997515848-1", "3"},
   CellID->273440836],
  Cell[568257, 16474, 1809, 47, 29, "ProofInfo",ExpressionUUID->"335661a6-db39-426e-a2d5-3a88bbf119ee",
   Evaluatable->True,
   CellTags->{"Proof|ID:36426955-1", "3"},
   CellID->1087560852],
  Cell[570069, 16523, 1817, 47, 29, "ProofInfo",ExpressionUUID->"790c583f-b1fd-4cae-b4c7-3b7f8318ee81",
   Evaluatable->True,
   CellTags->{"Proof|ID:1941170909-1", "3"},
   CellID->769645528],
  Cell[571889, 16572, 1817, 47, 29, "ProofInfo",ExpressionUUID->"978d16d3-2a8d-48c1-8119-e556b594fdbe",
   Evaluatable->True,
   CellTags->{"Proof|ID:980585996-1", "3"},
   CellID->1868666534],
  Cell[573709, 16621, 1820, 47, 29, "ProofInfo",ExpressionUUID->"55ba8ff4-e771-4565-ad2a-a535645579da",
   Evaluatable->True,
   CellTags->{"Proof|ID:1324045296-1", "3"},
   CellID->1667315755],
  Cell[575532, 16670, 1821, 47, 29, "ProofInfo",ExpressionUUID->"4270ef55-e968-4e30-88f5-189f15367a47",
   Evaluatable->True,
   CellTags->{"Proof|ID:1636712813-1", "3"},
   CellID->193380618],
  Cell[577356, 16719, 1816, 47, 29, "ProofInfo",ExpressionUUID->"ff8108f1-b1f1-4f1c-be7e-36f13ea484f5",
   Evaluatable->True,
   CellTags->{"Proof|ID:1334137643-1", "3"},
   CellID->95672954],
  Cell[579175, 16768, 1831, 47, 29, "ProofInfo",ExpressionUUID->"6c8fd24a-5787-423d-a5a0-9ee8249e79b1",
   Evaluatable->True,
   CellTags->{"Proof|ID:1036693907-1", "3"},
   CellID->197559568],
  Cell[581009, 16817, 1810, 47, 29, "ProofInfo",ExpressionUUID->"c77a6221-7b61-4579-83d1-c9ad096f840c",
   Evaluatable->True,
   CellTags->{"Proof|ID:204929684-1", "3"},
   CellID->459265611],
  Cell[582822, 16866, 1817, 47, 29, "ProofInfo",ExpressionUUID->"fcbb79cd-640b-433d-aca6-6f01bc0cd323",
   Evaluatable->True,
   CellTags->{"Proof|ID:849790927-1", "3"},
   CellID->1062590287],
  Cell[584642, 16915, 1795, 46, 29, "ProofInfo",ExpressionUUID->"03e8795e-cd22-4b6b-b60f-6a5c21b1baf5",
   Evaluatable->True,
   CellTags->{"Proof|ID:1609622809-1", "3"},
   CellID->1841417850],
  Cell[586440, 16963, 1791, 46, 29, "ProofInfo",ExpressionUUID->"fc5be3c0-cfbd-484d-840c-b987b55aab11",
   Evaluatable->True,
   CellTags->{"Proof|ID:1579837433-1", "3"},
   CellID->1890841345],
  Cell[588234, 17011, 1852, 47, 29, "ProofInfo",ExpressionUUID->"070c9e7a-2ffe-4eed-a409-8fe374d0f970",
   Evaluatable->True,
   CellTags->{"Proof|ID:2039310054-1", "3"},
   CellID->1409452885],
  Cell[590089, 17060, 1840, 47, 29, "ProofInfo",ExpressionUUID->"3255cca4-1d25-482e-bef3-695eb3e5fa77",
   Evaluatable->True,
   CellTags->{"Proof|ID:1867561751-1", "3"},
   CellID->1823633609],
  Cell[617762, 17922, 1783, 46, 29, "ProofInfo",ExpressionUUID->"41412e05-3d1b-47af-9a2c-69b7da569934",
   Evaluatable->True,
   CellTags->{"Proof|ID:889927681-1", "3"},
   CellID->989686867],
  Cell[619548, 17970, 1793, 46, 29, "ProofInfo",ExpressionUUID->"e718483a-9533-47fb-b2b3-21662a495fe5",
   Evaluatable->True,
   CellTags->{"Proof|ID:1181933579-1", "3"},
   CellID->14631120],
  Cell[621344, 18018, 1785, 46, 29, "ProofInfo",ExpressionUUID->"c083b88a-b5b4-4213-8c12-363f6380f03a",
   Evaluatable->True,
   CellTags->{"Proof|ID:1313006718-1", "3"},
   CellID->1424176457],
  Cell[623132, 18066, 1782, 46, 29, "ProofInfo",ExpressionUUID->"8dfe2f53-ea61-4e3d-9764-e5a54581d749",
   Evaluatable->True,
   CellTags->{"Proof|ID:405798501-1", "3"},
   CellID->1289581872],
  Cell[624917, 18114, 1782, 46, 29, "ProofInfo",ExpressionUUID->"15000b41-243b-47fb-9432-88147364a061",
   Evaluatable->True,
   CellTags->{"Proof|ID:142114837-1", "3"},
   CellID->1757475874],
  Cell[626702, 18162, 1752, 45, 29, "ProofInfo",ExpressionUUID->"ba81f3be-43a1-4405-a13e-20392ccc153a",
   Evaluatable->True,
   CellTags->{"Proof|ID:655509949-1", "3"},
   CellID->939658652],
  Cell[628457, 18209, 1752, 45, 29, "ProofInfo",ExpressionUUID->"a5e8bedb-4a86-49a2-916f-d64eed66bbf2",
   Evaluatable->True,
   CellTags->{"Proof|ID:149156198-1", "3"},
   CellID->125409990],
  Cell[630212, 18256, 1759, 45, 29, "ProofInfo",ExpressionUUID->"81eeefa4-874f-43cf-8a1a-bf67eab02b68",
   Evaluatable->True,
   CellTags->{"Proof|ID:1551567722-1", "3"},
   CellID->128243725],
  Cell[631974, 18303, 1764, 45, 29, "ProofInfo",ExpressionUUID->"27fb32d0-54c1-4430-9891-7032e6c9d015",
   Evaluatable->True,
   CellTags->{"Proof|ID:1618171348-1", "3"},
   CellID->1212022066],
  Cell[633741, 18350, 1790, 46, 29, "ProofInfo",ExpressionUUID->"0d2dd380-ba5d-4928-b373-142c2857c2c0",
   Evaluatable->True,
   CellTags->{"Proof|ID:263157914-1", "3"},
   CellID->1307442480],
  Cell[635534, 18398, 1782, 46, 29, "ProofInfo",ExpressionUUID->"1310c805-4a87-4c5e-91d3-1155f73282e3",
   Evaluatable->True,
   CellTags->{"Proof|ID:586501678-1", "3"},
   CellID->1257980868],
  Cell[637319, 18446, 1812, 47, 29, "ProofInfo",ExpressionUUID->"f7d7c577-0e4f-43f2-aab9-f884ae948b4c",
   Evaluatable->True,
   CellTags->{"Proof|ID:1211775432-1", "3"},
   CellID->1583175827],
  Cell[639134, 18495, 1812, 47, 29, "ProofInfo",ExpressionUUID->"06df1467-a1fc-4be5-b103-26e4a9577222",
   Evaluatable->True,
   CellTags->{"Proof|ID:1437446913-1", "3"},
   CellID->2050049889],
  Cell[640949, 18544, 1768, 45, 29, "ProofInfo",ExpressionUUID->"9ecc279b-a1b3-4c56-a8fa-e29147a8ea06",
   Evaluatable->True,
   CellTags->{"Proof|ID:1743954910-1", "3"},
   CellID->1362760740],
  Cell[642720, 18591, 1809, 47, 29, "ProofInfo",ExpressionUUID->"45cfb94c-9228-4f72-a7c6-3adc38573e3f",
   Evaluatable->True,
   CellTags->{"Proof|ID:732837784-1", "3"},
   CellID->1924200546],
  Cell[644532, 18640, 1831, 47, 29, "ProofInfo",ExpressionUUID->"d7f0bdda-8a9a-44b8-8d62-f6a17e32fa14",
   Evaluatable->True,
   CellTags->{"Proof|ID:520649721-1", "3"},
   CellID->1370228285],
  Cell[646366, 18689, 1768, 45, 29, "ProofInfo",ExpressionUUID->"dbca4d7d-fc2b-4f4e-9403-9ed1fcdfcd97",
   Evaluatable->True,
   CellTags->{"Proof|ID:1959934043-1", "3"},
   CellID->1728602517],
  Cell[648137, 18736, 1811, 47, 29, "ProofInfo",ExpressionUUID->"330a0d43-a3eb-4628-9ac8-18c5d5c04d74",
   Evaluatable->True,
   CellTags->{"Proof|ID:1023545665-1", "3"},
   CellID->124203596],
  Cell[649951, 18785, 1832, 47, 29, "ProofInfo",ExpressionUUID->"e55012c4-27e1-41ae-920d-8392fb2fc039",
   Evaluatable->True,
   CellTags->{"Proof|ID:1036748287-1", "3"},
   CellID->1046065391],
  Cell[651786, 18834, 1830, 47, 29, "ProofInfo",ExpressionUUID->"75c51d93-2b93-476d-8df7-56ee4ba944f9",
   Evaluatable->True,
   CellTags->{"Proof|ID:654257124-1", "3"},
   CellID->190015761],
  Cell[653619, 18883, 1786, 46, 29, "ProofInfo",ExpressionUUID->"bfb4263d-1542-4461-b96b-3cb80e17ef47",
   Evaluatable->True,
   CellTags->{"Proof|ID:788943865-1", "3"},
   CellID->1182421826],
  Cell[655408, 18931, 1786, 46, 29, "ProofInfo",ExpressionUUID->"7126e94e-5680-4178-a002-f91132148a3e",
   Evaluatable->True,
   CellTags->{"Proof|ID:811668796-1", "3"},
   CellID->1010368971],
  Cell[663030, 19140, 1758, 45, 29, "ProofInfo",ExpressionUUID->"60c20774-a0c4-41e0-8583-3cd982a256e3",
   Evaluatable->True,
   CellTags->{"Proof|ID:525687623-1", "3"},
   CellID->118494995],
  Cell[664791, 19187, 1762, 45, 29, "ProofInfo",ExpressionUUID->"e80da33b-c1dd-4198-bc6a-2a3ca4d75937",
   Evaluatable->True,
   CellTags->{"Proof|ID:537376125-1", "3"},
   CellID->998603711],
  Cell[666556, 19234, 1832, 47, 29, "ProofInfo",ExpressionUUID->"1cdbcc0c-c7a2-4a23-ad73-491bc0d57579",
   Evaluatable->True,
   CellTags->{"Proof|ID:1322774384-1", "3"},
   CellID->1961996116],
  Cell[668391, 19283, 1813, 47, 29, "ProofInfo",ExpressionUUID->"362f630c-4228-4cb1-b302-51e8b7bdf329",
   Evaluatable->True,
   CellTags->{"Proof|ID:552587340-1", "3"},
   CellID->1186950056],
  Cell[682196, 19714, 1765, 45, 29, "ProofInfo",ExpressionUUID->"4d04f4ae-48b6-4184-a3cd-f21b71cb61b2",
   Evaluatable->True,
   CellTags->{"Proof|ID:814265295-1", "3"},
   CellID->1805064061],
  Cell[683964, 19761, 1763, 45, 29, "ProofInfo",ExpressionUUID->"268be10f-dd81-47de-b132-ddea9a7f3a82",
   Evaluatable->True,
   CellTags->{"Proof|ID:1283984586-1", "3"},
   CellID->203354138],
  Cell[685730, 19808, 1756, 45, 29, "ProofInfo",ExpressionUUID->"c60cafa1-dc19-4490-a5d0-a50cbd0e6903",
   Evaluatable->True,
   CellTags->{"Proof|ID:1257505636-1", "3"},
   CellID->1997254040],
  Cell[687489, 19855, 1760, 45, 29, "ProofInfo",ExpressionUUID->"243af554-932a-474a-a189-73c259821534",
   Evaluatable->True,
   CellTags->{"Proof|ID:565842574-1", "3"},
   CellID->477474021],
  Cell[689252, 19902, 1792, 46, 29, "ProofInfo",ExpressionUUID->"231ca40f-e980-4b69-b7ca-c100eb04e88e",
   Evaluatable->True,
   CellTags->{"Proof|ID:685115154-1", "3"},
   CellID->1626158995],
  Cell[691047, 19950, 1785, 46, 29, "ProofInfo",ExpressionUUID->"e7c56a2e-bc98-4cf0-80e9-3bac79998304",
   Evaluatable->True,
   CellTags->{"Proof|ID:1570027835-1", "3"},
   CellID->1776911464],
  Cell[692835, 19998, 1781, 46, 29, "ProofInfo",ExpressionUUID->"860e2886-3792-4877-8a4c-304213f7d561",
   Evaluatable->True,
   CellTags->{"Proof|ID:771344911-1", "3"},
   CellID->141629000],
  Cell[694619, 20046, 1821, 47, 29, "ProofInfo",ExpressionUUID->"cecad2e4-4de8-4466-95cb-7b7894b699c6",
   Evaluatable->True,
   CellTags->{"Proof|ID:1251957142-1", "3"},
   CellID->150377480],
  Cell[696443, 20095, 1793, 46, 29, "ProofInfo",ExpressionUUID->"b6d3e804-b5ea-4af4-92d5-df3642a09ce3",
   Evaluatable->True,
   CellTags->{"Proof|ID:2088273463-1", "3"},
   CellID->1818369679],
  Cell[710255, 20529, 1758, 45, 29, "ProofInfo",ExpressionUUID->"d06f131e-6326-4e72-a0bc-30541cdf0aa3",
   Evaluatable->True,
   CellTags->{"Proof|ID:570770860-1", "3"},
   CellID->602566564],
  Cell[712016, 20576, 1766, 45, 29, "ProofInfo",ExpressionUUID->"619955c1-53c3-41d6-b1c8-daf0572378fc",
   Evaluatable->True,
   CellTags->{"Proof|ID:1108679522-1", "3"},
   CellID->1093017718],
  Cell[713785, 20623, 1784, 46, 29, "ProofInfo",ExpressionUUID->"6344d294-e9d4-420c-a9cb-bcd22dfbe228",
   Evaluatable->True,
   CellTags->{"Proof|ID:788559876-1", "3"},
   CellID->1966110247],
  Cell[715572, 20671, 1814, 47, 29, "ProofInfo",ExpressionUUID->"5334709a-119e-480c-b8ba-516e9e54b89b",
   Evaluatable->True,
   CellTags->{"Proof|ID:604821491-1", "3"},
   CellID->429588115],
  Cell[717389, 20720, 1791, 46, 29, "ProofInfo",ExpressionUUID->"4c1795f9-7bd4-48d4-ad29-b0688fdc3008",
   Evaluatable->True,
   CellTags->{"Proof|ID:1716105510-1", "3"},
   CellID->2002110999],
  Cell[719183, 20768, 1820, 47, 29, "ProofInfo",ExpressionUUID->"8dd5ad08-d868-4d2e-ad57-6ee353a4c58b",
   Evaluatable->True,
   CellTags->{"Proof|ID:1849777423-1", "3"},
   CellID->1702528487],
  Cell[721006, 20817, 1762, 45, 29, "ProofInfo",ExpressionUUID->"f5c12280-d2b6-42bc-8fbb-e7baab4c0376",
   Evaluatable->True,
   CellTags->{"Proof|ID:1569569101-1", "3"},
   CellID->1571614296],
  Cell[722771, 20864, 1781, 46, 29, "ProofInfo",ExpressionUUID->"f177dd05-a638-4859-bf62-074cc8860f82",
   Evaluatable->True,
   CellTags->{"Proof|ID:172277084-1", "3"},
   CellID->665117787],
  Cell[728528, 21029, 1808, 47, 29, "ProofInfo",ExpressionUUID->"dfe4d76c-665e-46e9-a153-3fd07c91d051",
   Evaluatable->True,
   CellTags->{"Proof|ID:628580588-1", "3"},
   CellID->202346487]},
 "Proof|ID:893774303-1"->{
  Cell[33425, 1002, 1787, 46, 29, "ProofInfo",ExpressionUUID->"26847eea-8373-4283-8a9d-d1e9558f9cde",
   Evaluatable->True,
   CellTags->{"Proof|ID:893774303-1", "3"},
   CellID->1532689697]},
 "\[SubsetEqual]I"->{
  Cell[36417, 1083, 1475, 46, 54, "FormalTextInputFormula",ExpressionUUID->"7031efb8-33bc-4ea6-bb2f-20619acc517e",
   CellTags->{"\[SubsetEqual]I", "intro"},
   CellID->334402131]},
 "intro"->{
  Cell[36417, 1083, 1475, 46, 54, "FormalTextInputFormula",ExpressionUUID->"7031efb8-33bc-4ea6-bb2f-20619acc517e",
   CellTags->{"\[SubsetEqual]I", "intro"},
   CellID->334402131],
  Cell[37895, 1131, 1324, 41, 50, "FormalTextInputFormula",ExpressionUUID->"fb900f0f-ac9a-44d1-979c-f15fb5f5be35",
   CellTags->{"\[SubsetEqual]D", "trans", "intro"},
   CellID->591515539],
  Cell[39222, 1174, 1344, 41, 50, "FormalTextInputFormula",ExpressionUUID->"e51dde54-c9d6-46ad-abea-f80296524d5c",
   CellTags->{"\[SubsetEqual]D reverse", "trans", "intro"},
   CellID->50542765],
  Cell[64737, 1914, 1315, 41, 50, "FormalTextInputFormula",ExpressionUUID->"ef16daca-313c-4d86-a78a-d1cac485fb01",
   CellTags->{"\[SubsetEqual] antisym", "intro"},
   CellID->833708774],
  Cell[80995, 2391, 1292, 41, 50, "FormalTextInputFormula",ExpressionUUID->"31eeffd9-8416-4535-8aef-65a137cf0890",
   CellTags->{"\[Subset]I", "intro"},
   CellID->32648321],
  Cell[141033, 4111, 1156, 36, 50, "FormalTextInputFormula",ExpressionUUID->"d3c585f0-0af8-4b77-b8f2-4a2e80d5ec5a",
   CellTags->{"PowI", "intro"},
   CellID->1005450222],
  Cell[159530, 4632, 1242, 39, 50, "FormalTextInputFormula",ExpressionUUID->"d27356cd-910b-4e27-a702-8f3e1f38e9d3",
   CellTags->{"complement I", "intro"},
   CellID->319529804],
  Cell[170592, 4949, 1352, 42, 50, "FormalTextInputFormula",ExpressionUUID->"f282f462-c6f2-4c42-93f9-8add211dc534",
   CellTags->{"\[Intersection]I", "intro"},
   CellID->742723078],
  Cell[199870, 5794, 1416, 44, 50, "FormalTextInputFormula",ExpressionUUID->"3dacd0ce-b618-4ccf-8cc9-6fc2a7468b22",
   CellTags->{"\[Union]I classical", "intro"},
   CellID->71686966],
  Cell[225668, 6529, 1318, 42, 50, "FormalTextInputFormula",ExpressionUUID->"f5cee395-7010-4ff0-ba10-f12db0c70eb9",
   CellTags->{"-I SET", "intro"},
   CellID->17112112],
  Cell[261050, 7555, 1513, 49, 50, "FormalTextInputFormula",ExpressionUUID->"94305d11-47e3-4dc3-a7b2-5f685d28a398",
   CellTags->{"InsertI classical", "intro"},
   CellID->794800661],
  Cell[296229, 8562, 860, 23, 47, "FormalTextInputFormula",ExpressionUUID->"154c747d-9b88-42a4-8a21-26f3dd0c0a24",
   CellTags->{"singleton I", "intro"},
   CellID->1706541997],
  Cell[313444, 9064, 1354, 43, 51, "FormalTextInputFormula",ExpressionUUID->"81f4ee41-7d1b-4661-86c1-cd51f6a030b1",
   CellTags->{"@I eq", "intro"},
   CellID->1595981918],
  Cell[350243, 10144, 1202, 37, 51, "FormalTextInputFormula",ExpressionUUID->"65d206a7-f149-48f0-8dc8-445143bf53a0",
   CellTags->{"RangeI eq", "intro"},
   CellID->686090881],
  Cell[364605, 10566, 1376, 42, 64, "FormalTextInputFormula",ExpressionUUID->"ba2298fa-7bbb-45c0-b5bb-c841cdf4f95a",
   CellTags->{"SetOf I", "intro"},
   CellID->783246906]},
 "\[SubsetEqual]D"->{
  Cell[37895, 1131, 1324, 41, 50, "FormalTextInputFormula",ExpressionUUID->"fb900f0f-ac9a-44d1-979c-f15fb5f5be35",
   CellTags->{"\[SubsetEqual]D", "trans", "intro"},
   CellID->591515539]},
 "trans"->{
  Cell[37895, 1131, 1324, 41, 50, "FormalTextInputFormula",ExpressionUUID->"fb900f0f-ac9a-44d1-979c-f15fb5f5be35",
   CellTags->{"\[SubsetEqual]D", "trans", "intro"},
   CellID->591515539],
  Cell[39222, 1174, 1344, 41, 50, "FormalTextInputFormula",ExpressionUUID->"e51dde54-c9d6-46ad-abea-f80296524d5c",
   CellTags->{"\[SubsetEqual]D reverse", "trans", "intro"},
   CellID->50542765],
  Cell[45836, 1383, 1331, 41, 50, "FormalTextInputFormula",ExpressionUUID->"d95e141f-73bd-461e-b3e2-1457b1c31792",
   CellTags->{"\[SubsetEqual] trans", "trans"},
   CellID->373774609],
  Cell[82290, 2434, 1296, 41, 50, "FormalTextInputFormula",ExpressionUUID->"b0ed0cc7-bc57-483e-a835-36a12066a7d1",
   CellTags->{"\[Subset]D", "trans"},
   CellID->457314425],
  Cell[87474, 2603, 1339, 41, 50, "FormalTextInputFormula",ExpressionUUID->"9525b07f-5513-4080-98ed-049a5ac6c399",
   CellTags->{"\[Subset]\[SubsetEqual] trans", "trans"},
   CellID->214375017],
  Cell[88816, 2646, 1337, 41, 50, "FormalTextInputFormula",ExpressionUUID->"c48fc25a-8ea6-4254-bb22-ab5d44b79f90",
   CellTags->{"\[SubsetEqual]\[Subset] trans", "trans"},
   CellID->705380086],
  Cell[90156, 2689, 1306, 41, 50, "FormalTextInputFormula",ExpressionUUID->"ab797c4b-7e7e-4267-afe1-61013ada1bc0",
   CellTags->{"\[Subset] trans", "trans"},
   CellID->122404649]},
 "\[SubsetEqual]D reverse"->{
  Cell[39222, 1174, 1344, 41, 50, "FormalTextInputFormula",ExpressionUUID->"e51dde54-c9d6-46ad-abea-f80296524d5c",
   CellTags->{"\[SubsetEqual]D reverse", "trans", "intro"},
   CellID->50542765]},
 "\[SubsetEqual]E classical"->{
  Cell[40569, 1217, 1749, 54, 50, "FormalTextInputFormula",ExpressionUUID->"fb93893a-e64d-4dfe-ae4e-00ff561e6191",
   CellTags->{"\[SubsetEqual]E classical", "elim"},
   CellID->389797744]},
 "elim"->{
  Cell[40569, 1217, 1749, 54, 50, "FormalTextInputFormula",ExpressionUUID->"fb93893a-e64d-4dfe-ae4e-00ff561e6191",
   CellTags->{"\[SubsetEqual]E classical", "elim"},
   CellID->389797744],
  Cell[69027, 2056, 1859, 59, 50, "FormalTextInputFormula",ExpressionUUID->"499b7ea4-c1ec-4a3c-93aa-52ab71002cec",
   CellTags->{"\[Equal]E classical SET", "elim"},
   CellID->187730850],
  Cell[83589, 2477, 1466, 45, 51, "FormalTextInputFormula",ExpressionUUID->"c1ed4c0b-f24a-4d7a-9cb7-e956d9316e61",
   CellTags->{"\[Subset]E", "elim"},
   CellID->9845002],
  Cell[110468, 3251, 1056, 32, 50, "FormalTextInputFormula",ExpressionUUID->"5e2a16f1-59da-4a82-8948-50266063e236",
   CellTags->{"{}E", "elim"},
   CellID->54906548],
  Cell[174251, 5067, 1520, 46, 51, "FormalTextInputFormula",ExpressionUUID->"d24a4225-19d0-415c-b284-7fe9384810ea",
   CellTags->{"\[Intersection]E", "elim"},
   CellID->31355863],
  Cell[201289, 5840, 1574, 49, 50, "FormalTextInputFormula",ExpressionUUID->"cc6aa25c-83a1-4025-8cd9-4fe49291cfe4",
   CellTags->{"\[Union]E", "elim"},
   CellID->734373493],
  Cell[229357, 6650, 1545, 47, 50, "FormalTextInputFormula",ExpressionUUID->"a15f38a2-1024-4357-b3ae-3d4405f34edf",
   CellTags->{"-E SET", "elim"},
   CellID->128535443],
  Cell[262566, 7606, 1547, 49, 50, "FormalTextInputFormula",ExpressionUUID->"e669061b-1a1d-4dbf-970c-ca9f87768bd7",
   CellTags->{"InsertE", "elim"},
   CellID->752025544],
  Cell[317254, 9190, 1528, 48, 55, "FormalTextInputFormula",ExpressionUUID->"27b53510-2aa4-42f2-ac57-bb62701ff6aa",
   CellTags->{"@E", "elim"},
   CellID->114344294],
  Cell[365984, 10610, 1544, 47, 64, "FormalTextInputFormula",ExpressionUUID->"6eb18b5a-43f0-4bf0-91cb-fa1b2deac1c6",
   CellTags->{"SetOf E", "elim"},
   CellID->1984656332]},
 "\[SubsetEqual] alt"->{
  Cell[42321, 1273, 1393, 43, 54, "FormalTextInputFormula",ExpressionUUID->"b00cb866-8d95-456e-b04f-0dc3663f9abf",
   CellTags->"\[SubsetEqual] alt",
   CellID->223583403]},
 "contraposition \[SubsetEqual]D"->{
  Cell[43717, 1318, 1302, 40, 50, "FormalTextInputFormula",ExpressionUUID->"81ae2f05-9998-4704-978b-a1ea11dc084a",
   CellTags->"contraposition \[SubsetEqual]D",
   CellID->324850200]},
 "\[SubsetEqual] refl"->{
  Cell[45022, 1360, 811, 21, 47, "FormalTextInputFormula",ExpressionUUID->"47260c40-dd6d-4c60-8569-2246238f2cef",
   CellTags->"\[SubsetEqual] refl",
   CellID->232739861]},
 "\[SubsetEqual] trans"->{
  Cell[45836, 1383, 1331, 41, 50, "FormalTextInputFormula",ExpressionUUID->"d95e141f-73bd-461e-b3e2-1457b1c31792",
   CellTags->{"\[SubsetEqual] trans", "trans"},
   CellID->373774609]},
 "Proof|ID:334402131-1"->{
  Cell[47410, 1433, 1790, 46, 29, "ProofInfo",ExpressionUUID->"300c87b6-2c78-4afd-9a22-f6e65b245b94",
   Evaluatable->True,
   CellTags->{"Proof|ID:334402131-1", "5"},
   CellID->624024459]},
 "5"->{
  Cell[47410, 1433, 1790, 46, 29, "ProofInfo",ExpressionUUID->"300c87b6-2c78-4afd-9a22-f6e65b245b94",
   Evaluatable->True,
   CellTags->{"Proof|ID:334402131-1", "5"},
   CellID->624024459],
  Cell[49203, 1481, 1791, 46, 29, "ProofInfo",ExpressionUUID->"c1bf422f-8799-415e-8d2a-f931451d75c6",
   Evaluatable->True,
   CellTags->{"Proof|ID:591515539-1", "5"},
   CellID->1850605281],
  Cell[50997, 1529, 1815, 47, 29, "ProofInfo",ExpressionUUID->"fbf77b68-1b7c-497c-86af-dfad1033b410",
   Evaluatable->True,
   CellTags->{"Proof|ID:50542765-1", "5"},
   CellID->940041033],
  Cell[52815, 1578, 1826, 47, 29, "ProofInfo",ExpressionUUID->"b5a607eb-73fe-4dbb-a4e6-6a0da3ab04fd",
   Evaluatable->True,
   CellTags->{"Proof|ID:389797744-1", "5"},
   CellID->1780566989],
  Cell[54644, 1627, 1812, 47, 29, "ProofInfo",ExpressionUUID->"a6e2338e-aaf5-4f15-b6ec-71a25b011cce",
   Evaluatable->True,
   CellTags->{"Proof|ID:223583403-1", "5"},
   CellID->1680503574],
  Cell[56459, 1676, 1835, 47, 29, "ProofInfo",ExpressionUUID->"576a4a46-e65f-47df-8867-6961c8be3b2d",
   Evaluatable->True,
   CellTags->{"Proof|ID:324850200-1", "5"},
   CellID->623913484],
  Cell[58297, 1725, 1814, 47, 29, "ProofInfo",ExpressionUUID->"5f2f6951-acfb-40e5-85f6-8302f0285cf1",
   Evaluatable->True,
   CellTags->{"Proof|ID:232739861-1", "5"},
   CellID->1074729877],
  Cell[60114, 1774, 1816, 47, 29, "ProofInfo",ExpressionUUID->"d361de80-4f01-4878-9ebd-2f690ccfb113",
   Evaluatable->True,
   CellTags->{"Proof|ID:373774609-1", "5"},
   CellID->1891328660]},
 "Proof|ID:591515539-1"->{
  Cell[49203, 1481, 1791, 46, 29, "ProofInfo",ExpressionUUID->"c1bf422f-8799-415e-8d2a-f931451d75c6",
   Evaluatable->True,
   CellTags->{"Proof|ID:591515539-1", "5"},
   CellID->1850605281]},
 "Proof|ID:50542765-1"->{
  Cell[50997, 1529, 1815, 47, 29, "ProofInfo",ExpressionUUID->"fbf77b68-1b7c-497c-86af-dfad1033b410",
   Evaluatable->True,
   CellTags->{"Proof|ID:50542765-1", "5"},
   CellID->940041033]},
 "Proof|ID:389797744-1"->{
  Cell[52815, 1578, 1826, 47, 29, "ProofInfo",ExpressionUUID->"b5a607eb-73fe-4dbb-a4e6-6a0da3ab04fd",
   Evaluatable->True,
   CellTags->{"Proof|ID:389797744-1", "5"},
   CellID->1780566989]},
 "Proof|ID:223583403-1"->{
  Cell[54644, 1627, 1812, 47, 29, "ProofInfo",ExpressionUUID->"a6e2338e-aaf5-4f15-b6ec-71a25b011cce",
   Evaluatable->True,
   CellTags->{"Proof|ID:223583403-1", "5"},
   CellID->1680503574]},
 "Proof|ID:324850200-1"->{
  Cell[56459, 1676, 1835, 47, 29, "ProofInfo",ExpressionUUID->"576a4a46-e65f-47df-8867-6961c8be3b2d",
   Evaluatable->True,
   CellTags->{"Proof|ID:324850200-1", "5"},
   CellID->623913484]},
 "Proof|ID:232739861-1"->{
  Cell[58297, 1725, 1814, 47, 29, "ProofInfo",ExpressionUUID->"5f2f6951-acfb-40e5-85f6-8302f0285cf1",
   Evaluatable->True,
   CellTags->{"Proof|ID:232739861-1", "5"},
   CellID->1074729877]},
 "Proof|ID:373774609-1"->{
  Cell[60114, 1774, 1816, 47, 29, "ProofInfo",ExpressionUUID->"d361de80-4f01-4878-9ebd-2f690ccfb113",
   Evaluatable->True,
   CellTags->{"Proof|ID:373774609-1", "5"},
   CellID->1891328660]},
 "\[Equal] iff SET"->{
  Cell[63011, 1854, 1723, 58, 54, "FormalTextInputFormula",ExpressionUUID->"cc3b44c7-7696-48b2-9a11-9b0e02cbd0de",
   CellTags->"\[Equal] iff SET",
   CellID->333794707]},
 "\[SubsetEqual] antisym"->{
  Cell[64737, 1914, 1315, 41, 50, "FormalTextInputFormula",ExpressionUUID->"ef16daca-313c-4d86-a78a-d1cac485fb01",
   CellTags->{"\[SubsetEqual] antisym", "intro"},
   CellID->833708774]},
 "\[Equal] iff \[SubsetEqual]"->{
  Cell[66055, 1957, 1539, 51, 50, "FormalTextInputFormula",ExpressionUUID->"734498b0-c5ef-4a95-a0a8-2bf22258d021",
   CellTags->"\[Equal] iff \[SubsetEqual]",
   CellID->1768369467]},
 "\[Equal]E SET"->{
  Cell[67597, 2010, 1427, 44, 51, "FormalTextInputFormula",ExpressionUUID->"39d27beb-c3fa-4672-8c21-115b260ef59e",
   CellTags->"\[Equal]E SET",
   CellID->129158904]},
 "\[Equal]E classical SET"->{
  Cell[69027, 2056, 1859, 59, 50, "FormalTextInputFormula",ExpressionUUID->"499b7ea4-c1ec-4a3c-93aa-52ab71002cec",
   CellTags->{"\[Equal]E classical SET", "elim"},
   CellID->187730850]},
 "Proof|ID:333794707-1"->{
  Cell[71130, 2124, 1808, 47, 29, "ProofInfo",ExpressionUUID->"d29d2a48-174c-424d-be8c-88ec8b4d64fe",
   Evaluatable->True,
   CellTags->{"Proof|ID:333794707-1", "3"},
   CellID->1506794386]},
 "Proof|ID:833708774-1"->{
  Cell[72941, 2173, 1818, 47, 29, "ProofInfo",ExpressionUUID->"5027dd32-9d55-40fe-a936-08172c79cad1",
   Evaluatable->True,
   CellTags->{"Proof|ID:833708774-1", "3"},
   CellID->74925535]},
 "Proof|ID:1768369467-1"->{
  Cell[74762, 2222, 1833, 47, 29, "ProofInfo",ExpressionUUID->"5f4442fc-9b36-4593-b740-1806f36d41e7",
   Evaluatable->True,
   CellTags->{"Proof|ID:1768369467-1", "3"},
   CellID->1488100858]},
 "Proof|ID:129158904-1"->{
  Cell[76598, 2271, 1787, 46, 29, "ProofInfo",ExpressionUUID->"46efcca5-5269-4858-be47-88d1f3ecc77b",
   Evaluatable->True,
   CellTags->{"Proof|ID:129158904-1", "3"},
   CellID->1301219521]},
 "Proof|ID:187730850-1"->{
  Cell[78388, 2319, 1821, 47, 29, "ProofInfo",ExpressionUUID->"e3d8ee28-9559-4c6b-b535-c0f73713378a",
   Evaluatable->True,
   CellTags->{"Proof|ID:187730850-1", "3"},
   CellID->815869071]},
 "\[Subset]I"->{
  Cell[80995, 2391, 1292, 41, 50, "FormalTextInputFormula",ExpressionUUID->"31eeffd9-8416-4535-8aef-65a137cf0890",
   CellTags->{"\[Subset]I", "intro"},
   CellID->32648321]},
 "\[Subset]D"->{
  Cell[82290, 2434, 1296, 41, 50, "FormalTextInputFormula",ExpressionUUID->"b0ed0cc7-bc57-483e-a835-36a12066a7d1",
   CellTags->{"\[Subset]D", "trans"},
   CellID->457314425]},
 "\[Subset]E"->{
  Cell[83589, 2477, 1466, 45, 51, "FormalTextInputFormula",ExpressionUUID->"c1ed4c0b-f24a-4d7a-9cb7-e956d9316e61",
   CellTags->{"\[Subset]E", "elim"},
   CellID->9845002]},
 "\[Subset] alt"->{
  Cell[85058, 2524, 1278, 41, 50, "FormalTextInputFormula",ExpressionUUID->"dbd4d924-a7ac-4077-824e-c31ad364a13b",
   CellTags->"\[Subset] alt",
   CellID->688844702]},
 "\[Subset] implies \[SubsetEqual]"->{
  Cell[86339, 2567, 1132, 34, 50, "FormalTextInputFormula",ExpressionUUID->"80efbdd3-5c09-4513-8814-fe0da4d53969",
   CellTags->"\[Subset] implies \[SubsetEqual]",
   CellID->53899447]},
 "\[Subset]\[SubsetEqual] trans"->{
  Cell[87474, 2603, 1339, 41, 50, "FormalTextInputFormula",ExpressionUUID->"9525b07f-5513-4080-98ed-049a5ac6c399",
   CellTags->{"\[Subset]\[SubsetEqual] trans", "trans"},
   CellID->214375017]},
 "\[SubsetEqual]\[Subset] trans"->{
  Cell[88816, 2646, 1337, 41, 50, "FormalTextInputFormula",ExpressionUUID->"c48fc25a-8ea6-4254-bb22-ab5d44b79f90",
   CellTags->{"\[SubsetEqual]\[Subset] trans", "trans"},
   CellID->705380086]},
 "\[Subset] trans"->{
  Cell[90156, 2689, 1306, 41, 50, "FormalTextInputFormula",ExpressionUUID->"ab797c4b-7e7e-4267-afe1-61013ada1bc0",
   CellTags->{"\[Subset] trans", "trans"},
   CellID->122404649]},
 "\[Subset] irrefl"->{
  Cell[91465, 2732, 919, 27, 47, "FormalTextInputFormula",ExpressionUUID->"8eacd303-9562-4ec0-b90b-56283033ded7",
   CellTags->"\[Subset] irrefl",
   CellID->490505039]},
 "Proof|ID:32648321-1"->{
  Cell[92627, 2768, 1774, 46, 29, "ProofInfo",ExpressionUUID->"6e659942-0c6d-405c-9b68-372d2c14a0ed",
   Evaluatable->True,
   CellTags->{"Proof|ID:32648321-1", "3"},
   CellID->925901741]},
 "Proof|ID:457314425-1"->{
  Cell[94404, 2816, 1781, 46, 29, "ProofInfo",ExpressionUUID->"c887437d-b8f9-4f4a-abc2-4c844d627714",
   Evaluatable->True,
   CellTags->{"Proof|ID:457314425-1", "3"},
   CellID->1854094453]},
 "Proof|ID:9845002-1"->{
  Cell[96188, 2864, 1771, 46, 29, "ProofInfo",ExpressionUUID->"9d0f6fb7-026a-419b-8d81-21560f62f383",
   Evaluatable->True,
   CellTags->{"Proof|ID:9845002-1", "3"},
   CellID->666993796]},
 "Proof|ID:688844702-1"->{
  Cell[97962, 2912, 1786, 46, 29, "ProofInfo",ExpressionUUID->"5f1c3502-0db1-4202-8b55-474b927d5261",
   Evaluatable->True,
   CellTags->{"Proof|ID:688844702-1", "3"},
   CellID->894137823]},
 "Proof|ID:53899447-1"->{
  Cell[99751, 2960, 1834, 47, 29, "ProofInfo",ExpressionUUID->"c6989024-1dd7-4cbd-bf7d-4edaa6f98d71",
   Evaluatable->True,
   CellTags->{"Proof|ID:53899447-1", "3"},
   CellID->1854942837]},
 "Proof|ID:214375017-1"->{
  Cell[101588, 3009, 1833, 47, 29, "ProofInfo",ExpressionUUID->"441a9d35-8e30-4162-9e21-8cf3471755aa",
   Evaluatable->True,
   CellTags->{"Proof|ID:214375017-1", "3"},
   CellID->187880605]},
 "Proof|ID:705380086-1"->{
  Cell[103424, 3058, 1833, 47, 29, "ProofInfo",ExpressionUUID->"5ba5cdc6-46d4-4ad2-94e9-7d5f31bc3fe9",
   Evaluatable->True,
   CellTags->{"Proof|ID:705380086-1", "3"},
   CellID->719967367]},
 "Proof|ID:122404649-1"->{
  Cell[105260, 3107, 1789, 46, 29, "ProofInfo",ExpressionUUID->"76b67bb5-6cbd-4d7e-8eed-928b19df89d2",
   Evaluatable->True,
   CellTags->{"Proof|ID:122404649-1", "3"},
   CellID->28123779]},
 "Proof|ID:490505039-1"->{
  Cell[107052, 3155, 1808, 47, 29, "ProofInfo",ExpressionUUID->"2887fda8-1b65-4980-bb82-685424955125",
   Evaluatable->True,
   CellTags->{"Proof|ID:490505039-1", "3"},
   CellID->1482758530]},
 "\[NotElement]{}"->{
  Cell[109645, 3227, 820, 22, 47, "FormalTextInputFormula",ExpressionUUID->"f40e860d-7927-4260-b558-399c721a0a03",
   CellTags->"\[NotElement]{}",
   CellID->109474291]},
 "{}E"->{
  Cell[110468, 3251, 1056, 32, 50, "FormalTextInputFormula",ExpressionUUID->"5e2a16f1-59da-4a82-8948-50266063e236",
   CellTags->{"{}E", "elim"},
   CellID->54906548]},
 "{}\[SubsetEqual]"->{
  Cell[111527, 3285, 824, 22, 47, "FormalTextInputFormula",ExpressionUUID->"c1739f2e-64f5-45b3-bfc9-a17967b39339",
   CellTags->"{}\[SubsetEqual]",
   CellID->289979004]},
 "\[Equal]{} I"->{
  Cell[112354, 3309, 1227, 39, 51, "FormalTextInputFormula",ExpressionUUID->"26118ba7-c5c4-4498-bf72-00252a9af324",
   CellTags->"\[Equal]{} I",
   CellID->828225892]},
 "\[Not]\[Subset]{}"->{
  Cell[113584, 3350, 941, 28, 47, "FormalTextInputFormula",ExpressionUUID->"9cd78d1f-85a0-4723-897c-277cb4c9e703",
   CellTags->"\[Not]\[Subset]{}",
   CellID->1987360397]},
 "Proof|ID:109474291-1"->{
  Cell[114767, 3387, 1791, 46, 29, "ProofInfo",ExpressionUUID->"ae3cd3a9-b24b-48b2-8218-78763782452d",
   Evaluatable->True,
   CellTags->{"Proof|ID:109474291-1", "3"},
   CellID->1750098064]},
 "Proof|ID:54906548-1"->{
  Cell[116561, 3435, 1746, 45, 29, "ProofInfo",ExpressionUUID->"7c42a337-6cab-46e3-bd77-c0a11197a7dc",
   Evaluatable->True,
   CellTags->{"Proof|ID:54906548-1", "3"},
   CellID->1933668764]},
 "Proof|ID:289979004-1"->{
  Cell[118310, 3482, 1808, 47, 29, "ProofInfo",ExpressionUUID->"6f47db7d-2333-492d-87ac-73a3393a968e",
   Evaluatable->True,
   CellTags->{"Proof|ID:289979004-1", "3"},
   CellID->2138873264]},
 "Proof|ID:828225892-1"->{
  Cell[120121, 3531, 1784, 46, 29, "ProofInfo",ExpressionUUID->"f789d536-f6cf-4bca-84d6-4532e0bed010",
   Evaluatable->True,
   CellTags->{"Proof|ID:828225892-1", "3"},
   CellID->157843443]},
 "Proof|ID:1987360397-1"->{
  Cell[121908, 3579, 1813, 47, 29, "ProofInfo",ExpressionUUID->"3cb753de-8224-4302-aee9-4b257cf1136a",
   Evaluatable->True,
   CellTags->{"Proof|ID:1987360397-1", "3"},
   CellID->1029679455]},
 "UNIV I"->{
  Cell[124979, 3662, 784, 21, 47, "FormalTextInputFormula",ExpressionUUID->"84774292-9b5d-4170-8a68-008bf0e627b8",
   CellTags->"UNIV I",
   CellID->313591036]},
 "\[SubsetEqual]UNIV"->{
  Cell[125766, 3685, 812, 21, 47, "FormalTextInputFormula",ExpressionUUID->"5a1c8aed-f913-4370-96ff-617b5a19afed",
   CellTags->"\[SubsetEqual]UNIV",
   CellID->223522958]},
 "\[Equal]UNIV I"->{
  Cell[126581, 3708, 1048, 31, 51, "FormalTextInputFormula",ExpressionUUID->"5edf972c-0fd1-45e3-886f-b3b8b285efab",
   CellTags->"\[Equal]UNIV I",
   CellID->256313502]},
 "UNIV\[NotEqual]{}"->{
  Cell[127632, 3741, 780, 20, 38, "FormalTextInputFormula",ExpressionUUID->"eb06d768-e9ca-4788-ad46-4aa781f6599b",
   CellTags->"UNIV\[NotEqual]{}",
   CellID->358676608]},
 "{}\[NotEqual]UNIV"->{
  Cell[128415, 3763, 776, 20, 38, "FormalTextInputFormula",ExpressionUUID->"f150aa79-de04-4fd6-8a60-7995eeec44d4",
   CellTags->"{}\[NotEqual]UNIV",
   CellID->13969540]},
 "Proof|ID:313591036-1"->{
  Cell[129434, 3792, 1758, 45, 29, "ProofInfo",ExpressionUUID->"f472fc82-802b-410b-868c-d07e017b0268",
   Evaluatable->True,
   CellTags->{"Proof|ID:313591036-1", "3"},
   CellID->1173747547]},
 "Proof|ID:223522958-1"->{
  Cell[131195, 3839, 1812, 47, 29, "ProofInfo",ExpressionUUID->"184057ce-28ca-4f86-bd4e-79e86a17715d",
   Evaluatable->True,
   CellTags->{"Proof|ID:223522958-1", "3"},
   CellID->2010038796]},
 "Proof|ID:256313502-1"->{
  Cell[133010, 3888, 1788, 46, 29, "ProofInfo",ExpressionUUID->"de3eb27b-5b7d-4ae2-988d-b2653d4d8d31",
   Evaluatable->True,
   CellTags->{"Proof|ID:256313502-1", "3"},
   CellID->741200714]},
 "Proof|ID:358676608-1"->{
  Cell[134801, 3936, 1810, 47, 29, "ProofInfo",ExpressionUUID->"824b928c-cb7f-4776-8b3e-3335d33de91a",
   Evaluatable->True,
   CellTags->{"Proof|ID:358676608-1", "3"},
   CellID->1104785395]},
 "Proof|ID:13969540-1"->{
  Cell[136614, 3985, 1803, 47, 29, "ProofInfo",ExpressionUUID->"49775606-3bad-4639-ad88-ce4f5704386f",
   Evaluatable->True,
   CellTags->{"Proof|ID:13969540-1", "3"},
   CellID->400553415]},
 "\[Element]Pow"->{
  Cell[139786, 4069, 1244, 40, 50, "FormalTextInputFormula",ExpressionUUID->"2d5efc8e-b5dc-4fd3-9859-b98a6c77774b",
   CellTags->"\[Element]Pow",
   CellID->91383453]},
 "PowI"->{
  Cell[141033, 4111, 1156, 36, 50, "FormalTextInputFormula",ExpressionUUID->"d3c585f0-0af8-4b77-b8f2-4a2e80d5ec5a",
   CellTags->{"PowI", "intro"},
   CellID->1005450222]},
 "PowD"->{
  Cell[142192, 4149, 1096, 34, 50, "FormalTextInputFormula",ExpressionUUID->"134a7df0-6a03-483f-8b3c-9ed6ba820c3f",
   CellTags->"PowD",
   CellID->71998340]},
 "{}\[Element]Pow"->{
  Cell[143291, 4185, 850, 23, 47, "FormalTextInputFormula",ExpressionUUID->"0039201d-a8f0-4273-ac3a-5b47720c154c",
   CellTags->"{}\[Element]Pow",
   CellID->135248918]},
 "self\[Element]Pow"->{
  Cell[144144, 4210, 835, 22, 47, "FormalTextInputFormula",ExpressionUUID->"3975308c-87da-4dbf-bd5e-7f174b4273f7",
   CellTags->"self\[Element]Pow",
   CellID->365430671]},
 "Pow\[NotEqual]{}"->{
  Cell[144982, 4234, 855, 23, 47, "FormalTextInputFormula",ExpressionUUID->"725fd266-ab88-47ba-adf8-00acdeb98c1b",
   CellTags->"Pow\[NotEqual]{}",
   CellID->761522039]},
 "Proof|ID:91383453-1"->{
  Cell[146081, 4266, 1782, 46, 29, "ProofInfo",ExpressionUUID->"135e6c9f-6b5f-45fe-bca0-d1d8524dfc01",
   Evaluatable->True,
   CellTags->{"Proof|ID:91383453-1", "3"},
   CellID->1669010862]},
 "Proof|ID:1005450222-1"->{
  Cell[147866, 4314, 1757, 45, 29, "ProofInfo",ExpressionUUID->"2eff29d4-750b-432e-b353-66e38bf0fded",
   Evaluatable->True,
   CellTags->{"Proof|ID:1005450222-1", "3"},
   CellID->969791405]},
 "Proof|ID:71998340-1"->{
  Cell[149626, 4361, 1749, 45, 29, "ProofInfo",ExpressionUUID->"e66761b0-9d29-4441-b48e-3fce10402d48",
   Evaluatable->True,
   CellTags->{"Proof|ID:71998340-1", "3"},
   CellID->1724899203]},
 "Proof|ID:135248918-1"->{
  Cell[151378, 4408, 1791, 46, 29, "ProofInfo",ExpressionUUID->"e9f8d063-f619-4ed9-811c-81d57ff90446",
   Evaluatable->True,
   CellTags->{"Proof|ID:135248918-1", "3"},
   CellID->417457554]},
 "Proof|ID:365430671-1"->{
  Cell[153172, 4456, 1811, 47, 29, "ProofInfo",ExpressionUUID->"88881ba8-d043-4491-afa7-e2b45acd577c",
   Evaluatable->True,
   CellTags->{"Proof|ID:365430671-1", "3"},
   CellID->1046907733]},
 "Proof|ID:761522039-1"->{
  Cell[154986, 4505, 1809, 47, 29, "ProofInfo",ExpressionUUID->"7af83b7e-0a64-42d4-9cfa-8fb1f1515d6b",
   Evaluatable->True,
   CellTags->{"Proof|ID:761522039-1", "3"},
   CellID->1384782616]},
 "\[Element]complement"->{
  Cell[158279, 4590, 1248, 40, 50, "FormalTextInputFormula",ExpressionUUID->"7d81673e-75fc-41aa-b4b7-569230380565",
   CellTags->"\[Element]complement",
   CellID->499422188]},
 "complement I"->{
  Cell[159530, 4632, 1242, 39, 50, "FormalTextInputFormula",ExpressionUUID->"d27356cd-910b-4e27-a702-8f3e1f38e9d3",
   CellTags->{"complement I", "intro"},
   CellID->319529804]},
 "complement D"->{
  Cell[160775, 4673, 1109, 35, 50, "FormalTextInputFormula",ExpressionUUID->"0bda5580-c651-4792-9b44-3829608ed193",
   CellTags->"complement D",
   CellID->7165872]},
 "Proof|ID:499422188-1"->{
  Cell[162127, 4717, 1816, 47, 29, "ProofInfo",ExpressionUUID->"4bb986a8-28da-4979-8914-a7bf05134176",
   Evaluatable->True,
   CellTags->{"Proof|ID:499422188-1", "3"},
   CellID->742874104]},
 "Proof|ID:319529804-1"->{
  Cell[163946, 4766, 1785, 46, 29, "ProofInfo",ExpressionUUID->"6b7cdde2-f7e1-4895-96bd-f5dba2fa343f",
   Evaluatable->True,
   CellTags->{"Proof|ID:319529804-1", "3"},
   CellID->692678955]},
 "Proof|ID:7165872-1"->{
  Cell[165734, 4814, 1776, 46, 29, "ProofInfo",ExpressionUUID->"1b5041f8-10cf-4c78-8caf-cee02595fafa",
   Evaluatable->True,
   CellTags->{"Proof|ID:7165872-1", "3"},
   CellID->214206169]},
 "\[Element]\[Intersection]"->{
  Cell[169237, 4905, 1352, 42, 50, "FormalTextInputFormula",ExpressionUUID->"faf9caf8-e1c5-4ad6-a348-5e25e999ed9e",
   CellTags->"\[Element]\[Intersection]",
   CellID->299832540]},
 "\[Intersection]I"->{
  Cell[170592, 4949, 1352, 42, 50, "FormalTextInputFormula",ExpressionUUID->"f282f462-c6f2-4c42-93f9-8add211dc534",
   CellTags->{"\[Intersection]I", "intro"},
   CellID->742723078]},
 "\[Intersection]D1"->{
  Cell[171947, 4993, 1149, 35, 50, "FormalTextInputFormula",ExpressionUUID->"728e5702-a968-4311-b864-1dd1f476124d",
   CellTags->"\[Intersection]D1",
   CellID->196903271]},
 "\[Intersection]D2"->{
  Cell[173099, 5030, 1149, 35, 50, "FormalTextInputFormula",ExpressionUUID->"2c742309-e67a-46f8-923c-ec686d1342d2",
   CellTags->"\[Intersection]D2",
   CellID->214165647]},
 "\[Intersection]E"->{
  Cell[174251, 5067, 1520, 46, 51, "FormalTextInputFormula",ExpressionUUID->"d24a4225-19d0-415c-b284-7fe9384810ea",
   CellTags->{"\[Intersection]E", "elim"},
   CellID->31355863]},
 "Proof|ID:299832540-1"->{
  Cell[176015, 5122, 1826, 47, 29, "ProofInfo",ExpressionUUID->"06ccfced-9f01-4ea1-b866-1dc0ce6111d4",
   Evaluatable->True,
   CellTags->{"Proof|ID:299832540-1", "3"},
   CellID->520490995]},
 "Proof|ID:742723078-1"->{
  Cell[177844, 5171, 1809, 47, 29, "ProofInfo",ExpressionUUID->"b655e7ec-ba70-4141-b8ce-f4c7c9313eb5",
   Evaluatable->True,
   CellTags->{"Proof|ID:742723078-1", "3"},
   CellID->1206322670]},
 "Proof|ID:196903271-1"->{
  Cell[179656, 5220, 1811, 47, 29, "ProofInfo",ExpressionUUID->"999e3753-e878-46ae-816d-c1c14791130d",
   Evaluatable->True,
   CellTags->{"Proof|ID:196903271-1", "3"},
   CellID->1182674228]},
 "Proof|ID:214165647-1"->{
  Cell[181470, 5269, 1811, 47, 29, "ProofInfo",ExpressionUUID->"263b3d20-5c8b-4d42-8cd7-bb5ecf6bba6c",
   Evaluatable->True,
   CellTags->{"Proof|ID:214165647-1", "3"},
   CellID->1685360901]},
 "Proof|ID:31355863-1"->{
  Cell[183284, 5318, 1802, 47, 29, "ProofInfo",ExpressionUUID->"e8ca3e41-bda4-4b23-8e4b-61f126bed335",
   Evaluatable->True,
   CellTags->{"Proof|ID:31355863-1", "3"},
   CellID->634954845]},
 "\[Intersection] lower1"->{
  Cell[186331, 5396, 884, 24, 50, "FormalTextInputFormula",ExpressionUUID->"5a07b132-a168-47df-8a0a-d4cbc2bdf2be",
   CellTags->"\[Intersection] lower1",
   CellID->330976770]},
 "\[Intersection] lower2"->{
  Cell[187218, 5422, 886, 24, 50, "FormalTextInputFormula",ExpressionUUID->"69552c22-c7e7-4e84-ae6d-0c027f7b7ca1",
   CellTags->"\[Intersection] lower2",
   CellID->567877347]},
 "\[Intersection] greatest"->{
  Cell[188107, 5448, 1333, 41, 50, "FormalTextInputFormula",ExpressionUUID->"da054ea7-7898-424b-8f20-ba26dcda115c",
   CellTags->"\[Intersection] greatest",
   CellID->370774399]},
 "Proof|ID:330976770-1"->{
  Cell[189684, 5498, 1821, 47, 29, "ProofInfo",ExpressionUUID->"c9408df1-52d0-43ba-9c83-f86df76887a4",
   Evaluatable->True,
   CellTags->{"Proof|ID:330976770-1", "3"},
   CellID->1901417663]},
 "Proof|ID:567877347-1"->{
  Cell[191508, 5547, 1821, 47, 29, "ProofInfo",ExpressionUUID->"cd544126-02a6-4a0d-b5ab-040100796126",
   Evaluatable->True,
   CellTags->{"Proof|ID:567877347-1", "3"},
   CellID->1604414518]},
 "Proof|ID:370774399-1"->{
  Cell[193332, 5596, 1825, 47, 29, "ProofInfo",ExpressionUUID->"067b7963-94d6-4a20-aae6-bc77ea005db6",
   Evaluatable->True,
   CellTags->{"Proof|ID:370774399-1", "3"},
   CellID->1352341786]},
 "\[Element]\[Union]"->{
  Cell[196280, 5676, 1325, 42, 50, "FormalTextInputFormula",ExpressionUUID->"f52d2f46-1098-465b-90bd-12324ecbac1f",
   CellTags->"\[Element]\[Union]",
   CellID->24302859]},
 "\[Union]I1"->{
  Cell[197608, 5720, 1126, 35, 50, "FormalTextInputFormula",ExpressionUUID->"5ea72cf8-f892-430b-956b-10c727c945bf",
   CellTags->"\[Union]I1",
   CellID->204712299]},
 "\[Union]I2"->{
  Cell[198737, 5757, 1130, 35, 50, "FormalTextInputFormula",ExpressionUUID->"55fd758c-2ba1-4a9c-a211-2ba0089a2f44",
   CellTags->"\[Union]I2",
   CellID->118362674]},
 "\[Union]I classical"->{
  Cell[199870, 5794, 1416, 44, 50, "FormalTextInputFormula",ExpressionUUID->"3dacd0ce-b618-4ccf-8cc9-6fc2a7468b22",
   CellTags->{"\[Union]I classical", "intro"},
   CellID->71686966]},
 "\[Union]E"->{
  Cell[201289, 5840, 1574, 49, 50, "FormalTextInputFormula",ExpressionUUID->"cc6aa25c-83a1-4025-8cd9-4fe49291cfe4",
   CellTags->{"\[Union]E", "elim"},
   CellID->734373493]},
 "Proof|ID:24302859-1"->{
  Cell[203106, 5898, 1806, 47, 29, "ProofInfo",ExpressionUUID->"5d01d301-0853-41ec-8b5b-f654db10d92f",
   Evaluatable->True,
   CellTags->{"Proof|ID:24302859-1", "3"},
   CellID->910429301]},
 "Proof|ID:204712299-1"->{
  Cell[204915, 5947, 1782, 46, 29, "ProofInfo",ExpressionUUID->"58cda68d-cdf0-4074-9690-d94a7ad8e87b",
   Evaluatable->True,
   CellTags->{"Proof|ID:204712299-1", "3"},
   CellID->2106560278]},
 "Proof|ID:118362674-1"->{
  Cell[206700, 5995, 1781, 46, 29, "ProofInfo",ExpressionUUID->"5ab74a01-dcb4-4410-adbb-2e6f4bc3d887",
   Evaluatable->True,
   CellTags->{"Proof|ID:118362674-1", "3"},
   CellID->430856106]},
 "Proof|ID:71686966-1"->{
  Cell[208484, 6043, 1808, 47, 29, "ProofInfo",ExpressionUUID->"c438dae3-6253-4324-8a41-6d2ae06ea6ec",
   Evaluatable->True,
   CellTags->{"Proof|ID:71686966-1", "3"},
   CellID->634618502]},
 "Proof|ID:734373493-1"->{
  Cell[210295, 6092, 1765, 45, 29, "ProofInfo",ExpressionUUID->"7192f13d-fafc-45a0-9eba-e5628635aa2e",
   Evaluatable->True,
   CellTags->{"Proof|ID:734373493-1", "3"},
   CellID->1988227689]},
 "\[Union] upper1"->{
  Cell[213344, 6169, 859, 23, 50, "FormalTextInputFormula",ExpressionUUID->"23354d54-d353-451c-96d2-7d23f93c1ab9",
   CellTags->"\[Union] upper1",
   CellID->776879610]},
 "\[Union] upper2"->{
  Cell[214206, 6194, 857, 23, 50, "FormalTextInputFormula",ExpressionUUID->"fd46a9f7-6c14-40c4-a4e9-9c816ccbf6cd",
   CellTags->"\[Union] upper2",
   CellID->463603332]},
 "\[Union] least"->{
  Cell[215066, 6219, 1308, 41, 50, "FormalTextInputFormula",ExpressionUUID->"2c6fd105-c6b3-4c91-930d-cdf36722c20d",
   CellTags->"\[Union] least",
   CellID->672172372]},
 "Proof|ID:776879610-1"->{
  Cell[216618, 6269, 1792, 46, 29, "ProofInfo",ExpressionUUID->"4273e686-36d1-479a-869b-b04747afdf48",
   Evaluatable->True,
   CellTags->{"Proof|ID:776879610-1", "3"},
   CellID->1308387989]},
 "Proof|ID:463603332-1"->{
  Cell[218413, 6317, 1791, 46, 29, "ProofInfo",ExpressionUUID->"a0289919-612a-43d3-83be-5e58d75618f3",
   Evaluatable->True,
   CellTags->{"Proof|ID:463603332-1", "3"},
   CellID->283700494]},
 "Proof|ID:672172372-1"->{
  Cell[220207, 6365, 1789, 46, 29, "ProofInfo",ExpressionUUID->"bacbec27-f6b0-4574-94e4-76900e1b7a43",
   Evaluatable->True,
   CellTags->{"Proof|ID:672172372-1", "3"},
   CellID->563063133]},
 "- alt SET"->{
  Cell[223104, 6444, 1247, 39, 54, "FormalTextInputFormula",ExpressionUUID->"e2f0a575-b7f7-478c-9cc1-7d579996f383",
   CellTags->"- alt SET",
   CellID->540115919]},
 "\[Element]-"->{
  Cell[224354, 6485, 1311, 42, 50, "FormalTextInputFormula",ExpressionUUID->"dbbc2d0c-ae52-42a9-a1a3-6ac77bbec4ef",
   CellTags->"\[Element]-",
   CellID->201090499]},
 "-I SET"->{
  Cell[225668, 6529, 1318, 42, 50, "FormalTextInputFormula",ExpressionUUID->"f5cee395-7010-4ff0-ba10-f12db0c70eb9",
   CellTags->{"-I SET", "intro"},
   CellID->17112112]},
 "-D1 SET"->{
  Cell[226989, 6573, 1114, 35, 50, "FormalTextInputFormula",ExpressionUUID->"62a66027-3b18-408d-9e80-b3a57cfe1bcf",
   CellTags->"-D1 SET",
   CellID->590579378]},
 "-D2 SET"->{
  Cell[228106, 6610, 1248, 38, 50, "FormalTextInputFormula",ExpressionUUID->"b24c8d74-3d7a-402c-958b-25766a575785",
   CellTags->"-D2 SET",
   CellID->984086]},
 "-E SET"->{
  Cell[229357, 6650, 1545, 47, 50, "FormalTextInputFormula",ExpressionUUID->"a15f38a2-1024-4357-b3ae-3d4405f34edf",
   CellTags->{"-E SET", "elim"},
   CellID->128535443]},
 "complement \[Equal] UNIV-"->{
  Cell[230905, 6699, 960, 28, 47, "FormalTextInputFormula",ExpressionUUID->"25eb9645-ea9d-42ae-8802-e919a21ef9e4",
   CellTags->"complement \[Equal] UNIV-",
   CellID->679084152]},
 "Proof|ID:540115919-1"->{
  Cell[232109, 6736, 1765, 45, 29, "ProofInfo",ExpressionUUID->"413f783b-5cb5-476c-8d17-7d845255998a",
   Evaluatable->True,
   CellTags->{"Proof|ID:540115919-1", "3"},
   CellID->2034290086]},
 "Proof|ID:201090499-1"->{
  Cell[233877, 6783, 1784, 46, 29, "ProofInfo",ExpressionUUID->"0e3f881b-3ec2-4218-8f49-710e98423e7e",
   Evaluatable->True,
   CellTags->{"Proof|ID:201090499-1", "3"},
   CellID->1196091729]},
 "Proof|ID:17112112-1"->{
  Cell[235664, 6831, 1753, 45, 29, "ProofInfo",ExpressionUUID->"fae8ecb6-0379-4df0-8561-e4c239de4b21",
   Evaluatable->True,
   CellTags->{"Proof|ID:17112112-1", "3"},
   CellID->1855793195]},
 "Proof|ID:590579378-1"->{
  Cell[237420, 6878, 1759, 45, 29, "ProofInfo",ExpressionUUID->"ff98ca43-6f95-4017-83fe-5e58556b956d",
   Evaluatable->True,
   CellTags->{"Proof|ID:590579378-1", "3"},
   CellID->87366359]},
 "Proof|ID:984086-1"->{
  Cell[239182, 6925, 1748, 45, 29, "ProofInfo",ExpressionUUID->"497b9d54-4fde-4ed0-af44-f8d1fb40d21a",
   Evaluatable->True,
   CellTags->{"Proof|ID:984086-1", "3"},
   CellID->738044893]},
 "Proof|ID:128535443-1"->{
  Cell[240933, 6972, 1758, 45, 29, "ProofInfo",ExpressionUUID->"9ab0c91e-bd51-4993-aa37-27e47b2db8cf",
   Evaluatable->True,
   CellTags->{"Proof|ID:128535443-1", "3"},
   CellID->187234111]},
 "Proof|ID:679084152-1"->{
  Cell[242694, 7019, 1826, 47, 29, "ProofInfo",ExpressionUUID->"526b809c-7826-439b-ac09-9eaa35adbac7",
   Evaluatable->True,
   CellTags->{"Proof|ID:679084152-1", "3"},
   CellID->887782668]},
 "-\[SubsetEqual]"->{
  Cell[245829, 7097, 856, 24, 50, "FormalTextInputFormula",ExpressionUUID->"d07f39c7-4b49-4a87-bb3c-ecff0503e8a1",
   CellTags->"-\[SubsetEqual]",
   CellID->917574097]},
 "-\[SubsetEqual] iff"->{
  Cell[246688, 7123, 1297, 41, 50, "FormalTextInputFormula",ExpressionUUID->"138bd892-f1bc-49e8-a2ee-0d8190289c18",
   CellTags->"-\[SubsetEqual] iff",
   CellID->34104065]},
 "\[Subset] implies \[Exists]\[Element]"->{
  Cell[247988, 7166, 1221, 37, 54, "FormalTextInputFormula",ExpressionUUID->"5f3dad52-4b63-4783-80a8-4673a24fc263",
   CellTags->"\[Subset] implies \[Exists]\[Element]",
   CellID->168218769]},
 "Proof|ID:917574097-1"->{
  Cell[249451, 7212, 1792, 46, 29, "ProofInfo",ExpressionUUID->"94ee77de-dbad-4d4e-97d6-1755cd58f119",
   Evaluatable->True,
   CellTags->{"Proof|ID:917574097-1", "3"},
   CellID->1547795340]},
 "Proof|ID:34104065-1"->{
  Cell[251246, 7260, 1808, 47, 29, "ProofInfo",ExpressionUUID->"19482a26-cb3e-4e1f-af6d-49608b2ed75c",
   Evaluatable->True,
   CellTags->{"Proof|ID:34104065-1", "3"},
   CellID->838116264]},
 "Proof|ID:168218769-1"->{
  Cell[253057, 7309, 1851, 47, 29, "ProofInfo",ExpressionUUID->"9127f290-b8cf-4129-98e6-27563e725e5a",
   Evaluatable->True,
   CellTags->{"Proof|ID:168218769-1", "3"},
   CellID->1909735428]},
 "\[Element]Insert"->{
  Cell[256129, 7392, 1579, 53, 50, "FormalTextInputFormula",ExpressionUUID->"87cdeabf-648d-464d-b035-5d143d98c87f",
   CellTags->"\[Element]Insert",
   CellID->833542610]},
 "InsertI1"->{
  Cell[257711, 7447, 1020, 30, 50, "FormalTextInputFormula",ExpressionUUID->"3dfc2ea8-5b58-4bed-bc5a-781776a474ee",
   CellTags->"InsertI1",
   CellID->265974781]},
 "InsertI1'"->{
  Cell[258734, 7479, 1156, 36, 50, "FormalTextInputFormula",ExpressionUUID->"a53e2216-ee84-4b9a-9367-f3854ab8dbba",
   CellTags->"InsertI1'",
   CellID->369554344]},
 "InsertI2"->{
  Cell[259893, 7517, 1154, 36, 50, "FormalTextInputFormula",ExpressionUUID->"eee79b47-d910-49b8-b643-3d87423df22a",
   CellTags->"InsertI2",
   CellID->276809760]},
 "InsertI classical"->{
  Cell[261050, 7555, 1513, 49, 50, "FormalTextInputFormula",ExpressionUUID->"94305d11-47e3-4dc3-a7b2-5f685d28a398",
   CellTags->{"InsertI classical", "intro"},
   CellID->794800661]},
 "InsertE"->{
  Cell[262566, 7606, 1547, 49, 50, "FormalTextInputFormula",ExpressionUUID->"e669061b-1a1d-4dbf-970c-ca9f87768bd7",
   CellTags->{"InsertE", "elim"},
   CellID->752025544]},
 "Insert alt"->{
  Cell[264116, 7657, 1246, 39, 54, "FormalTextInputFormula",ExpressionUUID->"2c22bdf3-e000-413b-9d7a-df4eabf07af9",
   CellTags->"Insert alt",
   CellID->60042161]},
 "Proof|ID:833542610-1"->{
  Cell[265606, 7705, 1809, 47, 29, "ProofInfo",ExpressionUUID->"69efc2d1-78ff-494c-9afc-e15c84128c04",
   Evaluatable->True,
   CellTags->{"Proof|ID:833542610-1", "3"},
   CellID->1603093206]},
 "Proof|ID:265974781-1"->{
  Cell[267418, 7754, 1763, 45, 29, "ProofInfo",ExpressionUUID->"e9466e7d-028a-4117-ac46-65b9a92ddf8e",
   Evaluatable->True,
   CellTags->{"Proof|ID:265974781-1", "3"},
   CellID->1283043907]},
 "Proof|ID:369554344-1"->{
  Cell[269184, 7801, 1764, 45, 29, "ProofInfo",ExpressionUUID->"9dffdc4b-0a67-4869-8254-6e3cf0381dfa",
   Evaluatable->True,
   CellTags->{"Proof|ID:369554344-1", "3"},
   CellID->111935385]},
 "Proof|ID:276809760-1"->{
  Cell[270951, 7848, 1762, 45, 29, "ProofInfo",ExpressionUUID->"53cab25a-3a17-4a43-97f3-cc2b778e43fc",
   Evaluatable->True,
   CellTags->{"Proof|ID:276809760-1", "3"},
   CellID->591244542]},
 "Proof|ID:794800661-1"->{
  Cell[272716, 7895, 1811, 47, 29, "ProofInfo",ExpressionUUID->"893df1bb-a5ff-4f0d-a86c-b9a2a51b065b",
   Evaluatable->True,
   CellTags->{"Proof|ID:794800661-1", "3"},
   CellID->1050155476]},
 "Proof|ID:752025544-1"->{
  Cell[274530, 7944, 1761, 45, 29, "ProofInfo",ExpressionUUID->"8908cda5-38e0-4ff0-806e-c21baafb2a94",
   Evaluatable->True,
   CellTags->{"Proof|ID:752025544-1", "3"},
   CellID->1183293999]},
 "Proof|ID:60042161-1"->{
  Cell[276294, 7991, 1775, 46, 29, "ProofInfo",ExpressionUUID->"acf349b9-ba5d-40be-9cc8-8cf74e4537e3",
   Evaluatable->True,
   CellTags->{"Proof|ID:60042161-1", "3"},
   CellID->807514983]},
 "\[SubsetEqual]Insert I"->{
  Cell[279014, 8065, 947, 25, 50, "FormalTextInputFormula",ExpressionUUID->"1378f62a-f8a1-4388-a3e7-d01a5c35792f",
   CellTags->"\[SubsetEqual]Insert I",
   CellID->499325579]},
 "\[SubsetEqual]Insert I2"->{
  Cell[279964, 8092, 1189, 36, 50, "FormalTextInputFormula",ExpressionUUID->"27164a92-0285-4493-bacb-7c74cb461d52",
   CellTags->"\[SubsetEqual]Insert I2",
   CellID->89181006]},
 "\[SubsetEqual]Insert"->{
  Cell[281156, 8130, 1550, 49, 50, "FormalTextInputFormula",ExpressionUUID->"fe809dfb-5875-4485-ae29-7091cd6e4d62",
   CellTags->"\[SubsetEqual]Insert",
   CellID->17294591]},
 "\[SubsetEqual]Insert iff"->{
  Cell[282709, 8181, 1608, 50, 56, "FormalTextInputFormula",ExpressionUUID->"ad82da62-4b0a-490c-8806-1851e7b37e9b",
   CellTags->"\[SubsetEqual]Insert iff",
   CellID->1726053781]},
 "Insert\[SubsetEqual] iff"->{
  Cell[284320, 8233, 1380, 43, 50, "FormalTextInputFormula",ExpressionUUID->"23802000-c362-4a0b-afa6-0ac37540f0d2",
   CellTags->"Insert\[SubsetEqual] iff",
   CellID->140408834]},
 "Proof|ID:499325579-1"->{
  Cell[285945, 8285, 1821, 47, 29, "ProofInfo",ExpressionUUID->"033295b6-03ea-4743-b02f-9a029d18a564",
   Evaluatable->True,
   CellTags->{"Proof|ID:499325579-1", "3"},
   CellID->1989694051]},
 "Proof|ID:89181006-1"->{
  Cell[287769, 8334, 1816, 47, 29, "ProofInfo",ExpressionUUID->"74457cd9-9415-4a5b-a6ff-c4224aff06d1",
   Evaluatable->True,
   CellTags->{"Proof|ID:89181006-1", "3"},
   CellID->169658549]},
 "Proof|ID:17294591-1"->{
  Cell[289588, 8383, 1809, 47, 29, "ProofInfo",ExpressionUUID->"539ab388-aaaf-402a-84e8-d34f4c88e778",
   Evaluatable->True,
   CellTags->{"Proof|ID:17294591-1", "3"},
   CellID->94829629]},
 "Proof|ID:1726053781-1"->{
  Cell[291400, 8432, 1828, 47, 29, "ProofInfo",ExpressionUUID->"a019316d-6449-4592-a608-6892d07b6388",
   Evaluatable->True,
   CellTags->{"Proof|ID:1726053781-1", "3"},
   CellID->1321149923]},
 "Proof|ID:140408834-1"->{
  Cell[293231, 8481, 1824, 47, 29, "ProofInfo",ExpressionUUID->"72ddb486-2189-446a-b4f0-dccfaf5bf1dc",
   Evaluatable->True,
   CellTags->{"Proof|ID:140408834-1", "3"},
   CellID->662871545]},
 "singleton I"->{
  Cell[296229, 8562, 860, 23, 47, "FormalTextInputFormula",ExpressionUUID->"154c747d-9b88-42a4-8a21-26f3dd0c0a24",
   CellTags->{"singleton I", "intro"},
   CellID->1706541997]},
 "singleton D"->{
  Cell[297092, 8587, 1112, 35, 50, "FormalTextInputFormula",ExpressionUUID->"390adf86-f8b8-4097-9da9-ab6551101d71",
   CellTags->"singleton D",
   CellID->1946555487]},
 "singleton inject"->{
  Cell[298207, 8624, 1146, 36, 51, "FormalTextInputFormula",ExpressionUUID->"cb45ee28-969d-4085-b7bf-1d4e9c010512",
   CellTags->"singleton inject",
   CellID->1642372132]},
 "Proof|ID:1706541997-1"->{
  Cell[299598, 8669, 1787, 46, 29, "ProofInfo",ExpressionUUID->"24b89c28-f6f2-4a2d-8f6c-d7faa64ba567",
   Evaluatable->True,
   CellTags->{"Proof|ID:1706541997-1", "3"},
   CellID->1610428281]},
 "Proof|ID:1946555487-1"->{
  Cell[301388, 8717, 1785, 46, 29, "ProofInfo",ExpressionUUID->"c797b61f-6a99-4d82-9375-e86eb679b1f6",
   Evaluatable->True,
   CellTags->{"Proof|ID:1946555487-1", "3"},
   CellID->44899502]},
 "Proof|ID:1642372132-1"->{
  Cell[303176, 8765, 1811, 47, 29, "ProofInfo",ExpressionUUID->"2d77d2ad-50ea-45a7-8a70-0483b95c475c",
   Evaluatable->True,
   CellTags->{"Proof|ID:1642372132-1", "3"},
   CellID->259575914]},
 "\[SubsetEqual] singleton"->{
  Cell[305901, 8838, 1604, 54, 50, "FormalTextInputFormula",ExpressionUUID->"ba3c94b4-891e-4275-8bbf-58fc308ba014",
   CellTags->"\[SubsetEqual] singleton",
   CellID->574835025]},
 "singleton \[SubsetEqual]"->{
  Cell[307508, 8894, 1257, 40, 50, "FormalTextInputFormula",ExpressionUUID->"987ead35-8676-46b5-86a2-3f01a4952119",
   CellTags->"singleton \[SubsetEqual]",
   CellID->316263559]},
 "Proof|ID:574835025-1"->{
  Cell[309009, 8943, 1824, 47, 29, "ProofInfo",ExpressionUUID->"b866f190-0698-4884-9f66-30d4006e9fd2",
   Evaluatable->True,
   CellTags->{"Proof|ID:574835025-1", "3"},
   CellID->991759232]},
 "Proof|ID:316263559-1"->{
  Cell[310836, 8992, 1824, 47, 29, "ProofInfo",ExpressionUUID->"61354163-ddd7-4786-856c-489f55e87f87",
   Evaluatable->True,
   CellTags->{"Proof|ID:316263559-1", "3"},
   CellID->811999371]},
 "@I eq"->{
  Cell[313444, 9064, 1354, 43, 51, "FormalTextInputFormula",ExpressionUUID->"81f4ee41-7d1b-4661-86c1-cd51f6a030b1",
   CellTags->{"@I eq", "intro"},
   CellID->1595981918]},
 "@I eq reverse"->{
  Cell[314801, 9109, 1322, 42, 51, "FormalTextInputFormula",ExpressionUUID->"31466d48-898a-47c6-9519-350697bfe016",
   CellTags->"@I eq reverse",
   CellID->796685651]},
 "@I"->{
  Cell[316126, 9153, 1125, 35, 51, "FormalTextInputFormula",ExpressionUUID->"0ce71d43-56f3-43ae-aa1a-139f021caf78",
   CellTags->"@I",
   CellID->1022035164]},
 "@E"->{
  Cell[317254, 9190, 1528, 48, 55, "FormalTextInputFormula",ExpressionUUID->"27b53510-2aa4-42f2-ac57-bb62701ff6aa",
   CellTags->{"@E", "elim"},
   CellID->114344294]},
 "\[Element]@"->{
  Cell[318785, 9240, 1482, 49, 55, "FormalTextInputFormula",ExpressionUUID->"87637558-3a97-43a1-b39c-38a749176477",
   CellTags->"\[Element]@",
   CellID->1795699175]},
 "@ cong"->{
  Cell[320270, 9291, 1657, 54, 55, "FormalTextInputFormula",ExpressionUUID->"6066f4bb-70bd-49b6-9c7f-ad807e7e6782",
   CellTags->"@ cong",
   CellID->724997673]},
 "@ ident I"->{
  Cell[321930, 9347, 1425, 46, 55, "FormalTextInputFormula",ExpressionUUID->"90ad9d86-50dd-4746-b1e9-0a5d82ed98c4",
   CellTags->"@ ident I",
   CellID->803636397]},
 "Proof|ID:1595981918-1"->{
  Cell[323599, 9402, 1759, 45, 29, "ProofInfo",ExpressionUUID->"b5119a56-9b6b-4dc9-baca-4b038f13e49d",
   Evaluatable->True,
   CellTags->{"Proof|ID:1595981918-1", "3"},
   CellID->638235692]},
 "Proof|ID:796685651-1"->{
  Cell[325361, 9449, 1788, 46, 29, "ProofInfo",ExpressionUUID->"8db8df19-fdf9-4fa7-b82c-0b2e4f83d460",
   Evaluatable->True,
   CellTags->{"Proof|ID:796685651-1", "3"},
   CellID->1222338081]},
 "Proof|ID:1022035164-1"->{
  Cell[327152, 9497, 1754, 45, 29, "ProofInfo",ExpressionUUID->"6d754d5f-fbc5-4f0b-9d25-e0f40502431f",
   Evaluatable->True,
   CellTags->{"Proof|ID:1022035164-1", "3"},
   CellID->1120566427]},
 "Proof|ID:114344294-1"->{
  Cell[328909, 9544, 1751, 45, 29, "ProofInfo",ExpressionUUID->"0d43d638-054b-4e7b-a14f-d856c779b0cc",
   Evaluatable->True,
   CellTags->{"Proof|ID:114344294-1", "3"},
   CellID->1155434313]},
 "Proof|ID:1795699175-1"->{
  Cell[330663, 9591, 1787, 46, 29, "ProofInfo",ExpressionUUID->"fc5a6e81-1d94-4b7a-9f62-d18cc426bf95",
   Evaluatable->True,
   CellTags->{"Proof|ID:1795699175-1", "3"},
   CellID->1210186748]},
 "Proof|ID:724997673-1"->{
  Cell[332453, 9639, 1758, 45, 29, "ProofInfo",ExpressionUUID->"c828c5b9-b548-4ffe-a01b-69ead7885f9f",
   Evaluatable->True,
   CellTags->{"Proof|ID:724997673-1", "3"},
   CellID->637413599]},
 "Proof|ID:803636397-1"->{
  Cell[334214, 9686, 1765, 45, 29, "ProofInfo",ExpressionUUID->"ed4bf583-b592-4cce-934f-774b532b3b39",
   Evaluatable->True,
   CellTags->{"Proof|ID:803636397-1", "3"},
   CellID->1659057562]},
 "@\[SubsetEqual] I"->{
  Cell[336549, 9749, 1458, 46, 55, "FormalTextInputFormula",ExpressionUUID->"5cc6be16-8334-4ca9-b65d-a945b6a3bac8",
   CellTags->"@\[SubsetEqual] I",
   CellID->1322094901]},
 "\[SubsetEqual]@ E"->{
  Cell[338010, 9797, 1523, 47, 55, "FormalTextInputFormula",ExpressionUUID->"b58a06f0-0329-40ed-ba74-3f6c05ed1672",
   CellTags->"\[SubsetEqual]@ E",
   CellID->1691291058]},
 "@\[Intersection]"->{
  Cell[339536, 9846, 1016, 28, 51, "FormalTextInputFormula",ExpressionUUID->"270a4307-6358-4635-9c32-2bef59493519",
   CellTags->"@\[Intersection]",
   CellID->1830553618]},
 "@-"->{
  Cell[340555, 9876, 953, 27, 51, "FormalTextInputFormula",ExpressionUUID->"b21e6dec-dd9c-482a-aafe-5ea92234662d",
   CellTags->"@-",
   CellID->398205279]},
 "Proof|ID:1322094901-1"->{
  Cell[341753, 9912, 1813, 47, 29, "ProofInfo",ExpressionUUID->"57da981f-f479-4764-8da0-d6a7a3c080ec",
   Evaluatable->True,
   CellTags->{"Proof|ID:1322094901-1", "3"},
   CellID->892892992]},
 "Proof|ID:1691291058-1"->{
  Cell[343569, 9961, 1813, 47, 29, "ProofInfo",ExpressionUUID->"7588b18e-aa27-40ac-8f3f-f9e9ae729362",
   Evaluatable->True,
   CellTags->{"Proof|ID:1691291058-1", "3"},
   CellID->903150912]},
 "Proof|ID:1830553618-1"->{
  Cell[345385, 10010, 1812, 47, 29, "ProofInfo",ExpressionUUID->"85d59c5c-1a6e-4476-8410-ecd019893f54",
   Evaluatable->True,
   CellTags->{"Proof|ID:1830553618-1", "3"},
   CellID->1214598633]},
 "Proof|ID:398205279-1"->{
  Cell[347200, 10059, 1751, 45, 29, "ProofInfo",ExpressionUUID->"16659fac-0e9b-4219-a757-d5db7b81bc81",
   Evaluatable->True,
   CellTags->{"Proof|ID:398205279-1", "3"},
   CellID->1641967303]},
 "RangeI eq"->{
  Cell[350243, 10144, 1202, 37, 51, "FormalTextInputFormula",ExpressionUUID->"65d206a7-f149-48f0-8dc8-445143bf53a0",
   CellTags->{"RangeI eq", "intro"},
   CellID->686090881]},
 "RangeI"->{
  Cell[351448, 10183, 867, 24, 51, "FormalTextInputFormula",ExpressionUUID->"4f9de9e6-85d8-491a-9fd8-f47ebc444eae",
   CellTags->"RangeI",
   CellID->1169658618]},
 "RangeE"->{
  Cell[352318, 10209, 1319, 41, 55, "FormalTextInputFormula",ExpressionUUID->"a0d58dda-dbda-43ce-b842-8bf69646e1dc",
   CellTags->"RangeE",
   CellID->1129746621]},
 "Range singleton D"->{
  Cell[353640, 10252, 1198, 37, 51, "FormalTextInputFormula",ExpressionUUID->"35257160-6070-44f3-b494-ab56075ee0e6",
   CellTags->"Range singleton D",
   CellID->1805361717]},
 "Proof|ID:686090881-1"->{
  Cell[355083, 10298, 1765, 45, 29, "ProofInfo",ExpressionUUID->"6e01bd25-a1a6-443d-b697-ce5e66313c66",
   Evaluatable->True,
   CellTags->{"Proof|ID:686090881-1", "3"},
   CellID->1329697308]},
 "Proof|ID:1169658618-1"->{
  Cell[356851, 10345, 1760, 45, 29, "ProofInfo",ExpressionUUID->"8e614f57-fd84-4b20-aada-099cb8f5fd99",
   Evaluatable->True,
   CellTags->{"Proof|ID:1169658618-1", "3"},
   CellID->65583140]},
 "Proof|ID:1129746621-1"->{
  Cell[358614, 10392, 1762, 45, 29, "ProofInfo",ExpressionUUID->"ec01b3f4-f18b-473a-adaa-b64c4b1794c1",
   Evaluatable->True,
   CellTags->{"Proof|ID:1129746621-1", "3"},
   CellID->1205185338]},
 "Proof|ID:1805361717-1"->{
  Cell[360379, 10439, 1813, 47, 29, "ProofInfo",ExpressionUUID->"796ac993-5d77-429a-8ec4-94a6354c467e",
   Evaluatable->True,
   CellTags->{"Proof|ID:1805361717-1", "3"},
   CellID->774665428]},
 "\[Element]SetOf"->{
  Cell[362993, 10511, 1609, 53, 62, "FormalTextInputFormula",ExpressionUUID->"7a7bc97e-48fb-447b-ba94-7c245728934d",
   CellTags->"\[Element]SetOf",
   CellID->718424401]},
 "SetOf I"->{
  Cell[364605, 10566, 1376, 42, 64, "FormalTextInputFormula",ExpressionUUID->"ba2298fa-7bbb-45c0-b5bb-c841cdf4f95a",
   CellTags->{"SetOf I", "intro"},
   CellID->783246906]},
 "SetOf E"->{
  Cell[365984, 10610, 1544, 47, 64, "FormalTextInputFormula",ExpressionUUID->"6eb18b5a-43f0-4bf0-91cb-fa1b2deac1c6",
   CellTags->{"SetOf E", "elim"},
   CellID->1984656332]},
 "Proof|ID:718424401-1"->{
  Cell[367772, 10666, 1791, 46, 29, "ProofInfo",ExpressionUUID->"62e40651-f209-4327-b387-490f4c46bb3d",
   Evaluatable->True,
   CellTags->{"Proof|ID:718424401-1", "3"},
   CellID->961492024]},
 "Proof|ID:783246906-1"->{
  Cell[369566, 10714, 1761, 45, 29, "ProofInfo",ExpressionUUID->"8842b459-89de-4cca-be91-b8c2e9a5746a",
   Evaluatable->True,
   CellTags->{"Proof|ID:783246906-1", "3"},
   CellID->1483483416]},
 "Proof|ID:1984656332-1"->{
  Cell[371330, 10761, 1764, 45, 29, "ProofInfo",ExpressionUUID->"3432589c-d2fc-4a23-a0c8-b28e516663cd",
   Evaluatable->True,
   CellTags->{"Proof|ID:1984656332-1", "3"},
   CellID->1243944656]},
 "SetOf\[Equal]Collect"->{
  Cell[374181, 10839, 1049, 31, 62, "FormalTextInputFormula",ExpressionUUID->"03aff8c6-ce79-4896-9c4b-ff5194aa4795",
   CellTags->"SetOf\[Equal]Collect",
   CellID->314660242]},
 "SetOf const"->{
  Cell[375233, 10872, 1279, 40, 62, "FormalTextInputFormula",ExpressionUUID->"43821a10-5e3f-462a-b9d8-7e55cb502399",
   CellTags->"SetOf const",
   CellID->222426804]},
 "SetOf\[Equal]{} iff"->{
  Cell[376515, 10914, 1585, 52, 62, "FormalTextInputFormula",ExpressionUUID->"7a67e7ee-094b-4cec-89ae-429266833c16",
   CellTags->"SetOf\[Equal]{} iff",
   CellID->1237361888]},
 "SetOf\[Not]"->{
  Cell[378103, 10968, 1300, 42, 62, "FormalTextInputFormula",ExpressionUUID->"bc81c816-825b-4560-8ed5-0038c9d1cc89",
   CellTags->"SetOf\[Not]",
   CellID->165006998]},
 "SetOf\[Or]"->{
  Cell[379406, 11012, 1732, 56, 62, "FormalTextInputFormula",ExpressionUUID->"3e4ae205-39dd-4613-a34d-75a3ca68670a",
   CellTags->"SetOf\[Or]",
   CellID->1855260115]},
 "SetOf\[Implies]"->{
  Cell[381141, 11070, 1710, 55, 62, "FormalTextInputFormula",ExpressionUUID->"ca2c40f5-3c9f-4d06-818b-6843e4fc238f",
   CellTags->"SetOf\[Implies]",
   CellID->985026514]},
 "Collect\[And]"->{
  Cell[382854, 11127, 1734, 54, 62, "FormalTextInputFormula",ExpressionUUID->"e7181290-0067-4cd3-bfdd-00152da0d246",
   CellTags->"Collect\[And]",
   CellID->636400335]},
 "SetOf mono iff"->{
  Cell[384591, 11183, 1589, 51, 62, "FormalTextInputFormula",ExpressionUUID->"8003e81d-2011-440e-b291-7344a40018ad",
   CellTags->"SetOf mono iff",
   CellID->859120159]},
 "Proof|ID:314660242-1"->{
  Cell[386423, 11243, 1816, 47, 29, "ProofInfo",ExpressionUUID->"90ce6af8-8154-479f-a67a-eb82f1369496",
   Evaluatable->True,
   CellTags->{"Proof|ID:314660242-1", "3"},
   CellID->462247040]},
 "Proof|ID:222426804-1"->{
  Cell[388242, 11292, 1784, 46, 29, "ProofInfo",ExpressionUUID->"23123a44-70d0-4b7e-af9d-66586f2c7c3b",
   Evaluatable->True,
   CellTags->{"Proof|ID:222426804-1", "3"},
   CellID->1447843469]},
 "Proof|ID:1237361888-1"->{
  Cell[390029, 11340, 1817, 47, 29, "ProofInfo",ExpressionUUID->"216bcc67-94ae-452d-9f9d-fb1577639bae",
   Evaluatable->True,
   CellTags->{"Proof|ID:1237361888-1", "3"},
   CellID->779088903]},
 "Proof|ID:165006998-1"->{
  Cell[391849, 11389, 1784, 46, 29, "ProofInfo",ExpressionUUID->"5156a1b2-c4e6-447c-8758-161115f33ced",
   Evaluatable->True,
   CellTags->{"Proof|ID:165006998-1", "3"},
   CellID->1390818865]},
 "Proof|ID:1855260115-1"->{
  Cell[393636, 11437, 1784, 46, 29, "ProofInfo",ExpressionUUID->"627a6140-2e49-4960-8e4a-725e778a5b7a",
   Evaluatable->True,
   CellTags->{"Proof|ID:1855260115-1", "3"},
   CellID->859972040]},
 "Proof|ID:985026514-1"->{
  Cell[395423, 11485, 1792, 46, 29, "ProofInfo",ExpressionUUID->"d35e54e6-1cd4-4c97-a6a7-67a76bf2b9ca",
   Evaluatable->True,
   CellTags->{"Proof|ID:985026514-1", "3"},
   CellID->2146486548]},
 "Proof|ID:636400335-1"->{
  Cell[397218, 11533, 1788, 46, 29, "ProofInfo",ExpressionUUID->"3f1308e2-2bf7-43a6-b603-e554f2db72a9",
   Evaluatable->True,
   CellTags->{"Proof|ID:636400335-1", "3"},
   CellID->1733058477]},
 "Proof|ID:859120159-1"->{
  Cell[399009, 11581, 1790, 46, 29, "ProofInfo",ExpressionUUID->"43fc8102-da0a-4d32-8477-1d0fb499a635",
   Evaluatable->True,
   CellTags->{"Proof|ID:859120159-1", "3"},
   CellID->1491492338]},
 "\[SubsetEqual]{}"->{
  Cell[401753, 11654, 1237, 40, 47, "FormalTextInputFormula",ExpressionUUID->"3055a66b-2e7f-420c-b9ad-a698d5a357e9",
   CellTags->"\[SubsetEqual]{}",
   CellID->485185733]},
 "\[ForAll]\[NotElement] conv"->{
  Cell[402993, 11696, 1290, 41, 51, "FormalTextInputFormula",ExpressionUUID->"6cb5ed56-c05b-44f8-91f8-042d5b96b3ca",
   CellTags->"\[ForAll]\[NotElement] conv",
   CellID->775376237]},
 "\[Exists]\[Element] conv"->{
  Cell[404286, 11739, 1286, 41, 51, "FormalTextInputFormula",ExpressionUUID->"c11276ab-818d-47e0-906b-fa4426b2d132",
   CellTags->"\[Exists]\[Element] conv",
   CellID->978065715]},
 "Proof|ID:485185733-1"->{
  Cell[405817, 11789, 1808, 47, 29, "ProofInfo",ExpressionUUID->"533d16e0-ce09-469a-92e6-88dd693c1a07",
   Evaluatable->True,
   CellTags->{"Proof|ID:485185733-1", "3"},
   CellID->851152370]},
 "Proof|ID:775376237-1"->{
  Cell[407628, 11838, 1830, 47, 29, "ProofInfo",ExpressionUUID->"d64e40de-107c-465e-a04b-e0d1316fc389",
   Evaluatable->True,
   CellTags->{"Proof|ID:775376237-1", "3"},
   CellID->497842656]},
 "Proof|ID:978065715-1"->{
  Cell[409461, 11887, 1824, 47, 29, "ProofInfo",ExpressionUUID->"3dea71aa-c69b-4215-9524-803e9c59e9f2",
   Evaluatable->True,
   CellTags->{"Proof|ID:978065715-1", "3"},
   CellID->118302286]},
 "Insert UNIV"->{
  Cell[412254, 11961, 949, 28, 47, "FormalTextInputFormula",ExpressionUUID->"2d432965-27dc-4fc0-99e8-0b89f9dcb74f",
   CellTags->"Insert UNIV",
   CellID->988950467]},
 "Insert\[Equal]\[Union]"->{
  Cell[413206, 11991, 1048, 31, 50, "FormalTextInputFormula",ExpressionUUID->"55070231-9ff3-4351-bf0a-aab3ba4c1427",
   CellTags->"Insert\[Equal]\[Union]",
   CellID->1780572558]},
 "Insert\[NotEqual]{}"->{
  Cell[414257, 12024, 1007, 30, 50, "FormalTextInputFormula",ExpressionUUID->"7464ffb2-77ec-46ff-a00f-bf067e26bbb5",
   CellTags->"Insert\[NotEqual]{}",
   CellID->1795020389]},
 "Insert absorb"->{
  Cell[415267, 12056, 1155, 36, 50, "FormalTextInputFormula",ExpressionUUID->"7472c75c-9564-488e-935f-bd257382c5ec",
   CellTags->"Insert absorb",
   CellID->2125656893]},
 "Insert absorb2"->{
  Cell[416425, 12094, 1099, 33, 50, "FormalTextInputFormula",ExpressionUUID->"f27f74b0-b865-4980-83e7-8f44091d7719",
   CellTags->"Insert absorb2",
   CellID->742456464]},
 "Insert comm"->{
  Cell[417527, 12129, 1171, 35, 51, "FormalTextInputFormula",ExpressionUUID->"fd79fcdc-631d-4c21-9d76-e467595a23ef",
   CellTags->"Insert comm",
   CellID->1038355606]},
 "Insert SetOf"->{
  Cell[418701, 12166, 1392, 44, 62, "FormalTextInputFormula",ExpressionUUID->"34779372-bb66-44fe-b2b6-9f99f0d10da0",
   CellTags->"Insert SetOf",
   CellID->657964489]},
 "Insert\[Intersection]Insert"->{
  Cell[420096, 12212, 1326, 40, 50, "FormalTextInputFormula",ExpressionUUID->"b767ae43-32fd-4d40-9226-48275463435c",
   CellTags->"Insert\[Intersection]Insert",
   CellID->1602428504]},
 "Insert disjoint"->{
  Cell[421425, 12254, 1599, 52, 50, "FormalTextInputFormula",ExpressionUUID->"be8df8a3-7c41-4241-8b84-f2d34ca0f3fb",
   CellTags->"Insert disjoint",
   CellID->621646007]},
 "disjoint Insert"->{
  Cell[423027, 12308, 1601, 52, 50, "FormalTextInputFormula",ExpressionUUID->"c9610130-bb90-4d52-a0f4-3b30a68d9d28",
   CellTags->"disjoint Insert",
   CellID->656967390]},
 "Proof|ID:988950467-1"->{
  Cell[424874, 12369, 1783, 46, 29, "ProofInfo",ExpressionUUID->"711e6551-2491-4f23-a69e-6ba95e9ee3c5",
   Evaluatable->True,
   CellTags->{"Proof|ID:988950467-1", "3"},
   CellID->207220836]},
 "Proof|ID:1780572558-1"->{
  Cell[426660, 12417, 1823, 47, 29, "ProofInfo",ExpressionUUID->"ce02b660-2efc-48d5-a410-83f64d3cb0a4",
   Evaluatable->True,
   CellTags->{"Proof|ID:1780572558-1", "3"},
   CellID->709620784]},
 "Proof|ID:1795020389-1"->{
  Cell[428486, 12466, 1817, 47, 29, "ProofInfo",ExpressionUUID->"a593fce4-239a-48ba-a634-0803bc3ce5f5",
   Evaluatable->True,
   CellTags->{"Proof|ID:1795020389-1", "3"},
   CellID->201322071]},
 "Proof|ID:2125656893-1"->{
  Cell[430306, 12515, 1791, 46, 29, "ProofInfo",ExpressionUUID->"1c164155-3908-47ea-80ff-28bb818b47e3",
   Evaluatable->True,
   CellTags->{"Proof|ID:2125656893-1", "3"},
   CellID->1679122200]},
 "Proof|ID:742456464-1"->{
  Cell[432100, 12563, 1789, 46, 29, "ProofInfo",ExpressionUUID->"fdfb52c1-1f7b-42b0-8e1f-bfb7551b6086",
   Evaluatable->True,
   CellTags->{"Proof|ID:742456464-1", "3"},
   CellID->292466109]},
 "Proof|ID:1038355606-1"->{
  Cell[433892, 12611, 1786, 46, 29, "ProofInfo",ExpressionUUID->"0017ff7e-6593-4c7d-a1d7-1bfa18dabc62",
   Evaluatable->True,
   CellTags->{"Proof|ID:1038355606-1", "3"},
   CellID->600513012]},
 "Proof|ID:657964489-1"->{
  Cell[435681, 12659, 1786, 46, 29, "ProofInfo",ExpressionUUID->"864a6093-7b93-41ad-b279-fcdce277baa9",
   Evaluatable->True,
   CellTags->{"Proof|ID:657964489-1", "3"},
   CellID->1781110406]},
 "Proof|ID:1602428504-1"->{
  Cell[437470, 12707, 1834, 47, 29, "ProofInfo",ExpressionUUID->"a1dd04b2-c9d9-4f5c-a60f-43df42997afa",
   Evaluatable->True,
   CellTags->{"Proof|ID:1602428504-1", "3"},
   CellID->1228710860]},
 "Proof|ID:621646007-1"->{
  Cell[439307, 12756, 1792, 46, 29, "ProofInfo",ExpressionUUID->"7df4f81d-c45a-43a2-a483-44ffc18f92a3",
   Evaluatable->True,
   CellTags->{"Proof|ID:621646007-1", "3"},
   CellID->1551927440]},
 "Proof|ID:656967390-1"->{
  Cell[441102, 12804, 1792, 46, 29, "ProofInfo",ExpressionUUID->"5a84cad7-a764-4fb6-b914-1258f0a32f54",
   Evaluatable->True,
   CellTags->{"Proof|ID:656967390-1", "3"},
   CellID->1616332482]},
 "\[Intersection] absorb"->{
  Cell[444152, 12881, 945, 27, 47, "FormalTextInputFormula",ExpressionUUID->"fc55d8fa-8b90-4254-897c-2a6575a0f480",
   CellTags->"\[Intersection] absorb",
   CellID->125816224]},
 "\[Intersection] absorb'"->{
  Cell[445100, 12910, 1074, 31, 50, "FormalTextInputFormula",ExpressionUUID->"78e30461-db5c-440a-a771-68e1af989dd5",
   CellTags->"\[Intersection] absorb'",
   CellID->848809596]},
 "\[Intersection] comm"->{
  Cell[446177, 12943, 1002, 29, 50, "FormalTextInputFormula",ExpressionUUID->"ed2c3178-573c-441d-b883-c094a6cdf7fc",
   CellTags->"\[Intersection] comm",
   CellID->147215930]},
 "\[Intersection] comm'"->{
  Cell[447182, 12974, 1148, 33, 50, "FormalTextInputFormula",ExpressionUUID->"9303ebaf-4bff-47f8-b699-5c2016271cf7",
   CellTags->"\[Intersection] comm'",
   CellID->902394792]},
 "\[Intersection] assoc"->{
  Cell[448333, 13009, 1153, 34, 50, "FormalTextInputFormula",ExpressionUUID->"201ed866-e06f-4435-b662-35ed14257a22",
   CellTags->"\[Intersection] assoc",
   CellID->568250332]},
 "\[Intersection] absorb1"->{
  Cell[449489, 13045, 1151, 35, 50, "FormalTextInputFormula",ExpressionUUID->"58f867f3-5e00-4674-b9e0-286a263e5f89",
   CellTags->"\[Intersection] absorb1",
   CellID->1826338130]},
 "\[Intersection] absorb2"->{
  Cell[450643, 13082, 1153, 35, 50, "FormalTextInputFormula",ExpressionUUID->"9457e9e2-d7e8-4e67-a0fb-022e6c713401",
   CellTags->"\[Intersection] absorb2",
   CellID->1588902597]},
 "\[Intersection]{}"->{
  Cell[451799, 13119, 975, 29, 47, "FormalTextInputFormula",ExpressionUUID->"ba8fd0f4-f2cd-4892-b0e0-8f45db7178ec",
   CellTags->"\[Intersection]{}",
   CellID->1253019475]},
 "{}\[Intersection]"->{
  Cell[452777, 13150, 975, 29, 47, "FormalTextInputFormula",ExpressionUUID->"a5ac7dd9-350f-4831-9b39-0ac63d3d6e99",
   CellTags->"{}\[Intersection]",
   CellID->1274251332]},
 "disjoint iff \[SubsetEqual]-"->{
  Cell[453755, 13181, 1324, 42, 50, "FormalTextInputFormula",ExpressionUUID->"6135b89a-9bc7-4d84-b9ab-fd95519b0153",
   CellTags->"disjoint iff \[SubsetEqual]-",
   CellID->903302990]},
 "disjoint iff \[NotEqual]"->{
  Cell[455082, 13225, 1482, 47, 57, "FormalTextInputFormula",ExpressionUUID->"76d4823b-68ee-4ac8-be6a-79c20b3ff6fb",
   CellTags->"disjoint iff \[NotEqual]",
   CellID->920640376]},
 "\[Intersection]UNIV"->{
  Cell[456567, 13274, 945, 27, 47, "FormalTextInputFormula",ExpressionUUID->"7810a2e8-7dec-41dc-8f02-f32b02736408",
   CellTags->"\[Intersection]UNIV",
   CellID->1023611974]},
 "UNIV\[Intersection]"->{
  Cell[457515, 13303, 943, 27, 47, "FormalTextInputFormula",ExpressionUUID->"187c8311-a111-480e-87b9-378d25295ce1",
   CellTags->"UNIV\[Intersection]",
   CellID->1195769099]},
 "\[Intersection]\[Union] dist left"->{
  Cell[458461, 13332, 1226, 35, 50, "FormalTextInputFormula",ExpressionUUID->"aec71586-6dd1-456e-8894-29dcd9143ebd",
   CellTags->"\[Intersection]\[Union] dist left",
   CellID->56424478]},
 "\[Intersection]\[Union] dist right"->{
  Cell[459690, 13369, 1235, 36, 50, "FormalTextInputFormula",ExpressionUUID->"252a8781-d3e3-43ef-8e2d-c57e47347aba",
   CellTags->"\[Intersection]\[Union] dist right",
   CellID->1624124987]},
 "\[Intersection]\[Equal]UNIV"->{
  Cell[460928, 13407, 1575, 52, 50, "FormalTextInputFormula",ExpressionUUID->"c2d8d38f-e73a-4c20-90e9-10e81ae9316a",
   CellTags->"\[Intersection]\[Equal]UNIV",
   CellID->158694025]},
 "\[SubsetEqual]\[Intersection]"->{
  Cell[462506, 13461, 1369, 42, 50, "FormalTextInputFormula",ExpressionUUID->"b2e42c5d-5b00-41ba-bc88-ffaa02e6c6f4",
   CellTags->"\[SubsetEqual]\[Intersection]",
   CellID->1837534884]},
 "Insert\[Intersection]"->{
  Cell[463878, 13505, 1485, 45, 58, "FormalTextInputFormula",ExpressionUUID->"37035f2d-8be1-4d9b-b346-2538b15706a2",
   CellTags->"Insert\[Intersection]",
   CellID->229432372]},
 "\[Intersection]Insert"->{
  Cell[465366, 13552, 1488, 45, 58, "FormalTextInputFormula",ExpressionUUID->"5baff602-e322-4acf-a322-433104f4e54a",
   CellTags->"\[Intersection]Insert",
   CellID->2031161567]},
 "Proof|ID:125816224-1"->{
  Cell[467098, 13606, 1821, 47, 29, "ProofInfo",ExpressionUUID->"cf2a0b05-947e-4e08-b681-dbb9bb955bbf",
   Evaluatable->True,
   CellTags->{"Proof|ID:125816224-1", "3"},
   CellID->1628613977]},
 "Proof|ID:848809596-1"->{
  Cell[468922, 13655, 1823, 47, 29, "ProofInfo",ExpressionUUID->"36472be4-f0d7-4752-bf1f-5e56ab4cfee6",
   Evaluatable->True,
   CellTags->{"Proof|ID:848809596-1", "3"},
   CellID->2030772045]},
 "Proof|ID:147215930-1"->{
  Cell[470748, 13704, 1816, 47, 29, "ProofInfo",ExpressionUUID->"e0bc9de7-8d38-449f-8c01-166784414f07",
   Evaluatable->True,
   CellTags->{"Proof|ID:147215930-1", "3"},
   CellID->548669029]},
 "Proof|ID:902394792-1"->{
  Cell[472567, 13753, 1819, 47, 29, "ProofInfo",ExpressionUUID->"0c6b49b0-74e6-4b95-a486-249347bf00e4",
   Evaluatable->True,
   CellTags->{"Proof|ID:902394792-1", "3"},
   CellID->1223279036]},
 "Proof|ID:568250332-1"->{
  Cell[474389, 13802, 1818, 47, 29, "ProofInfo",ExpressionUUID->"c9ae36de-51e6-4ab1-af37-a9c0f0d09f5f",
   Evaluatable->True,
   CellTags->{"Proof|ID:568250332-1", "3"},
   CellID->868040367]},
 "Proof|ID:1826338130-1"->{
  Cell[476210, 13851, 1825, 47, 29, "ProofInfo",ExpressionUUID->"263a06fe-fc74-4a56-8db9-29590e7b58ff",
   Evaluatable->True,
   CellTags->{"Proof|ID:1826338130-1", "3"},
   CellID->679262902]},
 "Proof|ID:1588902597-1"->{
  Cell[478038, 13900, 1826, 47, 29, "ProofInfo",ExpressionUUID->"e1bfc212-a53d-43e0-b061-4f254f379381",
   Evaluatable->True,
   CellTags->{"Proof|ID:1588902597-1", "3"},
   CellID->1210453596]},
 "Proof|ID:1253019475-1"->{
  Cell[479867, 13949, 1814, 47, 29, "ProofInfo",ExpressionUUID->"1a5c661b-1557-4af5-b166-96572bd68d8c",
   Evaluatable->True,
   CellTags->{"Proof|ID:1253019475-1", "3"},
   CellID->1968362112]},
 "Proof|ID:1274251332-1"->{
  Cell[481684, 13998, 1814, 47, 29, "ProofInfo",ExpressionUUID->"3cc072c7-64ff-40a0-8749-2ddcf91e538c",
   Evaluatable->True,
   CellTags->{"Proof|ID:1274251332-1", "3"},
   CellID->2131093972]},
 "Proof|ID:903302990-1"->{
  Cell[483501, 14047, 1831, 47, 29, "ProofInfo",ExpressionUUID->"8a174da6-9259-4456-a888-bca8c161fadd",
   Evaluatable->True,
   CellTags->{"Proof|ID:903302990-1", "3"},
   CellID->72270293]},
 "Proof|ID:920640376-1"->{
  Cell[485335, 14096, 1824, 47, 29, "ProofInfo",ExpressionUUID->"677ec8e8-79c5-4849-ab27-fd71ffc1c532",
   Evaluatable->True,
   CellTags->{"Proof|ID:920640376-1", "3"},
   CellID->321131777]},
 "Proof|ID:1023611974-1"->{
  Cell[487162, 14145, 1818, 47, 29, "ProofInfo",ExpressionUUID->"777b4649-e951-4bda-9678-a8b639463f5b",
   Evaluatable->True,
   CellTags->{"Proof|ID:1023611974-1", "3"},
   CellID->1877569687]},
 "Proof|ID:1195769099-1"->{
  Cell[488983, 14194, 1818, 47, 29, "ProofInfo",ExpressionUUID->"8338a43c-ea58-4e7d-85ec-4622d938afa3",
   Evaluatable->True,
   CellTags->{"Proof|ID:1195769099-1", "3"},
   CellID->1618369204]},
 "Proof|ID:56424478-1"->{
  Cell[490804, 14243, 1836, 47, 29, "ProofInfo",ExpressionUUID->"f1380f81-fa44-4763-b4e5-077e4ac37ebd",
   Evaluatable->True,
   CellTags->{"Proof|ID:56424478-1", "3"},
   CellID->400332023]},
 "Proof|ID:1624124987-1"->{
  Cell[492643, 14292, 1848, 47, 29, "ProofInfo",ExpressionUUID->"f946aa9b-4bc5-4b6b-a6d6-94285e3a76b4",
   Evaluatable->True,
   CellTags->{"Proof|ID:1624124987-1", "3"},
   CellID->1734538043]},
 "Proof|ID:158694025-1"->{
  Cell[494494, 14341, 1831, 47, 29, "ProofInfo",ExpressionUUID->"edb892ab-bd7f-474d-8831-ba95ca7c0306",
   Evaluatable->True,
   CellTags->{"Proof|ID:158694025-1", "3"},
   CellID->1315875576]},
 "Proof|ID:1837534884-1"->{
  Cell[496328, 14390, 1837, 47, 29, "ProofInfo",ExpressionUUID->"49d48a11-4604-4d84-b391-5e5974a2445a",
   Evaluatable->True,
   CellTags->{"Proof|ID:1837534884-1", "3"},
   CellID->978925165]},
 "Proof|ID:229432372-1"->{
  Cell[498168, 14439, 1818, 47, 29, "ProofInfo",ExpressionUUID->"4ac23cf1-4b24-4d31-90fa-4a0775d734ae",
   Evaluatable->True,
   CellTags->{"Proof|ID:229432372-1", "3"},
   CellID->387372574]},
 "Proof|ID:2031161567-1"->{
  Cell[499989, 14488, 1821, 47, 29, "ProofInfo",ExpressionUUID->"812c3dd1-7e84-453c-b984-6de8b7f632e4",
   Evaluatable->True,
   CellTags->{"Proof|ID:2031161567-1", "3"},
   CellID->895002178]},
 "\[Union] absorb"->{
  Cell[503127, 14567, 924, 27, 47, "FormalTextInputFormula",ExpressionUUID->"1dd930a7-a4a4-4184-a990-a1221b81347e",
   CellTags->"\[Union] absorb",
   CellID->510851566]},
 "\[Union] absorb'"->{
  Cell[504054, 14596, 1040, 31, 50, "FormalTextInputFormula",ExpressionUUID->"30dd6ca7-e166-4968-b230-3695f88edcbe",
   CellTags->"\[Union] absorb'",
   CellID->961990306]},
 "\[Union] comm"->{
  Cell[505097, 14629, 977, 29, 50, "FormalTextInputFormula",ExpressionUUID->"fffc7ebe-1903-470d-a423-01e5da9bdb37",
   CellTags->"\[Union] comm",
   CellID->2140682664]},
 "\[Union] comm'"->{
  Cell[506077, 14660, 1109, 33, 50, "FormalTextInputFormula",ExpressionUUID->"63cba1dd-7f2a-42ee-bd5b-dd610f3b54b3",
   CellTags->"\[Union] comm'",
   CellID->1655518259]},
 "\[Union] assoc"->{
  Cell[507189, 14695, 1106, 33, 50, "FormalTextInputFormula",ExpressionUUID->"e8e6d705-631c-4c9d-aa97-fefca4115ead",
   CellTags->"\[Union] assoc",
   CellID->945600630]},
 "\[Union] absorb1"->{
  Cell[508298, 14730, 1130, 35, 50, "FormalTextInputFormula",ExpressionUUID->"ad5beaab-93b9-48c6-b2c2-a8f8c1ff258d",
   CellTags->"\[Union] absorb1",
   CellID->837689454]},
 "\[Union] absorb2"->{
  Cell[509431, 14767, 1132, 35, 50, "FormalTextInputFormula",ExpressionUUID->"38deb9e2-550b-48aa-8740-428d4796982e",
   CellTags->"\[Union] absorb2",
   CellID->1727788553]},
 "\[Union]{}"->{
  Cell[510566, 14804, 935, 28, 47, "FormalTextInputFormula",ExpressionUUID->"d6414077-67ef-4ac9-8836-f56a0138097e",
   CellTags->"\[Union]{}",
   CellID->985635433]},
 "{}\[Union]"->{
  Cell[511504, 14834, 936, 28, 47, "FormalTextInputFormula",ExpressionUUID->"f0351117-b62d-4593-804a-5846ee59e902",
   CellTags->"{}\[Union]",
   CellID->1158821232]},
 "\[Union]UNIV"->{
  Cell[512443, 14864, 925, 27, 47, "FormalTextInputFormula",ExpressionUUID->"cc76f1fe-6b2e-417e-b83a-4a1e85689882",
   CellTags->"\[Union]UNIV",
   CellID->1958752872]},
 "UNIV\[Union]"->{
  Cell[513371, 14893, 924, 27, 47, "FormalTextInputFormula",ExpressionUUID->"2c1fe514-2923-4862-ae06-ebdba0500fae",
   CellTags->"UNIV\[Union]",
   CellID->715721472]},
 "Insert\[Union]"->{
  Cell[514298, 14922, 1112, 33, 50, "FormalTextInputFormula",ExpressionUUID->"88ad4619-c3ec-44e6-9cc9-c15783ae50f8",
   CellTags->"Insert\[Union]",
   CellID->727949230]},
 "\[Union]Insert"->{
  Cell[515413, 14957, 1114, 33, 50, "FormalTextInputFormula",ExpressionUUID->"1108794c-416f-48fd-94ec-fb8fcba1ca2e",
   CellTags->"\[Union]Insert",
   CellID->1429638428]},
 "\[Union]\[Intersection] dist left"->{
  Cell[516530, 14992, 1221, 35, 50, "FormalTextInputFormula",ExpressionUUID->"08d29689-ef28-4a06-890c-f9306ee30fd1",
   CellTags->"\[Union]\[Intersection] dist left",
   CellID->1116053495]},
 "\[Union]\[Intersection] dist right"->{
  Cell[517754, 15029, 1225, 36, 50, "FormalTextInputFormula",ExpressionUUID->"b2ce6988-93f6-4266-a5cd-26dc6dbf5711",
   CellTags->"\[Union]\[Intersection] dist right",
   CellID->1097000890]},
 "\[Union]\[Equal]{}"->{
  Cell[518982, 15067, 1612, 55, 50, "FormalTextInputFormula",ExpressionUUID->"36426bb0-b926-47c4-aeb4-4508b486267d",
   CellTags->"\[Union]\[Equal]{}",
   CellID->1016692823]},
 "\[Union]\[SubsetEqual]"->{
  Cell[520597, 15124, 1345, 42, 50, "FormalTextInputFormula",ExpressionUUID->"c2879c31-4a67-4636-b191-393fcad6e26d",
   CellTags->"\[Union]\[SubsetEqual]",
   CellID->997515848]},
 "Proof|ID:510851566-1"->{
  Cell[522187, 15175, 1791, 46, 29, "ProofInfo",ExpressionUUID->"eb249535-8173-40d6-bbdb-2c2aae352b07",
   Evaluatable->True,
   CellTags->{"Proof|ID:510851566-1", "3"},
   CellID->365686177]},
 "Proof|ID:961990306-1"->{
  Cell[523981, 15223, 1809, 47, 29, "ProofInfo",ExpressionUUID->"f296832d-3220-4ad0-9d80-345f79408b1f",
   Evaluatable->True,
   CellTags->{"Proof|ID:961990306-1", "3"},
   CellID->1321288291]},
 "Proof|ID:2140682664-1"->{
  Cell[525793, 15272, 1790, 46, 29, "ProofInfo",ExpressionUUID->"e2b5c4d8-ca86-4739-8add-11abf118bb58",
   Evaluatable->True,
   CellTags->{"Proof|ID:2140682664-1", "3"},
   CellID->678754593]},
 "Proof|ID:1655518259-1"->{
  Cell[527586, 15320, 1793, 46, 29, "ProofInfo",ExpressionUUID->"5592b9fd-2693-411b-becf-a4a020ad57be",
   Evaluatable->True,
   CellTags->{"Proof|ID:1655518259-1", "3"},
   CellID->1904743954]},
 "Proof|ID:945600630-1"->{
  Cell[529382, 15368, 1790, 46, 29, "ProofInfo",ExpressionUUID->"3fc23a3b-63ed-4e0f-9760-2469f4d0179b",
   Evaluatable->True,
   CellTags->{"Proof|ID:945600630-1", "3"},
   CellID->1060575105]},
 "Proof|ID:837689454-1"->{
  Cell[531175, 15416, 1808, 47, 29, "ProofInfo",ExpressionUUID->"a45e0c3a-1b21-41cc-887a-dad678f3e9fe",
   Evaluatable->True,
   CellTags->{"Proof|ID:837689454-1", "3"},
   CellID->363045913]},
 "Proof|ID:1727788553-1"->{
  Cell[532986, 15465, 1810, 47, 29, "ProofInfo",ExpressionUUID->"ccf3a6e2-23b5-4cd1-a76c-3ffdf84cc204",
   Evaluatable->True,
   CellTags->{"Proof|ID:1727788553-1", "3"},
   CellID->56107910]},
 "Proof|ID:985635433-1"->{
  Cell[534799, 15514, 1782, 46, 29, "ProofInfo",ExpressionUUID->"e1d8893c-5ca6-4930-940f-940100ee72b4",
   Evaluatable->True,
   CellTags->{"Proof|ID:985635433-1", "3"},
   CellID->1189519127]},
 "Proof|ID:1158821232-1"->{
  Cell[536584, 15562, 1784, 46, 29, "ProofInfo",ExpressionUUID->"84574fea-7e8d-4451-afb5-958ee98ab045",
   Evaluatable->True,
   CellTags->{"Proof|ID:1158821232-1", "3"},
   CellID->772575899]},
 "Proof|ID:1958752872-1"->{
  Cell[538371, 15610, 1789, 46, 29, "ProofInfo",ExpressionUUID->"57de62df-9c50-4468-8a9b-27db3a02f067",
   Evaluatable->True,
   CellTags->{"Proof|ID:1958752872-1", "3"},
   CellID->1073197046]},
 "Proof|ID:715721472-1"->{
  Cell[540163, 15658, 1785, 46, 29, "ProofInfo",ExpressionUUID->"091f7b25-0bd4-44cb-ae0f-90617fc64236",
   Evaluatable->True,
   CellTags->{"Proof|ID:715721472-1", "3"},
   CellID->645311516]},
 "Proof|ID:727949230-1"->{
  Cell[541951, 15706, 1790, 46, 29, "ProofInfo",ExpressionUUID->"f96d0823-3864-4900-af03-5095e0ae6fd2",
   Evaluatable->True,
   CellTags->{"Proof|ID:727949230-1", "3"},
   CellID->1612445395]},
 "Proof|ID:1429638428-1"->{
  Cell[543744, 15754, 1792, 46, 29, "ProofInfo",ExpressionUUID->"e1b4be1f-45b6-489f-ae2b-a23d03f34073",
   Evaluatable->True,
   CellTags->{"Proof|ID:1429638428-1", "3"},
   CellID->788463628]},
 "Proof|ID:1116053495-1"->{
  Cell[545539, 15802, 1845, 47, 29, "ProofInfo",ExpressionUUID->"10ae46d4-043b-4352-a8ed-04644b597001",
   Evaluatable->True,
   CellTags->{"Proof|ID:1116053495-1", "3"},
   CellID->826786679]},
 "Proof|ID:1097000890-1"->{
  Cell[547387, 15851, 1848, 47, 29, "ProofInfo",ExpressionUUID->"5e24b08f-bb6e-4cbf-80ba-ba0b65e7b37f",
   Evaluatable->True,
   CellTags->{"Proof|ID:1097000890-1", "3"},
   CellID->1930524663]},
 "Proof|ID:1016692823-1"->{
  Cell[549238, 15900, 1816, 47, 29, "ProofInfo",ExpressionUUID->"4e864b7c-43bd-41bd-b35b-e80644e4ba7c",
   Evaluatable->True,
   CellTags->{"Proof|ID:1016692823-1", "3"},
   CellID->1014581777]},
 "Proof|ID:997515848-1"->{
  Cell[551057, 15949, 1820, 47, 29, "ProofInfo",ExpressionUUID->"5bcae50a-1782-48b2-bce6-0bda8d2274c8",
   Evaluatable->True,
   CellTags->{"Proof|ID:997515848-1", "3"},
   CellID->273440836]},
 "disjoint complement"->{
  Cell[554213, 16029, 979, 29, 47, "FormalTextInputFormula",ExpressionUUID->"14526c5d-7e1a-491c-a18d-b2adee8c4fd1",
   CellTags->"disjoint complement",
   CellID->36426955]},
 "complement disjoint"->{
  Cell[555195, 16060, 979, 29, 47, "FormalTextInputFormula",ExpressionUUID->"0f6a4806-b037-4cb5-86be-f9e02d910c80",
   CellTags->"complement disjoint",
   CellID->1941170909]},
 "partition complement"->{
  Cell[556177, 16091, 956, 28, 47, "FormalTextInputFormula",ExpressionUUID->"cba2301e-f964-4dfa-9ffc-d9fc085df72c",
   CellTags->"partition complement",
   CellID->980585996]},
 "complement partition"->{
  Cell[557136, 16121, 957, 28, 47, "FormalTextInputFormula",ExpressionUUID->"26e2dded-cd7a-432f-9c23-b65e3a2865a1",
   CellTags->"complement partition",
   CellID->1324045296]},
 "complement complement"->{
  Cell[558096, 16151, 1029, 31, 50, "FormalTextInputFormula",ExpressionUUID->"d99264a8-25ec-40e3-a87a-33a8d7399ef1",
   CellTags->"complement complement",
   CellID->1636712813]},
 "complement \[Union]"->{
  Cell[559128, 16184, 1120, 33, 50, "FormalTextInputFormula",ExpressionUUID->"b414a419-93c9-4969-bb70-aa9e72fb3466",
   CellTags->"complement \[Union]",
   CellID->1334137643]},
 "complement \[Intersection]"->{
  Cell[560251, 16219, 1133, 33, 50, "FormalTextInputFormula",ExpressionUUID->"b67ff51c-86bc-4021-948a-f367e9f183e7",
   CellTags->"complement \[Intersection]",
   CellID->1036693907]},
 "complement Insert"->{
  Cell[561387, 16254, 1097, 34, 50, "FormalTextInputFormula",ExpressionUUID->"2ca9d86e-03aa-4873-a4f5-ea04b00662b0",
   CellTags->"complement Insert",
   CellID->204929684]},
 "\[SubsetEqual]- self"->{
  Cell[562487, 16290, 1243, 40, 47, "FormalTextInputFormula",ExpressionUUID->"a2f250c1-a6f8-48b1-917d-4f4472a06ad4",
   CellTags->"\[SubsetEqual]- self",
   CellID->849790927]},
 "complement UNIV"->{
  Cell[563733, 16332, 791, 21, 38, "FormalTextInputFormula",ExpressionUUID->"521382bc-26ea-430e-92b8-33377f2e1ed9",
   CellTags->"complement UNIV",
   CellID->1609622809]},
 "complement {}"->{
  Cell[564527, 16355, 786, 21, 38, "FormalTextInputFormula",ExpressionUUID->"d6188e01-f675-4a8a-b808-d55954fe886e",
   CellTags->"complement {}",
   CellID->1579837433]},
 "complement \[SubsetEqual] complement"->{
  Cell[565316, 16378, 1304, 41, 50, "FormalTextInputFormula",ExpressionUUID->"d6e1db89-6011-4f20-b209-e333d98038c8",
   CellTags->"complement \[SubsetEqual] complement",
   CellID->2039310054]},
 "complement \[Equal] complement"->{
  Cell[566623, 16421, 1389, 44, 50, "FormalTextInputFormula",ExpressionUUID->"ba85e675-30b3-4fb0-b8c4-fd034abf0994",
   CellTags->"complement \[Equal] complement",
   CellID->1867561751]},
 "Proof|ID:36426955-1"->{
  Cell[568257, 16474, 1809, 47, 29, "ProofInfo",ExpressionUUID->"335661a6-db39-426e-a2d5-3a88bbf119ee",
   Evaluatable->True,
   CellTags->{"Proof|ID:36426955-1", "3"},
   CellID->1087560852]},
 "Proof|ID:1941170909-1"->{
  Cell[570069, 16523, 1817, 47, 29, "ProofInfo",ExpressionUUID->"790c583f-b1fd-4cae-b4c7-3b7f8318ee81",
   Evaluatable->True,
   CellTags->{"Proof|ID:1941170909-1", "3"},
   CellID->769645528]},
 "Proof|ID:980585996-1"->{
  Cell[571889, 16572, 1817, 47, 29, "ProofInfo",ExpressionUUID->"978d16d3-2a8d-48c1-8119-e556b594fdbe",
   Evaluatable->True,
   CellTags->{"Proof|ID:980585996-1", "3"},
   CellID->1868666534]},
 "Proof|ID:1324045296-1"->{
  Cell[573709, 16621, 1820, 47, 29, "ProofInfo",ExpressionUUID->"55ba8ff4-e771-4565-ad2a-a535645579da",
   Evaluatable->True,
   CellTags->{"Proof|ID:1324045296-1", "3"},
   CellID->1667315755]},
 "Proof|ID:1636712813-1"->{
  Cell[575532, 16670, 1821, 47, 29, "ProofInfo",ExpressionUUID->"4270ef55-e968-4e30-88f5-189f15367a47",
   Evaluatable->True,
   CellTags->{"Proof|ID:1636712813-1", "3"},
   CellID->193380618]},
 "Proof|ID:1334137643-1"->{
  Cell[577356, 16719, 1816, 47, 29, "ProofInfo",ExpressionUUID->"ff8108f1-b1f1-4f1c-be7e-36f13ea484f5",
   Evaluatable->True,
   CellTags->{"Proof|ID:1334137643-1", "3"},
   CellID->95672954]},
 "Proof|ID:1036693907-1"->{
  Cell[579175, 16768, 1831, 47, 29, "ProofInfo",ExpressionUUID->"6c8fd24a-5787-423d-a5a0-9ee8249e79b1",
   Evaluatable->True,
   CellTags->{"Proof|ID:1036693907-1", "3"},
   CellID->197559568]},
 "Proof|ID:204929684-1"->{
  Cell[581009, 16817, 1810, 47, 29, "ProofInfo",ExpressionUUID->"c77a6221-7b61-4579-83d1-c9ad096f840c",
   Evaluatable->True,
   CellTags->{"Proof|ID:204929684-1", "3"},
   CellID->459265611]},
 "Proof|ID:849790927-1"->{
  Cell[582822, 16866, 1817, 47, 29, "ProofInfo",ExpressionUUID->"fcbb79cd-640b-433d-aca6-6f01bc0cd323",
   Evaluatable->True,
   CellTags->{"Proof|ID:849790927-1", "3"},
   CellID->1062590287]},
 "Proof|ID:1609622809-1"->{
  Cell[584642, 16915, 1795, 46, 29, "ProofInfo",ExpressionUUID->"03e8795e-cd22-4b6b-b60f-6a5c21b1baf5",
   Evaluatable->True,
   CellTags->{"Proof|ID:1609622809-1", "3"},
   CellID->1841417850]},
 "Proof|ID:1579837433-1"->{
  Cell[586440, 16963, 1791, 46, 29, "ProofInfo",ExpressionUUID->"fc5be3c0-cfbd-484d-840c-b987b55aab11",
   Evaluatable->True,
   CellTags->{"Proof|ID:1579837433-1", "3"},
   CellID->1890841345]},
 "Proof|ID:2039310054-1"->{
  Cell[588234, 17011, 1852, 47, 29, "ProofInfo",ExpressionUUID->"070c9e7a-2ffe-4eed-a409-8fe374d0f970",
   Evaluatable->True,
   CellTags->{"Proof|ID:2039310054-1", "3"},
   CellID->1409452885]},
 "Proof|ID:1867561751-1"->{
  Cell[590089, 17060, 1840, 47, 29, "ProofInfo",ExpressionUUID->"3255cca4-1d25-482e-bef3-695eb3e5fa77",
   Evaluatable->True,
   CellTags->{"Proof|ID:1867561751-1", "3"},
   CellID->1823633609]},
 "- alt SET 2"->{
  Cell[593536, 17145, 1004, 30, 50, "FormalTextInputFormula",ExpressionUUID->"6623863d-220c-43b7-b56a-d97b10f1a117",
   CellTags->"- alt SET 2",
   CellID->889927681]},
 "-\[Equal]{} iff"->{
  Cell[594543, 17177, 1261, 41, 50, "FormalTextInputFormula",ExpressionUUID->"60d9d2ad-a5de-4e53-8980-466c84af87c2",
   CellTags->"-\[Equal]{} iff",
   CellID->1181933579]},
 "- self SET"->{
  Cell[595807, 17220, 928, 28, 47, "FormalTextInputFormula",ExpressionUUID->"6781d634-3dc0-42fb-8e42-aff33a4f6cec",
   CellTags->"- self SET",
   CellID->1313006718]},
 "- idem SET"->{
  Cell[596738, 17250, 1115, 34, 50, "FormalTextInputFormula",ExpressionUUID->"e34deac0-2ee0-4195-b922-322864967d47",
   CellTags->"- idem SET",
   CellID->405798501]},
 "- triv SET"->{
  Cell[597856, 17286, 1166, 37, 50, "FormalTextInputFormula",ExpressionUUID->"ed752dac-cb8f-4130-b775-aa2d387899a4",
   CellTags->"- triv SET",
   CellID->142114837]},
 "-{}"->{
  Cell[599025, 17325, 902, 27, 47, "FormalTextInputFormula",ExpressionUUID->"f814e4e0-d5fd-4f2d-8049-0a0cc5ce82a9",
   CellTags->"-{}",
   CellID->655509949]},
 "{}-"->{
  Cell[599930, 17354, 924, 28, 47, "FormalTextInputFormula",ExpressionUUID->"b1c87120-9de9-411e-b7fb-51ab4e109a6f",
   CellTags->"{}-",
   CellID->149156198]},
 "-UNIV"->{
  Cell[600857, 17384, 919, 28, 47, "FormalTextInputFormula",ExpressionUUID->"c3b9274b-a806-42e8-a112-b152145efc29",
   CellTags->"-UNIV",
   CellID->1551567722]},
 "Insert-"->{
  Cell[601779, 17414, 1407, 44, 57, "FormalTextInputFormula",ExpressionUUID->"42e35235-185b-4a52-99ed-8e7cf0803c40",
   CellTags->"Insert-",
   CellID->1618171348]},
 "Insert- single"->{
  Cell[603189, 17460, 1091, 33, 50, "FormalTextInputFormula",ExpressionUUID->"a4ab1dc9-a2bf-463a-8074-9e4f2ca3cbf3",
   CellTags->"Insert- single",
   CellID->263157914]},
 "- disjoint"->{
  Cell[604283, 17495, 1018, 31, 50, "FormalTextInputFormula",ExpressionUUID->"50650e03-a3f8-46d0-b2ff-9a8d0cd54ee2",
   CellTags->"- disjoint",
   CellID->586501678]},
 "\[Union]- cancel"->{
  Cell[605304, 17528, 1087, 32, 50, "FormalTextInputFormula",ExpressionUUID->"bffe28f1-05a1-4afd-a4bd-cc179788a0cf",
   CellTags->"\[Union]- cancel",
   CellID->1211775432]},
 "-\[Union] cancel"->{
  Cell[606394, 17562, 1035, 31, 50, "FormalTextInputFormula",ExpressionUUID->"4b9cee39-401e-4340-85f0-5cc10f31ca56",
   CellTags->"-\[Union] cancel",
   CellID->1437446913]},
 "-\[Union]"->{
  Cell[607432, 17595, 1143, 35, 50, "FormalTextInputFormula",ExpressionUUID->"1364d8ff-9078-426a-83d8-d37785793f2d",
   CellTags->"-\[Union]",
   CellID->1743954910]},
 "-\[Intersection]"->{
  Cell[608578, 17632, 1158, 35, 50, "FormalTextInputFormula",ExpressionUUID->"14da9918-2877-431c-9676-774a594407e6",
   CellTags->"-\[Intersection]",
   CellID->732837784]},
 "-- \[Equal] \[Intersection]"->{
  Cell[609739, 17669, 1057, 31, 50, "FormalTextInputFormula",ExpressionUUID->"3ac7b94b-778f-462d-a022-df58ba35c477",
   CellTags->"-- \[Equal] \[Intersection]",
   CellID->520649721]},
 "\[Union]-"->{
  Cell[610799, 17702, 1138, 35, 50, "FormalTextInputFormula",ExpressionUUID->"19d77f44-c8da-4204-b8b0-44e14aee704c",
   CellTags->"\[Union]-",
   CellID->1959934043]},
 "\[Intersection]-"->{
  Cell[611940, 17739, 1111, 33, 50, "FormalTextInputFormula",ExpressionUUID->"a1bb30b1-8de9-47b4-8158-a68f8ab50910",
   CellTags->"\[Intersection]-",
   CellID->1023545665]},
 "\[Intersection]- dist left"->{
  Cell[613054, 17774, 1200, 35, 50, "FormalTextInputFormula",ExpressionUUID->"83e3520e-f1ff-4eac-a5ba-79ed51ea4e65",
   CellTags->"\[Intersection]- dist left",
   CellID->1036748287]},
 "\[Intersection]- dist right"->{
  Cell[614257, 17811, 1199, 35, 50, "FormalTextInputFormula",ExpressionUUID->"9c74b7bb-8043-4f3f-adef-5cedca906c9b",
   CellTags->"\[Intersection]- dist right",
   CellID->654257124]},
 "- complement"->{
  Cell[615459, 17848, 1020, 31, 50, "FormalTextInputFormula",ExpressionUUID->"ac1ce070-9cd7-4b47-9574-487965c33172",
   CellTags->"- complement",
   CellID->788943865]},
 "complement -"->{
  Cell[616482, 17881, 1034, 32, 50, "FormalTextInputFormula",ExpressionUUID->"f380d7e6-5d16-4293-becd-33ad9bfa3087",
   CellTags->"complement -",
   CellID->811668796]},
 "Proof|ID:889927681-1"->{
  Cell[617762, 17922, 1783, 46, 29, "ProofInfo",ExpressionUUID->"41412e05-3d1b-47af-9a2c-69b7da569934",
   Evaluatable->True,
   CellTags->{"Proof|ID:889927681-1", "3"},
   CellID->989686867]},
 "Proof|ID:1181933579-1"->{
  Cell[619548, 17970, 1793, 46, 29, "ProofInfo",ExpressionUUID->"e718483a-9533-47fb-b2b3-21662a495fe5",
   Evaluatable->True,
   CellTags->{"Proof|ID:1181933579-1", "3"},
   CellID->14631120]},
 "Proof|ID:1313006718-1"->{
  Cell[621344, 18018, 1785, 46, 29, "ProofInfo",ExpressionUUID->"c083b88a-b5b4-4213-8c12-363f6380f03a",
   Evaluatable->True,
   CellTags->{"Proof|ID:1313006718-1", "3"},
   CellID->1424176457]},
 "Proof|ID:405798501-1"->{
  Cell[623132, 18066, 1782, 46, 29, "ProofInfo",ExpressionUUID->"8dfe2f53-ea61-4e3d-9764-e5a54581d749",
   Evaluatable->True,
   CellTags->{"Proof|ID:405798501-1", "3"},
   CellID->1289581872]},
 "Proof|ID:142114837-1"->{
  Cell[624917, 18114, 1782, 46, 29, "ProofInfo",ExpressionUUID->"15000b41-243b-47fb-9432-88147364a061",
   Evaluatable->True,
   CellTags->{"Proof|ID:142114837-1", "3"},
   CellID->1757475874]},
 "Proof|ID:655509949-1"->{
  Cell[626702, 18162, 1752, 45, 29, "ProofInfo",ExpressionUUID->"ba81f3be-43a1-4405-a13e-20392ccc153a",
   Evaluatable->True,
   CellTags->{"Proof|ID:655509949-1", "3"},
   CellID->939658652]},
 "Proof|ID:149156198-1"->{
  Cell[628457, 18209, 1752, 45, 29, "ProofInfo",ExpressionUUID->"a5e8bedb-4a86-49a2-916f-d64eed66bbf2",
   Evaluatable->True,
   CellTags->{"Proof|ID:149156198-1", "3"},
   CellID->125409990]},
 "Proof|ID:1551567722-1"->{
  Cell[630212, 18256, 1759, 45, 29, "ProofInfo",ExpressionUUID->"81eeefa4-874f-43cf-8a1a-bf67eab02b68",
   Evaluatable->True,
   CellTags->{"Proof|ID:1551567722-1", "3"},
   CellID->128243725]},
 "Proof|ID:1618171348-1"->{
  Cell[631974, 18303, 1764, 45, 29, "ProofInfo",ExpressionUUID->"27fb32d0-54c1-4430-9891-7032e6c9d015",
   Evaluatable->True,
   CellTags->{"Proof|ID:1618171348-1", "3"},
   CellID->1212022066]},
 "Proof|ID:263157914-1"->{
  Cell[633741, 18350, 1790, 46, 29, "ProofInfo",ExpressionUUID->"0d2dd380-ba5d-4928-b373-142c2857c2c0",
   Evaluatable->True,
   CellTags->{"Proof|ID:263157914-1", "3"},
   CellID->1307442480]},
 "Proof|ID:586501678-1"->{
  Cell[635534, 18398, 1782, 46, 29, "ProofInfo",ExpressionUUID->"1310c805-4a87-4c5e-91d3-1155f73282e3",
   Evaluatable->True,
   CellTags->{"Proof|ID:586501678-1", "3"},
   CellID->1257980868]},
 "Proof|ID:1211775432-1"->{
  Cell[637319, 18446, 1812, 47, 29, "ProofInfo",ExpressionUUID->"f7d7c577-0e4f-43f2-aab9-f884ae948b4c",
   Evaluatable->True,
   CellTags->{"Proof|ID:1211775432-1", "3"},
   CellID->1583175827]},
 "Proof|ID:1437446913-1"->{
  Cell[639134, 18495, 1812, 47, 29, "ProofInfo",ExpressionUUID->"06df1467-a1fc-4be5-b103-26e4a9577222",
   Evaluatable->True,
   CellTags->{"Proof|ID:1437446913-1", "3"},
   CellID->2050049889]},
 "Proof|ID:1743954910-1"->{
  Cell[640949, 18544, 1768, 45, 29, "ProofInfo",ExpressionUUID->"9ecc279b-a1b3-4c56-a8fa-e29147a8ea06",
   Evaluatable->True,
   CellTags->{"Proof|ID:1743954910-1", "3"},
   CellID->1362760740]},
 "Proof|ID:732837784-1"->{
  Cell[642720, 18591, 1809, 47, 29, "ProofInfo",ExpressionUUID->"45cfb94c-9228-4f72-a7c6-3adc38573e3f",
   Evaluatable->True,
   CellTags->{"Proof|ID:732837784-1", "3"},
   CellID->1924200546]},
 "Proof|ID:520649721-1"->{
  Cell[644532, 18640, 1831, 47, 29, "ProofInfo",ExpressionUUID->"d7f0bdda-8a9a-44b8-8d62-f6a17e32fa14",
   Evaluatable->True,
   CellTags->{"Proof|ID:520649721-1", "3"},
   CellID->1370228285]},
 "Proof|ID:1959934043-1"->{
  Cell[646366, 18689, 1768, 45, 29, "ProofInfo",ExpressionUUID->"dbca4d7d-fc2b-4f4e-9403-9ed1fcdfcd97",
   Evaluatable->True,
   CellTags->{"Proof|ID:1959934043-1", "3"},
   CellID->1728602517]},
 "Proof|ID:1023545665-1"->{
  Cell[648137, 18736, 1811, 47, 29, "ProofInfo",ExpressionUUID->"330a0d43-a3eb-4628-9ac8-18c5d5c04d74",
   Evaluatable->True,
   CellTags->{"Proof|ID:1023545665-1", "3"},
   CellID->124203596]},
 "Proof|ID:1036748287-1"->{
  Cell[649951, 18785, 1832, 47, 29, "ProofInfo",ExpressionUUID->"e55012c4-27e1-41ae-920d-8392fb2fc039",
   Evaluatable->True,
   CellTags->{"Proof|ID:1036748287-1", "3"},
   CellID->1046065391]},
 "Proof|ID:654257124-1"->{
  Cell[651786, 18834, 1830, 47, 29, "ProofInfo",ExpressionUUID->"75c51d93-2b93-476d-8df7-56ee4ba944f9",
   Evaluatable->True,
   CellTags->{"Proof|ID:654257124-1", "3"},
   CellID->190015761]},
 "Proof|ID:788943865-1"->{
  Cell[653619, 18883, 1786, 46, 29, "ProofInfo",ExpressionUUID->"bfb4263d-1542-4461-b96b-3cb80e17ef47",
   Evaluatable->True,
   CellTags->{"Proof|ID:788943865-1", "3"},
   CellID->1182421826]},
 "Proof|ID:811668796-1"->{
  Cell[655408, 18931, 1786, 46, 29, "ProofInfo",ExpressionUUID->"7126e94e-5680-4178-a002-f91132148a3e",
   Evaluatable->True,
   CellTags->{"Proof|ID:811668796-1", "3"},
   CellID->1010368971]},
 "Pow {}"->{
  Cell[658976, 19017, 822, 23, 38, "FormalTextInputFormula",ExpressionUUID->"5be69f99-7725-46d4-bb6b-dbc97f3b3ab6",
   CellTags->"Pow {}",
   CellID->525687623]},
 "Pow UNIV"->{
  Cell[659801, 19042, 774, 21, 38, "FormalTextInputFormula",ExpressionUUID->"94af11de-54d0-4a76-a0f9-9d7f0fa4a28f",
   CellTags->"Pow UNIV",
   CellID->537376125]},
 "\[Union]Pow \[SubsetEqual]"->{
  Cell[660578, 19065, 1106, 32, 50, "FormalTextInputFormula",ExpressionUUID->"fd9a1629-3d23-485d-a17d-020a9e92a3d4",
   CellTags->"\[Union]Pow \[SubsetEqual]",
   CellID->1322774384]},
 "Pow\[Intersection]"->{
  Cell[661687, 19099, 1097, 32, 50, "FormalTextInputFormula",ExpressionUUID->"334b0879-21a1-493b-a548-ff7fa6aad20a",
   CellTags->"Pow\[Intersection]",
   CellID->552587340]},
 "Proof|ID:525687623-1"->{
  Cell[663030, 19140, 1758, 45, 29, "ProofInfo",ExpressionUUID->"60c20774-a0c4-41e0-8583-3cd982a256e3",
   Evaluatable->True,
   CellTags->{"Proof|ID:525687623-1", "3"},
   CellID->118494995]},
 "Proof|ID:537376125-1"->{
  Cell[664791, 19187, 1762, 45, 29, "ProofInfo",ExpressionUUID->"e80da33b-c1dd-4198-bc6a-2a3ca4d75937",
   Evaluatable->True,
   CellTags->{"Proof|ID:537376125-1", "3"},
   CellID->998603711]},
 "Proof|ID:1322774384-1"->{
  Cell[666556, 19234, 1832, 47, 29, "ProofInfo",ExpressionUUID->"1cdbcc0c-c7a2-4a23-ad73-491bc0d57579",
   Evaluatable->True,
   CellTags->{"Proof|ID:1322774384-1", "3"},
   CellID->1961996116]},
 "Proof|ID:552587340-1"->{
  Cell[668391, 19283, 1813, 47, 29, "ProofInfo",ExpressionUUID->"362f630c-4228-4cb1-b302-51e8b7bdf329",
   Evaluatable->True,
   CellTags->{"Proof|ID:552587340-1", "3"},
   CellID->1186950056]},
 "@\[Union]"->{
  Cell[671352, 19359, 1081, 33, 51, "FormalTextInputFormula",ExpressionUUID->"27842332-1915-4319-aa16-d1927fa59d6a",
   CellTags->"@\[Union]",
   CellID->814265295]},
 "@ ident"->{
  Cell[672436, 19394, 1047, 32, 51, "FormalTextInputFormula",ExpressionUUID->"ebf795ec-f3cd-4609-a17d-968049102958",
   CellTags->"@ ident",
   CellID->1283984586]},
 "@{}"->{
  Cell[673486, 19428, 925, 28, 48, "FormalTextInputFormula",ExpressionUUID->"c6a43110-bc9f-4894-aa96-288c7209b020",
   CellTags->"@{}",
   CellID->1257505636]},
 "@Insert"->{
  Cell[674414, 19458, 1116, 34, 51, "FormalTextInputFormula",ExpressionUUID->"d15f846d-e95a-4275-90e4-d985073cea8e",
   CellTags->"@Insert",
   CellID->565842574]},
 "@\[Equal]{} iff"->{
  Cell[675533, 19494, 1278, 42, 51, "FormalTextInputFormula",ExpressionUUID->"4c3aa816-2f51-4a8c-8b3d-a002c2ecd252",
   CellTags->"@\[Equal]{} iff",
   CellID->685115154]},
 "\[ForAll]@"->{
  Cell[676814, 19538, 1456, 47, 58, "FormalTextInputFormula",ExpressionUUID->"66f7bb38-30cc-49de-bc0e-de89ae07f96c",
   CellTags->"\[ForAll]@",
   CellID->1570027835]},
 "\[Exists]@"->{
  Cell[678273, 19587, 1453, 47, 58, "FormalTextInputFormula",ExpressionUUID->"8b19b12b-215f-4771-b7ad-e947689e6e14",
   CellTags->"\[Exists]@",
   CellID->771344911]},
 "SetOf\[Equal]@Collect"->{
  Cell[679729, 19636, 1134, 34, 62, "FormalTextInputFormula",ExpressionUUID->"79ebb868-37c7-4860-8e0b-65f7121ee750",
   CellTags->"SetOf\[Equal]@Collect",
   CellID->1251957142]},
 "SetOf\[Equal]@"->{
  Cell[680866, 19672, 1085, 33, 62, "FormalTextInputFormula",ExpressionUUID->"04de836f-2f8e-4c32-b3ee-1f22e4061854",
   CellTags->"SetOf\[Equal]@",
   CellID->2088273463]},
 "Proof|ID:814265295-1"->{
  Cell[682196, 19714, 1765, 45, 29, "ProofInfo",ExpressionUUID->"4d04f4ae-48b6-4184-a3cd-f21b71cb61b2",
   Evaluatable->True,
   CellTags->{"Proof|ID:814265295-1", "3"},
   CellID->1805064061]},
 "Proof|ID:1283984586-1"->{
  Cell[683964, 19761, 1763, 45, 29, "ProofInfo",ExpressionUUID->"268be10f-dd81-47de-b132-ddea9a7f3a82",
   Evaluatable->True,
   CellTags->{"Proof|ID:1283984586-1", "3"},
   CellID->203354138]},
 "Proof|ID:1257505636-1"->{
  Cell[685730, 19808, 1756, 45, 29, "ProofInfo",ExpressionUUID->"c60cafa1-dc19-4490-a5d0-a50cbd0e6903",
   Evaluatable->True,
   CellTags->{"Proof|ID:1257505636-1", "3"},
   CellID->1997254040]},
 "Proof|ID:565842574-1"->{
  Cell[687489, 19855, 1760, 45, 29, "ProofInfo",ExpressionUUID->"243af554-932a-474a-a189-73c259821534",
   Evaluatable->True,
   CellTags->{"Proof|ID:565842574-1", "3"},
   CellID->477474021]},
 "Proof|ID:685115154-1"->{
  Cell[689252, 19902, 1792, 46, 29, "ProofInfo",ExpressionUUID->"231ca40f-e980-4b69-b7ca-c100eb04e88e",
   Evaluatable->True,
   CellTags->{"Proof|ID:685115154-1", "3"},
   CellID->1626158995]},
 "Proof|ID:1570027835-1"->{
  Cell[691047, 19950, 1785, 46, 29, "ProofInfo",ExpressionUUID->"e7c56a2e-bc98-4cf0-80e9-3bac79998304",
   Evaluatable->True,
   CellTags->{"Proof|ID:1570027835-1", "3"},
   CellID->1776911464]},
 "Proof|ID:771344911-1"->{
  Cell[692835, 19998, 1781, 46, 29, "ProofInfo",ExpressionUUID->"860e2886-3792-4877-8a4c-304213f7d561",
   Evaluatable->True,
   CellTags->{"Proof|ID:771344911-1", "3"},
   CellID->141629000]},
 "Proof|ID:1251957142-1"->{
  Cell[694619, 20046, 1821, 47, 29, "ProofInfo",ExpressionUUID->"cecad2e4-4de8-4466-95cb-7b7894b699c6",
   Evaluatable->True,
   CellTags->{"Proof|ID:1251957142-1", "3"},
   CellID->150377480]},
 "Proof|ID:2088273463-1"->{
  Cell[696443, 20095, 1793, 46, 29, "ProofInfo",ExpressionUUID->"b6d3e804-b5ea-4af4-92d5-df3642a09ce3",
   Evaluatable->True,
   CellTags->{"Proof|ID:2088273463-1", "3"},
   CellID->1818369679]},
 "@ mono"->{
  Cell[699407, 20175, 1145, 36, 51, "FormalTextInputFormula",ExpressionUUID->"1f0cd3b1-4c97-4af8-889c-d482a16463ad",
   CellTags->"@ mono",
   CellID->570770860]},
 "Pow mono"->{
  Cell[700555, 20213, 1154, 36, 50, "FormalTextInputFormula",ExpressionUUID->"dc492e7f-ab0b-45d7-8e5f-894206d553d3",
   CellTags->"Pow mono",
   CellID->1108679522]},
 "Insert mono"->{
  Cell[701712, 20251, 1229, 38, 50, "FormalTextInputFormula",ExpressionUUID->"94e58813-71c0-439e-b734-f656ba867cb1",
   CellTags->"Insert mono",
   CellID->788559876]},
 "complement antimono"->{
  Cell[702944, 20291, 1151, 36, 50, "FormalTextInputFormula",ExpressionUUID->"e9f9e5f9-d013-4783-8a5e-8ae725783603",
   CellTags->"complement antimono",
   CellID->604821491]},
 "\[Union] mono"->{
  Cell[704098, 20329, 1348, 42, 50, "FormalTextInputFormula",ExpressionUUID->"f8f62021-ce16-4148-97bc-a826f1bd212d",
   CellTags->"\[Union] mono",
   CellID->1716105510]},
 "\[Intersection] mono"->{
  Cell[705449, 20373, 1376, 42, 50, "FormalTextInputFormula",ExpressionUUID->"e21c8cc3-a299-4b52-a555-2f06ebaa8254",
   CellTags->"\[Intersection] mono",
   CellID->1849777423]},
 "- mono"->{
  Cell[706828, 20417, 1318, 42, 50, "FormalTextInputFormula",ExpressionUUID->"efb2b44d-7bb0-461d-a404-c6b1978ae33a",
   CellTags->"- mono",
   CellID->1569569101]},
 "SetOf mono"->{
  Cell[708149, 20461, 1862, 59, 62, "FormalTextInputFormula",ExpressionUUID->"1be2d71c-9305-4560-9d81-fafc180c5b2b",
   CellTags->"SetOf mono",
   CellID->172277084]},
 "Proof|ID:570770860-1"->{
  Cell[710255, 20529, 1758, 45, 29, "ProofInfo",ExpressionUUID->"d06f131e-6326-4e72-a0bc-30541cdf0aa3",
   Evaluatable->True,
   CellTags->{"Proof|ID:570770860-1", "3"},
   CellID->602566564]},
 "Proof|ID:1108679522-1"->{
  Cell[712016, 20576, 1766, 45, 29, "ProofInfo",ExpressionUUID->"619955c1-53c3-41d6-b1c8-daf0572378fc",
   Evaluatable->True,
   CellTags->{"Proof|ID:1108679522-1", "3"},
   CellID->1093017718]},
 "Proof|ID:788559876-1"->{
  Cell[713785, 20623, 1784, 46, 29, "ProofInfo",ExpressionUUID->"6344d294-e9d4-420c-a9cb-bcd22dfbe228",
   Evaluatable->True,
   CellTags->{"Proof|ID:788559876-1", "3"},
   CellID->1966110247]},
 "Proof|ID:604821491-1"->{
  Cell[715572, 20671, 1814, 47, 29, "ProofInfo",ExpressionUUID->"5334709a-119e-480c-b8ba-516e9e54b89b",
   Evaluatable->True,
   CellTags->{"Proof|ID:604821491-1", "3"},
   CellID->429588115]},
 "Proof|ID:1716105510-1"->{
  Cell[717389, 20720, 1791, 46, 29, "ProofInfo",ExpressionUUID->"4c1795f9-7bd4-48d4-ad29-b0688fdc3008",
   Evaluatable->True,
   CellTags->{"Proof|ID:1716105510-1", "3"},
   CellID->2002110999]},
 "Proof|ID:1849777423-1"->{
  Cell[719183, 20768, 1820, 47, 29, "ProofInfo",ExpressionUUID->"8dd5ad08-d868-4d2e-ad57-6ee353a4c58b",
   Evaluatable->True,
   CellTags->{"Proof|ID:1849777423-1", "3"},
   CellID->1702528487]},
 "Proof|ID:1569569101-1"->{
  Cell[721006, 20817, 1762, 45, 29, "ProofInfo",ExpressionUUID->"f5c12280-d2b6-42bc-8fbb-e7baab4c0376",
   Evaluatable->True,
   CellTags->{"Proof|ID:1569569101-1", "3"},
   CellID->1571614296]},
 "Proof|ID:172277084-1"->{
  Cell[722771, 20864, 1781, 46, 29, "ProofInfo",ExpressionUUID->"f177dd05-a638-4859-bf62-074cc8860f82",
   Evaluatable->True,
   CellTags->{"Proof|ID:172277084-1", "3"},
   CellID->665117787]},
 "Cantor's paradox"->{
  Cell[725369, 20935, 1443, 39, 54, "FormalTextInputFormula",ExpressionUUID->"a0cf4926-7497-4691-ac9b-596d4394ea91",
   CellTags->"Cantor's paradox",
   CellID->628580588]},
 "Cantor's paradox 2"->{
  Cell[726815, 20976, 1469, 44, 55, "FormalTextInputFormula",ExpressionUUID->"c1f91a4d-9e0e-4fc8-b926-adf6ecfaa815",
   CellTags->"Cantor's paradox 2",
   CellID->289824169]},
 "Proof|ID:628580588-1"->{
  Cell[728528, 21029, 1808, 47, 29, "ProofInfo",ExpressionUUID->"dfe4d76c-665e-46e9-a153-3fd07c91d051",
   Evaluatable->True,
   CellTags->{"Proof|ID:628580588-1", "3"},
   CellID->202346487]},
 "Proof|ID:289824169-1"->{
  Cell[730339, 21078, 1812, 47, 29, "ProofInfo",ExpressionUUID->"64679309-62a4-49fc-8eec-bdb7212bfd48",
   Evaluatable->True,
   CellTags->{"Proof|ID:289824169-1", "4"},
   CellID->383070079]},
 "4"->{
  Cell[730339, 21078, 1812, 47, 29, "ProofInfo",ExpressionUUID->"64679309-62a4-49fc-8eec-bdb7212bfd48",
   Evaluatable->True,
   CellTags->{"Proof|ID:289824169-1", "4"},
   CellID->383070079]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"\[Element] def", 775145, 22038},
 {"\[SubsetEqual] def", 775338, 22042},
 {"\[Subset] def", 775529, 22046},
 {"Pow def", 775710, 22050},
 {"\[Intersection] def", 775896, 22054},
 {"\[Union] def", 776089, 22058},
 {"Minus set def", 776276, 22062},
 {"- set def", 776460, 22066},
 {"UNIV def", 776639, 22070},
 {"EMPTY def", 776817, 22074},
 {"Insert def", 776997, 22078},
 {"@ def", 777174, 22082},
 {"SetOf def", 777350, 22086},
 {"\[Element]Collect", 777538, 22090},
 {"Collect ident", 777730, 22094},
 {"Proof|ID:270104195-1", 777925, 22098},
 {"3", 778125, 22103},
 {"Proof|ID:893774303-1", 818735, 22944},
 {"\[SubsetEqual]I", 818950, 22949},
 {"intro", 819144, 22953},
 {"\[SubsetEqual]D", 821876, 22999},
 {"trans", 822079, 23003},
 {"\[SubsetEqual]D reverse", 823430, 23025},
 {"\[SubsetEqual]E classical", 823660, 23029},
 {"elim", 823862, 23033},
 {"\[SubsetEqual] alt", 825651, 23064},
 {"contraposition \[SubsetEqual]D", 825862, 23068},
 {"\[SubsetEqual] refl", 826074, 23072},
 {"\[SubsetEqual] trans", 826275, 23076},
 {"Proof|ID:334402131-1", 826489, 23080},
 {"5", 826689, 23085},
 {"Proof|ID:591515539-1", 828249, 23118},
 {"Proof|ID:50542765-1", 828468, 23123},
 {"Proof|ID:389797744-1", 828686, 23128},
 {"Proof|ID:223583403-1", 828906, 23133},
 {"Proof|ID:324850200-1", 829126, 23138},
 {"Proof|ID:232739861-1", 829345, 23143},
 {"Proof|ID:373774609-1", 829565, 23148},
 {"\[Equal] iff SET", 829781, 23153},
 {"\[SubsetEqual] antisym", 829982, 23157},
 {"\[Equal] iff \[SubsetEqual]", 830205, 23161},
 {"\[Equal]E SET", 830409, 23165},
 {"\[Equal]E classical SET", 830608, 23169},
 {"Proof|ID:333794707-1", 830824, 23173},
 {"Proof|ID:833708774-1", 831044, 23178},
 {"Proof|ID:1768369467-1", 831263, 23183},
 {"Proof|ID:129158904-1", 831484, 23188},
 {"Proof|ID:187730850-1", 831704, 23193},
 {"\[Subset]I", 831913, 23198},
 {"\[Subset]D", 832106, 23202},
 {"\[Subset]E", 832300, 23206},
 {"\[Subset] alt", 832494, 23210},
 {"\[Subset] implies \[SubsetEqual]", 832702, 23214},
 {"\[Subset]\[SubsetEqual] trans", 832925, 23218},
 {"\[SubsetEqual]\[Subset] trans", 833157, 23222},
 {"\[Subset] trans", 833375, 23226},
 {"\[Subset] irrefl", 833580, 23230},
 {"Proof|ID:32648321-1", 833777, 23234},
 {"Proof|ID:457314425-1", 833995, 23239},
 {"Proof|ID:9845002-1", 834213, 23244},
 {"Proof|ID:688844702-1", 834430, 23249},
 {"Proof|ID:53899447-1", 834648, 23254},
 {"Proof|ID:214375017-1", 834867, 23259},
 {"Proof|ID:705380086-1", 835087, 23264},
 {"Proof|ID:122404649-1", 835307, 23269},
 {"Proof|ID:490505039-1", 835526, 23274},
 {"\[NotElement]{}", 835742, 23279},
 {"{}E", 835923, 23283},
 {"{}\[SubsetEqual]", 836115, 23287},
 {"\[Equal]{} I", 836306, 23291},
 {"\[Not]\[Subset]{}", 836499, 23295},
 {"Proof|ID:109474291-1", 836700, 23299},
 {"Proof|ID:54906548-1", 836920, 23304},
 {"Proof|ID:289979004-1", 837140, 23309},
 {"Proof|ID:828225892-1", 837361, 23314},
 {"Proof|ID:1987360397-1", 837582, 23319},
 {"UNIV I", 837790, 23324},
 {"\[SubsetEqual]UNIV", 837977, 23328},
 {"\[Equal]UNIV I", 838172, 23332},
 {"UNIV\[NotEqual]{}", 838367, 23336},
 {"{}\[NotEqual]UNIV", 838564, 23340},
 {"Proof|ID:313591036-1", 838763, 23344},
 {"Proof|ID:223522958-1", 838984, 23349},
 {"Proof|ID:256313502-1", 839205, 23354},
 {"Proof|ID:358676608-1", 839425, 23359},
 {"Proof|ID:13969540-1", 839645, 23364},
 {"\[Element]Pow", 839857, 23369},
 {"PowI", 840037, 23373},
 {"PowD", 840221, 23377},
 {"{}\[Element]Pow", 840403, 23381},
 {"self\[Element]Pow", 840598, 23385},
 {"Pow\[NotEqual]{}", 840794, 23389},
 {"Proof|ID:91383453-1", 840992, 23393},
 {"Proof|ID:1005450222-1", 841213, 23398},
 {"Proof|ID:71998340-1", 841433, 23403},
 {"Proof|ID:135248918-1", 841653, 23408},
 {"Proof|ID:365430671-1", 841873, 23413},
 {"Proof|ID:761522039-1", 842094, 23418},
 {"\[Element]complement", 842315, 23423},
 {"complement I", 842511, 23427},
 {"complement D", 842710, 23431},
 {"Proof|ID:499422188-1", 842904, 23435},
 {"Proof|ID:319529804-1", 843124, 23440},
 {"Proof|ID:7165872-1", 843342, 23445},
 {"\[Element]\[Intersection]", 843565, 23450},
 {"\[Intersection]I", 843770, 23454},
 {"\[Intersection]D1", 843978, 23458},
 {"\[Intersection]D2", 844176, 23462},
 {"\[Intersection]E", 844373, 23466},
 {"Proof|ID:299832540-1", 844582, 23470},
 {"Proof|ID:742723078-1", 844802, 23475},
 {"Proof|ID:196903271-1", 845023, 23480},
 {"Proof|ID:214165647-1", 845244, 23485},
 {"Proof|ID:31355863-1", 845464, 23490},
 {"\[Intersection] lower1", 845685, 23495},
 {"\[Intersection] lower2", 845892, 23499},
 {"\[Intersection] greatest", 846101, 23503},
 {"Proof|ID:330976770-1", 846309, 23507},
 {"Proof|ID:567877347-1", 846530, 23512},
 {"Proof|ID:370774399-1", 846751, 23517},
 {"\[Element]\[Union]", 846970, 23522},
 {"\[Union]I1", 847161, 23526},
 {"\[Union]I2", 847345, 23530},
 {"\[Union]I classical", 847538, 23534},
 {"\[Union]E", 847740, 23538},
 {"Proof|ID:24302859-1", 847942, 23542},
 {"Proof|ID:204712299-1", 848161, 23547},
 {"Proof|ID:118362674-1", 848382, 23552},
 {"Proof|ID:71686966-1", 848601, 23557},
 {"Proof|ID:734373493-1", 848820, 23562},
 {"\[Union] upper1", 849036, 23567},
 {"\[Union] upper2", 849229, 23571},
 {"\[Union] least", 849421, 23575},
 {"Proof|ID:776879610-1", 849619, 23579},
 {"Proof|ID:463603332-1", 849840, 23584},
 {"Proof|ID:672172372-1", 850060, 23589},
 {"- alt SET", 850269, 23594},
 {"\[Element]-", 850453, 23598},
 {"-I SET", 850634, 23602},
 {"-D1 SET", 850821, 23606},
 {"-D2 SET", 850999, 23610},
 {"-E SET", 851173, 23614},
 {"complement \[Equal] UNIV-", 851378, 23618},
 {"Proof|ID:540115919-1", 851586, 23622},
 {"Proof|ID:201090499-1", 851807, 23627},
 {"Proof|ID:17112112-1", 852027, 23632},
 {"Proof|ID:590579378-1", 852247, 23637},
 {"Proof|ID:984086-1", 852463, 23642},
 {"Proof|ID:128535443-1", 852680, 23647},
 {"Proof|ID:679084152-1", 852900, 23652},
 {"-\[SubsetEqual]", 853115, 23657},
 {"-\[SubsetEqual] iff", 853312, 23661},
 {"\[Subset] implies \[Exists]\[Element]", 853531, 23665},
 {"Proof|ID:917574097-1", 853752, 23669},
 {"Proof|ID:34104065-1", 853972, 23674},
 {"Proof|ID:168218769-1", 854191, 23679},
 {"\[Element]Insert", 854408, 23684},
 {"InsertI1", 854596, 23688},
 {"InsertI1'", 854777, 23692},
 {"InsertI2", 854958, 23696},
 {"InsertI classical", 855147, 23700},
 {"InsertE", 855346, 23704},
 {"Insert alt", 855537, 23708},
 {"Proof|ID:833542610-1", 855730, 23712},
 {"Proof|ID:265974781-1", 855951, 23717},
 {"Proof|ID:369554344-1", 856172, 23722},
 {"Proof|ID:276809760-1", 856392, 23727},
 {"Proof|ID:794800661-1", 856612, 23732},
 {"Proof|ID:752025544-1", 856833, 23737},
 {"Proof|ID:60042161-1", 857053, 23742},
 {"\[SubsetEqual]Insert I", 857274, 23747},
 {"\[SubsetEqual]Insert I2", 857482, 23751},
 {"\[SubsetEqual]Insert", 857688, 23755},
 {"\[SubsetEqual]Insert iff", 857895, 23759},
 {"Insert\[SubsetEqual] iff", 858108, 23763},
 {"Proof|ID:499325579-1", 858316, 23767},
 {"Proof|ID:89181006-1", 858536, 23772},
 {"Proof|ID:17294591-1", 858754, 23777},
 {"Proof|ID:1726053781-1", 858973, 23782},
 {"Proof|ID:140408834-1", 859195, 23787},
 {"singleton I", 859406, 23792},
 {"singleton D", 859603, 23796},
 {"singleton inject", 859795, 23800},
 {"Proof|ID:1706541997-1", 859997, 23804},
 {"Proof|ID:1946555487-1", 860220, 23809},
 {"Proof|ID:1642372132-1", 860441, 23814},
 {"\[SubsetEqual] singleton", 860666, 23819},
 {"singleton \[SubsetEqual]", 860878, 23823},
 {"Proof|ID:574835025-1", 861086, 23827},
 {"Proof|ID:316263559-1", 861306, 23832},
 {"@I eq", 861511, 23837},
 {"@I eq reverse", 861705, 23841},
 {"@I", 861884, 23845},
 {"@E", 862053, 23849},
 {"\[Element]@", 862240, 23853},
 {"@ cong", 862422, 23857},
 {"@ ident I", 862601, 23861},
 {"Proof|ID:1595981918-1", 862795, 23865},
 {"Proof|ID:796685651-1", 863016, 23870},
 {"Proof|ID:1022035164-1", 863238, 23875},
 {"Proof|ID:114344294-1", 863460, 23880},
 {"Proof|ID:1795699175-1", 863682, 23885},
 {"Proof|ID:724997673-1", 863904, 23890},
 {"Proof|ID:803636397-1", 864124, 23895},
 {"@\[SubsetEqual] I", 864342, 23900},
 {"\[SubsetEqual]@ E", 864541, 23904},
 {"@\[Intersection]", 864739, 23908},
 {"@-", 864922, 23912},
 {"Proof|ID:1322094901-1", 865108, 23916},
 {"Proof|ID:1691291058-1", 865330, 23921},
 {"Proof|ID:1830553618-1", 865552, 23926},
 {"Proof|ID:398205279-1", 865775, 23931},
 {"RangeI eq", 865986, 23936},
 {"RangeI", 866177, 23940},
 {"RangeE", 866354, 23944},
 {"Range singleton D", 866543, 23948},
 {"Proof|ID:686090881-1", 866746, 23952},
 {"Proof|ID:1169658618-1", 866969, 23957},
 {"Proof|ID:1129746621-1", 867191, 23962},
 {"Proof|ID:1805361717-1", 867415, 23967},
 {"\[Element]SetOf", 867632, 23972},
 {"SetOf I", 867819, 23976},
 {"SetOf E", 868009, 23980},
 {"Proof|ID:718424401-1", 868212, 23984},
 {"Proof|ID:783246906-1", 868433, 23989},
 {"Proof|ID:1984656332-1", 868656, 23994},
 {"SetOf\[Equal]Collect", 868879, 23999},
 {"SetOf const", 869075, 24003},
 {"SetOf\[Equal]{} iff", 869270, 24007},
 {"SetOf\[Not]", 869466, 24011},
 {"SetOf\[Or]", 869652, 24015},
 {"SetOf\[Implies]", 869843, 24019},
 {"Collect\[And]", 870036, 24023},
 {"SetOf mono iff", 870228, 24027},
 {"Proof|ID:314660242-1", 870427, 24031},
 {"Proof|ID:222426804-1", 870648, 24036},
 {"Proof|ID:1237361888-1", 870871, 24041},
 {"Proof|ID:165006998-1", 871093, 24046},
 {"Proof|ID:1855260115-1", 871316, 24051},
 {"Proof|ID:985026514-1", 871538, 24056},
 {"Proof|ID:636400335-1", 871760, 24061},
 {"Proof|ID:859120159-1", 871982, 24066},
 {"\[SubsetEqual]{}", 872200, 24071},
 {"\[ForAll]\[NotElement] conv", 872408, 24075},
 {"\[Exists]\[Element] conv", 872624, 24079},
 {"Proof|ID:485185733-1", 872833, 24083},
 {"Proof|ID:775376237-1", 873054, 24088},
 {"Proof|ID:978065715-1", 873275, 24093},
 {"Insert UNIV", 873487, 24098},
 {"Insert\[Equal]\[Union]", 873684, 24102},
 {"Insert\[NotEqual]{}", 873891, 24106},
 {"Insert absorb", 874089, 24110},
 {"Insert absorb2", 874282, 24114},
 {"Insert comm", 874472, 24118},
 {"Insert SetOf", 874661, 24122},
 {"Insert\[Intersection]Insert", 874865, 24126},
 {"Insert disjoint", 875073, 24130},
 {"disjoint Insert", 875268, 24134},
 {"Proof|ID:988950467-1", 875468, 24138},
 {"Proof|ID:1780572558-1", 875690, 24143},
 {"Proof|ID:1795020389-1", 875913, 24148},
 {"Proof|ID:2125656893-1", 876136, 24153},
 {"Proof|ID:742456464-1", 876359, 24158},
 {"Proof|ID:1038355606-1", 876581, 24163},
 {"Proof|ID:657964489-1", 876803, 24168},
 {"Proof|ID:1602428504-1", 877026, 24173},
 {"Proof|ID:621646007-1", 877249, 24178},
 {"Proof|ID:656967390-1", 877471, 24183},
 {"\[Intersection] absorb", 877695, 24188},
 {"\[Intersection] absorb'", 877904, 24192},
 {"\[Intersection] comm", 878112, 24196},
 {"\[Intersection] comm'", 878318, 24200},
 {"\[Intersection] assoc", 878525, 24204},
 {"\[Intersection] absorb1", 878734, 24208},
 {"\[Intersection] absorb2", 878946, 24212},
 {"\[Intersection]{}", 879152, 24216},
 {"{}\[Intersection]", 879351, 24220},
 {"disjoint iff \[SubsetEqual]-", 879561, 24224},
 {"disjoint iff \[NotEqual]", 879778, 24228},
 {"\[Intersection]UNIV", 879986, 24232},
 {"UNIV\[Intersection]", 880189, 24236},
 {"\[Intersection]\[Union] dist left", 880406, 24240},
 {"\[Intersection]\[Union] dist right", 880637, 24244},
 {"\[Intersection]\[Equal]UNIV", 880864, 24248},
 {"\[SubsetEqual]\[Intersection]", 881085, 24252},
 {"Insert\[Intersection]", 881301, 24256},
 {"\[Intersection]Insert", 881508, 24260},
 {"Proof|ID:125816224-1", 881715, 24264},
 {"Proof|ID:848809596-1", 881937, 24269},
 {"Proof|ID:147215930-1", 882159, 24274},
 {"Proof|ID:902394792-1", 882380, 24279},
 {"Proof|ID:568250332-1", 882602, 24284},
 {"Proof|ID:1826338130-1", 882824, 24289},
 {"Proof|ID:1588902597-1", 883047, 24294},
 {"Proof|ID:1253019475-1", 883271, 24299},
 {"Proof|ID:1274251332-1", 883495, 24304},
 {"Proof|ID:903302990-1", 883718, 24309},
 {"Proof|ID:920640376-1", 883938, 24314},
 {"Proof|ID:1023611974-1", 884160, 24319},
 {"Proof|ID:1195769099-1", 884384, 24324},
 {"Proof|ID:56424478-1", 884606, 24329},
 {"Proof|ID:1624124987-1", 884827, 24334},
 {"Proof|ID:158694025-1", 885050, 24339},
 {"Proof|ID:1837534884-1", 885273, 24344},
 {"Proof|ID:229432372-1", 885495, 24349},
 {"Proof|ID:2031161567-1", 885717, 24354},
 {"\[Union] absorb", 885934, 24359},
 {"\[Union] absorb'", 886129, 24363},
 {"\[Union] comm", 886323, 24367},
 {"\[Union] comm'", 886515, 24371},
 {"\[Union] assoc", 886709, 24375},
 {"\[Union] absorb1", 886904, 24379},
 {"\[Union] absorb2", 887101, 24383},
 {"\[Union]{}", 887293, 24387},
 {"{}\[Union]", 887477, 24391},
 {"\[Union]UNIV", 887664, 24395},
 {"UNIV\[Union]", 887853, 24399},
 {"Insert\[Union]", 888043, 24403},
 {"\[Union]Insert", 888236, 24407},
 {"\[Union]\[Intersection] dist left", 888449, 24411},
 {"\[Union]\[Intersection] dist right", 888682, 24415},
 {"\[Union]\[Equal]{}", 888900, 24419},
 {"\[Union]\[SubsetEqual]", 889106, 24423},
 {"Proof|ID:510851566-1", 889313, 24427},
 {"Proof|ID:961990306-1", 889534, 24432},
 {"Proof|ID:2140682664-1", 889757, 24437},
 {"Proof|ID:1655518259-1", 889980, 24442},
 {"Proof|ID:945600630-1", 890203, 24447},
 {"Proof|ID:837689454-1", 890425, 24452},
 {"Proof|ID:1727788553-1", 890647, 24457},
 {"Proof|ID:985635433-1", 890868, 24462},
 {"Proof|ID:1158821232-1", 891091, 24467},
 {"Proof|ID:1958752872-1", 891314, 24472},
 {"Proof|ID:715721472-1", 891537, 24477},
 {"Proof|ID:727949230-1", 891758, 24482},
 {"Proof|ID:1429638428-1", 891981, 24487},
 {"Proof|ID:1116053495-1", 892204, 24492},
 {"Proof|ID:1097000890-1", 892427, 24497},
 {"Proof|ID:1016692823-1", 892651, 24502},
 {"Proof|ID:997515848-1", 892874, 24507},
 {"disjoint complement", 893094, 24512},
 {"complement disjoint", 893295, 24516},
 {"partition complement", 893499, 24520},
 {"complement partition", 893703, 24524},
 {"complement complement", 893909, 24528},
 {"complement \[Union]", 894115, 24532},
 {"complement \[Intersection]", 894326, 24536},
 {"complement Insert", 894535, 24540},
 {"\[SubsetEqual]- self", 894737, 24544},
 {"complement UNIV", 894937, 24548},
 {"complement {}", 895130, 24552},
 {"complement \[SubsetEqual] complement", 895344, 24556},
 {"complement \[Equal] complement", 895576, 24560},
 {"Proof|ID:36426955-1", 895791, 24564},
 {"Proof|ID:1941170909-1", 896013, 24569},
 {"Proof|ID:980585996-1", 896235, 24574},
 {"Proof|ID:1324045296-1", 896458, 24579},
 {"Proof|ID:1636712813-1", 896682, 24584},
 {"Proof|ID:1334137643-1", 896905, 24589},
 {"Proof|ID:1036693907-1", 897127, 24594},
 {"Proof|ID:204929684-1", 897349, 24599},
 {"Proof|ID:849790927-1", 897570, 24604},
 {"Proof|ID:1609622809-1", 897793, 24609},
 {"Proof|ID:1579837433-1", 898017, 24614},
 {"Proof|ID:2039310054-1", 898241, 24619},
 {"Proof|ID:1867561751-1", 898465, 24624},
 {"- alt SET 2", 898679, 24629},
 {"-\[Equal]{} iff", 898870, 24633},
 {"- self SET", 899061, 24637},
 {"- idem SET", 899246, 24641},
 {"- triv SET", 899431, 24645},
 {"-{}", 899609, 24649},
 {"{}-", 899779, 24653},
 {"-UNIV", 899951, 24657},
 {"Insert-", 900128, 24661},
 {"Insert- single", 900315, 24665},
 {"- disjoint", 900504, 24669},
 {"\[Union]- cancel", 900695, 24673},
 {"-\[Union] cancel", 900893, 24677},
 {"-\[Union]", 901084, 24681},
 {"-\[Intersection]", 901275, 24685},
 {"-- \[Equal] \[Intersection]", 901483, 24689},
 {"\[Union]-", 901684, 24693},
 {"\[Intersection]-", 901875, 24697},
 {"\[Intersection]- dist left", 902083, 24701},
 {"\[Intersection]- dist right", 902302, 24705},
 {"- complement", 902506, 24709},
 {"complement -", 902695, 24713},
 {"Proof|ID:889927681-1", 902892, 24717},
 {"Proof|ID:1181933579-1", 903114, 24722},
 {"Proof|ID:1313006718-1", 903336, 24727},
 {"Proof|ID:405798501-1", 903559, 24732},
 {"Proof|ID:142114837-1", 903781, 24737},
 {"Proof|ID:655509949-1", 904003, 24742},
 {"Proof|ID:149156198-1", 904224, 24747},
 {"Proof|ID:1551567722-1", 904446, 24752},
 {"Proof|ID:1618171348-1", 904669, 24757},
 {"Proof|ID:263157914-1", 904892, 24762},
 {"Proof|ID:586501678-1", 905114, 24767},
 {"Proof|ID:1211775432-1", 905337, 24772},
 {"Proof|ID:1437446913-1", 905561, 24777},
 {"Proof|ID:1743954910-1", 905785, 24782},
 {"Proof|ID:732837784-1", 906008, 24787},
 {"Proof|ID:520649721-1", 906230, 24792},
 {"Proof|ID:1959934043-1", 906453, 24797},
 {"Proof|ID:1023545665-1", 906677, 24802},
 {"Proof|ID:1036748287-1", 906900, 24807},
 {"Proof|ID:654257124-1", 907123, 24812},
 {"Proof|ID:788943865-1", 907344, 24817},
 {"Proof|ID:811668796-1", 907566, 24822},
 {"Pow {}", 907774, 24827},
 {"Pow UNIV", 907952, 24831},
 {"\[Union]Pow \[SubsetEqual]", 908150, 24835},
 {"Pow\[Intersection]", 908360, 24839},
 {"Proof|ID:525687623-1", 908563, 24843},
 {"Proof|ID:537376125-1", 908784, 24848},
 {"Proof|ID:1322774384-1", 909006, 24853},
 {"Proof|ID:552587340-1", 909229, 24858},
 {"@\[Union]", 909440, 24863},
 {"@ ident", 909621, 24867},
 {"@{}", 909797, 24871},
 {"@Insert", 909972, 24875},
 {"@\[Equal]{} iff", 910159, 24879},
 {"\[ForAll]@", 910349, 24883},
 {"\[Exists]@", 910535, 24887},
 {"SetOf\[Equal]@Collect", 910731, 24891},
 {"SetOf\[Equal]@", 910932, 24895},
 {"Proof|ID:814265295-1", 911132, 24899},
 {"Proof|ID:1283984586-1", 911355, 24904},
 {"Proof|ID:1257505636-1", 911578, 24909},
 {"Proof|ID:565842574-1", 911801, 24914},
 {"Proof|ID:685115154-1", 912022, 24919},
 {"Proof|ID:1570027835-1", 912245, 24924},
 {"Proof|ID:771344911-1", 912468, 24929},
 {"Proof|ID:1251957142-1", 912690, 24934},
 {"Proof|ID:2088273463-1", 912913, 24939},
 {"@ mono", 913122, 24944},
 {"Pow mono", 913301, 24948},
 {"Insert mono", 913486, 24952},
 {"complement antimono", 913681, 24956},
 {"\[Union] mono", 913878, 24960},
 {"\[Intersection] mono", 914077, 24964},
 {"- mono", 914269, 24968},
 {"SetOf mono", 914451, 24972},
 {"Proof|ID:570770860-1", 914646, 24976},
 {"Proof|ID:1108679522-1", 914868, 24981},
 {"Proof|ID:788559876-1", 915091, 24986},
 {"Proof|ID:604821491-1", 915313, 24991},
 {"Proof|ID:1716105510-1", 915535, 24996},
 {"Proof|ID:1849777423-1", 915759, 25001},
 {"Proof|ID:1569569101-1", 915983, 25006},
 {"Proof|ID:172277084-1", 916206, 25011},
 {"Cantor's paradox", 916423, 25016},
 {"Cantor's paradox 2", 916622, 25020},
 {"Proof|ID:628580588-1", 916825, 25024},
 {"Proof|ID:289824169-1", 917046, 25029},
 {"4", 917248, 25034}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[580, 22, 170, 2, 85, "Title",ExpressionUUID->"a9b3133e-7cc9-46ff-ac44-1507e25d0b1a",
 CellID->365833305],
Cell[753, 26, 546, 10, 55, "Subsubtitle",ExpressionUUID->"0ef2b046-6bef-49bb-a223-5aad136b4a98",
 CellID->644169820],
Cell[1302, 38, 489, 10, 25, "Text",ExpressionUUID->"b5abd81b-7369-46a0-9261-4dcb0df35cf8",
 CellID->1791917602],
Cell[1794, 50, 379, 7, 55, "Input",ExpressionUUID->"4f1f2d8e-c48d-4274-9cfe-b57e907c12d8",
 CellID->265914558],
Cell[2176, 59, 237, 4, 44, "TheoremaCode",ExpressionUUID->"4127ed3e-f512-414b-ac69-880b74601b7f",
 CellID->441817464],
Cell[2416, 65, 750, 19, 44, "TheoremaCode",ExpressionUUID->"2c3959e9-507d-4a08-8e71-abcb54f1b1bb",
 CellID->38326797],
Cell[3169, 86, 286, 5, 25, "Text",ExpressionUUID->"03254130-9804-400f-8000-122d1989abfe",
 CellID->537343700],
Cell[CellGroupData[{
Cell[3480, 95, 181, 2, 57, "Section",ExpressionUUID->"a5436c35-1f68-402a-962c-c5c418d1fe70",
 CellID->224383646],
Cell[3664, 99, 836, 17, 44, "TheoremaCode",ExpressionUUID->"4611e6d7-af3c-474e-aabd-e9db75f190d5",
 CellID->1023665245],
Cell[4503, 118, 102, 1, 12, "OpenEnvironment",ExpressionUUID->"24d6f0fa-57c1-4851-ba7e-68d350fd2d4d",
 CellID->639499678],
Cell[CellGroupData[{
Cell[4630, 123, 447, 9, 38, "EnvironmentHeader",ExpressionUUID->"ecc598b7-1902-44cd-8a91-8d65c055822a",
 CellID->102108795],
Cell[5080, 134, 1138, 35, 50, "FormalTextInputFormula",ExpressionUUID->"e62578c8-e0a6-4737-b48a-89439b421a97",
 CellTags->"\[Element] def",
 CellID->464571198],
Cell[6221, 171, 1492, 49, 54, "FormalTextInputFormula",ExpressionUUID->"962bc855-b87d-497f-828c-d9a1f6dcc296",
 CellTags->"\[SubsetEqual] def",
 CellID->86239678],
Cell[7716, 222, 1451, 48, 50, "FormalTextInputFormula",ExpressionUUID->"8386411c-09c9-4758-8f4c-116b52c47684",
 CellTags->"\[Subset] def",
 CellID->169597205],
Cell[9170, 272, 1148, 36, 51, "FormalTextInputFormula",ExpressionUUID->"903ff798-45d2-49ee-9483-412b4271e487",
 CellTags->"Pow def",
 CellID->15464031],
Cell[10321, 310, 1271, 39, 54, "FormalTextInputFormula",ExpressionUUID->"bab1ffcc-3be2-40cc-8c31-516831c21a46",
 CellTags->"\[Intersection] def",
 CellID->755145795],
Cell[11595, 351, 1251, 39, 54, "FormalTextInputFormula",ExpressionUUID->"331436ef-1768-43c6-a0e6-dd4797dde600",
 CellTags->"\[Union] def",
 CellID->157538211],
Cell[12849, 392, 1173, 37, 51, "FormalTextInputFormula",ExpressionUUID->"839c14a1-7e40-49b4-866c-82bfae447357",
 CellTags->"Minus set def",
 CellID->375270853],
Cell[14025, 431, 1059, 32, 50, "FormalTextInputFormula",ExpressionUUID->"6a8c6292-cb26-4cf3-9299-3dca65b6f0ad",
 CellTags->"- set def",
 CellID->559936356],
Cell[15087, 465, 825, 23, 51, "FormalTextInputFormula",ExpressionUUID->"1bad0884-8923-430a-9e48-8ee7cdd1b0ec",
 CellTags->"UNIV def",
 CellID->426543191],
Cell[15915, 490, 831, 23, 51, "FormalTextInputFormula",ExpressionUUID->"dce4d230-1c21-4c60-8667-799c9b8fd815",
 CellTags->"EMPTY def",
 CellID->217880718],
Cell[16749, 515, 1269, 40, 54, "FormalTextInputFormula",ExpressionUUID->"ce7cf6be-50c3-49f4-8235-e3df79addb62",
 CellTags->"Insert def",
 CellID->394538268],
Cell[18021, 557, 1389, 43, 58, "FormalTextInputFormula",ExpressionUUID->"e6744ab5-9386-45ac-a300-eaa9ae40a920",
 CellTags->"@ def",
 CellID->469719290],
Cell[19413, 602, 1330, 42, 59, "FormalTextInputFormula",ExpressionUUID->"0cad2b9d-2d0a-4eb0-9090-3c81a65f8eca",
 CellTags->"SetOf def",
 CellID->312695955],
Cell[20746, 646, 120, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"7a41264a-2f82-4561-acc1-3b324a8f2168",
 CellID->556256564]
}, Open  ]],
Cell[20881, 650, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"0cf713e8-dfe9-403e-ba40-61f0e48a1abf",
 CellID->227314424],
Cell[20987, 653, 539, 12, 44, "TheoremaCode",ExpressionUUID->"1fbd38a4-1162-45e5-b5af-ec2820303f0a",
 CellID->1335265679],
Cell[21529, 667, 301, 7, 44, "TheoremaCode",ExpressionUUID->"8b63823c-e860-4837-8842-556b21924da3",
 CellID->428073695],
Cell[21833, 676, 412, 8, 25, "Text",ExpressionUUID->"422b6fbe-f7f9-48af-a839-b22e536716c8",
 CellID->1333340164],
Cell[22248, 686, 863, 15, 44, "TheoremaCode",ExpressionUUID->"98ee92b3-488a-4383-af32-2881e63b7b60",
 CellID->811751068],
Cell[23114, 703, 3547, 101, 282, "TheoremaCode",ExpressionUUID->"75b4e402-a521-479a-86f4-2c7e87e93ace",
 CellID->501147791],
Cell[CellGroupData[{
Cell[26686, 808, 184, 2, 39, "Subsection",ExpressionUUID->"675abfcf-7eb5-49a1-9cc1-153cbc76f1d7",
 CellID->1880070792],
Cell[26873, 812, 1230, 28, 406, "Code",ExpressionUUID->"8d1d744e-1bd0-41cd-b05e-282671d25f26",
 CellID->875851793]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[28152, 846, 177, 2, 43, "Section",ExpressionUUID->"3e61c363-b8c6-4f55-aaf9-b8cae1db42f3",
 CellID->72433663],
Cell[CellGroupData[{
Cell[28354, 852, 170, 2, 39, "Subsection",ExpressionUUID->"dfc54ba8-8641-4008-add0-a24265bb944d",
 CellID->1814441028],
Cell[28527, 856, 102, 1, 12, "OpenEnvironment",ExpressionUUID->"43482684-a9f1-43ad-b7fc-f9a839feb641",
 CellID->639519359],
Cell[CellGroupData[{
Cell[28654, 861, 437, 9, 38, "EnvironmentHeader",ExpressionUUID->"b375ea09-9d2e-4e37-b575-cc7fcb241585",
 CellID->828673570],
Cell[29094, 872, 1142, 35, 50, "FormalTextInputFormula",ExpressionUUID->"346c1ec5-297b-426b-a764-6d4ceb56a4e8",
 CellTags->"\[Element]Collect",
 CellID->270104195],
Cell[30239, 909, 1128, 35, 51, "FormalTextInputFormula",ExpressionUUID->"cdd465a6-9272-4aa3-980d-ab6053ec12e3",
 CellTags->"Collect ident",
 CellID->893774303],
Cell[31370, 946, 121, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"edd89845-6d3c-4630-b07f-d96f82425765",
 CellID->1190941598]
}, Open  ]],
Cell[31506, 950, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"703bd2cc-6bf2-43bd-b8fc-cee7d0754941",
 CellID->441174219],
Cell[31612, 953, 1810, 47, 29, "ProofInfo",ExpressionUUID->"5aad6e8a-a738-407b-bc9f-9ca36dc727e9",
 Evaluatable->True,
 CellTags->{"Proof|ID:270104195-1", "3"},
 CellID->1363674537],
Cell[33425, 1002, 1787, 46, 29, "ProofInfo",ExpressionUUID->"26847eea-8373-4283-8a9d-d1e9558f9cde",
 Evaluatable->True,
 CellTags->{"Proof|ID:893774303-1", "3"},
 CellID->1532689697],
Cell[35215, 1050, 367, 7, 44, "TheoremaCode",ExpressionUUID->"1a27c0b9-a50b-4961-8848-e4d073b0d107",
 CellID->864640627]
}, Closed]],
Cell[CellGroupData[{
Cell[35619, 1062, 219, 3, 31, "Subsection",ExpressionUUID->"b06a8a2d-d680-4e9e-afa8-212a5e7d816e",
 CellID->58512089],
Cell[35841, 1067, 102, 1, 12, "OpenEnvironment",ExpressionUUID->"e4fa0d71-e146-423a-8659-d1459de1cdaa",
 CellID->179550781],
Cell[CellGroupData[{
Cell[35968, 1072, 446, 9, 38, "EnvironmentHeader",ExpressionUUID->"dc7f7121-0d67-41cc-b367-efa704e052f9",
 CellID->268601953],
Cell[36417, 1083, 1475, 46, 54, "FormalTextInputFormula",ExpressionUUID->"7031efb8-33bc-4ea6-bb2f-20619acc517e",
 CellTags->{"\[SubsetEqual]I", "intro"},
 CellID->334402131],
Cell[37895, 1131, 1324, 41, 50, "FormalTextInputFormula",ExpressionUUID->"fb900f0f-ac9a-44d1-979c-f15fb5f5be35",
 CellTags->{"\[SubsetEqual]D", "trans", "intro"},
 CellID->591515539],
Cell[39222, 1174, 1344, 41, 50, "FormalTextInputFormula",ExpressionUUID->"e51dde54-c9d6-46ad-abea-f80296524d5c",
 CellTags->{"\[SubsetEqual]D reverse", "trans", "intro"},
 CellID->50542765],
Cell[40569, 1217, 1749, 54, 50, "FormalTextInputFormula",ExpressionUUID->"fb93893a-e64d-4dfe-ae4e-00ff561e6191",
 CellTags->{"\[SubsetEqual]E classical", "elim"},
 CellID->389797744],
Cell[42321, 1273, 1393, 43, 54, "FormalTextInputFormula",ExpressionUUID->"b00cb866-8d95-456e-b04f-0dc3663f9abf",
 CellTags->"\[SubsetEqual] alt",
 CellID->223583403],
Cell[43717, 1318, 1302, 40, 50, "FormalTextInputFormula",ExpressionUUID->"81ae2f05-9998-4704-978b-a1ea11dc084a",
 CellTags->"contraposition \[SubsetEqual]D",
 CellID->324850200],
Cell[45022, 1360, 811, 21, 47, "FormalTextInputFormula",ExpressionUUID->"47260c40-dd6d-4c60-8569-2246238f2cef",
 CellTags->"\[SubsetEqual] refl",
 CellID->232739861],
Cell[45836, 1383, 1331, 41, 50, "FormalTextInputFormula",ExpressionUUID->"d95e141f-73bd-461e-b3e2-1457b1c31792",
 CellTags->{"\[SubsetEqual] trans", "trans"},
 CellID->373774609],
Cell[47170, 1426, 119, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"ed87274a-1db5-4b7c-b265-0710a3a11d35",
 CellID->58488055]
}, Open  ]],
Cell[47304, 1430, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"3bf15b2d-c8b2-428f-8ece-a9d38046fe20",
 CellID->666809684],
Cell[47410, 1433, 1790, 46, 29, "ProofInfo",ExpressionUUID->"300c87b6-2c78-4afd-9a22-f6e65b245b94",
 Evaluatable->True,
 CellTags->{"Proof|ID:334402131-1", "5"},
 CellID->624024459],
Cell[49203, 1481, 1791, 46, 29, "ProofInfo",ExpressionUUID->"c1bf422f-8799-415e-8d2a-f931451d75c6",
 Evaluatable->True,
 CellTags->{"Proof|ID:591515539-1", "5"},
 CellID->1850605281],
Cell[50997, 1529, 1815, 47, 29, "ProofInfo",ExpressionUUID->"fbf77b68-1b7c-497c-86af-dfad1033b410",
 Evaluatable->True,
 CellTags->{"Proof|ID:50542765-1", "5"},
 CellID->940041033],
Cell[52815, 1578, 1826, 47, 29, "ProofInfo",ExpressionUUID->"b5a607eb-73fe-4dbb-a4e6-6a0da3ab04fd",
 Evaluatable->True,
 CellTags->{"Proof|ID:389797744-1", "5"},
 CellID->1780566989],
Cell[54644, 1627, 1812, 47, 29, "ProofInfo",ExpressionUUID->"a6e2338e-aaf5-4f15-b6ec-71a25b011cce",
 Evaluatable->True,
 CellTags->{"Proof|ID:223583403-1", "5"},
 CellID->1680503574],
Cell[56459, 1676, 1835, 47, 29, "ProofInfo",ExpressionUUID->"576a4a46-e65f-47df-8867-6961c8be3b2d",
 Evaluatable->True,
 CellTags->{"Proof|ID:324850200-1", "5"},
 CellID->623913484],
Cell[58297, 1725, 1814, 47, 29, "ProofInfo",ExpressionUUID->"5f2f6951-acfb-40e5-85f6-8302f0285cf1",
 Evaluatable->True,
 CellTags->{"Proof|ID:232739861-1", "5"},
 CellID->1074729877],
Cell[60114, 1774, 1816, 47, 29, "ProofInfo",ExpressionUUID->"d361de80-4f01-4878-9ebd-2f690ccfb113",
 Evaluatable->True,
 CellTags->{"Proof|ID:373774609-1", "5"},
 CellID->1891328660],
Cell[61933, 1823, 288, 6, 44, "TheoremaCode",ExpressionUUID->"a9b217e6-c68f-4e1b-bb54-a0e9c95b2309",
 CellID->755107797]
}, Closed]],
Cell[CellGroupData[{
Cell[62258, 1834, 171, 2, 31, "Subsection",ExpressionUUID->"843f838a-acb8-41d2-a70e-2bf4ffcec857",
 CellID->445026059],
Cell[62432, 1838, 100, 1, 12, "OpenEnvironment",ExpressionUUID->"d5b300c3-9cb8-4aae-beb0-db6795feeb60",
 CellID->9238772],
Cell[CellGroupData[{
Cell[62557, 1843, 451, 9, 38, "EnvironmentHeader",ExpressionUUID->"47f6c4c3-2db8-4595-9230-fbe83022565a",
 CellID->172029412],
Cell[63011, 1854, 1723, 58, 54, "FormalTextInputFormula",ExpressionUUID->"cc3b44c7-7696-48b2-9a11-9b0e02cbd0de",
 CellTags->"\[Equal] iff SET",
 CellID->333794707],
Cell[64737, 1914, 1315, 41, 50, "FormalTextInputFormula",ExpressionUUID->"ef16daca-313c-4d86-a78a-d1cac485fb01",
 CellTags->{"\[SubsetEqual] antisym", "intro"},
 CellID->833708774],
Cell[66055, 1957, 1539, 51, 50, "FormalTextInputFormula",ExpressionUUID->"734498b0-c5ef-4a95-a0a8-2bf22258d021",
 CellTags->"\[Equal] iff \[SubsetEqual]",
 CellID->1768369467],
Cell[67597, 2010, 1427, 44, 51, "FormalTextInputFormula",ExpressionUUID->"39d27beb-c3fa-4672-8c21-115b260ef59e",
 CellTags->"\[Equal]E SET",
 CellID->129158904],
Cell[69027, 2056, 1859, 59, 50, "FormalTextInputFormula",ExpressionUUID->"499b7ea4-c1ec-4a3c-93aa-52ab71002cec",
 CellTags->{"\[Equal]E classical SET", "elim"},
 CellID->187730850],
Cell[70889, 2117, 120, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"57e6dcfc-053d-422d-a477-1e83166d1b37",
 CellID->790548136]
}, Open  ]],
Cell[71024, 2121, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"438c5a65-ac66-4fe8-b287-df4ac79616c0",
 CellID->146581488],
Cell[71130, 2124, 1808, 47, 29, "ProofInfo",ExpressionUUID->"d29d2a48-174c-424d-be8c-88ec8b4d64fe",
 Evaluatable->True,
 CellTags->{"Proof|ID:333794707-1", "3"},
 CellID->1506794386],
Cell[72941, 2173, 1818, 47, 29, "ProofInfo",ExpressionUUID->"5027dd32-9d55-40fe-a936-08172c79cad1",
 Evaluatable->True,
 CellTags->{"Proof|ID:833708774-1", "3"},
 CellID->74925535],
Cell[74762, 2222, 1833, 47, 29, "ProofInfo",ExpressionUUID->"5f4442fc-9b36-4593-b740-1806f36d41e7",
 Evaluatable->True,
 CellTags->{"Proof|ID:1768369467-1", "3"},
 CellID->1488100858],
Cell[76598, 2271, 1787, 46, 29, "ProofInfo",ExpressionUUID->"46efcca5-5269-4858-be47-88d1f3ecc77b",
 Evaluatable->True,
 CellTags->{"Proof|ID:129158904-1", "3"},
 CellID->1301219521],
Cell[78388, 2319, 1821, 47, 29, "ProofInfo",ExpressionUUID->"e3d8ee28-9559-4c6b-b535-c0f73713378a",
 Evaluatable->True,
 CellTags->{"Proof|ID:187730850-1", "3"},
 CellID->815869071]
}, Closed]],
Cell[CellGroupData[{
Cell[80246, 2371, 175, 2, 31, "Subsection",ExpressionUUID->"5c871da2-25b4-4e48-ac4a-37581c088cfe",
 CellID->65075678],
Cell[80424, 2375, 102, 1, 12, "OpenEnvironment",ExpressionUUID->"4188a6f5-1e35-4b9c-b1fe-18e85a2cf358",
 CellID->171209090],
Cell[CellGroupData[{
Cell[80551, 2380, 441, 9, 38, "EnvironmentHeader",ExpressionUUID->"75bb5aed-dd93-47b0-ae4f-be3c6a0b5f1a",
 CellID->513227992],
Cell[80995, 2391, 1292, 41, 50, "FormalTextInputFormula",ExpressionUUID->"31eeffd9-8416-4535-8aef-65a137cf0890",
 CellTags->{"\[Subset]I", "intro"},
 CellID->32648321],
Cell[82290, 2434, 1296, 41, 50, "FormalTextInputFormula",ExpressionUUID->"b0ed0cc7-bc57-483e-a835-36a12066a7d1",
 CellTags->{"\[Subset]D", "trans"},
 CellID->457314425],
Cell[83589, 2477, 1466, 45, 51, "FormalTextInputFormula",ExpressionUUID->"c1ed4c0b-f24a-4d7a-9cb7-e956d9316e61",
 CellTags->{"\[Subset]E", "elim"},
 CellID->9845002],
Cell[85058, 2524, 1278, 41, 50, "FormalTextInputFormula",ExpressionUUID->"dbd4d924-a7ac-4077-824e-c31ad364a13b",
 CellTags->"\[Subset] alt",
 CellID->688844702],
Cell[86339, 2567, 1132, 34, 50, "FormalTextInputFormula",ExpressionUUID->"80efbdd3-5c09-4513-8814-fe0da4d53969",
 CellTags->"\[Subset] implies \[SubsetEqual]",
 CellID->53899447],
Cell[87474, 2603, 1339, 41, 50, "FormalTextInputFormula",ExpressionUUID->"9525b07f-5513-4080-98ed-049a5ac6c399",
 CellTags->{"\[Subset]\[SubsetEqual] trans", "trans"},
 CellID->214375017],
Cell[88816, 2646, 1337, 41, 50, "FormalTextInputFormula",ExpressionUUID->"c48fc25a-8ea6-4254-bb22-ab5d44b79f90",
 CellTags->{"\[SubsetEqual]\[Subset] trans", "trans"},
 CellID->705380086],
Cell[90156, 2689, 1306, 41, 50, "FormalTextInputFormula",ExpressionUUID->"ab797c4b-7e7e-4267-afe1-61013ada1bc0",
 CellTags->{"\[Subset] trans", "trans"},
 CellID->122404649],
Cell[91465, 2732, 919, 27, 47, "FormalTextInputFormula",ExpressionUUID->"8eacd303-9562-4ec0-b90b-56283033ded7",
 CellTags->"\[Subset] irrefl",
 CellID->490505039],
Cell[92387, 2761, 119, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"67c188f5-2a05-4fc3-9299-224de209fdf7",
 CellID->27085985]
}, Open  ]],
Cell[92521, 2765, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"36f8c886-d198-48a0-99e7-3309c33bff47",
 CellID->716136528],
Cell[92627, 2768, 1774, 46, 29, "ProofInfo",ExpressionUUID->"6e659942-0c6d-405c-9b68-372d2c14a0ed",
 Evaluatable->True,
 CellTags->{"Proof|ID:32648321-1", "3"},
 CellID->925901741],
Cell[94404, 2816, 1781, 46, 29, "ProofInfo",ExpressionUUID->"c887437d-b8f9-4f4a-abc2-4c844d627714",
 Evaluatable->True,
 CellTags->{"Proof|ID:457314425-1", "3"},
 CellID->1854094453],
Cell[96188, 2864, 1771, 46, 29, "ProofInfo",ExpressionUUID->"9d0f6fb7-026a-419b-8d81-21560f62f383",
 Evaluatable->True,
 CellTags->{"Proof|ID:9845002-1", "3"},
 CellID->666993796],
Cell[97962, 2912, 1786, 46, 29, "ProofInfo",ExpressionUUID->"5f1c3502-0db1-4202-8b55-474b927d5261",
 Evaluatable->True,
 CellTags->{"Proof|ID:688844702-1", "3"},
 CellID->894137823],
Cell[99751, 2960, 1834, 47, 29, "ProofInfo",ExpressionUUID->"c6989024-1dd7-4cbd-bf7d-4edaa6f98d71",
 Evaluatable->True,
 CellTags->{"Proof|ID:53899447-1", "3"},
 CellID->1854942837],
Cell[101588, 3009, 1833, 47, 29, "ProofInfo",ExpressionUUID->"441a9d35-8e30-4162-9e21-8cf3471755aa",
 Evaluatable->True,
 CellTags->{"Proof|ID:214375017-1", "3"},
 CellID->187880605],
Cell[103424, 3058, 1833, 47, 29, "ProofInfo",ExpressionUUID->"5ba5cdc6-46d4-4ad2-94e9-7d5f31bc3fe9",
 Evaluatable->True,
 CellTags->{"Proof|ID:705380086-1", "3"},
 CellID->719967367],
Cell[105260, 3107, 1789, 46, 29, "ProofInfo",ExpressionUUID->"76b67bb5-6cbd-4d7e-8eed-928b19df89d2",
 Evaluatable->True,
 CellTags->{"Proof|ID:122404649-1", "3"},
 CellID->28123779],
Cell[107052, 3155, 1808, 47, 29, "ProofInfo",ExpressionUUID->"2887fda8-1b65-4980-bb82-685424955125",
 Evaluatable->True,
 CellTags->{"Proof|ID:490505039-1", "3"},
 CellID->1482758530]
}, Closed]],
Cell[CellGroupData[{
Cell[108897, 3207, 174, 2, 31, "Subsection",ExpressionUUID->"319ae35b-f354-4f8d-894c-8478b4686747",
 CellID->398691178],
Cell[109074, 3211, 101, 1, 12, "OpenEnvironment",ExpressionUUID->"6e852d7e-d449-4a55-9c75-03282f176ba7",
 CellID->22140062],
Cell[CellGroupData[{
Cell[109200, 3216, 442, 9, 38, "EnvironmentHeader",ExpressionUUID->"0f914343-0bd7-494c-be74-1d7419de28d4",
 CellID->225411204],
Cell[109645, 3227, 820, 22, 47, "FormalTextInputFormula",ExpressionUUID->"f40e860d-7927-4260-b558-399c721a0a03",
 CellTags->"\[NotElement]{}",
 CellID->109474291],
Cell[110468, 3251, 1056, 32, 50, "FormalTextInputFormula",ExpressionUUID->"5e2a16f1-59da-4a82-8948-50266063e236",
 CellTags->{"{}E", "elim"},
 CellID->54906548],
Cell[111527, 3285, 824, 22, 47, "FormalTextInputFormula",ExpressionUUID->"c1739f2e-64f5-45b3-bfc9-a17967b39339",
 CellTags->"{}\[SubsetEqual]",
 CellID->289979004],
Cell[112354, 3309, 1227, 39, 51, "FormalTextInputFormula",ExpressionUUID->"26118ba7-c5c4-4498-bf72-00252a9af324",
 CellTags->"\[Equal]{} I",
 CellID->828225892],
Cell[113584, 3350, 941, 28, 47, "FormalTextInputFormula",ExpressionUUID->"9cd78d1f-85a0-4723-897c-277cb4c9e703",
 CellTags->"\[Not]\[Subset]{}",
 CellID->1987360397],
Cell[114528, 3380, 119, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"843f2252-85c3-42c4-9fdb-f3591cb11305",
 CellID->40835667]
}, Open  ]],
Cell[114662, 3384, 102, 1, 16, "CloseEnvironment",ExpressionUUID->"e50fc8f6-c3fe-4b13-a863-8b649fb65e90",
 CellID->27978791],
Cell[114767, 3387, 1791, 46, 29, "ProofInfo",ExpressionUUID->"ae3cd3a9-b24b-48b2-8218-78763782452d",
 Evaluatable->True,
 CellTags->{"Proof|ID:109474291-1", "3"},
 CellID->1750098064],
Cell[116561, 3435, 1746, 45, 29, "ProofInfo",ExpressionUUID->"7c42a337-6cab-46e3-bd77-c0a11197a7dc",
 Evaluatable->True,
 CellTags->{"Proof|ID:54906548-1", "3"},
 CellID->1933668764],
Cell[118310, 3482, 1808, 47, 29, "ProofInfo",ExpressionUUID->"6f47db7d-2333-492d-87ac-73a3393a968e",
 Evaluatable->True,
 CellTags->{"Proof|ID:289979004-1", "3"},
 CellID->2138873264],
Cell[120121, 3531, 1784, 46, 29, "ProofInfo",ExpressionUUID->"f789d536-f6cf-4bca-84d6-4532e0bed010",
 Evaluatable->True,
 CellTags->{"Proof|ID:828225892-1", "3"},
 CellID->157843443],
Cell[121908, 3579, 1813, 47, 29, "ProofInfo",ExpressionUUID->"3cb753de-8224-4302-aee9-4b257cf1136a",
 Evaluatable->True,
 CellTags->{"Proof|ID:1987360397-1", "3"},
 CellID->1029679455],
Cell[123724, 3628, 458, 9, 44, "TheoremaCode",ExpressionUUID->"ea74a46f-344b-4e6a-946f-391f2dba6215",
 CellID->435688659]
}, Closed]],
Cell[CellGroupData[{
Cell[124219, 3642, 178, 2, 31, "Subsection",ExpressionUUID->"c4569365-eae7-43ae-9d6b-96c63302ed0c",
 CellID->243229164],
Cell[124400, 3646, 102, 1, 12, "OpenEnvironment",ExpressionUUID->"676f0161-cbcd-44b4-ac0d-3474620a07f7",
 CellID->418476274],
Cell[CellGroupData[{
Cell[124527, 3651, 449, 9, 38, "EnvironmentHeader",ExpressionUUID->"4b3b91bf-0c36-4782-bf4b-5136344ebe69",
 CellID->102519102],
Cell[124979, 3662, 784, 21, 47, "FormalTextInputFormula",ExpressionUUID->"84774292-9b5d-4170-8a68-008bf0e627b8",
 CellTags->"UNIV I",
 CellID->313591036],
Cell[125766, 3685, 812, 21, 47, "FormalTextInputFormula",ExpressionUUID->"5a1c8aed-f913-4370-96ff-617b5a19afed",
 CellTags->"\[SubsetEqual]UNIV",
 CellID->223522958],
Cell[126581, 3708, 1048, 31, 51, "FormalTextInputFormula",ExpressionUUID->"5edf972c-0fd1-45e3-886f-b3b8b285efab",
 CellTags->"\[Equal]UNIV I",
 CellID->256313502],
Cell[127632, 3741, 780, 20, 38, "FormalTextInputFormula",ExpressionUUID->"eb06d768-e9ca-4788-ad46-4aa781f6599b",
 CellTags->"UNIV\[NotEqual]{}",
 CellID->358676608],
Cell[128415, 3763, 776, 20, 38, "FormalTextInputFormula",ExpressionUUID->"f150aa79-de04-4fd6-8a60-7995eeec44d4",
 CellTags->"{}\[NotEqual]UNIV",
 CellID->13969540],
Cell[129194, 3785, 119, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"3b694a22-d12f-435d-ac03-bbcd673b2ff2",
 CellID->62064389]
}, Open  ]],
Cell[129328, 3789, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"a462eee5-f1a6-4424-9fc6-527262d5e68c",
 CellID->208900752],
Cell[129434, 3792, 1758, 45, 29, "ProofInfo",ExpressionUUID->"f472fc82-802b-410b-868c-d07e017b0268",
 Evaluatable->True,
 CellTags->{"Proof|ID:313591036-1", "3"},
 CellID->1173747547],
Cell[131195, 3839, 1812, 47, 29, "ProofInfo",ExpressionUUID->"184057ce-28ca-4f86-bd4e-79e86a17715d",
 Evaluatable->True,
 CellTags->{"Proof|ID:223522958-1", "3"},
 CellID->2010038796],
Cell[133010, 3888, 1788, 46, 29, "ProofInfo",ExpressionUUID->"de3eb27b-5b7d-4ae2-988d-b2653d4d8d31",
 Evaluatable->True,
 CellTags->{"Proof|ID:256313502-1", "3"},
 CellID->741200714],
Cell[134801, 3936, 1810, 47, 29, "ProofInfo",ExpressionUUID->"824b928c-cb7f-4776-8b3e-3335d33de91a",
 Evaluatable->True,
 CellTags->{"Proof|ID:358676608-1", "3"},
 CellID->1104785395],
Cell[136614, 3985, 1803, 47, 29, "ProofInfo",ExpressionUUID->"49775606-3bad-4639-ad88-ce4f5704386f",
 Evaluatable->True,
 CellTags->{"Proof|ID:13969540-1", "3"},
 CellID->400553415],
Cell[138420, 4034, 592, 10, 44, "TheoremaCode",ExpressionUUID->"e1462291-f7cd-49ae-ac8d-6c511fa5abc1",
 CellID->904482657]
}, Closed]],
Cell[CellGroupData[{
Cell[139049, 4049, 169, 2, 31, "Subsection",ExpressionUUID->"0f6ebc42-5c52-420e-a539-869385485218",
 CellID->683848815],
Cell[139221, 4053, 102, 1, 12, "OpenEnvironment",ExpressionUUID->"14fa329f-4d46-4367-a4e8-abe24002157e",
 CellID->578346443],
Cell[CellGroupData[{
Cell[139348, 4058, 435, 9, 38, "EnvironmentHeader",ExpressionUUID->"3e1294c2-57fa-432b-8aca-09f3e21dba79",
 CellID->167038458],
Cell[139786, 4069, 1244, 40, 50, "FormalTextInputFormula",ExpressionUUID->"2d5efc8e-b5dc-4fd3-9859-b98a6c77774b",
 CellTags->"\[Element]Pow",
 CellID->91383453],
Cell[141033, 4111, 1156, 36, 50, "FormalTextInputFormula",ExpressionUUID->"d3c585f0-0af8-4b77-b8f2-4a2e80d5ec5a",
 CellTags->{"PowI", "intro"},
 CellID->1005450222],
Cell[142192, 4149, 1096, 34, 50, "FormalTextInputFormula",ExpressionUUID->"134a7df0-6a03-483f-8b3c-9ed6ba820c3f",
 CellTags->"PowD",
 CellID->71998340],
Cell[143291, 4185, 850, 23, 47, "FormalTextInputFormula",ExpressionUUID->"0039201d-a8f0-4273-ac3a-5b47720c154c",
 CellTags->"{}\[Element]Pow",
 CellID->135248918],
Cell[144144, 4210, 835, 22, 47, "FormalTextInputFormula",ExpressionUUID->"3975308c-87da-4dbf-bd5e-7f174b4273f7",
 CellTags->"self\[Element]Pow",
 CellID->365430671],
Cell[144982, 4234, 855, 23, 47, "FormalTextInputFormula",ExpressionUUID->"725fd266-ab88-47ba-adf8-00acdeb98c1b",
 CellTags->"Pow\[NotEqual]{}",
 CellID->761522039],
Cell[145840, 4259, 120, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"8066a26b-19d3-45ab-959d-de9d5da23820",
 CellID->175636950]
}, Open  ]],
Cell[145975, 4263, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"8946ff36-3df7-4ac4-ad26-ce1ef17ad359",
 CellID->316659783],
Cell[146081, 4266, 1782, 46, 29, "ProofInfo",ExpressionUUID->"135e6c9f-6b5f-45fe-bca0-d1d8524dfc01",
 Evaluatable->True,
 CellTags->{"Proof|ID:91383453-1", "3"},
 CellID->1669010862],
Cell[147866, 4314, 1757, 45, 29, "ProofInfo",ExpressionUUID->"2eff29d4-750b-432e-b353-66e38bf0fded",
 Evaluatable->True,
 CellTags->{"Proof|ID:1005450222-1", "3"},
 CellID->969791405],
Cell[149626, 4361, 1749, 45, 29, "ProofInfo",ExpressionUUID->"e66761b0-9d29-4441-b48e-3fce10402d48",
 Evaluatable->True,
 CellTags->{"Proof|ID:71998340-1", "3"},
 CellID->1724899203],
Cell[151378, 4408, 1791, 46, 29, "ProofInfo",ExpressionUUID->"e9f8d063-f619-4ed9-811c-81d57ff90446",
 Evaluatable->True,
 CellTags->{"Proof|ID:135248918-1", "3"},
 CellID->417457554],
Cell[153172, 4456, 1811, 47, 29, "ProofInfo",ExpressionUUID->"88881ba8-d043-4491-afa7-e2b45acd577c",
 Evaluatable->True,
 CellTags->{"Proof|ID:365430671-1", "3"},
 CellID->1046907733],
Cell[154986, 4505, 1809, 47, 29, "ProofInfo",ExpressionUUID->"7af83b7e-0a64-42d4-9cfa-8fb1f1515d6b",
 Evaluatable->True,
 CellTags->{"Proof|ID:761522039-1", "3"},
 CellID->1384782616],
Cell[156798, 4554, 695, 11, 44, "TheoremaCode",ExpressionUUID->"1859160d-8dbb-44dd-b244-cb10bb4d705c",
 CellID->404830536]
}, Closed]],
Cell[CellGroupData[{
Cell[157530, 4570, 174, 2, 31, "Subsection",ExpressionUUID->"4adeb3cb-5bc8-4193-8ec2-a95f87a70931",
 CellID->84239928],
Cell[157707, 4574, 102, 1, 12, "OpenEnvironment",ExpressionUUID->"9364deea-7691-419d-b60e-955bdaa9477c",
 CellID->564974621],
Cell[CellGroupData[{
Cell[157834, 4579, 442, 9, 38, "EnvironmentHeader",ExpressionUUID->"564c50c6-2fb4-4244-8eb4-ce407adb7a92",
 CellID->203867575],
Cell[158279, 4590, 1248, 40, 50, "FormalTextInputFormula",ExpressionUUID->"7d81673e-75fc-41aa-b4b7-569230380565",
 CellTags->"\[Element]complement",
 CellID->499422188],
Cell[159530, 4632, 1242, 39, 50, "FormalTextInputFormula",ExpressionUUID->"d27356cd-910b-4e27-a702-8f3e1f38e9d3",
 CellTags->{"complement I", "intro"},
 CellID->319529804],
Cell[160775, 4673, 1109, 35, 50, "FormalTextInputFormula",ExpressionUUID->"0bda5580-c651-4792-9b44-3829608ed193",
 CellTags->"complement D",
 CellID->7165872],
Cell[161887, 4710, 119, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"9d9b35e8-0ba4-4422-975e-8ed6dca8e0af",
 CellID->18980456]
}, Open  ]],
Cell[162021, 4714, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"b3edc74f-edee-44e9-9fde-57bec6435130",
 CellID->203717016],
Cell[162127, 4717, 1816, 47, 29, "ProofInfo",ExpressionUUID->"4bb986a8-28da-4979-8914-a7bf05134176",
 Evaluatable->True,
 CellTags->{"Proof|ID:499422188-1", "3"},
 CellID->742874104],
Cell[163946, 4766, 1785, 46, 29, "ProofInfo",ExpressionUUID->"6b7cdde2-f7e1-4895-96bd-f5dba2fa343f",
 Evaluatable->True,
 CellTags->{"Proof|ID:319529804-1", "3"},
 CellID->692678955],
Cell[165734, 4814, 1776, 46, 29, "ProofInfo",ExpressionUUID->"1b5041f8-10cf-4c78-8caf-cee02595fafa",
 Evaluatable->True,
 CellTags->{"Proof|ID:7165872-1", "3"},
 CellID->214206169],
Cell[167513, 4862, 329, 6, 44, "TheoremaCode",ExpressionUUID->"30ba7cb1-88f9-4fd1-afb5-8439bab1d1f0",
 CellID->19405246],
Cell[167845, 4870, 600, 10, 44, "TheoremaCode",ExpressionUUID->"22060e06-cda4-491b-b33a-1575f31a4956",
 CellID->131828019]
}, Closed]],
Cell[CellGroupData[{
Cell[168482, 4885, 177, 2, 31, "Subsection",ExpressionUUID->"d344c0ef-7e76-40e8-b6fb-0ed849f63b7a",
 CellID->466527033],
Cell[168662, 4889, 102, 1, 12, "OpenEnvironment",ExpressionUUID->"76ce294e-31e6-41fa-94dc-d1b5f8ac0836",
 CellID->465207003],
Cell[CellGroupData[{
Cell[168789, 4894, 445, 9, 38, "EnvironmentHeader",ExpressionUUID->"87893071-e583-408a-8705-aec67b67f9af",
 CellID->373067187],
Cell[169237, 4905, 1352, 42, 50, "FormalTextInputFormula",ExpressionUUID->"faf9caf8-e1c5-4ad6-a348-5e25e999ed9e",
 CellTags->"\[Element]\[Intersection]",
 CellID->299832540],
Cell[170592, 4949, 1352, 42, 50, "FormalTextInputFormula",ExpressionUUID->"f282f462-c6f2-4c42-93f9-8add211dc534",
 CellTags->{"\[Intersection]I", "intro"},
 CellID->742723078],
Cell[171947, 4993, 1149, 35, 50, "FormalTextInputFormula",ExpressionUUID->"728e5702-a968-4311-b864-1dd1f476124d",
 CellTags->"\[Intersection]D1",
 CellID->196903271],
Cell[173099, 5030, 1149, 35, 50, "FormalTextInputFormula",ExpressionUUID->"2c742309-e67a-46f8-923c-ec686d1342d2",
 CellTags->"\[Intersection]D2",
 CellID->214165647],
Cell[174251, 5067, 1520, 46, 51, "FormalTextInputFormula",ExpressionUUID->"d24a4225-19d0-415c-b284-7fe9384810ea",
 CellTags->{"\[Intersection]E", "elim"},
 CellID->31355863],
Cell[175774, 5115, 120, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"52002887-2767-48d7-8efc-8b61dc177c28",
 CellID->940647718]
}, Open  ]],
Cell[175909, 5119, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"692f6ffe-50f7-4dac-8332-1ecce010a4bc",
 CellID->323591903],
Cell[176015, 5122, 1826, 47, 29, "ProofInfo",ExpressionUUID->"06ccfced-9f01-4ea1-b866-1dc0ce6111d4",
 Evaluatable->True,
 CellTags->{"Proof|ID:299832540-1", "3"},
 CellID->520490995],
Cell[177844, 5171, 1809, 47, 29, "ProofInfo",ExpressionUUID->"b655e7ec-ba70-4141-b8ce-f4c7c9313eb5",
 Evaluatable->True,
 CellTags->{"Proof|ID:742723078-1", "3"},
 CellID->1206322670],
Cell[179656, 5220, 1811, 47, 29, "ProofInfo",ExpressionUUID->"999e3753-e878-46ae-816d-c1c14791130d",
 Evaluatable->True,
 CellTags->{"Proof|ID:196903271-1", "3"},
 CellID->1182674228],
Cell[181470, 5269, 1811, 47, 29, "ProofInfo",ExpressionUUID->"263b3d20-5c8b-4d42-8cd7-bb5ecf6bba6c",
 Evaluatable->True,
 CellTags->{"Proof|ID:214165647-1", "3"},
 CellID->1685360901],
Cell[183284, 5318, 1802, 47, 29, "ProofInfo",ExpressionUUID->"e8ca3e41-bda4-4b23-8e4b-61f126bed335",
 Evaluatable->True,
 CellTags->{"Proof|ID:31355863-1", "3"},
 CellID->634954845],
Cell[185089, 5367, 655, 11, 44, "TheoremaCode",ExpressionUUID->"5a8634c2-f223-4970-b9d3-0023b971effd",
 CellID->558707431],
Cell[185747, 5380, 102, 1, 12, "OpenEnvironment",ExpressionUUID->"b8b32025-d4d0-4408-9a9b-adf9aee41347",
 CellID->154478456],
Cell[CellGroupData[{
Cell[185874, 5385, 454, 9, 38, "EnvironmentHeader",ExpressionUUID->"5994a4e0-47c8-4b41-8340-de158fdf537a",
 CellID->121705945],
Cell[186331, 5396, 884, 24, 50, "FormalTextInputFormula",ExpressionUUID->"5a07b132-a168-47df-8a0a-d4cbc2bdf2be",
 CellTags->"\[Intersection] lower1",
 CellID->330976770],
Cell[187218, 5422, 886, 24, 50, "FormalTextInputFormula",ExpressionUUID->"69552c22-c7e7-4e84-ae6d-0c027f7b7ca1",
 CellTags->"\[Intersection] lower2",
 CellID->567877347],
Cell[188107, 5448, 1333, 41, 50, "FormalTextInputFormula",ExpressionUUID->"da054ea7-7898-424b-8f20-ba26dcda115c",
 CellTags->"\[Intersection] greatest",
 CellID->370774399],
Cell[189443, 5491, 120, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"8ad37a03-2479-40ce-8bb8-0b77f545807f",
 CellID->181179688]
}, Open  ]],
Cell[189578, 5495, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"a1a3e511-7907-45f5-a063-0dffd7b3714c",
 CellID->916921185],
Cell[189684, 5498, 1821, 47, 29, "ProofInfo",ExpressionUUID->"c9408df1-52d0-43ba-9c83-f86df76887a4",
 Evaluatable->True,
 CellTags->{"Proof|ID:330976770-1", "3"},
 CellID->1901417663],
Cell[191508, 5547, 1821, 47, 29, "ProofInfo",ExpressionUUID->"cd544126-02a6-4a0d-b5ab-040100796126",
 Evaluatable->True,
 CellTags->{"Proof|ID:567877347-1", "3"},
 CellID->1604414518],
Cell[193332, 5596, 1825, 47, 29, "ProofInfo",ExpressionUUID->"067b7963-94d6-4a20-aae6-bc77ea005db6",
 Evaluatable->True,
 CellTags->{"Proof|ID:370774399-1", "3"},
 CellID->1352341786],
Cell[195160, 5645, 337, 6, 44, "TheoremaCode",ExpressionUUID->"ee7d02d6-2f91-43c3-b10a-841d0491cb48",
 CellID->563438415]
}, Closed]],
Cell[CellGroupData[{
Cell[195534, 5656, 172, 2, 31, "Subsection",ExpressionUUID->"eec10edc-69fe-43f8-a984-5fd292e58f17",
 CellID->392728112],
Cell[195709, 5660, 101, 1, 12, "OpenEnvironment",ExpressionUUID->"3b04699b-4a98-40f5-8d1d-c23bc26329a4",
 CellID->32649250],
Cell[CellGroupData[{
Cell[195835, 5665, 442, 9, 38, "EnvironmentHeader",ExpressionUUID->"641b9c20-4123-4a68-818a-9a5a4b3ee1b6",
 CellID->473756208],
Cell[196280, 5676, 1325, 42, 50, "FormalTextInputFormula",ExpressionUUID->"f52d2f46-1098-465b-90bd-12324ecbac1f",
 CellTags->"\[Element]\[Union]",
 CellID->24302859],
Cell[197608, 5720, 1126, 35, 50, "FormalTextInputFormula",ExpressionUUID->"5ea72cf8-f892-430b-956b-10c727c945bf",
 CellTags->"\[Union]I1",
 CellID->204712299],
Cell[198737, 5757, 1130, 35, 50, "FormalTextInputFormula",ExpressionUUID->"55fd758c-2ba1-4a9c-a211-2ba0089a2f44",
 CellTags->"\[Union]I2",
 CellID->118362674],
Cell[199870, 5794, 1416, 44, 50, "FormalTextInputFormula",ExpressionUUID->"3dacd0ce-b618-4ccf-8cc9-6fc2a7468b22",
 CellTags->{"\[Union]I classical", "intro"},
 CellID->71686966],
Cell[201289, 5840, 1574, 49, 50, "FormalTextInputFormula",ExpressionUUID->"cc6aa25c-83a1-4025-8cd9-4fe49291cfe4",
 CellTags->{"\[Union]E", "elim"},
 CellID->734373493],
Cell[202866, 5891, 120, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"a4411408-efb7-4cf4-83a2-657117340369",
 CellID->364554349]
}, Open  ]],
Cell[203001, 5895, 102, 1, 16, "CloseEnvironment",ExpressionUUID->"365c834d-b98f-42f8-8f01-89f88fb7b941",
 CellID->91957176],
Cell[203106, 5898, 1806, 47, 29, "ProofInfo",ExpressionUUID->"5d01d301-0853-41ec-8b5b-f654db10d92f",
 Evaluatable->True,
 CellTags->{"Proof|ID:24302859-1", "3"},
 CellID->910429301],
Cell[204915, 5947, 1782, 46, 29, "ProofInfo",ExpressionUUID->"58cda68d-cdf0-4074-9690-d94a7ad8e87b",
 Evaluatable->True,
 CellTags->{"Proof|ID:204712299-1", "3"},
 CellID->2106560278],
Cell[206700, 5995, 1781, 46, 29, "ProofInfo",ExpressionUUID->"5ab74a01-dcb4-4410-adbb-2e6f4bc3d887",
 Evaluatable->True,
 CellTags->{"Proof|ID:118362674-1", "3"},
 CellID->430856106],
Cell[208484, 6043, 1808, 47, 29, "ProofInfo",ExpressionUUID->"c438dae3-6253-4324-8a41-6d2ae06ea6ec",
 Evaluatable->True,
 CellTags->{"Proof|ID:71686966-1", "3"},
 CellID->634618502],
Cell[210295, 6092, 1765, 45, 29, "ProofInfo",ExpressionUUID->"7192f13d-fafc-45a0-9eba-e5628635aa2e",
 Evaluatable->True,
 CellTags->{"Proof|ID:734373493-1", "3"},
 CellID->1988227689],
Cell[212063, 6139, 698, 12, 44, "TheoremaCode",ExpressionUUID->"b498f909-0860-4ebc-9ce5-d68efdf15213",
 CellID->575866964],
Cell[212764, 6153, 102, 1, 12, "OpenEnvironment",ExpressionUUID->"ed69799d-0791-49c4-ae25-a6159e3bbe56",
 CellID->913765167],
Cell[CellGroupData[{
Cell[212891, 6158, 450, 9, 38, "EnvironmentHeader",ExpressionUUID->"0851ea46-50e5-45a0-a430-e7263c195598",
 CellID->861897061],
Cell[213344, 6169, 859, 23, 50, "FormalTextInputFormula",ExpressionUUID->"23354d54-d353-451c-96d2-7d23f93c1ab9",
 CellTags->"\[Union] upper1",
 CellID->776879610],
Cell[214206, 6194, 857, 23, 50, "FormalTextInputFormula",ExpressionUUID->"fd46a9f7-6c14-40c4-a4e9-9c816ccbf6cd",
 CellTags->"\[Union] upper2",
 CellID->463603332],
Cell[215066, 6219, 1308, 41, 50, "FormalTextInputFormula",ExpressionUUID->"2c6fd105-c6b3-4c91-930d-cdf36722c20d",
 CellTags->"\[Union] least",
 CellID->672172372],
Cell[216377, 6262, 120, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"8a5c25f6-6ffc-44e4-8865-e8c47b4e9639",
 CellID->384387934]
}, Open  ]],
Cell[216512, 6266, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"429a5b27-6a34-4289-a130-7628cd5d781c",
 CellID->137981057],
Cell[216618, 6269, 1792, 46, 29, "ProofInfo",ExpressionUUID->"4273e686-36d1-479a-869b-b04747afdf48",
 Evaluatable->True,
 CellTags->{"Proof|ID:776879610-1", "3"},
 CellID->1308387989],
Cell[218413, 6317, 1791, 46, 29, "ProofInfo",ExpressionUUID->"a0289919-612a-43d3-83be-5e58d75618f3",
 Evaluatable->True,
 CellTags->{"Proof|ID:463603332-1", "3"},
 CellID->283700494],
Cell[220207, 6365, 1789, 46, 29, "ProofInfo",ExpressionUUID->"bacbec27-f6b0-4574-94e4-76900e1b7a43",
 Evaluatable->True,
 CellTags->{"Proof|ID:672172372-1", "3"},
 CellID->563063133],
Cell[221999, 6413, 317, 6, 44, "TheoremaCode",ExpressionUUID->"8f1d7af7-f61b-48b9-b3bb-f3ee60948a33",
 CellID->853725528]
}, Closed]],
Cell[CellGroupData[{
Cell[222353, 6424, 174, 2, 31, "Subsection",ExpressionUUID->"4ca4939a-241f-4eb0-8fd4-8da4587bd5c2",
 CellID->614708801],
Cell[222530, 6428, 102, 1, 12, "OpenEnvironment",ExpressionUUID->"66a30973-35c5-465b-808e-ef3b010f0f47",
 CellID->850854527],
Cell[CellGroupData[{
Cell[222657, 6433, 444, 9, 38, "EnvironmentHeader",ExpressionUUID->"f05615d2-d878-40ce-a534-a569ea814ae7",
 CellID->180283027],
Cell[223104, 6444, 1247, 39, 54, "FormalTextInputFormula",ExpressionUUID->"e2f0a575-b7f7-478c-9cc1-7d579996f383",
 CellTags->"- alt SET",
 CellID->540115919],
Cell[224354, 6485, 1311, 42, 50, "FormalTextInputFormula",ExpressionUUID->"dbbc2d0c-ae52-42a9-a1a3-6ac77bbec4ef",
 CellTags->"\[Element]-",
 CellID->201090499],
Cell[225668, 6529, 1318, 42, 50, "FormalTextInputFormula",ExpressionUUID->"f5cee395-7010-4ff0-ba10-f12db0c70eb9",
 CellTags->{"-I SET", "intro"},
 CellID->17112112],
Cell[226989, 6573, 1114, 35, 50, "FormalTextInputFormula",ExpressionUUID->"62a66027-3b18-408d-9e80-b3a57cfe1bcf",
 CellTags->"-D1 SET",
 CellID->590579378],
Cell[228106, 6610, 1248, 38, 50, "FormalTextInputFormula",ExpressionUUID->"b24c8d74-3d7a-402c-958b-25766a575785",
 CellTags->"-D2 SET",
 CellID->984086],
Cell[229357, 6650, 1545, 47, 50, "FormalTextInputFormula",ExpressionUUID->"a15f38a2-1024-4357-b3ae-3d4405f34edf",
 CellTags->{"-E SET", "elim"},
 CellID->128535443],
Cell[230905, 6699, 960, 28, 47, "FormalTextInputFormula",ExpressionUUID->"25eb9645-ea9d-42ae-8802-e919a21ef9e4",
 CellTags->"complement \[Equal] UNIV-",
 CellID->679084152],
Cell[231868, 6729, 120, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"dee1df3a-a638-4d09-87d2-d08a6bcbe94c",
 CellID->353082906]
}, Open  ]],
Cell[232003, 6733, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"88ba20c7-9519-47e9-ab87-0b43883e097d",
 CellID->760389402],
Cell[232109, 6736, 1765, 45, 29, "ProofInfo",ExpressionUUID->"413f783b-5cb5-476c-8d17-7d845255998a",
 Evaluatable->True,
 CellTags->{"Proof|ID:540115919-1", "3"},
 CellID->2034290086],
Cell[233877, 6783, 1784, 46, 29, "ProofInfo",ExpressionUUID->"0e3f881b-3ec2-4218-8f49-710e98423e7e",
 Evaluatable->True,
 CellTags->{"Proof|ID:201090499-1", "3"},
 CellID->1196091729],
Cell[235664, 6831, 1753, 45, 29, "ProofInfo",ExpressionUUID->"fae8ecb6-0379-4df0-8561-e4c239de4b21",
 Evaluatable->True,
 CellTags->{"Proof|ID:17112112-1", "3"},
 CellID->1855793195],
Cell[237420, 6878, 1759, 45, 29, "ProofInfo",ExpressionUUID->"ff98ca43-6f95-4017-83fe-5e58556b956d",
 Evaluatable->True,
 CellTags->{"Proof|ID:590579378-1", "3"},
 CellID->87366359],
Cell[239182, 6925, 1748, 45, 29, "ProofInfo",ExpressionUUID->"497b9d54-4fde-4ed0-af44-f8d1fb40d21a",
 Evaluatable->True,
 CellTags->{"Proof|ID:984086-1", "3"},
 CellID->738044893],
Cell[240933, 6972, 1758, 45, 29, "ProofInfo",ExpressionUUID->"9ab0c91e-bd51-4993-aa37-27e47b2db8cf",
 Evaluatable->True,
 CellTags->{"Proof|ID:128535443-1", "3"},
 CellID->187234111],
Cell[242694, 7019, 1826, 47, 29, "ProofInfo",ExpressionUUID->"526b809c-7826-439b-ac09-9eaa35adbac7",
 Evaluatable->True,
 CellTags->{"Proof|ID:679084152-1", "3"},
 CellID->887782668],
Cell[244523, 7068, 735, 11, 44, "TheoremaCode",ExpressionUUID->"2c17d1a6-6bc5-408a-a897-98c78635ecd3",
 CellID->876677958],
Cell[245261, 7081, 102, 1, 12, "OpenEnvironment",ExpressionUUID->"cb496f36-e22c-4a14-8eb8-d92f7bfa2169",
 CellID->147803878],
Cell[CellGroupData[{
Cell[245388, 7086, 438, 9, 38, "EnvironmentHeader",ExpressionUUID->"3a73a610-399d-4dcc-a7b1-e3e7b30b481b",
 CellID->685751960],
Cell[245829, 7097, 856, 24, 50, "FormalTextInputFormula",ExpressionUUID->"d07f39c7-4b49-4a87-bb3c-ecff0503e8a1",
 CellTags->"-\[SubsetEqual]",
 CellID->917574097],
Cell[246688, 7123, 1297, 41, 50, "FormalTextInputFormula",ExpressionUUID->"138bd892-f1bc-49e8-a2ee-0d8190289c18",
 CellTags->"-\[SubsetEqual] iff",
 CellID->34104065],
Cell[247988, 7166, 1221, 37, 54, "FormalTextInputFormula",ExpressionUUID->"5f3dad52-4b63-4783-80a8-4673a24fc263",
 CellTags->"\[Subset] implies \[Exists]\[Element]",
 CellID->168218769],
Cell[249212, 7205, 120, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"97e9c61e-e45f-487a-9870-0ac5d2ba6006",
 CellID->114851357]
}, Open  ]],
Cell[249347, 7209, 101, 1, 16, "CloseEnvironment",ExpressionUUID->"f85aad04-0ff9-48d5-bea1-8e828e74391f",
 CellID->1543820],
Cell[249451, 7212, 1792, 46, 29, "ProofInfo",ExpressionUUID->"94ee77de-dbad-4d4e-97d6-1755cd58f119",
 Evaluatable->True,
 CellTags->{"Proof|ID:917574097-1", "3"},
 CellID->1547795340],
Cell[251246, 7260, 1808, 47, 29, "ProofInfo",ExpressionUUID->"19482a26-cb3e-4e1f-af6d-49608b2ed75c",
 Evaluatable->True,
 CellTags->{"Proof|ID:34104065-1", "3"},
 CellID->838116264],
Cell[253057, 7309, 1851, 47, 29, "ProofInfo",ExpressionUUID->"9127f290-b8cf-4129-98e6-27563e725e5a",
 Evaluatable->True,
 CellTags->{"Proof|ID:168218769-1", "3"},
 CellID->1909735428],
Cell[254911, 7358, 339, 7, 44, "TheoremaCode",ExpressionUUID->"43ad83ae-f229-40f6-816b-725713df2e53",
 CellID->1973892905]
}, Closed]],
Cell[CellGroupData[{
Cell[255287, 7370, 273, 4, 31, "Subsection",ExpressionUUID->"b2d4dae3-8fcd-4813-89b9-e475fa132b8a",
 CellID->468103083],
Cell[255563, 7376, 102, 1, 12, "OpenEnvironment",ExpressionUUID->"87eab77c-8cfa-4157-807e-4e76d59c7276",
 CellID->441329419],
Cell[CellGroupData[{
Cell[255690, 7381, 436, 9, 38, "EnvironmentHeader",ExpressionUUID->"65e5f855-cb62-45c6-af2e-1b5b1b8a6956",
 CellID->113481011],
Cell[256129, 7392, 1579, 53, 50, "FormalTextInputFormula",ExpressionUUID->"87cdeabf-648d-464d-b035-5d143d98c87f",
 CellTags->"\[Element]Insert",
 CellID->833542610],
Cell[257711, 7447, 1020, 30, 50, "FormalTextInputFormula",ExpressionUUID->"3dfc2ea8-5b58-4bed-bc5a-781776a474ee",
 CellTags->"InsertI1",
 CellID->265974781],
Cell[258734, 7479, 1156, 36, 50, "FormalTextInputFormula",ExpressionUUID->"a53e2216-ee84-4b9a-9367-f3854ab8dbba",
 CellTags->"InsertI1'",
 CellID->369554344],
Cell[259893, 7517, 1154, 36, 50, "FormalTextInputFormula",ExpressionUUID->"eee79b47-d910-49b8-b643-3d87423df22a",
 CellTags->"InsertI2",
 CellID->276809760],
Cell[261050, 7555, 1513, 49, 50, "FormalTextInputFormula",ExpressionUUID->"94305d11-47e3-4dc3-a7b2-5f685d28a398",
 CellTags->{"InsertI classical", "intro"},
 CellID->794800661],
Cell[262566, 7606, 1547, 49, 50, "FormalTextInputFormula",ExpressionUUID->"e669061b-1a1d-4dbf-970c-ca9f87768bd7",
 CellTags->{"InsertE", "elim"},
 CellID->752025544],
Cell[264116, 7657, 1246, 39, 54, "FormalTextInputFormula",ExpressionUUID->"2c22bdf3-e000-413b-9d7a-df4eabf07af9",
 CellTags->"Insert alt",
 CellID->60042161],
Cell[265365, 7698, 120, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"c6618d7d-8e37-46d8-809b-07a6244df8a0",
 CellID->418443738]
}, Open  ]],
Cell[265500, 7702, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"5cd13aa9-2f40-41ee-af41-e9247a2ba5b2",
 CellID->162133543],
Cell[265606, 7705, 1809, 47, 29, "ProofInfo",ExpressionUUID->"69efc2d1-78ff-494c-9afc-e15c84128c04",
 Evaluatable->True,
 CellTags->{"Proof|ID:833542610-1", "3"},
 CellID->1603093206],
Cell[267418, 7754, 1763, 45, 29, "ProofInfo",ExpressionUUID->"e9466e7d-028a-4117-ac46-65b9a92ddf8e",
 Evaluatable->True,
 CellTags->{"Proof|ID:265974781-1", "3"},
 CellID->1283043907],
Cell[269184, 7801, 1764, 45, 29, "ProofInfo",ExpressionUUID->"9dffdc4b-0a67-4869-8254-6e3cf0381dfa",
 Evaluatable->True,
 CellTags->{"Proof|ID:369554344-1", "3"},
 CellID->111935385],
Cell[270951, 7848, 1762, 45, 29, "ProofInfo",ExpressionUUID->"53cab25a-3a17-4a43-97f3-cc2b778e43fc",
 Evaluatable->True,
 CellTags->{"Proof|ID:276809760-1", "3"},
 CellID->591244542],
Cell[272716, 7895, 1811, 47, 29, "ProofInfo",ExpressionUUID->"893df1bb-a5ff-4f0d-a86c-b9a2a51b065b",
 Evaluatable->True,
 CellTags->{"Proof|ID:794800661-1", "3"},
 CellID->1050155476],
Cell[274530, 7944, 1761, 45, 29, "ProofInfo",ExpressionUUID->"8908cda5-38e0-4ff0-806e-c21baafb2a94",
 Evaluatable->True,
 CellTags->{"Proof|ID:752025544-1", "3"},
 CellID->1183293999],
Cell[276294, 7991, 1775, 46, 29, "ProofInfo",ExpressionUUID->"acf349b9-ba5d-40be-9cc8-8cf74e4537e3",
 Evaluatable->True,
 CellTags->{"Proof|ID:60042161-1", "3"},
 CellID->807514983],
Cell[278072, 8039, 316, 7, 44, "TheoremaCode",ExpressionUUID->"cb9b2ff7-9b03-4f49-a797-f9f6c3e9e115",
 CellID->2044667126],
Cell[278391, 8048, 102, 1, 12, "OpenEnvironment",ExpressionUUID->"c5bf61ce-25c6-4a30-a6b9-924af40e911f",
 CellID->840329746],
Cell[CellGroupData[{
Cell[278518, 8053, 493, 10, 38, "EnvironmentHeader",ExpressionUUID->"c0587a64-5e71-4530-bace-5373a518f661",
 CellID->955019610],
Cell[279014, 8065, 947, 25, 50, "FormalTextInputFormula",ExpressionUUID->"1378f62a-f8a1-4388-a3e7-d01a5c35792f",
 CellTags->"\[SubsetEqual]Insert I",
 CellID->499325579],
Cell[279964, 8092, 1189, 36, 50, "FormalTextInputFormula",ExpressionUUID->"27164a92-0285-4493-bacb-7c74cb461d52",
 CellTags->"\[SubsetEqual]Insert I2",
 CellID->89181006],
Cell[281156, 8130, 1550, 49, 50, "FormalTextInputFormula",ExpressionUUID->"fe809dfb-5875-4485-ae29-7091cd6e4d62",
 CellTags->"\[SubsetEqual]Insert",
 CellID->17294591],
Cell[282709, 8181, 1608, 50, 56, "FormalTextInputFormula",ExpressionUUID->"ad82da62-4b0a-490c-8806-1851e7b37e9b",
 CellTags->"\[SubsetEqual]Insert iff",
 CellID->1726053781],
Cell[284320, 8233, 1380, 43, 50, "FormalTextInputFormula",ExpressionUUID->"23802000-c362-4a0b-afa6-0ac37540f0d2",
 CellTags->"Insert\[SubsetEqual] iff",
 CellID->140408834],
Cell[285703, 8278, 121, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"605767b6-42d2-40b2-bbbf-59adf70a6090",
 CellID->1197911587]
}, Open  ]],
Cell[285839, 8282, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"13f5ce1e-b466-42c4-b2a1-1a8e64b5d2db",
 CellID->408625003],
Cell[285945, 8285, 1821, 47, 29, "ProofInfo",ExpressionUUID->"033295b6-03ea-4743-b02f-9a029d18a564",
 Evaluatable->True,
 CellTags->{"Proof|ID:499325579-1", "3"},
 CellID->1989694051],
Cell[287769, 8334, 1816, 47, 29, "ProofInfo",ExpressionUUID->"74457cd9-9415-4a5b-a6ff-c4224aff06d1",
 Evaluatable->True,
 CellTags->{"Proof|ID:89181006-1", "3"},
 CellID->169658549],
Cell[289588, 8383, 1809, 47, 29, "ProofInfo",ExpressionUUID->"539ab388-aaaf-402a-84e8-d34f4c88e778",
 Evaluatable->True,
 CellTags->{"Proof|ID:17294591-1", "3"},
 CellID->94829629],
Cell[291400, 8432, 1828, 47, 29, "ProofInfo",ExpressionUUID->"a019316d-6449-4592-a608-6892d07b6388",
 Evaluatable->True,
 CellTags->{"Proof|ID:1726053781-1", "3"},
 CellID->1321149923],
Cell[293231, 8481, 1824, 47, 29, "ProofInfo",ExpressionUUID->"72ddb486-2189-446a-b4f0-dccfaf5bf1dc",
 Evaluatable->True,
 CellTags->{"Proof|ID:140408834-1", "3"},
 CellID->662871545],
Cell[295058, 8530, 383, 7, 44, "TheoremaCode",ExpressionUUID->"313ba586-bc6d-4b7d-91aa-122f9e5bdcfe",
 CellID->781604907]
}, Closed]],
Cell[CellGroupData[{
Cell[295478, 8542, 174, 2, 31, "Subsection",ExpressionUUID->"4cb24e1a-2154-43e3-b47a-fb2ec281cec9",
 CellID->1045098640],
Cell[295655, 8546, 103, 1, 12, "OpenEnvironment",ExpressionUUID->"ecb922b7-b44e-49bc-aa68-afb8aafac54e",
 CellID->1232065355],
Cell[CellGroupData[{
Cell[295783, 8551, 443, 9, 38, "EnvironmentHeader",ExpressionUUID->"678aade1-04d8-4c3d-b135-cd7641fc0403",
 CellID->1042012196],
Cell[296229, 8562, 860, 23, 47, "FormalTextInputFormula",ExpressionUUID->"154c747d-9b88-42a4-8a21-26f3dd0c0a24",
 CellTags->{"singleton I", "intro"},
 CellID->1706541997],
Cell[297092, 8587, 1112, 35, 50, "FormalTextInputFormula",ExpressionUUID->"390adf86-f8b8-4097-9da9-ab6551101d71",
 CellTags->"singleton D",
 CellID->1946555487],
Cell[298207, 8624, 1146, 36, 51, "FormalTextInputFormula",ExpressionUUID->"cb45ee28-969d-4085-b7bf-1d4e9c010512",
 CellTags->"singleton inject",
 CellID->1642372132],
Cell[299356, 8662, 120, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"08a1df6f-26c7-4bce-bc36-fc4843beb8c6",
 CellID->776643782]
}, Open  ]],
Cell[299491, 8666, 104, 1, 16, "CloseEnvironment",ExpressionUUID->"fe979dd2-f120-4cb7-a66a-354d90db237a",
 CellID->1880661924],
Cell[299598, 8669, 1787, 46, 29, "ProofInfo",ExpressionUUID->"24b89c28-f6f2-4a2d-8f6c-d7faa64ba567",
 Evaluatable->True,
 CellTags->{"Proof|ID:1706541997-1", "3"},
 CellID->1610428281],
Cell[301388, 8717, 1785, 46, 29, "ProofInfo",ExpressionUUID->"c797b61f-6a99-4d82-9375-e86eb679b1f6",
 Evaluatable->True,
 CellTags->{"Proof|ID:1946555487-1", "3"},
 CellID->44899502],
Cell[303176, 8765, 1811, 47, 29, "ProofInfo",ExpressionUUID->"2d77d2ad-50ea-45a7-8a70-0483b95c475c",
 Evaluatable->True,
 CellTags->{"Proof|ID:1642372132-1", "3"},
 CellID->259575914],
Cell[304990, 8814, 329, 6, 44, "TheoremaCode",ExpressionUUID->"8e8464d3-eac7-4d6a-ad01-5226f6a46e5a",
 CellID->1158961922],
Cell[305322, 8822, 103, 1, 12, "OpenEnvironment",ExpressionUUID->"29362c51-68af-4f8a-9f81-27252ecddf5a",
 CellID->1073790972],
Cell[CellGroupData[{
Cell[305450, 8827, 448, 9, 38, "EnvironmentHeader",ExpressionUUID->"bf79ca0b-9ea0-44ab-857b-ef58982a7927",
 CellID->1080577362],
Cell[305901, 8838, 1604, 54, 50, "FormalTextInputFormula",ExpressionUUID->"ba3c94b4-891e-4275-8bbf-58fc308ba014",
 CellTags->"\[SubsetEqual] singleton",
 CellID->574835025],
Cell[307508, 8894, 1257, 40, 50, "FormalTextInputFormula",ExpressionUUID->"987ead35-8676-46b5-86a2-3f01a4952119",
 CellTags->"singleton \[SubsetEqual]",
 CellID->316263559],
Cell[308768, 8936, 120, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"139e9921-f326-44c0-a744-c2fd7ef29831",
 CellID->636142242]
}, Open  ]],
Cell[308903, 8940, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"bf48eb20-6c77-4107-b527-21098333b892",
 CellID->903528505],
Cell[309009, 8943, 1824, 47, 29, "ProofInfo",ExpressionUUID->"b866f190-0698-4884-9f66-30d4006e9fd2",
 Evaluatable->True,
 CellTags->{"Proof|ID:574835025-1", "3"},
 CellID->991759232],
Cell[310836, 8992, 1824, 47, 29, "ProofInfo",ExpressionUUID->"61354163-ddd7-4786-856c-489f55e87f87",
 Evaluatable->True,
 CellTags->{"Proof|ID:316263559-1", "3"},
 CellID->811999371]
}, Closed]],
Cell[CellGroupData[{
Cell[312697, 9044, 184, 2, 31, "Subsection",ExpressionUUID->"42bf86ae-570f-4237-bfc5-18eeb00d39a2",
 CellID->633366498],
Cell[312884, 9048, 101, 1, 12, "OpenEnvironment",ExpressionUUID->"e7046d20-e426-433f-ae66-ffea17fe847c",
 CellID->38110315],
Cell[CellGroupData[{
Cell[313010, 9053, 431, 9, 38, "EnvironmentHeader",ExpressionUUID->"1ba0e1b6-a89f-4a7b-8466-b6c7139908fa",
 CellID->805541198],
Cell[313444, 9064, 1354, 43, 51, "FormalTextInputFormula",ExpressionUUID->"81f4ee41-7d1b-4661-86c1-cd51f6a030b1",
 CellTags->{"@I eq", "intro"},
 CellID->1595981918],
Cell[314801, 9109, 1322, 42, 51, "FormalTextInputFormula",ExpressionUUID->"31466d48-898a-47c6-9519-350697bfe016",
 CellTags->"@I eq reverse",
 CellID->796685651],
Cell[316126, 9153, 1125, 35, 51, "FormalTextInputFormula",ExpressionUUID->"0ce71d43-56f3-43ae-aa1a-139f021caf78",
 CellTags->"@I",
 CellID->1022035164],
Cell[317254, 9190, 1528, 48, 55, "FormalTextInputFormula",ExpressionUUID->"27b53510-2aa4-42f2-ac57-bb62701ff6aa",
 CellTags->{"@E", "elim"},
 CellID->114344294],
Cell[318785, 9240, 1482, 49, 55, "FormalTextInputFormula",ExpressionUUID->"87637558-3a97-43a1-b39c-38a749176477",
 CellTags->"\[Element]@",
 CellID->1795699175],
Cell[320270, 9291, 1657, 54, 55, "FormalTextInputFormula",ExpressionUUID->"6066f4bb-70bd-49b6-9c7f-ad807e7e6782",
 CellTags->"@ cong",
 CellID->724997673],
Cell[321930, 9347, 1425, 46, 55, "FormalTextInputFormula",ExpressionUUID->"90ad9d86-50dd-4746-b1e9-0a5d82ed98c4",
 CellTags->"@ ident I",
 CellID->803636397],
Cell[323358, 9395, 120, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"a27a8bf6-b62b-4690-9fa2-88cba8d59229",
 CellID->369196157]
}, Open  ]],
Cell[323493, 9399, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"2add8c0f-f5fa-47f5-8a8e-55d3e9d8d11c",
 CellID->717570174],
Cell[323599, 9402, 1759, 45, 29, "ProofInfo",ExpressionUUID->"b5119a56-9b6b-4dc9-baca-4b038f13e49d",
 Evaluatable->True,
 CellTags->{"Proof|ID:1595981918-1", "3"},
 CellID->638235692],
Cell[325361, 9449, 1788, 46, 29, "ProofInfo",ExpressionUUID->"8db8df19-fdf9-4fa7-b82c-0b2e4f83d460",
 Evaluatable->True,
 CellTags->{"Proof|ID:796685651-1", "3"},
 CellID->1222338081],
Cell[327152, 9497, 1754, 45, 29, "ProofInfo",ExpressionUUID->"6d754d5f-fbc5-4f0b-9d25-e0f40502431f",
 Evaluatable->True,
 CellTags->{"Proof|ID:1022035164-1", "3"},
 CellID->1120566427],
Cell[328909, 9544, 1751, 45, 29, "ProofInfo",ExpressionUUID->"0d43d638-054b-4e7b-a14f-d856c779b0cc",
 Evaluatable->True,
 CellTags->{"Proof|ID:114344294-1", "3"},
 CellID->1155434313],
Cell[330663, 9591, 1787, 46, 29, "ProofInfo",ExpressionUUID->"fc5a6e81-1d94-4b7a-9f62-d18cc426bf95",
 Evaluatable->True,
 CellTags->{"Proof|ID:1795699175-1", "3"},
 CellID->1210186748],
Cell[332453, 9639, 1758, 45, 29, "ProofInfo",ExpressionUUID->"c828c5b9-b548-4ffe-a01b-69ead7885f9f",
 Evaluatable->True,
 CellTags->{"Proof|ID:724997673-1", "3"},
 CellID->637413599],
Cell[334214, 9686, 1765, 45, 29, "ProofInfo",ExpressionUUID->"ed4bf583-b592-4cce-934f-774b532b3b39",
 Evaluatable->True,
 CellTags->{"Proof|ID:803636397-1", "3"},
 CellID->1659057562],
Cell[335982, 9733, 103, 1, 12, "OpenEnvironment",ExpressionUUID->"98413215-ec07-4604-a4d8-fec106943722",
 CellID->1195143136],
Cell[CellGroupData[{
Cell[336110, 9738, 436, 9, 38, "EnvironmentHeader",ExpressionUUID->"fb5b4c3b-c7ca-4788-a384-5e7a4b4dacc1",
 CellID->1012321715],
Cell[336549, 9749, 1458, 46, 55, "FormalTextInputFormula",ExpressionUUID->"5cc6be16-8334-4ca9-b65d-a945b6a3bac8",
 CellTags->"@\[SubsetEqual] I",
 CellID->1322094901],
Cell[338010, 9797, 1523, 47, 55, "FormalTextInputFormula",ExpressionUUID->"b58a06f0-0329-40ed-ba74-3f6c05ed1672",
 CellTags->"\[SubsetEqual]@ E",
 CellID->1691291058],
Cell[339536, 9846, 1016, 28, 51, "FormalTextInputFormula",ExpressionUUID->"270a4307-6358-4635-9c32-2bef59493519",
 CellTags->"@\[Intersection]",
 CellID->1830553618],
Cell[340555, 9876, 953, 27, 51, "FormalTextInputFormula",ExpressionUUID->"b21e6dec-dd9c-482a-aafe-5ea92234662d",
 CellTags->"@-",
 CellID->398205279],
Cell[341511, 9905, 121, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"f837b76c-8d58-40cc-95d3-42a9bbaed8b8",
 CellID->2000684334]
}, Open  ]],
Cell[341647, 9909, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"6740f01c-0997-4fc5-8b37-c64b52955f49",
 CellID->460819985],
Cell[341753, 9912, 1813, 47, 29, "ProofInfo",ExpressionUUID->"57da981f-f479-4764-8da0-d6a7a3c080ec",
 Evaluatable->True,
 CellTags->{"Proof|ID:1322094901-1", "3"},
 CellID->892892992],
Cell[343569, 9961, 1813, 47, 29, "ProofInfo",ExpressionUUID->"7588b18e-aa27-40ac-8f3f-f9e9ae729362",
 Evaluatable->True,
 CellTags->{"Proof|ID:1691291058-1", "3"},
 CellID->903150912],
Cell[345385, 10010, 1812, 47, 29, "ProofInfo",ExpressionUUID->"85d59c5c-1a6e-4476-8410-ecd019893f54",
 Evaluatable->True,
 CellTags->{"Proof|ID:1830553618-1", "3"},
 CellID->1214598633],
Cell[347200, 10059, 1751, 45, 29, "ProofInfo",ExpressionUUID->"16659fac-0e9b-4219-a757-d5db7b81bc81",
 Evaluatable->True,
 CellTags->{"Proof|ID:398205279-1", "3"},
 CellID->1641967303]
}, Closed]],
Cell[CellGroupData[{
Cell[348988, 10109, 170, 2, 31, "Subsection",ExpressionUUID->"0ee84eed-3054-4410-8186-99c14e7a5127",
 CellID->487407862],
Cell[349161, 10113, 510, 13, 52, "TheoremaCode",ExpressionUUID->"1d7b6ce8-4f66-43d1-a7c5-bbedde1496c0",
 CellID->492828318],
Cell[349674, 10128, 101, 1, 12, "OpenEnvironment",ExpressionUUID->"ada86cdc-f85f-4fc3-9bce-913768b66ba6",
 CellID->20339922],
Cell[CellGroupData[{
Cell[349800, 10133, 440, 9, 38, "EnvironmentHeader",ExpressionUUID->"3c5e1ff0-a3f1-429d-94af-2405d5cdad7a",
 CellID->1961526212],
Cell[350243, 10144, 1202, 37, 51, "FormalTextInputFormula",ExpressionUUID->"65d206a7-f149-48f0-8dc8-445143bf53a0",
 CellTags->{"RangeI eq", "intro"},
 CellID->686090881],
Cell[351448, 10183, 867, 24, 51, "FormalTextInputFormula",ExpressionUUID->"4f9de9e6-85d8-491a-9fd8-f47ebc444eae",
 CellTags->"RangeI",
 CellID->1169658618],
Cell[352318, 10209, 1319, 41, 55, "FormalTextInputFormula",ExpressionUUID->"a0d58dda-dbda-43ce-b842-8bf69646e1dc",
 CellTags->"RangeE",
 CellID->1129746621],
Cell[353640, 10252, 1198, 37, 51, "FormalTextInputFormula",ExpressionUUID->"35257160-6070-44f3-b494-ab56075ee0e6",
 CellTags->"Range singleton D",
 CellID->1805361717],
Cell[354841, 10291, 120, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"7b7fc611-b97e-459b-b46f-5cbc8b7a53c9",
 CellID->730000772]
}, Open  ]],
Cell[354976, 10295, 104, 1, 16, "CloseEnvironment",ExpressionUUID->"b2bfdac4-916a-4690-9d8c-92eaeeb272a6",
 CellID->2127657008],
Cell[355083, 10298, 1765, 45, 29, "ProofInfo",ExpressionUUID->"6e01bd25-a1a6-443d-b697-ce5e66313c66",
 Evaluatable->True,
 CellTags->{"Proof|ID:686090881-1", "3"},
 CellID->1329697308],
Cell[356851, 10345, 1760, 45, 29, "ProofInfo",ExpressionUUID->"8e614f57-fd84-4b20-aada-099cb8f5fd99",
 Evaluatable->True,
 CellTags->{"Proof|ID:1169658618-1", "3"},
 CellID->65583140],
Cell[358614, 10392, 1762, 45, 29, "ProofInfo",ExpressionUUID->"ec01b3f4-f18b-473a-adaa-b64c4b1794c1",
 Evaluatable->True,
 CellTags->{"Proof|ID:1129746621-1", "3"},
 CellID->1205185338],
Cell[360379, 10439, 1813, 47, 29, "ProofInfo",ExpressionUUID->"796ac993-5d77-429a-8ec4-94a6354c467e",
 Evaluatable->True,
 CellTags->{"Proof|ID:1805361717-1", "3"},
 CellID->774665428]
}, Closed]],
Cell[CellGroupData[{
Cell[362229, 10491, 182, 2, 31, "Subsection",ExpressionUUID->"1bf74fab-e60d-4c22-8562-ea637d43a537",
 CellID->312522199],
Cell[362414, 10495, 103, 1, 12, "OpenEnvironment",ExpressionUUID->"7efc5894-c3fc-4ee4-85d6-d1111ba346ed",
 CellID->1020901598],
Cell[CellGroupData[{
Cell[362542, 10500, 448, 9, 38, "EnvironmentHeader",ExpressionUUID->"1ef8f7c3-8722-45c0-aadd-88cabf8fbe16",
 CellID->1926779225],
Cell[362993, 10511, 1609, 53, 62, "FormalTextInputFormula",ExpressionUUID->"7a7bc97e-48fb-447b-ba94-7c245728934d",
 CellTags->"\[Element]SetOf",
 CellID->718424401],
Cell[364605, 10566, 1376, 42, 64, "FormalTextInputFormula",ExpressionUUID->"ba2298fa-7bbb-45c0-b5bb-c841cdf4f95a",
 CellTags->{"SetOf I", "intro"},
 CellID->783246906],
Cell[365984, 10610, 1544, 47, 64, "FormalTextInputFormula",ExpressionUUID->"6eb18b5a-43f0-4bf0-91cb-fa1b2deac1c6",
 CellTags->{"SetOf E", "elim"},
 CellID->1984656332],
Cell[367531, 10659, 120, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"7b4d5aff-6d8b-4198-bec5-ce161dfa81f5",
 CellID->523083548]
}, Open  ]],
Cell[367666, 10663, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"98f8d90f-f7e3-46c9-9808-6cb2ecad7f44",
 CellID->663433186],
Cell[367772, 10666, 1791, 46, 29, "ProofInfo",ExpressionUUID->"62e40651-f209-4327-b387-490f4c46bb3d",
 Evaluatable->True,
 CellTags->{"Proof|ID:718424401-1", "3"},
 CellID->961492024],
Cell[369566, 10714, 1761, 45, 29, "ProofInfo",ExpressionUUID->"8842b459-89de-4cca-be91-b8c2e9a5746a",
 Evaluatable->True,
 CellTags->{"Proof|ID:783246906-1", "3"},
 CellID->1483483416],
Cell[371330, 10761, 1764, 45, 29, "ProofInfo",ExpressionUUID->"3432589c-d2fc-4a23-a0c8-b28e516663cd",
 Evaluatable->True,
 CellTags->{"Proof|ID:1984656332-1", "3"},
 CellID->1243944656]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[373143, 10812, 191, 2, 43, "Section",ExpressionUUID->"f4b65ffd-44d6-457b-82f5-f4dd34cd4a9d",
 CellID->176566893],
Cell[CellGroupData[{
Cell[373359, 10818, 181, 2, 39, "Subsection",ExpressionUUID->"7d55af62-2424-4ec7-a94f-63e121930d20",
 CellID->2439801],
Cell[373543, 10822, 102, 1, 12, "OpenEnvironment",ExpressionUUID->"4bbf9cbd-c25f-4f71-8364-536ea8b87204",
 CellID->339955184],
Cell[CellGroupData[{
Cell[373670, 10827, 508, 10, 38, "EnvironmentHeader",ExpressionUUID->"81de3f1e-7371-4d87-ae62-c696956b807f",
 CellID->1466695304],
Cell[374181, 10839, 1049, 31, 62, "FormalTextInputFormula",ExpressionUUID->"03aff8c6-ce79-4896-9c4b-ff5194aa4795",
 CellTags->"SetOf\[Equal]Collect",
 CellID->314660242],
Cell[375233, 10872, 1279, 40, 62, "FormalTextInputFormula",ExpressionUUID->"43821a10-5e3f-462a-b9d8-7e55cb502399",
 CellTags->"SetOf const",
 CellID->222426804],
Cell[376515, 10914, 1585, 52, 62, "FormalTextInputFormula",ExpressionUUID->"7a67e7ee-094b-4cec-89ae-429266833c16",
 CellTags->"SetOf\[Equal]{} iff",
 CellID->1237361888],
Cell[378103, 10968, 1300, 42, 62, "FormalTextInputFormula",ExpressionUUID->"bc81c816-825b-4560-8ed5-0038c9d1cc89",
 CellTags->"SetOf\[Not]",
 CellID->165006998],
Cell[379406, 11012, 1732, 56, 62, "FormalTextInputFormula",ExpressionUUID->"3e4ae205-39dd-4613-a34d-75a3ca68670a",
 CellTags->"SetOf\[Or]",
 CellID->1855260115],
Cell[381141, 11070, 1710, 55, 62, "FormalTextInputFormula",ExpressionUUID->"ca2c40f5-3c9f-4d06-818b-6843e4fc238f",
 CellTags->"SetOf\[Implies]",
 CellID->985026514],
Cell[382854, 11127, 1734, 54, 62, "FormalTextInputFormula",ExpressionUUID->"e7181290-0067-4cd3-bfdd-00152da0d246",
 CellTags->"Collect\[And]",
 CellID->636400335],
Cell[384591, 11183, 1589, 51, 62, "FormalTextInputFormula",ExpressionUUID->"8003e81d-2011-440e-b291-7344a40018ad",
 CellTags->"SetOf mono iff",
 CellID->859120159],
Cell[386183, 11236, 118, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"a43204a5-e280-4589-9668-2b3b6887ef3e",
 CellID->2284427]
}, Open  ]],
Cell[386316, 11240, 104, 1, 16, "CloseEnvironment",ExpressionUUID->"1f713404-7561-4a92-94ec-d3b14a2bd4b0",
 CellID->2066769854],
Cell[386423, 11243, 1816, 47, 29, "ProofInfo",ExpressionUUID->"90ce6af8-8154-479f-a67a-eb82f1369496",
 Evaluatable->True,
 CellTags->{"Proof|ID:314660242-1", "3"},
 CellID->462247040],
Cell[388242, 11292, 1784, 46, 29, "ProofInfo",ExpressionUUID->"23123a44-70d0-4b7e-af9d-66586f2c7c3b",
 Evaluatable->True,
 CellTags->{"Proof|ID:222426804-1", "3"},
 CellID->1447843469],
Cell[390029, 11340, 1817, 47, 29, "ProofInfo",ExpressionUUID->"216bcc67-94ae-452d-9f9d-fb1577639bae",
 Evaluatable->True,
 CellTags->{"Proof|ID:1237361888-1", "3"},
 CellID->779088903],
Cell[391849, 11389, 1784, 46, 29, "ProofInfo",ExpressionUUID->"5156a1b2-c4e6-447c-8758-161115f33ced",
 Evaluatable->True,
 CellTags->{"Proof|ID:165006998-1", "3"},
 CellID->1390818865],
Cell[393636, 11437, 1784, 46, 29, "ProofInfo",ExpressionUUID->"627a6140-2e49-4960-8e4a-725e778a5b7a",
 Evaluatable->True,
 CellTags->{"Proof|ID:1855260115-1", "3"},
 CellID->859972040],
Cell[395423, 11485, 1792, 46, 29, "ProofInfo",ExpressionUUID->"d35e54e6-1cd4-4c97-a6a7-67a76bf2b9ca",
 Evaluatable->True,
 CellTags->{"Proof|ID:985026514-1", "3"},
 CellID->2146486548],
Cell[397218, 11533, 1788, 46, 29, "ProofInfo",ExpressionUUID->"3f1308e2-2bf7-43a6-b603-e554f2db72a9",
 Evaluatable->True,
 CellTags->{"Proof|ID:636400335-1", "3"},
 CellID->1733058477],
Cell[399009, 11581, 1790, 46, 29, "ProofInfo",ExpressionUUID->"43fc8102-da0a-4d32-8477-1d0fb499a635",
 Evaluatable->True,
 CellTags->{"Proof|ID:859120159-1", "3"},
 CellID->1491492338],
Cell[400802, 11629, 321, 6, 44, "TheoremaCode",ExpressionUUID->"8ba5107c-90ea-4a93-8483-b37b37bd82a7",
 CellID->549704930],
Cell[401126, 11637, 103, 1, 12, "OpenEnvironment",ExpressionUUID->"36e3f8b0-332b-4db3-b8d3-0dec5ffafd3b",
 CellID->1157303599],
Cell[CellGroupData[{
Cell[401254, 11642, 496, 10, 38, "EnvironmentHeader",ExpressionUUID->"d310918b-c9e9-4184-9dad-829213bfcc3c",
 CellID->2127892673],
Cell[401753, 11654, 1237, 40, 47, "FormalTextInputFormula",ExpressionUUID->"3055a66b-2e7f-420c-b9ad-a698d5a357e9",
 CellTags->"\[SubsetEqual]{}",
 CellID->485185733],
Cell[402993, 11696, 1290, 41, 51, "FormalTextInputFormula",ExpressionUUID->"6cb5ed56-c05b-44f8-91f8-042d5b96b3ca",
 CellTags->"\[ForAll]\[NotElement] conv",
 CellID->775376237],
Cell[404286, 11739, 1286, 41, 51, "FormalTextInputFormula",ExpressionUUID->"c11276ab-818d-47e0-906b-fa4426b2d132",
 CellTags->"\[Exists]\[Element] conv",
 CellID->978065715],
Cell[405575, 11782, 120, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"f5257999-7982-4354-b4df-cb208c4d9524",
 CellID->965328407]
}, Open  ]],
Cell[405710, 11786, 104, 1, 16, "CloseEnvironment",ExpressionUUID->"b35b36d4-bd17-448f-8b9d-11f16caf926c",
 CellID->2063758297],
Cell[405817, 11789, 1808, 47, 29, "ProofInfo",ExpressionUUID->"533d16e0-ce09-469a-92e6-88dd693c1a07",
 Evaluatable->True,
 CellTags->{"Proof|ID:485185733-1", "3"},
 CellID->851152370],
Cell[407628, 11838, 1830, 47, 29, "ProofInfo",ExpressionUUID->"d64e40de-107c-465e-a04b-e0d1316fc389",
 Evaluatable->True,
 CellTags->{"Proof|ID:775376237-1", "3"},
 CellID->497842656],
Cell[409461, 11887, 1824, 47, 29, "ProofInfo",ExpressionUUID->"3dea71aa-c69b-4215-9524-803e9c59e9f2",
 Evaluatable->True,
 CellTags->{"Proof|ID:978065715-1", "3"},
 CellID->118302286],
Cell[411288, 11936, 381, 7, 44, "TheoremaCode",ExpressionUUID->"a44bb603-ab15-474b-97cc-c13669158c15",
 CellID->1615646849],
Cell[411672, 11945, 102, 1, 12, "OpenEnvironment",ExpressionUUID->"25e40008-4442-464c-be2b-75e1dcfb5b1f",
 CellID->159553041],
Cell[CellGroupData[{
Cell[411799, 11950, 452, 9, 38, "EnvironmentHeader",ExpressionUUID->"755f2645-02c2-414b-b25d-98e22fddc29f",
 CellID->1812937043],
Cell[412254, 11961, 949, 28, 47, "FormalTextInputFormula",ExpressionUUID->"2d432965-27dc-4fc0-99e8-0b89f9dcb74f",
 CellTags->"Insert UNIV",
 CellID->988950467],
Cell[413206, 11991, 1048, 31, 50, "FormalTextInputFormula",ExpressionUUID->"55070231-9ff3-4351-bf0a-aab3ba4c1427",
 CellTags->"Insert\[Equal]\[Union]",
 CellID->1780572558],
Cell[414257, 12024, 1007, 30, 50, "FormalTextInputFormula",ExpressionUUID->"7464ffb2-77ec-46ff-a00f-bf067e26bbb5",
 CellTags->"Insert\[NotEqual]{}",
 CellID->1795020389],
Cell[415267, 12056, 1155, 36, 50, "FormalTextInputFormula",ExpressionUUID->"7472c75c-9564-488e-935f-bd257382c5ec",
 CellTags->"Insert absorb",
 CellID->2125656893],
Cell[416425, 12094, 1099, 33, 50, "FormalTextInputFormula",ExpressionUUID->"f27f74b0-b865-4980-83e7-8f44091d7719",
 CellTags->"Insert absorb2",
 CellID->742456464],
Cell[417527, 12129, 1171, 35, 51, "FormalTextInputFormula",ExpressionUUID->"fd79fcdc-631d-4c21-9d76-e467595a23ef",
 CellTags->"Insert comm",
 CellID->1038355606],
Cell[418701, 12166, 1392, 44, 62, "FormalTextInputFormula",ExpressionUUID->"34779372-bb66-44fe-b2b6-9f99f0d10da0",
 CellTags->"Insert SetOf",
 CellID->657964489],
Cell[420096, 12212, 1326, 40, 50, "FormalTextInputFormula",ExpressionUUID->"b767ae43-32fd-4d40-9226-48275463435c",
 CellTags->"Insert\[Intersection]Insert",
 CellID->1602428504],
Cell[421425, 12254, 1599, 52, 50, "FormalTextInputFormula",ExpressionUUID->"be8df8a3-7c41-4241-8b84-f2d34ca0f3fb",
 CellTags->"Insert disjoint",
 CellID->621646007],
Cell[423027, 12308, 1601, 52, 50, "FormalTextInputFormula",ExpressionUUID->"c9610130-bb90-4d52-a0f4-3b30a68d9d28",
 CellTags->"disjoint Insert",
 CellID->656967390],
Cell[424631, 12362, 121, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"03e57289-00f9-42c5-a6ac-c9e4cdaf7111",
 CellID->1511772922]
}, Open  ]],
Cell[424767, 12366, 104, 1, 16, "CloseEnvironment",ExpressionUUID->"2e8b4d52-ee48-4749-82ff-cb4b4c371c49",
 CellID->1940050514],
Cell[424874, 12369, 1783, 46, 29, "ProofInfo",ExpressionUUID->"711e6551-2491-4f23-a69e-6ba95e9ee3c5",
 Evaluatable->True,
 CellTags->{"Proof|ID:988950467-1", "3"},
 CellID->207220836],
Cell[426660, 12417, 1823, 47, 29, "ProofInfo",ExpressionUUID->"ce02b660-2efc-48d5-a410-83f64d3cb0a4",
 Evaluatable->True,
 CellTags->{"Proof|ID:1780572558-1", "3"},
 CellID->709620784],
Cell[428486, 12466, 1817, 47, 29, "ProofInfo",ExpressionUUID->"a593fce4-239a-48ba-a634-0803bc3ce5f5",
 Evaluatable->True,
 CellTags->{"Proof|ID:1795020389-1", "3"},
 CellID->201322071],
Cell[430306, 12515, 1791, 46, 29, "ProofInfo",ExpressionUUID->"1c164155-3908-47ea-80ff-28bb818b47e3",
 Evaluatable->True,
 CellTags->{"Proof|ID:2125656893-1", "3"},
 CellID->1679122200],
Cell[432100, 12563, 1789, 46, 29, "ProofInfo",ExpressionUUID->"fdfb52c1-1f7b-42b0-8e1f-bfb7551b6086",
 Evaluatable->True,
 CellTags->{"Proof|ID:742456464-1", "3"},
 CellID->292466109],
Cell[433892, 12611, 1786, 46, 29, "ProofInfo",ExpressionUUID->"0017ff7e-6593-4c7d-a1d7-1bfa18dabc62",
 Evaluatable->True,
 CellTags->{"Proof|ID:1038355606-1", "3"},
 CellID->600513012],
Cell[435681, 12659, 1786, 46, 29, "ProofInfo",ExpressionUUID->"864a6093-7b93-41ad-b279-fcdce277baa9",
 Evaluatable->True,
 CellTags->{"Proof|ID:657964489-1", "3"},
 CellID->1781110406],
Cell[437470, 12707, 1834, 47, 29, "ProofInfo",ExpressionUUID->"a1dd04b2-c9d9-4f5c-a60f-43df42997afa",
 Evaluatable->True,
 CellTags->{"Proof|ID:1602428504-1", "3"},
 CellID->1228710860],
Cell[439307, 12756, 1792, 46, 29, "ProofInfo",ExpressionUUID->"7df4f81d-c45a-43a2-a483-44ffc18f92a3",
 Evaluatable->True,
 CellTags->{"Proof|ID:621646007-1", "3"},
 CellID->1551927440],
Cell[441102, 12804, 1792, 46, 29, "ProofInfo",ExpressionUUID->"5a84cad7-a764-4fb6-b914-1258f0a32f54",
 Evaluatable->True,
 CellTags->{"Proof|ID:656967390-1", "3"},
 CellID->1616332482],
Cell[442897, 12852, 665, 11, 44, "TheoremaCode",ExpressionUUID->"927fb3a2-4e6b-4a39-bce2-8bad25de40e1",
 CellID->695666758],
Cell[443565, 12865, 103, 1, 12, "OpenEnvironment",ExpressionUUID->"6615bd85-1b25-42bb-ab9b-eff19f9ef5d9",
 CellID->1749328760],
Cell[CellGroupData[{
Cell[443693, 12870, 456, 9, 38, "EnvironmentHeader",ExpressionUUID->"daddb5e7-e3b2-45a5-992b-b90b3f4110cc",
 CellID->388765566],
Cell[444152, 12881, 945, 27, 47, "FormalTextInputFormula",ExpressionUUID->"fc55d8fa-8b90-4254-897c-2a6575a0f480",
 CellTags->"\[Intersection] absorb",
 CellID->125816224],
Cell[445100, 12910, 1074, 31, 50, "FormalTextInputFormula",ExpressionUUID->"78e30461-db5c-440a-a771-68e1af989dd5",
 CellTags->"\[Intersection] absorb'",
 CellID->848809596],
Cell[446177, 12943, 1002, 29, 50, "FormalTextInputFormula",ExpressionUUID->"ed2c3178-573c-441d-b883-c094a6cdf7fc",
 CellTags->"\[Intersection] comm",
 CellID->147215930],
Cell[447182, 12974, 1148, 33, 50, "FormalTextInputFormula",ExpressionUUID->"9303ebaf-4bff-47f8-b699-5c2016271cf7",
 CellTags->"\[Intersection] comm'",
 CellID->902394792],
Cell[448333, 13009, 1153, 34, 50, "FormalTextInputFormula",ExpressionUUID->"201ed866-e06f-4435-b662-35ed14257a22",
 CellTags->"\[Intersection] assoc",
 CellID->568250332],
Cell[449489, 13045, 1151, 35, 50, "FormalTextInputFormula",ExpressionUUID->"58f867f3-5e00-4674-b9e0-286a263e5f89",
 CellTags->"\[Intersection] absorb1",
 CellID->1826338130],
Cell[450643, 13082, 1153, 35, 50, "FormalTextInputFormula",ExpressionUUID->"9457e9e2-d7e8-4e67-a0fb-022e6c713401",
 CellTags->"\[Intersection] absorb2",
 CellID->1588902597],
Cell[451799, 13119, 975, 29, 47, "FormalTextInputFormula",ExpressionUUID->"ba8fd0f4-f2cd-4892-b0e0-8f45db7178ec",
 CellTags->"\[Intersection]{}",
 CellID->1253019475],
Cell[452777, 13150, 975, 29, 47, "FormalTextInputFormula",ExpressionUUID->"a5ac7dd9-350f-4831-9b39-0ac63d3d6e99",
 CellTags->"{}\[Intersection]",
 CellID->1274251332],
Cell[453755, 13181, 1324, 42, 50, "FormalTextInputFormula",ExpressionUUID->"6135b89a-9bc7-4d84-b9ab-fd95519b0153",
 CellTags->"disjoint iff \[SubsetEqual]-",
 CellID->903302990],
Cell[455082, 13225, 1482, 47, 57, "FormalTextInputFormula",ExpressionUUID->"76d4823b-68ee-4ac8-be6a-79c20b3ff6fb",
 CellTags->"disjoint iff \[NotEqual]",
 CellID->920640376],
Cell[456567, 13274, 945, 27, 47, "FormalTextInputFormula",ExpressionUUID->"7810a2e8-7dec-41dc-8f02-f32b02736408",
 CellTags->"\[Intersection]UNIV",
 CellID->1023611974],
Cell[457515, 13303, 943, 27, 47, "FormalTextInputFormula",ExpressionUUID->"187c8311-a111-480e-87b9-378d25295ce1",
 CellTags->"UNIV\[Intersection]",
 CellID->1195769099],
Cell[458461, 13332, 1226, 35, 50, "FormalTextInputFormula",ExpressionUUID->"aec71586-6dd1-456e-8894-29dcd9143ebd",
 CellTags->"\[Intersection]\[Union] dist left",
 CellID->56424478],
Cell[459690, 13369, 1235, 36, 50, "FormalTextInputFormula",ExpressionUUID->"252a8781-d3e3-43ef-8e2d-c57e47347aba",
 CellTags->"\[Intersection]\[Union] dist right",
 CellID->1624124987],
Cell[460928, 13407, 1575, 52, 50, "FormalTextInputFormula",ExpressionUUID->"c2d8d38f-e73a-4c20-90e9-10e81ae9316a",
 CellTags->"\[Intersection]\[Equal]UNIV",
 CellID->158694025],
Cell[462506, 13461, 1369, 42, 50, "FormalTextInputFormula",ExpressionUUID->"b2e42c5d-5b00-41ba-bc88-ffaa02e6c6f4",
 CellTags->"\[SubsetEqual]\[Intersection]",
 CellID->1837534884],
Cell[463878, 13505, 1485, 45, 58, "FormalTextInputFormula",ExpressionUUID->"37035f2d-8be1-4d9b-b346-2538b15706a2",
 CellTags->"Insert\[Intersection]",
 CellID->229432372],
Cell[465366, 13552, 1488, 45, 58, "FormalTextInputFormula",ExpressionUUID->"5baff602-e322-4acf-a322-433104f4e54a",
 CellTags->"\[Intersection]Insert",
 CellID->2031161567],
Cell[466857, 13599, 120, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"22e8a15a-a1da-466c-b92f-7672a0e79078",
 CellID->139188812]
}, Open  ]],
Cell[466992, 13603, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"d5447440-d4bb-4b59-9c83-6f01b5bea9fb",
 CellID->595986403],
Cell[467098, 13606, 1821, 47, 29, "ProofInfo",ExpressionUUID->"cf2a0b05-947e-4e08-b681-dbb9bb955bbf",
 Evaluatable->True,
 CellTags->{"Proof|ID:125816224-1", "3"},
 CellID->1628613977],
Cell[468922, 13655, 1823, 47, 29, "ProofInfo",ExpressionUUID->"36472be4-f0d7-4752-bf1f-5e56ab4cfee6",
 Evaluatable->True,
 CellTags->{"Proof|ID:848809596-1", "3"},
 CellID->2030772045],
Cell[470748, 13704, 1816, 47, 29, "ProofInfo",ExpressionUUID->"e0bc9de7-8d38-449f-8c01-166784414f07",
 Evaluatable->True,
 CellTags->{"Proof|ID:147215930-1", "3"},
 CellID->548669029],
Cell[472567, 13753, 1819, 47, 29, "ProofInfo",ExpressionUUID->"0c6b49b0-74e6-4b95-a486-249347bf00e4",
 Evaluatable->True,
 CellTags->{"Proof|ID:902394792-1", "3"},
 CellID->1223279036],
Cell[474389, 13802, 1818, 47, 29, "ProofInfo",ExpressionUUID->"c9ae36de-51e6-4ab1-af37-a9c0f0d09f5f",
 Evaluatable->True,
 CellTags->{"Proof|ID:568250332-1", "3"},
 CellID->868040367],
Cell[476210, 13851, 1825, 47, 29, "ProofInfo",ExpressionUUID->"263a06fe-fc74-4a56-8db9-29590e7b58ff",
 Evaluatable->True,
 CellTags->{"Proof|ID:1826338130-1", "3"},
 CellID->679262902],
Cell[478038, 13900, 1826, 47, 29, "ProofInfo",ExpressionUUID->"e1bfc212-a53d-43e0-b061-4f254f379381",
 Evaluatable->True,
 CellTags->{"Proof|ID:1588902597-1", "3"},
 CellID->1210453596],
Cell[479867, 13949, 1814, 47, 29, "ProofInfo",ExpressionUUID->"1a5c661b-1557-4af5-b166-96572bd68d8c",
 Evaluatable->True,
 CellTags->{"Proof|ID:1253019475-1", "3"},
 CellID->1968362112],
Cell[481684, 13998, 1814, 47, 29, "ProofInfo",ExpressionUUID->"3cc072c7-64ff-40a0-8749-2ddcf91e538c",
 Evaluatable->True,
 CellTags->{"Proof|ID:1274251332-1", "3"},
 CellID->2131093972],
Cell[483501, 14047, 1831, 47, 29, "ProofInfo",ExpressionUUID->"8a174da6-9259-4456-a888-bca8c161fadd",
 Evaluatable->True,
 CellTags->{"Proof|ID:903302990-1", "3"},
 CellID->72270293],
Cell[485335, 14096, 1824, 47, 29, "ProofInfo",ExpressionUUID->"677ec8e8-79c5-4849-ab27-fd71ffc1c532",
 Evaluatable->True,
 CellTags->{"Proof|ID:920640376-1", "3"},
 CellID->321131777],
Cell[487162, 14145, 1818, 47, 29, "ProofInfo",ExpressionUUID->"777b4649-e951-4bda-9678-a8b639463f5b",
 Evaluatable->True,
 CellTags->{"Proof|ID:1023611974-1", "3"},
 CellID->1877569687],
Cell[488983, 14194, 1818, 47, 29, "ProofInfo",ExpressionUUID->"8338a43c-ea58-4e7d-85ec-4622d938afa3",
 Evaluatable->True,
 CellTags->{"Proof|ID:1195769099-1", "3"},
 CellID->1618369204],
Cell[490804, 14243, 1836, 47, 29, "ProofInfo",ExpressionUUID->"f1380f81-fa44-4763-b4e5-077e4ac37ebd",
 Evaluatable->True,
 CellTags->{"Proof|ID:56424478-1", "3"},
 CellID->400332023],
Cell[492643, 14292, 1848, 47, 29, "ProofInfo",ExpressionUUID->"f946aa9b-4bc5-4b6b-a6d6-94285e3a76b4",
 Evaluatable->True,
 CellTags->{"Proof|ID:1624124987-1", "3"},
 CellID->1734538043],
Cell[494494, 14341, 1831, 47, 29, "ProofInfo",ExpressionUUID->"edb892ab-bd7f-474d-8831-ba95ca7c0306",
 Evaluatable->True,
 CellTags->{"Proof|ID:158694025-1", "3"},
 CellID->1315875576],
Cell[496328, 14390, 1837, 47, 29, "ProofInfo",ExpressionUUID->"49d48a11-4604-4d84-b391-5e5974a2445a",
 Evaluatable->True,
 CellTags->{"Proof|ID:1837534884-1", "3"},
 CellID->978925165],
Cell[498168, 14439, 1818, 47, 29, "ProofInfo",ExpressionUUID->"4ac23cf1-4b24-4d31-90fa-4a0775d734ae",
 Evaluatable->True,
 CellTags->{"Proof|ID:229432372-1", "3"},
 CellID->387372574],
Cell[499989, 14488, 1821, 47, 29, "ProofInfo",ExpressionUUID->"812c3dd1-7e84-453c-b984-6de8b7f632e4",
 Evaluatable->True,
 CellTags->{"Proof|ID:2031161567-1", "3"},
 CellID->895002178],
Cell[501813, 14537, 731, 12, 44, "TheoremaCode",ExpressionUUID->"5de92f59-1f70-45b1-9192-629fc3c2efab",
 CellID->701072721],
Cell[502547, 14551, 103, 1, 12, "OpenEnvironment",ExpressionUUID->"c18afc6d-00fa-4f34-983d-08f8d5fb3b70",
 CellID->1941586236],
Cell[CellGroupData[{
Cell[502675, 14556, 449, 9, 38, "EnvironmentHeader",ExpressionUUID->"23182386-327e-4081-ae82-c7469e1c2fe7",
 CellID->1625693420],
Cell[503127, 14567, 924, 27, 47, "FormalTextInputFormula",ExpressionUUID->"1dd930a7-a4a4-4184-a990-a1221b81347e",
 CellTags->"\[Union] absorb",
 CellID->510851566],
Cell[504054, 14596, 1040, 31, 50, "FormalTextInputFormula",ExpressionUUID->"30dd6ca7-e166-4968-b230-3695f88edcbe",
 CellTags->"\[Union] absorb'",
 CellID->961990306],
Cell[505097, 14629, 977, 29, 50, "FormalTextInputFormula",ExpressionUUID->"fffc7ebe-1903-470d-a423-01e5da9bdb37",
 CellTags->"\[Union] comm",
 CellID->2140682664],
Cell[506077, 14660, 1109, 33, 50, "FormalTextInputFormula",ExpressionUUID->"63cba1dd-7f2a-42ee-bd5b-dd610f3b54b3",
 CellTags->"\[Union] comm'",
 CellID->1655518259],
Cell[507189, 14695, 1106, 33, 50, "FormalTextInputFormula",ExpressionUUID->"e8e6d705-631c-4c9d-aa97-fefca4115ead",
 CellTags->"\[Union] assoc",
 CellID->945600630],
Cell[508298, 14730, 1130, 35, 50, "FormalTextInputFormula",ExpressionUUID->"ad5beaab-93b9-48c6-b2c2-a8f8c1ff258d",
 CellTags->"\[Union] absorb1",
 CellID->837689454],
Cell[509431, 14767, 1132, 35, 50, "FormalTextInputFormula",ExpressionUUID->"38deb9e2-550b-48aa-8740-428d4796982e",
 CellTags->"\[Union] absorb2",
 CellID->1727788553],
Cell[510566, 14804, 935, 28, 47, "FormalTextInputFormula",ExpressionUUID->"d6414077-67ef-4ac9-8836-f56a0138097e",
 CellTags->"\[Union]{}",
 CellID->985635433],
Cell[511504, 14834, 936, 28, 47, "FormalTextInputFormula",ExpressionUUID->"f0351117-b62d-4593-804a-5846ee59e902",
 CellTags->"{}\[Union]",
 CellID->1158821232],
Cell[512443, 14864, 925, 27, 47, "FormalTextInputFormula",ExpressionUUID->"cc76f1fe-6b2e-417e-b83a-4a1e85689882",
 CellTags->"\[Union]UNIV",
 CellID->1958752872],
Cell[513371, 14893, 924, 27, 47, "FormalTextInputFormula",ExpressionUUID->"2c1fe514-2923-4862-ae06-ebdba0500fae",
 CellTags->"UNIV\[Union]",
 CellID->715721472],
Cell[514298, 14922, 1112, 33, 50, "FormalTextInputFormula",ExpressionUUID->"88ad4619-c3ec-44e6-9cc9-c15783ae50f8",
 CellTags->"Insert\[Union]",
 CellID->727949230],
Cell[515413, 14957, 1114, 33, 50, "FormalTextInputFormula",ExpressionUUID->"1108794c-416f-48fd-94ec-fb8fcba1ca2e",
 CellTags->"\[Union]Insert",
 CellID->1429638428],
Cell[516530, 14992, 1221, 35, 50, "FormalTextInputFormula",ExpressionUUID->"08d29689-ef28-4a06-890c-f9306ee30fd1",
 CellTags->"\[Union]\[Intersection] dist left",
 CellID->1116053495],
Cell[517754, 15029, 1225, 36, 50, "FormalTextInputFormula",ExpressionUUID->"b2ce6988-93f6-4266-a5cd-26dc6dbf5711",
 CellTags->"\[Union]\[Intersection] dist right",
 CellID->1097000890],
Cell[518982, 15067, 1612, 55, 50, "FormalTextInputFormula",ExpressionUUID->"36426bb0-b926-47c4-aeb4-4508b486267d",
 CellTags->"\[Union]\[Equal]{}",
 CellID->1016692823],
Cell[520597, 15124, 1345, 42, 50, "FormalTextInputFormula",ExpressionUUID->"c2879c31-4a67-4636-b191-393fcad6e26d",
 CellTags->"\[Union]\[SubsetEqual]",
 CellID->997515848],
Cell[521945, 15168, 121, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"828c525d-3bea-47cc-aeab-fe08a6d1713b",
 CellID->1899020357]
}, Open  ]],
Cell[522081, 15172, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"d3d9e47a-fe79-4339-be10-f2548eae4b65",
 CellID->172320552],
Cell[522187, 15175, 1791, 46, 29, "ProofInfo",ExpressionUUID->"eb249535-8173-40d6-bbdb-2c2aae352b07",
 Evaluatable->True,
 CellTags->{"Proof|ID:510851566-1", "3"},
 CellID->365686177],
Cell[523981, 15223, 1809, 47, 29, "ProofInfo",ExpressionUUID->"f296832d-3220-4ad0-9d80-345f79408b1f",
 Evaluatable->True,
 CellTags->{"Proof|ID:961990306-1", "3"},
 CellID->1321288291],
Cell[525793, 15272, 1790, 46, 29, "ProofInfo",ExpressionUUID->"e2b5c4d8-ca86-4739-8add-11abf118bb58",
 Evaluatable->True,
 CellTags->{"Proof|ID:2140682664-1", "3"},
 CellID->678754593],
Cell[527586, 15320, 1793, 46, 29, "ProofInfo",ExpressionUUID->"5592b9fd-2693-411b-becf-a4a020ad57be",
 Evaluatable->True,
 CellTags->{"Proof|ID:1655518259-1", "3"},
 CellID->1904743954],
Cell[529382, 15368, 1790, 46, 29, "ProofInfo",ExpressionUUID->"3fc23a3b-63ed-4e0f-9760-2469f4d0179b",
 Evaluatable->True,
 CellTags->{"Proof|ID:945600630-1", "3"},
 CellID->1060575105],
Cell[531175, 15416, 1808, 47, 29, "ProofInfo",ExpressionUUID->"a45e0c3a-1b21-41cc-887a-dad678f3e9fe",
 Evaluatable->True,
 CellTags->{"Proof|ID:837689454-1", "3"},
 CellID->363045913],
Cell[532986, 15465, 1810, 47, 29, "ProofInfo",ExpressionUUID->"ccf3a6e2-23b5-4cd1-a76c-3ffdf84cc204",
 Evaluatable->True,
 CellTags->{"Proof|ID:1727788553-1", "3"},
 CellID->56107910],
Cell[534799, 15514, 1782, 46, 29, "ProofInfo",ExpressionUUID->"e1d8893c-5ca6-4930-940f-940100ee72b4",
 Evaluatable->True,
 CellTags->{"Proof|ID:985635433-1", "3"},
 CellID->1189519127],
Cell[536584, 15562, 1784, 46, 29, "ProofInfo",ExpressionUUID->"84574fea-7e8d-4451-afb5-958ee98ab045",
 Evaluatable->True,
 CellTags->{"Proof|ID:1158821232-1", "3"},
 CellID->772575899],
Cell[538371, 15610, 1789, 46, 29, "ProofInfo",ExpressionUUID->"57de62df-9c50-4468-8a9b-27db3a02f067",
 Evaluatable->True,
 CellTags->{"Proof|ID:1958752872-1", "3"},
 CellID->1073197046],
Cell[540163, 15658, 1785, 46, 29, "ProofInfo",ExpressionUUID->"091f7b25-0bd4-44cb-ae0f-90617fc64236",
 Evaluatable->True,
 CellTags->{"Proof|ID:715721472-1", "3"},
 CellID->645311516],
Cell[541951, 15706, 1790, 46, 29, "ProofInfo",ExpressionUUID->"f96d0823-3864-4900-af03-5095e0ae6fd2",
 Evaluatable->True,
 CellTags->{"Proof|ID:727949230-1", "3"},
 CellID->1612445395],
Cell[543744, 15754, 1792, 46, 29, "ProofInfo",ExpressionUUID->"e1b4be1f-45b6-489f-ae2b-a23d03f34073",
 Evaluatable->True,
 CellTags->{"Proof|ID:1429638428-1", "3"},
 CellID->788463628],
Cell[545539, 15802, 1845, 47, 29, "ProofInfo",ExpressionUUID->"10ae46d4-043b-4352-a8ed-04644b597001",
 Evaluatable->True,
 CellTags->{"Proof|ID:1116053495-1", "3"},
 CellID->826786679],
Cell[547387, 15851, 1848, 47, 29, "ProofInfo",ExpressionUUID->"5e24b08f-bb6e-4cbf-80ba-ba0b65e7b37f",
 Evaluatable->True,
 CellTags->{"Proof|ID:1097000890-1", "3"},
 CellID->1930524663],
Cell[549238, 15900, 1816, 47, 29, "ProofInfo",ExpressionUUID->"4e864b7c-43bd-41bd-b35b-e80644e4ba7c",
 Evaluatable->True,
 CellTags->{"Proof|ID:1016692823-1", "3"},
 CellID->1014581777],
Cell[551057, 15949, 1820, 47, 29, "ProofInfo",ExpressionUUID->"5bcae50a-1782-48b2-bce6-0bda8d2274c8",
 Evaluatable->True,
 CellTags->{"Proof|ID:997515848-1", "3"},
 CellID->273440836],
Cell[552880, 15998, 745, 13, 44, "TheoremaCode",ExpressionUUID->"c4aaa5c8-a5cb-4e9a-b9c5-af3f0a10c650",
 CellID->631533417],
Cell[553628, 16013, 103, 1, 12, "OpenEnvironment",ExpressionUUID->"33c844af-b832-438b-bdfb-0ed6b8d72678",
 CellID->1347919889],
Cell[CellGroupData[{
Cell[553756, 16018, 454, 9, 38, "EnvironmentHeader",ExpressionUUID->"4ba8a9ba-0141-45de-abc8-2e3ff7352704",
 CellID->2078327963],
Cell[554213, 16029, 979, 29, 47, "FormalTextInputFormula",ExpressionUUID->"14526c5d-7e1a-491c-a18d-b2adee8c4fd1",
 CellTags->"disjoint complement",
 CellID->36426955],
Cell[555195, 16060, 979, 29, 47, "FormalTextInputFormula",ExpressionUUID->"0f6a4806-b037-4cb5-86be-f9e02d910c80",
 CellTags->"complement disjoint",
 CellID->1941170909],
Cell[556177, 16091, 956, 28, 47, "FormalTextInputFormula",ExpressionUUID->"cba2301e-f964-4dfa-9ffc-d9fc085df72c",
 CellTags->"partition complement",
 CellID->980585996],
Cell[557136, 16121, 957, 28, 47, "FormalTextInputFormula",ExpressionUUID->"26e2dded-cd7a-432f-9c23-b65e3a2865a1",
 CellTags->"complement partition",
 CellID->1324045296],
Cell[558096, 16151, 1029, 31, 50, "FormalTextInputFormula",ExpressionUUID->"d99264a8-25ec-40e3-a87a-33a8d7399ef1",
 CellTags->"complement complement",
 CellID->1636712813],
Cell[559128, 16184, 1120, 33, 50, "FormalTextInputFormula",ExpressionUUID->"b414a419-93c9-4969-bb70-aa9e72fb3466",
 CellTags->"complement \[Union]",
 CellID->1334137643],
Cell[560251, 16219, 1133, 33, 50, "FormalTextInputFormula",ExpressionUUID->"b67ff51c-86bc-4021-948a-f367e9f183e7",
 CellTags->"complement \[Intersection]",
 CellID->1036693907],
Cell[561387, 16254, 1097, 34, 50, "FormalTextInputFormula",ExpressionUUID->"2ca9d86e-03aa-4873-a4f5-ea04b00662b0",
 CellTags->"complement Insert",
 CellID->204929684],
Cell[562487, 16290, 1243, 40, 47, "FormalTextInputFormula",ExpressionUUID->"a2f250c1-a6f8-48b1-917d-4f4472a06ad4",
 CellTags->"\[SubsetEqual]- self",
 CellID->849790927],
Cell[563733, 16332, 791, 21, 38, "FormalTextInputFormula",ExpressionUUID->"521382bc-26ea-430e-92b8-33377f2e1ed9",
 CellTags->"complement UNIV",
 CellID->1609622809],
Cell[564527, 16355, 786, 21, 38, "FormalTextInputFormula",ExpressionUUID->"d6188e01-f675-4a8a-b808-d55954fe886e",
 CellTags->"complement {}",
 CellID->1579837433],
Cell[565316, 16378, 1304, 41, 50, "FormalTextInputFormula",ExpressionUUID->"d6e1db89-6011-4f20-b209-e333d98038c8",
 CellTags->"complement \[SubsetEqual] complement",
 CellID->2039310054],
Cell[566623, 16421, 1389, 44, 50, "FormalTextInputFormula",ExpressionUUID->"ba85e675-30b3-4fb0-b8c4-fd034abf0994",
 CellTags->"complement \[Equal] complement",
 CellID->1867561751],
Cell[568015, 16467, 121, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"74c9067f-78fa-4dbb-9151-be9dc9d5ecee",
 CellID->2026674483]
}, Open  ]],
Cell[568151, 16471, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"95396ec1-e8c6-46ac-924c-a6db80e2509f",
 CellID->263470190],
Cell[568257, 16474, 1809, 47, 29, "ProofInfo",ExpressionUUID->"335661a6-db39-426e-a2d5-3a88bbf119ee",
 Evaluatable->True,
 CellTags->{"Proof|ID:36426955-1", "3"},
 CellID->1087560852],
Cell[570069, 16523, 1817, 47, 29, "ProofInfo",ExpressionUUID->"790c583f-b1fd-4cae-b4c7-3b7f8318ee81",
 Evaluatable->True,
 CellTags->{"Proof|ID:1941170909-1", "3"},
 CellID->769645528],
Cell[571889, 16572, 1817, 47, 29, "ProofInfo",ExpressionUUID->"978d16d3-2a8d-48c1-8119-e556b594fdbe",
 Evaluatable->True,
 CellTags->{"Proof|ID:980585996-1", "3"},
 CellID->1868666534],
Cell[573709, 16621, 1820, 47, 29, "ProofInfo",ExpressionUUID->"55ba8ff4-e771-4565-ad2a-a535645579da",
 Evaluatable->True,
 CellTags->{"Proof|ID:1324045296-1", "3"},
 CellID->1667315755],
Cell[575532, 16670, 1821, 47, 29, "ProofInfo",ExpressionUUID->"4270ef55-e968-4e30-88f5-189f15367a47",
 Evaluatable->True,
 CellTags->{"Proof|ID:1636712813-1", "3"},
 CellID->193380618],
Cell[577356, 16719, 1816, 47, 29, "ProofInfo",ExpressionUUID->"ff8108f1-b1f1-4f1c-be7e-36f13ea484f5",
 Evaluatable->True,
 CellTags->{"Proof|ID:1334137643-1", "3"},
 CellID->95672954],
Cell[579175, 16768, 1831, 47, 29, "ProofInfo",ExpressionUUID->"6c8fd24a-5787-423d-a5a0-9ee8249e79b1",
 Evaluatable->True,
 CellTags->{"Proof|ID:1036693907-1", "3"},
 CellID->197559568],
Cell[581009, 16817, 1810, 47, 29, "ProofInfo",ExpressionUUID->"c77a6221-7b61-4579-83d1-c9ad096f840c",
 Evaluatable->True,
 CellTags->{"Proof|ID:204929684-1", "3"},
 CellID->459265611],
Cell[582822, 16866, 1817, 47, 29, "ProofInfo",ExpressionUUID->"fcbb79cd-640b-433d-aca6-6f01bc0cd323",
 Evaluatable->True,
 CellTags->{"Proof|ID:849790927-1", "3"},
 CellID->1062590287],
Cell[584642, 16915, 1795, 46, 29, "ProofInfo",ExpressionUUID->"03e8795e-cd22-4b6b-b60f-6a5c21b1baf5",
 Evaluatable->True,
 CellTags->{"Proof|ID:1609622809-1", "3"},
 CellID->1841417850],
Cell[586440, 16963, 1791, 46, 29, "ProofInfo",ExpressionUUID->"fc5be3c0-cfbd-484d-840c-b987b55aab11",
 Evaluatable->True,
 CellTags->{"Proof|ID:1579837433-1", "3"},
 CellID->1890841345],
Cell[588234, 17011, 1852, 47, 29, "ProofInfo",ExpressionUUID->"070c9e7a-2ffe-4eed-a409-8fe374d0f970",
 Evaluatable->True,
 CellTags->{"Proof|ID:2039310054-1", "3"},
 CellID->1409452885],
Cell[590089, 17060, 1840, 47, 29, "ProofInfo",ExpressionUUID->"3255cca4-1d25-482e-bef3-695eb3e5fa77",
 Evaluatable->True,
 CellTags->{"Proof|ID:1867561751-1", "3"},
 CellID->1823633609],
Cell[591932, 17109, 1030, 18, 67, "TheoremaCode",ExpressionUUID->"c86e651e-48af-4cbc-9361-9d33d054749b",
 CellID->1273666900],
Cell[592965, 17129, 103, 1, 12, "OpenEnvironment",ExpressionUUID->"d002514c-7838-4dd5-8a4f-d36656ebd8b4",
 CellID->1686577370],
Cell[CellGroupData[{
Cell[593093, 17134, 440, 9, 38, "EnvironmentHeader",ExpressionUUID->"3e915c3f-6136-4bc8-b8eb-74392ebc93b9",
 CellID->85132320],
Cell[593536, 17145, 1004, 30, 50, "FormalTextInputFormula",ExpressionUUID->"6623863d-220c-43b7-b56a-d97b10f1a117",
 CellTags->"- alt SET 2",
 CellID->889927681],
Cell[594543, 17177, 1261, 41, 50, "FormalTextInputFormula",ExpressionUUID->"60d9d2ad-a5de-4e53-8980-466c84af87c2",
 CellTags->"-\[Equal]{} iff",
 CellID->1181933579],
Cell[595807, 17220, 928, 28, 47, "FormalTextInputFormula",ExpressionUUID->"6781d634-3dc0-42fb-8e42-aff33a4f6cec",
 CellTags->"- self SET",
 CellID->1313006718],
Cell[596738, 17250, 1115, 34, 50, "FormalTextInputFormula",ExpressionUUID->"e34deac0-2ee0-4195-b922-322864967d47",
 CellTags->"- idem SET",
 CellID->405798501],
Cell[597856, 17286, 1166, 37, 50, "FormalTextInputFormula",ExpressionUUID->"ed752dac-cb8f-4130-b775-aa2d387899a4",
 CellTags->"- triv SET",
 CellID->142114837],
Cell[599025, 17325, 902, 27, 47, "FormalTextInputFormula",ExpressionUUID->"f814e4e0-d5fd-4f2d-8049-0a0cc5ce82a9",
 CellTags->"-{}",
 CellID->655509949],
Cell[599930, 17354, 924, 28, 47, "FormalTextInputFormula",ExpressionUUID->"b1c87120-9de9-411e-b7fb-51ab4e109a6f",
 CellTags->"{}-",
 CellID->149156198],
Cell[600857, 17384, 919, 28, 47, "FormalTextInputFormula",ExpressionUUID->"c3b9274b-a806-42e8-a112-b152145efc29",
 CellTags->"-UNIV",
 CellID->1551567722],
Cell[601779, 17414, 1407, 44, 57, "FormalTextInputFormula",ExpressionUUID->"42e35235-185b-4a52-99ed-8e7cf0803c40",
 CellTags->"Insert-",
 CellID->1618171348],
Cell[603189, 17460, 1091, 33, 50, "FormalTextInputFormula",ExpressionUUID->"a4ab1dc9-a2bf-463a-8074-9e4f2ca3cbf3",
 CellTags->"Insert- single",
 CellID->263157914],
Cell[604283, 17495, 1018, 31, 50, "FormalTextInputFormula",ExpressionUUID->"50650e03-a3f8-46d0-b2ff-9a8d0cd54ee2",
 CellTags->"- disjoint",
 CellID->586501678],
Cell[605304, 17528, 1087, 32, 50, "FormalTextInputFormula",ExpressionUUID->"bffe28f1-05a1-4afd-a4bd-cc179788a0cf",
 CellTags->"\[Union]- cancel",
 CellID->1211775432],
Cell[606394, 17562, 1035, 31, 50, "FormalTextInputFormula",ExpressionUUID->"4b9cee39-401e-4340-85f0-5cc10f31ca56",
 CellTags->"-\[Union] cancel",
 CellID->1437446913],
Cell[607432, 17595, 1143, 35, 50, "FormalTextInputFormula",ExpressionUUID->"1364d8ff-9078-426a-83d8-d37785793f2d",
 CellTags->"-\[Union]",
 CellID->1743954910],
Cell[608578, 17632, 1158, 35, 50, "FormalTextInputFormula",ExpressionUUID->"14da9918-2877-431c-9676-774a594407e6",
 CellTags->"-\[Intersection]",
 CellID->732837784],
Cell[609739, 17669, 1057, 31, 50, "FormalTextInputFormula",ExpressionUUID->"3ac7b94b-778f-462d-a022-df58ba35c477",
 CellTags->"-- \[Equal] \[Intersection]",
 CellID->520649721],
Cell[610799, 17702, 1138, 35, 50, "FormalTextInputFormula",ExpressionUUID->"19d77f44-c8da-4204-b8b0-44e14aee704c",
 CellTags->"\[Union]-",
 CellID->1959934043],
Cell[611940, 17739, 1111, 33, 50, "FormalTextInputFormula",ExpressionUUID->"a1bb30b1-8de9-47b4-8158-a68f8ab50910",
 CellTags->"\[Intersection]-",
 CellID->1023545665],
Cell[613054, 17774, 1200, 35, 50, "FormalTextInputFormula",ExpressionUUID->"83e3520e-f1ff-4eac-a5ba-79ed51ea4e65",
 CellTags->"\[Intersection]- dist left",
 CellID->1036748287],
Cell[614257, 17811, 1199, 35, 50, "FormalTextInputFormula",ExpressionUUID->"9c74b7bb-8043-4f3f-adef-5cedca906c9b",
 CellTags->"\[Intersection]- dist right",
 CellID->654257124],
Cell[615459, 17848, 1020, 31, 50, "FormalTextInputFormula",ExpressionUUID->"ac1ce070-9cd7-4b47-9574-487965c33172",
 CellTags->"- complement",
 CellID->788943865],
Cell[616482, 17881, 1034, 32, 50, "FormalTextInputFormula",ExpressionUUID->"f380d7e6-5d16-4293-becd-33ad9bfa3087",
 CellTags->"complement -",
 CellID->811668796],
Cell[617519, 17915, 121, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"51f15a8e-2618-4e52-a726-990229061e99",
 CellID->1477814126]
}, Open  ]],
Cell[617655, 17919, 104, 1, 16, "CloseEnvironment",ExpressionUUID->"693560ca-d387-49be-b62c-5b7317c69a34",
 CellID->1940925515],
Cell[617762, 17922, 1783, 46, 29, "ProofInfo",ExpressionUUID->"41412e05-3d1b-47af-9a2c-69b7da569934",
 Evaluatable->True,
 CellTags->{"Proof|ID:889927681-1", "3"},
 CellID->989686867],
Cell[619548, 17970, 1793, 46, 29, "ProofInfo",ExpressionUUID->"e718483a-9533-47fb-b2b3-21662a495fe5",
 Evaluatable->True,
 CellTags->{"Proof|ID:1181933579-1", "3"},
 CellID->14631120],
Cell[621344, 18018, 1785, 46, 29, "ProofInfo",ExpressionUUID->"c083b88a-b5b4-4213-8c12-363f6380f03a",
 Evaluatable->True,
 CellTags->{"Proof|ID:1313006718-1", "3"},
 CellID->1424176457],
Cell[623132, 18066, 1782, 46, 29, "ProofInfo",ExpressionUUID->"8dfe2f53-ea61-4e3d-9764-e5a54581d749",
 Evaluatable->True,
 CellTags->{"Proof|ID:405798501-1", "3"},
 CellID->1289581872],
Cell[624917, 18114, 1782, 46, 29, "ProofInfo",ExpressionUUID->"15000b41-243b-47fb-9432-88147364a061",
 Evaluatable->True,
 CellTags->{"Proof|ID:142114837-1", "3"},
 CellID->1757475874],
Cell[626702, 18162, 1752, 45, 29, "ProofInfo",ExpressionUUID->"ba81f3be-43a1-4405-a13e-20392ccc153a",
 Evaluatable->True,
 CellTags->{"Proof|ID:655509949-1", "3"},
 CellID->939658652],
Cell[628457, 18209, 1752, 45, 29, "ProofInfo",ExpressionUUID->"a5e8bedb-4a86-49a2-916f-d64eed66bbf2",
 Evaluatable->True,
 CellTags->{"Proof|ID:149156198-1", "3"},
 CellID->125409990],
Cell[630212, 18256, 1759, 45, 29, "ProofInfo",ExpressionUUID->"81eeefa4-874f-43cf-8a1a-bf67eab02b68",
 Evaluatable->True,
 CellTags->{"Proof|ID:1551567722-1", "3"},
 CellID->128243725],
Cell[631974, 18303, 1764, 45, 29, "ProofInfo",ExpressionUUID->"27fb32d0-54c1-4430-9891-7032e6c9d015",
 Evaluatable->True,
 CellTags->{"Proof|ID:1618171348-1", "3"},
 CellID->1212022066],
Cell[633741, 18350, 1790, 46, 29, "ProofInfo",ExpressionUUID->"0d2dd380-ba5d-4928-b373-142c2857c2c0",
 Evaluatable->True,
 CellTags->{"Proof|ID:263157914-1", "3"},
 CellID->1307442480],
Cell[635534, 18398, 1782, 46, 29, "ProofInfo",ExpressionUUID->"1310c805-4a87-4c5e-91d3-1155f73282e3",
 Evaluatable->True,
 CellTags->{"Proof|ID:586501678-1", "3"},
 CellID->1257980868],
Cell[637319, 18446, 1812, 47, 29, "ProofInfo",ExpressionUUID->"f7d7c577-0e4f-43f2-aab9-f884ae948b4c",
 Evaluatable->True,
 CellTags->{"Proof|ID:1211775432-1", "3"},
 CellID->1583175827],
Cell[639134, 18495, 1812, 47, 29, "ProofInfo",ExpressionUUID->"06df1467-a1fc-4be5-b103-26e4a9577222",
 Evaluatable->True,
 CellTags->{"Proof|ID:1437446913-1", "3"},
 CellID->2050049889],
Cell[640949, 18544, 1768, 45, 29, "ProofInfo",ExpressionUUID->"9ecc279b-a1b3-4c56-a8fa-e29147a8ea06",
 Evaluatable->True,
 CellTags->{"Proof|ID:1743954910-1", "3"},
 CellID->1362760740],
Cell[642720, 18591, 1809, 47, 29, "ProofInfo",ExpressionUUID->"45cfb94c-9228-4f72-a7c6-3adc38573e3f",
 Evaluatable->True,
 CellTags->{"Proof|ID:732837784-1", "3"},
 CellID->1924200546],
Cell[644532, 18640, 1831, 47, 29, "ProofInfo",ExpressionUUID->"d7f0bdda-8a9a-44b8-8d62-f6a17e32fa14",
 Evaluatable->True,
 CellTags->{"Proof|ID:520649721-1", "3"},
 CellID->1370228285],
Cell[646366, 18689, 1768, 45, 29, "ProofInfo",ExpressionUUID->"dbca4d7d-fc2b-4f4e-9403-9ed1fcdfcd97",
 Evaluatable->True,
 CellTags->{"Proof|ID:1959934043-1", "3"},
 CellID->1728602517],
Cell[648137, 18736, 1811, 47, 29, "ProofInfo",ExpressionUUID->"330a0d43-a3eb-4628-9ac8-18c5d5c04d74",
 Evaluatable->True,
 CellTags->{"Proof|ID:1023545665-1", "3"},
 CellID->124203596],
Cell[649951, 18785, 1832, 47, 29, "ProofInfo",ExpressionUUID->"e55012c4-27e1-41ae-920d-8392fb2fc039",
 Evaluatable->True,
 CellTags->{"Proof|ID:1036748287-1", "3"},
 CellID->1046065391],
Cell[651786, 18834, 1830, 47, 29, "ProofInfo",ExpressionUUID->"75c51d93-2b93-476d-8df7-56ee4ba944f9",
 Evaluatable->True,
 CellTags->{"Proof|ID:654257124-1", "3"},
 CellID->190015761],
Cell[653619, 18883, 1786, 46, 29, "ProofInfo",ExpressionUUID->"bfb4263d-1542-4461-b96b-3cb80e17ef47",
 Evaluatable->True,
 CellTags->{"Proof|ID:788943865-1", "3"},
 CellID->1182421826],
Cell[655408, 18931, 1786, 46, 29, "ProofInfo",ExpressionUUID->"7126e94e-5680-4178-a002-f91132148a3e",
 Evaluatable->True,
 CellTags->{"Proof|ID:811668796-1", "3"},
 CellID->1010368971],
Cell[657197, 18979, 1200, 20, 67, "TheoremaCode",ExpressionUUID->"7e78a0f2-fb38-42a7-95d1-866199b8bd64",
 CellID->673690670],
Cell[658400, 19001, 103, 1, 12, "OpenEnvironment",ExpressionUUID->"ac145a3d-05f8-4b42-b9f9-1c1e2108743f",
 CellID->1252732680],
Cell[CellGroupData[{
Cell[658528, 19006, 445, 9, 38, "EnvironmentHeader",ExpressionUUID->"176ffabc-7757-47f4-8501-78507d8a2ca1",
 CellID->1075678771],
Cell[658976, 19017, 822, 23, 38, "FormalTextInputFormula",ExpressionUUID->"5be69f99-7725-46d4-bb6b-dbc97f3b3ab6",
 CellTags->"Pow {}",
 CellID->525687623],
Cell[659801, 19042, 774, 21, 38, "FormalTextInputFormula",ExpressionUUID->"94af11de-54d0-4a76-a0f9-9d7f0fa4a28f",
 CellTags->"Pow UNIV",
 CellID->537376125],
Cell[660578, 19065, 1106, 32, 50, "FormalTextInputFormula",ExpressionUUID->"fd9a1629-3d23-485d-a17d-020a9e92a3d4",
 CellTags->"\[Union]Pow \[SubsetEqual]",
 CellID->1322774384],
Cell[661687, 19099, 1097, 32, 50, "FormalTextInputFormula",ExpressionUUID->"334b0879-21a1-493b-a548-ff7fa6aad20a",
 CellTags->"Pow\[Intersection]",
 CellID->552587340],
Cell[662787, 19133, 121, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"d90dd32d-9143-4137-8c32-c73557871bc7",
 CellID->1029449578]
}, Open  ]],
Cell[662923, 19137, 104, 1, 16, "CloseEnvironment",ExpressionUUID->"9af747af-3746-4b40-af63-c3cbec04840d",
 CellID->2047020259],
Cell[663030, 19140, 1758, 45, 29, "ProofInfo",ExpressionUUID->"60c20774-a0c4-41e0-8583-3cd982a256e3",
 Evaluatable->True,
 CellTags->{"Proof|ID:525687623-1", "3"},
 CellID->118494995],
Cell[664791, 19187, 1762, 45, 29, "ProofInfo",ExpressionUUID->"e80da33b-c1dd-4198-bc6a-2a3ca4d75937",
 Evaluatable->True,
 CellTags->{"Proof|ID:537376125-1", "3"},
 CellID->998603711],
Cell[666556, 19234, 1832, 47, 29, "ProofInfo",ExpressionUUID->"1cdbcc0c-c7a2-4a23-ad73-491bc0d57579",
 Evaluatable->True,
 CellTags->{"Proof|ID:1322774384-1", "3"},
 CellID->1961996116],
Cell[668391, 19283, 1813, 47, 29, "ProofInfo",ExpressionUUID->"362f630c-4228-4cb1-b302-51e8b7bdf329",
 Evaluatable->True,
 CellTags->{"Proof|ID:552587340-1", "3"},
 CellID->1186950056],
Cell[670207, 19332, 567, 9, 44, "TheoremaCode",ExpressionUUID->"94ec2d25-14f2-4619-add2-d57a3be773f6",
 CellID->1895915908],
Cell[670777, 19343, 103, 1, 12, "OpenEnvironment",ExpressionUUID->"a42f4f35-009b-4e43-8cdb-d23c7c2628a3",
 CellID->1424840277],
Cell[CellGroupData[{
Cell[670905, 19348, 444, 9, 38, "EnvironmentHeader",ExpressionUUID->"b9fec212-daaa-46f8-a94b-39fa9c10f439",
 CellID->579114060],
Cell[671352, 19359, 1081, 33, 51, "FormalTextInputFormula",ExpressionUUID->"27842332-1915-4319-aa16-d1927fa59d6a",
 CellTags->"@\[Union]",
 CellID->814265295],
Cell[672436, 19394, 1047, 32, 51, "FormalTextInputFormula",ExpressionUUID->"ebf795ec-f3cd-4609-a17d-968049102958",
 CellTags->"@ ident",
 CellID->1283984586],
Cell[673486, 19428, 925, 28, 48, "FormalTextInputFormula",ExpressionUUID->"c6a43110-bc9f-4894-aa96-288c7209b020",
 CellTags->"@{}",
 CellID->1257505636],
Cell[674414, 19458, 1116, 34, 51, "FormalTextInputFormula",ExpressionUUID->"d15f846d-e95a-4275-90e4-d985073cea8e",
 CellTags->"@Insert",
 CellID->565842574],
Cell[675533, 19494, 1278, 42, 51, "FormalTextInputFormula",ExpressionUUID->"4c3aa816-2f51-4a8c-8b3d-a002c2ecd252",
 CellTags->"@\[Equal]{} iff",
 CellID->685115154],
Cell[676814, 19538, 1456, 47, 58, "FormalTextInputFormula",ExpressionUUID->"66f7bb38-30cc-49de-bc0e-de89ae07f96c",
 CellTags->"\[ForAll]@",
 CellID->1570027835],
Cell[678273, 19587, 1453, 47, 58, "FormalTextInputFormula",ExpressionUUID->"8b19b12b-215f-4771-b7ad-e947689e6e14",
 CellTags->"\[Exists]@",
 CellID->771344911],
Cell[679729, 19636, 1134, 34, 62, "FormalTextInputFormula",ExpressionUUID->"79ebb868-37c7-4860-8e0b-65f7121ee750",
 CellTags->"SetOf\[Equal]@Collect",
 CellID->1251957142],
Cell[680866, 19672, 1085, 33, 62, "FormalTextInputFormula",ExpressionUUID->"04de836f-2f8e-4c32-b3ee-1f22e4061854",
 CellTags->"SetOf\[Equal]@",
 CellID->2088273463],
Cell[681954, 19707, 121, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"283ba5a0-14a8-496b-82f9-5a0772fad864",
 CellID->1593529551]
}, Open  ]],
Cell[682090, 19711, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"ce63b603-f078-4331-aa75-0494586678ae",
 CellID->182572638],
Cell[682196, 19714, 1765, 45, 29, "ProofInfo",ExpressionUUID->"4d04f4ae-48b6-4184-a3cd-f21b71cb61b2",
 Evaluatable->True,
 CellTags->{"Proof|ID:814265295-1", "3"},
 CellID->1805064061],
Cell[683964, 19761, 1763, 45, 29, "ProofInfo",ExpressionUUID->"268be10f-dd81-47de-b132-ddea9a7f3a82",
 Evaluatable->True,
 CellTags->{"Proof|ID:1283984586-1", "3"},
 CellID->203354138],
Cell[685730, 19808, 1756, 45, 29, "ProofInfo",ExpressionUUID->"c60cafa1-dc19-4490-a5d0-a50cbd0e6903",
 Evaluatable->True,
 CellTags->{"Proof|ID:1257505636-1", "3"},
 CellID->1997254040],
Cell[687489, 19855, 1760, 45, 29, "ProofInfo",ExpressionUUID->"243af554-932a-474a-a189-73c259821534",
 Evaluatable->True,
 CellTags->{"Proof|ID:565842574-1", "3"},
 CellID->477474021],
Cell[689252, 19902, 1792, 46, 29, "ProofInfo",ExpressionUUID->"231ca40f-e980-4b69-b7ca-c100eb04e88e",
 Evaluatable->True,
 CellTags->{"Proof|ID:685115154-1", "3"},
 CellID->1626158995],
Cell[691047, 19950, 1785, 46, 29, "ProofInfo",ExpressionUUID->"e7c56a2e-bc98-4cf0-80e9-3bac79998304",
 Evaluatable->True,
 CellTags->{"Proof|ID:1570027835-1", "3"},
 CellID->1776911464],
Cell[692835, 19998, 1781, 46, 29, "ProofInfo",ExpressionUUID->"860e2886-3792-4877-8a4c-304213f7d561",
 Evaluatable->True,
 CellTags->{"Proof|ID:771344911-1", "3"},
 CellID->141629000],
Cell[694619, 20046, 1821, 47, 29, "ProofInfo",ExpressionUUID->"cecad2e4-4de8-4466-95cb-7b7894b699c6",
 Evaluatable->True,
 CellTags->{"Proof|ID:1251957142-1", "3"},
 CellID->150377480],
Cell[696443, 20095, 1793, 46, 29, "ProofInfo",ExpressionUUID->"b6d3e804-b5ea-4af4-92d5-df3642a09ce3",
 Evaluatable->True,
 CellTags->{"Proof|ID:2088273463-1", "3"},
 CellID->1818369679],
Cell[698239, 20143, 389, 7, 44, "TheoremaCode",ExpressionUUID->"5a20411d-76e9-4691-abcb-bdc2ec43704d",
 CellID->1437902574]
}, Closed]],
Cell[CellGroupData[{
Cell[698665, 20155, 179, 2, 31, "Subsection",ExpressionUUID->"d997d30d-e899-432a-997a-4295f73a9abf",
 CellID->397026471],
Cell[698847, 20159, 103, 1, 12, "OpenEnvironment",ExpressionUUID->"601b9d42-41d9-4abd-82cb-f93913ac8545",
 CellID->1898688307],
Cell[CellGroupData[{
Cell[698975, 20164, 429, 9, 38, "EnvironmentHeader",ExpressionUUID->"bc300775-4c6c-4e35-88da-615cf64925c8",
 CellID->619323678],
Cell[699407, 20175, 1145, 36, 51, "FormalTextInputFormula",ExpressionUUID->"1f0cd3b1-4c97-4af8-889c-d482a16463ad",
 CellTags->"@ mono",
 CellID->570770860],
Cell[700555, 20213, 1154, 36, 50, "FormalTextInputFormula",ExpressionUUID->"dc492e7f-ab0b-45d7-8e5f-894206d553d3",
 CellTags->"Pow mono",
 CellID->1108679522],
Cell[701712, 20251, 1229, 38, 50, "FormalTextInputFormula",ExpressionUUID->"94e58813-71c0-439e-b734-f656ba867cb1",
 CellTags->"Insert mono",
 CellID->788559876],
Cell[702944, 20291, 1151, 36, 50, "FormalTextInputFormula",ExpressionUUID->"e9f9e5f9-d013-4783-8a5e-8ae725783603",
 CellTags->"complement antimono",
 CellID->604821491],
Cell[704098, 20329, 1348, 42, 50, "FormalTextInputFormula",ExpressionUUID->"f8f62021-ce16-4148-97bc-a826f1bd212d",
 CellTags->"\[Union] mono",
 CellID->1716105510],
Cell[705449, 20373, 1376, 42, 50, "FormalTextInputFormula",ExpressionUUID->"e21c8cc3-a299-4b52-a555-2f06ebaa8254",
 CellTags->"\[Intersection] mono",
 CellID->1849777423],
Cell[706828, 20417, 1318, 42, 50, "FormalTextInputFormula",ExpressionUUID->"efb2b44d-7bb0-461d-a404-c6b1978ae33a",
 CellTags->"- mono",
 CellID->1569569101],
Cell[708149, 20461, 1862, 59, 62, "FormalTextInputFormula",ExpressionUUID->"1be2d71c-9305-4560-9d81-fafc180c5b2b",
 CellTags->"SetOf mono",
 CellID->172277084],
Cell[710014, 20522, 120, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"4fa1ee10-87ae-491c-9c41-de19ffabf718",
 CellID->833259476]
}, Open  ]],
Cell[710149, 20526, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"83c10aeb-6402-4a2a-b9fe-80469e4d7fc4",
 CellID->786833653],
Cell[710255, 20529, 1758, 45, 29, "ProofInfo",ExpressionUUID->"d06f131e-6326-4e72-a0bc-30541cdf0aa3",
 Evaluatable->True,
 CellTags->{"Proof|ID:570770860-1", "3"},
 CellID->602566564],
Cell[712016, 20576, 1766, 45, 29, "ProofInfo",ExpressionUUID->"619955c1-53c3-41d6-b1c8-daf0572378fc",
 Evaluatable->True,
 CellTags->{"Proof|ID:1108679522-1", "3"},
 CellID->1093017718],
Cell[713785, 20623, 1784, 46, 29, "ProofInfo",ExpressionUUID->"6344d294-e9d4-420c-a9cb-bcd22dfbe228",
 Evaluatable->True,
 CellTags->{"Proof|ID:788559876-1", "3"},
 CellID->1966110247],
Cell[715572, 20671, 1814, 47, 29, "ProofInfo",ExpressionUUID->"5334709a-119e-480c-b8ba-516e9e54b89b",
 Evaluatable->True,
 CellTags->{"Proof|ID:604821491-1", "3"},
 CellID->429588115],
Cell[717389, 20720, 1791, 46, 29, "ProofInfo",ExpressionUUID->"4c1795f9-7bd4-48d4-ad29-b0688fdc3008",
 Evaluatable->True,
 CellTags->{"Proof|ID:1716105510-1", "3"},
 CellID->2002110999],
Cell[719183, 20768, 1820, 47, 29, "ProofInfo",ExpressionUUID->"8dd5ad08-d868-4d2e-ad57-6ee353a4c58b",
 Evaluatable->True,
 CellTags->{"Proof|ID:1849777423-1", "3"},
 CellID->1702528487],
Cell[721006, 20817, 1762, 45, 29, "ProofInfo",ExpressionUUID->"f5c12280-d2b6-42bc-8fbb-e7baab4c0376",
 Evaluatable->True,
 CellTags->{"Proof|ID:1569569101-1", "3"},
 CellID->1571614296],
Cell[722771, 20864, 1781, 46, 29, "ProofInfo",ExpressionUUID->"f177dd05-a638-4859-bf62-074cc8860f82",
 Evaluatable->True,
 CellTags->{"Proof|ID:172277084-1", "3"},
 CellID->665117787]
}, Closed]],
Cell[CellGroupData[{
Cell[724589, 20915, 195, 2, 31, "Subsection",ExpressionUUID->"4333526b-cfc3-4819-9486-4332bd86473c",
 CellID->19992184],
Cell[724787, 20919, 102, 1, 12, "OpenEnvironment",ExpressionUUID->"0e335947-1101-49d4-8136-3b5403b8318a",
 CellID->318971740],
Cell[CellGroupData[{
Cell[724914, 20924, 452, 9, 38, "EnvironmentHeader",ExpressionUUID->"7cf961ec-f8c5-4433-9b54-1fa29b06c3e0",
 CellID->28949169],
Cell[725369, 20935, 1443, 39, 54, "FormalTextInputFormula",ExpressionUUID->"a0cf4926-7497-4691-ac9b-596d4394ea91",
 CellTags->"Cantor's paradox",
 CellID->628580588],
Cell[726815, 20976, 1469, 44, 55, "FormalTextInputFormula",ExpressionUUID->"c1f91a4d-9e0e-4fc8-b926-adf6ecfaa815",
 CellTags->"Cantor's paradox 2",
 CellID->289824169],
Cell[728287, 21022, 120, 1, 24, "EndEnvironmentMarker",ExpressionUUID->"8c847f23-e35a-47ff-85a3-50f610dea31d",
 CellID->373064460]
}, Open  ]],
Cell[728422, 21026, 103, 1, 16, "CloseEnvironment",ExpressionUUID->"ce4d5482-79dc-43ac-ab6b-69d6eb56b85e",
 CellID->626869507],
Cell[728528, 21029, 1808, 47, 29, "ProofInfo",ExpressionUUID->"dfe4d76c-665e-46e9-a153-3fd07c91d051",
 Evaluatable->True,
 CellTags->{"Proof|ID:628580588-1", "3"},
 CellID->202346487],
Cell[730339, 21078, 1812, 47, 29, "ProofInfo",ExpressionUUID->"64679309-62a4-49fc-8eec-bdb7212bfd48",
 Evaluatable->True,
 CellTags->{"Proof|ID:289824169-1", "4"},
 CellID->383070079],
Cell[732154, 21127, 502, 13, 25, "Text",ExpressionUUID->"45f9ec30-3522-4e1a-abe9-001e7f7cf017",
 CellID->1219761075],
Cell[CellGroupData[{
Cell[732681, 21144, 625, 14, 44, "TheoremaCode",ExpressionUUID->"8c1e3ac1-5f29-4015-995f-0d13a830c93e",
 CellID->1767667081],
Cell[733309, 21160, 2658, 76, 35, "Print",ExpressionUUID->"d47b3a55-0bb4-4693-b519-7da41c2bf2a7",
 CellID->1500264271]
}, Open  ]]
}, Closed]]
}, Closed]]
}, Open  ]]
}
]
*)

